unit AlmostFloat;

interface

{$I VCL_VERSIONS.INC}

uses SysUtils, Math;

type EAlmostFloatBad = class(Exception);

function AlmostFloatFromStr(s: String; var Mantissa: Extended; var Exponent: Integer): Extended;

implementation

function RemoveSpaces(const s: String): String;
var i: Integer;
begin
    Result := s;
    for i := Length(s) downto 1 do
        if s[i] <= ' ' then Delete(Result, i, 1);
end;

function AlmostFloatFromStr(s: String; var Mantissa: Extended; var Exponent: Integer): Extended;
var p, q, i: Integer;
    Negative: Boolean;
    m: String;
begin
    s := TrimLeft(s);
    if s = '' then
        raise EAlmostFloatBad.Create('     ');
    Negative := s[1] = '-';
    if s[1] in ['-', '+'] then Delete(s, 1, 1);
    s := TrimLeft(s);
    p := pos('e', s);
    if p <= 0 then p := pos('E', s);
    if p <= 0 then p := pos('', s); //  
    if p <= 0 then p := pos('', s); //  
    if p <= 0 then p := pos('d', s);
    if p <= 0 then p := pos('D', s);
    if p <= 0 then p := pos('', s); //  
    if p <= 0 then p := pos('', s); //  

    if p <= 0 then
    begin
        p := pos('*', s);
        q := pos('x', s); if (q > 0) and ((q < p) or (p <= 0)) then p := q;
        q := pos('X', s); if (q > 0) and ((q < p) or (p <= 0)) then p := q;
    end;

    if p <= 0 then
    begin
        for i := 1 to Length(s) do
            if (s[i] <= ' ') or (s[i] in ['+', '-']) then
            begin
                p := i;
                break;
            end;
    end;

    if p > 0 then
    begin
        m := Copy(s, 1, p - 1);
        Delete(s, 1, p - 1);
        s := Trim(s);
    end
    else
    begin
        m := s;
        s := '';
    end;
    p := pos(',', m);
    if p <= 0 then p := pos('.', m);
    if p > 0 then m[p] := DecimalSeparator;
    Mantissa := StrToFloat(m);
    if Negative then Mantissa := -Mantissa;

    Exponent := 0;
    if s <> '' then
    begin
        CASE s[1] OF
            'e', 'E', '', '',
                'd', 'D', '', '':
                begin
                    Delete(s, 1, 1);
                    s := Trim(s);
                    Exponent := StrToInt(RemoveSpaces(s));
                end;
            '*', 'x', 'X', '', '':
                begin
                    Delete(s, 1, 1);
                    s := Trim(s);
                    if Copy(s, 1, 2) <> '10' then
                        raise EAlmostFloatBad.Create('    10: ' + s);
                    Delete(s, 1, 2);
                    while (s <> '') and not (s[1] in ['-', '+', '0'..'9']) do
                        Delete(s, 1, 1);
                    s := Trim(s);
                    Exponent := StrToInt(RemoveSpaces(s));
                end;
            '1':
                begin
                    if (Copy(s, 1, 2) = '10') and (Length(s) > 3) then
                    begin
                        Delete(s, 1, 2);
                        while (s <> '') and not (s[1] in ['-', '+', '0'..'9']) do
                            Delete(s, 1, 1);
                        s := Trim(s);
                        if s = '' then Exponent := 10
                        else Exponent := StrToInt(RemoveSpaces(s));
                    end
                    else
                    begin
                        Exponent := StrToInt(RemoveSpaces(s));
                    end;
                end;
        else
            while (s <> '') and not (s[1] in ['-', '+', '0'..'9']) do
                Delete(s, 1, 1);
            Exponent := StrToInt(RemoveSpaces(s));
        END;
    end;

    Result := Mantissa * IntPower(10, Exponent);
end;

end.
