//** TColorListView.
//**   , 2001
//**
//**  ,     VCL Delphi.
//**  Vladimir Kladov, 2000-2010
//**
//**      .
//**     /     
//**  2          
//**   GNU,  Free Software Foundation.
//**
//**       ,    
//** ,       ,   
//**         
//**   .      
//**     GNU.
//**
//**          
//**   GNU.     ,     Free
//** Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
//** MA 02111-1307, USA.
//******************************************************************************

{
  TColorListView (C) by Alexander Shackhailo, 2001

  List view extension, which allows to change color of header
  buttons (in report style), and has events for dragging header
  divider and double click header divider. )

  Updated:     , 2001-2010
}

unit ColorListView;

interface

{$I VCL_VERSIONS.INC}

////{$DEFINE STD_LISTVIEW_HDR_DRAW}
{$IFNDEF STD_LISTVIEW_HDR_DRAW}
{$DEFINE LVHDR_DRAW_VIA_NORMALLABEL}
{$ENDIF}

//{$DEFINE DRAW_HDR_BTN}

uses
    Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
    ComCtrls, CommCtrl, StdCtrls, imglist, math
    {$IFDEF LVHDR_DRAW_VIA_NORMALLABEL}, NormalBands{$ENDIF};

const
    WM_USER_SAVECOLUMNWIDTHS = WM_USER + 77;

type

    TOnVScroll = procedure(Sender: TObject; Show: boolean) of object;
    TOnHScroll = procedure(Sender: TObject; Show: boolean) of object;
    TNotifyHeaderDividerDblClk = procedure(Sender: TObject; columnIdx: Integer) of object;
    TNotifyHeaderPaint = procedure(Sender: TObject; Canvas: TCanvas; R: TRect; ColIdx: Integer) of object;
    TNotifyHeaderColor = function(Sender: TObject; ColIdx: Integer): TColor of object;
    THdrWrapping = (hwWords, hwLetters, hwSyllabels, hwNone);

    TColorListView = class(TListView)
    private
        FHdrColor,
            fSelColor,
            fPswColor: TColor;
        fCliWidth: integer;
        fItmAlign: TAlignment;
        fVVisible: boolean;
        fOnVScroll: TOnVScroll;
        FOnHdrTrack: TNotifyEvent;
        FOnHdrDividerDblClk: TNotifyHeaderDividerDblClk;
        FAutoExpandColumn: Integer;
        FAutoExpanding: Boolean;
        fSelColorText: TColor;
        fBoldIfFocus: Boolean;
        FSelColorNoFocus: TColor;
        FSelColorTextNoFocus: TColor;
        FOnHdrPaint: TNotifyHeaderPaint;
        FIndentWidth: Integer;
        FChangingOrder: Boolean;
        FSelectedColumn: Integer;
        FOnGetHdrHeight: TNotifyEvent;
        FHdrFontScale: Boolean;
        FHdrWrapping: THdrWrapping;
        FHdrFont: TFont;
        FHdrFontSame: Boolean;
        FReadOnly: Boolean;
        fOnHScroll: TOnHScroll;
        FOnHdrColor: TNotifyHeaderColor;
        procedure SetHdrColor(const Value: TColor);
        procedure SetAutoExpandColumn(const Value: Integer);
        procedure SetSelColorNoFocus(const Value: TColor);
        procedure SetSelColorTextNoFocus(const Value: TColor);
        function GetIndentWidth: Integer;
        procedure SetIndentWidth(const Value: Integer);
        function GetColumnOrder(ColIdx: Integer): Integer;
        procedure DoSetColumnOrder(ColIdx: Integer; const Value: Integer);
        procedure SetColumnOrder(ColIdx: Integer; Value: Integer);
        function GetColumnWidth(ColIdx: Integer): Integer;
        procedure SetSelectedColumn(const Value: Integer);
        procedure SetOnGetHdrHeight(const Value: TNotifyEvent);
        procedure SetHdrFontScale(const Value: Boolean);
        procedure SetHdrWrapping(const Value: THdrWrapping);
        procedure SetHdrFont(const Value: TFont);
        function GetHdrFontSame: Boolean;
        procedure SetHdrFontSame(const Value: Boolean);
        function get_read_only: Boolean;
        procedure set_read_only(const Value: Boolean);
        { Private declarations }
    protected
        { Protected declarations }
        fNeverDoubleBuffered: Boolean;
        FOldHandler: function(W, M, wParam, lParam: DWORD): Integer; stdcall;
        FLbtnPressed: Boolean;
        FTrackingHeader, FtrackingHdr: Boolean;
        //FColumnResizeDisabled: Boolean;
        FOrderArray: TList;
        FColumnWidths: TList;
        FHdrHeight: Integer;
        FUpdateCount: Integer;
        procedure WMAutoExpand(var M: TMessage); message WM_USER + 66;
        procedure WndProc(var M: TMessage); override;
        procedure DoAutoExpand(ExcludeColumn: Integer); virtual;
        procedure CustDrawItem(
            Sender: TCustomListView;
            Item: TListItem;
            State: TCustomDrawState;
            var DefaultDraw: Boolean); virtual;
        procedure CustDraw(Rect: TRect;
            Item: TListItem;
            Capt: string;
            SubItemIdx: Integer;
            Stat: TCustomDrawState;
            Alig: TAlignment);
        procedure GetColors(var ForeColor, BackColor: TColor; var FontStyle: TFontStyles;
            Item: TListItem; SubItemIdx: integer; Stat: TCustomDrawState);
        function GetSubItemRect(ItemIdx, ColIdx: Integer): TRect; virtual;
        function HdrMarginLeft(ColIdx: Integer): Integer; virtual;
        function HdrMarginRight(ColIdx: Integer): Integer; virtual;
        function FilterButtonClick(IgnoreCursorPos: Boolean; ItemIdx: Integer = -1): Boolean; virtual;
        function OnFilterButton: Boolean; virtual;
        procedure HdrHitTest(var HTI: THDHitTestInfo); virtual;
        function ProcessWMSetCursor(Wnd: HWnd; lParam: Integer): Boolean; virtual;
        function ItemIndent(Item: TListItem): Integer; virtual;
        function RightHdrClick(Msg: DWORD; HdrWnd: HWnd): Boolean;
        function HdrRightClick(ColIdx: Integer; Down: Boolean): Boolean; virtual;
        function EnabledAutoExpand: Boolean; virtual;
        procedure SaveColumnWidths;
        function GetColumnVisible(ColIdx: Integer): Boolean; virtual;
        procedure PaintHdr(W: Hwnd; DC: HDC); virtual;
    public
        { Public declarations }
        constructor Create(AOwner: TComponent); override;
        destructor Destroy; override;
        property CheckBoxes;
        //property ColumnResizeDisabled: Boolean read FColumnResizeDisabled;
        procedure InvalidateHeader;
        property ColumnOrder[ColIdx: Integer]: Integer read GetColumnOrder write SetColumnOrder;
        property ColumnWidth[ColIdx: Integer]: Integer read GetColumnWidth;
        function GetHeaderCtlWindow: HWnd;
        {* Handle   Header  . }
        property SelectedColumn: Integer read FSelectedColumn write SetSelectedColumn;
        {* Column selected (selected cell index in current line). }
    published
        { Published declarations }
        property HdrColor: TColor read FHdrColor write SetHdrColor;
        property SelColor: Tcolor read fSelColor write fSelColor;
        property SelColorText: TColor read fSelColorText write fSelColorText;
        property PswColor: TColor read fPswColor write fPswColor;
        property SelColorNoFocus: TColor read FSelColorNoFocus write SetSelColorNoFocus;
        property SelColorTextNoFocus: TColor read FSelColorTextNoFocus write SetSelColorTextNoFocus;
        property ItmAlign: TAlignment read fItmAlign write fItmAlign;
        property OnHdrTrack: TNotifyEvent read FOnHdrTrack write FOnHdrTrack;
        property OnHdrDividerDblClk: TNotifyHeaderDividerDblClk read FOnHdrDividerDblClk write FOnHdrDividerDblClk;
        property OnHdrPaint: TNotifyHeaderPaint read FOnHdrPaint write FOnHdrPaint;
        property OnHdrColor: TNotifyHeaderColor read FOnHdrColor write FOnHdrColor;
        property OnVScroll: TOnVScroll read fOnVScroll write fOnVScroll;
        property OnHScroll: TOnHScroll read fOnHScroll write fOnHScroll;
        property AutoExpandColumn: Integer read FAutoExpandColumn write SetAutoExpandColumn;
        property BoldIfFocused: Boolean read fBoldIfFocus write fBoldIfFocus;
        property IndentWidth: Integer read GetIndentWidth write SetIndentWidth;
    protected
        HdrImgList: THandle;
        FDisableGetDispInfo: Integer;
        FMultiHeader: String;
    public
        property LBtnPressed: Boolean read FLBtnPressed;
    published
        property OnGetHdrHeight: TNotifyEvent read FOnGetHdrHeight write SetOnGetHdrHeight;
        property HdrFontScale: Boolean read FHdrFontScale write SetHdrFontScale;
        property HdrWrapping: THdrWrapping read FHdrWrapping write SetHdrWrapping;
        property HdrFont: TFont read FHdrFont write SetHdrFont;
        property HdrFontSame: Boolean read GetHdrFontSame write SetHdrFontSame
            default TRUE;
        property ReadOnly: Boolean read get_read_only write set_read_only;
        procedure HdrLabelDestroy(Sender: TObject);
    end;

procedure BtnDraw(C: TCanvas; R: TRect; Pressed: Boolean; const Text: String;
    ImgIdx: Integer; ImgL: TCustomImageList; LV: TColorListView; ColIdx: Integer);

//procedure Register;

implementation

uses Extctrls, ShellAPi, utPhotoExc;

{procedure Register;
begin
  RegisterComponents('Additional', [TColorListView]);
end;}

{ TColorListView }

constructor TColorListView.Create(AOwner: TComponent);
begin
    inherited;
    FHdrFont := TFont.Create;
    FHdrFontSame := TRUE;
    FHdrColor := clBtnFace;
    fSelColor := clHighLight;
    fSelColorText := clHighLightText;
    fSelColorNoFocus := clBtnFace;
    fSelColorTextNofocus := $FFFFFF xor ColorToRGB(clBtnFace);
    fPswColor := clWindow;
    fItmAlign := taLeftJustify;
    fBoldIfFocus := TRUE;
    FAutoExpandColumn := -2;
    FHdrFontScale := TRUE;
    FHdrWrapping := hwSyllabels;
    OnCustomDrawItem := CustDrawItem;
end;

{$IFDEF LVHDR_DRAW_VIA_NORMALLABEL}
type TNormalLabel2 = class(TNormalLabel)
    end;
var HdrLabel: TNormalLabel2;
    {$ENDIF}

procedure BtnDraw(C: TCanvas; R: TRect; Pressed: Boolean; const Text: String;
    ImgIdx: Integer; ImgL: TCustomImageList; LV: TColorListView; ColIdx: Integer);
{$IFDEF LVHDR_DRAW_VIA_NORMALLABEL}
{$ELSE}
var x0, L: Integer;
    Sz: TSize;
    {$ENDIF}
var C1, C2: TColor;
    y0: Integer;
    Rgn: HRgn;
    hasRgn: Integer;
begin
    if (LV.FMultiHeader <> '') then
    begin
        C.Pen.Color := ColorToRGB(clWindowText);
        C.MoveTo(R.Right, R.Top);

        C1 := ColorToRGB(clBtnHighlight);
        C2 := ColorToRGB(clBtnFace);
        if Pressed then
        begin
            C2 := ColorToRGB(clBtnHighlight);
            C1 := ColorToRGB(clBtnShadow);

            C.LineTo(R.Left, R.Top);
            C.LineTo(R.Left, R.Bottom);
            Inc(R.Left);
            Inc(R.Top);

        end
        else
        begin

            Dec(R.Right);
            Dec(R.Bottom);
            C.MoveTo(R.Right, R.Top);
            C.LineTo(R.Right, R.Bottom);
            C.LineTo(R.Left, R.Bottom);
            Dec(R.Right);
            Dec(R.Bottom);

        end;
        C.Pen.Color := C1;
        C.MoveTo(R.Right, R.Top);
        C.LineTo(R.Left, R.Top);
        C.LineTo(R.Left, R.Bottom);
        C.Pen.Color := C2;
        C.LineTo(R.Right, R.Bottom);
        C.LineTo(R.Right, R.Top);
        C.Pen.Color := clBtnShadow;
    end;

    {$IFDEF LVHDR_DRAW_VIA_NORMALLABEL}

    if HdrLabel = nil then
    begin
        HdrLabel := TNormalLabel2(TNormalLabel.Create(LV));
        HdrLabel.Visible := FALSE;
        HdrLabel.Color := LV.HdrColor;
        HdrLabel.Transparent := (LV as TColorListView).FMultiHeader = '';
        //HdrLabel.WrapLetters := TRUE;
        HdrLabel.InFontScale := TRUE;
        HdrLabel.InFontMinSize := 7;
        //HdrLabel.InScale := TRUE;
        HdrLabel.AutoSize := FALSE;
        HdrLabel.Layout := tlCenter;
        HdrLabel.MarginTop := -3;
        HdrLabel.MarginBottom := -3;
        HdrLabel.OnDestroy := LV.HdrLabelDestroy;
    end;

    HdrLabel.Enabled := LV.Enabled;
    HdrLabel.MarginLeft := LV.HdrMarginLeft(ColIdx);
    HdrLabel.MarginRight := LV.HdrMarginLeft(ColIdx);
    if LV.HdrFontSame then
        HdrLabel.Font.Assign(LV.Font)
    else
        HdrLabel.Font.Assign(LV.HdrFont);
    HdrLabel.Caption := Text;
    HdrLabel.InFontScale := LV.HdrFontScale;
    HdrLabel.WordWrap := FALSE;
    HdrLabel.WrapSyllabel := FALSE;
    HdrLabel.WrapLetters := FALSE;
    CASE LV.HdrWrapping OF
        hwWords: HdrLabel.WordWrap := TRUE;
        hwSyllabels: HdrLabel.WrapSyllabel := TRUE;
        hwLetters: HdrLabel.WrapLetters := TRUE;
    END;
    if (ImgL <> nil) and (ImgIdx >= 0) then
    begin
        y0 := R.Top + (R.Bottom - R.Top - ImgL.Height) div 2;
        ImgL.Draw(C, R.Left + 2, y0, ImgIdx);
        inc(R.Left, 4 + ImgL.Width);
    end;
    Rgn := CreateRectRgn(0, 0, 0, 0);
    hasRgn := GetClipRgn(C.Handle, Rgn);
    IntersectClipRect(C.Handle, R.Left, R.Top, R.Right, R.Bottom);
    TRY
        HdrLabel.Color := C.Brush.Color;
        HdrLabel.PaintToCanvas(C, R);
    FINALLY
        if hasRgn <> 0 then
            SelectClipRgn(C.Handle, Rgn)
        else
            SelectClipRgn(C.Handle, 0);
        DeleteObject(Rgn);
    END;

    {$ELSE  HDR_DRAW_VIA_NORMALLABEL}

    Inc(R.Left);
    Inc(R.Top);
    Sz := C.TextExtent(Text);
    L := LV.HdrMarginLeft(ColIdx) + R.Left;
    if ImgL <> nil then
        if ImgIdx >= 0 then
            L := L + 4 + ImgL.Width;

    Dec(R.Right, LV.HdrMarginRight(ColIdx));

    if L < R.Right then
    begin
        x0 := L + (R.Right - L - Sz.cx) div 2;
        if x0 < L then x0 := L;
        y0 := R.Top + (R.Bottom - R.Top - Sz.cy) div 2;
        if y0 < R.Top then y0 := R.Top;

        C.TextRect(R, x0, y0, Text);
        if ImgL <> nil then
            if ImgIdx >= 0 then
            begin
                y0 := R.Top + (R.Bottom - R.Top - ImgL.Height) div 2;
                ImgL.Draw(C, R.Left + 2, y0, ImgIdx);
            end;
    end;
    {$ENDIF HDR_DRAW_VIA_NORMALLABEL}

end;

procedure TColorListView.PaintHdr(W: HWnd; DC: HDC);
var R, R1, R2: TRect;
    Cnv: TCanvas;
    I: Integer;
    MP: TPoint;
    //Buf: array[ 0..511 ] of Char;
    Buf: String;
    //Item: THDItem;
    DrawPressed: Boolean;
    HColor: TColor;
begin
    if ForceCanvas <> 0 then
        DC := ForceCanvas;
    windows.GetClientRect(W, R);
    GetCursorPos(MP);
    windows.ScreenToClient(W, MP);
    Cnv := TCanvas.Create;
    try
        Cnv.Handle := DC;
        Cnv.Font.Assign(HdrFont);
        I := SendMessage(W, HDM_GETITEMCOUNT, 0, 0);
        for I := 0 to I - 1 do
        begin
            {Item.Mask := HDI_TEXT;
            Item.cchTextMax := Sizeof( Buf );
            Item.pszText := @ Buf[ 0 ];
            SendMessage( W, HDM_GETITEM, I, Integer( @Item ) );}
            Buf := Columns[I].Caption;
            SendMessage(W, HDM_GETITEMRECT, I, Integer(@R1));
            if R.Left < R1.Right then
                R.Left := R1.Right;
            if FHdrColor <> clNone then
                HColor := FHdrColor
            else
                HColor := clBtnFace;
            if Assigned(OnHdrColor) then
            begin
                HColor := OnHdrColor(Self, I);
                Cnv.Brush.Color := HColor;
                Cnv.Brush.Style := bsSolid;
                R2 := R1;
                InflateRect(R2, -2, -1);
                Cnv.FillRect(R2); //<<<<<<<<<<<<
            end else
            begin
                Cnv.Brush.Color := HColor;
                Cnv.Brush.Style := bsSolid;
                {$IFDEF DRAW_HDR_BTN}
                Cnv.FillRect(R1); //<<<<<<<<<<<<
                {$ENDIF DRAW_HDR_BTN}
            end;
            R2 := R1;
            //InflateRect( R2, -3, -3 );
            DrawPressed := not FTrackingHeader
                //and not FTrackingHdr
            and FLBtnPressed
                and PtInRect(R1, MP);
            if DrawPressed then
                if Windows.GetCursor <> LoadCursor(0, IDC_ARROW) then
                    //if  Screen.Cursor = crVSplit then
                    DrawPressed := FALSE;
            BtnDraw(Cnv, R1, DrawPressed,
                Buf, Columns[I].ImageIndex,
                SmallImages, Self, I);
            if FHdrColor <> clNone then
                Cnv.Brush.Color := FHdrColor
            else
                Cnv.Brush.Color := HColor;
            Cnv.Brush.Style := bsSolid;
            if Assigned(OnHdrPaint) then
                OnHdrPaint(Self, Cnv, R1, I);
        end;
        {$IFDEF DRAW_HDR_BTN}
        Cnv.FillRect(R);
        {$ENDIF DRAW_HDR_BTN}
    finally
        Cnv.Free;
    end;
end;

function NewHandler(W: HWnd; Msg, wParam, lParam: DWORD): Integer; stdcall;
var LV: TColorListView;
    PS: TPaintStruct;
    DC: HDC;
    HTI: PHdHitTestInfo;
    MsgRec: TMsg;
    HL: PHDLayout;
    HIp: PHDItem;
    i: Integer;
    HI: THDItem;
    {$IFDEF LOG_FLV}
    Method: DWORD;
    {$ENDIF}
begin
    LV := Pointer(GetProp(W, '_HDR_OWN_LV_'));
    Assert(LV <> nil, 'COLOR HEADER: !!! Could not find owner listview !!!');
    case Msg of
        WM_NCDESTROY:
            begin
                SetWindowLong(W, GWL_WNDPROC, Integer(@LV.FOldHandler));
                Result := LV.FOldHandler(W, Msg, wParam, lParam);
                LV.FOldHandler := nil;
                Exit;
            end;
        WM_PAINT:
            begin
                {$IFDEF LOG_FLV}
                LogFileOutput('c:\log_flv.txt', 'FastListView_WM_Paint');
                {$ENDIF}
                if LV.FUpdateCount > 0 then
                begin
                    Result := 0;
                    ValidateRect(W, nil);
                    {$IFDEF LOG_FLV}
                    LogFileOutput('c:\log_flv.txt', 'FastListView UpdateCount=' + IntToStr(LV.FUpdateCount) + ' - exit');
                    {$ENDIF}
                    Exit;
                end;
                if not (csDesigning in LV.ComponentState) then
                    for i := 0 to LV.Columns.Count - 1 do
                    begin
                        HI.Mask := HDI_TEXT;
                        HI.pszText := LPSTR_TEXTCALLBACK;
                        SendMessage(W, HDM_SETITEM, i, Integer(@HI));
                    end;
                if ForceCanvas <> 0 then
                begin
                    {$IFDEF LOG_FLV}
                    LogFileOutput('c:\log_flv.txt', 'FastListView - PaintHdr');
                    {$ENDIF}
                    inc(LV.FDisableGetDispInfo);
                    LV.PaintHdr(W, ForceCanvas);
                    dec(LV.FDisableGetDispInfo);
                    {$IFDEF LOG_FLV}
                    LogFileOutput('c:\log_flv.txt', 'FastListView - PaintHdr exit');
                    {$ENDIF}
                end else
                    if wParam <> 0 then
                    begin
                        {$IFDEF LOG_FLV}
                        LogFileOutput('c:\log_flv.txt', 'FastListView wParam <> 0 - PaintHdr');
                        {$ENDIF}
                        inc(LV.FDisableGetDispInfo);
                        LV.PaintHdr(W, wParam);
                        dec(LV.FDisableGetDispInfo);
                        {$IFDEF LOG_FLV}
                        LogFileOutput('c:\log_flv.txt', 'FastListView wParam <> 0 - PaintHdr exit');
                        {$ENDIF}
                    end else
                    begin
                        {$IFDEF LOG_FLV}
                        LogFileOutput('c:\log_flv.txt', 'FastListView paint OldHdr');
                        {$ENDIF}
                        if LV <> nil then
                        TRY
                            {$IFDEF LOG_FLV}
                            LogFileOutput('c:\log_flv.txt', 'FastListView paint OldHdr 1');
                            {$ENDIF}
                            inc(LV.FDisableGetDispInfo);
                            TRY
                                {$IFDEF LOG_FLV}
                                LogFileOutput('c:\log_flv.txt', 'FastListView paint OldHdr 2');
                                {$ENDIF}
                                if LV.FMultiHeader = '' then
                                begin
                                    {$IFDEF LOG_FLV}
                                    LogFileOutput('c:\log_flv.txt', 'FastListView paint OldHdr 3');
                                    {$ENDIF}
                                    if Assigned(LV.FOldHandler) then
                                    begin
                                        {$IFDEF LOG_FLV}
                                        asm
                                MOV EAX, [LV]
                                MOV EAX, [EAX].TMethod.Code
                                MOV [Method], EAX
                            //Method := TMethod(@LV.FOldHandler);
                                        end;
                                        LogFileOutput('c:\log_flv.txt', 'FastListView paint OldHdr 4' +
                                            ' LV=' + IntToHex(Integer(LV), 8) +
                                            ' LV.FOldHandler=' + IntToHex(
                                            Integer(Method), 8));
                                        {$ENDIF}
                                        LV.FOldHandler(W, Msg, wParam, lParam);
                                    end;
                                end;
                                {$IFDEF LOG_FLV}
                                LogFileOutput('c:\log_flv.txt', 'FastListView paint OldHdr 5');
                                {$ENDIF}
                                {DC :=}BeginPaint(W, PS);
                                {$IFDEF LOG_FLV}
                                LogFileOutput('c:\log_flv.txt', 'FastListView paint OldHdr 6');
                                {$ENDIF}
                                DC := GetWindowDC(W); //<<<<<<<<<<<<<<<<<
                                {$IFDEF LOG_FLV}
                                LogFileOutput('c:\log_flv.txt', 'FastListView paint OldHdr 7');
                                {$ENDIF}
                                TRY
                                    {$IFDEF LOG_FLV}
                                    LogFileOutput('c:\log_flv.txt', 'FastListView paint OldHdr 8');
                                    {$ENDIF}
                                    LV.PaintHdr(W, DC);
                                    {$IFDEF LOG_FLV}
                                    LogFileOutput('c:\log_flv.txt', 'FastListView paint OldHdr 9');
                                    {$ENDIF}
                                FINALLY
                                    ReleaseDC(W, DC); //<<<<<<<<<<<<<<<<<<<<
                                    EndPaint(W, PS);
                                END;
                            FINALLY
                                dec(LV.FDisableGetDispInfo);
                            END;
                        EXCEPT on e: Exception do
                            begin
                                {$IFDEF LOG_FLV}
                                LogFileOutput('c:\log_flv.txt', ' ' + e.ClassName +
                                    ' ' + e.Message);
                                {$ENDIF}
                            end;
                        END;
                        {$IFDEF LOG_FLV}
                        LogFileOutput('c:\log_flv.txt', 'FastListView paint OldHdr exit');
                        {$ENDIF}
                    end;
                Result := 0;
                {$IFDEF LOG_FLV}
                LogFileOutput('c:\log_flv.txt', 'FastListView paint EXIT');
                {$ENDIF}
                Exit;
            end;
        WM_ERASEBKGND:
            begin
                if LV.FUpdateCount > 0 then
                begin
                    Result := 1;
                    Exit;
                end;
            end;
        WM_NCPAINT:
            begin
                if LV.FUpdateCount > 0 then
                begin
                    Result := 0;
                    Exit;
                end;
            end;
        WM_LBUTTONDOWN:
            begin
                LV.FLbtnPressed := TRUE;
                if LV.FilterButtonClick(FALSE) then
                begin
                    while PeekMessage(MsgRec, LV.Handle, WM_USER + 66, WM_USER + 66, pm_remove) do ;
                    Result := 0;
                    Exit;
                end;
            end;
        WM_LBUTTONUP:
            begin
                LV.FLbtnPressed := FALSE;
                LV.InvalidateHeader;
                if LV.OnFilterButton then
                begin
                    Result := 0;
                    Exit;
                end;
            end;
        WM_RBUTTONDOWN, WM_RBUTTONUP:
            begin
                if LV.RightHdrClick(Msg, W) then
                begin
                    Result := 0;
                    Exit;
                end;
            end;
        HDM_HITTEST:
            begin
                Result := LV.FOldHandler(W, Msg, wParam, lParam);
                HTI := Pointer(lParam);
                if not (csDesigning in LV.ComponentState) and
                    (LongBool(HTI.Flags and (HHT_ONDIVIDER or HHT_ONDIVOPEN))) then
                    LV.HdrHitTest(HTI^);
                Exit;
            end;
        HDM_LAYOUT:
            begin
                Result := LV.FOldHandler(W, Msg, wParam, lParam);
                if Assigned(LV.OnGetHdrHeight) then
                    LV.OnGetHdrHeight(LV);
                if (LV.FHdrHeight <> 0) and
                    not (csDesigning in LV.ComponentState) then
                begin
                    HL := Pointer(lParam);
                    i := max(HL.WindowPos.cy, LV.FHdrHeight);
                    HL.Rect.Top := HL.Rect.Top + i - HL.WindowPos.cy;
                    HL.WindowPos.cy := i;
                end;
                Exit;
            end;
        HDM_SETITEM:
            begin
                HIp := Pointer(lParam);
                if HIp.Mask and HDI_TEXT <> 0 then
                    HIp.pszText := LPSTR_TEXTCALLBACK;
            end;
        WM_SETCURSOR:
            begin
                if LV.ProcessWMSetCursor(W, lParam) then
                begin
                    Result := 1;
                    Exit;
                end;
            end;
    end;
    Result := LV.FOldHandler(W, Msg, wParam, lParam);
end;

procedure TColorListView.DoAutoExpand(ExcludeColumn: Integer);
var I: Integer;
    W: Integer;
    ExpandCol: Integer;
    WasFocused: Boolean;
    CW: Integer;
begin
    if fTrackingHeader then exit;
    if FLBtnPressed then exit;
    ExpandCol := FAutoExpandColumn;
    if ExpandCol = ExcludeColumn then
        ExpandCol := Columns.Count - 1;
    if ExpandCol < 0 then Exit;
    if ExpandCol >= Columns.Count then Exit;
    if FAutoExpanding then Exit;
    if not EnabledAutoExpand then Exit;
    if FChangingOrder then Exit;

    FAutoExpanding := TRUE;
    TRY

        if assigned(fOnVScroll) then begin
            if fCliWidth = 0 then fCliWidth := ClientWidth;
            if fCliWidth > ClientWidth then begin
                if not fVVisible then begin
                    fVVisible := True;
                    fOnVScroll(self, True);
                end;
            end else
                if fCliWidth < ClientWidth then begin
                    if fVVisible then begin
                        fVVisible := False;
                        fOnVScroll(self, False);
                    end;
                end;
            fCliWidth := ClientWidth;
        end;
        W := 0;
        for I := 0 to Columns.Count - 1 do
        begin
            if I <> ExpandCol then
                if GetColumnVisible(I) then
                    W := W + Columns[I].Width;
        end;
        if Parent = nil then
            CW := Width - 4
        else
            CW := ClientWidth;
        if CW - W > 10 then
            if (Columns.Count > ExpandCol) and
                (Columns[ExpandCol].Width <> CW - W) then
            begin
                WasFocused := HandleAllocated and (GetFocus = Handle);
                if ExcludeColumn <> -2 then
                    Visible := FALSE;
                TRY
                    Columns[ExpandCol].Width := CW - W;
                FINALLY
                    if not Visible then
                    begin
                        Visible := TRUE;
                        if WasFocused then
                            Windows.SetFocus(Handle);
                    end;
                END;
            end;
        SaveColumnWidths;

    FINALLY
        FAutoExpanding := FALSE;
    END;
end;

procedure TColorListView.CustDraw(Rect: TRect;
    Item: TListItem;
    Capt: string;
    SubItemIdx: Integer;
    Stat: TCustomDrawState;
    Alig: TAlignment);
var f: integer;
    R: TRect;
    T: string;
    ForeColor, BackColor: TColor;
    FontStyle: TFontStyles;
begin
    f := DT_NOPREFIX;
    R := Rect;
    if Alig = taLeftJustify then begin
        f := f or DT_LEFT;
    end else
        if Alig = taRightJustify then begin
            f := f or DT_RIGHT;
            R.Right := R.Right - 5;
        end else
            if Alig = taCenter then begin
                f := f or DT_CENTER;
            end;

    FontStyle := Font.Style;
    GetColors(ForeColor, BackColor, FontStyle, Item, SubItemIdx, Stat);
    Canvas.Font.Style := FontStyle;
    Canvas.Font.Color := ForeColor;
    Canvas.Brush.Color := BackColor;
    Canvas.Brush.Style := bsSolid;

    Canvas.FillRect(Rect);

    R.Left := R.Left + 4;
    T := Capt;
    if T <> '' then begin
        if T[Length(T)] = #0 then begin
            T := Copy(T, 1, Length(T) - 1);
        end;
    end;
    //Canvas.Handle := 0; //     VCL !!! ?????????
    DrawText(Canvas.Handle, PChar(T), Length(T), R,
        f or DT_END_ELLIPSIS or DT_SINGLELINE);
    Canvas.Brush.Style := bsSolid;
end;

procedure TColorListView.CustDrawItem(
    Sender: TCustomListView;
    Item: TListItem;
    State: TCustomDrawState;
    var DefaultDraw: Boolean);
var R, R1, R0: TRect;
    T: TRect;
    O, Frame: TRect;
    I: integer;
    L: integer;
    P: TCustomImageList;
    S: TCustomImageList;
    W: HWnd;
    D: integer;
    FrColor: TColor;
    Offset: Integer;
begin
    W := GetWindow(Handle, GW_CHILD);
    DefaultDraw := False;
    R := Item.DisplayRect(drBounds);
    R1 := R;
    T := R;
    D := R.Left;
    if ViewStyle = vsReport then
        SendMessage(W, HDM_GETITEMRECT, 0, Integer(@T));
    R.Left := T.Left;
    S := nil;
    if StateImages <> nil then begin
        S := StateImages;
    end;
    P := nil;
    if LargeImages <> nil then begin
        P := LargeImages;
    end else
        if SmallImages <> nil then begin
            P := SmallImages;
        end;
    Frame := R;
    Frame.Left := R1.Left;
    if (S <> nil) then begin
        R.Left := D + S.Width + 4;
    end;
    if (P <> nil) then begin
        R.Left := D + P.Width + 4;
    end;
    if (S = nil) and (P = nil) then
        R.Left := D + 2;
    L := D + 2;
    O := R;
    R.Right := D + T.Right;
    if S <> nil then
        R.Left := R.Left + 2 + S.Width;
    R0 := R;
    Offset := 0;
    if ItemIndent(Item) > 0 then
        Offset := ItemIndent(Item) * IndentWidth;
    R0.Left := R.Left + Offset;
    CustDraw(R0, Item, Item.Caption, -1, State, fItmAlign);
    Canvas.FillRect(Rect(Frame.Left, Frame.Top, Frame.Left + 2, Frame.Bottom));
    if S <> nil then
    begin
        Canvas.FillRect(Rect(L - 2 + Offset, R.Top, L + S.Width + 2 + Offset, R.Bottom));
        if Item.StateIndex > -1 then
        begin
            I := R.Top + (R.Bottom - R.Top - S.Height) div 2;
            S.Draw(Canvas, L + Offset, I, Item.StateIndex, True);
        end;
        L := L + 2 + S.Width;
    end;
    if P <> nil then
    begin
        if S = nil then
            Canvas.FillRect(Rect(L - 2 + Offset, R.Top, L + P.Width + 2 + Offset, R.Bottom))
        else
            Canvas.FillRect(Rect(L + Offset, R.Top, L + P.Width + 2 + Offset, R.Bottom));
        if Item.ImageIndex > -1 then
        begin
            I := R.Top + (R.Bottom - R.Top - P.Height) div 2;
            P.Draw(Canvas, L + Offset, I, Item.ImageIndex, True);
        end;
    end;
    if ViewStyle = vsReport then
    begin
        for i := 0 to Item.SubItems.Count - 1 do
        begin
            if i >= Columns.Count - 1 then break;
            SendMessage(W, HDM_GETITEMRECT, I + 1, Integer(@T));
            R.Left := D + T.Left;
            R.Right := D + T.Right;
            CustDraw(R, Item, Item.SubItems[i], i, State, Column[i + 1].Alignment);
        end;
        for i := Item.SubItems.Count to Columns.Count - 2 do
        begin
            SendMessage(W, HDM_GETITEMRECT, I + 1, Integer(@T));
            R.Left := D + T.Left;
            R.Right := D + T.Right;
            CustDraw(R, Item, '', i, State, Column[i + 1].Alignment);
        end;
    end;
    Inc(Frame.Left, Offset);
    if Item.Focused and Item.Selected then
    begin
        if SelectedColumn >= 0 then
        begin
            Frame := GetSubItemRect(Item.Index, SelectedColumn);
        end;
        if GridLines then
            Dec(Frame.Bottom);
        FrColor := $FFFFFF xor ColorToRGB(Canvas.Brush.Color);
        Frame3D(Canvas, Frame, FrColor, FrColor, 1);
    end;
end;

procedure TColorListView.SetAutoExpandColumn(const Value: Integer);
begin
    FAutoExpandColumn := Value;
    if FAutoExpandColumn >= Columns.Count then
        FAutoExpandColumn := Columns.Count - 1;
    DoAutoExpand(-1);
end;

procedure TColorListView.SetHdrColor(const Value: TColor);
var W: HWnd;
begin
    FHdrColor := Value;
    Invalidate;
    HandleNeeded;
    W := GetWindow(Handle, GW_CHILD);
    if W = 0 then Exit;
    if not Assigned(FOldHandler) and not (csDesigning in ComponentState) then
    begin
        SetProp(W, '_HDR_OWN_LV_', DWORD(Self));
        Integer(@FOldHandler) := GetWindowLong(W, GWL_WNDPROC);
        SetWindowLong(W, GWL_WNDPROC, Integer(@NewHandler));
    end;
end;

procedure TColorListView.WMAutoExpand(var M: TMessage);
begin
    DoAutoExpand(M.wParam);
    inherited;
end;

procedure TColorListView.WndProc(var M: TMessage);
type
    TNMHeader = packed record
        hdr: TNMHdr;
        iItem: Integer;
        iButton: Integer;
        pitem: PHDItem;
    end;
    PNMHeader = ^TNMHeader;
var Header: PNMHeader;
    procedure PostAutoExpand(ExcludeCol: Integer = -1);
    var Msg: TMsg;
    begin
        if not FAutoExpanding then
            if not PeekMessage(Msg, Handle, WM_USER + 66, WM_USER + 66, pm_noremove) then
                PostMessage(Handle, WM_USER + 66, ExcludeCol, 0);
    end;
var P: TPoint;
    HHT: THdHitTestInfo;
    HdrWnd: HWnd;
    I: Integer;
    DIA: PNMHDispInfoA;
    DIW: PNMHDispInfoW;
    //HI: THDItem;
    PP: PPoint;
    R: TRect;
begin
    if M.Msg = WM_NOTIFY then
    begin
        case TWMNotify(M).NMHdr.code of
            HDN_GETDISPINFOA:
                begin
                    DIA := PNMHDispInfoA(M.LParam);
                    DIA.mask := HDI_DI_SETITEM;
                    M.Result := 0;
                    Exit;
                end;
            HDN_GETDISPINFOW:
                begin
                    {$IFDEF DRAW_HDR_BTN}
                    DIW := PNMHDispInfoW(M.LParam);
                    DIW.mask := HDI_DI_SETITEM;
                    M.Result := 0;
                    {$ELSE}
                    DIW := PNMHDispInfoW(M.LParam);
                    DIW.iImage := -1; //  
                    //DIW.mask := HDI_TEXT;
                    {if  DIW.mask and HDI_IMAGE = HDI_IMAGE then
                    begin
                        HI.Mask := HDI_TEXT;
                        HI.pszText := LPSTR_TEXTCALLBACK;
                        SendMessage( GetHeaderCtlWindow, HDM_SETITEM, DIW.iItem,
                                     Integer( @ HI ) );
                    end;}
                    if DIW.mask and HDI_TEXT <> 0 then
                        if DIW.pszText <> nil then
                        begin
                            if (DIW.iItem >= 0) and (DIW.iItem < Columns.Count)
                                and (FDisableGetDispInfo = 0) then
                            begin
                                I := Length(Columns[DIW.iItem].Caption);
                                if I > DIW.cchTextMax then
                                    I := DIW.cchTextMax;
                                Move(PWideChar(WideString(Columns[DIW.iItem].Caption))^,
                                    DIW.pszText^, 2 * I);
                                DIW.pszText[I] := #0;
                                //DIW.cchTextMax := I;
                            end
                            else
                            begin
                                if DIW.cchTextMax > 0 then
                                    DIW.pszText^ := #0;
                            end;
                        end;
                    M.Result := 0;
                    {$ENDIF}
                    Exit;
                end;
            HDN_ENDTRACK, HDN_ENDTRACKW:
                begin
                    FTrackingHeader := FALSE;
                    FTrackingHdr := FALSE;
                    if Assigned(FOnHdrTrack) then
                        FOnHdrTrack(Self);
                    Header := Pointer(M.LParam);
                    PostAutoExpand(Header.iItem);
                    SaveColumnWidths;
                end;
            HDN_TRACK, HDN_TRACKW, HDN_ITEMCHANGING, HDN_ITEMCHANGINGW:
                begin
                    if (TWMNotify(M).NMHdr.code = HDN_TRACK) or
                        (TWMNotify(M).NMHdr.code = HDN_TRACKW) then
                        FTrackingHeader := TRUE;
                    FTrackingHdr := TRUE;
                    if Assigned(FOnHdrTrack) then
                        FOnHdrTrack(Self);
                    //PostAutoExpand;
                    SaveColumnWidths;
                end;
            HDN_DIVIDERDBLCLICK, HDN_DIVIDERDBLCLICKW:
                begin
                    Header := Pointer(M.LParam);
                    if Assigned(FOnHdrDividerDblClk) then
                        FOnHdrDividerDblClk(Self, Header.iItem);
                    PostAutoExpand(Header.iItem);
                    if Assigned(FOnHdrDividerDblClk) then
                        Exit;
                end;
            HDN_ENDDRAG:
                begin
                    Header := Pointer(M.lParam);
                    {if Assigned( FOnHdrEndDrag ) then
                      if FOnHdrEndDrag( Self, Header.iItem ) then
                        Exit;}
                    M.Result := 1;
                    GetCursorPos(P);
                    HdrWnd := GetWindow(Handle, GW_CHILD);
                    Windows.ScreenToClient(HdrWnd, P);
                    HHT.Point := P;
                    SendMessage(HdrWnd, HDM_HITTEST, 0, Integer(@HHT));
                    if (HHT.Item >= 0) and (HHT.Item < Columns.Count) and
                        (HHT.Item <> Header.iItem) then
                        ColumnOrder[Header.iItem] := ColumnOrder[HHT.Item];
                    SaveColumnWidths;
                    Exit;
                end;
        end;
    end
    else
        CASE M.Msg OF
            WM_SIZE: begin
                    PostAutoExpand(-2);
                    SaveColumnWidths;
                end;
            WM_USER_SAVECOLUMNWIDTHS:
                begin
                    if FColumnWidths = nil then
                        FColumnWidths := TList.Create;
                    FColumnWidths.Clear;
                    for I := 0 to Columns.Count - 1 do
                        FColumnWidths.Add(Pointer(Columns[I].Width));
                end;
            LVM_GETORIGIN:
                begin
                    inherited;
                    if ShowColumnHeaders then
                    begin
                        GetWindowRect(GetHeaderCtlWindow, R);
                        PP := Pointer(M.LParam);
                        PP.y := R.Bottom;
                    end;
                end;
            WM_PAINT:
                begin
                    if ForceCanvas <> 0 then
                    begin
                        HdrWnd := GetWindow(Handle, GW_CHILD);
                        PaintHdr(HdrWnd, ForceCanvas);
                    end;
                end;
            WM_HSCROLL:
                begin
                    if Assigned(OnHScroll) then
                        OnHScroll(Self, TRUE);
                end;
        END;
    inherited;
end;

procedure TColorListView.GetColors(var ForeColor, BackColor: TColor; var FontStyle: TFontStyles;
    Item: TListItem; SubItemIdx: integer; Stat: TCustomDrawState);
begin
    FontStyle := Font.Style;
    if Item.Selected then
    begin
        if Focused then
        begin
            BackColor := SelColor;
            ForeColor := SelColorText;
        end
        else
        begin
            BackColor := SelColorNoFocus;
            ForeColor := SelColorTextNoFocus;
        end;
    end else
    begin
        BackColor := fPswColor;
        ForeColor := clWindowText;
    end;
    if Item.Focused and BoldIfFocused then
        FontStyle := FontStyle + [fsBold];
end;

{procedure TColorListView.SetDoubleBuffered(const Value: Boolean);
begin
  //fNeverDoubleBuffered := Value;
  //   DoubleBuffered  TRUE =  FALSE
end;}

procedure TColorListView.SetSelColorNoFocus(const Value: TColor);
begin
    FSelColorNoFocus := Value;
end;

procedure TColorListView.SetSelColorTextNoFocus(const Value: TColor);
begin
    FSelColorTextNoFocus := Value;
end;

function TColorListView.HdrMarginLeft(ColIdx: Integer): Integer;
begin
    Result := 0;
end;

function TColorListView.HdrMarginRight(ColIdx: Integer): Integer;
begin
    Result := 0;
end;

function TColorListView.FilterButtonClick(IgnoreCursorPos: Boolean; ItemIdx: Integer = -1): Boolean;
begin
    Result := FALSE;
end;

function TColorListView.OnFilterButton: Boolean;
begin
    Result := FALSE;
end;

procedure TColorListView.HdrHitTest(var HTI: THDHitTestInfo);
begin
    //    
end;

function TColorListView.ProcessWMSetCursor(Wnd: HWnd;
    lParam: Integer): Boolean;
begin
    Result := FALSE;
end;

function TColorListView.GetIndentWidth: Integer;
begin
    Result := FIndentWidth;
    if Result <= 0 then
    begin
        Result := 0;
        if StateImages <> nil then
            Result := Result + StateImages.Width;
        if Result = 0 then
            if SmallImages <> nil then
                Result := Result + SmallImages.Width;
        if Result = 0 then
            if LargeImages <> nil then
                Result := Result + LargeImages.Width;
        if Result = 0 then
            Result := 20;
    end;
end;

procedure TColorListView.SetIndentWidth(const Value: Integer);
begin
    FIndentWidth := Value;
end;

function TColorListView.ItemIndent(Item: TListItem): Integer;
begin
    Result := Item.Indent;
end;

function TColorListView.HdrRightClick(ColIdx: Integer;
    Down: Boolean): Boolean;
begin
    Result := FALSE;
end;

function TColorListView.RightHdrClick(Msg: DWORD; HdrWnd: HWnd): Boolean;
var P: TPoint;
    I: Integer;
    R: TRect;
begin
    GetCursorPos(P);
    Windows.ScreenToClient(HdrWnd, P);
    for I := 0 to Columns.Count - 1 do
    begin
        SendMessage(HdrWnd, HDM_GETITEMRECT, I, Integer(@R));
        if PtInRect(R, P) then
        begin
            Result := HdrRightClick(I, Msg = WM_RBUTTONDOWN);
            Exit;
        end;
    end;
    Result := FALSE;
end;

procedure TColorListView.InvalidateHeader;
var HdrWnd: HWnd;
    ClassNameBuf: array[0..49] of Char;
begin
    HdrWnd := childWindowFromPoint(Handle, Point(0, 0));

    if (HdrWnd <> 0) and (HdrWnd <> Handle) then
    begin

        GetClassName(HdrWnd, ClassNameBuf, 50);
        if ClassNameBuf = 'SysHeader32' then
        begin
            InvalidateRect(HdrWnd, nil, FALSE);
        end;

    end;

end;

function TColorListView.GetColumnOrder(ColIdx: Integer): Integer;
{var HDItem: THDItem;
    HdrWnd: HWnd;}
begin
    if (FOrderArray <> nil) and (FOrderArray.Count > ColIdx) then
        Result := Integer(FOrderArray[ColIdx])
    else
        Result := -1;
    {HdrWnd := GetWindow( Handle, GW_CHILD );
    HDItem.Mask := HDI_ORDER;
    if SendMessage( HdrWnd, HDM_GETITEM, ColIdx, Integer( @HDItem ) ) <> 0 then
      Result := HDItem.iOrder
    else
      Result := ColIdx;}
end;

procedure TColorListView.DoSetColumnOrder(ColIdx: Integer;
    const Value: Integer);
var HDItem: THDItem;
    HdrWnd: HWnd;
begin
    FChangingOrder := TRUE;
    TRY
        if Value >= Columns.Count then Exit;
        HdrWnd := GetWindow(Handle, GW_CHILD);
        HDItem.Mask := HDI_ORDER;
        HDItem.iOrder := Value;
        SendMessage(HdrWnd, HDM_SETITEM, ColIdx, Integer(@HDItem));
    FINALLY
        FChangingOrder := FALSE;
    END;
end;

procedure TColorListView.SetColumnOrder(ColIdx: Integer; Value: Integer);
var I: Integer;
    AllMinus1: Boolean;

    procedure CheckValueForColumn;
    var I: Integer;
    begin
        if Value > Integer(FOrderArray[ColIdx]) then
        begin
            for I := 0 to FOrderArray.Count - 1 do
            begin
                if (Integer(FOrderArray[I]) > Integer(FOrderArray[ColIdx])) and
                    (Integer(FOrderArray[I]) <= Value) then
                    FOrderArray[I] := Pointer(Integer(FOrderArray[I]) - 1);
            end;
        end
        else
        begin
            for I := 0 to FOrderArray.Count - 1 do
            begin
                if (Integer(FOrderArray[I]) >= Value) and
                    (Integer(FOrderArray[I]) < Integer(FOrderArray[ColIdx])) then
                    FOrderArray[I] := Pointer(Integer(FOrderArray[I]) + 1);
            end;
        end;
    end;

begin
    if Value >= Columns.Count then Exit;
    if FOrderArray = nil then
    begin
        FOrderArray := TList.Create;
        for I := 0 to Columns.Count - 1 do
            FOrderArray.Add(Pointer(I));
    end;
    TRY
        while FOrderArray.Count < Columns.Count do
            FOrderArray.Add(Pointer(-1));
        if Value = Integer(FOrderArray[ColIdx]) then Exit;
        if Value = -1 then
        begin
            Value := ColIdx;
            CheckValueForColumn;
            if ColIdx = Integer(FOrderArray[ColIdx]) then
                Value := -1;
        end
        else
        begin
            CheckValueForColumn;
        end;
        FOrderArray[ColIdx] := Pointer(Value);
        for I := 0 to FOrderArray.Count - 1 do
        begin
            if Integer(FOrderArray[I]) <> -1 then
                DoSetColumnOrder(I, Integer(FOrderArray[I]));
        end;
    FINALLY
        Invalidate;
        TRY
            AllMinus1 := TRUE;
            if FOrderArray <> nil then
                for I := 0 to FOrderArray.Count - 1 do
                    if Integer(FOrderArray[I]) <> -1 then
                    begin
                        AllMinus1 := FALSE;
                        break;
                    end;
            if AllMinus1 then
            begin
                //ShowMessage( 'Natural order on!' );
                FreeAndNil(FOrderArray);
            end;
        FINALLY
        END;
    END;
end;

function TColorListView.EnabledAutoExpand: Boolean;
begin
    Result := TRUE;
end;

destructor TColorListView.Destroy;
begin
    FOrderArray.Free;
    FColumnWidths.Free;
    inherited;
    if HdrImgList <> 0 then
    begin
        ImageList_Destroy(HdrImgList); HdrImgList := 0;
        FreeAndNil(HdrLabel);
    end;
end;

procedure TColorListView.SaveColumnWidths;
var Msg: TMsg;
begin
    if not PeekMessage(Msg, Handle, WM_USER_SAVECOLUMNWIDTHS, WM_USER_SAVECOLUMNWIDTHS,
        pm_noremove) then
        PostMessage(Handle, WM_USER_SAVECOLUMNWIDTHS, 0, 0);
end;

function TColorListView.GetColumnWidth(ColIdx: Integer): Integer;
begin
    if FColumnWidths <> nil then
        Result := Integer(FColumnWidths[ColIdx])
    else
        Result := Columns[ColIdx].Width;
end;

function TColorListView.GetColumnVisible(ColIdx: Integer): Boolean;
begin
    Result := TRUE;
end;

function TColorListView.GetHeaderCtlWindow: HWnd;
var ClsBuf: array[0..127] of Char;
begin
    Result := GetWindow(Handle, GW_CHILD);
    REPEAT
        if Result = 0 then Exit;
        GetClassName(Result, ClsBuf, Sizeof(ClsBuf));
        if ClsBuf = 'SysHeader32' then Exit;
        Result := GetWindow(Result, GW_HWNDNEXT);
    UNTIL FALSE;
end;

procedure TColorListView.SetSelectedColumn(const Value: Integer);
begin
    FSelectedColumn := Value;
    Invalidate;
end;

function TColorListView.GetSubItemRect(ItemIdx, ColIdx: Integer): TRect;
begin
    Result := Items[ItemIdx].DisplayRect(drBounds);
end;

procedure TColorListView.SetOnGetHdrHeight(const Value: TNotifyEvent);
begin
    FOnGetHdrHeight := Value;
end;

procedure TColorListView.SetHdrFontScale(const Value: Boolean);
begin
    FHdrFontScale := Value;
    Invalidate;
end;

procedure TColorListView.SetHdrWrapping(const Value: THdrWrapping);
begin
    FHdrWrapping := Value;
    Invalidate;
end;

procedure TColorListView.SetHdrFont(const Value: TFont);
begin
    FHdrFont := Value;
    Invalidate;
end;

function TColorListView.GetHdrFontSame: Boolean;
begin
    Result := FHdrFontSame
        and (HdrFont.Name = Font.Name)
        and (HdrFont.Size = Font.Size)
        and (HdrFont.Height = Font.Height)
        and (HdrFont.Style = Font.Style)
        and (HdrFont.Color = Font.Color)
        and (HdrFont.Charset = Font.Charset)
        and (HdrFont.Pitch = Font.Pitch);
end;

procedure TColorListView.SetHdrFontSame(const Value: Boolean);
begin
    if Value then
        HdrFont := Font;
    FHdrFontSame := Value;
end;

procedure TColorListView.HdrLabelDestroy(Sender: TObject);
begin
    HdrLabel := nil;
end;

function TColorListView.get_read_only: Boolean;
begin
    TRY
        Result := FReadOnly; // and inherited ReadOnly;
    EXCEPT on E: Exception do
        begin
            Result := FALSE;
            LogFileOutput('c:\readonly.txt',
                'Exception TColorListView.get_read_only: ' +
                E.Message);
        end;
    END;
end;

procedure TColorListView.set_read_only(const Value: Boolean);
const FT: array[Boolean] of String = ('FALSE', 'TRUE');
begin
    TRY
        FReadOnly := Value;
        //inherited ReadOnly := Value;
    EXCEPT on E: Exception do
        begin
            LogFileOutput('c:\readonly.txt',
                'Exception TColorListView.set_read_only: ' +
                E.Message);
            {MessageBox( 0, PChar( '    ReadOnly   ' +
            FT[Value] + '  ' +  Name + ': TColorListview, ' + E.Message ),
            nil, MB_OK );}
        end;

    END;
end;

end.
