//**  ,     VCL Delphi.
//**  Vladimir Kladov, 2000-2010
//**
//**      .
//**     /     
//**  2          
//**   GNU,  Free Software Foundation.
//**
//**       ,    
//** ,       ,   
//**         
//**   .      
//**     GNU.
//**
//**          
//**   GNU.     ,     Free
//** Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
//** MA 02111-1307, USA.
//******************************************************************************

{ TExcelReport -      MS Excel
  (C) by vladimir Kladov, 2006
  --------------------------------------------------------
  ,  , .  .     
   .
}

{$I VCL_VERSIONS.INC}

unit ExcelReport;

interface

uses Windows, Classes, SysUtils, Graphics{$IFDEF _D6orHigher}, Variants{$ENDIF};

const
    noFrames = 0;
    frameBottom = 8;
    frameAround = 15;

type
    TExcelReport = class(TObject)
    private
        FDefaultFrame: Integer;
        FFrameWidth: Integer;
        FVerticalAlignment: Integer;
        FFontBold: Boolean;
        FNextCellFontItalic: Boolean;
        FNextCellFontUnderline: Boolean;
        FNextCellFontName: String;
        FNextCellFontSize: Double;
        FCellFontSize: Double;
        procedure SetDefaultFrame(const Value: Integer);
        procedure SetFrameWidth(const Value: Integer);
        procedure SetVerticalAlignment(const Value: Integer);
        procedure SetFontBold(const Value: Boolean);
        procedure SetNextCellFontItalic(const Value: Boolean);
        procedure SetNextCellFontUnderline(const Value: Boolean);
        procedure SetNextCellFontName(const Value: String);
        procedure SetNextCellFontSize(const Value: Double);
        procedure SetCellFontSize(const Value: Double);
    protected
        CellPosition: TPoint;
        CellSize: TSize;
        FNextNumFormat: Boolean;
        RestoreFontName: String;
    public
        WorkSheet: Variant; //  Excel.  ,    ,
        //       
        //   MS Excel.

        function NewWorksheet: Variant;
        function SelectWorksheet(idx: Integer): Variant;
        property DefaultFrame: Integer read FDefaultFrame write SetDefaultFrame;
        //       
        // 15 -  , 1 - , 2 - , 4 - , 8 -  

        property FrameWidth: Integer read FFrameWidth write SetFrameWidth;
        //  

        procedure FrameCells(x1, y1, x2, y2: Integer; w: Integer); overload;
        procedure FrameCells(x1: Char; y1: Integer; x2: Char; y2: Integer; w: Integer); overload;

        property NextCellFontBold: Boolean read FFontBold write SetFontBold;
        property NextCellFontItalic: Boolean read FNextCellFontItalic write SetNextCellFontItalic;
        property NextCellFontUnderline: Boolean read FNextCellFontUnderline write SetNextCellFontUnderline;
        property NextCellFontName: String read FNextCellFontName write SetNextCellFontName;
        property NextCellFontSize: Double read FNextCellFontSize write SetNextCellFontSize;

        property CellFontSize: Double read FCellFontSize write SetCellFontSize;

        procedure Cell(x1, y1, x2, y2: Integer; const Text: String;
            Frames: Integer = -1; //  -1,   DefaultFrame
            TextAlign: TAlignment = taCenter; ForceText: Boolean = FALSE); overload;
        //    ( x2 > x1  y2 > y1,   )

        procedure Cell(x1: Char; y1: Integer; x2: Char; y2: Integer; const Text: String;
            Frames: Integer = -1; //  -1,   DefaultFrame
            TextAlign: TAlignment = taCenter; ForceText: Boolean = FALSE); overload;
        //    ( x2 > x1  y2 > y1,   )

        procedure CellRight(szx, szy: Integer; const Text: String;
            Frames: Integer = -1; //  -1,   DefaultFrame
            TextAlign: TAlignment = taCenter; ForceText: Boolean = FALSE);
        //     

        procedure Right(szx, szy: Integer; const Text: String;
            Frames: Integer = -1; //  -1,   DefaultFrame
            TextAlign: TAlignment = taCenter; ForceText: Boolean = FALSE);
        //  ,  CellRight

        procedure CellDown(szx, szy: Integer; const Text: String;
            Frames: Integer = -1; //  -1,   DefaultFrame
            TextAlign: TAlignment = taCenter);
        //     

        procedure Down(szx, szy: Integer; const Text: String;
            Frames: Integer = -1; //  -1,   DefaultFrame
            TextAlign: TAlignment = taCenter);
        //  $&  CellDown

        procedure CellOffset(offx, offy, szx, szy: Integer; const Text: String;
            Frames: Integer = -1; //  -1,   DefaultFrame
            TextAlign: TAlignment = taCenter);
        //        

        procedure Offset(offx, offy, szx, szy: Integer; const Text: String;
            Frames: Integer = -1; //  -1,   DefaultFrame
            TextAlign: TAlignment = taCenter);
        //  ,  CellOffset

        procedure NextNumeric;
        property VerticalAlignment: Integer read FVerticalAlignment write SetVerticalAlignment;

        procedure WrapText(x1, x2: Integer);
        procedure PageOrientation(orientation: Integer);
        procedure Zoom(z: Integer);

        constructor Create;
        destructor Destroy; override;

        procedure RowHeight(nRow: Integer; h: Double);
        procedure ColWidth(nCol: Integer; w: Double); overload;
        procedure ColWidth(nCol: Char; w: Double); overload;

        function Range(x1, y1, x2, y2: Integer): Variant; overload;
        function Range(x1: Char; y1: Integer; x2: Char; y2: Integer): Variant; overload;

        procedure ShowWindow;


    end;


implementation

uses ComObj, ActiveX;

function IsFloatNum(const s: String): Boolean;
var i: Integer;
begin
    Result := FALSE;
    if s = '' then Exit;
    for i := 1 to Length(s) do
        if not (s[i] in ['0'..'9', '-', ',', '.']) then Exit;
    TRY
        StrToFloat(s);
        Result := TRUE;
    EXCEPT
    END;
end;

{ TExcelReport }

procedure TExcelReport.Cell(x1, y1, x2, y2: Integer; const Text: String;
    Frames: Integer = -1; TextAlign: TAlignment = taCenter; ForceText: Boolean = FALSE);
    function ReplaceAllDigsTo0(const s: String): String;
    var i: Integer;
    begin
        Result := s;
        for i := 1 to Length(s) do
            if s[i] in ['0'..'9'] then
                Result[i] := '0'
                    {else if s[ i ] = ',' then
                  Result[ i ] := '.'};
        while (Length(Result) > 1) and (Result[1] = '0') do
            Delete(Result, 1, 1);
    end;
var R: Variant;
begin
    CellPosition := Point(x1, y1);
    CellSize.cx := x2 - x1 + 1;
    CellSize.cy := y2 - y1 + 1;
    R := WorkSheet.Range[WorkSheet.Cells[y1, x1], WorkSheet.Cells[y2, x2]];
    R.MergeCells := TRUE;
    R.WrapText := TRUE;
    if NextCellFontBold then
    begin
        R.Font.Bold := 1;
        NextCellFontBold := FALSE;
    end;
    if NextCellFontItalic then
    begin
        R.Font.Italic := 1;
        NextCellFontItalic := FALSE;
    end;
    if NextCellFontUnderline then
    begin
        R.Font.Underline := 2;
        NextCellFontUnderline := FALSE;
    end;
    if NextCellFontName <> '' then
    begin
        RestoreFontName := R.Font.Name;
        R.Font.Name := NextCellFontName;
        NextCellFontName := '';
    end
    else
        if RestoreFontName <> '' then
        begin
            R.Font.Name := RestoreFontName;
            RestoreFontName := '';
        end;
    if NextCellFontSize > 0 then
    begin
        R.Font.Size := NextCellFontSize;
        NextCellFontSize := 0;
    end
    else if CellFontSize <> 0 then
        R.Font.Size := CellFontSize;
    if ForceText OR
        (Text <> '') and not (Text[1] in ['0'..'9']) or (pos('.', Text) > 0) or
        not IsFloatNum(Text) then
    begin
        R.NumberFormat := '@';
        if Text <> '' then
            R.Value := Text;
    end
    else
        if not ForceText and IsFloatNum(Text) then
        begin
            R.NumberFormat := ReplaceAllDigsTo0(Text);
            if Text <> '' then
                R.Value := StrToFloat(Text);
        end;
    if Frames < 0 then Frames := DefaultFrame;
    if Frames <> 0 then
    begin
        R.Borders.LineStyle := 1;
        if FrameWidth > 0 then
            R.Borders.Weight := FrameWidth;
        if Frames and 1 = 0 then R.Borders[1].Color := clWhite;
        if Frames and 2 = 0 then R.Borders[2].Color := clWhite;
        if Frames and 4 = 0 then R.Borders[3].Color := clWhite;
        if Frames and 8 = 0 then R.Borders[4].Color := clWhite;
    end;
    if TextAlign <> taLeftJustify then
        CASE TextAlign OF
            taLeftJustify: R.HorizontalAlignment := 1;
            taCenter: R.HorizontalAlignment := 3;
            taRightJustify: R.HorizontalAlignment := 4;
        END;
    CASE VerticalAlignment OF
        0: ;
        1: R.VerticalAlignment := 1;
        2: R.VerticalAlignment := 2;
    END;
    R := Unassigned;
end;

procedure TExcelReport.CellDown(szx, szy: Integer; const Text: String;
    Frames: Integer; TextAlign: TAlignment);
begin
    Cell(CellPosition.x, CellPosition.y + CellSize.cy,
        CellPosition.x + szx - 1, CellPosition.y + CellSize.cy + szy - 1,
        Text, Frames, TextAlign);
end;

procedure TExcelReport.CellOffset(offx, offy, szx, szy: Integer;
    const Text: String; Frames: Integer; TextAlign: TAlignment);
begin
    Cell(CellPosition.x + offx, CellPosition.y + offy,
        CellPosition.x + offx + szx - 1, CellPosition.y + offy + szy - 1,
        Text, Frames, TextAlign);
end;

procedure TExcelReport.CellRight(szx, szy: Integer; const Text: String;
    Frames: Integer = -1; TextAlign: TAlignment = taCenter; ForceText: Boolean = FALSE);
begin
    Cell(CellPosition.x + CellSize.cx, CellPosition.y,
        CellPosition.x + CellSize.cx + szx - 1, CellPosition.y + szy - 1,
        Text, Frames, TextAlign, ForceText);
end;

procedure TExcelReport.Zoom(z: Integer);
var P: Variant;
begin
    P := WorkSheet.PageSetup;
    //P.FitToPagesWide := 1;
    P.Zoom := z;
end;

procedure TExcelReport.FrameCells(x1, y1, x2, y2, w: Integer);
var R: Variant;
begin
    R := WorkSheet.Range[WorkSheet.Cells[y1, x1], WorkSheet.Cells[y2, x2]];
    R.BorderAround(1, w);
end;

procedure TExcelReport.NextNumeric;
begin
    FNextNumFormat := TRUE;
end;

procedure TExcelReport.PageOrientation(orientation: Integer);
var P: Variant;
begin
    P := WorkSheet.PageSetup;
    P.Orientation := orientation;
end;

procedure TExcelReport.SetDefaultFrame(const Value: Integer);
begin
    FDefaultFrame := Value;
end;

procedure TExcelReport.SetFontBold(const Value: Boolean);
begin
    FFontBold := Value;
end;

procedure TExcelReport.SetFrameWidth(const Value: Integer);
begin
    FFrameWidth := Value;
end;

procedure TExcelReport.SetNextCellFontItalic(const Value: Boolean);
begin
    FNextCellFontItalic := Value;
end;

procedure TExcelReport.SetNextCellFontUnderline(const Value: Boolean);
begin
    FNextCellFontUnderline := Value;
end;

procedure TExcelReport.SetVerticalAlignment(const Value: Integer);
begin
    FVerticalAlignment := Value;
end;

procedure TExcelReport.WrapText(x1, x2: Integer);
var R: Variant;
begin
    R := WorkSheet.Range[WorkSheet.Cells[1, x1], WorkSheet.Cells[1, x2]];
    R.Columns.WrapText := TRUE;
end;

destructor TExcelReport.Destroy;
begin
    WorkSheet := Unassigned;
    inherited;
    CoUninitialize;
end;

function TExcelReport.NewWorksheet: Variant;
begin
    Worksheet := Worksheet.Application.Worksheets.Add;
    Result := Worksheet;
    //Worksheet.Select;
end;

function TExcelReport.SelectWorksheet(idx: Integer): Variant;
begin
    Worksheet := Worksheet.Application.Worksheets[idx];
    Result := Worksheet;
end;

procedure TExcelReport.SetNextCellFontName(const Value: String);
begin
    FNextCellFontName := Value;
end;

procedure TExcelReport.SetNextCellFontSize(const Value: Double);
begin
    FNextCellFontSize := Value;
end;

procedure TExcelReport.Cell(x1: Char; y1: Integer; x2: Char; y2: Integer;
    const Text: String; Frames: Integer; TextAlign: TAlignment;
    ForceText: Boolean);
begin
    if x1 >= 'a' then
        x1 := UpperCase(x1)[1];
    Cell(Ord(x1) - Ord('A') + 1, y1, Ord(x2) - Ord('A') + 1, y2,
        Text, Frames, TextAlign, ForceText);
end;

procedure TExcelReport.Down(szx, szy: Integer; const Text: String;
    Frames: Integer; TextAlign: TAlignment);
begin
    CellDown(szx, szy, Text, Frames, TextAlign);
end;

procedure TExcelReport.Offset(offx, offy, szx, szy: Integer;
    const Text: String; Frames: Integer; TextAlign: TAlignment);
begin
    CellOffset(offx, offy, szx, szy, Text, Frames, TextAlign);
end;

procedure TExcelReport.Right(szx, szy: Integer; const Text: String;
    Frames: Integer; TextAlign: TAlignment; ForceText: Boolean);
begin
    CellRight(szx, szy, Text, Frames, TextAlign, ForceText);
end;

procedure TExcelReport.RowHeight(nRow: Integer; h: Double);
var R: Variant;
begin
    R := WorkSheet.Rows[nRow];
    R.RowHeight := h;
end;

procedure TExcelReport.ColWidth(nCol: Integer; w: Double);
var R: Variant;
begin
    R := WorkSheet.Columns[nCol];
    R.ColumnWidth := w;
end;

procedure TExcelReport.ColWidth(nCol: Char; w: Double);
var R: Variant;
    i: Integer;
begin
    if nCol >= 'a' then
        nCol := UpperCase(nCol)[1];
    i := Integer(Ord(nCol) - Ord('A') + 1);
    R := WorkSheet.Columns[i];
    R.ColumnWidth := w;
end;

function TExcelReport.Range(x1, y1, x2, y2: Integer): Variant;
begin
    Result := WorkSheet.Range[WorkSheet.Cells[y1, x1],
        WorkSheet.Cells[y2, x2]];
end;

function TExcelReport.Range(x1: Char; y1: Integer; x2: Char;
    y2: Integer): Variant;
begin
    if x1 >= 'a' then
        x1 := UpperCase(x1)[1];
    if x2 >= 'a' then
        x2 := UpperCase(x2)[1];
    Result := Range(Ord(x1) - Ord('A') + 1, y1,
        Ord(x2) - Ord('A') + 1, y2);
end;

procedure TExcelReport.ShowWindow;
begin
    WorkSheet.Application.Show;
end;

procedure TExcelReport.FrameCells(x1: Char; y1: Integer; x2: Char; y2,
    w: Integer);
begin
    if x1 >= 'a' then
        x1 := UpperCase(x1)[1];
    if x2 >= 'a' then
        x2 := UpperCase(x2)[1];
    FrameCells(Ord(x1) - Ord('A') + 1, y1, Ord(x2) - Ord('A') + 1, y2, w);
end;

procedure TExcelReport.SetCellFontSize(const Value: Double);
begin
    FCellFontSize := Value;
end;

constructor TExcelReport.Create;
begin
    inherited;
    CoInitialize(nil);
end;

end.
