//**  ,     VCL Delphi.
//**  Vladimir Kladov, 2000-2010
//**
//**      .
//**     /     
//**  2          
//**   GNU,  Free Software Foundation.
//**
//**       ,    
//** ,       ,   
//**         
//**   .      
//**     GNU.
//**
//**          
//**   GNU.     ,     Free
//** Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
//** MA 02111-1307, USA.
//******************************************************************************

unit FLVCompEditor;

interface

{$I VCL_VERSIONS.INC}

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  comctrls, imglist, commctrl, QSort, ColorListView, menus, Extctrls, registry,
  {$IFDEF _D6} designintf {$IFDEF IN_PACKAGE}, DesignEditors {$ENDIF}
  {$ELSE}
     {$IFDEF _D7orHigher} designintf {$IFDEF IN_PACKAGE}, DesignEditors {$ENDIF}
     {$ELSE} dsgnintf {$ENDIF}
  {$ENDIF},
  FastListView;

type
  TFastLVEditor = class( TComponentEditor )
  private
  protected
    procedure EditColumns;
    procedure ImportListView;
    procedure DoImport( LV: TOldListView; FLV: TFastListView; DeleteOld: Boolean );
  public
    procedure Edit; override;
    procedure ExecuteVerb(Index: Integer); override;
    function GetVerb(Index: Integer): string; override;
    function GetVerbCount: Integer; override;
  end;

  TFastLVColumnsEditor = class( TPropertyEditor )
  private
  protected
  public
    procedure Edit; override;
    function GetAttributes: TPropertyAttributes; override;
  end;

  TFastListView_Access = class( TFastListView )
  end;

  TFLVColumn = class( TComponent )
  private
    FVisible: Boolean;
    FFiltered: Boolean;
    FAutosize: Boolean;
    FMinWidth: Integer;
    FWidth: Integer;
    FMaxWidth: Integer;
    FImageIndex: Integer;
    FOrder: Integer;
    FCaption: String;
    FAlignment: TAlignment;
    FSortAttrs: TColumnSortAttrs;
    FFieldName: String;
    FCanEdit: Boolean;
    FEditKind: TEditKind;
    FComboItems: TStrings;
    procedure SetAlignment(const Value: TAlignment);
    procedure SetAutosize(const Value: Boolean);
    procedure SetCaption(const Value: String);
    procedure SetFiltered(const Value: Boolean);
    procedure SetImageIndex(const Value: Integer);
    procedure SetMaxWidth(const Value: Integer);
    procedure SetMinWidth(const Value: Integer);
    procedure SetOrder(const Value: Integer);
    procedure SetSortAttrs(const Value: TColumnSortAttrs);
    procedure SetVisible(const Value: Boolean);
    procedure SetWidth(const Value: Integer);
    procedure SetFieldName(const Value: String);
    procedure SetCanEdit(const Value: Boolean);
    procedure SetComboItems(const Value: TStrings);
    procedure SetEditKind(const Value: TEditKind);
  protected
    procedure Changed;
  public
    FDesigner: TObject;
    destructor Destroy; override;
  published
    property Caption: String read FCaption write SetCaption;
    property Alignment: TAlignment read FAlignment write SetAlignment;
    property Autosize: Boolean read FAutosize write SetAutosize;
    property ImageIndex: Integer read FImageIndex write SetImageIndex;
    property MaxWidth: Integer read FMaxWidth write SetMaxWidth;
    property MinWidth: Integer read FMinWidth write SetMinWidth;
    property Width: Integer read FWidth write SetWidth;
    property SortAttrs: TColumnSortAttrs read FSortAttrs write SetSortAttrs;
    property Filtered: Boolean read FFiltered write SetFiltered;
    property Visible: Boolean read FVisible write SetVisible;
    property Order: Integer read FOrder write SetOrder;
    property FieldName: String read FFieldName write SetFieldName;
    property CanEdit: Boolean read FCanEdit write SetCanEdit;
    property EditKind: TEditKind read FEditKind write SetEditKind;
    property ComboItems: TStrings read FComboItems write SetComboItems;
  end;

  {TStringsProperty = class( TDefaultEditor )
  public
    procedure EditProperty(PropertyEditor: TPropertyEditor;
      var Continue, FreeEditor: Boolean); override;
  end;}

function GetFLVColumnsCount( FLV: TObject ): Integer;
function GetFLVColumn_SortAttrs( FLV: TObject; ColIdx: Integer ): TColumnSortAttrs;
procedure SetFLVColumn_SortAttrs( FLV: TObject; ColIdx: Integer; Value: TColumnSortAttrs );
function GetFLVColumn_Filtered( FLV: TObject; ColIdx: Integer ): Boolean;
procedure SetFLVColumn_Filtered( FLV: TObject; ColIdx: Integer; Value: Boolean );
function GetFLVColumn_Visible( FLV: TObject; ColIdx: Integer ): Boolean;
procedure SetFLVColumn_Visible( FLV: TObject; ColIdx: Integer; Value: Boolean );
function GetFLVColumn_Order( FLV: TObject; ColIdx: Integer ): Integer;
procedure SetFLVColumn_Order( FLV: TObject; ColIdx: Integer; Value: Integer );
function GetFLVColumn_FieldName( FLV: TObject; ColIdx: Integer ): String;
procedure SetFLVColumn_FieldName( FLV: TObject; ColIdx: Integer; Value: String );

function GetFLVColumn_CanEdit( FLV: TObject; ColIdx: Integer ): Boolean;
procedure SetFLVColumn_CanEdit( FLV: TObject; ColIdx: Integer; Value: Boolean );
function GetFLVColumn_EditKind( FLV: TObject; ColIdx: Integer ): TEditKind;
procedure SetFLVColumn_EditKind( FLV: TObject; ColIdx: Integer; Value: TEditKind );
function GetFLVColumn_ComboItems( FLV: TObject; ColIdx: Integer ): TStrings;
procedure SetFLVColumn_ComboItems( FLV: TObject; ColIdx: Integer; Value: TStrings );

procedure FLVColumn_DeleteFieldName( FLV: TObject; ColIdx: Integer );
procedure FLVColumn_InsertFieldName( FLV: TObject; ColIdx: Integer; const NewName: String );
procedure FLVColumn_MoveFieldName( FLV: TObject; FromIndex, ToIndex: Integer );

procedure Register;

implementation

uses
  WhatToImport, FLVResources, fmuFastLVColsEdit;

procedure Register;
begin
  RegisterComponentEditor( TFastListView, TFastLVEditor );
  RegisterPropertyEditor( TypeInfo( TListColumns ), TFastListView, 'Columns',
                          TFastLVColumnsEditor );
  {RegisterPropertyEditor( TypeInfo( TStrings ), TFastListItem, 'ComboItems',
                          TStringsProperty );}
  //RegisterComponentEditor( TListColumn, TStringsProperty );
end;

type
  TFastListView1 = class( TFastListView )
  end;

function GetFLVColumnsCount( FLV: TObject ): Integer;
begin
  Result := (FLV as TFastListView).Columns.Count;
end;

function GetFLVColumn_SortAttrs( FLV: TObject; ColIdx: Integer ): TColumnSortAttrs;
begin
  Result := (FLV as TFastListView).ColumnSortAttrs[ ColIdx ];
end;

procedure SetFLVColumn_SortAttrs( FLV: TObject; ColIdx: Integer; Value: TColumnSortAttrs );
begin
  (FLV as TFastListView).SetSortColumnAttributes( [ ColIdx ], Value );
end;

function GetFLVColumn_Filtered( FLV: TObject; ColIdx: Integer ): Boolean;
begin
  Result := (FLV as TFastListView).ColumnFiltered[ ColIdx ];
end;

procedure SetFLVColumn_Filtered( FLV: TObject; ColIdx: Integer; Value: Boolean );
begin
  {if Value then
    Windows.Beep( 500, 500 );}
  (FLV as TFastListView).ColumnFiltered[ ColIdx ] := Value;
end;

function GetFLVColumn_Visible( FLV: TObject; ColIdx: Integer ): Boolean;
begin
  Result := (FLV as TFastListView).ColumnVisible[ ColIdx ];
end;

procedure SetFLVColumn_Visible( FLV: TObject; ColIdx: Integer; Value: Boolean );
begin
  (FLV as TFastListView).ColumnVisible[ ColIdx ] := Value;
end;

function GetFLVColumn_Order( FLV: TObject; ColIdx: Integer ): Integer;
begin
  Result := (FLV as TFastListView).ColumnOrder[ ColIdx ];
end;

procedure SetFLVColumn_Order( FLV: TObject; ColIdx: Integer; Value: Integer );
begin
  (FLV as TFastListView).ColumnOrder[ ColIdx ] := Value;
end;

function GetFLVColumn_FieldName( FLV: TObject; ColIdx: Integer ): String;
begin
  Result := (FLV as TFastListView).ColumnFieldNames[ ColIdx ];
end;

procedure SetFLVColumn_FieldName( FLV: TObject; ColIdx: Integer; Value: String );
begin
  (FLV as TFastListView).ColumnFieldNames[ ColIdx ] := Value;
end;

function GetFLVColumn_CanEdit( FLV: TObject; ColIdx: Integer ): Boolean;
begin
  Result := (FLV as TFastListView).ColumnCanEdit[ ColIdx ];
end;

procedure SetFLVColumn_CanEdit( FLV: TObject; ColIdx: Integer; Value: Boolean );
begin
  (FLV as TFastListView).ColumnCanEdit[ ColIdx ] := Value;
end;

function GetFLVColumn_EditKind( FLV: TObject; ColIdx: Integer ): TEditKind;
begin
  Result := (FLV as TFastListView).ColumnEditKind[ ColIdx ];
end;

procedure SetFLVColumn_EditKind( FLV: TObject; ColIdx: Integer; Value: TEditKind );
begin
  (FLV as TFastListView).ColumnEditKind[ ColIdx ] := Value;
end;

function GetFLVColumn_ComboItems( FLV: TObject; ColIdx: Integer ): TStrings;
begin
  Result := (FLV as TFastListView).ColumnComboItems[ ColIdx ];
  //ShowMessage( 'items:' + IntToStr( Result.Count ) );
end;

procedure SetFLVColumn_ComboItems( FLV: TObject; ColIdx: Integer; Value: TStrings );
begin
  (FLV as TFastListView).ColumnComboItems[ ColIdx ] := Value;
end;

procedure FLVColumn_DeleteFieldName( FLV: TObject; ColIdx: Integer );
begin
  TFastListview1(FLV).FColumnNames.Delete( ColIdx );
end;

procedure FLVColumn_InsertFieldName( FLV: TObject; ColIdx: Integer; const NewName: String );
begin
  if TFastListview1(FLV).FColumnNames = nil then
    TFastListview1(FLV).FColumnNames := TStringList.Create;
  while ColIdx >= TFastListview1(FLV).FColumnNames.Count do
    TFastListview1(FLV).FColumnNames.Add( '' );
  TFastListview1(FLV).FColumnNames.Insert( ColIdx, NewName );
end;

procedure FLVColumn_MoveFieldName( FLV: TObject; FromIndex, ToIndex: Integer );
begin
  TFastListview1(FLV).FColumnNames.Move( FromIndex, ToIndex );
end;

procedure ShowColumnsEditor( Comp: TPersistent );
begin
  if not (Comp is TFastListView) then
  begin
    ShowMessage( 'Is not TFastListView' );
    Exit;
  end;
  if TFastListView_Access( Comp ).FColumnsEditor_DesignTime = nil then
    TFastListView_Access( Comp ).FColumnsEditor_DesignTime :=
      TfmFLVColsEdit.Create( Application );
  TfmFLVColsEdit( TFastListView_Access( Comp ).FColumnsEditor_DesignTime )
    .Caption := TFastListView( Comp ).Name + '.Columns';
  TfmFLVColsEdit( TFastListView_Access( Comp ).FColumnsEditor_DesignTime )
    .DesignColumns := (Comp as TFastListView).Columns;
  TfmFLVColsEdit( TFastListView_Access( Comp ).FColumnsEditor_DesignTime )
    .DesignObject := Comp;

  TfmFLVColsEdit( TFastListView_Access( Comp ).FColumnsEditor_DesignTime )
    .Show;
end;

{ TFastLVEditor }

procedure TFastLVEditor.DoImport(LV: TOldListView; FLV: TFastListView; DeleteOld: Boolean);
var NewName: String;
    NewParent: TWinControl;
    NewBoundsRect: TRect;
    NewTabOrder: Integer;
    NewComponentIndex: Integer;
    I: Integer;
    Col: TListColumn;
    {$IFDEF _D6orHigher}

    {$ELSE}
    LVI: IComponent;
    {$ENDIF}

begin
  try

  with FLV do
  BEGIN

    if LV = nil then Exit;
    NewName := LV.Name;
    NewParent := LV.Parent;
    NewBoundsRect := LV.BoundsRect;
    NewTabOrder := LV.TabOrder;
    NewComponentIndex := LV.ComponentIndex;

    I := 1;
    while I > 0 do
    begin
      try
        LV.Name := NewName + '_Old' + IntToStr( I );
        I := -1;
      except
      end;
      Inc( I );
    end;
    Name := NewName;

    Items.Clear;
    Columns.Clear;
    for I := 0 to LV.Columns.Count - 1 do
    begin
      Col := Columns.Add;
      Col.Width := LV.Columns[ I ].Width;
      Col.autosize := LV.Columns[ I ].autosize;
      Col.Caption := LV.Columns[ I ].Caption;
      Col.DisplayName := LV.Columns[ I ].DisplayName;
      Col.ImageIndex := LV.Columns[ I ].ImageIndex;
      Col.MaxWidth := LV.Columns[ I ].MaxWidth;
      Col.MinWidth := LV.Columns[ I ].MinWidth;
      Col.Alignment := LV.Columns[ I ].Alignment;
    end;

    I := 1;
    while I > 0 do
    begin
      try
        LV.Name := NewName + '_Old' + IntToStr( I );
        I := -1;
      except
      end;
      Inc( I );
    end;
    Name := NewName;

    if NewParent <> nil then
      Parent := NewParent
    else
    if LV.ParentWindow <> 0 then
      ParentWindow := LV.ParentWindow
    else
    begin
      {$IFDEF _D6orHigher}
      //ParentWindow := Designer.ContainerWindow.Handle;
      {$ELSE}
      ParentWindow := Designer.ContainerWindow.Handle;
      {$ENDIF}
      windows.Beep( 400, 200 );
    end;



    AllocBy := LV.AllocBy;
    BorderStyle := LV.BorderStyle;
    Checkboxes := LV.Checkboxes;
    ColumnClick := LV.ColumnClick;
    FlatScrollBars := LV.FlatScrollBars;
    FullDrag := LV.FullDrag;
    GridLines := LV.GridLines;
    HideSelection := LV.HideSelection;
    HotTrack := LV.HotTrack;
    HotTrackStyles := LV.HotTrackStyles;
    IconOptions := LV.IconOptions;
    LargeImages := LV.LargeImages;
    MultiSelect := LV.MultiSelect;
    OwnerDraw := LV.OwnerDraw;
    ReadOnly := LV.ReadOnly;
    RowSelect := LV.RowSelect;
    ShowColumnHeaders := LV.ShowColumnHeaders;
    ShowWorkAreas := LV.ShowWorkAreas;
    SmallImages := LV.SmallImages;
    SortType := LV.SortType;
    StateImages := LV.StateImages;
    ViewStyle := LV.ViewStyle;
    BorderWidth := LV.BorderWidth;
    Ctl3D := LV.Ctl3D;
    HelpContext := LV.HelpContext;
    TabStop := LV.TabStop;
    BiDiMode := LV.BiDiMode;
    Color := LV.Color;
    Cursor := LV.Cursor;

    Constraints.MaxHeight := LV.Constraints.MaxHeight;
    Constraints.MaxWidth := LV.Constraints.MaxWidth;
    Constraints.MinHeight := LV.Constraints.MinHeight;
    Constraints.MinWidth := LV.Constraints.MinWidth;
    Anchors := LV.Anchors;
    Align := LV.Align;
    DragCursor := LV.DragCursor;
    DragKind := LV.DragKind;
    DragMode := LV.DragMode;
    Enabled := LV.Enabled;
    Font := LV.Font;
    Hint := LV.Hint;
    ParentBiDiMode := LV.ParentBiDiMode;
    ParentColor := LV.Parentcolor;
    ParentFont := LV.ParentFont;
    ParentShowHint := LV.ParentShowHint;
    PopupMenu := LV.PopupMenu;
    ShowHint := LV.showHint;
    Visible := LV.Visible;
    Tag := LV.Tag;

    OnAdvancedCustomDraw := LV.OnAdvancedCustomDraw;
    OnAdvancedCustomDrawItem := LV.OnAdvancedCustomDrawItem;
    //nAdvancedCustomDrawSubItem := LV.OnAdvancedCustomDrawSubItem;
    OnChange := LV.OnChange;
    OnChanging := LV.OnChanging;
    OnColumnClick := LV.OnColumnClick;
    OnColumnDragged := LV.OnColumnDragged;
    OnColumnRightClick := LV.OnColumnRightClick;
    OnCompare := TFastLVCompareEvent( LV.OnCompare );
    OnCustomDraw := LV.OnCustomDraw;
    OnCustomDrawItem := LV.OnCustomDrawItem;
    //OnCustomDrawSubItem := LV.OnCustomDrawSubItem;
    if Assigned( LV.OnData ) or Assigned( LV.OnDataFind ) or Assigned( LV.OnDataHint )
       or Assigned( LV.OnDataStateChange ) or Assigned( OnDrawItem ) then
    begin
      ShowMessage( '     : '#13 +
                   'OnData, OnDataFind, OnDataHint, OnDataStateChange, ' +
                   'OnDrawItem, OnSelectItem, OnDeletion, OnCustomDrawSubItem - .');
    end;
    //OnDeletion := LV.OnDeletion;
    //OnDrawItem := LV.OnDrawItem;
    //OnEdited := LV.OnEdited;
    //OnEditing := LV.OnEditing;
    OnGetImageIndex := LV.OnGetImageIndex;
    OnGetSubItemImage := LV.OnGetSubItemImage;
    OnInfoTip := LV.OnInfoTip;
    OnInsert := LV.OnInsert;
    //OnSelectItem := LV.OnSelectItem;
    OnEnter := LV.OnEnter;
    OnExit := LV.OnExit;
    OnKeyDown := LV.OnKeyDown;
    OnKeyPress := LV.OnKeyPress;
    OnKeyUp := LV.OnKeyUp;
    OnClick := LV.OnClick;
    OnContextPopup := LV.OnContextPopup;
    OnDataFind := LV.OnDataFind;
    OnDblClick := LV.OnDblClick;
    OnDragDrop := LV.OnDragDrop;
    OnDragOver := LV.OnDragOver;
    OnEndDock := LV.OnEndDock;
    OnMouseDown := LV.OnMouseDown;
    OnMouseMove := LV.OnMouseMove;
    OnMouseUp := LV.OnMouseUp;
    OnResize := LV.OnResize;
    OnStartDock := LV.OnStartDock;
    OnStartDrag := LV.OnStartDrag;


    Left := NewBoundsRect.Left;
    Top := NewBoundsRect.Top;
    Width := NewBoundsRect.Right - NewBoundsRect.Left;
    Height := NewBoundsRect.Bottom - NewBoundsRect.Top;

    TabOrder := NewTabOrder;
    ComponentIndex := NewComponentIndex;

    if DeleteOld then
    begin
      try

        {$IFDEF _D6orHigher}
        LV.Free;
        {$ELSE}
        LVI := MakeIComponent( LV );
        LVI.DestroyObject;
        {$ENDIF}

      finally
      end;
    end;

    END;

    windows.Beep( 600, 600 );

  except

  end;


end;

procedure TFastLVEditor.Edit;
begin
  EditColumns;
end;

procedure TFastLVEditor.EditColumns;
var Comp: TPersistent;
begin
  Comp := {$IFDEF _D6orHigher} GetComponent {$ELSE} Component {$ENDIF};
  ShowColumnsEditor( Comp );
end;

procedure TFastLVEditor.ExecuteVerb(Index: Integer);
begin
  if Component = nil then Exit;
  case Index of
  0: EditColumns;
  1: ImportListView;
  end;
end;

function TFastLVEditor.GetVerb(Index: Integer): string;
begin
  case Index of
  0: Result := CompEditEditCols;
  1: Result := CompEditImportTListView;
  end;
end;

function TFastLVEditor.GetVerbCount: Integer;
begin
  Result := 2;
end;

procedure TFastLVEditor.ImportListView;
var WTI: TfmWhatToImport;
    F: TForm;
    Fr: TFrame;
    C: TComponent;
    I, J: Integer;
    DeleteOld: Boolean;
begin
  if Component.Owner = nil then Exit;
  if not( Component.Owner is TForm ) and not( Component.Owner is TFrame ) then
  begin
    ShowMessage( Err0101 );
    Exit;
  end;
  F := nil;
  Fr := nil;
  if Component.Owner is TForm
    then F := Component.Owner as TForm
    else Fr := Component.Owner as TFrame;
  WTI := TfmWhatToImport.Create( Application );
  //       TListView      :
  if Assigned( F ) then
  for I := 0 to F.ComponentCount-1 do
  begin
    C := F.Components[ I ];
    if C is TOldListView then
    begin
      J := WTI.cbNames.Items.Add( C.Name + ': ' + C.ClassName );
      WTI.cbNames.Items.Objects[ J ] := C;
    end;
  end
    else
  for I := 0 to Fr.ComponentCount-1 do
  begin
    C := Fr.Components[ I ];
    if C is TOldListView then
    begin
      J := WTI.cbNames.Items.Add( C.Name + ': ' + C.ClassName );
      WTI.cbNames.Items.Objects[ J ] := C;
    end;
  end;
  if WTI.cbNames.Items.Count = 1 then
  begin
    WTI.cbNames.ItemIndex := 0;
    WTI.btnOK.Enabled := TRUE;
  end;
  WTI.ShowModal;
  if WTI.ModalResult = mrOK then
  begin
    C := TComponent( WTI.cbNames.Items.Objects[ WTI.cbNames.ItemIndex ] );
    DeleteOld := WTI.chkDeleteOld.Checked;
    WTI.Free;
    WTI := nil;
    DoImport( C as TOldListView, Component as TFastListView, DeleteOld );
  end;
  WTI.Free;
end;

{ TFastLVColumnsEditor }

procedure TFastLVColumnsEditor.Edit;
var Comp: TPersistent;
begin
  Comp := GetComponent( 0 );
  ShowColumnsEditor( Comp );
end;

function TFastLVColumnsEditor.GetAttributes: TPropertyAttributes;
begin
  Result := [ paDialog ];
end;

{ TFLVColumn }

procedure TFLVColumn.Changed;
begin
  if Assigned( FDesigner ) then
    TfmFLVColsEdit( FDesigner ).ColumnChanged( Self );
end;

destructor TFLVColumn.Destroy;
begin
  FComboItems.Free;
  inherited;
end;

procedure TFLVColumn.SetAlignment(const Value: TAlignment);
begin
  FAlignment := Value;
  Changed;
end;

procedure TFLVColumn.SetAutosize(const Value: Boolean);
begin
  FAutosize := Value;
  Changed;
end;

procedure TFLVColumn.SetCanEdit(const Value: Boolean);
begin
  FCanEdit := Value;
  Changed;
end;

procedure TFLVColumn.SetCaption(const Value: String);
begin
  FCaption := Value;
  Changed;
end;

procedure TFLVColumn.SetComboItems(const Value: TStrings);
begin
  if FComboItems = nil then
  begin
    if Value = nil then Exit;
    FComboItems := TStringList.Create;
  end;
  FComboItems.Clear;
  if Value <> nil then
    FComboItems.Assign( Value );
  if FComboItems.Count = 0 then
    FreeAndNil( FComboItems );
  Changed;
end;

procedure TFLVColumn.SetEditKind(const Value: TEditKind);
begin
  FEditKind := Value;
  Changed;
end;

procedure TFLVColumn.SetFieldName(const Value: String);
begin
  FFieldName := Value;
  Changed;
end;

procedure TFLVColumn.SetFiltered(const Value: Boolean);
begin
  FFiltered := Value;
  Changed;
end;

procedure TFLVColumn.SetImageIndex(const Value: Integer);
begin
  FImageIndex := Value;
  Changed;
end;

procedure TFLVColumn.SetMaxWidth(const Value: Integer);
begin
  FMaxWidth := Value;
  Changed;
end;

procedure TFLVColumn.SetMinWidth(const Value: Integer);
begin
  FMinWidth := Value;
  Changed;
end;

procedure TFLVColumn.SetOrder(const Value: Integer);
begin
  FOrder := Value;
  Changed;
end;

procedure TFLVColumn.SetSortAttrs(const Value: TColumnSortAttrs);
begin
  FSortAttrs := Value;
  Changed;
end;

procedure TFLVColumn.SetVisible(const Value: Boolean);
begin
  FVisible := Value;
  if not Value then Width := 0
               else if Width = 0 then Width := 50;
  Changed;
end;

procedure TFLVColumn.SetWidth(const Value: Integer);
begin
  FWidth := Value;
  Changed;
end;

{ TStringsProperty }

{procedure TStringsProperty.EditProperty(PropertyEditor: TPropertyEditor;
  var Continue, FreeEditor: Boolean);
var
  PropName: string;
begin
  PropName := PropertyEditor.GetName;
  if (CompareText(PropName, 'STRINGS') = 0) then begin
    PropertyEditor.Edit;
    Continue := False;
  end
    else inherited;
end;}

end.
