//**  ,     VCL Delphi.
//**  Vladimir Kladov, 2000-2010
//**
//**      .
//**     /     
//**  2          
//**   GNU,  Free Software Foundation.
//**
//**       ,    
//** ,       ,   
//**         
//**   .      
//**     GNU.
//**
//**          
//**   GNU.     ,     Free
//** Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
//** MA 02111-1307, USA.
//******************************************************************************

{------------------------------------------------------------
         "    "
       :  ..

       : 2 ()
       : 
-------------------------------------------------------------
}
unit FLVProgress;

interface

{$I VCL_VERSIONS.INC}

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  ExtCtrls, StdCtrls, ComCtrls;

type
  TCancelButton = class( TButton )
  private
  protected
    OldCursor: TCursor;
    MouseInControl: Boolean;
    procedure CMMouseEnter( var Msg: TMessage ); message CM_MOUSEENTER;
    procedure CMMouseLeave( var Msg: TMessage ); message CM_MOUSELEAVE;
    procedure WMMouseMove( var Msg: TWMMouseMove ); message WM_MOUSEMOVE;
    procedure WMMouseLeave( var Msg: TMessage ); message WM_MOUSELEAVE;
  public
  published
  end;

  TFLV_fmProgress = class(TForm)
    Panel1: TPanel;
    ProgressBar1: TProgressBar;
    btnCancel: TButton;
    Label1: TLabel;
    Label2: TLabel;
    procedure btnCancelClick(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure FormShow(Sender: TObject);
  private
    FCancelled: Boolean;
    function GetProgress: Integer;
    procedure SetProgress(const Value: Integer);
    { Private declarations }
  public
    { Public declarations }
    property Cancelled: Boolean read FCancelled;
    property Progress: Integer read GetProgress write SetProgress;
    procedure StepIt;
  end;

{var
  fmProgress: TfmProgress;}

implementation

uses FLVResources;

{$R *.DFM}

{ TFLV_fmProgress }

function TFLV_fmProgress.GetProgress: Integer;
begin
  Result := ProgressBar1.Position;
end;

procedure TFLV_fmProgress.SetProgress(const Value: Integer);
begin
  ProgressBar1.Position := Value;
  Update;
end;

procedure TFLV_fmProgress.btnCancelClick(Sender: TObject);
begin
  FCancelled := TRUE;
  Label1.Caption := '  ';
  Update;
end;

procedure TFLV_fmProgress.StepIt;
begin
  ProgressBar1.Position := ProgressBar1.Position + 1;
  BringToFront;
  Application.ProcessMessages;
end;

{ TCancelButton }

procedure TCancelButton.CMMouseEnter(var Msg: TMessage);
begin
  OldCursor := Screen.Cursor;
  Screen.Cursor := crArrow;
end;

procedure TCancelButton.CMMouseLeave(var Msg: TMessage);
begin
  Screen.Cursor := OldCursor;
end;

procedure TFLV_fmProgress.FormCreate(Sender: TObject);
var BCancel: TCancelButton;
begin
  btnCancel.Caption := ProgressCancel;
  Caption           := ProgressCaption;
  BCancel := TCancelButton.Create( Self );
  BCancel.Parent := btnCancel.Parent;
  BCancel.Caption := btnCancel.Caption;
  BCancel.Cancel := TRUE;
  BCancel.OnClick := btnCancel.OnClick;
  BCancel.BoundsRect := btnCancel.BoundsRect;
  BCancel.Cursor := crArrow;
end;

function _TrackMouseEvent(lpEventTrack: PTrackMouseEvent): BOOL; stdcall;
              external 'comctl32.dll' name '_TrackMouseEvent';

procedure TCancelButton.WMMouseLeave(var Msg: TMessage);
begin
  MouseInControl := FALSE;
  Perform( CM_MOUSELEAVE, 0, 0 );
end;

procedure TCancelButton.WMMouseMove(var Msg: TWMMouseMove);
var Track: TTrackMouseEvent;
begin
  inherited;
  if not MouseInControl then
  begin
    MouseInControl := TRUE;
    Track.cbSize := Sizeof( Track );
    Track.dwFlags := TME_LEAVE;
    Track.hwndTrack := Handle;
    //Track.dwHoverTime := 0;
    _TrackMouseEvent( @ Track );
    Perform( CM_MOUSEENTER, 0, 0 );
  end;
end;

procedure TFLV_fmProgress.FormShow(Sender: TObject);
begin
    if  Trim( Copy( Label1.Caption, 1, 1 ) ) = '' then
        Label1.Caption := ' ';
    if  Trim( Label2.Caption ) = '' then
        Label2.Caption := ', ';
end;

end.

