//**  ,     VCL Delphi.
//**  Vladimir Kladov, 2000-2010
//**
//**      .
//**     /     
//**  2          
//**   GNU,  Free Software Foundation.
//**
//**       ,    
//** ,       ,   
//**         
//**   .      
//**     GNU.
//**
//**          
//**   GNU.     ,     Free
//** Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
//** MA 02111-1307, USA.
//******************************************************************************

{
 TFastTable (C) by Vladimir Kladov, 2005

           , 
 Delphi   " ",      
     ,   ,
     ,    
   ,      .
         ,   100%-
           ,
  Delphi     ,  
        .    ,  
 ,      , 
      -.
}

unit FastTable;

interface

{$I VCL_VERSIONS.inc}

{$O+}

uses Windows, Classes, SysUtils, DBTables, AdoSQLUnit, db, Math, FltList;

type
    TFastIndex = class;
    PTFastIndex = ^TFastIndex;

    TFastTable = class
        Columns: TStringList;
        Lines: TList;
        LineObjects: TList;
    private
        FAutoAllocateColumns: Boolean;
        function GetValues(line: Integer; const column: String): String;
        procedure SetValues(line: Integer; const column, Value: String);
        function GetCount: Integer;
        function GetData(i: Integer): Pointer;
        procedure SetData(i: Integer; const Value: Pointer);
        function GetIntVal(line: Integer; const column: String): Integer;
        procedure SetIntVal(line: Integer; const column: String;
            const Value: Integer);
        function GetValByColIdx(line, colidx: Integer): String;
        function GetValByColIdx_StrOnly(line, colidx: Integer): String;
        function GetIntValByColIdx(line, colidx: Integer): Integer;
        procedure SetValByColIdx(line, colidx: Integer; const Value: String);
        procedure SetIntValByColIdx(line, colidx: Integer;
            const Value: Integer);
        function GetFltVals(line: Integer; const column: String): Extended;
        procedure SetFltVals(line: Integer; const column: String;
            const Value: Extended);
        function GetFltValByColIdx(line, colidx: Integer): Extended;
        procedure SetFltValByColIdx(line, colidx: Integer;
            const Value: Extended);
        function GetSValue(line: Integer; const column: String;
            var fast: TFastIndex): String;
        procedure SetSValue(line: Integer; const column: String;
            var fast: TFastIndex; const Value: String);
        function GetIValue(line: Integer; const column: String;
            var fast: TFastIndex): Integer;
        procedure SetIValue(line: Integer; const column: String;
            var fast: TFastIndex; const Value: Integer);
        function GetFValue(line: Integer; const column: String;
            var fast: TFastIndex): Extended;
        procedure SetFValue(line: Integer; const column: String;
            var fast: TFastIndex; const Value: Extended);
    protected
        Caches: TList;
        FloatLists: TList;
        FColumnNameCache: array[0..65535] of SmallInt;
        FColumnNameCache_Ready: Boolean;
    public
        constructor Create;
        destructor Destroy; override;
        procedure Clear;
        procedure Assign(line: Integer; Q: TQuery); overload;
        procedure AssignBase(line: Integer; Q: DBTables.TQuery);
        procedure Assign(line: Integer; T: TFastTable; T_line: Integer); overload;
        procedure AssignFirst(line: Integer; Q: TQuery); overload;
        procedure AssignFast(line: Integer; Q: TQuery);
        procedure Assign(Q: TQuery); overload;
        property Values[line: Integer; const column: String]: String
        read GetValues write SetValues; default;
        property SS[line: Integer; const column: String; var fast: TFastIndex]: String
        read GetSValue write SetSValue;
        property ValByColIdx[line, colidx: Integer]: String read GetValByColIdx
        write SetValByColIdx;
        property IntVal[line: Integer; const column: String]: Integer
        read GetIntVal write SetIntVal;
        property II[line: Integer; const column: String; var fast: TFastIndex]: Integer
        read GetIValue write SetIValue;
        property IntValByColIdx[line: Integer; colidx: Integer]: Integer
        read GetIntValByColIdx write SetIntValByColIdx;
        property FltVal[line: Integer; const column: String]: Extended
        read GetFltVals write SetFltVals;
        property FF[line: Integer; const column: String; var fast: TFastIndex]: Extended
        read GetFValue write SetFValue;
        property FltValByColIdx[line: Integer; colidx: Integer]: Extended
        read GetFltValByColIdx write SetFltValByColIdx;
        property Count: Integer read GetCount;
        property Data[i: Integer]: Pointer read GetData write SetData;
        property AutoAllocateColumns: Boolean read FAutoAllocateColumns
            write FAutoAllocateColumns;
        function IndexOfColumn(const column: String): Integer;
        procedure ResetBadRefs;
    public
        procedure SaveStatsToFile(const Filename: String; Append: Boolean);
        procedure SaveStatsToStream(Stream: TStream);
        procedure Delete(idx: Integer);
        procedure SwapLines(i1, i2: Integer);
    protected
        ListOfFastIndeces: TList;
        function FirstChar(const s: String): Char;
    public
        procedure ReleaseIndeces(const ListOfReleasingIndeces: array of PTFastIndex);
        function LineAsString(idx: Integer; ColumnNames: Boolean = TRUE): String;
    end;

    TFastIndex = class(TObject)
    protected
        fColIndex: Integer;
    end;

implementation

function CheckSum16(const S: String): Word;
asm
  xchg ecx, eax
  xor  eax, eax
  JECXZ @@exit
  mov  edx, ecx
@@loop:
  movzx ecx, word ptr [edx]
  cmp  cl, 0
  jz   @@exit
  inc  edx
  inc  edx
  ror  ax, 1
  xor  ax, cx
  //and  cl, ch
  cmp  ch, 0
  jnz  @@loop
@@exit:
  and  eax, $7FFF
end;

{ TFastTable }

procedure TFastTable.Assign(line: Integer; Q: TQuery);
var i: Integer;
begin
    while line >= Lines.Count do
    begin
        if Lines.Count >= Lines.Capacity then
            Lines.Capacity := Max(1024, (Lines.Count + 1) * 2);
        Lines.Add(TStringList.Create);
        LineObjects.Add(nil);
    end;
    for i := 0 to Q.FieldCount - 1 do
    begin
        if not Q.NullFieldN[i] then
            CASE
                {$IFDEF USE_NEWQUERY} // ODBC_DIRECT}
            Q.FieldTypes[I]
                {$ELSE}
            Q.FieldDefs[I].DataType
                {$ENDIF}
            OF
                ftCurrency, ftBCD: FltVal[line,
                    {$IFDEF USE_NEWQUERY} Q.FieldNames[i]
                    {$ELSE} Q.FieldDefs[i].Name
                    {$ENDIF} ] :=
                    {$IFDEF USE_NEWQUERY} // ODBC_DIRECT}
                    Q.DFieldN[I];
                {$ELSE}
                    Q.Fields[I].AsFloat;
                {$ENDIF}
                ftDate:
                    begin
                        FltVal[line, Q.FieldNames[i]] :=
                            {$IFDEF USE_NEWQUERY} // ODBC_DIRECT}
                        Q.FFieldN[I];
                        {$ELSE}
                        Q.Fields[I].AsFloat;
                        {$ENDIF}
                        Values[line, Q.FieldNames[i]] :=
                            {$IFDEF USE_NEWQUERY} // ODBC_DIRECT}
                        FormatDateTime('dd.mm.yyyyy', Q.DFieldN[I]);
                        {$ELSE}
                        FormatDateTime('dd.mm.yyyyy', Q.Fields[I].AsDateTime);
                        {$ENDIF}
                    end;
                ftTime: Values[line, Q.FieldNames[i]] :=
                    {$IFDEF USE_NEWQUERY} // ODBC_DIRECT}
                    FormatDateTime('hh:nn:ss', Q.DFieldN[I]);
                {$ELSE}
                    FormatDateTime('hh:nn:ss', Q.Fields[I].AsDateTime);
                {$ENDIF}
                ftDateTime:
                    begin
                        FltVal[line, Q.FieldNames[i]] :=
                            {$IFDEF USE_NEWQUERY} // ODBC_DIRECT}
                        Q.FFieldN[I];
                        {$ELSE}
                        Q.Fields[I].AsFloat;
                        {$ENDIF}
                        if Trunc(Q.DFieldN[i]) = Q.DFieldN[i] then
                            Values[line, Q.FieldNames[i]] :=
                                {$IFDEF USE_NEWQUERY} // ODBC_DIRECT}
                            FormatDateTime('dd.mm.yyyyy', Q.FFieldN[I])
                                {$ELSE}
                            FormatDateTime('dd.mm.yyyyy', Q.Fields[I].AsDateTime)
                                {$ENDIF}
                        else
                            Values[line, Q.FieldNames[i]] :=
                                {$IFDEF USE_NEWQUERY} // ODBC_DIRECT}
                            FormatDateTime('dd.mm.yyyyy hh:nn:ss', Q.DFieldN[I])
                                {$ELSE}
                            FormatDateTime('dd.mm.yyyyy hh:nn:ss', Q.Fields[I].AsDateTime)
                                {$ENDIF}
                            ;
                    end;
                ftInteger, ftSmallint, ftWord, ftLargeint, ftAutoInc:
                    IntVal[line, Q.FieldNames[i]] :=
                        {$IFDEF USE_NEWQUERY} // ODBC_DIRECT}
                    Q.IFieldN[I];
                {$ELSE}
                    Q.Fields[I].AsInteger;
                {$ENDIF}
                ftBoolean:
                    IntVal[line, Q.FieldNames[i]] :=
                        {$IFDEF USE_NEWQUERY} // ODBC_DIRECT}
                    Q.IFieldN[I];
                {$ELSE}
                    Integer(Q.Fields[I].AsBoolean);
                {$ENDIF}
                //ftCurrency,
                ftFloat: FltVal[line, Q.FieldNames[i]] :=
                    {$IFDEF USE_NEWQUERY} // ODBC_DIRECT}
                    Q.FFieldN[I];
                {$ELSE}
                    Q.Fields[I].AsFloat;
                {$ENDIF}
            else Values[line, Q.FieldNames[i]] :=
                {$IFDEF USE_NEWQUERY} // ODBC_DIRECT}
                Q.SFieldN[I];
                {$ELSE}
                Q.Fields[I].AsString;
                {$ENDIF}
            END
        else
            CASE
                {$IFDEF USE_NEWQUERY} // ODBC_DIRECT}
            Q.FieldTypes[I]
                {$ELSE}
            Q.FieldDefs[I].DataType
                {$ENDIF}
            OF
                ftCurrency, ftBCD, ftDate, ftTime, ftDateTime,
                    ftFloat: FltVal[line, Q.FieldNames[i]] := 0.0;
                ftInteger, ftSmallint, ftWord, ftLargeint, ftAutoInc:
                    IntVal[line, Q.FieldNames[i]] := 0;
                ftBoolean: IntVal[line, Q.FieldNames[i]] := 0;
            else Values[line, Q.FieldNames[i]] := '';
            END;
    end;
end;

procedure TFastTable.Assign(Q: TQuery);
var i: Integer;
    wasAuto: Boolean;
begin
    i := Count - 1;
    wasAuto := FAutoAllocateColumns;
    FAutoAllocateColumns := TRUE;
    {if  (Count = 0) and (Columns.Count = 0) and not Q.Eof then
    begin
        AssignFirst( 0, Q );
        REPEAT
            Q.Next;
            if  not Q.Eof then
                AssignFast( Count, Q );
        UNTIL Q.Eof;
    end
      else}
    begin
        while not Q.Eof do
        begin
            inc(i);
            Assign(i, Q);
            Q.Next;
        end;
    end;
    FAutoAllocateColumns := wasAuto;
end;

procedure TFastTable.Assign(line: Integer; T: TFastTable; T_line: Integer);
var i: Integer;
    LineList: TStringList;
begin
    while line >= Lines.Count do
    begin
        if Lines.Count >= Lines.Capacity then
            Lines.Capacity := Max(1024, (Lines.Count + 1) * 2);
        Lines.Add(TStringList.Create);
        LineObjects.Add(nil);
    end;
    for i := 0 to T.Columns.Count - 1 do
    begin
        Values[line, T.Columns[i]] := T.ValByColIdx[T_line, i];
        if (T.FloatLists.Count > i) and
            (T.FloatLists[i] <> nil) and
            (FirstChar(T.GetValByColIdx_StrOnly(T_line, i)) = #2) then
        TRY
            FltVal[line, T.Columns[i]] := T.FltValByColIdx[T_line, i]
        EXCEPT
        END
        else
            if (T.Lines.Count > T_line) then
            begin
                LineList := T.Lines[T_line];
                if LineList.Count > i then
                begin
                    if (Integer(LineList.Objects[i]) <> 0)
                        or (IntToStr(Integer(LineList.Objects[i])) = T.ValByColIdx[T_line, i]) then
                        IntVal[line, T.Columns[i]] := Integer(LineList.Objects[i]);
                end;
            end;
    end;
end;

procedure TFastTable.AssignBase(line: Integer; Q: DBTables.TQuery);
var i: Integer;
begin
    while line >= Lines.Count do
    begin
        if Lines.Count >= Lines.Capacity then
            Lines.Capacity := Max(1024, (Lines.Count + 1) * 2);
        Lines.Add(TStringList.Create);
        LineObjects.Add(nil);
    end;
    for i := 0 to Q.FieldCount - 1 do
    begin
        if not Q.Fields[i].isNull then
            CASE Q.FieldDefs[I].DataType
            OF
                ftCurrency, ftBCD: FltVal[line, Q.FieldDefs[i].Name] :=
                   Q.Fields[I].AsFloat;
                ftDate:
                    begin
                        FltVal[line, Q.Fields[i].Name] :=
                            Q.Fields[I].AsFloat;
                        Values[line, Q.Fields[i].Name] :=
                            FormatDateTime('dd.mm.yyyyy', Q.Fields[I].AsDateTime);
                    end;
                ftTime: Values[line, Q.Fields[i].Name] :=
                    FormatDateTime('hh:nn:ss', Q.Fields[I].AsDateTime);
                ftDateTime:
                    begin
                        FltVal[line, Q.Fields[i].Name] :=
                            Q.Fields[I].AsFloat;
                        if Trunc(Q.Fields[i].AsDateTime) = Q.Fields[i].AsDateTime
                        then
                            Values[line, Q.Fields[i].Name] :=
                                FormatDateTime('dd.mm.yyyyy', Q.Fields[I].AsDateTime)
                        else
                            Values[line, Q.Fields[i].Name] :=
                                FormatDateTime('dd.mm.yyyyy hh:nn:ss', Q.Fields[I].AsDateTime)
                            ;
                    end;
                ftInteger, ftSmallint, ftWord, ftLargeint, ftAutoInc:
                    IntVal[line, Q.Fields[i].Name] :=
                        Q.Fields[I].AsInteger;
                ftBoolean:
                    IntVal[line, Q.Fields[i].Name] :=
                        Integer(Q.Fields[I].AsBoolean);
                //ftCurrency,
                ftFloat: FltVal[line, Q.Fields[i].Name] :=
                    Q.Fields[I].AsFloat;
            else Values[line, Q.Fields[i].Name] :=
                Q.Fields[I].AsString;
            END
        else
            CASE
                Q.FieldDefs[I].DataType
            OF
                ftCurrency, ftBCD, ftDate, ftTime, ftDateTime,
                    ftFloat: FltVal[line, Q.Fields[i].Name] := 0.0;
                ftInteger, ftSmallint, ftWord, ftLargeint, ftAutoInc:
                    IntVal[line, Q.Fields[i].Name] := 0;
                ftBoolean: IntVal[line, Q.Fields[i].Name] := 0;
            else Values[line, Q.Fields[i].Name] := '';
            END;
    end;
end;

procedure TFastTable.AssignFast(line: Integer; Q: TQuery);
var i: Integer;
begin
    while line >= Lines.Count do
    begin
        if Lines.Count >= Lines.Capacity then
            Lines.Capacity := Max(1024, (Lines.Count + 1) * 2);
        Lines.Add(TStringList.Create);
        LineObjects.Add(nil);
    end;
    for i := 0 to Q.FieldCount - 1 do
    begin
        if not Q.NullFieldN[i] then
            CASE
                {$IFDEF USE_NEWQUERY} // ODBC_DIRECT}
            Q.FieldTypes[I]
                {$ELSE}
            Q.FieldDefs[I].DataType
                {$ENDIF}
            OF
                ftCurrency: ValByColIdx[line, i] :=
                    {$IFDEF USE_NEWQUERY} // ODBC_DIRECT}
                    Format('%.2f', [Q.FFieldN[I]]);
                {$ELSE}
                    Format('%.2f', [Q.Fields[I].AsFloat]);
                {$ENDIF}
                ftDate: ValByColIdx[line, i] :=
                    {$IFDEF USE_NEWQUERY} // ODBC_DIRECT}
                    FormatDateTime('dd.mm.yyyyy', Q.FFieldN[I]);
                {$ELSE}
                    FormatDateTime('dd.mm.yyyyy', Q.Fields[I].AsDateTime);
                {$ENDIF}
                ftTime: ValByColIdx[line, i] :=
                    {$IFDEF USE_NEWQUERY} // ODBC_DIRECT}
                    FormatDateTime('hh:nn:ss', Q.DFieldN[I]);
                {$ELSE}
                    FormatDateTime('hh:nn:ss', Q.Fields[I].AsDateTime);
                {$ENDIF}
                ftDateTime:
                    if Trunc(Q.DFieldN[i]) = Q.DFieldN[i] then
                        ValByColIdx[line, i] :=
                            {$IFDEF USE_NEWQUERY} // ODBC_DIRECT}
                        FormatDateTime('dd.mm.yyyyy', Q.FFieldN[I])
                            {$ELSE}
                        FormatDateTime('dd.mm.yyyyy', Q.Fields[I].AsDateTime)
                            {$ENDIF}
                    else
                        ValByColIdx[line, i] :=
                            {$IFDEF USE_NEWQUERY} // ODBC_DIRECT}
                        FormatDateTime('dd.mm.yyyyy hh:nn:ss', Q.DFieldN[I])
                            {$ELSE}
                        FormatDateTime('dd.mm.yyyyy hh:nn:ss', Q.Fields[I].AsDateTime)
                            {$ENDIF}
                        ;
                ftInteger, ftSmallint, ftWord, ftLargeint, ftAutoInc:
                    IntValByColIdx[line, i] :=
                        {$IFDEF USE_NEWQUERY} // ODBC_DIRECT}
                    Q.IFieldN[I];
                {$ELSE}
                    Q.Fields[I].AsInteger;
                {$ENDIF}
                ftFloat: FltValByColIdx[line, i] :=
                    {$IFDEF USE_NEWQUERY} // ODBC_DIRECT}
                    Q.FFieldN[I];
                {$ELSE}
                    Q.Fields[I].AsFloat;
                {$ENDIF}
            else ValByColIdx[line, i] :=
                {$IFDEF USE_NEWQUERY} // ODBC_DIRECT}
                Q.SFieldN[I];
                {$ELSE}
                Q.Fields[I].AsString;
                {$ENDIF}
            END
        else
            ValByColIdx[line, i] := '';
    end;
end;

procedure TFastTable.AssignFirst(line: Integer; Q: TQuery);
var i: Integer;
    j: Integer;
    //LineList: TStringList;
begin
    while line >= Lines.Count do
    begin
        if Lines.Count >= Lines.Capacity then
            Lines.Capacity := Max(1024, (Lines.Count + 1) * 2);
        Lines.Add(TStringList.Create);
        LineObjects.Add(nil);
    end;
    for i := 0 to Q.FieldCount - 1 do
    begin
        j := Columns.Add(Q.FieldNames[i]);
        if not Q.NullFieldN[i] then
            CASE
                {$IFDEF USE_NEWQUERY} // ODBC_DIRECT}
            Q.FieldTypes[I]
                {$ELSE}
            Q.FieldDefs[I].DataType
                {$ENDIF}
            OF
                ftCurrency: ValByColIdx[line, j] :=
                    {$IFDEF USE_NEWQUERY} // ODBC_DIRECT}
                    Format('%.2f', [Q.DFieldN[I]]);
                {$ELSE}
                    Format('%.2f', [Q.Fields[I].AsFloat]);
                {$ENDIF}
                ftDate: ValByColIdx[line, j] :=
                    {$IFDEF USE_NEWQUERY} // ODBC_DIRECT}
                    FormatDateTime('dd.mm.yyyyy', Q.DFieldN[I]);
                {$ELSE}
                    FormatDateTime('dd.mm.yyyyy', Q.Fields[I].AsDateTime);
                {$ENDIF}
                ftTime: ValByColIdx[line, j] :=
                    {$IFDEF USE_NEWQUERY} // ODBC_DIRECT}
                    FormatDateTime('hh:nn:ss', Q.DFieldN[I]);
                {$ELSE}
                    FormatDateTime('hh:nn:ss', Q.Fields[I].AsDateTime);
                {$ENDIF}
                ftInteger, ftSmallint, ftWord, ftLargeint, ftAutoInc:
                    IntValByColIdx[line, j] :=
                        {$IFDEF USE_NEWQUERY} // ODBC_DIRECT}
                    Q.IFieldN[I];
                {$ELSE}
                    Q.Fields[I].AsInteger;
                {$ENDIF}
                ftFloat: FltValByColIdx[line, j] :=
                    {$IFDEF USE_NEWQUERY} // ODBC_DIRECT}
                    Q.FFieldN[I];
                {$ELSE}
                    Q.Fields[I].AsFloat;
                {$ENDIF}
            else ValByColIdx[line, j] :=
                {$IFDEF USE_NEWQUERY} // ODBC_DIRECT}
                Q.SFieldN[I];
                {$ELSE}
                Q.Fields[I].AsString;
                {$ENDIF}
            END
        else
            ValByColIdx[line, j] := '';
    end;
    ResetBadRefs;
end;

procedure TFastTable.Clear;
var i: Integer;
begin
    FColumnNameCache_Ready := FALSE;
    Columns.Clear;
    for i := Lines.Count - 1 downto 0 do
        TObject(Lines[i]).Free;
    Lines.Clear;
    for i := 0 to Caches.Count - 1 do
        TObject(Caches[i]).Free;
    Caches.Clear;
    for i := 0 to FloatLists.Count - 1 do
        TObject(FloatLists[i]).Free;
    FloatLists.Clear;
    ReleaseIndeces([]);
end;

constructor TFastTable.Create;
begin
    inherited;
    Columns := TStringList.Create;
    Lines := TList.Create;
    LineObjects := TList.Create;
    Caches := TList.Create;
    FloatLists := TList.Create;
    AutoAllocateColumns := TRUE;
    ListOfFastIndeces := TList.Create;
end;

procedure TFastTable.Delete(idx: Integer);
var i: Integer;
begin
    if idx < 0 then Exit;
    if idx >= Count then Exit;
    Lines.Delete(idx);
    LineObjects.Delete(idx);
    for i := 0 to FloatLists.Count - 1 do
    begin
        if FloatLists[i] <> nil then
        begin
            if TFloatList(FloatLists[i]).Count > idx then
                TFloatList(FloatLists[i]).Delete(idx);
        end;
    end;
end;

destructor TFastTable.Destroy;
begin
    Clear;
    Columns.Free;
    Lines.Free;
    LineObjects.Free;
    Caches.Free;
    FloatLists.Free;
    ListOfFastIndeces.Free;
    inherited;
end;

function TFastTable.FirstChar(const s: String): Char;
begin
    if s = '' then Result := #0
    else Result := s[1];
end;

function TFastTable.GetCount: Integer;
begin
    Result := Lines.Count;
end;

function TFastTable.GetData(i: Integer): Pointer;
begin
    Result := LineObjects[i];
end;

function TFastTable.GetFltValByColIdx(line, colidx: Integer): Extended;
var j: Integer;
    F: TFloatList;
begin
    j := colidx;
    Result := 0;
    if j < 0 then Exit;
    if j >= FloatLists.Count then Exit;
    F := FloatLists[j];
    Result := F[line];
end;

function TFastTable.GetFltVals(line: Integer;
    const column: String): Extended;
var j, i, k: Integer;
    F: TFloatList;
    s: String;
begin
    j := IndexOfColumn(column);
    Result := 0;
    if j < 0 then
    begin
        asm
      int 3
        end;
        Exit;
    end;
    if j >= FloatLists.Count then Exit;
    F := FloatLists[j];
    if F <> nil then
        Result := F[line]
    else
    begin
        s := GetValByColIdx(line, j);
        TRY
            k := 0;
            for i := 1 to Length(s) do
                if s[i] = '.' then
                    inc(k);
            if k = 2 then
                Result := StrToDate(s)
            else
            begin
                for i := 1 to Length(s) do
                    if s[i] in ['.', ','] then
                        s[i] := DecimalSeparator;
                Result := StrToFloat(s);
            end;
            //SetFltValByColIdx( line, j, Result );
        EXCEPT
            Result := 0;
        END;
    end;
end;

function TFastTable.GetFValue(line: Integer; const column: String;
    var fast: TFastIndex): Extended;
var j, i, k: Integer;
    F: TFloatList;
    s: String;
begin
    if fast <> nil then
        j := fast.fColIndex
    else
    begin
        j := IndexOfColumn(column);
        if j >= 0 then
        begin
            fast := TFastIndex.Create;
            ListOfFastIndeces.Add(@fast);
            fast.fColIndex := j;
        end;
    end;
    Result := 0;
    if j < 0 then
    begin
        {asm
          int 3
        end;}
        Exit;
    end;
    if j >= FloatLists.Count then
        F := nil
    else
        F := FloatLists[j];
    if F <> nil then
        Result := F[line]
    else
    begin
        s := GetValByColIdx(line, j);
        TRY
            k := 0;
            for i := 1 to Length(s) do
                if s[i] = '.' then
                    inc(k);
            if k = 2 then
                Result := StrToDate(s)
            else
            begin
                for i := 1 to Length(s) do
                    if s[i] in ['.', ','] then
                        s[i] := DecimalSeparator;
                Result := StrToFloat(s);
            end;
            //SetFltValByColIdx( line, j, Result );
        EXCEPT
            Result := 0;
        END;
    end;
end;

function TFastTable.GetIntVal(line: Integer;
    const column: String): Integer;
var LineList: TStringList;
    j: Integer;
begin
    LineList := Lines[line];
    j := IndexOfColumn(column);
    Result := 0;
    if j < 0 then
    begin
        asm
      //int 3
        end;
        Exit;
    end;
    if j >= LineList.Count then Exit;
    Result := Integer(LineList.Objects[j]);
end;

function TFastTable.GetIntValByColIdx(line, colidx: Integer): Integer;
var LineList: TStringList;
    j: Integer;
begin
    LineList := Lines[line];
    j := colidx;
    Result := 0;
    if j < 0 then Exit;
    if j >= LineList.Count then Exit;
    Result := Integer(LineList.Objects[j]);
end;

function TFastTable.GetIValue(line: Integer; const column: String;
    var fast: TFastIndex): Integer;
var LineList: TStringList;
    j: Integer;
    //s: String;
begin
    LineList := Lines[line];
    if fast <> nil then
        j := fast.fColIndex
    else
        j := IndexOfColumn(column);
    Result := 0;
    if j < 0 then
    begin
        asm
      //int 3
        end;
        Exit;
    end
    else
        if fast = nil then
        begin
            fast := TFastIndex.Create;
            ListOfFastIndeces.Add(@fast);
            fast.fColIndex := j;
        end;
    if j >= LineList.Count then Exit;
    Result := Integer(LineList.Objects[j]);
    {if  Result = 0 then
    begin
        s := LineList[j];
        if  (s <> '') and (s[1] <> #1) then
        begin
            TRY
                Result := StrToInt( s );
            EXCEPT
                Result := 0;
            END;
        end;
    end;}
end;

function TFastTable.GetSValue(line: Integer; const column: String;
    var fast: TFastIndex): String;
var LineList: TStringList;
    j: Integer;
begin
    Result := '';
    if (Lines.Count <= line) or (line < 0) then
        Exit; {>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>}
    LineList := Lines[line];
    if fast <> nil then
    begin
        j := fast.fColIndex;
        if j < 0 then Exit;
        if j >= LineList.Count then Exit;
    end
    else
    begin
        j := IndexOfColumn(column);
        if j < 0 then Exit;
        if j >= LineList.Count then Exit;
        fast := TFastIndex.Create;
        ListOfFastIndeces.Add(@fast);
        fast.fColIndex := j;
    end;
    Result := LineList[j];
    if Result = ''#1 then
    begin
        j := GetIntValByColIdx(line, j);
        Result := IntToStr(j);
    end
    else
        if Result = ''#2 then
        begin
            Result := FloatToStr(GetFltValByColIdx(line, j));
        end;
end;

function TFastTable.GetValByColIdx(line, colidx: Integer): String;
var LineList: TStringList;
    j: Integer;
begin
    Result := '';
    if line >= Count then Exit;
    LineList := Lines[line];
    j := colidx;
    if j < 0 then Exit;
    if j >= LineList.Count then Exit;
    Result := LineList[j];
    //if Result = ''#1 then
    if (Result <> '') then
    begin
        if Result[1] = #1 then
        begin
            j := GetIntValByColIdx(line, colidx);
            Result := IntToStr(j);
        end
        else
            if Result[1] = #2 then
            begin
                Result := FloatToStr(GetFltValByColIdx(line, colidx));
            end;
    end;
end;

function TFastTable.GetValByColIdx_StrOnly(line, colidx: Integer): String;
var LineList: TStringList;
    j: Integer;
begin
    Result := '';
    if line >= Count then Exit;
    LineList := Lines[line];
    j := colidx;
    if j < 0 then Exit;
    if j >= LineList.Count then Exit;
    Result := LineList[j];
end;

function TFastTable.GetValues(line: Integer; const column: String): String;
var LineList: TStringList;
    j: Integer;
begin
    LineList := Lines[line];
    j := IndexOfColumn(column);
    Result := '';
    if j < 0 then Exit;
    if j >= LineList.Count then Exit;
    Result := LineList[j];
    if Result = ''#1 then
    begin
        j := GetIntValByColIdx(line, j);
        //if j <> 0 then
        Result := IntToStr(j);
    end
    else
        if Result = ''#2 then
        begin
            Result := FloatToStr(GetFltValByColIdx(line, j));
        end;
end;

function TFastTable.IndexOfColumn(const column: String): Integer;
var C: Word;
    i: Integer;
begin
    C := CheckSum16(column);
    if not FColumnNameCache_Ready then
    begin
        FillChar(FColumnNameCache, Sizeof(FColumnNameCache), 0);
        for i := 0 to Columns.Count - 1 do
            FColumnNameCache[CheckSum16(Columns[i])] := i + 1;
        FColumnNameCache_Ready := TRUE;
    end;
    Result := FColumnNameCache[C] - 1;
    if Result = -3 then
    begin
        Result := -1;
        Exit;
    end;
    if (Result >= 0) and (Columns[Result] = column) then Exit;
    for i := 0 to Columns.Count - 1 do
        if AnsiCompareText(column, Columns[i]) = 0 then
        begin
            if Result = -1 then
                FColumnNameCache[C] := i + 1;
            Result := i;
            Exit;
        end;
    if Result = -1 then
        FColumnNameCache[C] := -2;
    Result := -1;
end;

function TFastTable.LineAsString(idx: Integer;
    ColumnNames: Boolean): String;
var i: Integer;
begin
    Result := '';
    for i := 0 to Columns.Count - 1 do
    begin
        if Result <> '' then Result := Result + '; ';
        if ColumnNames then
            Result := Result + Columns[i] + '=';
        Result := Result + ValByColIdx[idx, i];
        if AnsiSameText(Copy(Columns[i], 1, 4), '')
            or AnsiSameText(Columns[i], '') then
        TRY
            Result := Result + '(' + FormatDateTime('dd.mm.yyyy',
                StrToFloat(ValByColIdx[idx, i])) + ')';
        EXCEPT
        END;
    end;
end;

procedure TFastTable.ReleaseIndeces(
    const ListOfReleasingIndeces: array of PTFastIndex);
var i, j: Integer;
    fast: ^TFastIndex;
    found: Boolean;
begin
    if Length(ListOfReleasingIndeces) = 0 then
    begin
        for i := 0 to ListOfFastIndeces.Count - 1 do
        begin
            fast := ListOfFastIndeces[i];
            fast^.Free;
            fast^ := nil;
        end;
        ListOfFastIndeces.Clear;
    end
    else
        for i := 0 to High(ListOfReleasingIndeces) do
        begin
            PTFastIndex(fast) := ListOfReleasingIndeces[i];
            //ListOfFastIndeces.Remove( fast );
            found := FALSE;
            for j := ListOfFastIndeces.Count - 1 downto 0 do
                if ListOfFastIndeces[j] = fast then
                begin
                    ListOfFastIndeces.Delete(j);
                    found := TRUE;
                end;
            if not found then
            begin
                asm
                nop
                end;
            end;
            fast^.Free;
            fast^ := nil;
        end;
end;

procedure TFastTable.ResetBadRefs;
var i: Integer;
begin
    for i := 0 to High(FColumnNameCache) do
        if FColumnNameCache[i] < 0 then
            FColumnNameCache[i] := 0;
end;

procedure TFastTable.SaveStatsToFile(const Filename: String; Append: Boolean);
var F: TStream;
begin
    if FileExists(Filename) then
        F := TFileStream.Create(Filename, fmOpenReadWrite)
    else
        F := TFileStream.Create(Filename, fmCreate);
    if Append then
        F.Seek(0, soFromEnd);
    SaveStatsToStream(F);
    F.Free;
end;

procedure TFastTable.SaveStatsToStream(Stream: TStream);
var SL: TStringList;
    i, c, cap: Integer;
begin
    SL := TStringList.Create;
    TRY
        c := 0;
        cap := 0;
        for i := 0 to Caches.Count - 1 do
        begin
            c := c + TStringList(Caches[i]).Count;
            cap := cap + TStringList(Caches[i]).Capacity;
        end;
        SL.Add('*** Count=' + IntToStr(Count) +
            ' Columns.Count=' + IntToStr(Columns.Count) +
            ' Cache.Count=' + IntToStr(c) +
            ' Cache.Capacity=' + IntToStr(cap));
        for c := 0 to Caches.Count - 1 do
            for i := 0 to TStringList(Caches[c]).Count - 1 do
            begin
                SL.Add(IntToHex(DWORD(TStrings(Caches[c]).Objects[i]), 8) + ' ' +
                    TStringList(Caches[c])[i]);
            end;
        SL.SaveToStream(Stream);
    FINALLY
        SL.Free;
    END;
end;

procedure TFastTable.SetData(i: Integer; const Value: Pointer);
begin
    LineObjects[i] := Value;
end;

procedure TFastTable.SetFltValByColIdx(line, colidx: Integer;
    const Value: Extended);
var LineList: TStringList;
    j: Integer;
    F: TFloatList;
begin
    LineList := Lines[line];
    j := colidx;
    if j < 0 then Exit;
    while j >= LineList.Count do
        LineList.Add('');
    while j >= FloatLists.Count do
        FloatLists.Add(nil);
    if FloatLists[j] = nil then
        FloatLists[j] := TFloatList.Create;
    F := FloatLists[j];
    F[line] := Value;
    LineList[j] := ''#2;
end;

procedure TFastTable.SetFltVals(line: Integer; const column: String;
    const Value: Extended);
var LineList: TStringList;
    j: Integer;
    F: TFloatList;
begin
    SetValues(line, column, ''#2);
    LineList := Lines[line];
    j := IndexOfColumn(column);
    if j < 0 then Exit;
    while j >= LineList.Count do
        LineList.Add('');
    while j >= FloatLists.Count do
        FloatLists.Add(nil);
    if FloatLists[j] = nil then
        FloatLists[j] := TFloatList.Create;
    F := FloatLists[j];
    while F.Count <= line do
        F.Add(0);
    F[line] := Value;
end;

procedure TFastTable.SetFValue(line: Integer; const column: String;
    var fast: TFastIndex; const Value: Extended);
var LineList: TStringList;
    j: Integer;
    F: TFloatList;
begin
    SetSValue(line, column, fast, ''#2);
    if fast = nil then Exit;
    LineList := Lines[line];
    j := fast.fColIndex; //IndexOfColumn( column );
    //if j < 0 then Exit;
    while j >= LineList.Count do
        LineList.Add('');
    while j >= FloatLists.Count do
        FloatLists.Add(nil);
    if FloatLists[j] = nil then
        FloatLists[j] := TFloatList.Create;
    F := FloatLists[j];
    while F.Count <= line do
        F.Add(0);
    F[line] := Value;
end;

procedure TFastTable.SetIntVal(line: Integer; const column: String;
    const Value: Integer);
var LineList: TStringList;
    j: Integer;
begin
    SetValues(line, column, ''#1);
    LineList := Lines[line];
    j := IndexOfColumn(column);
    if j < 0 then Exit;
    while j >= LineList.Count do
        LineList.Add('');
    LineList.Objects[j] := Pointer(Value);
end;

procedure TFastTable.SetIntValByColIdx(line, colidx: Integer;
    const Value: Integer);
var LineList: TStringList;
    j: Integer;
begin
    LineList := Lines[line];
    j := colidx;
    if j < 0 then Exit;
    //if j >= LineList.Count then Exit;
    while j >= LineList.Count do
        LineList.Add('');
    LineList.Objects[j] := Pointer(Value);
    LineList[j] := ''#1;
end;

procedure TFastTable.SetIValue(line: Integer; const column: String;
    var fast: TFastIndex; const Value: Integer);
var LineList: TStringList;
    j: Integer;
begin
    SetSValue(line, column, fast, ''#1);
    if fast = nil then Exit;
    LineList := Lines[line];
    j := fast.fColIndex; //IndexOfColumn( column );
    //if j < 0 then Exit;
    while j >= LineList.Count do
        LineList.Add('');
    LineList.Objects[j] := Pointer(Value);
end;

procedure TFastTable.SetSValue(line: Integer; const column: String;
    var fast: TFastIndex; const Value: String);
var j_col: Integer;
begin
    while line >= Count do
    begin
        Lines.Add(TStringList.Create);
        LineObjects.Add(nil);
    end;

    if fast <> nil then
        j_col := fast.fColIndex
    else
        j_col := IndexOfColumn(column);
    if j_col < 0 then
    begin
        if not FAutoAllocateColumns then Exit;
        j_col := Columns.Add(column);
        ResetBadRefs;
    end;
    if fast = nil then
    begin
        fast := TFastIndex.Create;
        ListOfFastIndeces.Add(@fast);
        fast.fColIndex := j_col;
    end;
    SetValByColIdx(line, j_col, Value);
end;

procedure TFastTable.SetValByColIdx(line, colidx: Integer;
    const Value: String);
var LineList: TStringList;
    i, j, k, m, j_col: Integer;
    crc: DWORD;
    S: String;
    Cache: TStringList;
begin
    while line >= Count do
    begin
        Lines.Add(TStringList.Create);
        LineObjects.Add(nil);
    end;

    j_col := colidx;

    while j_col >= Caches.Count do
        Caches.Add(TStringList.Create);
    Cache := Caches[j_col];

    S := Value;
    if S = '' then
    else
        if (line > 0) and (GetValByColIdx_StrOnly(line - 1, j_col) = S) then
            S := GetValByColIdx_StrOnly(line - 1, j_col)
        else
        begin

            crc := 0;
            for i := 1 to Length(Value) do
                crc := (crc shr 10 or crc shl 6) xor Ord(Value[i]);
            k := -1;
            i := 0;
            j := Cache.Count - 1;
            while j - i > 10 do
            begin
                if DWORD(Cache.Objects[i]) = crc then
                    break;
                m := (i + j) div 2;
                if DWORD(Cache.Objects[m]) = crc then
                begin
                    i := m;
                    break;
                end;
                if crc < DWORD(Cache.Objects[m]) then
                    j := m
                else
                    i := m;
            end;
            while (i > 0) and (Cache.Objects[i - 1] = Cache.Objects[i]) do
                dec(i);
            for i := i to Cache.Count - 1 do
            begin
                if (DWORD(Cache.Objects[i]) = crc) and (Cache[i] = Value) then
                begin
                    k := i; break;
                end
                else
                    if (DWORD(Cache.Objects[i]) > crc) then
                    begin
                        k := i;
                        if Cache.Capacity <= Cache.Count then
                            Cache.Capacity := Max(1024, Cache.Count * 2);
                        Cache.InsertObject(k, Value, Pointer(crc));
                        break;
                    end;
            end;
            if k < 0 then
            begin
                k := Cache.Count;
                if Cache.Capacity <= Cache.Count then
                    Cache.Capacity := Max(1024, Cache.Count * 2);
                Cache.AddObject(Value, Pointer(crc));
            end;

            S := Cache[k];

        end;

    LineList := Lines[line];
    if j_col >= LineList.Count then
        LineList.Capacity := j_col + 1; //LineList.Count+1;
    while j_col >= LineList.Count do
        LineList.Add('');
    LineList[j_col] := S;
end;

procedure TFastTable.SetValues(line: Integer; const column, Value: String);
var //LineList: TStringList;
    //i, j, k, m, j_col: Integer;
    //crc: DWORD;
    //S: String;
    //Cache: TStringList;
    j_col: Integer;
begin
    while line >= Count do
    begin
        Lines.Add(TStringList.Create);
        LineObjects.Add(nil);
    end;

    j_col := IndexOfColumn(column);
    if j_col < 0 then
    begin
        if not FAutoAllocateColumns then Exit;
        j_col := Columns.Add(column);
        ResetBadRefs;
    end;
    SetValByColIdx(line, j_col, Value);

    {while j_col >= Caches.Count do
      Caches.Add( TStringList.Create );
    Cache := Caches[ j_col ];

    S := Value;
    if S = '' then
    else
    if (line > 0) and (Values[ line-1, column ] = S) then
      S := Values[ line-1, column ]
    else
    begin

      crc := 0;
      for i := 1 to Length( Value ) do
        crc := ( crc shr 10 or crc shl 6 ) xor Ord( Value[ i ] );
      k := -1;
      i := 0;
      j := Cache.Count-1;
      while j - i > 10 do
      begin
        if DWORD( Cache.Objects[ i ] ) = crc then
          break;
        m := (i + j) div 2;
        if DWORD( Cache.Objects[ m ] ) = crc then
        begin
          i := m;
          break;
        end;
        if crc < DWORD( Cache.Objects[ m ] ) then
          j := m
        else
          i := m;
      end;
      while (i > 0) and ( Cache.Objects[ i-1 ] = Cache.Objects[ i ] ) do
        dec( i );
      for i := i to Cache.Count-1 do
      begin
        if (DWORD( Cache.Objects[ i ] ) = crc) and (Cache[ i ] = Value) then
        begin
          k := i; break;
        end
          else
        if (DWORD( Cache.Objects[ i ] ) > crc) then
        begin
          k := i;
          if Cache.Capacity <= Cache.Count then
            Cache.Capacity := Max( 1024, Cache.Count * 2 );
          Cache.InsertObject( k, Value, Pointer( crc ) );
          break;
        end;
      end;
      if k < 0 then
      begin
        k := Cache.Count;
        if Cache.Capacity <= Cache.Count then
          Cache.Capacity := Max( 1024, Cache.Count * 2 );
        Cache.AddObject( Value, Pointer( crc ) );
      end;

      S := Cache[ k ];

    end;

    LineList := Lines[ line ];
    if j_col >= LineList.Count then
      LineList.Capacity := Columns.Count;
    while j_col >= LineList.Count do
      LineList.Add( '' );
    LineList[ j_col ] := S;}
end;

procedure TFastTable.SwapLines(i1, i2: Integer);
var Tmp: Pointer;
    i: Integer;
    F: TFloatList;
    Ftmp: Extended;
begin
    Tmp := Lines[i1];
    Lines[i1] := Lines[i2];
    Lines[i2] := Tmp;
    if FloatLists <> nil then
        for i := 0 to FloatLists.Count - 1 do
        begin
            F := FloatLists[i];
            if F = nil then continue;
            if (F.Count > i1) or (F.Count > i2) then
            begin
                {if  F.Count <= i1 then
                    F[i1] := 0.0;
                if  F.Count <= i2 then
                    F[i2] := 0.0;}
                Ftmp := F[i1];
                F[i1] := F[i2];
                F[i2] := Ftmp;
            end;
        end;
end;

end.
