//**  ,     VCL Delphi.
//**  Vladimir Kladov, 2000-2010
//**
//**      .
//**     /     
//**  2          
//**   GNU,  Free Software Foundation.
//**
//**       ,    
//** ,       ,   
//**         
//**   .      
//**     GNU.
//**
//**          
//**   GNU.     ,     Free
//** Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
//** MA 02111-1307, USA.
//******************************************************************************

{-------------------------------------------------------------------------------
FastTreeView component for Delphi VCL (C) by Vladimir Kladov, 2005
      ,   TreeView -  
       TreeView  .
     ListView, ..   
,    .  :
+      (    listview),
+    ButtonImages (ImageList,    
   [+]  [-]),
+    (, ) ,  
+      .
-------------------------------------------------------------------------------}

unit FastTreeView;

interface

{$I VCL_VERSIONS.INC}

uses Windows, messages, Classes, SysUtils, comctrls, commctrl, controls, imglist,
    graphics, QSort, forms, stdctrls, utPhotoExc;

var DisableFrameAroundFocusedTreeView: Boolean = TRUE;

type
    TFastTreeView = class;

    TFastTreeItem = class
    private
        //FCount: Integer;
        FOwner: TFastTreeView;
        FItems: TList;
        FText: String;
        FImageIndex: Integer;
        FTotalCount: Integer;
        FVisibleCount: Integer;
        FCollapsed: Boolean;
        FParentNode: TFastTreeItem;
        FPrevNode: TFastTreeItem;
        FNextNode: TFastTreeItem;
        FData: Pointer;
        FDeleting: Boolean;
        FOverlayIndex: Integer;
        FStateIndex: Integer;
        FSelectedIndex: Integer;
        FHasChildren: Boolean;
        FFontColor: TColor;
        function GetItem(idx: Integer): TFastTreeItem;
        procedure SetText(const Value: String);
        procedure SetImageIndex(const Value: Integer);
        procedure SetCollapsed(const Value: Boolean);
        function GetVisible: Boolean;
        procedure SetVisible(const Value: Boolean);
        function GetExpanded: Boolean;
        procedure SetExpanded(const Value: Boolean);
        procedure SetFocused(const Value: Boolean);
        function GetFocused: Boolean;
        function ListItem: TListItem;
        function GetIndex: Integer;
        procedure SetOverlayIndex(const Value: Integer);
        function GetParentNode: TFastTreeItem;
        procedure SetSelected(const Value: Boolean);
        function GetSelected: Boolean;
        procedure SetStateIndex(const Value: Integer);
        procedure SetSelectedIndex(const Value: Integer);
        function GetCount: Integer;
        function GetHasChildren: Boolean;
        procedure SetHasChildren(const Value: Boolean);
        function Get_FontColor: TColor;
        procedure Set_FontColor(const Value: TColor);
    protected
        procedure Unlink;
    public
        constructor Create(AOwner: TFastTreeView; const S: String);
        destructor Destroy; override;
        property Count: Integer read GetCount; //   
        property TotalCount: Integer read FTotalCount;
        property Owner: TFastTreeView read FOwner;
        property Item[idx: Integer]: TFastTreeItem read GetItem;
        function VisibleIndex: Integer;
        property HasChildren: Boolean read GetHasChildren write SetHasChildren;
    public //   
        property Text: String read FText write SetText;
        property ImageIndex: Integer read FImageIndex write SetImageIndex;
        property OverlayIndex: Integer read FOverlayIndex write SetOverlayIndex;
        property StateIndex: Integer read FStateIndex write SetStateIndex;
        property SelectedIndex: Integer read FSelectedIndex write SetSelectedIndex;
        property Data: Pointer read FData write FData;
        property Collapsed: Boolean read FCollapsed write SetCollapsed;
        property Expanded: Boolean read GetExpanded write SetExpanded;
        procedure Expand(Recurse: Boolean);
        procedure Collapse(Recurse: Boolean);
        function DisplayRect(TextOnly: Boolean): TRect;
        property Focused: Boolean read GetFocused write SetFocused;
        property Selected: Boolean read GetSelected write SetSelected;
    public // 
        property Parent: TFastTreeItem read GetParentNode;
        property Visible: Boolean read GetVisible write SetVisible;
        function GetFirstChild: TFastTreeItem;
        function GetLastChild: TFastTreeItem;
        function GetNextChild(AfterChild: TFastTreeItem): TFastTreeItem;
        function GetPrevChild(BeforeChild: TFastTreeItem): TFastTreeItem;
        property GetNextSibling: TFastTreeItem read FNextNode;
        property GetPrevSibling: TFastTreeItem read FPrevNode;
        function GetNext: TFastTreeItem;
        function GetPrev: TFastTreeItem;
        function GetNextVisible: TFastTreeItem;
        function GetPrevVisible: TFastTreeItem;
        function HasAsParent(Node: TFastTreeItem): Boolean;
        function HasSiblingsAsParent(Node: TFastTreeItem): Boolean;
        function IndexOf(Value: TFastTreeItem): Integer;
        procedure MakeVisible;
        function Level: Integer;
        property Index: Integer read GetIndex;
    PUBLIC // , 
        function NextSelected: TFastTreeItem;
    public //  
        procedure Delete;
        procedure DeleteChildren;
        property Deleting: Boolean read FDeleting;
        procedure MoveTo(Destination: TFastTreeItem; Mode: TNodeAttachMode);
        procedure Sort(Recurse: Boolean);
    public // 
        function EditLabel: Boolean;
        property FontColor: TColor
            read Get_FontColor
            write Set_FontColor;
    end;

    TFastTreeNodes = class
    private
        FOwner: TFastTreeView;
        FRoot: TFastTreeItem;
        function GetItem(Idx: Integer): TFastTreeItem;
        function GetVisibleItems(idx: Integer): TFastTreeItem;
        function GetVisibleCount: Integer;
        function GetCount: Integer;
    protected
        property VisibleItemsCount: Integer read GetVisibleCount;
        property VisibleItems[idx: Integer]: TFastTreeItem read GetVisibleItems;
    public
        constructor Create(AOwner: TFastTreeView);
        destructor Destroy; override;
        property Count: Integer read GetCount;
        property Owner: TFastTreeView read FOwner;
        property Item[Idx: Integer]: TFastTreeItem read GetItem; default;
    public
        function Add(Node: TFastTreeItem; const S: string): TFastTreeItem;
        function Insert(Node: TFastTreeItem; const S: string): TFastTreeItem;
        function AddObject(Node: TFastTreeItem; const S: string; P: Pointer): TFastTreeItem;
        function InsertObject(Node: TFastTreeItem; const S: string; P: Pointer): TFastTreeItem;
        function AddChild(Node: TFastTreeItem; const S: string): TFastTreeItem;
        function AddChildObject(Node: TFastTreeItem; const S: string; P: Pointer): TFastTreeItem;
        function AddChildFirst(Node: TFastTreeItem; const S: string): TFastTreeItem;
        function AddChildObjectFirst(Node: TFastTreeItem; const S: string; P: Pointer): TFastTreeItem;
        procedure Delete(Node: TFastTreeItem);
        procedure Clear;
    protected
        FUpdateCount: Integer;
    public
        procedure BeginUpdate;
        procedure EndUpdate;
        function IndexOf(Node: TFastTreeItem): Integer;
    end;

    TFastTreeViewCollapsingEvent = procedure(Sender: TObject; Node: TFastTreeItem;
        var AllowCollapse: Boolean) of object;
    TFastTreeViewExpandingEvent = procedure(Sender: TObject; Node: TFastTreeItem;
        var AllowExpand: Boolean) of object;
    TFastTreeViewExpandEvent = procedure(Sender: TObject; Node: TFastTreeItem)
        of object;
    TFastTreeViewChangingEvent = procedure(Sender: TObject; Node: TFastTreeItem;
        var AllowChange: Boolean) of object;
    TFastTreeViewChangedEvent = procedure(Sender: TObject; Node: TFastTreeItem)
        of object;
    TFastTreeViewDeletionEvent = procedure(Sender: TObject; Node: TFastTreeItem)
        of object;
    TFastTreeViewCompareEvent = procedure(Sender: TObject;
        Node1, Node2: TFastTreeItem; Data: Integer; var Compare: Integer) of object;
    TOnEndEdit = procedure(Sender: TObject; Node: TFastTreeItem; var S: String)
        of object;

    TFastTreeView = class(TCustomListView)
    private
        FNodes: TFastTreeNodes;
        FIndent: Integer;
        FOnCollapsing: TFastTreeViewCollapsingEvent;
        FOnCollapsed: TFastTreeViewExpandEvent;
        FOnExpanding: TFastTreeViewExpandingEvent;
        FOnExpanded: TFastTreeViewExpandEvent;
        FOnChanged: TFastTreeViewChangedEvent;
        FOnChanging: TFastTreeViewChangingEvent;
        FShowButtons: Boolean;
        FButtonImages: TCustomImageList;
        FImages, F_StateImages: TCustomImageList;
        FFocusedLineWidth: Integer;
        FLineWidth: Integer;
        FFocusedLineColor: TColor;
        FLineColor: TColor;
        FOnDeletion: TFastTreeViewDeletionEvent;
        FOnCompare: TFastTreeViewCompareEvent;
        FMultiSelect: Boolean;
        FReadOnly: Boolean;
        FOnEndEdit: TOnEndEdit;
        FSelectWhenExpandCollapseByMouse: Boolean;
        procedure SetIndent(const Value: Integer);
        function GetVisibleNodes(idx: Integer): TFastTreeItem;
        function GetImages: TCustomImageList;
        procedure SetImages(const Value: TCustomImageList);
        function GetStateImages: TCustomImageList;
        procedure SetStateImages(const Value: TCustomImageList);
        procedure SetShowButtons(const Value: Boolean);
        procedure SetButtonImages(const Value: TCustomImageList);
        function GetSelectedNode: TFastTreeItem;
        procedure SetSelectedNode(const Value: TFastTreeItem);
        procedure SetFocusedLineColor(const Value: TColor);
        procedure SetFocusedLineWidth(const Value: Integer);
        procedure SetLineColor(const Value: TColor);
        procedure SetLineWidth(const Value: Integer);
        procedure Set_MultiSelect(const Value: Boolean);
        procedure SetReadOnly(const Value: Boolean);
        procedure SetSelectWhenExpandCollapseByMouse(const Value: Boolean);
    protected
        FBitmap: TBitmap;
        FImageChangeLink: TChangeLink;
        FStateChangeLink: TChangeLink;
        FButtonsChangeLink: TChangeLink;
        FFakeImages: TImageList;
        FEdit: TEdit;
        procedure InvalidateItem(item: TFastTreeItem);
        procedure InvalidateFromItem(item: TFastTreeItem);
        procedure InvalidateIndents;
        procedure AdjustVisibleCount;
        function InheritedItemRect(i: Integer): TRect;
        function InheritedItems(i: Integer): TListItem;
        procedure DoDrawItem(Sender: TCustomListView;
            Item: TListItem; Rect: TRect; State: TOwnerDrawState);
        procedure WndProc(var M: TMessage); override;
        procedure AdjustItemHeight;
        //procedure InheritedSelectItem(Sender: TObject; Item: TListItem;
        //  Selected: Boolean);
        procedure SetInheritedItemsCount(Value: Integer);
        procedure InheritedItems_BeginUpdate;
        procedure InheritedItems_EndUpdate;
        procedure UnselectAllSelected;
        procedure EndEdit;
        procedure WMNCPaint(var M: TMessage); message WM_NCPAINT;
        procedure CMEnter(var M: TCMEnter); message CM_ENTER;
        procedure CMLeave(var M: TCMExit); message CM_EXIT;
        procedure WMVScroll(var M: TWMVScroll); message WM_VSCROLL;
    protected
        FInternalCollapsing: Integer;
    protected
        FEditItem: TFastTreeItem;
        DestroyingTree: Boolean;
        procedure EditExit(Sender: TObject);
        procedure EditChar(Sender: TObject; var Key: Char);
    public
        destructor Destroy; override;
        property Items: TFastTreeNodes read FNodes;
        function VisibleItemsCount: Integer;
        property VisibleItems[idx: Integer]: TFastTreeItem read GetVisibleNodes;
        constructor Create(AOwner: TComponent); override;
        procedure FullCollapse;
        procedure FullExpand;
        function GetHitTestInfoAt(X, Y: Integer): THitTests;
        function GetNodeAt(X, Y: Integer): TFastTreeItem;
        property Selected: TFastTreeItem read GetSelectedNode write SetSelectedNode;
        function AlphaSort: Boolean;
        procedure Fix;
        function InheritedNextSelected(Node: TFastTreeItem): TFastTreeItem;
    published
        property Indent: Integer read FIndent write SetIndent;
        property OnCollapsing: TFastTreeViewCollapsingEvent read FOnCollapsing
            write FOnCollapsing;
        property OnCollapsed: TFastTreeViewExpandEvent read FOnCollapsed
            write FOnCollapsed;
        property OnExpanding: TFastTreeViewExpandingEvent read FOnExpanding
            write FOnExpanding;
        property OnExpanded: TFastTreeViewExpandEvent read FOnExpanded
            write FOnExpanded;
        property OnChanging: TFastTreeViewChangingEvent read FOnChanging
            write FOnChanging;
        property OnChange: TFastTreeViewChangedEvent read FOnChanged
            write FOnChanged;
        property OnDeletion: TFastTreeViewDeletionEvent read FOnDeletion
            write FOnDeletion;
        property OnCompare: TFastTreeViewCompareEvent read FOnCompare
            write FOnCompare;
        property Images: TCustomImageList read GetImages write SetImages;
        property StateImages: TCustomImageList read GetStateImages write SetStateImages;
        property ShowButtons: Boolean read FShowButtons write SetShowButtons;
    published //   Object Inspector':
        property Align;
        property Anchors;
        property BorderStyle;
        property BorderWidth;
        property Color;
        property Constraints;
        property Cursor;
        property Enabled;
        property Font;
        property Height;
        property HelpContext;
        property Hint;
        property Left;
        property Name;
        property ParentColor;
        property ParentFont;
        property PopupMenu;
        property ShowColumnHeaders; //!!!  list view -   .
        property ShowHint;
        property TabOrder;
        property TabStop;
        property Tag;
        property Top;
        property Visible;
        property Width;
        property OnClick;
        property OnDblClick;
        property OnEnter;
        property OnExit;
        property OnKeyDown;
        property OnKeyPress;
        property OnKeyUp;
        property OnMouseDown;
        property OnMouseMove;
        property OnMouseUp;
    published //  
        property ButtonImages: TCustomImageList read FButtonImages write SetButtonImages;
        property FocusedLineColor: TColor read FFocusedLineColor write SetFocusedLineColor;
        property FocusedLineWidth: Integer read FFocusedLineWidth write SetFocusedLineWidth;
        property LineColor: TColor read FLineColor write SetLineColor;
        property LineWidth: Integer read FLineWidth write SetLineWidth;
        property MultiSelect: Boolean read FMultiSelect write Set_MultiSelect;
        property ReadOnly: Boolean read FReadOnly write SetReadOnly;
        property OnEndEdit: TOnEndEdit read FOnEndEdit write FOnEndEdit;
        property SelectWhenExpandCollapseByMouse: Boolean read FSelectWhenExpandCollapseByMouse write SetSelectWhenExpandCollapseByMouse;
    end;

procedure Register;

{$IFNDEF NOREDECLARETREEVIEW}
type
    TTreeView = TFastTreeView;
    TTreeNode = TFastTreeItem;
    {$ENDIF}

implementation

procedure Register;
begin
    RegisterComponents('Kladov', [TFastTreeView]);
end;

{$R FTVStdButtons.res}

{ TFastTreeNodes }

var FStandardButtons: TImageList = nil;

function StandardButtons: TImageList;
begin
    if FStandardButtons = nil then
    begin
        FStandardButtons := TImageList.Create(nil);
        FStandardButtons.Width := 16;
        FStandardButtons.Height := 16;
        FStandardButtons.ResourceLoad(rtBitmap, 'FASTTREEVIEWSTDBUTTONS', clFuchsia)
    end;
    Result := FStandardButtons;
end;

function TFastTreeNodes.Add(Node: TFastTreeItem;
    const S: string): TFastTreeItem;
begin
    Result := AddObject(Node, S, nil);
end;

function TFastTreeNodes.AddChild(Node: TFastTreeItem;
    const S: string): TFastTreeItem;
begin
    Result := AddChildObject(Node, S, nil);
end;

function TFastTreeNodes.AddChildFirst(Node: TFastTreeItem; const S: string): TFastTreeItem;
begin
    Result := AddChildObjectFirst(Node, S, nil);
end;

function TFastTreeNodes.AddChildObject(Node: TFastTreeItem;
    const S: string; P: Pointer): TFastTreeItem;
var addvisible, addvisible1: Boolean;
begin
    Result := TFastTreeItem.Create(Owner, S);
    Result.Data := P;
    if Node = nil then Node := FRoot;
    Result.FParentNode := Node;
    Node.FItems.Add(Result);
    //inc( Node.FCount );
    if Node.Count > 1 then
    begin
        Node.Item[Node.Count - 2].FNextNode := Result;
        Result.FPrevNode := Node.Item[Node.Count - 2];
    end;
    addvisible1 := (Node.Visible or (Node = FRoot)) and not Node.FCollapsed;
    inc(Node.FTotalCount);
    if addvisible1 then
        inc(Node.FVisibleCount);
    Node := Node.FParentNode;
    addvisible := addvisible1 and (Node <> nil) and not Node.FCollapsed;
    while Node <> nil do
    begin
        inc(Node.FTotalCount);
        addvisible := addvisible and not Node.FCollapsed;
        if addvisible then
            inc(Node.FVisibleCount);
        Node := Node.FParentNode;
    end;
    if addvisible1 then
        Owner.AdjustVisibleCount;
end;

function TFastTreeNodes.AddChildObjectFirst(Node: TFastTreeItem;
    const S: string; P: Pointer): TFastTreeItem;
var addvisible, addvisible1: Boolean;
begin
    if Node = nil then Node := FRoot;
    Result := TFastTreeItem.Create(Owner, S);
    Result.FParentNode := Node;
    Result.Data := P;
    Node.FItems.Insert(0, Result);
    addvisible1 := (Node.Visible or (Node = FRoot)) and not Node.FCollapsed;
    //inc( Node.FCount );
    if Node.Count > 1 then
    begin
        Node.Item[1].FPrevNode := Result;
        Result.FNextNode := Node.Item[1];
    end;
    inc(Node.FTotalCount);
    if addvisible1 then
        inc(Node.FVisibleCount);
    Node := Node.FParentNode;
    addvisible := addvisible1 and (Node <> nil) and not Node.FCollapsed;
    while Node <> nil do
    begin
        addvisible := addvisible and not Node.FCollapsed;
        inc(Node.FTotalCount);
        if addvisible then inc(Node.FVisibleCount);
        Node := Node.FParentNode;
    end;
    if addvisible1 then
        Owner.AdjustVisibleCount;
end;

function TFastTreeNodes.AddObject(Node: TFastTreeItem;
    const S: string; P: Pointer): TFastTreeItem;
begin
    if Node <> nil then
        Node := Node.FParentNode
    else
        Node := FRoot;
    Result := AddChildObject(Node, S, P);
end;

procedure TFastTreeNodes.BeginUpdate;
begin
    inc(FUpdateCount);
    Owner.InheritedItems_BeginUpdate;
    //todo:
end;

procedure TFastTreeNodes.Clear;
begin
    if Self = nil then
    begin
        asm
      nop
        end;
        Exit;
    end;
    if Count > 0 then
    begin
        if not (csDestroying in Owner.ComponentState) then
        begin
            Owner.Selected := nil;
            Owner.ItemFocused := nil;
            BeginUpdate;
            TRY
                Owner.SetInheritedItemsCount(0);
                FRoot.DeleteChildren;
            FINALLY
                EndUpdate;
            END;
            Owner.ViewStyle := vsIcon;
            Owner.ViewStyle := vsReport;
            Application.ProcessMessages;
        end;
    end;
end;

constructor TFastTreeNodes.Create(AOwner: TFastTreeView);
begin
    inherited Create;
    FOwner := AOwner;
    FRoot := TFastTreeItem.Create(Owner, '');
    FRoot.FCollapsed := FALSE;
end;

procedure TFastTreeNodes.Delete(Node: TFastTreeItem);
begin
    Node.Delete;
end;

destructor TFastTreeNodes.Destroy;
begin
    FRoot.Free;
    inherited;
end;

procedure TFastTreeNodes.EndUpdate;
begin
    //todo:
    Dec(FUpdateCount);
    if FUpdateCount <= 0 then
    begin
        Owner.AdjustVisibleCount;
        Owner.Invalidate;
    end;
    Owner.InheritedItems_EndUpdate;
end;

function TFastTreeNodes.GetCount: Integer;
begin
    Result := 0;
    TRY
        if (Self = nil) or (FRoot = nil) or Owner.DestroyingTree then
        begin
            asm
        nop
            end;
            Exit;
        end;
        Result := FRoot.FTotalCount;
    EXCEPT
    END;
end;

function TFastTreeNodes.GetItem(Idx: Integer): TFastTreeItem;
begin
    result := nil;
    if Idx < 0 then exit;
    if Idx >= Count then exit;
    Result := FRoot.FItems[0];
    while Idx > 0 do
    begin
        if (Result.Count > 0) and (Idx < 1 + Result.FTotalCount) then
        begin
            Result := Result.FItems[0];
            dec(Idx);
        end
        else
        begin
            dec(Idx, 1 + Result.FTotalCount);
            Result := Result.GetNextSibling;
        end;
    end;
end;

function TFastTreeNodes.GetVisibleCount: Integer;
begin
    Result := FRoot.FVisibleCount;
end;

function TFastTreeNodes.GetVisibleItems(idx: Integer): TFastTreeItem;
begin
    result := nil;
    if Idx < 0 then exit;
    if Idx >= VisibleItemsCount then exit;
    Result := FRoot.Item[0];
    if Result = nil then Exit;
    while Idx > 0 do
    begin
        if (Result <> nil) and (Result.Count > 0) and
            (Result.Collapsed and (Idx <= 0) or
            not Result.Collapsed and
            (Idx < 1 + Result.FVisibleCount)) then
        begin
            Result := Result.Item[0];
            dec(Idx);
        end
        else
            if Result <> nil then
            begin
                if Result.FCollapsed then
                    dec(idx)
                else
                    dec(Idx, 1 + Result.FVisibleCount);
                Result := Result.FNextNode;
            end;
        if Result = nil then break;
    end;
end;

function TFastTreeNodes.IndexOf(Node: TFastTreeItem): Integer;
var i: Integer;
begin
    for i := 0 to Count - 1 do
    begin
        if Item[i] = Node then
        begin
            Result := i;
            Exit;
        end;
    end;
    Result := -1;
end;

function TFastTreeNodes.Insert(Node: TFastTreeItem;
    const S: string): TFastTreeItem;
begin
    Result := InsertObject(Node, S, nil);
end;

function TFastTreeNodes.InsertObject(Node: TFastTreeItem; const S: string;
    P: Pointer): TFastTreeItem;
var addvisible1, addvisible: Boolean;
    i: Integer;
    Prnt: TFastTreeItem;
begin
    if Node <> nil then
        Prnt := Node.FParentNode
    else
        Prnt := FRoot;
    Result := TFastTreeItem.Create(Owner, S);
    Result.FParentNode := Prnt;
    Result.Data := P;
    i := Prnt.IndexOf(Node);
    if i < 0 then
    begin
        Prnt.FItems.Add(Result);
        if Prnt.Count > 1 then
        begin
            Prnt.Item[Prnt.Count - 2].FNextNode := Result;
            Result.FPrevNode := Prnt.Item[Prnt.Count - 2];
        end;
    end
    else
    begin
        Result.FPrevNode := Node.FPrevNode;
        if Result.FPrevNode <> nil then
            Result.FPrevNode.FNextNode := Result;
        Result.FNextNode := Node;
        Node.FPrevNode := Result;
        Prnt.FItems.Insert(i, Result);
    end;
    //inc( Node.Count );
    addvisible1 := Prnt.Visible or (Prnt = FRoot);
    if addvisible1 then
        inc(Prnt.FVisibleCount);
    inc(Prnt.FTotalCount);
    Prnt := Prnt.FParentNode;
    addvisible := addvisible1 and (Prnt <> nil) and not Prnt.FCollapsed;
    while Prnt <> nil do
    begin
        inc(Prnt.FTotalCount);
        if addvisible then inc(Prnt.FVisibleCount);
        addvisible := addvisible and not Prnt.FCollapsed;
        Prnt := Prnt.FParentNode;
    end;
    if addvisible1 then
        Owner.AdjustVisibleCount;
end;

{ TFastTreeItem }

procedure TFastTreeItem.Collapse(Recurse: Boolean);
var i: Integer;
begin
    if Self = nil then Exit;
    Collapsed := TRUE;
    if Recurse then
        for i := 0 to Count - 1 do
            Item[i].Collapse(TRUE);
end;

constructor TFastTreeItem.Create(AOwner: TFastTreeView; const S: String);
begin
    inherited Create;
    FCollapsed := TRUE;
    FOwner := AOwner;
    FText := S;
    FItems := TList.Create;
    FFontColor := clNone;
end;

procedure TFastTreeItem.Delete;
begin
    Free;
end;

procedure TFastTreeItem.DeleteChildren;
var i: Integer;
begin
    for i := Count - 1 downto 0 do
        TObject(FItems.Items[i]).Free;
    //FCount := 0;
    FVisibleCount := 0;
    FTotalCount := 0;
end;

destructor TFastTreeItem.Destroy;
begin
    if Assigned(Owner.OnDeletion) then
        Owner.OnDeletion(Owner, Self);
    FDeleting := TRUE;
    Unlink;
    DeleteChildren;
    FItems.Free;
    inherited;
end;

function TFastTreeItem.DisplayRect(TextOnly: Boolean): TRect;
var i: Integer;
begin
    i := VisibleIndex;
    Result := Rect(0, 0, 0, 0);
    if i < 0 then exit;
    Result := Owner.InheritedItemRect(i);
    Result.Left := Result.Left + Level * Owner.Indent;
    if TextOnly then
    begin
        if Assigned(Owner.Images) then
            inc(Result.Left, Owner.Images.Width);
        if Assigned(Owner.StateImages) then
            inc(Result.Left, Owner.StateImages.Width);
        if Owner.ShowButtons then
            if Assigned(Owner.ButtonImages) then
                inc(Result.Left, Owner.ButtonImages.Width)
            else
                inc(Result.Left, 16);
    end;
end;

function TFastTreeItem.EditLabel: Boolean;
var R: TRect;
begin
    FOwner.Selected := Self;
    //Result := FOwner.InheritedItems( VisibleIndex ).EditCaption;
    if FOwner.FEdit = nil then
    begin
        FOwner.FEdit := TEdit.Create(FOwner);
        FOwner.FEdit.Parent := FOwner;
        FOwner.FEdit.OnExit := FOwner.EditExit;
        FOwner.FEdit.OnKeyPress := FOwner.EditChar;
    end;
    FOwner.FEditItem := Self;
    R := DisplayRect(TRUE);
    if R.Right > FOwner.ClientWidth then
        R.Right := FOwner.ClientWidth;
    FOwner.FEdit.SetBounds(R.Left, R.Top, R.Right - R.Left, R.Bottom - R.Top);
    FOwner.FEdit.Text := Text;
    FOwner.FEdit.Show;
    FOwner.FEdit.SetFocus;
    Result := TRUE;
end;

procedure TFastTreeItem.Expand(Recurse: Boolean);
var i: Integer;
begin
    if Self = nil then Exit;
    Collapsed := FALSE;
    if Recurse then
        for i := 0 to Count - 1 do
            Item[i].Expand(TRUE);
end;

function TFastTreeItem.GetCount: Integer;
begin
    Result := FItems.Count;
end;

function TFastTreeItem.GetExpanded: Boolean;
begin
    Result := not FCollapsed;
end;

function TFastTreeItem.GetFirstChild: TFastTreeItem;
begin
    Result := nil;
    if Count = 0 then Exit;
    Result := FItems[0];
end;

function TFastTreeItem.GetFocused: Boolean;
var F1: Integer;
    i: Integer;
begin
    F1 := Owner.Perform(LVM_GETNEXTITEM, -1, LVNI_FOCUSED);
    Result := FALSE;
    if F1 = -1 then Exit;
    i := VisibleIndex;
    Result := F1 = i;
end;

function TFastTreeItem.GetHasChildren: Boolean;
begin
    Result := FHasChildren or (Count > 0);
end;

function TFastTreeItem.GetIndex: Integer;
begin
    Result := FParentNode.IndexOf(Self);
end;

function TFastTreeItem.GetItem(idx: Integer): TFastTreeItem;
begin
    //Assert( (idx >= 0) and (idx < Count), 'Tree '+Owner.Name+
    //        ' items index out of bounds (' + IntToStr( idx ) + ')' );
    Result := nil;
    if idx < 0 then Exit;
    if idx >= Count then Exit;
    Result := FItems[idx];
end;

function TFastTreeItem.GetLastChild: TFastTreeItem;
begin
    Result := nil;
    if Count = 0 then Exit;
    Result := FItems[Count - 1];
end;

function TFastTreeItem.GetNext: TFastTreeItem;
var FromParent: TFastTreeItem;
begin
    if Count > 0 then
        Result := GetFirstChild
    else
    begin
        Result := FNextNode;
        if Result = nil then
        begin
            FromParent := Self;
            while FromParent <> nil do
            begin
                if FromParent.Count > 0 then
                begin
                    Result := FromParent.FNextNode;
                    if Result <> nil then Exit;
                end;
                FromParent := FromParent.FParentNode;
            end;
        end;
    end;
end;

function TFastTreeItem.GetNextChild(
    AfterChild: TFastTreeItem): TFastTreeItem;
begin
    if AfterChild = nil then
        Result := GetFirstChild()
    else
        Result := AfterChild.GetNextSibling;
end;

function TFastTreeItem.GetNextVisible: TFastTreeItem;
var FromParent: TFastTreeItem;
begin
    FromParent := Self;
    Result := nil;
    while not FromParent.Visible do
        FromParent := FromParent.FParentNode;
    if (FromParent <> nil) and
        not FromParent.Collapsed and
        (FromParent.Count > 0) then
    begin
        Result := FromParent.GetFirstChild;
        Exit;
    end;
    while FromParent <> nil do
    begin
        if FromParent.GetNextSibling <> nil then
        begin
            Result := FromParent.GetNextSibling;
            Exit;
        end;
        FromParent := FromParent.FParentNode;
    end;
end;

function TFastTreeItem.GetParentNode: TFastTreeItem;
begin
    Result := FParentNode;
    if Result = nil then Exit;
    if Result.FParentNode = nil then
        Result := nil;
end;

function TFastTreeItem.GetPrev: TFastTreeItem;
begin
    Result := GetPrevSibling;
    if Result = nil then Result := Parent
    else
    begin
        while Result.Count > 0 do
            Result := Result.GetLastChild;
    end;
end;

function TFastTreeItem.GetPrevChild(
    BeforeChild: TFastTreeItem): TFastTreeItem;
begin
    if BeforeChild = nil then
        Result := GetLastChild()
    else
        Result := BeforeChild.GetPrevSibling;
end;

function TFastTreeItem.GetPrevVisible: TFastTreeItem;
begin
    if not Visible then
    begin
        Result := Parent;
        while not Result.Visible do
            Result := Result.Parent;
        Exit;
    end;
    Result := GetPrevSibling;
    if Result = nil then Result := Parent
    else
    begin
        while not Result.Collapsed and (Result.Count > 0) do
            Result := Result.GetLastChild;
    end;
end;

function TFastTreeItem.GetSelected: Boolean;
begin
    Result := ListItem.Selected;
end;

function TFastTreeItem.GetVisible: Boolean;
var Node: TFastTreeItem;
begin
    Result := FALSE;
    Node := FParentNode;
    while Node <> nil do
    begin
        if Node.Collapsed then Exit;
        Node := Node.FParentNode;
    end;
    Result := TRUE;
end;

function TFastTreeItem.Get_FontColor: TColor;
begin
    Result := FFontColor;
end;

function TFastTreeItem.HasAsParent(Node: TFastTreeItem): Boolean;
var P: TFastTreeItem;
begin
    Result := TRUE;
    if Node = nil then Exit;
    P := FParentNode;
    while P <> nil do
    begin
        if P = Node then Exit;
        P := P.FParentNode;
    end;
    Result := FALSE;
end;

function TFastTreeItem.HasSiblingsAsParent(Node: TFastTreeItem): Boolean;
begin
    while Node <> nil do
    begin
        Result := HasAsParent(Node) or (Node = Self);
        if Result then Exit;
        Node := Node.GetNextSibling;
    end;
    Result := FALSE;
end;

function TFastTreeItem.IndexOf(Value: TFastTreeItem): Integer;
begin
    Result := FItems.IndexOf(Value);
end;

function TFastTreeItem.Level: Integer;
var Node: TFastTreeItem;
begin
    Node := Self;
    Result := -2;
    while Node <> nil do
    begin
        inc(Result);
        Node := Node.FParentNode;
    end;
end;

function TFastTreeItem.ListItem: TListItem;
begin
    Result := nil;
    if not Visible then Exit;
    Result := Owner.InheritedItems(VisibleIndex);
end;

procedure TFastTreeItem.MakeVisible;
begin
    Visible := TRUE;
    // ,    
    ListItem.MakeVisible(FALSE);
end;

procedure TFastTreeItem.MoveTo(Destination: TFastTreeItem;
    Mode: TNodeAttachMode);
    procedure DoAddChild(beforeidx: Integer);
    var becomevisible: Boolean;
        P: TFastTreeItem;
    begin
        becomevisible := (Destination.Visible or (Destination.FParentNode = nil))
            and not Destination.Collapsed;
        Destination.FItems.Insert(beforeidx, Self);
        FParentNode := Destination;
        P := Destination;
        //inc( Destination.FCount );
        while P <> nil do
        begin
            inc(P.FTotalCount, 1 + FTotalCount);
            if becomevisible then
                inc(P.FVisibleCount, 1 + FVisibleCount);
            P := P.FParentNode;
        end;
        if Destination.Count > 1 then
        begin
            FPrevNode := Destination.Item[beforeidx - 1];
            FNextNode := Destination.Item[beforeidx + 1];
            if FPrevNode <> nil then
                FPrevNode.FNextNode := Self;
            if FNextNode <> nil then
                FNextNode.FPrevNode := Self;
        end;
        if becomevisible then
            Owner.AdjustVisibleCount;
    end;
var P: TFastTreeItem;
begin
    Owner.InvalidateFromItem(Self);
    Unlink;
    CASE Mode OF
        naAdd:
            begin
                if Destination = nil then
                    Destination := Owner.FNodes.FRoot
                else
                    Destination := Destination.FParentNode;
                DoAddChild(Destination.Count);
            end;
        naAddFirst:
            begin
                if Destination = nil then
                    Destination := Owner.FNodes.FRoot
                else
                    Destination := Destination.FParentNode;
                DoAddChild(0);
            end;
        naAddChild:
            begin
                if Destination = nil then
                    Destination := Owner.FNodes.FRoot;
                DoAddChild(Destination.Count);
            end;
        naAddChildFirst:
            begin
                if Destination = nil then
                    Destination := Owner.FNodes.FRoot;
                DoAddChild(0);
            end;
        naInsert:
            begin
                P := Destination;
                if Destination = nil then
                    Destination := Owner.FNodes.FRoot
                else
                    Destination := Destination.FParentNode;
                if P = nil then
                    DoAddChild(Destination.Count)
                else
                    DoAddChild(P.Index);
            end;
    END;
    Owner.InvalidateFromItem(Self);
end;

function TFastTreeItem.NextSelected: TFastTreeItem;
begin
    Result := nil;
    if not Owner.MultiSelect then Exit;
    Result := Owner.InheritedNextSelected(Self);
end;

procedure TFastTreeItem.SetCollapsed(const Value: Boolean);
var Allow: Boolean;
    P: TFastTreeItem;
    //WasSelected: TFastTreeItem;
    OwnTree: TFastTreeView;
begin
    if Self = nil then Exit;
    if FCollapsed = Value then Exit;
    OwnTree := Owner;
    Allow := TRUE;
    if OwnTree.FInternalCollapsing = 0 then
    begin
        if Value then
        begin
            if Assigned(OwnTree.OnCollapsing) then
                OwnTree.OnCollapsing(OwnTree, Self, Allow);
        end
        else
        begin
            if Assigned(OwnTree.OnExpanding) then
                OwnTree.OnExpanding(OwnTree, Self, Allow);
        end;
        if not Allow then Exit;
    end;

    //WasSelected := OwnTree.Selected;
    inc(OwnTree.FInternalCollapsing);
    TRY
        //OwnTree.Selected := nil;
        //OwnTree.Selected := OwnTree.Items[0];
        OwnTree.Items[0].MakeVisible;

        FCollapsed := Value;
        if Value then
        begin
            if (OwnTree.Selected <> nil) and OwnTree.Selected.HasAsParent(Self)
                or OwnTree.MultiSelect then
                OwnTree.Selected := Self;
            P := FParentNode;
            while P <> nil do
            begin
                dec(P.FVisibleCount, FVisibleCount);
                P := P.FParentNode;
            end;
            OwnTree.AdjustVisibleCount;
            if OwnTree.FInternalCollapsing <= 1 then
            begin
                if Assigned(OwnTree.OnCollapsed) then
                    OwnTree.OnCollapsed(OwnTree, Self);
            end;
        end
        else
        begin
            if (FVisibleCount = 0) and (Count > 0) then
            begin
                FVisibleCount := Count;
            end;
            P := FParentNode;
            while P <> nil do
            begin
                inc(P.FVisibleCount, FVisibleCount);
                P := P.FParentNode;
            end;
            OwnTree.AdjustVisibleCount;
            if Assigned(OwnTree.OnExpanded) then
                OwnTree.OnExpanded(OwnTree, Self);
        end;
        //OwnTree.Selected := nil;
    FINALLY
        dec(OwnTree.FInternalCollapsing);
    END;
    //OwnTree.Selected := WasSelected;
    if OwnTree.Selected <> nil then
        OwnTree.Selected.MakeVisible;
end;

procedure TFastTreeItem.SetExpanded(const Value: Boolean);
begin
    Collapsed := not Value;
end;

procedure TFastTreeItem.SetFocused(const Value: Boolean);
var NodeF: TFastTreeItem;
    //FireChanged: Boolean;
begin
    if not Visible then Exit;
    NodeF := Owner.Selected;
    //FireChanged := NodeF <> Self;
    if NodeF <> nil then
        Owner.InvalidateItem(NodEF);
    Owner.ItemFocused := ListItem;
    if (Owner.FFocusedLineColor <> Owner.FLineColor) or
        (Owner.FFocusedLineWidth <> Owner.FLineWidth) then
        Owner.InvalidateIndents;
    //if FireChanged and Assigned( Owner.OnChange ) then
    //  Owner.OnChange( Owner, Self );
end;

procedure TFastTreeItem.SetHasChildren(const Value: Boolean);
var WasHasChildren: Boolean;
begin
    if FHasChildren = Value then Exit;
    WasHasChildren := HasChildren;
    FHasChildren := Value;
    if WasHasChildren <> HasChildren then
    begin
        Owner.InvalidateItem(Self);
        Owner.InvalidateIndents;
    end;
end;

procedure TFastTreeItem.SetImageIndex(const Value: Integer);
begin
    if FImageIndex = Value then Exit;
    FImageIndex := Value;
    Owner.InvalidateItem(Self);
end;

procedure TFastTreeItem.SetOverlayIndex(const Value: Integer);
begin
    FOverlayIndex := Value;
    Owner.InvalidateItem(Self);
end;

procedure TFastTreeItem.SetSelected(const Value: Boolean);
var Allow: Boolean;
begin
    if Owner.MultiSelect then
        ListItem.Selected := Value
    else
    begin
        if ListItem.Selected = Value then Exit;
        Allow := TRUE;
        if (Owner.FInternalCollapsing = 0) and Assigned(Owner.OnChanging) then
            Owner.OnChanging(Owner, Self, Allow);
        if not Allow then Exit;
        ListItem.Selected := Value;
        if (Owner.FocusedLineColor <> Owner.LineColor)
            or (Owner.FocusedLineWidth <> Owner.LineWidth) then
            Owner.Invalidate
        else
            Owner.InvalidateItem(Self);
    end;
    if (Owner.FInternalCollapsing = 0) and
        not Owner.MultiSelect and Assigned(Owner.OnChange) then
        Owner.OnChange(Owner, Self);
end;

procedure TFastTreeItem.SetSelectedIndex(const Value: Integer);
begin
    if FSelectedIndex = Value then Exit;
    FSelectedIndex := Value;
    Owner.InvalidateItem(Self);
end;

procedure TFastTreeItem.SetStateIndex(const Value: Integer);
begin
    if FStateIndex = Value then Exit;
    FStateIndex := Value;
    Owner.InvalidateItem(Self);
end;

procedure TFastTreeItem.SetText(const Value: String);
begin
    if FText = Value then Exit;
    FText := Value;
    Owner.InvalidateItem(Self);
end;

procedure TFastTreeItem.SetVisible(const Value: Boolean);
var L: TList;
    P: TFastTreeItem;
    i: Integer;
begin
    L := TList.Create;
    TRY
        P := FParentNode;
        while P <> nil do
        begin
            //P.Collapsed := FALSE;
            L.Add(P);
            P := P.Parent;
        end;
        for i := L.Count - 1 downto 0 do
        begin
            P := L[i];
            P.Collapsed := FALSE;
        end;
    FINALLY
        L.Free;
    END;
end;

function CompareNodes(const Data: Pointer; const e1, e2: Dword): Integer;
var Node: TFastTreeItem;
    i1, i2: TFastTreeItem;
begin
    Node := Data;
    i1 := Node.FItems[e1];
    i2 := Node.FItems[e2];
    if Assigned(Node.Owner.OnCompare) then
        Node.Owner.OnCompare(Node.Owner, i1, i2, 0, Result)
    else
        Result := AnsiCompareStr(i1.Text, i2.Text);
end;

procedure SwapNodes(const Data: Pointer; const e1, e2: Dword);
var Node: TFastTreeItem;
    i1, i2: TFastTreeItem;
begin
    Node := Data;
    i1 := Node.FItems[e1];
    i2 := Node.FItems[e2];
    Node.FItems[e1] := i2;
    Node.FItems[e2] := i1;
end;

procedure TFastTreeItem.Set_FontColor(const Value: TColor);
begin
    if FFontColor = Value then Exit; {>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>}
    FFontColor := Value;
    Owner.InvalidateItem(Self);
end;

procedure TFastTreeItem.Sort(Recurse: Boolean);
var i: Integer;
begin
    if FItems.Count = 0 then Exit;
    SortData(Self, FItems.Count, CompareNodes, SwapNodes);
    if not Recurse then Exit;
    for i := 0 to FItems.Count - 1 do
    begin
        Item[i].FPrevNode := Item[i - 1];
        Item[i].FNextNode := Item[i + 1];
        Item[i].Sort(TRUE);
    end;
end;

procedure TFastTreeItem.Unlink;
var P: TFastTreeItem;
    WasVisible: Boolean;
begin
    WasVisible := Visible;
    if FPrevNode <> nil then
        FPrevNode.FNextNode := FNextNode;
    if FNextNode <> nil then
        FNextNode.FPrevNode := FPrevNode;
    P := FParentNode;
    FNextNode := nil;
    FPrevNode := nil;
    FParentNode := nil;
    if P <> nil then
    begin
        P.FItems.Remove(Self);
        //dec( P.Count );
        if P.FCollapsed then
            dec(P.FVisibleCount, 1 + FVisibleCount);
    end;
    while P <> nil do
    begin
        dec(P.FTotalCount, 1 + FTotalCount);
        if WasVisible then
            dec(P.FVisibleCount, 1 + FVisibleCount);
        P := P.FParentNode;
    end;
    if WasVisible then
        Owner.AdjustVisibleCount;
end;

function TFastTreeItem.VisibleIndex: Integer;
var P, N: TFastTreeItem;
begin
    Result := -1;
    if not Visible then Exit;
    N := Self;
    P := FParentNode;
    while P <> nil do
    begin
        while N <> nil do
        begin
            inc(Result);
            N := N.GetPrevSibling;
            if N = nil then break;
            if not N.Collapsed then
                inc(Result, N.FVisibleCount);
        end;
        N := P;
        P := P.FParentNode;
    end;
end;

{ TFastTreeView }

procedure TFastTreeView.AdjustItemHeight;
var h: Integer;
begin
    inherited SmallImages := nil;
    FreeAndNil(FFakeImages);
    FFakeImages := TImageList.Create(nil);
    FFakeImages.Width := 1;
    h := Font.Height;
    if Assigned(Images) and (h < Images.Height) then
        h := Images.Height;
    if Assigned(StateImages) and (h < StateImages.Height) then
        h := StateImages.Height;
    if Assigned(ButtonImages) and (h < ButtonImages.Height) then
        h := ButtonImages.Height;
    FFakeImages.Height := h;
    inherited SmallImages := FFakeImages;
end;

procedure TFastTreeView.AdjustVisibleCount;
var i: Integer;
    Chg: Boolean;
begin
    if DestroyingTree then Exit;
    if FNodes = nil then Exit;
    if FNodes.FUpdateCount > 0 then Exit;
    i := VisibleItemsCount;
    if inherited Items.Count = i then Exit;
    if Parent <> nil then
    begin
        Chg := inherited Items.Count <> i;
        inherited Items.Count := i;
        if Visible and (FNodes.FUpdateCount <= 0) then
            Invalidate;
        if Chg then
            {if  (inherited TopItem <> nil)
            and (inherited TopItem.Index = 0)
            and (inherited TopItem.DisplayRect(drBounds).Top > 0)
            or  (inherited TopItem = nil) then}
        begin
            {if  inherited TopItem <> nil then
                inherited Scroll( 0, -inherited TopItem.DisplayRect(drBounds).Top );}
            inherited ViewStyle := vsList;
            inherited ViewStyle := vsReport;
            Application.ProcessMessages;
        end;
    end;
end;

function TFastTreeView.AlphaSort: Boolean;
begin
    FNodes.FRoot.Sort(TRUE);
    Result := TRUE;
    Invalidate;
end;

constructor TFastTreeView.Create(AOwner: TComponent);
begin
    inherited Create(AOwner);
    FNodes := TFastTreeNodes.Create(Self);
    FIndent := 20;
    FShowButtons := TRUE;
    OwnerData := TRUE;
    OwnerDraw := TRUE;
    ViewStyle := vsReport;
    ShowColumnHeaders := FALSE;
    inherited OnDrawItem := DoDrawItem;
    //inherited OnSelectItem := InheritedSelectItem;
    Columns.Add;
    Columns[0].Width := 100;
    inherited ReadOnly := TRUE;
    FLineColor := clBtnFace;
    FLineWidth := 1;
    FFocusedLineColor := clHighlight;
    FFocusedLineWidth := 1;
    FImageChangeLink := TChangeLink.Create;
    FStateChangeLink := TChangeLink.Create;
    FButtonsChangeLink := TChangeLink.Create;
    FFakeImages := TImageList.Create(nil);
    FFakeImages.Width := 1;
    FFakeImages.Height := Font.Height;
    FSelectWhenExpandCollapseByMouse := TRUE;
end;

destructor TFastTreeView.Destroy;
var delNodes: TFastTreeNodes;
begin
    DestroyingTree := TRUE;
    TRY
        if HandleAllocated then
            DestroyWindowHandle;
    EXCEPT
    END;
    Items.Clear;
    delNodes := FNodes;
    FreeAndNil(FImageChangeLink);
    FreeAndNil(FStateChangeLink);
    FreeAndNil(FButtonsChangeLink);
    inherited SmallImages := nil;
    FreeAndNil(FFakeImages);
    TRY
        inherited;
    EXCEPT
    END;
    delNodes.Free;
end;

procedure TFastTreeView.DoDrawItem(Sender: TCustomListView;
    Item: TListItem; Rect: TRect; State: TOwnerDrawState);
var R, R2: TRect;
    i, j, x, x2, bw: Integer;
    Node, NodeF, NodeN: TFastTreeItem;
    do_draw: Boolean;
    DestCanvas: TCanvas;
    SaveFont: TFont;
    SaveBrush: TBrush;
    btntxt: String;
begin
    i := Item.Index;
    Node := VisibleItems[i];
    if Assigned(FBitmap) and
        ((FBitmap.Width < ClientWidth) or (FBitmap.Height < Rect.Bottom - Rect.Top)) then
        FreeAndNil(FBitmap);
    if not Assigned(FBitmap) then
    begin
        FBitmap := TBitmap.Create;
        FBitmap.Width := ClientWidth;
        FBitmap.Height := Rect.Bottom - Rect.Top;
    end;
    R := Rect;
    if (ForceCanvas = Canvas.Handle) then
    begin
        DestCanvas := Canvas;
    end
    else
    begin
        DestCanvas := FBitmap.Canvas;
        OffsetRect(R, -R.Left, -R.Top);
    end;
    DestCanvas.Brush.Color := Color;
    DestCanvas.FillRect(R);
    DestCanvas.Font.Assign(Font);
    if Node <> nil then
    begin
        inc(R.Left, Indent * Node.Level);
        x := R.Left;
        if Assigned(ButtonImages) then bw := ButtonImages.Width
        else bw := 16;
        //  
        if Node.Level >= 0 then
        begin
            //         (   
            //   ,  )
            NodeF := Selected;
            do_draw := TRUE;
            if (NodeF <> nil) and
                ((NodeF = Node) or NodeF.HasAsParent(Node)) then
            begin
                DestCanvas.Pen.Color := FFocusedLineColor;
                DestCanvas.Pen.Width := FFocusedLineWidth;
                if FFocusedLineWidth <= 0 then do_draw := FALSE;
            end
            else
            begin
                DestCanvas.Pen.Color := FLineColor;
                DestCanvas.Pen.Width := FLineWidth;
                if FLineWidth <= 0 then do_draw := FALSE;
            end;
            if do_draw then
            begin
                DestCanvas.MoveTo(x + bw, R.Top {+} + (R.Bottom - R.Top) div 2);
                DestCanvas.LineTo(x + bw div 2, R.Top {+} + (R.Bottom - R.Top) div 2);
                DestCanvas.LineTo(x + bw div 2, R.Top {+} - 1);
            end;
            if Node.GetNextSibling <> nil then
            begin
                j := i;
                if (NodeF <> nil) then j := NodeF.VisibleIndex;
                if (NodeF <> nil) and
                    NodeF.HasAsParent(Node.Parent) and
                    (NodeF.VisibleIndex > i) and
                    not NodeF.HasAsParent(Node) then
                begin
                    DestCanvas.Pen.Color := FFocusedLineColor;
                    DestCanvas.Pen.Width := FFocusedLineWidth;
                    if FFocusedLineWidth > 0 then
                    begin
                        DestCanvas.MoveTo(x + bw div 2, 0 {+} + R.Bottom);
                        DestCanvas.LineTo(x + bw div 2, R.Top {+} - 1);
                    end;
                end
                else
                    if (NodeF <> nil) and
                        NodeF.HasAsParent(Node.Parent) and
                        not (NodeF.HasAsParent(Node)) and
                        (j > i) then
                    begin
                        DestCanvas.Pen.Color := FFocusedLineColor;
                        DestCanvas.Pen.Width := FFocusedLineWidth;
                        if FFocusedLineWidth > 0 then
                        begin
                            DestCanvas.MoveTo(x + bw div 2, 0 {+} + R.Bottom);
                            DestCanvas.LineTo(x + bw div 2, R.Top {+} - 1);
                        end;
                    end
                    else
                    begin
                        DestCanvas.Pen.Color := FLineColor;
                        DestCanvas.Pen.Width := FLineWidth;
                        if FLineWidth > 0 then
                        begin
                            DestCanvas.MoveTo(x + bw div 2, 0 {+} + R.Bottom);
                            DestCanvas.LineTo(x + bw div 2, R.Top {+} + (R.Bottom - R.Top) div 2);
                        end;
                    end;
            end;
            //     ,   ,   
            //  
            //NodeP := Node;
            NodeN := Node.FParentNode;
            while NodeN <> nil do
            begin
                if NodeN.GetNextSibling <> nil then
                begin
                    do_draw := TRUE;
                    //    NodeN  NodeN.GetNextSibling
                    if (NodeF <> nil) and
                        //not NodeF.HasChildrenAsParent( NodeN.Parent )
                    NodeF.HasSiblingsAsParent(NodeN.GetNextSibling)
                        {( NodeF.HasAsParent( NodeN ) and
                      not NodeF.HasAsParent( NodeP ) and (NodeF <> NodeP)
                      OR
                      NodeF.HasChildrenAsParent( NodeN.Parent )
                      and (NodeF.VisibleIndex >= Node.VisibleIndex)
                    )}
                    then
                        {
                        if (NodeF <> nil) and
                           ( //(NodeN.GetNextSibling = NodeF) or
                              NodeF.HasAsParent( NodeN ) and
                              not NodeF.HasAsParent( NodeP ) ) then
                        }
                    begin
                        DestCanvas.Pen.Color := FFocusedLineColor;
                        DestCanvas.Pen.Width := FFocusedLineWidth;
                        if FFocusedLineWidth <= 0 then do_draw := FALSE;
                    end
                    else
                    begin
                        DestCanvas.Pen.Color := FLineColor;
                        DestCanvas.Pen.Width := FLineWidth;
                        if FLineWidth <= 0 then do_draw := FALSE;
                    end;
                    x2 := x - Indent * (Node.Level - NodeN.Level) + bw div 2;
                    if do_draw then
                    begin
                        DestCanvas.MoveTo(x2, 0 {+} + R.Bottom);
                        DestCanvas.LineTo(x2, R.Top {+} - 1);
                    end;
                end;
                //NodeP := NodeN;
                NodeN := NodeN.Parent;
            end;
        end;
        if ShowButtons then
        begin
            //----  
            if (Node.Count > 0) or Node.HasChildren then
            begin
                if (ForceCanvas = Canvas.Handle) then
                begin
                    SaveFont := TFont.Create;
                    SaveBrush := TBrush.Create;
                    TRY
                        SaveFont.Assign(DestCanvas.Font);
                        SaveBrush.Assign(DestCanvas.Brush);
                        DestCanvas.Font.Name := 'Courier New';
                        DestCanvas.Font.Size := 6;
                        DestCanvas.Font.Color := clRed;
                        DestCanvas.Font.Style := [];
                        DestCanvas.Brush.Style := bsSolid;
                        DestCanvas.Brush.Color := clYellow;
                        if Node.Collapsed or (Node.Count = 0) then
                            btntxt := '[+]'
                        else
                            btntxt := '[-]';
                        DestCanvas.TextOut(x,
                            R.Top + (Rect.Bottom - Rect.Top - StandardButtons.Height) div 2 + 3,
                            btntxt);
                    FINALLY
                        DestCanvas.Font.Assign(SaveFont);
                        DestCanvas.Brush.Assign(SaveBrush);
                        SaveFont.Free;
                        SaveBrush.Free;
                    END;
                end
                else
                    if Assigned(ButtonImages) then
                        ButtonImages.Draw(DestCanvas, x,
                            R.Top + (Rect.Bottom - Rect.Top - ButtonImages.Height) div 2,
                            Integer(Node.Collapsed or (Node.Count = 0)))
                    else
                    begin
                        StandardButtons.Draw(DestCanvas, x,
                            R.Top + (Rect.Bottom - Rect.Top - StandardButtons.Height) div 2,
                            Integer(Node.Collapsed));
                    end;
            end;
            inc(x, bw);
        end;
        if Assigned(Images) then
        begin
            if Node.ImageIndex >= 0 then
            begin
                if ForceCanvas = Canvas.Handle then
                begin
                    SaveFont := TFont.Create;
                    SaveBrush := TBrush.Create;
                    TRY
                        SaveFont.Assign(DestCanvas.Font);
                        SaveBrush.Assign(DestCanvas.Brush);
                        DestCanvas.Font.Name := 'Arial Narrow';
                        DestCanvas.Font.Size := 8;
                        DestCanvas.Font.Color := clRed;
                        if Node.ImageIndex = 0 then
                            DestCanvas.Font.Color := clBlue;
                        DestCanvas.Font.Style := [];
                        DestCanvas.Brush.Style := bsSolid;
                        DestCanvas.Brush.Color := clYellow;
                        if Node.FontColor = clNone then
                            DestCanvas.Font.Color := clWindowText
                        else
                            DestCanvas.Font.Color := Node.FontColor;
                        DestCanvas.TextOut(x, R.Top + (Rect.Bottom - Rect.Top - Images.Height) div 2,
                            '<' + IntToStr(Node.ImageIndex) + '>');
                    FINALLY
                        DestCanvas.Font.Assign(SaveFont);
                        DestCanvas.Brush.Assign(SaveBrush);
                        SaveFont.Free;
                        SaveBrush.Free;
                    END;
                end
                else
                begin
                    Images.Draw(DestCanvas, x, R.Top + (Rect.Bottom - Rect.Top - Images.Height) div 2,
                        Node.ImageIndex);
                end;
            end;
            inc(x, Images.Width);
        end;
        if Assigned(StateImages) and (Node.StateIndex > 0) then
        begin
            if Node.StateIndex > 0 then
            begin
                if ForceCanvas = Canvas.Handle then
                begin
                    SaveFont := TFont.Create;
                    SaveBrush := TBrush.Create;
                    TRY
                        SaveFont.Assign(DestCanvas.Font);
                        SaveBrush.Assign(DestCanvas.Brush);
                        DestCanvas.Font.Name := 'Arial Narrow';
                        DestCanvas.Font.Size := 8;
                        DestCanvas.Font.Color := clRed;
                        DestCanvas.Font.Style := [];
                        DestCanvas.Brush.Style := bsSolid;
                        DestCanvas.Brush.Color := clYellow;
                        DestCanvas.TextOut(x, R.Top + (Rect.Bottom - Rect.Top - Images.Height) div 2,
                            '<' + IntToStr(Node.StateIndex) + '>');
                    FINALLY
                        DestCanvas.Font.Assign(SaveFont);
                        DestCanvas.Brush.Assign(SaveBrush);
                        SaveFont.Free;
                        SaveBrush.Free;
                    END;
                end
                else
                begin
                    StateImages.Draw(DestCanvas, x,
                        R.Top + (Rect.Bottom - Rect.Top - StateImages.Height) div 2,
                        Node.StateIndex);
                end;
            end;
            inc(x, StateImages.Width);
        end;
        if Node.Selected then
        begin
            if Focused then
            begin
                DestCanvas.Brush.Color := clHighlight;
                DestCanvas.Font.Color := clHighlightText;
            end
            else
            begin
                DestCanvas.Brush.Color := clBtnFace;
                DestCanvas.Font.Color := clBtnText;
            end;
        end;
        DestCanvas.Font.Style := [];
        if Node.FontColor <> clNone then
        begin
            DestCanvas.Font.Color := Node.FontColor;
            DestCanvas.Font.Style := [fsBold];
        end;
        x2 := x + DestCanvas.TextWidth(Node.Text) + 4;
        R2 := R;
        R2.Left := x; R2.Right := x2;
        DestCanvas.TextRect(R2, R2.Left, R2.Top, Node.Text);
        if Node.Focused {and MultiSelect} then
            DestCanvas.DrawFocusRect(R2);
    end;
    if DestCanvas <> Canvas then
        BitBlt(Sender.Canvas.Handle, Rect.Left, Rect.Top, Rect.Right - Rect.Left,
            Rect.Bottom - Rect.Top, FBitmap.Canvas.Handle, 0, 0, SRCCOPY);
end;

procedure TFastTreeView.EditChar(Sender: TObject; var Key: Char);
begin
    if Assigned(OnKeyPress) then
        OnKeyPress(Self, Key);
    CASE Key OF
        #13: EndEdit;
    END;
end;

procedure TFastTreeView.EditExit(Sender: TObject);
var S: String;
begin
    S := FEdit.Text;
    if Assigned(OnEndEdit) then
        OnEndEdit(Self, FEditItem, S);
    FEditItem.Text := S;
    FEdit.Hide;
end;

procedure TFastTreeView.EndEdit;
begin
    if Assigned(FEditItem) then
        FEdit.Hide;
end;

procedure TFastTreeView.Fix;
begin
    inherited ViewStyle := vsList;
    inherited ViewStyle := vsReport;
    Application.ProcessMessages;
end;

procedure TFastTreeView.FullCollapse;
var i: Integer;
    Node: TFastTreeItem;
begin
    for i := 0 to FNodes.FRoot.Count - 1 do
    begin
        Node := FNodes.FRoot.Item[i];
        Node.Collapse(TRUE);
    end;
end;

procedure TFastTreeView.FullExpand;
var i: Integer;
    Node: TFastTreeItem;
begin
    for i := 0 to FNodes.FRoot.Count - 1 do
    begin
        Node := FNodes.FRoot.Item[i];
        Node.Expand(TRUE);
    end;
end;

function TFastTreeView.GetHitTestInfoAt(X, Y: Integer): THitTests;
var R: TRect;
    Node: TFastTreeItem;
    x1: Integer;
begin
    R := ClientRect;
    Result := [];
    if Y < R.Top then Result := Result + [htAbove];
    if Y >= R.Bottom then Result := Result + [htBelow];
    if X < R.Left then Result := Result + [htToLeft];
    if X >= R.Right then Result := Result + [htToRight];
    if Result <> [] then Exit;
    Node := GetNodeAt(X, Y);
    if Node = nil then Result := [htNowhere]
    else
    begin
        Result := [htOnItem];
        R := Node.DisplayRect(FALSE);
        if X < R.Left then Result := Result + [htOnIndent]
        else
        begin
            x1 := R.Left;
            if ShowButtons then
            begin
                if Assigned(ButtonImages) then inc(x1, ButtonImages.Width)
                else inc(x1, 16);
                if X < x1 then
                begin
                    Result := Result + [htOnButton]; Exit;
                end;
            end;
            if Assigned(Images) then
            begin
                inc(x1, Images.Width);
                if X < x1 then
                begin
                    Result := Result + [htOnIcon]; Exit;
                end;
            end;
            if Assigned(StateImages) and (Node.StateIndex > 0) then
            begin
                inc(x1, StateImages.Width);
                if X < x1 then
                begin
                    Result := Result + [htOnStateIcon]; Exit;
                end;
            end;
        end;
    end;
end;

function TFastTreeView.GetImages: TCustomImageList;
begin
    Result := FImages;
end;

function TFastTreeView.GetNodeAt(X, Y: Integer): TFastTreeItem;
var LI: TListItem;
begin
    Result := nil;
    LI := GetItemAt(X, Y);
    if LI = nil then Exit;
    Result := VisibleItems[LI.Index];
end;

function TFastTreeView.GetSelectedNode: TFastTreeItem;
var i: Integer;
begin
    //if MultiSelect then
    i := Perform(LVM_GETNEXTITEM, -1, LVNI_SELECTED)
        //else
    //  i := Perform( LVM_GETNEXTITEM, -1, LVNI_FOCUSED )
    ;
    Result := nil;
    if i < 0 then Exit;
    Result := VisibleItems[i];
end;

function TFastTreeView.GetStateImages: TCustomImageList;
begin
    Result := F_StateImages;
end;

function TFastTreeView.GetVisibleNodes(idx: Integer): TFastTreeItem;
begin
    Result := FNodes.VisibleItems[idx];
end;

function TFastTreeView.InheritedItemRect(i: Integer): TRect;
begin
    Result.Left := LVIR_BOUNDS;
    Perform(LVM_GETITEMRECT, i, Integer(@Result));
end;

function TFastTreeView.InheritedItems(i: Integer): TListItem;
begin
    Result := inherited Items[i];
end;

{procedure TFastTreeView.InheritedSelectItem(Sender: TObject;
  Item: TListItem; Selected: Boolean);
var Node: TFastTreeItem;
begin
  if not Selected then Exit;
  Node := VisibleItems[ Item.index ];
  Node.Focused := TRUE;
end;}

procedure TFastTreeView.InheritedItems_BeginUpdate;
begin
    inherited Items.BeginUpdate;
end;

procedure TFastTreeView.InheritedItems_EndUpdate;
begin
    inherited Items.EndUpdate;
end;

function TFastTreeView.InheritedNextSelected(Node: TFastTreeItem): TFastTreeItem;
var LI: TListItem;
begin
    LI := Node.ListItem;
    LI := LI.ListView.GetNextItem(LI, sdAll, [isSelected]);
    if LI <> nil then Result := VisibleItems[LI.Index]
    else Result := nil;
end;

procedure TFastTreeView.InvalidateFromItem(item: TFastTreeItem);
var R: TRect;
begin
    //Invalidate; //  
    R := InheritedItemRect(item.VisibleIndex);
    if R.Bottom <= R.Top then Exit;
    R.Bottom := ClientHeight;
    InvalidateRect(Handle, @R, TRUE);
end;

procedure TFastTreeView.InvalidateIndents;
var i: Integer;
    Node: TFastTreeItem;
    R: TRect;
begin
    i := Perform(LVM_GETTOPINDEX, 0, 0);
    for i := i to i + Perform(LVM_GETCOUNTPERPAGE, 0, 0) + 1 do
    begin
        Node := VisibleItems[i];
        if Node = nil then break;
        R.Left := LVIR_BOUNDS;
        Perform(LVM_GETITEMRECT, i, Integer(@R));
        R.Right := R.Left + Indent * Node.Level;
        if Assigned(ButtonImages) then
            inc(R.Right, ButtonImages.Width)
        else
            inc(R.Right, 16);
        InvalidateRect(Handle, @R, FALSE);
    end;
end;

procedure TFastTreeView.InvalidateItem(item: TFastTreeItem);
var R: TRect;
begin
    //Invalidate; //  
    if FNodes.FUpdateCount <= 0 then
    begin
        R := InheritedItemRect(item.VisibleIndex);
        InvalidateRect(Handle, @R, TRUE);
    end;
end;

procedure TFastTreeView.SetButtonImages(const Value: TCustomImageList);
begin
    if FButtonImages <> nil then
        FButtonImages.UnRegisterChanges(FButtonsChangeLink);
    FButtonImages := Value;
    if FButtonImages <> nil then
    begin
        FButtonImages.RegisterChanges(FButtonsChangeLink);
        FButtonImages.FreeNotification(Self);
    end;
    AdjustItemHeight;
    Invalidate;
end;

procedure TFastTreeView.SetFocusedLineColor(const Value: TColor);
begin
    if FFocusedLineColor = Value then Exit;
    FFocusedLineColor := Value;
    Invalidate;
end;

procedure TFastTreeView.SetFocusedLineWidth(const Value: Integer);
begin
    if FFocusedLineWidth = Value then Exit;
    FFocusedLineWidth := Value;
    Invalidate;
end;

procedure TFastTreeView.SetImages(const Value: TCustomImageList);
begin
    if Images <> nil then
        Images.UnRegisterChanges(FImageChangeLink);
    FImages := Value;
    if Images <> nil then
    begin
        Images.RegisterChanges(FImageChangeLink);
        Images.FreeNotification(Self);
        //SetImageList(Images.Handle, TVSIL_NORMAL)
    end
        //else SetImageList(0, TVSIL_NORMAL);
        ;
    AdjustItemHeight;
    Invalidate;
end;

procedure TFastTreeView.SetIndent(const Value: Integer);
begin
    FIndent := Value;
    Invalidate;
end;

procedure TFastTreeView.SetInheritedItemsCount(Value: Integer);
begin
    if Value = 0 then
    begin
        if inherited Items.Count > 0 then
        begin
            inherited Items[inherited Items.Count - 1].MakeVisible(TRUE);
            inherited Items[0].MakeVisible(TRUE);
        end;
    end;
    ListView_SetItemCountEx(Handle, Value, 0);
    inherited Items.Count := Value;
end;

procedure TFastTreeView.SetLineColor(const Value: TColor);
begin
    if FLineColor = Value then Exit;
    FLineColor := Value;
    Invalidate;
end;

procedure TFastTreeView.SetLineWidth(const Value: Integer);
begin
    if FLineWidth = Value then Exit;
    FLineWidth := Value;
    Invalidate;
end;

procedure TFastTreeView.Set_MultiSelect(const Value: Boolean);
begin
    FMultiSelect := Value;
    inherited MultiSelect := Value;
end;

procedure TFastTreeView.SetReadOnly(const Value: Boolean);
begin
    FReadOnly := Value;
    //inherited ReadOnly := Value;
end;

procedure TFastTreeView.SetSelectedNode(const Value: TFastTreeItem);
var Allow: Boolean;
begin
    Allow := TRUE;
    if (FInternalCollapsing = 0) and
        (Value <> nil) and Assigned(OnChanging) then
        OnChanging(Self, Value, Allow);
    if not Allow then Exit;
    if Value = nil then
    begin
        UnselectAllSelected;
        if (Selected <> nil) and not MultiSelect then
            Selected.Focused := FALSE;
    end
    else
    begin
        if not MultiSelect then
        begin
            UnselectAllSelected;
            inherited Selected := nil;
        end;
        Value.MakeVisible;
        Value.Selected := TRUE;
        Value.Focused := TRUE;
    end;
    //!!if (Value <> nil) and Assigned( OnChange ) then
    //!!  OnChange( Self, Value );
end;

procedure TFastTreeView.SetShowButtons(const Value: Boolean);
begin
    if FShowButtons = Value then Exit;
    FShowButtons := Value;
    Invalidate;
end;

procedure TFastTreeView.SetStateImages(const Value: TCustomImageList);
begin
    if F_StateImages <> nil then
        F_StateImages.UnRegisterChanges(FStateChangeLink);
    F_StateImages := Value;
    if F_StateImages <> nil then
    begin
        F_StateImages.RegisterChanges(FStateChangeLink);
        F_StateImages.FreeNotification(Self);
    end;
    AdjustItemHeight;
    Invalidate;
end;

procedure TFastTreeView.UnselectAllSelected;
var N, Nx: TFastTreeItem;
begin
    N := Selected;
    while N <> nil do
    begin
        Nx := N.NextSelected;
        N.Selected := FALSE;
        InvalidateItem(N);
        N := Nx;
    end;
end;

function TFastTreeView.VisibleItemsCount: Integer;
begin
    Result := FNodes.VisibleItemsCount;
end;

procedure TFastTreeView.WndProc(var M: TMessage);
    procedure MouseHit(IsDblClk: Boolean);
    var Node, NodeF: TFastTreeItem;
        x, y: Integer;
    begin
        x := SmallInt(LoWord(M.LParam));
        y := SmallInt(HiWord(M.LParam));
        Node := GetNodeAt(x, y);
        if Node <> nil then
        begin
            if (IsDblClk or (htOnButton in GetHitTestInfoAt(x, y)))
                and (Node.HasChildren) then
            begin
                if SelectWhenExpandCollapseByMouse then
                    Selected := Node;
                NodeF := Selected;
                if (NodeF <> nil) and NodeF.HasAsParent(Node) or (NodeF = nil) then
                    //Node.Focused := TRUE;
                    Selected := Node;
                Node.Collapsed := not Node.Collapsed
            end
            else
            begin
                //Node.Focused := TRUE;
                if MultiSelect and ((MK_SHIFT or MK_CONTROL) and M.wParam <> 0) then
                    //Node.Selected := TRUE
                else
                begin
                    Selected := nil;
                    Selected := Node;
                end;
            end;
        end;
    end;
var NewSel, N: TFastTreeItem;
    i: Integer;
    NMHdr: PNMListView;
    //LVDisp: PLVDispInfo;
begin
    CASE M.Msg OF
        WM_SIZE:
            if Columns.Count = 1 then
                Columns[0].Width := ClientWidth - 2;
        WM_LBUTTONDOWN:
            begin
                M.Result := 0;
                Windows.SetFocus(Handle);
                MouseHit(FALSE);
                if not (csDesigning in ComponentState)
                    and not MultiSelect then
                    Exit;
            end;
        WM_LBUTTONDBLCLK:
            begin
                M.Result := 0;
                Windows.SetFocus(Handle);
                MouseHit(TRUE);
                if Assigned(OnDblClick) then
                    OnDblClick(Self);
                if not (csDesigning in ComponentState) then
                    Exit;
            end;
        WM_MOUSEWHEEL:
            begin
                asm nop
                end;
            end;
        WM_KEYDOWN:
            begin
                NewSel := Selected;
                CASE M.WParam OF
                    VK_UP: if Selected = nil then
                        begin
                            if Items.Count > 0 then
                                NewSel := Items[0];
                        end
                        else
                        begin
                            NewSel := Selected.GetPrevVisible;
                            M.WParam := 0;
                            if (NewSel <> nil) and (NewSel <> Selected) then
                            begin
                                Selected := nil;
                                Selected := NewSel;
                                while (Selected = nil) and (NewSel <> nil) do
                                begin
                                    NewSel := NewSel.GetPrevVisible;
                                    Selected := NewSel;
                                end;
                                if (Selected = nil) and (Items.Count > 0) then
                                    Selected := Items[0];
                                if Selected <> nil then
                                    Selected.MakeVisible;
                                Exit;
                            end;
                        end;
                    VK_DOWN:
                        if Selected = nil then
                        begin
                            if Items.Count > 0 then
                                NewSel := Items[0];
                        end
                        else
                            NewSel := Selected.GetNextVisible;
                    VK_LEFT:
                        begin
                            if Selected = nil then
                            begin
                                if Items.Count > 0 then
                                    NewSel := Items[0];
                            end
                            else
                                if (Selected <> nil) and (Selected.Count > 0) and
                                    not Selected.Collapsed then
                                    Selected.Collapse(FALSE)
                                else
                                    if Selected <> nil then
                                        NewSel := Selected.Parent;
                        end;
                    VK_RIGHT:
                        begin
                            if Selected = nil then
                            begin
                                if Items.Count > 0 then
                                    NewSel := Items[0];
                            end
                            else
                                if (Selected <> nil) and ((Selected.Count > 0) or Selected.HasChildren) and
                                    Selected.Collapsed then
                                    Selected.Expand(FALSE)
                                else
                                    if Selected.Count > 0 then
                                        NewSel := Selected.Item[0];
                        end;
                    VK_HOME:
                        begin
                            if Items.Count > 0 then
                                NewSel := Items[0];
                        end;
                    VK_END: begin
                            if Items.Count > 0 then
                                NewSel := Items[Items.Count - 1];
                        end;
                    VK_PRIOR:
                        begin
                            if Items.Count > 0 then
                            begin
                                if NewSel = nil then NewSel := Items[0];
                                for i := 1 to Perform(LVM_GETCOUNTPERPAGE, 0, 0) do
                                begin
                                    N := NewSel.GetPrevVisible;
                                    if N = nil then break;
                                    NewSel := N;
                                end;
                            end;
                        end;
                    VK_NEXT:
                        begin
                            if Items.Count > 0 then
                            begin
                                if NewSel = nil then NewSel := Items[0];
                                for i := 1 to Perform(LVM_GETCOUNTPERPAGE, 0, 0) do
                                begin
                                    N := NewSel.GetNextVisible;
                                    if N = nil then break;
                                    NewSel := N;
                                end;
                            end;
                        end;
                else Exit;
                END;
                M.WParam := 0;
                if (NewSel <> nil) and (NewSel <> Selected) then
                begin
                    Selected := nil;
                    Selected := NewSel;
                    while (Selected = nil) and (NewSel <> nil) do
                    begin
                        NewSel := NewSel.GetNextVisible;
                        Selected := NewSel;
                    end;
                    if (Selected = nil) and (Items.Count > 0) then
                        Selected := Items[0];
                    if Selected <> nil then
                        Selected.MakeVisible;
                end;
            end;
        CN_NOTIFY:
            begin
                if MultiSelect then
                begin
                    NMHdr := Pointer(M.LParam);
                    CASE NMHdr.hdr.code OF
                        LVN_ITEMCHANGED:
                            begin
                                if assigned(OnChange) and
                                    (NMHdr.uOldState and LVIS_SELECTED = 0) and
                                    (NMHdr.uNewState and LVIS_SELECTED <> 0) then
                                    OnChange(Self, VisibleItems[NMHdr.iItem]);
                            end;
                        LVN_ODSTATECHANGED:
                            begin
                                if assigned(OnChange) and
                                    (PNMLVODStateChange(NMHdr).uOldState and LVIS_SELECTED = 0) and
                                    (PNMLVODStateChange(NMHdr).uNewState and LVIS_SELECTED <> 0) then
                                begin
                                    for i := PNMLVODStateChange(NMHdr).iFrom to PNMLVODStateChange(NMHdr).iTo do
                                        OnChange(Self, VisibleItems[i]);
                                end;
                            end;
                    END;
                end
                    (*else
                  if not FReadOnly then
                  begin
                    NMHdr := Pointer( M.LParam );
                    CASE NMHdr.hdr.code OF
                    {LVN_BEGINLABELEDIT:
                      begin
                        LVDisp := Pointer( NMHdr );
                        E := Perform( LVM_GETEDITCONTROL, 0, 0 );
                        i := LVDisp.item.iItem;
                        PostMessage( Handle, CM_BASE+100, i, E );
                      end;}
                    LVN_ENDLABELEDIT:
                      begin
                        LVDisp := Pointer( NMHdr );
                        i := LVDisp.item.iItem;
                        if LVDisp.item.pszText <> nil then
                        begin
                          VisibleItems[ i ].Text := LVDisp.item.pszText;
                          M.Result := 1;
                        end;
                      end;
                    END;
                  end*);
            end;
        {CM_BASE+100:
          begin
            i := M.WParam;
            E := M.LParam;
            N := VisibleItems[ i ];
            GetWindowRect( E, R );
            R.TopLeft := ScreenToClient( R.TopLeft );
            if N <> nil then
            begin
              OffsetRect( R, Indent * N.Level, 0 );
              SetWindowPos( E, 0, R.Left, R.Top, 0, 0, SWP_NOSIZE or SWP_NOZORDER )
            end;
          end;}
    END;
    inherited;
end;

procedure TFastTreeView.SetSelectWhenExpandCollapseByMouse(
    const Value: Boolean);
begin
    FSelectWhenExpandCollapseByMouse := Value;
end;

procedure TFastTreeView.WMNCPaint(var M: TMessage);
var R: TRect;
    dc: HDC;
    Br: TBrush;
begin
    if Focused
        and not DisableFrameAroundFocusedTreeView then
    begin
        inherited;
        R := BoundsRect;
        OffsetRect(R, -R.Left, -R.top);
        dc := GetDCEx(Handle, M.wParam, DCX_WINDOW or DCX_INTERSECTRGN or $10000);
        Br := TBrush.Create;
        try
            Try
                Br.Color := clBlue;
                Windows.FrameRect(dc, R, Br.Handle);
            Except
            end;
        finally
            ReleaseDC(Handle, dc);
            Br.Free;
        end;
    end
    else
        inherited;
end;

procedure TFastTreeView.CMEnter(var M: TCMEnter);
begin
    if not DisableFrameAroundFocusedTreeView then
        RedrawWindow(Handle, nil, 0, RDW_ERASE or RDW_FRAME or RDW_INVALIDATE or RDW_UPDATENOW);
end;

procedure TFastTreeView.CMLeave(var M: TCMExit);
begin
    if not DisableFrameAroundFocusedTreeView then
        RedrawWindow(Handle, nil, 0, RDW_ERASE or RDW_FRAME or RDW_INVALIDATE or RDW_UPDATENOW);
end;

procedure TFastTreeView.WMVScroll(var M: TWMVScroll);
begin
    inherited;
    if GridLines then
        Invalidate;
end;

initialization
finalization
    FStandardButtons.Free;

end.
