//**  ,     VCL Delphi.
//**  Vladimir Kladov, 2000-2010
//**
//**      .
//**     /     
//**  2          
//**   GNU,  Free Software Foundation.
//**
//**       ,    
//** ,       ,   
//**         
//**   .      
//**     GNU.
//**
//**          
//**   GNU.     ,     Free
//** Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
//** MA 02111-1307, USA.
//******************************************************************************

{
  TFloatList -    .
         ,
     -      
  (..      ).
      TList  , ..
    TList -  ,     
   32-  4 ,      
  (  Extended  10   . , .. 
    4,  TList      
  ).

   , 28  2005.
}

unit FltList;

interface

{$I VCL_VERSIONS.inc}

uses Classes;

type
    TFloatArray = array[0..999999] of Extended;
    PFloatArray = ^TFloatArray;

    TFloatList = class(TObject)
    private
        function GetItems(Idx: Integer): Extended;
        procedure SetCapacity(Value: Integer);
        procedure SetItems(Idx: Integer; const Value: Extended);
    protected
        FItems: PFloatArray;
        FCapacity: Integer;
        FCount: Integer;
        procedure Grow;
    public
        destructor Destroy; override;
        procedure Clear;
        property Count: Integer read FCount;
        property Capacity: Integer read FCapacity write SetCapacity;
        procedure Add(Value: Extended);
        procedure Insert(Idx: Integer; Value: Extended);
        procedure Delete(Idx: Integer);
        procedure DeleteRange(Idx, Cnt: Integer);
        property Items[Idx: Integer]: Extended read GetItems write SetItems; default;
    end;

implementation

{ TFloatList }

procedure TFloatList.Add(Value: Extended);
begin
    Insert(FCount, Value);
end;

procedure TFloatList.Clear;
begin
    if FCount = 0 then Exit;
    FreeMem(FItems); FItems := nil;
    FCount := 0;
    FCapacity := 0;
end;

procedure TFloatList.Delete(Idx: Integer);
begin
    DeleteRange(Idx, 1);
end;

procedure TFloatList.DeleteRange(Idx, Cnt: Integer);
begin
    if Idx < 0 then Exit;
    if Idx >= FCount then Exit;
    if Idx + Cnt >= FCount then
        Cnt := FCount - Cnt;
    if (Idx = 0) and (Cnt = FCount) then
        Clear
    else
    begin
        Move(FItems[Idx + Cnt], FItems[Idx], Sizeof(Extended) * (FCount - (Idx + Cnt)));
        Dec(FCount, Cnt);
    end;
end;

destructor TFloatList.Destroy;
begin
    Clear;
    inherited;
end;

function TFloatList.GetItems(Idx: Integer): Extended;
begin
    Result := FItems[Idx];
end;

procedure TFloatList.Grow;
begin
    if FCapacity = 0 then
        Capacity := 4
    else
        if FCapacity < 64 then
            Capacity := FCapacity * 2
        else
            Capacity := FCapacity + FCapacity div 4;
end;

procedure TFloatList.Insert(Idx: Integer; Value: Extended);
begin
    if FCount >= FCapacity then Grow;
    if (FCount > 0) and (Idx < FCount - 1) then
        Move(FItems[Idx], FItems[Idx + 1], Sizeof(Extended) * (FCount - Idx));
    FItems[Idx] := Value;
    inc(FCount);
end;

procedure TFloatList.SetCapacity(Value: Integer);
begin
    if Value < FCount then Value := FCount;
    if Value <> FCapacity then
        ReallocMem(FItems, Sizeof(Extended) * Value);
    FCapacity := Value;
end;

procedure TFloatList.SetItems(Idx: Integer; const Value: Extended);
begin
    if Idx = FCount then Add(Value)
    else
        if Idx > FCount then Exit
        else
            if Idx < 0 then Exit
            else
                FItems[Idx] := Value;
end;

end.
