//------------------------------------------------------------------------------
//  
//  :
// -  TVars -     ID  
// -  CalcExpression( Formula: String; Vars: TVars; SumList: TList = nil ): Extended
//               
//
//  .. 13.07.2004 - 
//             22.07.2004 -   ID  TVars
//             27.07.2004 -    SumList,
//                            :
//             CalcExpression( Formula: String; VarList: TList ): Extended
//             03.08.2004 -   Description   
//                            ( Descriptions[ ],   Add)
//                        -    <, >, <=, >=, <>, =
//------------------------------------------------------------------------------
unit FormulaInterpreterUnit;

interface

{$INCLUDE VCL_Versions.inc}
uses Windows, Classes, SysUtils, FltList{$IFDEF _D6orHigher}, Variants{$ENDIF};

var
    DebugMode: Boolean = FALSE;

type
    EFormula_DupVar = class(Exception);
    EFormula_NoVar = class(Exception);
    EFormula_Syntax = class(Exception);
    EFormula_Condition = class(Exception);

    //----------------------------------------------------------------------------
    //   .     (A <> a)!
    TVars = class
        FNames: TStringList;
        FDescriptions: TStringList;
        FValues: TFloatList;
    private
        function GetCount: Integer;
        procedure SetNames(Idx: Integer; const Value: String);
        procedure SetValues(IdxOrName: Variant; const Value: Extended);
        procedure SetID(Idx: Variant; const Value: Integer);
        procedure SetDescr(Idx: Integer; const Value: String);
    protected
        function GetID(Idx: Variant): Integer; virtual;
        function GetNames(Idx: Integer): String; virtual;
        function GetValues(IdxOrName: Variant): Extended; virtual;
        //--------------------------------------------------------------------------
        //   
        procedure WhenSummed(ToWhat: TVars; const AddText: String); virtual;
        function GetDescr(Idx: Integer): String; virtual;
    public
        //--------------------------------------------------------------------------
        //     (.    implementation)
        constructor Create; virtual;
        destructor Destroy; override;
        property Count: Integer read GetCount;
        property Names[Idx: Integer]: String read GetNames write SetNames;
        property Descriptions[Idx: Integer]: String read GetDescr write SetDescr;
        property Values[IdxOrName: Variant]: Extended read GetValues write SetValues;
        default;
        property ID[Idx: Variant]: Integer read GetID write SetID;
        function Add(const Name: String; Value: Extended; ID: Integer;
            const Description: String; const op: String): Pointer; virtual;
        procedure Delete(Name: Variant);
        function IndexOf(const Name: String): Integer;
        procedure Clear; virtual;
        procedure SwapItems(i1, i2: Integer); virtual;
        procedure Assign(From: TVars); virtual;
    end;

    //----------------------------------------------------------------------------
    //  ,  ( )   
    //    . ,    ..,
    //        () ,
    //      ,   
    //      .
    TAllowSum = procedure(Sender, SumParam: TVars;
        var AllowSum: Boolean; calc_by_batch: Boolean = TRUE) of object;


    //------------------------------------------------------------------------------
    //     
function CalcExpression(const Formula: String; //  
    Vars: TVars; //  
    SumList: TList = nil; //    
    OnAllowSum: TAllowSum = nil;
    Assume0: Boolean = FALSE;
    calc_by_batch: Boolean = TRUE) //   
: Extended; overload;

function CalcExpression(const Formula: String; //  
    VarsList: TList; //   
    OnAllowSum: TAllowSum = nil) //   
: Extended; overload;

procedure LogFileOutPut(const Filename, Text: String);

implementation

uses ParamUnit;


procedure LogFileOutPut(const Filename, Text: String);
{$IFDEF DEBUG_SUM}
var F: TFileStream;
    Flg: DWORD;
    {$ENDIF}
begin
    {$IFDEF DEBUG_SUM}
    Flg := fmOpenReadWrite;
    if not FileExists(Filename) then
        Flg := Flg or fmCreate;
    F := TFileStream.Create(Filename, Flg);
    F.Position := F.Size;
    F.Write((Text + #13#10)[1], Length(Text) + 2);
    F.Free;
    {$ENDIF}
end;


//------------------------------------------------------------------------------
//     

function CalcExpression(const Formula: String; Vars: TVars;
    SumList: TList = nil; OnAllowSum: TAllowSum = nil;
    Assume0: Boolean = FALSE; calc_by_batch: Boolean = TRUE): Extended;

//----------------------------------------------------------------------------
//  
    procedure SkipSpaces(var P: Integer);
    begin
        while (P <= Length(Formula)) and (Formula[P] <= ' ') do Inc(P);
    end;

    function CalcSubExpr(var P: Integer; Vars: TVars;
        const SumList: TList = nil; Assume0: Boolean = FALSE): Extended; forward;

    //----------------------------------------------------------------------------
    // / 
    function GetOperand(var P: Integer; Vars: TVars;
        const SumList: TList = nil; Assume0: Boolean = FALSE;
        calc_by_batch: Boolean = TRUE): Extended;
    var Ready: Boolean;
        Negative: Boolean;
        DotFound: Boolean;
        ExpFound: Boolean;
        P0, P1, PStart, I: Integer;
        SName: String;
        V: TVars;
        VList: TList;
        Allow_sum: Boolean;
        Val, ToVal: Double;

        procedure DoCalculation;
        var I: Integer;
        begin
            P1 := P;
            Result := CalcSubExpr(P, V, VList, TRUE);
            if SumList <> nil then
            begin
                for I := 0 to VList.Count - 1 do
                    if VList[I] <> V then
                    begin
                        Allow_sum := TRUE;
                        if Assigned(OnAllowSum) then
                            OnAllowSum(V, VList[I], Allow_sum, calc_by_batch);
                        if Allow_sum then
                        begin
                            P := P1;
                            TRY
                                Val := CalcSubExpr(P, VList[I], VList, TRUE);
                                ToVal := Result;
                                Result := Result + Val;
                            EXCEPT on E: Exception do
                                begin
                                    {$IFDEF DEBUG_SUM}
                                    if DebugMode then
                                        LogFileOutPut(
                                            GetLogsPath + 'sum.log', '***    - ' +
                                            ' = <' + formula + '>  ' +
                                            IntToStr(P) + '  #' + IntToStr(I) +
                                            '  ' + IntToStr(VList.Count) +
                                            '   = ' + FloatToStr(Result) +
                                            ' : ' + E.message);
                                    {$ENDIF}
                                    raise;
                                end;
                            END;
                            if DebugMode then
                                TVars(VList[I]).WhenSummed(V, ' Result = ' +
                                    Format('%f', [ToVal]) + '  = ' +
                                    Format('%f', [Val]) +
                                    '  Result = ' + Format('%f', [Result]));
                        end;
                    end;
            end;
            if DebugMode then
                V.WhenSummed(V, '<START> Formula = ' + Formula +
                    ' Result = ' + Format('%f', [Result]));
            SkipSpaces(P);
            if (P > Length(Formula)) or (Formula[P] <> ')') then
                raise EFormula_Syntax.Create('      : ' +
                    Formula);
            Inc(P);
        end;
    begin
        P0 := P;
        Ready := FALSE;
        Negative := FALSE;
        Result := 0;
        while not Ready and (P <= Length(Formula)) do
        begin
            //------------------------------------------------------------------------
            //  , ,    '+'-
            while (P <= Length(Formula)) and (
                (Formula[P] <= ' ') or (Formula[P] = '+')) do Inc(P);

            if P > Length(Formula) then
                raise EFormula_Syntax.Create('     ' + Formula);
            CASE Formula[P] OF
                '(': //-------------------------------------------------------------------
                    //   
                    begin
                        Inc(P);
                        Result := CalcSubExpr(P, Vars, SumList, Assume0);
                        SkipSpaces(P);
                        if (P > Length(Formula)) or (Formula[P] <> ')') then
                            raise EFormula_Syntax.Create('    c ' +
                                ' ' + IntToStr(P) + '>' + Copy(Formula, P, MaxInt));
                        Inc(P); //   
                        Ready := TRUE;
                    end;
                '-': //-------------------------------------------------------------------
                    //  '-'     
                    begin
                        Negative := not Negative;
                        Inc(P); //   '-'
                    end;
                '0'..'9', '.', ',': //----------------------------------------------------------
                    //   (       )
                    begin
                        PStart := P;
                        DotFound := Formula[P] in ['.', ','];
                        ExpFound := FALSE;
                        Inc(P);
                        while (P <= Length(Formula)) and (
                            (Formula[P] in ['0'..'9']) or (not DotFound and
                            (Formula[P] in ['.', ',']))
                            or (not ExpFound and (Formula[P] in ['e', 'E']))
                            or (ExpFound and (Formula[P - 1] in ['e', 'E']) and
                            (Formula[P] in ['+', '-']))) do
                        begin
                            CASE Formula[P] OF
                                'e', 'E': begin
                                        ExpFound := TRUE; DotFound := TRUE;
                                    end;
                                '.': DotFound := TRUE;
                            END;
                            Inc(P);
                        end;
                        Result := StrToFloat(StringReplace(Trim(Copy(Formula, PStart, P - PStart)),
                            '.', DecimalSeparator, []));
                        Ready := TRUE;
                    end;
            else //-------------------------------------------------------------------
                //   
                begin
                    PStart := P;
                    while (P <= Length(Formula)) and (Formula[P] > ' ') and
                        not (Formula[P] in ['+', '-', '*', '/', '(', ')', '<', '>', '=', ':', '?']) do
                        Inc(P);
                    SName := Copy(Formula, PStart, P - PStart);
                    SkipSpaces(P);
                    {if (CompareText( SName, 'OP' ) = 0) and (
                         (P >= Length( Formula )) or not(Formula[ P ] in ['A'..'Z','a'..'z',
                            ''..'',''..'', '0'..'9'])
                       ) then
                    begin
                      Result := ;
                    end
                      else}
                    if (CompareText(SName, 'sum') = 0) and (P <= Length(Formula)) and
                        (Formula[P] = '(') then
                    begin //  :     
                        //   TVars  
                        Inc(P);
                        V := Vars;
                        VList := SumList;
                        I := SumList.IndexOf(V);
                        if I < 0 then
                        begin
                            SumList.Insert(0, V);
                            DoCalculation;
                            SumList.Delete(0);
                        end
                        else
                            DoCalculation;
                    end
                    else
                    begin
                        if Assume0 then
                        begin
                            TRY
                                Result := Vars[SName];
                            EXCEPT
                                Result := 0;
                            END;
                        end
                        else
                            Result := Vars[SName];
                    end;
                    Ready := TRUE;
                end;
            END;
        end;
        if not Ready then
            raise EFormula_Syntax.Create('     ' + IntToStr(P0) +
                '   ' + Formula);
        if Negative then Result := -Result;
    end;

    //----------------------------------------------------------------------------
    //    ')'
    function CalcSubExpr(var P: Integer; Vars: TVars;
        const SumList: TList = nil; Assume0: Boolean = FALSE): Extended;
    var Opd1, Opd2: Extended; // 
        Op1, Op2: String; //  (+,-,*,/)
        Opd2Start: Integer;
        Cond: Boolean;
    begin
        //--------------------------------------------------------------------------
        //   
        Opd1 := GetOperand(P, Vars, SumList, Assume0, calc_by_batch);

        //--------------------------------------------------------------------------
        //      ,     
        Result := Opd1;
        SkipSpaces(P);
        if (Trim(Copy(Formula, P, MaxInt)) = '') or
            (P <= Length(Formula)) and (Formula[P] = ')') then Exit;

        while TRUE do
        begin
            //--------------------------------------------------------------------------
            //  
            Op1 := Formula[P];
            if (Op1[1] in ['<', '>']) and (P < Length(Formula)) and (Formula[P + 1] = '=') or
                (Op1[1] = '<') and (P < Length(Formula)) and (Formula[P + 1] = '>') then
            begin
                Inc(P);
                Op1 := Op1 + Formula[P];
            end;
            if not (Op1[1] in ['+', '-', '*', '/', '<', '>', '=', '?', ':']) then
                raise EFormula_Syntax.Create('     ' + Op1 +
                    '  ' + IntToStr(P) + '   ' + Formula);
            Inc(P);

            //--------------------------------------------------------------------------
            //   
            Opd2Start := P;
            Opd2 := GetOperand(P, Vars, SumList, Assume0, calc_by_batch);

            //--------------------------------------------------------------------------
            //    ,   ,   
            //   ,     
            SkipSpaces(P);
            if Trim(Copy(Formula, P, MaxInt)) <> '' then
            begin
                Op2 := Formula[P];
                if (Op2[1] in ['<', '>']) and (P < Length(Formula)) and (Formula[P + 1] = '=') or
                    (Op2[1] = '<') and (P < Length(Formula)) and (Formula[P + 1] = '>') then
                begin
                    Inc(P);
                    Op2 := Op2 + Formula[P];
                end;
                if (Op1[1] in ['-', '+', ':']) and (Op2[1] in ['*', '/', '?']) or
                    (Op1[1] in ['<', '=', '>']) and not (Op2[1] in ['<', '=', '>', '?', ')']) then
                begin
                    P := Opd2Start;
                    Opd2 := CalcSubExpr(P, Vars, SumList);
                end;
            end;

            //--------------------------------------------------------------------------
            //    
            CASE Op1[1] OF
                '+': Result := Opd1 + Opd2;
                '-': Result := Opd1 - Opd2;
                '*': Result := Opd1 * Opd2;
                '/': Result := Opd1 / Opd2;
                '?': if Opd1 <> 0 then
                    begin
                        Result := Opd2;
                        if Op2 = ':' then
                        begin
                            Inc(P);
                            GetOperand(P, Vars, SumList); //   
                        end;
                    end
                    else begin
                        if Op2 = ':' then
                        begin
                            Inc(P);
                            Result := GetOperand(P, Vars, SumList, Assume0, calc_by_batch);
                        end
                        else
                            Result := 0;
                    end;
            else
                Cond := (Op1 = '<') and (Opd1 < Opd2) or
                    (Op1 = '<=') and (Opd1 <= Opd2) or
                    (Op1 = '<>') and (Opd1 <> Opd2) or
                    (Op1 = '>') and (Opd1 > Opd2) or
                    (Op1 = '>=') and (Opd1 >= Opd2) or
                    (Op1 = '=') and (Opd1 = Opd2);
                if not Cond and (pos('?', Formula) <= 0) then
                    raise EFormula_Condition.Create('  : ' + Formula);
                if Cond then Result := 1
                else Result := 0;
            END;

            //--------------------------------------------------------------------------
            //    ,  
            SkipSpaces(P);
            if Trim(Copy(Formula, P, MaxInt)) = '' then break;
            Op2 := Formula[P];
            if Op2 = ')' then break;
            Opd1 := Result;
        end;

    end;

var P: Integer; //   
begin
    P := 1;
    Result := CalcSubExpr(P, Vars, SumList);
    if (P <= Length(Formula)) and (Trim(Copy(Formula, P, MaxInt)) <> '') then
        raise EFormula_Syntax.Create('    ' + Formula +
            '   ' + IntToStr(P) + '>' + Copy(Formula, P, MaxInt));
end;

function CalcExpression(const Formula: String; VarsList: TList;
    OnAllowSum: TAllowSum = nil): Extended; overload;
begin
    Result := CalcExpression(Formula, TObject(VarsList[0]) as TVars,
        VarsList, OnAllowSum);
end;

{ TVars }

function TVars.Add(const Name: String; Value: Extended; ID: Integer;
    const Description: String; const op: String): Pointer;
//------------------------------------------------------------------------------
//  .     .
//var I: Integer;
begin
    {I := IndexOf( Name );
    if (I >= 0) then
      raise EFormula_DupVar.Create( ' ' + Name + '   -  .' );}
    FNames.AddObject(Name, Pointer(ID));
    FDescriptions.Add(Description);
    FValues.Add(Value);
    Result := nil;
end;

procedure TVars.Assign(From: TVars);
var I: Integer;
begin
    if From = Self then Exit;
    if From = nil then Clear
    else
    begin
        FNames.Assign(From.FNames);
        FDescriptions.Assign(From.FDescriptions);
        //SetLength( FVals, Length( From.FVals ) );
        for I := 0 to From.FValues.Count - 1 do
            Values[I] := From.Values[I];
    end;
end;

procedure TVars.Clear;
//------------------------------------------------------------------------------
// 
begin
    FNames.Clear;
    FDescriptions.Clear;
    FValues.Clear;
end;

constructor TVars.Create;
//------------------------------------------------------------------------------
// :    
begin
    inherited;
    FNames := TStringList.Create;
    FDescriptions := TStringList.Create;
    FValues := TFloatList.Create;
end;

procedure TVars.Delete(Name: Variant);
//------------------------------------------------------------------------------
//     
var Idx: Integer;
begin
    if VarType(Name) = varString then
        Idx := IndexOf(Name)
    else
        Idx := Name;
    if (Idx < 0) or (Idx >= Count) then
        raise EFormula_NoVar.Create(' ' + Name + '  .');
    FNames.Delete(Idx);
    FDescriptions.Delete(Idx);
    FValues.Delete(Idx);
end;

destructor TVars.Destroy;
//------------------------------------------------------------------------------
// :  
begin
    Clear;
    FNames.Free;
    FDescriptions.Free;
    FValues.Free;
    inherited;
end;

function TVars.GetCount: Integer;
//------------------------------------------------------------------------------
//   Count
begin
    Result := FNames.Count;
end;

function TVars.GetDescr(Idx: Integer): String;
//------------------------------------------------------------------------------
//      Idx
begin
    Result := '';
    if Count > Idx then
        Result := FDescriptions[Idx];
end;

function TVars.GetID(Idx: Variant): Integer;
//------------------------------------------------------------------------------
//  ID 
var I: Integer;
begin
    if VarType(Idx) = varString then
        I := IndexOf(Idx)
    else
        I := Idx;
    Result := Integer(FNames.Objects[I]);
end;

function TVars.GetNames(Idx: Integer): String;
//------------------------------------------------------------------------------
//      Idx
begin
    Result := '';
    if Count > Idx then
        Result := FNames[Idx];
end;

function TVars.GetValues(IdxOrName: Variant): Extended;
//------------------------------------------------------------------------------
//   
// :   ,   
var I: Integer;
begin
    if VarType(IdxOrName) = varString then
        I := IndexOf(IdxOrName)
    else
        I := IdxOrName;
    if (I < 0) or (I >= Count) then
        raise EFormula_NoVar.Create(' ' + IdxOrName + '  .');
    Result := FValues[I];
end;

function TVars.IndexOf(const Name: String): Integer;
//------------------------------------------------------------------------------
//    
begin
    Result := FNames.IndexOf(Name);
end;

procedure TVars.SetDescr(Idx: Integer; const Value: String);
//------------------------------------------------------------------------------
//     
begin
    FDescriptions[Idx] := Value;
end;

procedure TVars.SetID(Idx: Variant; const Value: Integer);
//------------------------------------------------------------------------------
//  ID 
var I: Integer;
begin
    if VarType(Idx) = varString then
        I := IndexOf(Idx)
    else
        I := Idx;
    FNames.Objects[I] := Pointer(Value);
end;

procedure TVars.SetNames(Idx: Integer; const Value: String);
//------------------------------------------------------------------------------
//     
begin
    FNames[Idx] := Value;
end;

procedure TVars.SetValues(IdxOrName: Variant; const Value: Extended);
//------------------------------------------------------------------------------
//        
var I: Integer;
begin
    if VarType(IdxOrName) = varString then
        I := IndexOf(IdxOrName)
    else
        I := IdxOrName;
    if (I < 0) or (I >= Count) then
        raise EFormula_NoVar.Create(' ' + IdxOrNAme + '  .');
    while FValues.Count <= I do
        FValues.Add(0);
    FValues[I] := Value;
end;

procedure TVars.SwapItems(i1, i2: Integer);
//------------------------------------------------------------------------------
//    (    )
var TmpS: String;
    TmpD: Double;
begin
    TmpS := FNames[i1];
    FNames[i1] := FNames[i2];
    FNames[i2] := TmpS;

    TmpS := FDescriptions[i1];
    FDescriptions[i1] := FDescriptions[i2];
    FDescriptions[i2] := TmpS;

    TmpD := Values[i1];
    Values[i1] := Values[i2];
    Values[i2] := TmpD;
end;

procedure TVars.WhenSummed(ToWhat: TVars; const AddText: String);
begin
    //    -   
end;

end.
