//**  ,     VCL Delphi.
//**  Vladimir Kladov, 2000-2010
//**
//**      .
//**     /     
//**  2          
//**   GNU,  Free Software Foundation.
//**
//**       ,    
//** ,       ,   
//**         
//**   .      
//**     GNU.
//**
//**          
//**   GNU.     ,     Free
//** Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
//** MA 02111-1307, USA.
//******************************************************************************

unit Knopka;

interface

{$I VCL_VERSIONS.inc}

uses
    Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
    StdCtrls, extCtrls, NormalBands, UniPic, Math, UxTheme, buttons, ImgList;

const FocusAnimLength = 200;

type
    TKnopkaLayout = (klLabelRight, klLabelBottom, klLabelLeft, klLabelTop);
    TKnopkaWrap = (kwSyllabels, kwWords, kwLetters, kwNone);
    TKnopkaTextScale = (ktsInFont, ktsInScale, ktsNone);

    TNLabel = class(TNormalLabel)
    end;

    TThreadAuto = class(TThread)
        fOnClick: TNotifyEvent;
        fSender: TControl;
        fExited: Boolean;
        fException: String;
        procedure Execute; override;
    end;

    TKnopka = class(TButton) //TWinControl)
    private
        FLayout: TKnopkaLayout;
        //FAutoSize: Boolean;
        FCaption: String;
        FMarginsPic: Integer;
        FMarginsText: Integer;
        FWrapping: TKnopkaWrap;
        FPicTransparent: Boolean;
        FTextScale: TKnopkaTextScale;
        FFocusDotFrameAroundText: Boolean;
        FRepeatFirst: Integer;
        FRepeatDelay: Integer;
        FRepeatPress: Boolean;
        FTextAlignment: TAlignment;
        FImagesDisabled: TImageList;
        FDisableWhileClick: Boolean;
        FDisableFormWhileClick: Boolean;
        FSpeedButton: Boolean;
        FStdImgShowSize: Integer;
        FFlat: Boolean;
        FModalResult: TModalResult;
        FAnimation: Boolean;
        FAutoThreaded: Boolean;
        FColor: TColor;
        FClickRestoreFocus: Boolean;
        procedure SetLayout(const Value: TKnopkaLayout);
        //procedure SetAutoSize(const Value: Boolean);
        procedure SetCaption(const Value: String);
        procedure SetModalResult(const Value: TModalResult);
        procedure SetAnimation(const Value: Boolean);
        procedure SetAutoThreaded(const Value: Boolean);
        procedure SetColor(const Value: TColor);
        procedure SetClickRestoreFocus(const Value: Boolean);
        function Get_Default: Boolean;
        procedure Set_Default(const Value: Boolean);
    protected
        FPicture: TPicture;
        procedure SetPicture(const Value: TPicture);
        function GetPicture: TPicture;
        procedure CreateWnd; override;
        procedure CreateParams(var aParams: TCreateParams); override;
    protected
        procedure Loaded; override;
        function GetImageList: TImageList;
        procedure SetImageList(const Value: TImageList);
        function GetImageIndex: Integer;
        procedure SetImageIndex(const Value: Integer);
        procedure SetMarginsPic(const Value: Integer);
        procedure SetMarginsText(const Value: Integer);
        procedure SetWrapping(const Value: TKnopkaWrap);
        procedure SetPicTransparent(const Value: Boolean);
        procedure SetTextScale(const Value: TKnopkaTextScale);
        procedure SetFocusDotFrameAroundText(const Value: Boolean);
        procedure SetRepeatFirst(const Value: Integer);
        procedure SetRepeatDelay(const Value: Integer);
        procedure SetRepeatPress(const Value: Boolean);
        procedure SetTextAlignment(const Value: TAlignment);
        procedure SetImagesDisabled(const Value: TImageList);
        procedure SetDisableWhileClick(const Value: Boolean);
        procedure SetDisableFormWhileClick(const Value: Boolean);
        procedure SetSpeedButton(const Value: Boolean);
        procedure SetStdImage(const Value: TStdImage);
        procedure SetStdImgShowSize(const Value: Integer);
        function GetStdImage: TStdImage;
        procedure SetFlat(const Value: Boolean);
        { Private declarations }
    protected
        { Protected declarations }
        NLabel: TNLabel;
        UPic: TUniPic;
        FTimer: DWORD;
        FAniTimer: DWORD;
        FAniStart: DWORD;
        FLastHWnd: THandle;
        FDrawing_NoCaptionNow: Boolean;
        FIsPushed: Boolean;
        procedure AdjustSizeAuto;
        procedure WMSetText(var M: TMessage); message WM_SETTEXT;
        procedure WMGetText(var M: TMessage); message WM_GETTEXT;
        procedure WMTextLen(var M: TMessage); message WM_GETTEXTLENGTH;
        procedure WMPaint(var M: TMessage); message WM_PAINT;
        procedure WMPrint(var M: TMessage); message WM_PRINTCLIENT;
        procedure WM_LButtonDown(var M: TMsg); message WM_LBUTTONDOWN;
        procedure WM_LButtonUp(var M: TMsg); message WM_LBUTTONUP;
        procedure WM_LButtonDblClk(var M: TMsg); message WM_LBUTTONDBLCLK;
        procedure WM_KeyDown(var M: TMsg); message WM_KEYDOWN;
        procedure WM_Size(var M: TMsg); message WM_SIZE;
        procedure WM_Timer(var M: TMessage); message WM_TIMER;
        procedure WMClose(var M: TMessage); message WM_DESTROY;
    protected
        FocusedTime: DWORD;
        procedure WMSetFocus(var M: TMessage); message WM_SETFOCUS;
        procedure WMKillFocus(var M: TMessage); message WM_KILLFOCUS;
    protected
        procedure WMCaptureChanged(var M: TMessage); message WM_CAPTURECHANGED;
        procedure UPicResize(Sender: TObject);
        procedure SetEnabled(E: Boolean); override;
        //procedure CreateWnd( ); override;
        procedure PassFocusNextControl;
    protected
        FProcessingOnClick: Boolean;
        FEnableAfterOnClick: Boolean;
        //FCancel: Boolean;
        //FDefault: Boolean;
        procedure WMEnable(var M: TMessage); message WM_ENABLE;
        function ParentForm: TForm;
        function AllParentsVisibleAndEnabled: Boolean;
        procedure CNCTLCOLORBTN(var Msg: TMessage); message CN_CTLCOLORBTN;
        procedure WndProc(var M: TMessage); override;
    public
        { Public declarations }
        constructor Create(AOwner: TComponent); override;
        destructor Destroy; override;
        procedure Click; override;
        procedure GetAPictureToUPic(Sender: TObject; var P: TPicture);
        property EnableAfterClick: Boolean read FEnableAfterOnClick write FEnableAfterOnClick;
        procedure Invalidate; override;
    published
        { Published declarations }
        property DisableWhileClick: Boolean read FDisableWhileClick write SetDisableWhileClick
            default TRUE;
        property DisableFormWhileClick: Boolean read FDisableFormWhileClick write SetDisableFormWhileClick
            default TRUE;
        property Layout: TKnopkaLayout read FLayout write SetLayout
            default klLabelRight;
        property TextAlignment: TAlignment read FTextAlignment write SetTextAlignment
            default taCenter;
        //property AutoSize: Boolean read FAutoSize write SetAutoSize;
        property Caption: String read FCaption write SetCaption;
        property Picture: TPicture read GetPicture write SetPicture;
        property PicTransparent: Boolean read FPicTransparent write SetPicTransparent
            default TRUE;
        property ImageList: TImageList read GetImageList write SetImageList;
        property ImagesDisabled: TImageList read FImagesDisabled write SetImagesDisabled;
        property ImageIndex: Integer read GetImageIndex write SetImageIndex
            default 0;
        property StdImage: TStdImage read GetStdImage write SetStdImage
            default histBACK;
        property StdImgShowSize: Integer read FStdImgShowSize write SetStdImgShowSize
            default 0;
        property MarginsPic: Integer read FMarginsPic write SetMarginsPic
            default 8;
        property MarginsText: Integer read FMarginsText write SetMarginsText
            default 2;
        property Wrapping: TKnopkaWrap read FWrapping write SetWrapping
            default kwSyllabels;
        property TextScale: TKnopkaTextScale read FTextScale write SetTextScale
            default ktsInFont;
        property FocusDotFrameAroundText: Boolean read FFocusDotFrameAroundText
            write SetFocusDotFrameAroundText default FALSE;
        property RepeatFirst: Integer read FRepeatFirst write SetRepeatFirst
            default 400;
        property RepeatDelay: Integer read FRepeatDelay write SetRepeatDelay
            default 200;
        property RepeatPress: Boolean read FRepeatPress write SetRepeatPress
            default FALSE;
        property Align;
        property SpeedButton: Boolean read FSpeedButton write SetSpeedButton
            default FALSE;
        property ModalResult: TModalResult read FModalResult write SetModalResult;
        property Flat: Boolean read FFlat write SetFlat
            default FALSE;
        //property Glyph: TBitmap read GetGlyph write SetGlyph;
        //property Cancel: Boolean read FCancel write FCancel default FALSE;
        property Default: Boolean read Get_Default write Set_Default default FALSE;
        property OnClick;
        property TabOrder;
        property TabStop;
        property Visible;
        property ParentShowHint;
        property ShowHint;
        property Enabled;
        property Animation: Boolean read FAnimation write SetAnimation
            default TRUE;
        property Color: TColor read FColor write SetColor
            default clBtnFace;
        property AutoThreaded: Boolean read FAutoThreaded write SetAutoThreaded;
        property ClickRestoreFocus: Boolean read FClickRestoreFocus write SetClickRestoreFocus
            default TRUE;
    end;

    TPnKnopka = class(TPanel)
    private
        FSpaceRight: Integer;
        FSpaceLeft: Integer;
        FSpaceBottom: Integer;
        FSpaceTop: Integer;
        procedure SetSpaceBottom(const Value: Integer);
        procedure SetSpaceLeft(const Value: Integer);
        procedure SetSpaceRight(const Value: Integer);
        procedure SetSpaceTop(const Value: Integer);
        function GetLayout: TKnopkaLayout;
        procedure SetLayout(const Value: TKnopkaLayout);
        function GetTextAlignment: TAlignment;
        procedure SetTextAlignment(const Value: TAlignment);
        function GetCaption: String;
        procedure SetCaption(const Value: String);
        function GetPicture: TPicture;
        procedure SetPicture(const Value: TPicture);
        function GetImageList: TImageList;
        procedure SetImageList(const Value: TImageList);
        function GetImgDisabled: TImageList;
        procedure SetImgDisabled(const Value: TImageList);
        function GetImageIndex: Integer;
        procedure SetImageIndex(const Value: Integer);
        function GetMarginsPic: Integer;
        procedure SetMarginsPic(const Value: Integer);
        function GetMarginsText: Integer;
        procedure SetMarginsText(const Value: Integer);
        function GetKnopkaWrap: TKnopkaWrap;
        procedure SetKnopkaWrap(const Value: TKnopkaWrap);
        function GetTextScale: TKnopkaTextScale;
        procedure SetTextScale(const Value: TKnopkaTextScale);
        function GetDotFrame: Boolean;
        procedure SetDotFrame(const Value: Boolean);
        function GetRepeatFirst: Integer;
        procedure SetRepeatFirst(const Value: Integer);
        function GetRepeatDelay: Integer;
        procedure SetRepeatDelay(const Value: Integer);
        function GetRepeatPress: Boolean;
        procedure SetRepeatPress(const Value: Boolean);
        function GetActionA: TBasicAction;
        procedure SetAction(const Value: TBasicAction);
        function GetAutoSize0: Boolean;
        function GetBorderWidth0: Integer;
        function GetBiDiModeK: TBiDiMode;
        procedure SetBiDiModeK(const Value: TBiDiMode);
        function GetEnabledK: Boolean;
        procedure SetEnabledK(const Value: Boolean);
        function GetTabStopK: Boolean;
        procedure SetTabStopK(const Value: Boolean);
        function GetOnClickK: TNotifyEvent;
        procedure SetOnClickK(const Value: TNotifyEvent);
        function GetOnMouseDownK: TMouseEvent;
        procedure SetOnMouseDownK(const Value: TMouseEvent);
        function GetOnMouseMoveK: TMouseMoveEvent;
        procedure SetOnMouseMoveK(const Value: TMouseMoveEvent);
        function GetOnMouseUpK: TMouseEvent;
        procedure SetOnMouseUpK(const Value: TMouseEvent);
        function GetOnDblClickK: TNotifyEvent;
        procedure SetOnDblClickK(const Value: TNotifyEvent);
        function GetOnKeyDown: TKeyEvent;
        procedure SetOnKeyDown(const Value: TKeyEvent);
        function GetOnKeyUp: TKeyEvent;
        procedure SetOnKeyUp(const Value: TKeyEvent);
        function GetOnKeyPress: TKeyPressEvent;
        procedure SetOnKeyPress(const Value: TKeyPressEvent);
        function GetDisableFormWhileClick: Boolean;
        function GetDisableWhileClick: Boolean;
        procedure SetDisableFormWhileClick(const Value: Boolean);
        procedure SetDisableWhileClick(const Value: Boolean);
        function GetSpeedButton: Boolean;
        procedure SetSpeedButton(const Value: Boolean);
        function GetModalResult: TModalResult;
        procedure SetModalResult(const Value: TModalResult);
        function GetStdImg: TStdImage;
        procedure SetStdImg(const Value: TStdImage);
        function GetStdImgShowSz: Integer;
        procedure SetStdImgShowSz(const Value: Integer);
        function GetIsCancel: Boolean;
        function GetIsDefault: Boolean;
        procedure SetIsCancel(const Value: Boolean);
        procedure SetIsDefault(const Value: Boolean);
        function GetFlat: Boolean;
        procedure SetFlat(const Value: Boolean);
        function GetEnableAfterOnClick: Boolean;
        procedure SetEnableAfterOnClick(const Value: Boolean);
        function GetAnimation: Boolean;
        procedure SetAnimation(const Value: Boolean);
        function GetColorKnopka: TColor;
        procedure SetColorKnopka(const Value: TColor);
        function Get_AutoThreaded: Boolean;
        procedure Set_AutoThreaded(const Value: Boolean);
    protected
        Knopka: TKnopka;
        FDrawing_NoCaptionNow: Boolean;
        procedure AdjustKnopkaPosition;
        procedure WMSetText(var M: TMessage); message WM_SETTEXT;
        procedure WMGetText(var M: TMessage); message WM_GETTEXT;
        procedure WMTextLen(var M: TMessage); message WM_GETTEXTLENGTH;
        procedure WMPaint(var M: TMessage); message WM_PAINT;
    public
        constructor Create(AOwner: TComponent); override;
        procedure Click; override;
        property EnableAfterClick: Boolean read GetEnableAfterOnClick
            write SetEnableAfterOnClick;
    published
        property SpaceLeft: Integer read FSpaceLeft write SetSpaceLeft
            default 0;
        property SpaceRight: Integer read FSpaceRight write SetSpaceRight
            default 0;
        property SpaceTop: Integer read FSpaceTop write SetSpaceTop
            default 0;
        property SpaceBottom: Integer read FSpaceBottom write SetSpaceBottom
            default 0;
        property Layout: TKnopkaLayout read GetLayout write SetLayout
            default klLabelRight;
        property TextAlignment: TAlignment read GetTextAlignment write SetTextAlignment
            default taCenter;
        //property AutoSize: Boolean read GetAutoSize write SetAutoSize;
        property Caption: String read GetCaption write SetCaption;
        property Picture: TPicture read GetPicture write SetPicture;
        property ImageList: TImageList read GetImageList write SetImageList;
        property ImagesDisabled: TImageList read GetImgDisabled write SetImgDisabled;
        property ImageIndex: Integer read GetImageIndex write SetImageIndex
            default 0;
        property MarginsPic: Integer read GetMarginsPic write SetMarginsPic
            default 8;
        property MarginsText: Integer read GetMarginsText write SetMarginsText
            default 2;
        property Wrapping: TKnopkaWrap read GetKnopkaWrap write SetKnopkaWrap
            default kwSyllabels;
        property TextScale: TKnopkaTextScale read GetTextScale write SetTextScale
            default ktsInFont;
        property FocusDotFrameAroundText: Boolean read GetDotFrame write SetDotFrame
            default FALSE;
        property RepeatFirst: Integer read GetRepeatFirst write SetRepeatFirst
            default 400;
        property RepeatDelay: Integer read GetRepeatDelay write SetRepeatDelay
            default 200;
        property RepeatPress: Boolean read GetRepeatPress write SetRepeatPress
            default FALSE;
        property DisableWhileClick: Boolean read GetDisableWhileClick write SetDisableWhileClick
            default TRUE;
        property DisableFormWhileClick: Boolean read GetDisableFormWhileClick write SetDisableFormWhileClick
            default TRUE;
    published
        property Action: TBasicAction read GetActionA write SetAction;
        property Alignment: TAlignment read GetTextAlignment write SetTextAlignment
            default taCenter;
        property AutoSize: Boolean read GetAutoSize0
            default FALSE;
        property BorderWidth: Integer read GetBorderWidth0;
        property BiDiMode: TBiDiMode read GetBiDiModeK write SetBiDiModeK
            default bdLeftToRight;
        property Enabled: Boolean read GetEnabledK write SetEnabledK
            default TRUE;
        property Cancel: Boolean read GetIsCancel write SetIsCancel
            default FALSE;
        property Default: Boolean read GetIsDefault write SetIsDefault
            default FALSE;
        property TabStop: Boolean read GetTabStopK write SetTabStopK
            default TRUE;
        property OnClick: TNotifyEvent read GetOnClickK write SetOnClickK;
        property OnMouseDown: TMouseEvent read GetOnMouseDownK write SetOnMouseDownK;
        property OnMouseMove: TMouseMoveEvent read GetOnMouseMoveK write SetOnMouseMoveK;
        property OnMouseUp: TMouseEvent read GetOnMouseUpK write SetOnMouseUpK;
        property OnDblClick: TNotifyEvent read GetOnDblClickK write SetOnDblClickK;
        property OnKeyDown: TKeyEvent read GetOnKeyDown write SetOnKeyDown;
        property OnKeyUp: TKeyEvent read GetOnKeyUp write SetOnKeyUp;
        property OnKeyPress: TKeyPressEvent read GetOnKeyPress write SetOnKeyPress;
        property SpeedButton: Boolean read GetSpeedButton write SetSpeedButton
            default FALSE;
        property ModalResult: TModalResult read GetModalResult write SetModalResult
            default mrNone;
        property StdImage: TStdImage read GetStdImg write SetStdImg
            default histBACK;
        property StdImgShowSize: Integer read GetStdImgShowSz write SetStdImgShowSz
            default 0;
        property Flat: Boolean read GetFlat write SetFlat
            default FALSE;
        property Animation: Boolean read GetAnimation write SetAnimation
            default TRUE;
        property ColorKnopka: TColor read GetColorKnopka write SetColorKnopka;
        property AutoThreaded: Boolean read Get_AutoThreaded write Set_AutoThreaded;
    end;

    TGrKnopka = class(TGraphicControl)
    private
        FDisableWhileClick: Boolean;
        FDisableFormWhileClick: Boolean;
        FModalResult: TModalResult;
        FImagesDisabled: TImageList;
        FStdImgShowSz: Integer;
        FRepeatPress: Boolean;
        FRepeatDelay: Integer;
        FRepeatFirst: Integer;
        FFlat: Boolean;
        FLayout: TKnopkaLayout;
        FDefault: Boolean;
        FCancel: Boolean;
        FHelpContext: THelpContext;
        FMarginsPic: Integer;
        FMarginsText: Integer;
        FPicTransparent: Boolean;
        FTextScale: TKnopkaTextScale;
        FWrapping: TKnopkaWrap;
        FTextAlignment: TAlignment;
        procedure SetDisableFormWhileClick(const Value: Boolean);
        procedure SetDisableWhileClick(const Value: Boolean);
        procedure SetModalResult(const Value: TModalResult);
        function GetPicture: TPicture;
        procedure SetPicture(const Value: TPicture);
        function GetImageList: TImageList;
        procedure SetImageList(const Value: TImageList);
        procedure SetImagesDisabled(const Value: TImageList);
        function GetImageIndex: Integer;
        procedure SetImageIndex(const Value: Integer);
        function GetStdImage: TStdImage;
        procedure SetStdImage(const Value: TStdImage);
        procedure SetStdImgShowSz(const Value: Integer);
        procedure SetRepeatPress(const Value: Boolean);
        procedure SetRepeatDelay(const Value: Integer);
        procedure SetRepeatFirst(const Value: Integer);
        procedure SetFlat(const Value: Boolean);
        procedure SetLayout(const Value: TKnopkaLayout);
        procedure SetCancel(const Value: Boolean);
        procedure SetDefault(const Value: Boolean);
        //procedure SetFocusDotFrameAroundText(const Value: Boolean);
        procedure SetHelpContext(const Value: THelpContext);
        procedure SetMarginsPic(const Value: Integer);
        procedure SetMarginsText(const Value: Integer);
        procedure SetPicTransparent(const Value: Boolean);
        procedure SetTextScale(const Value: TKnopkaTextScale);
        procedure SetWrapping(const Value: TKnopkaWrap);
        procedure SetTextAlignment(const Value: TAlignment);
    protected
        procedure Paint; override;
        function ParentForm: TCustomForm;
        procedure WMMouseMove(var M: TWMMouse); message WM_MOUSEMOVE;
        procedure WMEraseBkgd(var M: TWMEraseBkgnd); message WM_ERASEBKGND;
    protected
        UPic: TUniPic;
        FProcessingOnClick: Boolean;
        FEnableAfterOnClick: Boolean;
        FIsPushed: Boolean;
        FTimePressed: DWORD;
        FWaitInterval: DWORD;
        FTimer: DWORD;
        procedure Click; override;
        procedure WMLButtonDown(var M: TWMMouse); message WM_LBUTTONDOWN;
        procedure WMLButtonUp(var M: TWMMouse); message WM_LBUTTONUP;
        procedure WMLDblClick(var M: TWMLButtonDblClk); message WM_LBUTTONDBLCLK;
        procedure CMTextChanged(var Message: TMessage); message CM_TEXTCHANGED;
    public
        constructor Create(AOwner: TComponent); override;
        destructor Destroy; override;
        procedure Invalidate; override;
        property EnableAfterClick: Boolean read FEnableAfterOnClick write FEnableAfterOnClick;
    published
        property Caption;
        property DisableWhileClick: Boolean read FDisableWhileClick write SetDisableWhileClick
            default TRUE;
        property DisableFormWhileClick: Boolean read FDisableFormWhileClick write SetDisableFormWhileClick
            default TRUE;
        property OnClick;
        property ModalResult: TModalResult read FModalResult write SetModalResult
            default mrNone;
        property Picture: TPicture read GetPicture write SetPicture;
        property ImageList: TImageList read GetImageList write SetImageList;
        property ImagesDisabled: TImageList read FImagesDisabled write SetImagesDisabled;
        property ImageIndex: Integer read GetImageIndex write SetImageIndex
            default 0;
        property StdImage: TStdImage read GetStdImage write SetStdImage
            default histBACK;
        property StdImgShowSize: Integer read FStdImgShowSz write SetStdImgShowSz
            default 0;
        property RepeatPress: Boolean read FRepeatPress write SetRepeatPress
            default FALSE;
        property RepeatDelay: Integer read FRepeatDelay write SetRepeatDelay
            default 200;
        property RepeatFirst: Integer read FRepeatFirst write SetRepeatFirst
            default 400;
        property Flat: Boolean read FFlat write SetFlat
            default FALSE;
        property Layout: TKnopkaLayout read FLayout write SetLayout
            default klLabelRight;
        property Align;
        property Cancel: Boolean read FCancel write SetCancel
            default FALSE;
        property Default: Boolean read FDefault write SetDefault
            default FALSE;
        property Enabled;
        property Font;
        property ParentFont;
        property Color;
        property ParentColor;
        property HelpContext: THelpContext read FHelpContext write SetHelpContext;
        property Hint;
        property ShowHint;
        property ParentShowHint;
        property MarginsPic: Integer read FMarginsPic write SetMarginsPic
            default 8;
        property MarginsText: Integer read FMarginsText write SetMarginsText
            default 2;
        property PicTransparent: Boolean read FPicTransparent write SetPicTransparent
            default TRUE;
        property TextScale: TKnopkaTextScale read FTextScale write SetTextScale
            default ktsInFont;
        property Wrapping: TKnopkaWrap read FWrapping write SetWrapping
            default kwSyllabels;
        property TextAlignment: TAlignment read FTextAlignment write SetTextAlignment
            default taCenter;
        property Alignment: TAlignment read FTextAlignment write SetTextAlignment
            default taCenter;
        property OnMouseDown;
        property OnMouseUp;
        property OnMouseMove;
        property ParentBiDiMode;
        property PopupMenu;
        property Visible;
    end;

procedure Register;
{$R Knopka.dcr}

procedure GrayBitmap(B: TBitmap; ExceptColor: TColor);

var DisableAllIconsOnKnopka: Boolean = FALSE;
    DisableBtnAnimation: Boolean = FALSE;
    Knopka0: TKnopka; //   TGrKnopka
    ThemeOK: Boolean;

implementation

uses utPhotoExc;

procedure Register;
begin
    RegisterComponents('Kladov', [TKnopka, TPnKnopka, TGrKnopka]);
end;

var
    TableGray: array[0..255] of Byte;

procedure GrayBitmap(B: TBitmap; ExceptColor: TColor);
var x, y, z: Integer;
    Dst: ^TRGBQuad;
    ExceptRGB: TRGBQuad;
begin
    if B.PixelFormat <> pf32bit then Exit;

    if TableGray[255] = 0 then
    begin
        for x := 0 to 255 do
        begin
            TableGray[x] := Trunc(255 * power(x / 255.0, 0.2));
        end;
    end;

    ExceptRGB.rgbBlue := ExceptColor and 255;
    ExceptRGB.rgbGreen := (ExceptColor shr 8) and 255;
    ExceptRGB.rgbRed := (ExceptColor shr 16) and 255;
    ExceptRGB.rgbReserved := 0;
    for y := 0 to B.Height - 1 do
    begin
        Dst := B.ScanLine[y];
        for x := 0 to B.Width - 1 do
        begin
            if (DWORD(Dst^) and $FFFFFF) <> DWORD(ExceptRGB) then
            begin
                z := (Dst^.rgbBlue + Dst^.rgbGreen * 2 + Dst^.rgbRed * 4) div 8;
                z := TableGray[z];
                Dst^.rgbBlue := z;
                Dst^.rgbGreen := z;
                Dst^.rgbRed := z;
            end;
            inc(Dst);
        end;
    end;
end;

var TrackingGrKnopka: TGrKnopka;
    Timer: DWORD;

procedure TimerTracking(W: HWnd; uMsg, idEvent, dwTime: DWORD); stdcall;
var P: TPoint;
begin
    if TrackingGrKnopka <> nil then
    begin
        GetCursorPos(P);
        P := TrackingGrKnopka.ScreenToClient(P);
        if (P.x < 0) or (P.x >= TrackingGrKnopka.Width)
            or (P.y < 0) or (P.y >= TrackingGrKnopka.Height) then
        begin
            TrackingGrKnopka.Invalidate;
            TrackingGrKnopka := nil;
            if Timer <> 0 then
                KillTimer(0, Timer);
            Timer := 0;
        end;
    end;
end;

procedure SetTrackingGrKnopka(K: TGrKnopka);
begin
    if TrackingGrKnopka <> nil then
    begin
        if TrackingGrKnopka <> K then
            TrackingGrKnopka.Invalidate;
        if Timer <> 0 then
            KillTimer(0, Timer);
    end;
    TrackingGrKnopka := K;
    Timer := SetTimer(0, 0, 100, @TimerTracking);
end;

{ TKnopka }

procedure TKnopka.AdjustSizeAuto;
var dW, dH, W, H, pW, pH: Integer;
begin
    if not AutoSize then Exit;
    pW := UPic.Width;
    pH := UPic.Height;
    if DisableAllIconsOnKnopka and (Caption <> '')
        OR (UPic.ImageList = nil)
        and ((UPic.Picture.Graphic = nil)
        OR (UPic.Picture.Graphic.Empty))
        and (StdImgShowSize = 0) then
    begin
        pW := 0;
        pH := 0;
    end;
    dW := Width - ClientWidth;
    dH := Height - ClientHeight;
    CASE Layout OF
        klLabelTop, klLabelBottom:
            begin
                W := Max(NLabel.DesiredWidth + MarginsText * 2, pW);
                H := NLabel.Height + MarginsText * 2 + pH;
            end;
        //klLabelLeft, klLabelRight:
    else
        begin
            H := Max(NLabel.DesiredHeight + MarginsText * 2, pH);
            W := NLabel.Width + MarginsText * 2 + pW;
        end;
    END;
    Width := W + dW;
    Height := H + dH;
end;

function TKnopka.AllParentsVisibleAndEnabled: Boolean;
var C: TControl;
begin
    Result := FALSE;
    C := Self;
    while C <> nil do
    begin
        if not C.Visible or not C.Enabled then Exit;
        if C is TForm then break;
        C := C.Parent;
    end;
    Result := TRUE;
end;

procedure TKnopka.Click;
var ParentTop: TControl;
    SelfCtl: TControl;
    Form: TCustomForm;
    e: Boolean;
    Re_Focus: Boolean;
    th: TThreadAuto;
begin
    if FProcessingOnClick then
    begin
        raise Exception.Create('    ' + Caption);
    end;
    FProcessingOnClick := TRUE;
    e := Enabled;
    if DisableWhileClick then
    begin
        Enabled := FALSE;
        if UPic <> nil then
            UPic.Enabled := TRUE;
    end;
    FEnableAfterOnClick := e;
    ParentTop := nil;
    if DisableFormWhileClick then
    begin
        ParentTop := Parent;
        while (ParentTop <> nil) and (ParentTop.Parent <> nil) do
            ParentTop := ParentTop.Parent;
        if (ParentTop <> nil)
            and (ParentTop.Visible) then
            ParentTop.Enabled := FALSE
        else
            ParentTop := nil;
    end;
    Re_Focus := ClickRestoreFocus;
    th := nil;
    TRY
        Invalidate;
        Update;
        if FAnimation then
        begin
            FAniStart := GetTickCount;
            if FAniTimer = 0 then
                FAniTimer := SetTimer(Handle, 117, 100, nil);
        end;
        //////////////////////////////////////////////////////////////////////////
        //inherited;
        Form := GetParentForm(Self);
        SelfCtl := Self;
        if (Parent <> nil)
            and (Parent is TPnKnopka) then
            SelfCtl := Parent;
        if Form <> nil then Form.ModalResult := ModalResult;
        //inherited Click;
        if AutoThreaded and Assigned(OnClick) then
        begin
            th := TThreadAuto.Create(TRUE);
            th.fOnClick := OnClick;
            th.fSender := Self;
            th.Resume;
            while not th.fExited and not th.Terminated do
            begin
                Sleep(50);
                if Animation then
                    Invalidate;
                Application.ProcessMessages;
            end;
            if th.fException <> '' then
                raise Exception.Create(th.fException);
            FreeAndNil(th);
        end
        else
        begin
            if Assigned(OnClick) and (Action <> nil)
                and (@OnClick <> @Action.OnExecute) then
                OnClick(SelfCtl)
            else if not (csDesigning in ComponentState) and (ActionLink <> nil) then
                ActionLink.Execute
            else if Assigned(OnClick) then
                OnClick(SelfCtl);
        end;
        //////////////////////////////////////////////////////////////////////////
    FINALLY
        FProcessingOnClick := FALSE;
        FIsPushed := FALSE;
        if (FAniTimer <> 0) and not Default then
        begin
            KillTimer(Handle, FAniTimer);
            FAniTimer := 0;
        end;
        if DisableWhileClick then
            Enabled := FEnableAfterOnClick;
        Invalidate;
        Update;
        if ClickRestoreFocus
            and DisableFormWhileClick
            and (ParentTop <> nil)
            and (ParentTop.Visible) then
        begin
            ParentTop.Enabled := TRUE;
            if not ((ParentTop is TForm) and
                ((ParentTop as TForm).ActiveControl <> nil) and
                ((ParentTop as TForm).ActiveControl <> Self)) then
                Windows.SetFocus((ParentTop as TWinControl).Handle)
            else
                if (ParentTop is TForm)
                    and ((ParentForm as TForm).ActiveControl <> nil) then
                begin
                    Windows.SetFocus((ParentTop as TForm).ActiveControl.Handle);
                    Re_Focus := FALSE;
                end;
        end;
        th.Free;
        Invalidate;
    END;
    if not SpeedButton
        and Enabled
        and AllParentsVisibleAndEnabled
        and Re_Focus then
    begin
        SetFocus;
        Invalidate;
    end;
end;

procedure TKnopka.CNCTLCOLORBTN(var Msg: TMessage);
begin
    //inherited;
    Msg.Result := 0;
    Invalidate;
end;

constructor TKnopka.Create(AOwner: TComponent);
begin
    inherited;
    Color := clBtnFace;
    ClickRestoreFocus := TRUE;
    FAnimation := TRUE;
    NLabel := TNLabel.Create(Self);
    UPic := TUniPic.Create(Self);
    FPicture := TPicture.Create;
    UPic.OnGetPicture := GetAPictureToUPic;

    //NLabel.Parent := Self;

    NLabel.Alignment := taCenter;
    NLabel.Layout := tlCenter;
    NLabel.WrapSyllabel := TRUE;
    NLabel.InScale := TRUE;
    NLabel.Transparent := TRUE;

    {--------}
    NLabel.FrameRadius := 2;
    NLabel.FrameStyle := fsNone;

    NLabel.ParentFont := TRUE;
    NLabel.ParentColor := TRUE;

    UPic.OnResize := UPicResize;
    UPic.Transparent := FALSE;

    FMarginsPic := 8;
    MarginsText := 2;

    Width := 90;
    Height := 25;
    Layout := klLabelRight;
    PicTransparent := TRUE;
    TextScale := ktsInFont;
    TextAlignment := taCenter;
    //FocusDotFrameAroundText := TRUE;

    RepeatFirst := 400;
    RepeatDelay := 200;
    //Caption := 'OK';

    DisableWhileClick := TRUE;
    DisableFormWhileClick := TRUE;
end;

procedure TKnopka.CreateParams(var aParams: TCreateParams);
begin
    inherited;
    ControlStyle := ControlStyle - [csDoubleClicks];
    //aParams.Style := aParams.Style or BS_OWNERDRAW;
end;

procedure TKnopka.CreateWnd;
begin
    inherited;
    NLabel.Caption := FCaption;
end;

destructor TKnopka.Destroy;
begin
    FreeAndNil(FPicture);
    inherited;
end;

procedure TKnopka.GetAPictureToUPic(Sender: TObject; var P: TPicture);
begin
    P := FPicture;
end;

{function TKnopka.GetGlyph: TBitmap;
begin
    Result := nil;
    if  (Picture.Graphic <> nil) and not Picture.Graphic.Empty
    and (Picture.Graphic is TBitmap) then
        Result := Picture.Bitmap;
end;}

function TKnopka.GetImageIndex: Integer;
begin
    Result := UPic.ImageIndex;
end;

function TKnopka.GetImageList: TImageList;
begin
    Result := UPic.ImageList;
end;

function TKnopka.GetPicture: TPicture;
begin
    Result := FPicture;
end;

{procedure TKnopka.SetAutoSize(const Value: Boolean);
begin
  FAutoSize := Value;
  NLabel.AutoSize := Value;
  if  Value then
      AdjustSizeAuto;
end;}

function TKnopka.GetStdImage: TStdImage;
begin
    Result := UPic.StdImage;
end;

function TKnopka.Get_Default: Boolean;
begin
    Result := inherited Default;
end;

procedure TKnopka.Invalidate;
begin
    inherited;
end;

procedure TKnopka.Loaded;
begin
    inherited;
    if FPicture <> nil then
        UPic.Picture.Assign(FPicture);
end;

function TKnopka.ParentForm: TForm;
var C: TControl;
begin
    C := Self;
    while C <> nil do
    begin
        C := C.Parent;
        if C is TForm then
        begin
            Result := C as TForm;
            Exit;
        end;
    end;
    Result := nil;
end;

procedure TKnopka.PassFocusNextControl;
var TopParent: TControl;
begin
    TopParent := Self;
    while TopParent.Parent <> nil do
        TopParent := TopParent.Parent;
    TopParent.Perform(WM_NEXTDLGCTL, 0, 0);
end;

procedure TKnopka.SetAnimation(const Value: Boolean);
begin
    FAnimation := Value;
    if Value then
        DisableFormWhileClick := FALSE;
end;

procedure TKnopka.SetAutoThreaded(const Value: Boolean);
begin
    FAutoThreaded := Value;
end;

procedure TKnopka.SetCaption(const Value: String);
begin
    HandleNeeded;
    FCaption := Value;
    NLabel.Caption := Value;
    if AutoSize then
        AdjustSizeAuto;
    Invalidate;
end;

procedure TKnopka.SetClickRestoreFocus(const Value: Boolean);
begin
    FClickRestoreFocus := Value;
end;

procedure TKnopka.SetColor(const Value: TColor);
begin
    FColor := Value;
    Invalidate;
end;

procedure TKnopka.SetDisableFormWhileClick(const Value: Boolean);
begin
    FDisableFormWhileClick := Value;
end;

procedure TKnopka.SetDisableWhileClick(const Value: Boolean);
begin
    FDisableWhileClick := Value;
end;

procedure TKnopka.SetEnabled(E: Boolean);
begin
    inherited;
    UPic.Enabled := E;
    Invalidate;
end;

procedure TKnopka.SetFlat(const Value: Boolean);
var F: DWORD;
begin
    FFlat := Value;
    HandleNeeded;
    F := GetWindowLong(Handle, GWL_STYLE);
    if Value then
        F := F or BS_FLAT
    else
        F := F and not BS_FLAT;
    SetWindowLong(Handle, GWL_STYLE, F);
    Invalidate;
end;

procedure TKnopka.SetFocusDotFrameAroundText(const Value: Boolean);
begin
    FFocusDotFrameAroundText := Value;
    Invalidate;
end;

{procedure TKnopka.SetGlyph(const Value: TBitmap);
begin
    if  (Picture.Graphic = nil) or (Picture.Graphic.Empty) then
        Picture.Assign( Value );
end;}

procedure TKnopka.SetImageIndex(const Value: Integer);
begin
    UPic.ImageIndex := Value;
    Invalidate;
end;

procedure TKnopka.SetImageList(const Value: TImageList);
begin
    UPic.ImageList := Value;
    Invalidate;
end;

procedure TKnopka.SetImagesDisabled(const Value: TImageList);
begin
    FImagesDisabled := Value;
    Invalidate;
end;

procedure TKnopka.SetLayout(const Value: TKnopkaLayout);
begin
    FLayout := Value;
    if AutoSize then
        AdjustSizeAuto;
    Invalidate;
end;

procedure TKnopka.SetMarginsPic(const Value: Integer);
begin
    FMarginsPic := Value;
    Invalidate;
end;

procedure TKnopka.SetMarginsText(const Value: Integer);
begin
    FMarginsText := Value;
    NLabel.MarginLeft := Value;
    NLabel.MarginTop := Value;
    NLabel.MarginRight := Value;
    NLabel.MarginBottom := Value;
    Invalidate;
end;

procedure TKnopka.SetModalResult(const Value: TModalResult);
begin
    FModalResult := Value;
end;

procedure TKnopka.SetPicTransparent(const Value: Boolean);
begin
    FPicTransparent := Value;
    Invalidate;
end;

procedure TKnopka.SetPicture(const Value: TPicture);
var Bmp: TBitmap;
begin
    if not (csLoading in ComponentState)
        and (csDesigning in ComponentState) then
    begin
        if Assigned(Value) and Assigned(Value.Graphic)
            and not Value.Graphic.Empty
            and (Value.Graphic is TBitmap)
            and (Value.Graphic.Width = 2 * Value.Graphic.Height) then
        begin
            case MessageBox(0, 'Divide glyph half? (Only left side will be assigned if yes).',
                'Loading picture', MB_YESNOCANCEL) of
                ID_YES: begin
                        Bmp := TBitmap.Create;
                        TRY
                            Bmp.Assign(Value.Bitmap);
                            Bmp.Width := Bmp.Height;
                            FPicture.Assign(Bmp);
                        FINALLY
                            Bmp.Free;
                        END;
                        Invalidate;
                        Exit;
                    end;
                ID_NO: ;
                ID_CANCEL: Exit;
            end;
        end;
    end;
    FPicture.Assign(Value.Graphic);
    if (FPicture.Graphic <> nil)
        and not (FPicture.Graphic.Empty) then
        StdImgShowSize := 0;
    Invalidate;
end;

procedure TKnopka.SetRepeatDelay(const Value: Integer);
begin
    FRepeatDelay := Value;
end;

procedure TKnopka.SetRepeatFirst(const Value: Integer);
begin
    FRepeatFirst := Value;
end;

procedure TKnopka.SetRepeatPress(const Value: Boolean);
begin
    FRepeatPress := Value;
end;

procedure TKnopka.SetSpeedButton(const Value: Boolean);
begin
    FSpeedButton := Value;
    if Value and not (csDesigning in ComponentState)
        and HandleAllocated and (GetFocus = Handle) then
        PassFocusNextControl;
    if TRUE then
        TabStop := FALSE
    else
        TabStop := TRUE;
end;

procedure TKnopka.SetStdImage(const Value: TStdImage);
begin
    UPic.StdImage := Value;
    Invalidate;
end;

procedure TKnopka.SetStdImgShowSize(const Value: Integer);
begin
    FStdImgShowSize := Value;
    UPic.StdImgSmall := Value < 24;
    Invalidate;
end;

procedure TKnopka.SetTextAlignment(const Value: TAlignment);
begin
    FTextAlignment := Value;
    NLabel.Alignment := Value;
    Invalidate;
end;

procedure TKnopka.SetTextScale(const Value: TKnopkaTextScale);
begin
    FTextScale := Value;
    CASE Value OF
        ktsInFont: begin
                NLabel.InScale := FALSE;
                NLabel.InFontScale := TRUE;
            end;
        ktsInScale: begin
                NLabel.InScale := TRUE;
                NLabel.InFontScale := FALSE;
            end;
        ktsNone: begin
                NLabel.InScale := FALSE;
                NLabel.InFontScale := FALSE;
            end;
    END;
    Invalidate;
end;

procedure TKnopka.SetWrapping(const Value: TKnopkaWrap);
begin
    FWrapping := Value;
    CASE Value OF
        kwWords: begin
                NLabel.WordWrap := TRUE;
                NLabel.WrapLetters := FALSE;
                NLabel.WrapSyllabel := FALSE;
            end;
        kwSyllabels: begin
                NLabel.WordWrap := FALSE;
                NLabel.WrapLetters := FALSE;
                NLabel.WrapSyllabel := TRUE;
            end;
        kwLetters: begin
                NLabel.WordWrap := FALSE;
                NLabel.WrapLetters := TRUE;
                NLabel.WrapSyllabel := FALSE;
            end;
        kwNone: begin
                NLabel.WordWrap := FALSE;
                NLabel.WrapLetters := FALSE;
                NLabel.WrapSyllabel := FALSE;
            end;
    END;
    Invalidate;
end;

procedure TKnopka.Set_Default(const Value: Boolean);
begin
    inherited Default := Value;
    Invalidate;
    if Value and not (csDesigning in ComponentState) then
    begin
        FAniStart := GetTickCount;
        if (FAniTimer = 0) and FAnimation then
            FAniTimer := SetTimer(Handle, 117, 100, nil);
    end;
end;

procedure TKnopka.UPicResize(Sender: TObject);
begin
    Invalidate;
end;

procedure TKnopka.WMCaptureChanged(var M: TMessage);
var H: THandle;
begin
    if not (csDesigning in ComponentState) and
        (M.LParam = 0) and (FLastHWnd <> 0) then
    begin
        H := FLastHWnd;
        FLastHWnd := 0;
        Windows.SetFocus(H);
    end;
    inherited;
end;

procedure TKnopka.WMClose(var M: TMessage);
begin
    if FTimer <> 0 then
    begin
        KillTimer(Handle, FTimer);
        FTimer := 0;
    end;
    if FAniTimer <> 0 then
    begin
        KillTimer(Handle, FAniTimer);
        FAniTimer := 0;
    end;
    inherited;
end;

procedure TKnopka.WMEnable(var M: TMessage);
begin
    if FProcessingOnClick and (DisableFormWhileClick or DisableWhileClick) then
    begin
        FEnableAfterOnClick := M.WParam <> 0;
        M.Result := 0;
    end
    else
        inherited;
end;

procedure TKnopka.WMGetText(var M: TMessage);
var Buf: PChar;
    L, Lcopy: Integer;
begin
    Buf := PChar(M.lParam);
    L := M.wParam;
    if FDrawing_NoCaptionNow then
        L := 0;
    M.Result := 0;
    if (Buf <> nil) and (L > 0) then
    begin
        Lcopy := L - 1;
        if Lcopy > Length(FCaption) then
            Lcopy := Length(FCaption);
        move(FCaption[1], Buf[0], Lcopy);
        Buf[Lcopy] := #0;
        M.Result := Lcopy;
    end;
end;

procedure TKnopka.WMKillFocus(var M: TMessage);
begin
    inherited;
    Invalidate;
end;

procedure TKnopka.WMPaint(var M: TMessage);
var PicDrawn: Boolean;
    procedure DoPaint;
    var DC: HDC;
        C: TCanvas;
        R1, R2: TRect;
        PicPresent: Boolean;
        B: TBitmap;
        F: DWORD;
        theme: THandle;
        P: TPoint;
        x, h, w: Integer;
        ants_anim: Boolean;
        Fm: TCustomForm;
    begin
        PicDrawn := FALSE;
        if M.Msg = WM_PAINT then
        begin
            if ForceCanvas <> 0 then
                DC := ForceCanvas
            else
                DC := GetDC(Handle);
        end
        else
            DC := M.WParam;
        C := TCanvas.Create;
        C.Handle := DC;
        C.Brush.Color := Color;
        R1 := Rect(0, 0, Width, Height);
        C.FillRect(R1);

        FDrawing_NoCaptionNow := TRUE;

        theme := 0;
        if (ThemeOK) and IsThemeActive() then
            theme := OpenThemeData(Handle, 'button');
        if theme <> 0 then
        begin
            F := PBS_NORMAL;
            if Default then
                F := F or PBS_DEFAULTED;
            if not Enabled then
                F := F or PBS_DISABLED;
            if FProcessingOnClick or FIsPushed then
                F := F or PBS_PRESSED;
            GetCursorPos(P);
            P := ScreenToClient(P);
            {if  PtInRect( R1, P ) then
                F := F or PBS_HOT;}
            DrawThemeBackground(theme, DC, BP_PUSHBUTTON,
                F, R1, nil);
            CloseThemeData(theme);
            R2 := R1;
        end
        else
        begin
            F := DFCS_BUTTONPUSH;
            if Flat then
                F := DFCS_FLAT or F
            else
                F := DFCS_BUTTONPUSH;
            {if  Focused then
                F := F or DFCS_HOT;}
            if not Enabled then
                F := F or DFCS_INACTIVE;
            if FProcessingOnClick or FIsPushed then
                F := F or DFCS_PUSHED;
            F := F or DFCS_ADJUSTRECT;
            R2 := R1;
            if Default or Focused then
            begin
                C.Brush.Color := clWindowText;
                if Focused then
                    C.Brush.Color := clBlue;
                C.FrameRect(R1);
                {C.MoveTo( R1.Left, R1.Bottom-1 );
                C.LineTo( R1.Right-1, R1.Bottom-1 );
                C.LineTo( R1.Right-1, R1.Top );}
                InflateRect(R1, -1, -1);
            end;
            DrawFrameControl(DC, R1, DFC_BUTTON, F);
            C.Brush.Color := Color;
            C.Brush.Style := bsSolid;
            C.FillRect(R1);
            R1 := R2;
        end;

        if not DisableBtnAnimation then
            if (FAniTimer <> 0)
                and (GetParentForm(Self).Enabled)
                then
            begin
                if FProcessingOnClick then
                begin
                    x := Abs(Integer(GetTickCount - FAniStart)) div 100;
                    x := x mod 8;
                    h := ClientHeight;
                    C.Pen.Color := clNavy;
                    while x < Width + h do
                    begin
                        C.MoveTo(x, 0);
                        C.LineTo(x - h, h);
                        inc(x, 8);
                    end;
                end;
                if (FocusedTime <> 0) then
                begin
                    x := Abs(Integer(GetTickCount - FocusedTime)) * ClientWidth div FocusAnimLength;
                    h := ClientHeight;
                    C.Pen.Color := clNavy + (((clBlue - clNavy) shr 16) * x div ClientWidth) shl 16;
                    C.MoveTo(x, 0);
                    C.LineTo(x, h);
                    inc(x, 7);
                    C.Pen.Width := 2;
                    C.MoveTo(x, 0);
                    C.LineTo(x, h);
                    C.Pen.Width := 0;
                    inc(x, 7);
                    C.MoveTo(x, 0);
                    C.LineTo(x, h);
                end;
                ants_anim := FALSE;
                if Focused and (GetForegroundWindow = GetParentForm(Self).Handle) then
                    ants_anim := TRUE
                else
                    if Default then
                    begin
                        Fm := GetParentForm(Self);
                        if Assigned(Fm.ActiveControl)
                            and (Fm.ActiveControl.Focused)
                            and ((Fm.ActiveControl is TButton) or (Fm.ActiveControl is TBitBtn))
                            and (Fm.ActiveControl <> Self) then
                        else
                            ants_anim := TRUE;
                    end;
                if not Enabled then
                    ants_anim := FALSE;
                if ants_anim then
                begin
                    x := Abs(Integer(GetTickCount - FAniStart)) div 200;
                    h := ClientHeight;
                    w := ClientWidth;
                    {x := x mod 4;
                    //C.Pen.Color := clNavy;
                    while x < w do
                    begin
                        C.Pixels[ x, 0 ] := clSilver;
                        C.Pixels[ w-x, h-1 ] := clSilver;
                        inc( x, 4 );
                    end;}
                    x := x mod 4;
                    while x < h do
                    begin
                        C.Pixels[w - 1, x] := clSilver;
                        C.Pixels[0, h - x] := clSilver;
                        inc(x, 4);
                    end;
                end;
            end;

        //inherited;
        FDrawing_NoCaptionNow := FALSE;
        if DisableAllIconsOnKnopka and (Caption <> '') then
            PicPresent := FALSE
        else
        begin
            if ImageList <> nil then
                UPic.ImageList := ImageList;
            PicPresent := (UPic.ImageList <> nil)
                or (UPic.Picture <> nil)
                and (UPic.Picture.Graphic <> nil)
                and (not UPic.Picture.Graphic.Empty)
                or (StdImgShowSize > 0);
        end;
        if PicPresent then
        begin
            if UPic.ImageList <> nil then
            begin
                UPic.Width := UPic.ImageList.Width;
                UPic.Height := UPic.ImageList.Height;
            end
            else
                if (UPic.Picture.Graphic <> nil)
                    and not UPic.Picture.Graphic.Empty then
                begin
                    UPic.Width := UPic.Picture.Width;
                    UPic.Height := UPic.Picture.Height;
                end
                else
                begin
                    UPic.Width := StdImgShowSize;
                    UPic.Height := StdImgShowSize;
                end;
            if Caption = '' then
            begin
                InflateRect(R2, -FMarginsPic, -FMarginsPic);
            end
            else
                CASE Layout OF
                    klLabelTop:
                        begin
                            R2.Bottom := R2.Bottom - FMarginsPic;
                            R1.Bottom := R2.Bottom - UPic.Height;
                            R2.Top := R1.Bottom;
                            R2.Left := R2.Left + FMarginsPic;
                            R2.Right := R2.Right - FMarginsPic;
                        end;
                    klLabelBottom:
                        begin
                            R1.Top := R1.Top + FMarginsPic + UPic.Height;
                            R2.Bottom := R1.Top;
                            R2.Top := R2.Top + FMarginsPic;
                            R2.Left := R2.Left + FMarginsPic;
                            R2.Right := R2.Right - FMarginsPic;
                        end;
                    klLabelRight:
                        begin
                            R1.Left := R1.Left + FMarginsPic + UPic.Width;
                            R2.Right := R1.Left;
                            R2.Left := R2.Left + FMarginsPic;
                            R2.Top := R2.Top + FMarginsPic;
                            R2.Bottom := R2.Bottom - FMarginsPic;
                        end;
                    klLabelLeft:
                        begin
                            R1.Right := R1.Right - FMarginsPic - UPic.Width;
                            R2.Right := R2.Right - FMarginsPic;
                            R2.Left := R2.Right - UPic.Width;
                            R2.Top := R2.Top + FMarginsPic;
                            R2.Bottom := R2.Bottom - FMarginsPic;
                        end;
                END;
            if (Layout in [klLabelTop, klLabelBottom])
                or (Caption = '') then
                OffsetRect(R2, (R2.Right - R2.Left - UPic.Width) div 2, 0);
            if (Layout in [klLabelLeft, klLabelRight])
                or (Caption = '') then
                OffsetRect(R2, 0, (R2.Bottom - R2.Top - UPic.Height) div 2);
        end
        else
        begin
            InflateRect(R1, -FMarginsText, -FMarginsText);
        end;
        NLabel.Color := Color;
        NLabel.Font.Assign(Font);
        if Focused and FocusDotFrameAroundText then
        begin
            NLabel.FrameStyle := fsDotted;
            NLabel.FrameInflate := 4;
        end
        else
        begin
            NLabel.FrameStyle := fsNone;
        end;
        if not Enabled then
            NLabel.Font.Color := clBtnHighlight;
        NLabel.PaintToCanvas(C, R1);
        if not Enabled then
        begin
            OffsetRect(R1, -1, -1);
            NLabel.Font.Color := clBtnShadow;
            NLabel.PaintToCanvas(C, R1);
        end;
        NLabel.BoundsRect := R1;
        if PicPresent then
        begin
            if not Enabled and (ImagesDisabled <> nil) then
            begin
                ImagesDisabled.DrawingStyle := dsTransparent;
                ImagesDisabled.Draw(C, R2.Left, R2.Top, ImageIndex, TRUE)
            end
            else if ImageList <> nil then
            begin
                ImageList.DrawingStyle := dsTransparent;
                ImageList.Draw(C, R2.Left, R2.Top, ImageIndex, Enabled)
            end
            else
                if (UPic.Picture.Graphic <> nil)
                    and not UPic.Picture.Graphic.Empty then
                begin
                    UPic.Picture.Graphic.Transparent := PicTransparent;
                    if not Enabled then
                    begin
                        B := TBitmap.Create();
                        TRY
                            B.Width := Upic.Picture.Width;
                            B.Height := UPic.Picture.Height;
                            B.PixelFormat := pf32bit;
                            B.Canvas.Brush.Color := ColorToRGB(Color);
                            B.Canvas.FillRect(Rect(0, 0, B.Width, B.Height));
                            B.Canvas.Draw(0, 0, UPic.Picture.Graphic);
                            GrayBitmap(B, ColorToRGB(Color));
                            B.Transparent := TRUE;
                            C.Draw(R2.Left, R2.Top, B);
                        FINALLY
                            B.Free;
                        END;
                    end
                    else
                        C.Draw(R2.Left, R2.Top, UPic.Picture.Graphic);
                end
                else
                begin
                    //UPic.Transparent := TRUE;
                    UPic.Color := Color;
                    UPic.PaintToCanvas(C, R2);
                end;
            PicDrawn := TRUE;
        end;
        if (M.Msg = WM_PAINT) and (DC <> ForceCanvas) then
            ReleaseDC(Handle, DC);
        C.Free;
    end;
begin
    TRY
        DoPaint;
        if (Caption = '') then
            if ForceCanvas <> 0 then
            begin
                DoPaint;
            end;
    EXCEPT
        DoPaint;
    END;
    M.Result := 0;
    ValidateRect(Handle, nil);
end;

procedure TKnopka.WMPrint(var M: TMessage);
begin
    WMPaint(M);
end;

procedure TKnopka.WMSetFocus(var M: TMessage);
begin
    if not (csDesigning in ComponentState) and
        SpeedButton then
    begin
        M.Result := 0;
        FLastHWnd := M.WParam;
    end
    else
    begin
        FocusedTime := GetTickCount;
        if FAniTimer <> 0 then
        begin
            KillTimer(Handle, FAniTimer);
            FAniTimer := 0;
        end;
        if (FAniTimer = 0) and FAnimation then
        begin
            FAniStart := GetTickCount;
            FAniTimer := SetTimer(Handle, 117, 30, nil);
        end;
        inherited;
    end;
    Invalidate;
end;

procedure TKnopka.WMSetText(var M: TMessage);
begin
    FCaption := PChar(M.lParam);
end;

procedure TKnopka.WMTextLen(var M: TMessage);
begin
    if not FDrawing_NoCaptionNow then
        M.Result := Length(FCaption)
    else
        M.Result := 0;
end;

procedure TKnopka.WM_KeyDown(var M: TMsg);
begin
    inherited;
    Invalidate;
end;

procedure TKnopka.WM_LButtonDblClk(var M: TMsg);
begin
    inherited;
    Invalidate;
end;

procedure TKnopka.WM_LButtonDown(var M: TMsg);
begin
    inherited;
    FIsPushed := TRUE;
    Invalidate;
    if RepeatPress then
    begin
        SetCapture(Handle);
        FTimer := SetTimer(Handle, 1, RepeatFirst, nil);
    end;
end;

procedure TKnopka.WM_LButtonUp(var M: TMsg);
begin
    if FTimer <> 0 then
    begin
        KillTimer(Handle, FTimer);
        FTimer := 0;
        ReleaseCapture;
        Click;
        Invalidate;
    end
    else
    begin
        inherited;
        if FIsPushed then
        begin
            FIsPushed := FALSE;
        end;
        Invalidate;
    end;
end;

procedure TKnopka.WM_Size(var M: TMsg);
begin
    inherited;
    Invalidate;
end;

procedure TKnopka.WM_Timer(var M: TMessage);
begin
    if M.WParam = 117 then
    begin
        //Invalidate;
        if (FocusedTime = 0) and (FAniTimer <> 0)
            and (not Default or not Visible or not Enabled or
            (GetParentForm(Self) <> nil) and
            (not GetParentForm(Self).Visible {or
            not GetParentForm( Self ).Enabled}
            ))
            and not FProcessingOnClick then
        begin
            KillTimer(Handle, FAniTimer);
            FAniTimer := 0;
        end;
        if (FocusedTime <> 0) and (GetTickCount - FocusedTime > FocusAnimLength) then
        begin
            FocusedTime := 0;
            if not FProcessingOnClick then
                if FAniTimer <> 0 then
                begin
                    KillTimer(Handle, FAniTimer);
                    FAniTimer := 0;
                end;
            if (FAniTimer = 0) and FAnimation then
                if Focused or Default then
                begin
                    FAniStart := GetTickCount;
                    FAniTimer := SetTimer(Handle, 117, 100, nil);
                end;
        end;
    end
    else
    begin
        Click;
        M.Result := 0;
        if M.WParam = 1 then
        begin
            KillTimer(Handle, FTimer);
            FTimer := SetTimer(Handle, 2, RepeatDelay, nil);
        end;
        SetCapture(Handle);
    end;
    Invalidate;
    //inherited;
end;

{var Buf: array[ 0..255 ] of Word;
    Ind: Integer = 0;}

procedure TKnopka.WndProc(var M: TMessage);
{var s: String;
    i: Integer;
    F: THandle;
    L: DWORD;}
begin
    inherited;
    {if  M.Msg = $B00C then
    asm
        nop
    end;}
    {Buf[Ind] := M.Msg;
    inc( Ind );
    if  (Ind >= Sizeof(Buf)) or (GetKeyState( VK_SHIFT ) < 0) then
    begin
        s := '';
        for i := 0 to High(Buf) do
        begin
            s := s + ' ' + IntToHex( Buf[i], 4 );
        end;
        s := s + #13#10;
        Ind := 0;
        F := CreateFile( 'c:\log_knopka_msgs.txt', GENERIC_WRITE or GENERIC_READ,
             FILE_SHARE_READ, nil, OPEN_ALWAYS, FILE_ATTRIBUTE_NORMAL, 0 );
        if  F <> INVALID_HANDLE_VALUE then
        begin
            WriteFile( F, s[1], Length(s), L, nil );
            CloseHandle(F);
        end;
    end;}
end;

{ TPnKnopka }

procedure TPnKnopka.AdjustKnopkaPosition;
begin
    if Parent <> nil then
        Knopka.BoundsRect := Rect(SpaceLeft, SpaceTop,
            ClientWidth - SpaceRight, ClientHeight - SpaceBottom);
end;

procedure TPnKnopka.Click;
begin
    if not Enabled then Exit;
    Knopka.Click;
end;

constructor TPnKnopka.Create(AOwner: TComponent);
begin
    inherited;
    BevelOuter := bvNone;
    Width := 90;
    Height := 25;
    Knopka := TKnopka.Create(Self);
    Knopka.Parent := Self;
    AdjustKnopkaPosition;
end;

{function TPnKnopka.GetAutoSize: Boolean;
begin
    Result := Knopka.AutoSize;
end;}

function TPnKnopka.GetActionA: TBasicAction;
begin
    Result := Knopka.Action;
end;

function TPnKnopka.GetAnimation: Boolean;
begin
    Result := Knopka.Animation;
end;

function TPnKnopka.GetAutoSize0: Boolean;
begin
    Result := FALSE;
end;

function TPnKnopka.GetBiDiModeK: TBiDiMode;
begin
    Result := Knopka.BiDiMode;
end;

function TPnKnopka.GetBorderWidth0: Integer;
begin
    Result := 0;
end;

function TPnKnopka.GetCaption: String;
begin
    Result := Knopka.Caption;
end;

function TPnKnopka.GetColorKnopka: TColor;
begin
    Result := Knopka.Color;
end;

function TPnKnopka.GetDisableFormWhileClick: Boolean;
begin
    Result := Knopka.DisableFormWhileClick;
end;

function TPnKnopka.GetDisableWhileClick: Boolean;
begin
    Result := Knopka.DisableWhileClick;
end;

function TPnKnopka.GetDotFrame: Boolean;
begin
    Result := Knopka.FocusDotFrameAroundText;
end;

function TPnKnopka.GetEnableAfterOnClick: Boolean;
begin
    Result := Knopka.EnableAfterClick;
end;

function TPnKnopka.GetEnabledK: Boolean;
begin
    Result := Knopka.Enabled;
end;

function TPnKnopka.GetFlat: Boolean;
begin
    Result := Knopka.Flat;
end;

function TPnKnopka.GetImageIndex: Integer;
begin
    Result := Knopka.ImageIndex;
end;

function TPnKnopka.GetImageList: TImageList;
begin
    Result := Knopka.ImageList;
end;

function TPnKnopka.GetImgDisabled: TImageList;
begin
    Result := Knopka.ImagesDisabled;
end;

function TPnKnopka.GetIsCancel: Boolean;
begin
    Result := Knopka.Cancel;
end;

function TPnKnopka.GetIsDefault: Boolean;
begin
    Result := Knopka.Default;
end;

function TPnKnopka.GetKnopkaWrap: TKnopkaWrap;
begin
    Result := Knopka.Wrapping;
end;

function TPnKnopka.GetLayout: TKnopkaLayout;
begin
    Result := Knopka.Layout;
end;

function TPnKnopka.GetMarginsPic: Integer;
begin
    Result := Knopka.MarginsPic;
end;

function TPnKnopka.GetMarginsText: Integer;
begin
    Result := Knopka.MArginsText;
end;

function TPnKnopka.GetModalResult: TModalResult;
begin
    Result := Knopka.ModalResult;
end;

function TPnKnopka.GetOnClickK: TNotifyEvent;
begin
    Result := Knopka.OnClick;
end;

function TPnKnopka.GetOnDblClickK: TNotifyEvent;
begin
    Result := Knopka.OnDblClick;
end;

function TPnKnopka.GetOnKeyDown: TKeyEvent;
begin
    Result := Knopka.OnKeyDown;
end;

function TPnKnopka.GetOnKeyPress: TKeyPressEvent;
begin
    Result := Knopka.OnKeyPress;
end;

function TPnKnopka.GetOnKeyUp: TKeyEvent;
begin
    Result := Knopka.OnKeyUp;
end;

function TPnKnopka.GetOnMouseDownK: TMouseEvent;
begin
    Result := Knopka.OnMouseDown;
end;

function TPnKnopka.GetOnMouseMoveK: TMouseMoveEvent;
begin
    Result := Knopka.OnMouseMove;
end;

function TPnKnopka.GetOnMouseUpK: TMouseEvent;
begin
    Result := Knopka.OnMouseUp;
end;

function TPnKnopka.GetPicture: TPicture;
begin
    Result := Knopka.Picture;
end;

function TPnKnopka.GetRepeatDelay: Integer;
begin
    Result := Knopka.RepeatDelay;
end;

function TPnKnopka.GetRepeatFirst: Integer;
begin
    Result := Knopka.RepeatFirst;
end;

function TPnKnopka.GetRepeatPress: Boolean;
begin
    Result := Knopka.RepeatPress;
end;

function TPnKnopka.GetSpeedButton: Boolean;
begin
    Result := Knopka.SpeedButton;
end;

function TPnKnopka.GetStdImg: TStdImage;
begin
    Result := Knopka.StdImage;
end;

function TPnKnopka.GetStdImgShowSz: Integer;
begin
    Result := Knopka.StdImgShowSize;
end;

function TPnKnopka.GetTabStopK: Boolean;
begin
    Result := Knopka.TabStop;
end;

function TPnKnopka.GetTextAlignment: TAlignment;
begin
    Result := Knopka.TextAlignment;
end;

function TPnKnopka.GetTextScale: TKnopkaTextScale;
begin
    Result := Knopka.TextScale;
end;

{procedure TPnKnopka.SetAutoSize(const Value: Boolean);
begin
    Knopka.AutoSize := Value;
    if  Value then
    begin
        Width := SpaceLeft + Knopka.Width + SpaceRight;
        Height := SpaceTop + Knopka.Height + SpaceBottom;
    end;
end;}

function TPnKnopka.Get_AutoThreaded: Boolean;
begin
    Result := Knopka.AutoThreaded;
end;

procedure TPnKnopka.SetAction(const Value: TBasicAction);
begin
    Knopka.Action := Value;
end;

procedure TPnKnopka.SetAnimation(const Value: Boolean);
begin
    Knopka.Animation := Value;
end;

procedure TPnKnopka.SetBiDiModeK(const Value: TBiDiMode);
begin
    Knopka.BiDiMode := Value;
end;

procedure TPnKnopka.SetCaption(const Value: String);
begin
    TRY
        inherited Caption := '';
        Knopka.Caption := Value;
    EXCEPT
        inherited Caption := '';
        Knopka.Caption := Value;
    END;
    {if  AutoSize then
        AutoSize := TRUE;}
end;

procedure TPnKnopka.SetColorKnopka(const Value: TColor);
begin
    Knopka.Color := Value;
end;

procedure TPnKnopka.SetDisableFormWhileClick(const Value: Boolean);
begin
    Knopka.DisableFormWhileClick := Value;
end;

procedure TPnKnopka.SetDisableWhileClick(const Value: Boolean);
begin
    Knopka.DisableWhileClick := Value;
end;

procedure TPnKnopka.SetDotFrame(const Value: Boolean);
begin
    Knopka.FocusDotFrameAroundText := Value;
end;

procedure TPnKnopka.SetEnableAfterOnClick(const Value: Boolean);
begin
    Knopka.EnableAfterClick := Value;
end;

procedure TPnKnopka.SetEnabledK(const Value: Boolean);
begin
    Knopka.Enabled := Value;
    inherited Enabled := Value;
end;

procedure TPnKnopka.SetFlat(const Value: Boolean);
begin
    Knopka.Flat := Value;
end;

procedure TPnKnopka.SetImageIndex(const Value: Integer);
begin
    Knopka.ImageIndex := Value;
    {if  AutoSize then
        AutoSize := TRUE;}
end;

procedure TPnKnopka.SetImageList(const Value: TImageList);
begin
    Knopka.ImageList := Value;
    {if  AutoSize then
        AutoSize := TRUE;}
end;

procedure TPnKnopka.SetImgDisabled(const Value: TImageList);
begin
    Knopka.ImagesDisabled := Value;
    {if  AutoSize then
        AutoSize := TRUE;}
end;

procedure TPnKnopka.SetIsCancel(const Value: Boolean);
begin
    Knopka.Cancel := Value;
end;

procedure TPnKnopka.SetIsDefault(const Value: Boolean);
begin
    Knopka.Default := Value;
end;

procedure TPnKnopka.SetKnopkaWrap(const Value: TKnopkaWrap);
begin
    Knopka.Wrapping := Value;
    {if  AutoSize then
        AutoSize := TRUE;}
end;

procedure TPnKnopka.SetLayout(const Value: TKnopkaLayout);
begin
    Knopka.Layout := Value;
    {if  AutoSize then
        AutoSize := TRUE;}
end;

procedure TPnKnopka.SetMarginsPic(const Value: Integer);
begin
    Knopka.MarginsPic := Value;
    {if  AutoSize then
        AutoSize := TRUE;}
end;

procedure TPnKnopka.SetMarginsText(const Value: Integer);
begin
    Knopka.MarginsText := Value;
    {if  AutoSize then
        AutoSize := TRUE;}
end;

procedure TPnKnopka.SetModalResult(const Value: TModalResult);
begin
    Knopka.ModalResult := Value;
end;

procedure TPnKnopka.SetOnClickK(const Value: TNotifyEvent);
begin
    Knopka.OnClick := Value;
end;

procedure TPnKnopka.SetOnDblClickK(const Value: TNotifyEvent);
begin
    Knopka.OnDblClick := Value;
    inherited OnDblClick := Value;
end;

procedure TPnKnopka.SetOnKeyDown(const Value: TKeyEvent);
begin
    Knopka.OnKeyDown := Value;
end;

procedure TPnKnopka.SetOnKeyPress(const Value: TKeyPressEvent);
begin
    Knopka.OnKeyPress := Value;
end;

procedure TPnKnopka.SetOnKeyUp(const Value: TKeyEvent);
begin
    Knopka.OnKeyUp := Value;
end;

procedure TPnKnopka.SetOnMouseDownK(const Value: TMouseEvent);
begin
    Knopka.OnMouseDown := Value;
    inherited OnMouseDown := Value;
end;

procedure TPnKnopka.SetOnMouseMoveK(const Value: TMouseMoveEvent);
begin
    Knopka.OnMouseMove := Value;
    inherited OnMouseMove := Value;
end;

procedure TPnKnopka.SetOnMouseUpK(const Value: TMouseEvent);
begin
    Knopka.OnMouseUp := Value;
    inherited OnMouseUp := Value;
end;

procedure TPnKnopka.SetPicture(const Value: TPicture);
begin
    Knopka.Picture := Value;
    {if  AutoSize then
        AutoSize := TRUE;}
end;

procedure TPnKnopka.SetRepeatDelay(const Value: Integer);
begin
    Knopka.RepeatDelay := Value;
end;

procedure TPnKnopka.SetRepeatFirst(const Value: Integer);
begin
    Knopka.RepeatFirst := Value;
end;

procedure TPnKnopka.SetRepeatPress(const Value: Boolean);
begin
    Knopka.RepeatPress := Value;
end;

procedure TPnKnopka.SetSpaceBottom(const Value: Integer);
begin
    FSpaceBottom := Value;
    AdjustKnopkaPosition;
    {if  AutoSize then
        AutoSize := TRUE;}
end;

procedure TPnKnopka.SetSpaceLeft(const Value: Integer);
begin
    FSpaceLeft := Value;
    AdjustKnopkaPosition;
    {if  AutoSize then
        AutoSize := TRUE;}
end;

procedure TPnKnopka.SetSpaceRight(const Value: Integer);
begin
    FSpaceRight := Value;
    AdjustKnopkaPosition;
    {if  AutoSize then
        AutoSize := TRUE;}
end;

procedure TPnKnopka.SetSpaceTop(const Value: Integer);
begin
    FSpaceTop := Value;
    AdjustKnopkaPosition;
    {if  AutoSize then
        AutoSize := TRUE;}
end;

procedure TPnKnopka.SetSpeedButton(const Value: Boolean);
begin
    Knopka.SpeedButton := Value;
end;

procedure TPnKnopka.SetStdImg(const Value: TStdImage);
begin
    Knopka.StdImage := Value;
end;

procedure TPnKnopka.SetStdImgShowSz(const Value: Integer);
begin
    Knopka.StdImgShowSize := Value;
end;

procedure TPnKnopka.SetTabStopK(const Value: Boolean);
begin
    Knopka.TabStop := Value;
end;

procedure TPnKnopka.SetTextAlignment(const Value: TAlignment);
begin
    Knopka.TextAlignment := Value;
end;

procedure TPnKnopka.SetTextScale(const Value: TKnopkaTextScale);
begin
    Knopka.TextScale := Value;
    {if  AutoSize then
        AutoSize := TRUE;}
end;

procedure TPnKnopka.Set_AutoThreaded(const Value: Boolean);
begin
    Knopka.AutoThreaded := Value;
end;

procedure TPnKnopka.WMGetText(var M: TMessage);
begin
    Knopka.FDrawing_NoCaptionNow := FDrawing_NoCaptionNow;
    M.Result := Knopka.Perform(WM_GETTEXT, M.WParam, M.LParam);
end;

procedure TPnKnopka.WMPaint(var M: TMessage);
begin
    AdjustKnopkaPosition;
    FDrawing_NoCaptionNow := TRUE;
    inherited;
    FDrawing_NoCaptionNow := FALSE;
end;

procedure TPnKnopka.WMSetText(var M: TMessage);
begin
    M.Result := Knopka.Perform(WM_SETTEXT, M.WParam, M.LParam);
end;


procedure TPnKnopka.WMTextLen(var M: TMessage);
begin
    Knopka.FDrawing_NoCaptionNow := FDrawing_NoCaptionNow;
    if not FDrawing_NoCaptionNow then
        M.Result := Knopka.Perform(WM_GETTEXTLENGTH, M.WParam, M.LParam)
    else
        M.Result := 0;
end;

{ TGrKnopka }

procedure TGrKnopka.Click;
var ParentTop: TControl;
    SelfCtl: TControl;
    Form: TCustomForm;
begin
    FWaitInterval := 0;
    FEnableAfterOnClick := Enabled;
    if DisableWhileClick then
        Enabled := FALSE;
    ParentTop := nil;
    if DisableFormWhileClick then
    begin
        ParentTop := Parent;
        while (ParentTop <> nil) and (ParentTop.Parent <> nil) do
            ParentTop := ParentTop.Parent;
        if (ParentTop <> nil)
            and (ParentTop.Visible) then
            ParentTop.Enabled := FALSE
        else
            ParentTop := nil;
    end;
    FProcessingOnClick := TRUE;
    TRY
        Invalidate;
        Update;
        //////////////////////////////////////////////////////////////////////////
        //inherited;
        Form := GetParentForm(Self);
        SelfCtl := Self;
        if (Parent <> nil)
            and (Parent is TPnKnopka) then
            SelfCtl := Parent;
        if Form <> nil then Form.ModalResult := ModalResult;
        //inherited Click;
        if Assigned(OnClick) and (Action <> nil)
            and (@OnClick <> @Action.OnExecute) then
            OnClick(SelfCtl)
        else if not (csDesigning in ComponentState) and (ActionLink <> nil) then
            ActionLink.Execute
        else if Assigned(OnClick) then
            OnClick(SelfCtl);
    FINALLY
        FProcessingOnClick := FALSE;
        if DisableWhileClick then
            Enabled := FEnableAfterOnClick;
        if DisableFormWhileClick
            and (ParentTop <> nil)
            and (ParentTop.Visible) then
            ParentTop.Enabled := TRUE;
        FIsPushed := FALSE;
        //////////////////////////////////////////////////////////////////////////
        Invalidate;
        Update;
        Invalidate;
    END;
end;

procedure TGrKnopka.CMTextChanged(var Message: TMessage);
begin
    inherited;
    Invalidate;
end;

constructor TGrKnopka.Create(AOwner: TComponent);
begin
    inherited;
    UPic := TUniPic.Create(Self);
    UPic.Visible := FALSE;

    Width := 90;
    Height := 25;

    RepeatFirst := 400;
    RepeatDelay := 200;
    MarginsPic := 8;
    MarginsText := 4;

    Layout := klLabelRight;
    PicTransparent := TRUE;
    TextScale := ktsInFont;
    TextAlignment := taCenter;

    ControlStyle := ControlStyle - [csDoubleClicks];
end;

destructor TGrKnopka.Destroy;
begin
    if FTimer <> 0 then
    begin
        KillTimer(0, Integer(@Self));
        FTimer := 0;
    end;
    inherited;
end;

function TGrKnopka.GetImageIndex: Integer;
begin
    Result := UPic.ImageIndex;
end;

function TGrKnopka.GetImageList: TImageList;
begin
    Result := UPic.ImageList;
end;

function TGrKnopka.GetPicture: TPicture;
begin
    Result := UPic.Picture;
end;

function TGrKnopka.GetStdImage: TStdImage;
begin
    Result := UPic.StdImage;
end;

procedure TGrKnopka.Invalidate;
var R: TRect;
begin
    inherited;
    if Parent = nil then Exit;
    if not Parent.HandleAllocated then Exit;
    R := Rect(Left, Top, Left + Width, Top + Height);
    InvalidateRect(Parent.Handle, @R, TRUE);
end;

procedure TGrKnopka.Paint;
var P: TPoint;
    F: DWORD;
    //DC, PDC: HDC;
    //Bitmap0: HBitmap;
begin
    //inherited;
      //---???---//
      //Parent.DoubleBuffered := TRUE;
    if Knopka0 = nil then
    begin
        Knopka0 := TKnopka.Create(nil);
        Knopka0.Visible := FALSE;
        Knopka0.AutoSize := FALSE;
        Knopka0.Wrapping := kwWords;
        Knopka0.TextScale := ktsInFont;
    end;

    Knopka0.Parent := Parent;
    Knopka0.Caption := Caption;
    Knopka0.BoundsRect := Rect(0, 0, Width, Height); // BoundsRect;
    Knopka0.Font := Font;
    Knopka0.Enabled := Enabled;
    Knopka0.Picture := Picture;
    Knopka0.ImageList := ImageList;
    Knopka0.ImageIndex := ImageIndex;
    Knopka0.ImagesDisabled := ImagesDisabled;
    Knopka0.PicTransparent := PicTransparent;
    Knopka0.StdImage := StdImage;
    Knopka0.StdImgShowSize := StdImgShowSize;
    Knopka0.FLat := Flat;
    Knopka0.Layout := Layout;
    Knopka0.Font := Font;
    Knopka0.HelpContext := HelpContext;
    Knopka0.MarginsPic := MarginsPic;
    Knopka0.MarginsText := MarginsText;
    Knopka0.TextScale := TextScale;
    Knopka0.Wrapping := Wrapping;
    Knopka0.TextAlignment := TextAlignment;
    Knopka0.FIsPushed := FIsPushed;

    GetCursorPos(P);
    P := ScreenToClient(P);
    F := 0;
    if (P.x >= 0) and (P.x < Width)
        and (P.y >= 0) and (P.y < Height) then
    begin
        if GetKeyState(VK_LBUTTON) < 0 then
            F := F or MK_LBUTTON;
        if GetKeyState(VK_RBUTTON) < 0 then
            F := F or MK_RBUTTON;
        if GetKeyState(VK_MBUTTON) < 0 then
            F := F or MK_MBUTTON;
        Knopka0.Perform(WM_MOUSEMOVE, F, WORD(P.x) or (WORD(P.y) shl 16));
        SetTrackingGrKnopka(Self);
    end;

    if FProcessingOnClick or FIsPushed then
        Knopka0.Perform(BM_SETSTATE, 1, 0);

    //Bitmap0 := CreateCompatibleBitmap( Canvas.Handle, Width, Height );
    //DC := CreateCompatibleDC( Canvas.Handle );
    //PDC := SelectObject( DC, Bitmap0 );

    Knopka0.FDrawing_NoCaptionNow := TRUE;
    Knopka0.Perform(WM_PRINTCLIENT, Canvas.Handle, //DC,
        PRF_CLIENT or PRF_NONCLIENT or PRF_ERASEBKGND);
    Knopka0.FDrawing_NoCaptionNow := FALSE;
    Knopka0.Caption := '';

    if FProcessingOnClick or FIsPushed then
        Knopka0.Perform(BM_SETSTATE, 0, 0);
    Knopka0.Parent := nil;

    //BitBlt( Canvas.Handle, 0, 0, Width, Height,
    //        DC, 0, 0, SRCCOPY );
    //DeleteObject( SelectObject( PDC, Bitmap0 ) );
    //DeleteObject( Bitmap0 );
end;

function TGrKnopka.ParentForm: TCustomForm;
var C: TControl;
begin
    C := Self;
    while C <> nil do
    begin
        if C is TCustomForm then break;
        C := C.Parent;
    end;
    Result := C as TCustomForm;
end;

procedure TGrKnopka.SetCancel(const Value: Boolean);
begin
    FCancel := Value;
end;

procedure TGrKnopka.SetDefault(const Value: Boolean);
begin
    FDefault := Value;
end;

procedure TGrKnopka.SetDisableFormWhileClick(const Value: Boolean);
begin
    FDisableFormWhileClick := Value;
end;

procedure TGrKnopka.SetDisableWhileClick(const Value: Boolean);
begin
    FDisableWhileClick := Value;
end;

procedure TGrKnopka.SetFlat(const Value: Boolean);
begin
    FFlat := Value;
end;

procedure TGrKnopka.SetHelpContext(const Value: THelpContext);
begin
    FHelpContext := Value;
end;

procedure TGrKnopka.SetImageIndex(const Value: Integer);
begin
    UPic.ImageIndex := Value;
    Invalidate;
end;

procedure TGrKnopka.SetImageList(const Value: TImageList);
begin
    UPic.ImageList := Value;
    Invalidate;
end;

procedure TGrKnopka.SetImagesDisabled(const Value: TImageList);
begin
    FImagesDisabled := Value;
end;

procedure TGrKnopka.SetLayout(const Value: TKnopkaLayout);
begin
    FLayout := Value;
end;

procedure TGrKnopka.SetMarginsPic(const Value: Integer);
begin
    FMarginsPic := Value;
end;

procedure TGrKnopka.SetMarginsText(const Value: Integer);
begin
    FMarginsText := Value;
end;

procedure TGrKnopka.SetModalResult(const Value: TModalResult);
begin
    FModalResult := Value;
end;

procedure TGrKnopka.SetPicTransparent(const Value: Boolean);
begin
    FPicTransparent := Value;
end;

procedure TGrKnopka.SetPicture(const Value: TPicture);
var Bmp: TBitmap;
begin
    if not (csLoading in ComponentState)
        and (csDesigning in ComponentState) then
    begin
        if Assigned(Value) and Assigned(Value.Graphic)
            and (Value.Graphic.Width = 2 * Value.Graphic.Height) then
        begin
            case MessageBox(0, 'Divide glyph half? (Only left side will be assigned if yes).',
                'Loading picture', MB_YESNOCANCEL) of
                ID_YES: begin
                        Bmp := TBitmap.Create;
                        TRY
                            Bmp.Assign(Value.Bitmap);
                            Bmp.Width := Bmp.Height;
                            UPic.Picture.Assign(Bmp);
                        FINALLY
                            Bmp.Free;
                        END;
                        Invalidate;
                        Exit;
                    end;
                ID_NO: ;
                ID_CANCEL: Exit;
            end;
        end;
    end;
    UPic.Picture := Value;
    Invalidate;
end;

procedure TGrKnopka.SetRepeatDelay(const Value: Integer);
begin
    FRepeatDelay := Value;
end;

procedure TGrKnopka.SetRepeatFirst(const Value: Integer);
begin
    FRepeatFirst := Value;
end;

procedure TGrKnopka.SetRepeatPress(const Value: Boolean);
begin
    FRepeatPress := Value;
end;

procedure TGrKnopka.SetStdImage(const Value: TStdImage);
begin
    UPic.StdImage := Value;
    Invalidate;
end;

procedure TGrKnopka.SetStdImgShowSz(const Value: Integer);
begin
    FStdImgShowSz := Value;
    Invalidate;
end;

procedure TGrKnopka.SetTextAlignment(const Value: TAlignment);
begin
    FTextAlignment := Value;
end;

procedure TGrKnopka.SetTextScale(const Value: TKnopkaTextScale);
begin
    FTextScale := Value;
end;

procedure TGrKnopka.SetWrapping(const Value: TKnopkaWrap);
begin
    FWrapping := Value;
end;

procedure TGrKnopka.WMEraseBkgd(var M: TWMEraseBkgnd);
begin
    //M.Result := 1;
    inherited;
end;

var TimeredGrKnopka: TGrKnopka;

procedure GrKnopkaTimerProc(
    hwnd: DWORD;
    uMsg: DWORD;
    idKnopka: TGrKnopka;
    dwTime: DWORD
    ); stdcall;
begin
    idKnopka := TimeredGrKnopka;
    if (idKnopka.FWaitInterval > 0)
        and (idKnopka.Enabled) then
    begin
        if GetTickCount >= idKnopka.FTimePressed + idKnopka.FWaitInterval then
        begin
            idKnopka.Click;
            idKnopka.FTimePressed := GetTickCount;
            idKnopka.FWaitInterval := idKnopka.RepeatDelay;
        end;
    end;
end;

procedure TGrKnopka.WMLButtonDown(var M: TWMMouse);
begin
    inherited;
    FIsPushed := TRUE;
    Invalidate;
    if RepeatPress then
    begin
        SetCapture(Parent.Handle);
        FTimePressed := GetTickCount;
        FWaitInterval := RepeatFirst;

        TimeredGrKnopka := Self;
        FTimer := SetTimer(0, Integer(Self), RepeatDelay, @GrKnopkaTimerProc);
    end;
end;

procedure TGrKnopka.WMLButtonUp(var M: TWMMouse);
begin
    FWaitInterval := 0;
    if FTimer <> 0 then
    begin
        KillTimer(0, FTimer);
        FTimer := 0;
    end;
    inherited;
    FIsPushed := FALSE;
    Invalidate;
end;

procedure TGrKnopka.WMLDblClick(var M: TWMLButtonDblClk);
begin
    inherited;
    //Invalidate;
    Click;
    //FIsPushed := TRUE;
    if RepeatPress and (FTimer = 0) then
    begin
        SetCapture(Parent.Handle);
        FTimePressed := GetTickCount;
        FWaitInterval := RepeatFirst;

        TimeredGrKnopka := Self;
        FTimer := SetTimer(0, Integer(Self), RepeatDelay, @GrKnopkaTimerProc);
    end;
end;

procedure TGrKnopka.WMMouseMove(var M: TWMMouse);
begin
    inherited;

    if (GetKeyState(VK_LBUTTON) >= 0)
        and (FWaitInterval > 0) then
    begin
        FWaitInterval := 0;
        if FTimer <> 0 then
        begin
            KillTimer(0, FTimer);
            FTimer := 0;
        end;
        FIsPushed := FALSE;
        Invalidate;
    end;

    //Invalidate;
    if (FWaitInterval > 0)
        and Enabled then
    begin
        if GetTickCount >= FTimePressed + FWaitInterval then
        begin
            Click;
            FTimePressed := GetTickCount;
            FWaitInterval := RepeatDelay;
        end;
    end;
end;

{ TThreadAuto }

procedure TThreadAuto.Execute;
begin
    try
        fExited := FALSE;
        try
            fOnClick(fSender);
        except on e: Exception do
            begin
                fException := e.Message;
            end;
        end;
    finally
        fExited := TRUE;
    end;
end;

initialization
    ThemeOK := InitThemeLibrary;

finalization
    FreeAndNil(Knopka0);
    if ThemeOK then
        FreeThemeLibrary;

end.
