//**  ,     VCL Delphi.
//**  Vladimir Kladov, 2000-2010
//**
//**      .
//**     /     
//**  2          
//**   GNU,  Free Software Foundation.
//**
//**       ,    
//** ,       ,   
//**         
//**   .      
//**     GNU.
//**
//**          
//**   GNU.     ,     Free
//** Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
//** MA 02111-1307, USA.
//******************************************************************************

{************************************************}
{                                                }
{   TNormalBand  TNormalLabel  }
{   "  "" "            }
{   Copyright (c) 2000-2003 by Vladimir Kladov   }
{                                                }
{            8.09.2003         }
{                                                }
{************************************************}
{           24.03.2005         }
{        DiagonalSplit         }
{ Transparent  TNormalLabel.                  }
{************************************************}
{           19.02.2010         }
{************************************************}

unit NormalBands;
{* TNormalBand, TNormalLabel -      
                                .
|<br><pre>
  (C) by Vladimir Kladov, 2002

  ,    ()     Delphi,
    TWinControl.        .
  (   rich-edit-,    
      ).
  ,      .     
       .
|</pre>}

{$I VCL_VERSIONS.INC}

interface

{$DEFINE BAND_PARENT_FONT}

uses
    Windows, Messages, SysUtils, Classes, Forms, Dialogs, Graphics, ExtCtrls, Controls,
    StdCtrls, math;

const
    Sqrt2 = 1.4142135623730950488016887242097;

type
    TFrame = (frLeft, frTop, frRight, frBottom);
    {*  . }
    TFrames = set of TFrame;
    {*   . }
    TFrameStyle = (fsSolid, fsDotted, fsNone);

    TNormalBand = class(TPanel)
        {* ,    . ,     ,
               .   ( 
           TNormalReport.OnPrint)  Caption   ()
           -  ,      TNormalReport.AddBand,
           TNormalReport.AddRight, TNormalReport.AddFooter    . }
    private
        FFrames: TFrames;
        FFrameWidth: Integer;
        FDoubleBuffered: Boolean;
        FTransparent: Boolean;
        FOnPrint: TNotifyEvent;
        procedure SetFrames(const Value: TFrames);
        procedure SetFrameWidth(const Value: Integer);
        procedure SetDoubleBuffered(const Value: Boolean);
        procedure SetTransparent(const Value: Boolean);
        procedure SetOnPrint(const Value: TNotifyEvent);
    protected
        procedure Paint; override;
    public
        constructor Create(AOwner: TComponent); override;
        function DesiredHeight: Integer;
        {*  .    ,  ,
                ,    ( )
                    .
              //    
            WordWrap. }
    published
        property Frames: TFrames read FFrames write SetFrames;
        {* .   ,     , 
            ,   ,      
           ( ). }
        property FrameWidth: Integer read FFrameWidth write SetFrameWidth;
        {*  . }
        property DoubleBuffered: Boolean read FDoubleBuffered write SetDoubleBuffered;
        property Transparent: Boolean read FTransparent write SetTransparent;
        property OnPrint: TNotifyEvent read FOnPrint write SetOnPrint;
    end;

    TTextOrientation = (toNormal, to90CW, to90ACW);

    TNormalLabel = class(TNormalBand)
        {*   .        
              Frames, ,   .  ,
                ,     
              . }
    private
        FFrames: TFrames;
        FFrameWidth: Integer;
        FLayout: TTextLayout;
        FWordWrap: Boolean;
        FAutosize: Boolean;
        FWrapSyllabel: Boolean;
        FMarginLeft: Integer;
        FMarginTop: Integer;
        FMarginBottom: Integer;
        FMarginRight: Integer;
        FFrameRadius: Integer;
        FDoubleBuffered: Boolean;
        FOnMouseLeave: TNotifyEvent;
        FOnMouseEnter: TNotifyEvent;
        FFrameColor: TColor;
        FMouseTransparent: Boolean;
        FTextOrientation: TTextOrientation;
        FDiagonalSplit: Char;
        FTransparent: Boolean;
        FWrapLetters: Boolean;
        FInScale: Boolean;
        FInFontScale: Boolean;
        FInFontMinSize: Integer;
        FTextRotateViaBitmap: Boolean;
        FInProportionalScale: Boolean;
        FFrameStyle: TFrameStyle;
        FFrameInflate: Integer;
        FOnDestroy: TNotifyEvent;
        procedure SetFrames(const Value: TFrames);
        procedure SetFrameWidth(const Value: Integer);
        procedure SetLayout(const Value: TTextLayout);
        procedure SetWordWrap(const Value: Boolean);
        procedure SetAutosize(const Value: Boolean); reintroduce;
        procedure SetCaption(const Value: TCaption);
        function GetCaption: TCaption;
        function GetParentFont: Boolean;
        procedure SetParentFont(const Value: Boolean);
        procedure SetWrapSyllabel(const Value: Boolean);
        procedure SetMarginBottom(const Value: Integer);
        procedure SetMarginLeft(const Value: Integer);
        procedure SetMarginRight(const Value: Integer);
        procedure SetMarginTop(const Value: Integer);
        procedure SetFrameRadius(const Value: Integer);
        procedure SetDoubleBuffered(const Value: Boolean);
        procedure SetOnMouseEnter(const Value: TNotifyEvent);
        procedure SetOnMouseLeave(const Value: TNotifyEvent);
        procedure SetFrameColor(const Value: TColor);
        procedure SetMouseTransparent(const Value: Boolean);
        procedure SetTextOrientation(const Value: TTextOrientation);
        procedure SetDiagonalSplit(const Value: Char);
        procedure SetTransparent(const Value: Boolean);
        procedure SetWrapLetters(const Value: Boolean);
        procedure SetInScale(const Value: Boolean);
        procedure SetInFontScale(const Value: Boolean);
        procedure SetInFontMinSize(const Value: Integer);
        procedure SetTextRotateViaBitmap(const Value: Boolean);
        procedure SetInProportionalScale(const Value: Boolean);
        procedure SetFrameStyle(const Value: TFrameStyle);
        procedure SetFrameInflate(const Value: Integer);
    protected
        FPaintWrapped: function(Canvas: TCanvas; var Rect: TRect;
            Calc_Rect: Boolean): Integer of object;
        FPainting, FPaintingWrapped: Boolean;
        CurRgn1, CurRgn2, CurRgn3: Integer;
        FRectLastPaint: TRect;
        FTextLastPaint: String;
        FFontSizeLastPaint: Integer;
        procedure PaintToCanvas(ACanvas: TCanvas; const Rect: TRect);
        procedure Paint; override;
        {*  ,     
             TNormalLabel. }

        function PaintWrapped(ACanvas: TCanvas; var Rect: TRect; Calc_Rect: Boolean): Integer;
        {*         .
               (   
               ). }
        function PaintWrapByLetters(ACanvas: TCanvas; var Rect: TRect; Calc_Rect: Boolean): Integer;
        {*         .
               (   
               ). }

        function CanAutoSize(var W, H: Integer): Boolean; override;
        function DesiredSize(var W, H: Integer; IfAutoSize: Boolean = TRUE): Boolean;
        procedure Resize; override;
        procedure DoAutosize;
        procedure CMFontChanged(var Message: TMessage); message CM_FONTCHANGED;
        procedure CMMouseEnter(var M: TMessage); message CM_MOUSEENTER;
        procedure CMMouseLeave(var M: TMessage); message CM_MOUSELEAVE;
        procedure HitTest(var M: TMessage); message WM_NCHITTEST;
        procedure Loaded; override;
    public
        constructor Create(AOwner: TComponent); override;
        property Canvas;

        function DesiredHeight: Integer;
        {*  .    ,  ,
                ,    ( )
                    .
              //    
            WordWrap. }
        function DesiredWidth: Integer;
        function ActualMarginLeft: Integer;
        function ActualMarginRight: Integer;
        function ActualMarginTop: Integer;
        function ActualMarginBottom: Integer;
        function ActualFrameRadius: Integer;
        function ActualFrameColor: TColor;
        destructor Destroy; override;
    published
        property Frames: TFrames read FFrames write SetFrames;
        {* . }

        property FrameWidth: Integer read FFrameWidth write SetFrameWidth;
        {*  . }

        property FrameRadius: Integer read FFrameRadius write SetFrameRadius;
        {*   .  0,  . 
           2 * FrameRadius > Height,    = Height div 2.
              > 0,     
            SQRT( 2 ) * ,      .  }

        property FrameStyle: TFrameStyle read FFrameStyle write SetFrameStyle;
        property FrameInflate: Integer read FFrameInflate write SetFrameInflate;

        property FrameColor: TColor read FFrameColor write SetFrameColor;
        {*  .  clNone ( ),    . }

        property Alignment;
        property Layout: TTextLayout read FLayout write SetLayout;
        {*    . }
        property TextOrientation: TTextOrientation read FTextOrientation write SetTextOrientation;
        property TextRotateViaBitmap: Boolean read FTextRotateViaBitmap write SetTextRotateViaBitmap;

        property WordWrap: Boolean read FWordWrap write SetWordWrap;
        {*   . }
        property WrapSyllabel: Boolean read FWrapSyllabel write SetWrapSyllabel;
        {*   .  ,    ,
              Alignment. }
        property WrapLetters: Boolean read FWrapLetters write SetWrapLetters;
        {*   .    . }

        property InScale: Boolean read FInScale write SetInScale;
        {*         ,  
              .  ,   WordWrap  WrapSyllabel. }
        property InProportionalScale: Boolean read FInProportionalScale write SetInProportionalScale;

        property InFontScale: Boolean read FInFontScale write SetInFontScale;
        {*      (  ,  
            InFontMinSize),  ,      ,
             . }
        property InFontMinSize: Integer read FInFontMinSize write SetInFontMinSize;


        property Autosize: Boolean read FAutosize write SetAutosize;
        {*   . }

        property Caption: TCaption read GetCaption write SetCaption;
        {*  () . }

        property ParentFont: Boolean read GetParentFont write SetParentFont;
        {* TRUE,    . }

        property MarginLeft: Integer read FMarginLeft write SetMarginLeft;
        {*      .   2 . }

        property MarginRight: Integer read FMarginRight write SetMarginRight;
        {*      .   2 . }

        property MarginTop: Integer read FMarginTop write SetMarginTop;
        {*      .   2 . }

        property MarginBottom: Integer read FMarginBottom write SetMarginBottom;
        {*      .   2 . }

        property DoubleBuffered: Boolean read FDoubleBuffered write SetDoubleBuffered;

        property OnMouseEnter: TNotifyEvent read FOnMouseEnter write SetOnMouseEnter;
        property OnMouseLeave: TNotifyEvent read FOnMouseLeave write SetOnMouseLeave;
        property MouseTransparent: Boolean read FMouseTransparent write SetMouseTransparent;

        property DiagonalSplit: Char read FDiagonalSplit write SetDiagonalSplit;
        property Transparent: Boolean read FTransparent write SetTransparent;
        property OnDestroy: TNotifyEvent read FOnDestroy write FOnDestroy;
    protected
        procedure AssignTo(C: TPersistent); override;
        procedure assignNormalLabel(L: TNormalLabel);
    end;

implementation

uses RotBmp;

var TmpBmp: TBitmap; //    11

procedure SwapInt(var I, J: Integer);
var Tmp: Integer;
begin
    Tmp := I; I := J; J := Tmp;
end;

function IsVowel(C: Char): Boolean;
begin
    Result := C in ['a', 'e', 'i', 'o', 'u', 'y', 'A', 'E', 'I', 'O', 'U', 'Y',
        '', '', '', '', '', '', '', '', '', '',
        '', '', '', '', '', '', '', '', '', ''];
end;

function LastSyllableWidth(const Word: String; OnEndOfWord: Boolean): Integer;
var I: Integer;
begin
    Result := 0;
    I := 1;
    if OnEndOfWord then I := 2;
    for I := Length(Word) - I downto 1 do
    begin
        if IsVowel(Word[I]) and not ((I > 1) and IsVowel(Word[I - 1])) then
        begin
            Result := Length(Word) - I;
            if Result = 1 then
                if not IsVowel(Word[Length(Word)]) or OnEndOfWord then
                    continue; //        .
            if Result >= 3 then
                if not IsVowel(Word[I + 1]) and not IsVowel(Word[I + 2]) then
                begin
                    Dec(Result);
                    if (I + 2 <= Length(Word)) and (Word[I + 2] in ['', '', '', '']) then
                        Dec(Result);
                end;
            if (Result >= 1) and not OnEndOfWord or (Result >= 2) and OnEndOfWord then
                Exit;
        end;
        Result := 0;
    end;
end;

{ TNormalBand }

constructor TNormalBand.Create(AOwner: TComponent);
begin
    inherited;
    Color := clWhite;
    {$IFDEF BAND_PARENT_FONT}
    if (AOwner <> nil) and (AOwner is TForm) and
        ((AOwner as TForm).Color <> clWhite) then
    begin
        Font.Assign((AOwner as TForm).Font);
        if Font.Name = 'MS Sans Serif' then
            Font.Name := 'Arial';
        if Font.Size = 8 then
            Font.Size := 10;
        if Font.Color = clWindowText then
            Font.Color := clBlack;
    end;
    {$ELSE}
    Font.Name := 'Arial';
    Font.Size := 10;
    Font.Color := clBlack;
    {$ENDIF}
    BorderStyle := bsNone;
    BevelOuter := bvNone;
    BevelInner := bvNone;
    Caption := ' ';
    FFrameWidth := 1;
end;

function TNormalBand.DesiredHeight: Integer;
var I, H: Integer;
    C: TComponent;
begin
    Result := Height;
    for I := 0 to Owner.ComponentCount - 1 do
    begin
        C := Owner.Components[I];
        if (C is TNormalLabel) and
            ((C as TNormalLabel).Parent = Self) and
            ((C as TNormalLabel).Align in [alLeft, alRight, alClient]) then
        begin
            H := (C as TNormalLabel).DesiredHeight;
            if H > Result then
                Result := H;
        end;
    end;
end;

procedure TNormalBand.Paint;
begin
    if Transparent then
        Exit;
    inherited;
    Canvas.Brush.Color := Font.Color;
    if frLeft in Frames then
        Canvas.FillRect(Rect(0, 0, FrameWidth, Height));
    if frTop in Frames then
        Canvas.FillRect(Rect(0, 0, Width, FrameWidth));
    if frRight in Frames then
        Canvas.FillRect(Rect(Width - FrameWidth, 0, Width, Height));
    if frBottom in Frames then
        Canvas.FillRect(Rect(0, Height - FrameWidth, Width, Height));
end;

procedure TNormalBand.SetDoubleBuffered(const Value: Boolean);
begin
    FDoubleBuffered := Value;
end;

procedure TNormalBand.SetFrames(const Value: TFrames);
begin
    FFrames := Value;
    Invalidate;
end;

procedure TNormalBand.SetFrameWidth(const Value: Integer);
begin
    FFrameWidth := Value;
    Invalidate;
end;

procedure TNormalBand.SetOnPrint(const Value: TNotifyEvent);
begin
    FOnPrint := Value;
end;

procedure TNormalBand.SetTransparent(const Value: Boolean);
begin
    FTransparent := Value;
end;

{ TNormalLabel }

function TNormalLabel.ActualFrameColor: TColor;
begin
    Result := FFrameColor;
    if Result = clNone then
        Result := Font.Color;
end;

function TNormalLabel.ActualFrameRadius: Integer;
begin
    if FrameRadius < Height div 2 then Result := FrameRadius
    else Result := Height div 2;
end;

function TNormalLabel.ActualMarginBottom: Integer;
begin
    Result := MarginBottom + FrameWidth;
    if FrameRadius > 0 then
        Result := Result + Round((1 - 1 / Sqrt2) * ActualFrameRadius);
end;

function TNormalLabel.ActualMarginLeft: Integer;
begin
    Result := MarginLeft + FrameWidth;
    if FrameRadius > 0 then
        Result := Result + Round((1 - 1 / Sqrt2) * ActualFrameRadius);
end;

function TNormalLabel.ActualMarginRight: Integer;
begin
    Result := MarginRight + FrameWidth;
    if FrameRadius > 0 then
        Result := Result + Round((1 - 1 / Sqrt2) * ActualFrameRadius);
end;

function TNormalLabel.ActualMarginTop: Integer;
begin
    Result := MarginTop + FrameWidth;
    if FrameRadius > 0 then
        Result := Result + Round((1 - 1 / Sqrt2) * ActualFrameRadius);
end;

procedure TNormalLabel.assignNormalLabel(L: TNormalLabel);
begin
    Align := L.Align;
    Alignment := L.Alignment;
    Anchors := L.Anchors;
    BevelInner := L.BevelInner;
    BevelOuter := L.BevelOuter;
    BiDiMode := L.BiDiMode;
    BorderStyle := L.BorderStyle;
    BorderWidth := L.BorderWidth;
    Caption := L.Caption;
    Color := L.Color;
    Constraints.Assign(L.Constraints);
    Cursor := L.Cursor;
    DiagonalSplit := L.DiagonalSplit;
    DockSite := L.DockSite;
    DoubleBuffered := L.DoubleBuffered;
    DragCursor := L.DragCursor;
    DragKind := L.DragKind;
    DragMode := L.DragMode;
    Enabled := L.Enabled;
    Font.Assign(L.Font);
    FrameColor := L.FrameColor;
    FrameRadius := L.FrameRadius;
    Frames := L.Frames;
    FrameWidth := L.FrameWidth;
    FullRepaint := L.FullRepaint;
    Height := L.Height;
    HelpContext := L.HelpContext;
    Hint := L.Hint;
    InFontScale := L.InFontScale;
    Layout := L.Layout;
    Left := L.Left;
    Locked := L.Locked;
    MarginBottom := L.MarginBottom;
    MarginLeft := L.MarginLeft;
    MarginRight := L.MarginRight;
    MarginTop := L.MarginTop;
    MouseTransparent := L.MouseTransparent;
    //Name := ;
    ParentBiDiMode := L.ParentBiDiMode;
    ParentColor := L.ParentColor;
    ParentFont := L.ParentFont;
    ParentShowHint := L.ParentShowHint;
    PopupMenu := L.PopupMenu;
    ShowHint := L.ShowHint;
    TabOrder := L.TabOrder;
    TabStop := L.TabStop;
    Tag := L.Tag;
    TextOrientation := L.TextOrientation;
    Top := L.Top;
    Transparent := L.Transparent;
    UseDockManager := L.UseDockManager;
    Visible := L.Visible;
    Width := L.Width;
    WordWrap := L.WordWrap;
    WrapLetters := L.WrapLetters;
    WrapSyllabel := L.WrapSyllabel;
end;

procedure TNormalLabel.AssignTo(C: TPersistent);
var L: TNormalLabel;
begin
    if C is TNormalLabel then
    begin
        L := C as TNormalLabel;
        L.assignNormalLabel(Self);
    end
    else
        inherited;
end;

function TNormalLabel.CanAutoSize(var W, H: Integer): Boolean;
begin
    if (Parent = nil) or (not Parent.HandleAllocated) then
        Result := FALSE
    else
        Result := DesiredSize(W, H);
end;

procedure TNormalLabel.CMFontChanged(var Message: TMessage);
begin
    inherited;
    DoAutosize;
end;

procedure TNormalLabel.CMMouseEnter(var M: TMessage);
begin
    if Assigned(OnMouseEnter) then
        OnMouseEnter(Self);
end;

procedure TNormalLabel.CMMouseLeave(var M: TMessage);
begin
    if Assigned(OnMouseLeave) then
        OnMouseLeave(Self);
end;

constructor TNormalLabel.Create(AOwner: TComponent);
begin
    inherited;
    MarginLeft := 2;
    MarginTop := 2;
    MarginRight := 2;
    MarginBottom := 2;
    if (AOwner <> nil) and (AOwner is TForm) and
        ((AOwner as TForm).Color <> clWhite) then
    begin
        Color := clWhite;
        Font.Assign((AOwner as TForm).Font);
        Font.Color := clBlack;
        {Font.Name := 'Arial';
        Font.Size := 10;}
    end;
    inherited ParentFont := TRUE;
    BevelOuter := bvNone;
    BevelInner := bvNone;
    FFrameWidth := 1;
    FFrameColor := clNone;
    FInFontMinSize := 4;
    Autosize := TRUE;
    FInProportionalScale := TRUE;
end;

function TNormalLabel.DesiredHeight: Integer;
var W, H: Integer;
begin
    W := Width;
    H := Height;
    DesiredSize(W, H, FALSE);
    Result := H;
end;

//var tst_cnt: Integer;

function TNormalLabel.DesiredSize(var W, H: Integer; IfAutoSize: Boolean = TRUE): Boolean;
var R, R1: TRect;
    F: DWORD;
    DTP: TDrawTextParams;
begin
    Result := FALSE;
    if TextOrientation <> toNormal then
        SwapInt(W, H);
    TRY

        if IfAutoSize then
            if not FAutosize then Exit;
        R := Rect(0, 0, W, H);
        if WrapLetters then
        begin
            W := ClientWidth;
            TmpBmp.Canvas.Font.Assign(Font);
            H := PaintWrapByLetters(TmpBmp.Canvas, R, TRUE);
        end
        ELSE
            if WrapSyllabel then
            begin
                if Parent = nil then
                    W := Width
                else
                    W := ClientWidth;
                TmpBmp.Canvas.Font.Assign(Font);
                H := PaintWrapped(TmpBmp.Canvas, R, TRUE);
            end
            ELSE
                if Trim(Caption) <> '' then
                begin
                    //Tmp.Canvas.Font.Assign( Font );
                    TmpBmp.Canvas.Font.Assign(Font);
                    F := 0;
                    case Layout of
                        tlTop: F := DT_TOP;
                        tlCenter: F := DT_VCENTER;
                        tlBottom: F := DT_BOTTOM;
                    end;
                    case Alignment of
                        taLeftJustify: F := F or DT_LEFT;
                        taCenter: F := F or DT_CENTER;
                        taRightJustify: F := F or DT_RIGHT;
                    end;
                    case WordWrap of
                        FALSE: F := F or DT_SINGLELINE;
                        TRUE: F := F or DT_WORDBREAK;
                    end;
                    R := ClientRect;
                    Inc(R.Top, ActualMarginTop);
                    Inc(R.Left, ActualMarginLeft);
                    Dec(R.Right, ActualMarginRight);
                    Dec(R.Bottom, ActualMarginBottom);
                    DTP.cbSize := Sizeof(DTP);
                    DTP.iTabLength := 8;
                    DTP.iLeftMargin := ActualMarginLeft;
                    DTP.iRightMargin := ActualMarginRight;
                    DTP.uiLengthDrawn := 0;
                    R1 := R;
                    Windows.DrawTextEx(TmpBmp.Canvas.Handle, PChar(Caption), Length(Caption),
                        R1, F or DT_CALCRECT, @DTP);
                    if not (Align in [alTop, alBottom, alClient]) or not IfAutoSize then
                        W := R1.Right - R1.Left + ActualMarginLeft + ActualMarginRight;
                    if not (Align in [alLeft, alRight, alClient]) or not IfAutoSize then
                        H := R1.Bottom - R1.Top + ActualMarginTop + ActualMarginBottom;
                end;
    FINALLY
        if TextOrientation <> toNormal then
            SwapInt(W, H);
    END;
    Result := (W <> Self.Width) or (H <> Self.Height);
end;

function TNormalLabel.DesiredWidth: Integer;
var W, H: Integer;
begin
    W := Width;
    H := Height;
    DesiredSize(W, H, FALSE);
    Result := W;
end;

destructor TNormalLabel.Destroy;
begin
    if Assigned(FOnDestroy) then
        FOnDestroy(Self);
    inherited;
end;

procedure TNormalLabel.DoAutosize;
var W, H: Integer;
begin
    if Parent = nil then Exit;
    if csLoading in ComponentState then Exit;
    if not FAutosize then Exit;
    if Align = alClient then Exit;
    W := Width;
    H := Height;
    if CanAutoSize(W, H) then
    begin
        Width := W;
        Height := H;
    end;
end;

function TNormalLabel.GetCaption: TCaption;
begin
    Result := inherited Caption;
end;

function TNormalLabel.GetParentFont: Boolean;
begin
    Result := inherited ParentFont;
end;

procedure TNormalLabel.HitTest(var M: TMessage);
begin
    if MouseTransparent and not (csDesigning in ComponentState) then
        M.Result := HTTRANSPARENT
    else
        inherited;
end;

procedure TNormalLabel.Loaded;
begin
    inherited;
    DoAutosize;
end;

procedure TNormalLabel.Paint;
var MF, MFDC: HDC;
    Bmp2, Bmp3: TBitmap;
    DC0: HDC;
    R: TRect;
    C, C2: TCanvas;
    NLtmp: TNormalLabel;
begin
    if TextOrientation = toNormal then
        PaintToCanvas(Canvas, ClientRect)
    else
        if not TextRotateViaBitmap then
        begin
            DC0 := GetDC(0);
            //MF := 0;
            MFDC := 0;
            Bmp2 := nil;
            try
                R := Rect(0, 0,
                    MulDiv(Height, GetDeviceCaps(DC0, VERTSIZE) * 100, GetDeviceCaps(DC0, VERTRES)),
                    MulDiv(Width, GetDeviceCaps(DC0, HORZSIZE) * 100, GetDeviceCaps(DC0, HORZRES))
                    );
                MF := CreateEnhMetaFile(DC0, nil, @R, '');
                ReleaseDC(0, DC0);

                R := Rect(0, 0, R.Bottom, R.Right);
                Bmp2 := TBitmap.Create;
                Bmp2.Width := Height * 10;
                Bmp2.Height := Width * 10;
                Bmp2.PixelFormat := pf32bit;
                C := TCanvas.Create;
                C2 := Bmp2.Canvas;
                C.Handle := MF;
                C.Refresh;
                C2.Refresh;
                NLtmp := TNormalLabel.Create(nil);
                TRY
                    C.Font.Assign(Font);
                    C.Brush.Color := Canvas.Brush.Color;

                    NLtmp.Assign(Self);
                    NLtmp.Width := Height;
                    NLtmp.Height := Width;
                    NLtmp.TextOrientation := toNormal;
                    NLtmp.PaintToCanvas(C, Rect(0, 0, NLtmp.Width, NLtmp.Height));
                    MFDC := CloseEnhMetaFile(MF);
                    MF := 0;

                    PlayEnhMetaFile(C2.Handle, MFDC, Rect(0, 0, Bmp2.Width, Bmp2.Height));
                    DeleteEnhMetaFile(MFDC);
                    MFDC := 0;

                    Bmp3 := TBitmap.Create;
                    try
                        Bmp3.PixelFormat := Bmp2.PixelFormat;
                        Bmp3.Width := ClientRect.Bottom;
                        Bmp3.Height := ClientRect.Right;
                        SetStretchBltMode(Bmp3.Canvas.Handle, halftone);
                        Bmp3.Canvas.CopyRect(Rect(0, 0, Bmp3.Width, Bmp3.Height),
                            Bmp2.Canvas, Rect(0, 0, Bmp2.Width, Bmp2.Height));

                        //Bmp3.SaveToFile( 'C:\test.bmp' );

                        CASE TextOrientation OF
                            to90CW: RotateBmpRight(Bmp3);
                            to90ACW: RotateBmpLeft(Bmp3);
                        END;
                        SetStretchBltMode(Canvas.Handle, halftone);
                        Canvas.CopyRect(ClientRect, Bmp3.Canvas,
                            Rect(0, 0, Bmp3.Width, Bmp3.Height));

                    finally
                        Bmp3.Free;
                    end;


                FINALLY
                    NLtmp.Free;
                    if MF <> 0 then
                        DeleteEnhMetaFile(MF);

                    C.Free;
                END;
            finally
                if DC0 <> 0 then
                    ReleaseDC(0, DC0);
                if MFDC <> 0 then
                    DeleteEnhMetaFile(MFDC);
                Bmp2.Free;
            end;
        end
        else
        begin
            Bmp2 := TBitmap.Create;
            NLtmp := TNormalLabel.Create(nil);
            TRY

                NLtmp.Assign(Self);
                NLtmp.Width := Height;
                NLtmp.Height := Width;
                NLtmp.TextOrientation := toNormal;
                Bmp2.Width := Height;
                Bmp2.Height := Width;
                Bmp2.PixelFormat := pf32bit;
                NLtmp.PaintToCanvas(Bmp2.Canvas, Rect(0, 0, Bmp2.Width, Bmp2.Height));
                CASE TextOrientation OF
                    to90CW: RotateBmpRight(Bmp2);
                    to90ACW: RotateBmpLeft(Bmp2);
                END;
                Canvas.CopyRect(ClientRect, Bmp2.Canvas, ClientRect);
            FINALLY
                Bmp2.Free;
                NLtmp.Free;
            END;
        end;
end;

procedure TNormalLabel.PaintToCanvas(ACanvas: TCanvas; const Rect: TRect);
var F: DWORD;
    R, R1, R0: TRect;
    DTP: TDrawTextParams;
    OldPen: TPen;
    Rgn: HRgn;
    B: TBitmap;
    DC0, MF, MFDC: HDC;
    C: TCanvas;
    W, H: Integer;
    KW, KH: Double;
    Cap: String;
label TryAgain;
begin
    Cap := Caption;
    if (Length(Cap) > 10) and (Cap[1] = '') and (Cap[2] = '') then
        asm
      nop
        end;
    ACanvas.Font.Assign(Font);
    TryAgain:
    if  InFontScale and (FTextLastPaint = Caption) and
        (FRectLastPaint.Right = Width) and (FRectLastPaint.Bottom = Height) then
        ACanvas.Font.Size := FFontSizeLastPaint;

    ACanvas.Brush.Color := Color;
    if Transparent then
        ACanvas.Brush.Style := bsClear
    else
        ACanvas.Brush.Style := bsSolid;
    if FrameRadius > 0 then
    begin
        if not FPainting and (
            (CurRgn1 <> Width) or (CurRgn2 <> Height) or (CurRgn3 <> ActualFrameRadius)) then
        begin
            FPainting := TRUE;
            CurRgn1 := Rect.Right - Rect.Left;
            CurRgn2 := Rect.Bottom - Rect.Top;
            CurRgn3 := ActualFrameRadius;
            TRY
                Rgn := CreateRoundRectRgn(Rect.Left, Rect.Top,
                    Rect.Right - Rect.Left + 1, Rect.Bottom - Rect.Top + 1,
                    ActualFrameRadius * 2, ActualFrameRadius * 2);
                if Parent <> nil then
                    SetWindowRgn(Handle, Rgn, TRUE);
            FINALLY
                FPainting := FALSE;
            END;
        end;

        OldPen := TPen.Create;
        TRY
            OldPen.Assign(ACanvas.Pen);
            begin
                ACanvas.Pen.Color := FrameColor; // ACanvas.Brush.Color;
                ACanvas.Pen.Width := 0; // FrameWidth;
                CASE FrameStyle OF
                    fsSolid: ACanvas.Pen.Style := psSolid;
                    fsDotted: ACanvas.Pen.Style := psDot;
                else ACanvas.Pen.Style := psClear;
                END;
                if (FrameWidth > 0) and
                    (ColorToRGB(ActualFrameColor) <> ColorToRGB(Brush.Color)) then
                    ACanvas.Brush.Color := ActualFrameColor
                else
                    ACanvas.Brush.Color := Color;
                if Transparent then
                    ACanvas.Brush.Style := bsClear;
                R := Rect;
                InflateRect(R, -FrameInflate, -FrameInflate);
                ACanvas.RoundRect(R.Left, R.Top,
                    R.Right {- R.Left+1},
                    R.Bottom {- R.Top+1},
                    ActualFrameRadius * 2, ActualFrameRadius * 2);
                //Canvas.FillRect( Rect( 0, 0, Width, Height ) );
            end;
            if (FrameWidth > 0) and
                (ColorToRGB(ActualFrameColor) <> ColorToRGB(Brush.Color))
                and not Transparent then
            begin
                ACanvas.Brush.Color := Color;
                ACanvas.RoundRect(Rect.Left + FrameWidth, Rect.Top + FrameWidth,
                    Rect.Right {- Rect.Left} - FrameWidth,
                    Rect.Bottom {- Rect.Top} - FrameWidth,
                    ActualFrameRadius * 2 - FrameWidth, ActualFrameRadius * 2 - FrameWidth);
            end;
            ACanvas.Brush.Style := bsClear;
        FINALLY
            ACanvas.Pen.Assign(OldPen);
            OldPen.Free;
        END;
    end
    else
    begin
        if not Transparent then
            ACanvas.FillRect(Rect);
        if not (DiagonalSplit in [' ', #0]) then
        begin
            OldPen := TPen.Create;
            TRY
                OldPen.Assign(ACanvas.Pen);
                begin
                    ACanvas.Pen.Color := ACanvas.Font.Color;
                    ACanvas.Pen.Width := FrameWidth;
                    ACanvas.Pen.Style := psSolid;
                    if DiagonalSplit = '\' then
                    begin
                        ACanvas.MoveTo(Rect.Left, Rect.Top);
                        ACanvas.LineTo(Rect.Right, Rect.Bottom);
                    end
                    else
                    begin
                        ACanvas.MoveTo(Rect.Right, Rect.Top);
                        ACanvas.LineTo(Rect.Left, Rect.Bottom);
                    end;
                end;
            FINALLY
                ACanvas.Pen.Assign(OldPen);
                OldPen.Free;
            END;
        end;
    end;

    R := Rect;
    F := 0;
    if Trim(Caption) <> '' then
    begin
        R1 := R;
        if WrapLetters then
            PaintWrapByLetters(ACanvas, R1, TRUE)
        else
            if WrapSyllabel then
                PaintWrapped(ACanvas, R1, TRUE)
            else
            begin
                Inc(R.Top, ActualMarginTop);
                Inc(R.Left, ActualMarginLeft);
                Dec(R.Right, ActualMarginRight);
                Dec(R.Bottom, ActualMarginBottom);
                case Layout of
                    tlTop: F := DT_TOP;
                    tlCenter: F := DT_VCENTER;
                    tlBottom: F := DT_BOTTOM;
                end;
                case Alignment of
                    taLeftJustify: F := F or DT_LEFT;
                    taCenter: F := F or DT_CENTER;
                    taRightJustify: F := F or DT_RIGHT;
                end;
                case WordWrap of
                    FALSE: begin
                            F := F or DT_SINGLELINE;
                            if (Alignment <> taLeftJustify) and
                                (R.Right - R.Left <= ACanvas.TextWidth(Caption)) then
                                F := F and not (DT_CENTER or DT_RIGHT);
                        end;
                    TRUE: F := F or DT_WORDBREAK;
                end;
                DTP.cbSize := Sizeof(DTP);
                DTP.iTabLength := 8;
                DTP.iLeftMargin := 0;
                DTP.iRightMargin := 0;
                DTP.uiLengthDrawn := 0;
                if FrameRadius > 0 then
                    ACanvas.Brush.Style := bsClear
                else
                begin
                    ACanvas.Brush.Color := Color;
                    ACanvas.Brush.Style := bsSolid;
                end;
                if (Layout <> tlTop) or InFontScale or InScale then
                begin
                    R1 := R;
                    Windows.DrawText(ACanvas.Handle, PChar(Caption), Length(Caption),
                        R1, F or DT_CALCRECT);
                    if R1.Bottom - R1.Top < R.Bottom - R.Top then
                    begin
                        if Layout = tlCenter then
                            OffsetRect(R1, 0, ((R.Bottom - R.Top) - (R1.Bottom - R1.Top)) div 2)
                        else
                            if Layout = tlBottom then
                                OffsetRect(R1, 0, (R.Bottom - R.Top) - (R1.Bottom - R1.Top));
                    end;
                    if R1.Right - R1.Left < R.Right - R.Left then
                    begin
                        if Alignment = taCenter then
                            OffsetRect(R1, ((R.Right - R.Left) - (R1.Right - R1.Left)) div 2, 0)
                        else
                            if Alignment = taRightJustify then
                                OffsetRect(R1, ((R.Right - R.Left) - (R1.Right - R1.Left)), 0);
                    end;
                end;
            end;
        if InFontScale and (
            (R1.Bottom - R1.Top > R.Bottom - R.Top) or
            (R1.Right - R1.Left > R.Right - R.Left)) then
        begin
            if ACanvas.Font.Size > InFontMinSize then
            begin
                ACanvas.Font.Size := ACanvas.Font.Size - 1;
                goto TryAgain;
            end;
        end;
        if  InFontScale then
        begin
            FRectLastPaint := Classes.Rect(0, 0, Width, Height);
            FTextLastPaint := Caption;
            FFontSizeLastPaint := ACanvas.Font.Size;
        end;
        if Transparent then
            ACanvas.Brush.Style := bsClear;
        if InScale and (
            (R1.Bottom - R1.Top > R.Bottom - R.Top) or
            (R1.Right - R1.Left > R.Right - R.Left)) then
        begin
            if TextRotateViaBitmap then
            begin
                B := TBitmap.Create();
                TRY
                    B.Width := R1.Right - R1.Left;
                    B.Height := R1.Bottom - R1.Top;
                    B.Canvas.Brush.Color := Color;
                    B.Canvas.Font.Assign(Font);
                    R0 := Classes.Rect(0, 0, B.Width, B.Height);
                    B.Canvas.FillRect(R0);
                    Windows.DrawTextEx(B.Canvas.Handle, PChar(Caption), Length(Caption),
                        R0, F, @DTP);
                    SetStretchBltMode(ACanvas.Handle, halftone);
                    StretchBlt(ACanvas.Handle, R.Left, R.Top, R.Right - R.Left, R.Bottom - R.Top,
                        B.Canvas.Handle, 0, 0, B.Width, B.Height, SRCCOPY);
                FINALLY
                    B.Free;
                END;
            end
            else
            begin
                DC0 := GetDC(0);
                W := R1.Right - R1.Left + 4;
                H := R1.Bottom - R1.Top + 4;
                R := Classes.Rect(0, 0,
                    MulDiv(W, GetDeviceCaps(DC0, HORZSIZE) * 100, GetDeviceCaps(DC0, HORZRES)),
                    MulDiv(H, GetDeviceCaps(DC0, VERTSIZE) * 100, GetDeviceCaps(DC0, VERTRES))
                    );
                MF := CreateEnhMetaFile(DC0, nil, @R, '');
                ReleaseDC(0, DC0);
                C := TCanvas.Create;
                C.Handle := MF;
                C.Refresh;
                TRY
                    C.Font.Assign(Font);
                    C.Brush.Color := Canvas.Brush.Color;

                    R0 := Classes.Rect(0, 0, W, H);
                    if not Transparent then
                        C.FillRect(R0)
                    else
                        C.Brush.Style := bsClear;
                    if not Enabled then //|||||||||||||||||||
                        C.Font.Color := clBtnHighlight;
                    Windows.DrawTextEx(C.Handle, PChar(Caption), Length(Caption),
                        R0, F, @DTP);

                    MFDC := CloseEnhMetaFile(MF);

                    R := Rect;
                    inc(R.Left, MarginLeft);
                    inc(R.Top, MarginTop);
                    dec(R.Right, MarginRight);
                    dec(R.Bottom, MarginBottom);
                    W := R0.Right - R0.Left;
                    H := R0.Bottom - R0.Top;
                    if InProportionalScale then
                    begin
                        KW := (R.Right - R.Left) / W;
                        KH := (R.Bottom - R.Top) / H;
                        if KW < KH then
                        begin
                            H := Trunc((R.Bottom - R.Top) / KH);
                            CASE Layout OF
                                tlCenter: OffsetRect(R, 0, (R.Bottom - R.Top - H) div 2);
                                tlBottom: OffsetRect(R, 0, (R.Bottom - R.Top - H));
                            END;
                            R.Bottom := R.Top + H;
                        end
                        else
                        begin
                            if KW > 1 then KW := KH;
                            W := Trunc((R.Right - R.Left) * KW);
                            CASE Alignment OF
                                taCenter: OffsetRect(R, (R.Right - R.Left - W) div 2, 0);
                                taRightJustify: OffsetRect(R, (R.Right - R.Left - W), 0);
                            END;
                            R.Right := R.Left + W;
                        end;
                        R1 := R;
                    end
                    else
                        R1 := Rect;
                    SetStretchBltMode(ACanvas.Handle, halftone);
                    PlayEnhMetaFile(ACanvas.Handle, MFDC, R1);

                    if FrameRadius > 0 then
                    begin
                        OldPen := TPen.Create;
                        TRY
                            OldPen.Assign(ACanvas.Pen);
                            if (FrameWidth > 0) and
                                (ColorToRGB(ActualFrameColor) <> ColorToRGB(Brush.Color)) then
                            begin
                                if FrameColor = clNone then
                                    ACanvas.Pen.Color := Font.Color
                                else
                                    ACanvas.Pen.Color := FrameColor;
                                ACanvas.Brush.Style := bsClear;
                                ACanvas.Pen.Style := psSolid;
                                ACanvas.Pen.Width := 0;
                                if FrameWidth > 0 then
                                    ACanvas.Pen.Width := FrameWidth;
                                SetStretchBltMode(ACanvas.Handle, BLACKONWHITE);
                                ACanvas.RoundRect(Rect.Left, Rect.Top,
                                    Rect.Right {- Rect.Left} - FrameWidth + 1,
                                    Rect.Bottom {- Rect.Top} - FrameWidth + 1,
                                    ActualFrameRadius * 2 - FrameWidth, ActualFrameRadius * 2 - FrameWidth);
                            end;
                            ACanvas.Brush.Style := bsSolid;
                        FINALLY
                            ACanvas.Pen.Assign(OldPen);
                            OldPen.Free;
                        END;
                    end;

                FINALLY
                    DeleteEnhMetaFile(MF);
                    C.Free;
                END;
            end;
        end
        else
        begin
            if WrapLetters then
                PaintWrapByLetters(ACanvas, R, FALSE)
            else
                if WrapSyllabel then
                    PaintWrapped(ACanvas, R, FALSE)
                else
                begin
                    if not Enabled then //|||||||||||||||||||
                        ACanvas.Font.Color := clBtnHighlight;
                    Windows.DrawTextEx(ACanvas.Handle, PChar(Caption), Length(Caption),
                        R, F, @DTP);
                    if not Enabled then
                    begin
                        ACanvas.Font.Color := clBtnShadow;
                        OffsetRect(R, -1, -1);
                        Windows.DrawTextEx(ACanvas.Handle, PChar(Caption), Length(Caption),
                            R, F, @DTP);
                    end;
                end;
        end;
        {begin
            if  not Enabled then     //|||||||||||||||||||
                ACanvas.Font.Color := clBtnHighlight;
            Windows.DrawTextEx( ACanvas.Handle, PChar( Caption ), Length( Caption ),
                              R, F, @ DTP );
            if  not Enabled then
            begin
                ACanvas.Font.Color := clBtnShadow;
                OffsetRect( R1, -1, -1 );
                Windows.DrawTextEx( ACanvas.Handle, PChar( Caption ), Length( Caption ),
                                R1, F, @ DTP );
            end;
        end;}
    end; // if  Trim( Caption ) <> '' //
    if FrameRadius = 0 then
    begin
        ACanvas.Brush.Color := ActualFrameColor;
        if frLeft in Frames then
            ACanvas.FillRect(Classes.Rect(Rect.Left, Rect.Top, Rect.Left + FrameWidth,
                Rect.Bottom));
        if frTop in Frames then
            ACanvas.FillRect(Classes.Rect(Rect.Left, Rect.Top, Rect.Right,
                Rect.Top + FrameWidth));
        if frRight in Frames then
            ACanvas.FillRect(Classes.Rect(Rect.Right - FrameWidth, Rect.Top,
                Rect.Right, Rect.Bottom));
        if frBottom in Frames then
            ACanvas.FillRect(Classes.Rect(Rect.Left, Rect.Bottom - FrameWidth,
                Rect.Right, Rect.Bottom));
    end;
    ACanvas.Brush.Style := bsSolid;
end;

function TNormalLabel.PaintWrapByLetters(ACanvas: TCanvas;
    var Rect: TRect; Calc_Rect: Boolean): Integer;
var X, X0, Y, Y0, I, TabWidth, SpaceWidth, H, W, j: Integer;
    Word: String;
    LineStarts: TList;
    Lines: TStringList;
    Stage: Integer;
    XOffset, YOffset: Integer;
    FinalRect: TRect;

    procedure NewLine;
    begin
        if Stage = 1 then
        begin
            LineStarts.Add(Pointer(X));
            X := 0;
        end
        else
        begin
            if X > FinalRect.Right then
                FinalRect.Right := X;
            X := ActualMarginLeft;
            if LineStarts.Count > 0 then
            begin
                X := Integer(LineStarts[0]);
                X0 := X;
                LineStarts.Delete(0);
            end;
        end;
    end;
    procedure AddWord(X, Y: Integer; const s: String);
    begin
        Lines.AddObject(s, Pointer(X + (Y shl 16)));
    end;
begin
    Result := 0; //   .
    FinalRect := Rect;
    if Trim(Caption) = '' then Exit;
    if (Length(Caption) > 10) and (Caption[1] = '') and (Caption[2] = '') then
        asm
      nop
        end;
    X := ActualMarginLeft;

    TabWidth := ACanvas.TextWidth('        ');
    SpaceWidth := ACanvas.TextWidth(' ');
    H := ACanvas.TextHeight('Ap^_/\');
    //Canvas.Brush.Style := bsClear;
    LineStarts := TList.Create;
    Lines := TStringList.Create;
    TRY
        Y := 0;
        for Stage := 1 to 2 do
        begin //   1     
            I := 1;
            if Stage = 2 then
            begin
                Y0 := -1;
                XOffset := 0;
                YOffset := 0;
                CASE Layout OF
                    tlTop: ;
                    tlCenter: YOffset := (Rect.Bottom - Rect.Top - Y) div 2;
                    tlBottom: YOffset := (Rect.Bottom - Rect.Top - ActualMarginBottom - Y);
                END;
                if Y < 0 then Y := 0;
                for j := 0 to Lines.Count - 1 do
                begin
                    X := LoWord(Integer(Lines.Objects[j]));
                    Y := HiWord(Integer(Lines.Objects[j])) + YOffset
                        + Rect.Top;
                    if Y <> Y0 then
                    begin
                        Y0 := Y;
                        W := 0;
                        if LineStarts.Count > 0 then
                        begin
                            W := Integer(LineStarts[0]);
                            LineStarts.Delete(0);
                        end;
                        CASE Alignment OF
                            taLeftJustify: XOffset := Rect.Left + ActualMarginLeft;
                            taCenter: XOffset := Rect.Left + (Rect.Right - Rect.Left - W) div 2;
                            taRightJustify: XOffset := Rect.Left + (Rect.Right - Rect.Left - W - ActualMarginRight);
                        END;
                    end;
                    X := X + XOffset;
                    ACanvas.TextOut(X, Y, Lines[j]);
                end;
                X := Rect.Left + ActualMarginLeft;
                if LineStarts.Count > 0 then
                begin
                    X := Integer(LineStarts[0]);
                    LineStarts.Delete(0);
                end;
            end
            else
            begin
                Y := 0;
                X := 0;
                X0 := X;
                while I <= Length(Caption) do
                begin
                    CASE Caption[I] OF
                        #9: //  .      .
                            //      X   
                            //  ,       .
                            begin
                                X := ((X - X0) div TabWidth + 1) * TabWidth;
                                if X > Rect.Right - Rect.Left - ActualMarginRight then
                                begin
                                    NewLine;
                                    Y := Y + H;
                                end;
                                Inc(I);
                            end;
                        #13: begin
                                if (I < Length(Caption)) and (Caption[I + 1] = #10) then
                                    Inc(I);
                                NewLine;
                                Y := Y + H;
                            end;
                        #1..#8, #10..#12, #14..' ': //  .   ,   
                            //      X   .
                            //     -,   
                            //  X,      ,   
                            //   .
                            begin
                                while TRUE do
                                begin
                                    X := X + SpaceWidth;
                                    Inc(I); if I > Length(Caption) then break;
                                    if Caption[I] in [#9, #33..#255] then break;
                                end;
                                if X > Rect.Right - Rect.Left - ActualMarginRight then
                                begin
                                    NewLine;
                                    Y := Y + H;
                                end;
                            end;
                    else //  .       ,
                        //    .
                        begin
                            Word := Copy(Caption, I, MaxInt);
                            I := I + Length(Word);
                            //  , I   ,    
                            //  .   :
                            while TRUE do
                            begin
                                W := ACanvas.TextWidth(Word);
                                if (X + W <= Rect.Right - Rect.Left - ActualMarginRight)
                                    or (Length(Word) <= 1) then
                                begin
                                    Word := Trim(Word);
                                    if Stage = 2 then
                                    begin
                                        if not Calc_Rect then
                                            ACanvas.TextOut(X, Y, Word)
                                    end
                                    else
                                        AddWord(X, Y, Word);
                                    X := X + ACanvas.TextWidth(Word);
                                    break;
                                end;
                                //     .    .
                                if X > X0 then
                                begin
                                    NewLine;
                                    Y := Y + H;
                                end;
                                while (Length(Word) > 1) and
                                    (X + ACanvas.TextWidth(Word) >
                                    Rect.Right - Rect.Left - ActualMarginRight) do
                                begin
                                    Word := Copy(Word, 1, Length(Word) - 1);
                                    Dec(I);
                                end;
                                W := ACanvas.TextWidth(Word);
                                if (X + W <= Rect.Right - Rect.Left - ActualMarginRight)
                                    or (Length(Word) = 1) then
                                begin
                                    Word := Trim(Word);
                                    if Stage = 2 then
                                    begin
                                        if not Calc_Rect then
                                            ACanvas.TextOut(X, Y, Word)
                                    end
                                    else
                                        AddWord(X, Y, Word);
                                    X := X + W;
                                    Word := '';
                                    break;
                                end;
                                if Word = '' then break;
                            end;
                            if X > Rect.Right - Rect.Left - ActualMarginRight then
                            begin
                                NewLine;
                                Y := Y + H;
                            end;
                        end;
                    END;
                end;
                NewLine;
                Y := Y + H;
                FinalRect.Bottom := Y;
                Result := Y + ActualMarginTop + ActualMarginBottom;
            end;
        end;
    FINALLY
        LineStarts.Free;
        Lines.Free;
    END;
    Rect := FinalRect;
end;

function TNormalLabel.PaintWrapped(ACanvas: TCanvas; var Rect: TRect;
    Calc_Rect: Boolean): Integer;
var X, X0, Y, I, TabWidth, SpaceWidth, HyphenWidth, W, H,
    XOffset, YOffset, Y0: Integer;
    Word: String;
    Firstdiv, WasHypHen: Boolean;
    LineStarts: TList;
    Lines: TStringList;
    FinalRect: TRect;
    cap: String;

    procedure DoPaintWrapped;
    label Divide_Syllabel;
    var Stage, j, L: Integer;
        procedure NewLine;
        begin
            if Stage = 1 then
            begin
                LineStarts.Add(Pointer(X));
                X := 0;
            end
            else
            begin
                if X > FinalRect.Right then
                    FinalRect.Right := X;
                X := ActualMarginLeft;
                if LineStarts.Count > 0 then
                begin
                    X := Integer(LineStarts[0]);
                    X0 := X;
                    LineStarts.Delete(0);
                end;
            end;
        end;
        procedure AddWord(X, Y: Integer; const s: String);
        begin
            Lines.AddObject(s, Pointer(X + (Y shl 16)));
        end;
    begin
        FinalRect := Rect;
        if Trim(Cap) = '' then Exit;
        X := ActualMarginLeft;
        LineStarts := TList.Create;
        Lines := TStringList.Create;
        TRY
            Y := 0;
            for Stage := 1 to 2 do
            begin
                I := 1;
                if Stage = 2 then
                begin
                    Y0 := -1;
                    XOffset := 0;
                    YOffset := 0;
                    CASE Layout OF
                        tlTop: ;
                        tlCenter: YOffset := (Rect.Bottom - Rect.Top - Y) div 2;
                        tlBottom: YOffset := (Rect.Bottom - Rect.Top - ActualMarginBottom - Y);
                    END;
                    if YOffset < 0 then YOffset := 0;
                    for j := 0 to Lines.Count - 1 do
                    begin
                        X := LoWord(Integer(Lines.Objects[j]));
                        Y := HiWord(Integer(Lines.Objects[j])) + YOffset
                            + Rect.Top;
                        if Y <> Y0 then
                        begin
                            Y0 := Y;
                            W := 0;
                            if LineStarts.Count > 0 then
                            begin
                                W := Integer(LineStarts[0]);
                                LineStarts.Delete(0);
                            end;
                            CASE Alignment OF
                                taLeftJustify: XOffset := Rect.Left + ActualMarginLeft;
                                taCenter: XOffset := Rect.Left + (Rect.Right - Rect.Left - W) div 2;
                                taRightJustify: XOffset := Rect.Left + (Rect.Right - Rect.Left - W - ActualMarginRight);
                            END;
                        end;
                        X := X + XOffset;
                        if not Calc_Rect then
                            ACanvas.TextOut(X, Y, Lines[j]);
                    end;
                    X := Rect.Left + ActualMarginLeft;
                    if LineStarts.Count > 0 then
                    begin
                        X := Integer(LineStarts[0]);
                        LineStarts.Delete(0);
                    end;
                end
                else
                begin
                    X := 0;
                    X0 := X;
                    TabWidth := ACanvas.TextWidth('        ');
                    SpaceWidth := ACanvas.TextWidth(' ');
                    H := ACanvas.TextHeight('Ap^_/\');
                    Y := 0;
                    //Canvas.Brush.Style := bsClear;
                    while I <= Length(Cap) do
                    begin
                        CASE Cap[I] OF
                            #9: //  .      .
                                //      X   
                                //  ,       .
                                begin
                                    X := (X div TabWidth + 1) * TabWidth;
                                    if X > Rect.Right - Rect.Left - ActualMarginRight then
                                    begin
                                        NewLine;
                                        Y := Y + H;
                                    end;
                                    Inc(I);
                                end;
                            #13: begin
                                    if (I < Length(Cap)) and (Cap[I + 1] = #10) then
                                        Inc(I);
                                    NewLine;
                                    Y := Y + H;
                                end;
                            #1..#8, #10..#12, #14..' ': //  .   ,   
                                //      X   .
                                //     -,   
                                //  X,      ,   
                                //   .
                                begin
                                    while TRUE do
                                    begin
                                        X := X + SpaceWidth;
                                        Inc(I); if I > Length(Cap) then break;
                                        if Cap[I] in [#9, #33..#255] then break;
                                    end;
                                    if X > Rect.Right - Rect.Left - ActualMarginRight then
                                    begin
                                        NewLine;
                                        Y := Y + H;
                                    end;
                                end;
                        else //  .       , 
                            //     .   ,    
                            //  ,       .  
                            //     ,     , 
                            //    .
                            begin
                                Word := Cap[I];
                                Inc(I);
                                while (I <= Length(Cap)) and
                                    (
                                    (Cap[I] in ['0'..'9', 'A'..'Z', 'a'..'z', ''..'', ''..'', '', ''])
                                    or
                                    (Word[1] in ['0'..'9', '-', '+']) and
                                    (Cap[I] in [',', {'.',} '^', '=', {'-',} ':'])
                                    ) do
                                begin
                                    Word := Word + Cap[I];
                                    Inc(I);
                                end;
                                //  , I   ,    
                                //  .   :
                                Firstdiv := TRUE;
                                while TRUE do
                                begin
                                    if X + ACanvas.TextWidth(Word) <= Rect.Right - Rect.Left - ActualMarginRight then
                                    begin
                                        Word := Trim(Word);
                                        if Stage = 2 then
                                        begin
                                            if not Calc_Rect then
                                                ACanvas.TextOut(X, Y, Word);
                                        end
                                        else
                                            AddWord(X, Y, Word);
                                        X := X + ACanvas.TextWidth(Word);
                                        break;
                                    end;
                                    Divide_Syllabel:
                                    //     .    .
                                    L := LastSyllableWidth(Word, Firstdiv);
                                    if (L <= 0) or (L >= Length(Word)) then
                                    begin
                                        if X > X0 then
                                        begin
                                            NewLine;
                                            Y := Y + H;
                                        end;
                                    end;
                                    HyphenWidth := 0;
                                    while X + ACanvas.TextWidth(Word) + HyphenWidth >
                                        Rect.Right - Rect.Left - ActualMarginRight do
                                    begin
                                        L := LastSyllableWidth(Word, Firstdiv);
                                        if (L > 0) and (L < Length(Word) - 1) then
                                        begin
                                            Word := Copy(Word, 1, Length(Word) - L);
                                            Firstdiv := FALSE;
                                            Dec(I, L);
                                            HyphenWidth := ACanvas.TextWidth('-');
                                        end
                                        else
                                        begin
                                            if X > X0 then
                                            begin
                                                NewLine;
                                                Y := Y + H;
                                                goto Divide_Syllabel;
                                            end;
                                            WasHypHen := HyphenWidth > 0;
                                            HyphenWidth := -1;
                                            for L := Length(Word) downto 1 do
                                            begin
                                                if X + ACanvas.TextWidth(Copy(Word, 1, Length(Word) - L))
                                                    <= Rect.Right - Rect.Left - ActualMarginRight then
                                                begin
                                                    if L = Length(Word) then
                                                        if WasHypHen then
                                                            HyphenWidth := -ACanvas.TextWidth('-');
                                                    Word := Copy(Word, 1, L);
                                                    Dec(I, Length(Word) - L);
                                                    Firstdiv := FALSE;
                                                    break;
                                                end;
                                            end;
                                            if Word <> '' then break;
                                        end;
                                    end;
                                    if (X + ACanvas.TextWidth(Word) + HyphenWidth <=
                                        Rect.Right - Rect.Left - ActualMarginRight) or
                                        (HyphenWidth < 0) then
                                    begin
                                        Word := Trim(Word);
                                        if Stage = 2 then
                                        begin
                                            if not Calc_Rect then
                                                ACanvas.TextOut(X, Y, Word)
                                        end
                                        else
                                            AddWord(X, Y, Word);
                                        Inc(X, ACanvas.TextWidth(Word));
                                        if (HyphenWidth > 0) or (HyphenWidth < -1) then
                                        begin
                                            if Stage = 2 then
                                            begin
                                                if not Calc_Rect then
                                                    ACanvas.TextOut(X, Y, '-');
                                            end
                                            else
                                                AddWord(X, Y, '-');
                                            X := X + Abs(HyphenWidth);
                                            NewLine;
                                            Y := Y + H;
                                        end;
                                        Word := '';
                                        break;
                                    end;
                                    if Word = '' then break;
                                end;
                                if X > Rect.Right - Rect.Left - ActualMarginRight then
                                begin
                                    NewLine;
                                    Y := Y + H;
                                end;
                            end;
                        END;
                    end;
                    NewLine;
                    Y := Y + H;
                    Result := Y + ActualMarginTop + ActualMarginBottom;
                    FinalRect.Bottom := Y;
                end;
            end;
        FINALLY
            LineStarts.Free;
            Lines.Free;
        END;
        Rect := FinalRect;
    end;
begin
    Result := 0; //   .
    Cap := '';
    TRY
        Cap := Caption;
        if (Length(Cap) > 10) and (Cap[1] = '') and (Cap[2] = '') then
            asm
          nop
            end;
        DoPaintWrapped;
    EXCEPT
        asm
          nop
        end;
        {TRY
            Cap := Caption;
        EXCEPT
        END;
        DoPaintWrapped;}
    END;
end;

procedure TNormalLabel.Resize;
var W, H: Integer;
begin
    W := Width;
    H := Height;
    if CanAutoSize(W, H) then
    begin
        Self.Width := W;
        Self.Height := H;
    end
    else
        inherited;
end;

procedure TNormalLabel.SetAutosize(const Value: Boolean);
begin
    FAutosize := Value;
    DoAutosize;
end;

procedure TNormalLabel.SetCaption(const Value: TCaption);
begin
    inherited Caption := Value;
    DoAutoSize;
    Invalidate;
end;

procedure TNormalLabel.SetDiagonalSplit(const Value: Char);
begin
    FDiagonalSplit := Value;
    Invalidate;
end;

procedure TNormalLabel.SetDoubleBuffered(const Value: Boolean);
begin
    FDoubleBuffered := Value;
end;

procedure TNormalLabel.SetFrameColor(const Value: TColor);
begin
    FFrameColor := Value;
    Invalidate;
end;

procedure TNormalLabel.SetFrameInflate(const Value: Integer);
begin
    FFrameInflate := Value;
    Invalidate;
end;

procedure TNormalLabel.SetFrameRadius(const Value: Integer);
begin
    FFrameRadius := Value;
    CurRgn1 := 0;
    CurRgn2 := 0;
    CurRgn3 := 0;
    if (Value = 0) and (Parent <> nil) then
        SetWindowRgn(Handle, 0, TRUE);
    DoAutosize;
    Update;
    Invalidate;
end;

procedure TNormalLabel.SetFrames(const Value: TFrames);
begin
    FFrames := Value;
    Invalidate;
end;

procedure TNormalLabel.SetFrameStyle(const Value: TFrameStyle);
begin
    FFrameStyle := Value;
    Invalidate;
end;

procedure TNormalLabel.SetFrameWidth(const Value: Integer);
begin
    FFrameWidth := Value;
    DoAutosize;
    Invalidate;
end;

procedure TNormalLabel.SetInFontMinSize(const Value: Integer);
begin
    if FInFontMinSize = Value then Exit;
    FInFontMinSize := Value;
    if InFontScale then
        Invalidate;
end;

procedure TNormalLabel.SetInFontScale(const Value: Boolean);
begin
    if FInFontScale = Value then Exit;
    FInFontScale := Value;
    Invalidate;
end;

procedure TNormalLabel.SetInProportionalScale(const Value: Boolean);
begin
    FInProportionalScale := Value;
    Invalidate;
end;

procedure TNormalLabel.SetInScale(const Value: Boolean);
begin
    if FInScale = Value then Exit;
    FInScale := Value;
    Invalidate;
end;

procedure TNormalLabel.SetLayout(const Value: TTextLayout);
begin
    if FLayout = Value then Exit;
    FLayout := Value;
    DoAutosize;
    Invalidate;
end;

procedure TNormalLabel.SetMarginBottom(const Value: Integer);
begin
    FMarginBottom := Value;
    DoAutosize;
    Invalidate;
end;

procedure TNormalLabel.SetMarginLeft(const Value: Integer);
begin
    FMarginLeft := Value;
    DoAutosize;
    Invalidate;
end;

procedure TNormalLabel.SetMarginRight(const Value: Integer);
begin
    FMarginRight := Value;
    DoAutosize;
    Invalidate;
end;

procedure TNormalLabel.SetMarginTop(const Value: Integer);
begin
    FMarginTop := Value;
    DoAutosize;
    Invalidate;
end;

procedure TNormalLabel.SetMouseTransparent(const Value: Boolean);
begin
    FMouseTransparent := Value;
end;

procedure TNormalLabel.SetOnMouseEnter(const Value: TNotifyEvent);
begin
    FOnMouseEnter := Value;
end;

procedure TNormalLabel.SetOnMouseLeave(const Value: TNotifyEvent);
begin
    FOnMouseLeave := Value;
end;

procedure TNormalLabel.SetParentFont(const Value: Boolean);
begin
    inherited ParentFont := Value;
    DoAutosize;
    Invalidate;
end;

procedure TNormalLabel.SetTextOrientation(const Value: TTextOrientation);
begin
    if FTextOrientation = Value then Exit;
    FTextOrientation := Value;
    DoAutosize;
    Invalidate;
end;

procedure TNormalLabel.SetTextRotateViaBitmap(const Value: Boolean);
begin
    if FTextRotateViaBitmap = Value then Exit;
    FTextRotateViaBitmap := Value;
    Invalidate;
end;

procedure TNormalLabel.SetTransparent(const Value: Boolean);
begin
    FTransparent := Value;
    Invalidate;
end;

procedure TNormalLabel.SetWordWrap(const Value: Boolean);
begin
    if FWordWrap = Value then Exit;
    FWordWrap := Value;
    DoAutosize;
    Invalidate;
end;

procedure TNormalLabel.SetWrapLetters(const Value: Boolean);
begin
    if FWrapLetters = Value then Exit;
    FWrapLetters := Value;
    DoAutosize;
    Invalidate;
end;

procedure TNormalLabel.SetWrapSyllabel(const Value: Boolean);
begin
    FWrapSyllabel := Value;
    FPaintWrapped := PaintWrapped;
    DoAutosize;
    Invalidate;
end;

initialization
    TmpBmp := TBitmap.Create;
    TmpBmp.Width := 1;
    TmpBmp.Height := 1;

finalization
    TmpBmp.Free;

end.

