//**  ,     VCL Delphi.
//**  Vladimir Kladov, 2000-2010
//**
//**      .
//**     /     
//**  2          
//**   GNU,  Free Software Foundation.
//**
//**       ,    
//** ,       ,   
//**         
//**   .      
//**     GNU.
//**
//**          
//**   GNU.     ,     Free
//** Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
//** MA 02111-1307, USA.
//******************************************************************************

{ TfmNormalPreview -      .
  (C) by Vladimir Kladov, 2002

  v1.0.1 (17.10.2002) -  
                      (   )

  19.02.2010 -   ,   WrapByLetters,
             WrapSyllabel,   
}

unit NormalPreview;

interface

{$I VCL_VERSIONS.INC}

uses
    Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
    ComCtrls, ToolWin, ExtCtrls, StdCtrls, Printers, Menus,
    ImgList, Buttons, math;

type
    TfmNormalPreview = class(TForm)
        Panel1: TPanel;
        ToolBar1: TToolBar;
        TB1: TToolButton;
        TB2: TToolButton;
        TB3: TToolButton;
        TB4: TToolButton;
        TB7: TToolButton;
        TB8: TToolButton;
        Label1: TLabel;
        ScrollBox1: TScrollBox;
        TB5: TToolButton;
        PopupMenu1: TPopupMenu;
        N1: TMenuItem;
        N2: TMenuItem;
        N3: TMenuItem;
        PaintPanel: TPanel;
        PaintBox1: TPaintBox;
        BelowPanel: TPanel;
        PopupMenu2: TPopupMenu;
        N4: TMenuItem;
        TB6: TToolButton;
        ImageList1: TImageList;
        ToolButton9: TToolButton;
        ToolButton10: TToolButton;
        ToolButton11: TToolButton;
        PrinterSetupDialog1: TPrinterSetupDialog;
        ImageList2: TImageList;
        Panel2: TPanel;
        ToolBar2: TToolBar;
        TB9: TToolButton;
        ToolButton13: TToolButton;
        N5: TMenuItem;
        procedure FormShow(Sender: TObject);
        procedure PaintBox1Paint(Sender: TObject);
        procedure TB8Click(Sender: TObject);
        procedure Fit_Page(Sender: TObject);
        procedure Fit_Width(Sender: TObject);
        procedure As_Is(Sender: TObject);
        procedure FormDestroy(Sender: TObject);
        procedure ScrollBox1Resize(Sender: TObject);
        procedure PopupMenu1Popup(Sender: TObject);
        procedure Print_Page(Sender: TObject);
        procedure Print_Setup(Sender: TObject);
        procedure FormCreate(Sender: TObject);
        procedure N5Click(Sender: TObject);
        procedure First_Page(Sender: TObject);
        procedure Prev_Page(Sender: TObject);
        procedure Last_Page(Sender: TObject);
        procedure Next_Page(Sender: TObject);
        procedure Print_All_Close(Sender: TObject);
        procedure Horizontal_Expand(Sender: TObject);
        procedure Vertical_Expand(Sender: TObject);
        procedure FormKeyDown(Sender: TObject; var Key: Word;
            Shift: TShiftState);
    private
        FCurrentPage: Integer;
        FReport: TObject;
        FFitMode: Integer;
        FPreviewBuf: TBitmap;
        FPreviewPage: Integer;
        procedure SetCurrentPage(const Value: Integer);
        function GetPrinter: TPrinter;
        procedure SetFitMode(const Value: Integer);
        { Private declarations }
    protected
        procedure EnableButtons;
        function ScaleToScreenX(X: Integer): Integer;
        procedure AdjustFitMode;
        procedure PreparePages;
    public
        { Public declarations }
        PrintDialog1: TPrintDialog;
        property CurrentPage: Integer read FCurrentPage write SetCurrentPage;
        property Report: TObject read FReport write FReport;
        property Printer: TPrinter read GetPrinter;
        property FitMode: Integer read FFitMode write SetFitMode;
    protected
        procedure WM_Wheel(var M: TMessage); message WM_MOUSEWHEEL;
    end;

var
    fmNormalPreview: TfmNormalPreview;

implementation

uses
    NormalReport, NormalResources, NormalRange;

{$O+}

procedure DrawStretched(DstBmp, SrcBmp: TBitmap);
var Dst, Src: PInteger;
    y, x, Ysrc, Ysrc2, Xsrc, Xsrc2: Integer;
    y1, x1: Integer;
    R, G, B, N, Ny, Nx: Integer;
begin
    Ysrc := 0;
    for y := 0 to DstBmp.Height - 1 do
    begin
        Ysrc2 := Ysrc + SrcBmp.Height * 256 div DstBmp.Height;
        Dst := DstBmp.ScanLine[y];
        Xsrc := 0;
        for x := 0 to DstBmp.Width - 1 do
        begin
            Xsrc2 := Xsrc + SrcBmp.Width * 256 div DstBmp.Width;
            R := 0;
            G := 0;
            B := 0;
            N := 0;
            Ny := 256 - (Ysrc and 255);
            for y1 := Ysrc div 256 to (Ysrc2 - 1) div 256 do
            begin
                Src := Pointer(Integer(SrcBmp.ScanLine[y1]) + 4 * (Xsrc div 256));
                Nx := 256 - (Xsrc and 255);
                for x1 := Xsrc div 256 to (Xsrc2 - 1) div 256 do
                begin
                    R := R + (Src^ and 255) * Ny * Nx;
                    G := G + ((Src^ shr 8) and 255) * Ny * Nx;
                    B := B + ((Src^ shr 16) and 255) * Ny * Nx;
                    N := N + Ny * Nx;
                    if x1 = (Xsrc2 - 1) div 256 - 1 then
                        Nx := Xsrc2 and 255
                    else
                        Nx := 256;
                    Inc(Src);
                end;
                if y1 = (Ysrc2 - 1) div 256 - 1 then
                    Ny := Ysrc2 and 255
                else
                    Ny := 256;
            end;
            if N > 0 then
            begin
                Dst^ := (R div N) or ((G div N) shl 8) or ((B div N) shl 16);
                inc(Dst);
            end;
            Xsrc := Xsrc2;
        end;
        Ysrc := Ysrc2;
    end;
end;

type
    TWindowsVersion = (wv31, wv95, wv98, wvNT, wvY2K, wvXP, wvLongHorn);
    {* Windows versions constants. }
    TWindowsVersions = Set of TWindowsVersion;
    {* Set of Windows version (e.g. to define a range of versions supported by the
       application). }
var SaveWinVer: Byte = $FF;

    //[function WinVer]

function WinVer: TWindowsVersion;
{* Returns Windows version. }
var OVI: TOsVersionInfo;
begin
    if SaveWinVer <> $FF then Result := TWindowsVersion(SaveWinVer)
    else
    begin
        OVI.dwOSVersionInfoSize := Sizeof(OVI);
        GetVersionEx(OVI);
        with OVI do
            if dwPlatformId = VER_PLATFORM_WIN32_NT then
            begin
                Result := wvNT;
                if dwMajorVersion >= 6 then
                    Result := wvLongHorn
                else begin
                    if dwMajorVersion >= 5 then
                        if dwMinorVersion >= 1 then
                            Result := wvXP
                        else
                            Result := wvY2K;
                end;
            end
            else
                if dwPlatformId = VER_PLATFORM_WIN32_WINDOWS then
                begin
                    Result := wv95;
                    if (dwMajorVersion > 4) or (dwMajorVersion = 4)
                        and (dwMinorVersion >= 10) then
                        Result := wv98;
                end
                else
                    Result := wv31; // Windows 3.1 (WIN32s)
        SaveWinVer := Ord(Result);
    end;
end;


{$R *.DFM}

{ TfmNormalPreview }

procedure TfmNormalPreview.SetCurrentPage(const Value: Integer);
begin
    FCurrentPage := Value;
    AdjustFitMode;
    EnableButtons;
    PaintBox1.Invalidate;
end;

procedure TfmNormalPreview.First_Page(Sender: TObject);
begin
    CurrentPage := 0;
    EnableButtons;
end;

procedure TfmNormalPreview.Prev_Page(Sender: TObject);
begin
    if CurrentPage > 0 then
        CurrentPage := CurrentPage - 1;
    EnableButtons;
end;

procedure TfmNormalPreview.Next_Page(Sender: TObject);
begin
    if CurrentPage < TNormalReport(Report).PagesCount - 1 then
        CurrentPage := CurrentPage + 1;
    EnableButtons;
end;

procedure TfmNormalPreview.Last_Page(Sender: TObject);
begin
    CurrentPage := TNormalReport(Report).PagesCount - 1;
    EnableButtons;
end;

procedure TfmNormalPreview.FormShow(Sender: TObject);
begin
    EnableButtons;
    if TNormalReport(Report).PagesCount > 0 then
        SetCurrentPage(0);
end;

procedure TfmNormalPreview.EnableButtons;
begin
    if TNormalReport(Report).PagesCount = 0 then
    begin
        Label1.Caption := rsNoData;
        TB1.Enabled := FALSE;
        TB2.Enabled := FALSE;
        TB3.Enabled := FALSE;
        TB4.Enabled := FALSE;
        PaintBox1.Width := 0;
        TB9.Enabled := FALSE;
        Exit;
    end;
    Label1.Caption := rsPage + ' ' + IntToStr(CurrentPage + 1) +
        rsFrom + IntToStr(TNormalReport(Report).PagesCount) + '   ';
    TB1.Enabled := CurrentPage > 0;
    TB2.Enabled := CurrentPage > 0;
    TB3.Enabled := CurrentPage < TNormalReport(Report).PagesCount - 1;
    TB4.Enabled := CurrentPage < TNormalReport(Report).PagesCount - 1;
    TB9.Enabled := TRUE;
end;

procedure TfmNormalPreview.PaintBox1Paint(Sender: TObject);
var MF: HEnhMetafile;
    Tmp: TBitmap;
begin
    PaintBox1.Canvas.Brush.Color := clWhite;
    PaintBox1.Canvas.FillRect(PaintBox1.ClientRect);
    if CurrentPage >= TNormalReport(Report).PagesCount then Exit;
    MF := TNormalReport(Report).Images[CurrentPage];
    AdjustFitMode;
    if (FPreviewPage <> CurrentPage) or (FPreviewBuf = nil) or
        (FPreviewBuf.Width <> PaintBox1.Width) or (FPreviewBuf.Height <> PaintBox1.Height) then
    begin
        FPreviewBuf.Free;
        FPreviewBuf := TBitmap.Create;
        FPreviewBuf.PixelFormat := pf32bit;
        FPreviewBuf.Width := PaintBox1.Width;
        FPreviewBuf.Height := PaintBox1.Height;
        FPreviewPage := CurrentPage;
        if FPreviewBuf.Width >= TNormalReport(Report).PagePixelsSize.cx then
            PlayEnhMetaFile(FPreviewBuf.Canvas.Handle, MF,
                Rect(0, 0, FPreviewBuf.Width, FPreviewBuf.Height))
        else
        begin
            Tmp := TBitmap.Create;
            TRY
                Tmp.PixelFormat := pf32bit;
                Tmp.Width := TNormalReport(Report).PagePixelsSize.cx;
                Tmp.Height := TNormalReport(Report).PagePixelsSize.cy;
                Tmp.Canvas.Brush.Color := clWhite;
                Tmp.Canvas.FillRect(Rect(0, 0, Tmp.Width, Tmp.Height));
                PlayEnhMetaFile(Tmp.Canvas.Handle, MF,
                    Rect(0, 0, Tmp.Width, Tmp.Height));
                Tmp.Canvas.Pixels[0, Tmp.Height - 1] := $DFDFDF;
                //Tmp.PixelFormat := pf24bit;

                if WinVer <= wvNT then
                    DrawStretched(FPreviewBuf, Tmp)
                else
                begin
                    SetStretchBltMode(FPreviewBuf.Canvas.Handle, HALFTONE);
                    SetBrushOrgEx(FPreviewBuf.Canvas.Handle, 0, 0, nil);
                    StretchBlt(FPreviewBuf.Canvas.Handle, 0, 0, FPreviewBuf.Width, FPreviewBuf.Height,
                        Tmp.Canvas.Handle, 0, 0, Tmp.Width, Tmp.Height, SRCCOPY);
                end;

                //FPreviewBuf.SaveToFile( 'C:\TestBmp.bmp' );
            FINALLY
                Tmp.Free;
            END;
        end;
        PaintBox1.Canvas.Draw(0, 0, FPreviewBuf);
    end
    else
    begin
        PaintBox1.Canvas.Draw(0, 0, FPreviewBuf);
    end;
end;

procedure TfmNormalPreview.Print_All_Close(Sender: TObject);
begin
    if PrintDialog1 = nil then
    begin
        PrintDialog1 := TPrintDialog.Create(Self);
    end;

    PrintDialog1.MinPage := 1;
    PrintDialog1.MaxPage := TNormalReport(Report).PagesCount;
    PrintDialog1.FromPage := 1;
    PrintDialog1.ToPage := TNormalReport(Report).PagesCount;
    if (Sender = nil) or PrintDialog1.Execute then
    begin
        if Assigned(Report) then
            (Report as TNormalReport).PrintPages(PrintDialog1.FromPage - 1,
                Printdialog1.ToPage - PrintDialog1.FromPage + 1);
        Close;
    end;
end;

procedure TfmNormalPreview.TB8Click(Sender: TObject);
begin
    Close;
end;

function TfmNormalPreview.ScaleToScreenX(X: Integer): Integer;
var Dc0: HDC;
begin
    Dc0 := GetDC(0);
    Result := Round(
        GetDeviceCaps(Dc0, LogPixelsX) / Printer.PageWidth * 72 * X);
    ReleaseDC(0, Dc0);
end;

function TfmNormalPreview.GetPrinter: TPrinter;
begin
    if Assigned(Report) then
        Result := TNormalReport(Report).Printer
    else
        Result := Printers.Printer;
end;

procedure TfmNormalPreview.Fit_Page(Sender: TObject);
begin
    FitMode := 0;
end;

procedure TfmNormalPreview.Fit_Width(Sender: TObject);
begin
    FitMode := 1;
end;

procedure TfmNormalPreview.As_Is(Sender: TObject);
begin
    FitMode := 2;
end;

procedure TfmNormalPreview.SetFitMode(const Value: Integer);
begin
    FFitMode := Value;
    AdjustFitMode;
end;

procedure TfmNormalPreview.AdjustFitMode;
var K, D: Extended;
begin
    if TNormalReport(Report).PagesCount = 0 then Exit;
    K := TNormalReport(Report).PagePixelsSize.cx;
    D := TNormalReport(Report).PagePixelsSize.cy;
    K := K / D;
    case FFitMode of
        0: begin //  
                SetScrollPos(ScrollBox1.Handle, SB_VERT, 0, TRUE);
                BelowPanel.Height := 0;
                PaintPanel.Height := ScrollBox1.ClientHeight;
                PaintPanel.Width := Round(K * PaintPanel.Height);
                ScrollBox1.Color := clBtnFace;
            end;
        1: begin //  
                BelowPanel.Height := GetSystemMetrics(SM_CYHSCROLL);
                SetScrollPos(ScrollBox1.Handle, SB_HORZ, 0, TRUE);
                PaintPanel.Width := ScrollBox1.ClientWidth - 1;
                PaintPanel.Height := Round(PaintPanel.Width / K);
                if PaintPanel.Width < ScrollBox1.ClientWidth - 1 then
                begin
                    PaintPanel.Width := ScrollBox1.ClientWidth - 1;
                    PaintPanel.Height := Round(PaintPanel.Width / K);
                end;
                ScrollBox1.Color := clWhite;
            end;
        2: begin
                BelowPanel.Height := 0;
                PaintPanel.Width := TNormalReport(Report).PagePixelsSize.cx;
                PaintPanel.Height := TNormalReport(Report).PagePixelsSize.cy;
                ScrollBox1.Color := clBtnFace;
            end;
    end;
    PaintBox1.Width := PaintPanel.Width;
    PaintBox1.Height := PaintPanel.Height - BelowPanel.Height;
end;

procedure TfmNormalPreview.FormDestroy(Sender: TObject);
begin
    FPreviewBuf.Free;
end;

procedure TfmNormalPreview.ScrollBox1Resize(Sender: TObject);
begin
    AdjustFitMode;
end;

procedure TfmNormalPreview.PopupMenu1Popup(Sender: TObject);
begin
    case FFitMode of
        0: N1.Checked := TRUE;
        1: N2.Checked := TRUE;
        2: N3.Checked := TRUE;
    end;
end;

procedure TfmNormalPreview.Horizontal_Expand(Sender: TObject);
var R: TRect;
begin
    R := TB9.BoundsRect;
    R.Top := R.Bottom;
    R.TopLeft := Toolbar1.ClientToScreen(R.TopLeft);
    PopupMenu1.Popup(R.Left, R.Top);
end;

procedure TfmNormalPreview.Print_Page(Sender: TObject);
begin
    if TNormalReport(Report).PagesCount > 0 then
    begin
        TNormalReport(Report).ResetProgress;
        TNormalReport(Report).PrintPages(CurrentPage, 1);
    end;
end;

procedure TfmNormalPreview.Print_Setup(Sender: TObject);
begin
    if PrinterSetupDialog1.Execute then
        PreparePages;
end;

procedure TfmNormalPreview.PreparePages;
begin
    FCurrentPage := 0;
    if Assigned(TNormalReport(Report).OnPrint) then
    begin
        TNormalReport(Report).Clear;
        TNormalReport(Report).Stage := 1;
        TNormalReport(Report).OnPrint(Self);
        if TNormalReport(report).PagesCount = 0 then Exit;
        TNormalReport(Report).Clear;
        TNormalReport(Report).Stage := 2;
        TNormalReport(Report).OnPrint(Self);
    end;
    CurrentPage := 0;
    AdjustFitMode;
end;

procedure TfmNormalPreview.Vertical_Expand(Sender: TObject);
var R: TRect;
begin
    SystemParametersInfo(SPI_GETWORKAREA, 0, @R, 0);
    Top := R.Top;
    Height := R.Bottom - R.Top;
end;

procedure TfmNormalPreview.FormCreate(Sender: TObject);
begin
    //   :
    Caption := rsPreviewCaption;
    N1.Caption := rsN1;
    N2.Caption := rsN2;
    N3.Caption := rsN3;
    N4.Caption := rsN4;
    N5.Caption := rsN5;


    TB1.Hint := rsTB1;
    TB2.Hint := rsTB2;
    TB3.Hint := rsTB3;
    TB4.Hint := rsTB4;
    TB5.Hint := rsTB5;
    TB6.Hint := rsTB6;
    TB7.Hint := rsTB7;
    TB8.Hint := rsTB8;
    TB9.Hint := rsTB9;

    Label1.Caption := rsLabel1;

    //
end;

procedure TfmNormalPreview.N5Click(Sender: TObject);
var F: TfmNormalRange;
begin
    F := TfmNormalRange.Create(Application);
    TRY
        F.Edit1.Text := IntToStr(CurrentPage + 1);
        F.Edit2.Text := IntToStr(CurrentPage + 1);
        F.Edit3.Text := '1';
        F.TotalPages := TNormalReport(Report).PagesCount;
        if F.ShowModal = mrOK then
        begin
            F.Hide;
            TNormalReport(Report).ResetProgress;
            TNormalReport(Report).PrintPages(StrToInt(F.Edit1.Text) - 1,
                StrToInt(F.Edit3.Text));
        end;
    FINALLY
        F.Free;
    END;
end;

procedure TfmNormalPreview.WM_Wheel(var M: TMessage);
var WheelDelta: SmallInt;
begin
    inherited;
    WheelDelta := -Integer(M.wParam) div 65536;
    ScrollBox1.VertScrollBar.Position :=
        ScrollBox1.VertScrollBar.Position + WheelDelta;
end;

procedure TfmNormalPreview.FormKeyDown(Sender: TObject; var Key: Word;
    Shift: TShiftState);
begin
    case Key of
        VK_NEXT: Next_Page(nil);
        VK_PRIOR: Prev_Page(nil);
        VK_UP: ScrollBox1.VertScrollBar.Position :=
            Max(0, ScrollBox1.VertScrollBar.Position - 100);
        VK_DOWN: ScrollBox1.VertScrollBar.Position :=
            Min(ScrollBox1.VertScrollBar.Range - ScrollBox1.ClientHeight,
                ScrollBox1.VertScrollBar.Position + 100);
        VK_LEFT: ScrollBox1.HorzScrollBar.Position :=
            Max(0, ScrollBox1.HorzScrollBar.Position - 100);
        VK_RIGHT: ScrollBox1.HorzScrollBar.Position :=
            Min(ScrollBox1.HorzScrollBar.Range - ScrollBox1.ClientWidth,
                ScrollBox1.HorzScrollBar.Position + 100);
        VK_HOME: First_Page(nil);
        VK_END: Last_Page(nil);
        VK_RETURN: Print_All_Close(Self);
        VK_ESCAPE: Close;
        VK_SPACE: if WindowState = wsNormal then
                WindowState := wsMaximized
            else
                WindowState := wsNormal;
    end;
end;

end.
