//**  ,     VCL Delphi.
//**  Vladimir Kladov, 2000-2010
//**
//**      .
//**     /     
//**  2          
//**   GNU,  Free Software Foundation.
//**
//**       ,    
//** ,       ,   
//**         
//**   .      
//**     GNU.
//**
//**          
//**   GNU.     ,     Free
//** Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
//** MA 02111-1307, USA.
//******************************************************************************

unit NormalReport;
{* TNormalReport -  .
|<br><pre>
  (C) by Vladimir Kladov, 2002
================================================================================

  v1.0.1 (17.10.2002) -  
                        (   )
  v1.1   (24.10.2002) -    ,
                            ( :
			 ..)
  v1.2   (27.11.2002) -    VCL ( API),
			      ;
			NormalLabel   NormalBand (
			  TGraphicControl).
  v1.3   (5.12.2002)  -   ,   
			   /;
			-    :
			   .
  v1.4   (11.12.2002) -  Autosize  TNormalLabel,   true.
			 ParentFont=true    TNormalLabel.
  v1.5   (15.01.2003) -     WordWrap  TNormalLabel,
			 WordWrap=false,      
			 .
  v2.0   (24.07.2003) -     (WordSyllabel)  
                           (MarginLeft, MarginRight, MarginTop,
                        MarginBottom -   2 )  
                        TNormalLabel.
  v2.5   (27.12.2007) -  SamePageCount -   
                            
  v2.9   (19.02.2010) - , , 

     Delphi.   
    .    -  
,   ,       ()
 ,   ,    ...
  -  .   ,    
,    .       
    : ,     ,
   ,        
...     ,    
      Delphi (
 TNormalBand  TNormalLabel).     
    ,   Printer   
.     []  ,  
     [] .

    Key Objects Library   
  .         VCL.

 

e-mail: vk@kolmck.net (   Subkect -    
         - KOL,      !)
WebPage: http://kolmck.net

================================================================================
|</pre>}

interface

{$I VCL_VERSIONS.INC}
/////////////{$DEFINE NRLOGS} // to debug 2 stage printings

uses Windows, Messages, Forms, Dialogs, Classes, Printers, Controls, Graphics,
     SysUtils, NormalPreview, NormalProgress;

type
  ENormalReportException = class( Exception )
  end;

  TMF = HDC; // DC     

  TPaperSize = ( psA4, psA5, psA6, psA3, psLetter, psCustom );
  TBandLayout = ( blLeft, blCenter, blRight, blExpandRight );

  TProcOfObj = procedure of object;

  TPreviewScale = ( psFitPage, psFitWidth, psAsIs );

  TNormalReport = class( TComponent )
  {*  .    ,     
     OnPrint (    ,     
         ,    AddBand  .)
     |<br> .  TNormalBand, TNormalLabel.  }
  private
    FPageTop: Boolean;
    FY: Integer;
    FOnNewPage: TNotifyEvent;
    FPrinter: TPrinter;
    FX: Integer;
    FPrinting: Boolean;
    FPages: TList;
    FImages: TList;
    FStage: Integer;
    FOnPrint: TNotifyEvent;
    FPreviewForm: TfmNormalPreview;
    FBottom: Integer;
    FPagePixelsSize: TSize;
    FPaperSize: TPaperSize;
    FCustomPaperSize: TSize;
    FPageWidth: Integer;
    FPageHeight: Integer;
    FDoubleBufferedPreview: Boolean;
    FCurBandHeight: Integer;
    FOnEndBand: TNotifyEvent;
    fEndBandHandling: Boolean;
    fNewPageHandling: Boolean;
    FExactScale: Boolean;
    FDocName: String;
    FTwoStagePrinting: Boolean;
    FShowProgress: Boolean;
    FOrientation: TPrinterOrientation;
    FSetOrientation: Boolean;
    FPreviewScale: TPreviewScale;
    FOnCreatePreviewForm: TNotifyEvent;
    FOnShowPreviewForm: TNotifyEvent;
    FSamePageCount: Boolean;
    FDirectAcceptNewPage: Boolean;
    FOnNewLine: TNotifyEvent;
    FCurrentPageIdx: Integer;
    procedure SetPageTop(const Value: Boolean);
    procedure SetPrinter(const Value: TPrinter);
    function GetPagesCount: Integer;
    function GetPreviewForm: TfmNormalPreview;
    procedure SetPreviewForm(const Value: TfmNormalPreview);
    function GetCurrentPage: TMF;
    function GetPrinter: TPrinter;
    function GetPagePixelsSize: TSize;
    function GetOrientation: TPrinterOrientation;
    procedure SetPaperSize(const Value: TPaperSize);
    procedure SetCustomPaperSize(const Value: TSize);
    function GetPageHeight: Integer;
    function GetPageWidth: Integer;
    procedure GetPageWidthHeight;
    procedure SetX(const Value: Integer);
    procedure SetY(const Value: Integer);
    function GetPage: Integer;
    function GetImages(Idx: Integer): HEnhMetafile;
    function GetPages(Idx: Integer): TMF;
    function GetPageCount: Integer;
    procedure SetTwoStagePrinting(const Value: Boolean);
    procedure SetShowProgress(const Value: Boolean);
    procedure SetOrientation(const Value: TPrinterOrientation);
    procedure SetPreviewScale(const Value: TPreviewScale);
    procedure SetSamePageCount(const Value: Boolean);
    procedure SetDirectAcceptNewPage(const Value: Boolean);
    procedure SetPage(const Value: Integer);
  protected
    FFastAndDraft: Boolean;
    FPagesAfterStage1: Integer;
    FProgress: TfmNormalProgress;
    FCancelled: Boolean;
    function AddPage: TMF;
    function PaintBand( MF: TMF; Band: TWinControl; X, Y: Integer ): Integer;
    function ScaleX( X: Integer ): Integer;
    function ScaleY( Y: Integer ): Integer;
    procedure PreparePages;
    property PageCount: Integer read GetPageCount;
  protected
    CommandsList: TList;
    FCmdCount, FCmdPage: Integer;
    FCmdCount1: Integer;
    AddingPage: Boolean;
    procedure CountCommand( before_command: Boolean );
    procedure LogOut( const s: String );
  public
    Constructor Create( AOwner: TComponent ); override;
    Destructor Destroy; override;
    procedure Clear;
    {*  .    .  
           . }

    procedure ResetProgress;
    {*   (  ,   
         ).      
       . }

    procedure DoPrintPreview( Proc: TProcOfObj );
    {*   Preview, PreviewModal  Print. }

    procedure DoPrint;
    {*   Print. }

    procedure DoPreview;
    {*   Preview. }

    procedure DoPreviewModal;
    {*   PreviewModal. }

    property PreviewForm: TfmNormalPreview read GetPreviewForm write SetPreviewForm;
    {*  (  )   .
              
         .   -   
        ,      
          .  }

    property PagePixelsSize: TSize read GetPagePixelsSize;
    {*     (, ). }

    property PaperSize: TPaperSize read FPaperSize write SetPaperSize;
    {*  . }

    property CustomPaperSize: TSize read FCustomPaperSize write SetCustomPaperSize;
    {*    (    )  
       PaperSize=psCustom. }

    property Page: Integer read GetPage write SetPage;
    {*    (    1). 
       , ,   OnPrint  OnNewPage  ,
          . }

    property PageWidth: Integer read GetPageWidth;
    {*    ,   TPrinter 
         . .   PaperSize. }

    property PageHeight: Integer read GetPageHeight;
    {*    ,   TPrinter 
         . .   PaperSize. }

    property CurrentPage: TMF read GetCurrentPage;
    {*    ( HDC).    
             API  .
    }

    property Printer: TPrinter read GetPrinter write SetPrinter;
    {*   .   ,   
        (  ). }

    property PageTop: Boolean read FPageTop write SetPageTop;
    {* TRUE,    .
       |<br>:  !
    }

    property X: Integer read FX write SetX;
    {*     . }

    property Y: Integer read FY write SetY;
    {*     . }

    property Pages[ Idx: Integer ]: TMF read GetPages;
    {*     (   
        ,  HDC). }

    property Images[ Idx: Integer ]: HEnhMetafile read GetImages;
    {*       . }

    property Cancelled: Boolean read FCancelled;
    {* TRUE,   (  )   -
       ..,       . }

    procedure Cancel( const Reason: String );
    {*   (  ).     
         ,       . }

  public
    LastX, LastY: Integer;
    {*  X, Y,       }

  public
    procedure AddBand( Band: TWinControl; BandLayout: TBandLayout = blLeft );
    {*  "", ..  ,  . 
          BandLayout.    
        TNormalBand,    TWinControl. }

    procedure AddFooter( Band: TWinControl; BandLayout: TBandLayout = blLeft );
    {*     .    
        Y    (  ). }

    procedure AddRight(Band: TWinControl; AllowWrapBand: Boolean = TRUE);
    {*       .   
               
       . }

    procedure NewPage;
    {*    . }

    procedure NewLine;
    {*    .     
        AddBandRight. }

    property Bottom: Integer read FBottom;
    {*        . }

    property PagesCount: Integer read GetPagesCount;
    {*     (  ). }

    procedure PrintPages( Start, Count: Integer );
    {*    . }

    procedure Print;
    {*     . }

    procedure Preview;
    {*   (). }

    procedure PreviewModal;
    {*   (). . }

    procedure Abort;
    {*   . }

    property Printing: Boolean read FPrinting;
    {* TRUE,    . }

  public
    property Stage: Integer read FStage write FStage;
    {*  .   -  -.  -
             - .
               
       OnPrint  ,      
                  .
       |<br> : -       ,
                ,
             . ,
              .  }
  published

    property OnPrint: TNotifyEvent read FOnPrint write FOnPrint;
    {*  ,     . 
             ,    
       :    ,    .. }

    property OnNewPage: TNotifyEvent read FOnNewPage write FOnNewPage;
    {* ,       , 
           .     ,
          (AddBand)   (AddFooter)  
       ,  - ,   , , 
          ( ,    
                 ). }

    property OnNewLine: TNotifyEvent read FOnNewLine write FOnNewLine;
    {* ,       , 
              
       AddRight,     NewLine.    
       ,    (AddRight)  
       . }

    property OnEndBand: TNotifyEvent read FOnEndBand write FOnEndBand;
    {*        , 
           -  AddRight (   
       ,      -  AddBand). }

  public
    function HeightAvailable: Integer;
    {*  ,    (     Y
           Bottom). }

  published
    property DoubleBufferedPreview: Boolean read FDoubleBufferedPreview write FDoubleBufferedPreview;
    {*       (
        TRUE). }

    property ExactScalePrint: Boolean read FExactScale write FExactScale;
    {*   TRUE,   . }

    property DocumentName: String read FDocName write FDocName;
    {*  ,      
              . }

    property TwoStagePrinting: Boolean read FTwoStagePrinting write SetTwoStagePrinting;
    {* TRUE,   - .   FALSE. 
       TRUE,  On:Print       
                .   
         OnPrint     , 
        ,  ,  ,    
       .       
        . =-      
       ,       , 
         ,      
         . ,       
           ,    
            . }

    property ShowProgress: Boolean read FShowProgress write SetShowProgress;
    {* TRUE,         .
         TRUE. }

    property Orientation: TPrinterOrientation read GetOrientation write SetOrientation;
    {*  . }

    property PreviewScale: TPreviewScale read FPreviewScale write SetPreviewScale;
    {*     . }

    property OnCreatePreviewForm: TNotifyEvent read FOnCreatePreviewForm write FOnCreatePreviewForm;
    {*       . 
            
       .  }

    property OnShowPreviewForm: TNotifyEvent read FOnShowPreviewForm write FOnShowPreviewForm;
    {*  ,     
       . }

    property SamePageCount: Boolean read FSamePageCount write SetSamePageCount;
    {*   TRUE.    
            .    
               
        .       NewPage 
               
              
       (AddBandXXXX). }
    property DirectAcceptNewPage: Boolean read FDirectAcceptNewPage write SetDirectAcceptNewPage;
    property CommandCount: Integer read FCmdCount;
  end;

procedure Register;

{$R *.DCR}

implementation

uses
  NormalBands, NormalResources;

procedure Register;
begin
  RegisterComponents( 'Kladov', [ TNormalReport, TNormalBand, TNormalLabel ] );
end;

{ TNormalReport }

procedure TNormalReport.Abort;
begin
  Clear;
  if Assigned( FPrinter ) then
  begin
    if Printer.Printing then
      Printer.Abort;
  end;
end;

procedure TNormalReport.AddBand(Band: TWinControl; BandLayout: TBandLayout = blLeft);
var MF: TMF;
    OldW: Integer;
begin
  {$IFDEF NRLOGS}
    LogOut( 'AddBand ' + Band.Name + '(' + IntToStr( Band.Width ) + 'x' + IntToStr( Band.Height ) + ')' );
  {$ENDIF}
  if Band = nil then
  begin
    asm
      int 3
    end;
    Exit;
  end;

  if  (Band is TNormalBand)
  and Assigned(TNormalBand(Band).OnPrint) then
      TNormalBand(Band).OnPrint(Self);

  CountCommand( TRUE );

  if FCurBandHeight > 0 then
  begin
    if not fEndBandHandling then
    if Assigned( OnEndBand ) then
    begin
      fEndBandHandling := TRUE;
      OnEndBand( Self );
      fEndBandHandling := FALSE;
    end;
    FX := 0;
    FY := FY + FCurBandHeight;
  end;
  if Band.Height > HeightAvailable then
    NewPage;
  MF := CurrentPage;
  case BandLayout of
  blLeft:  FY := FY + PaintBand( MF, Band, X, Y );
  blRight: FY := FY + PaintBand( MF, Band, PagePixelsSize.cx - Band.Width, Y );
  blCenter: FY := FY + PaintBand( MF, Band, (PagePixelsSize.cx - Band.Width) div 2, Y );
  blExpandRight: begin
                   OldW := Band.Width;
                   try
                     Band.Width := PagePixelsSize.cx - X;
                     FY := FY + PaintBand( MF, Band, X, Y );
                   finally
                     Band.Width := OldW;
                   end;
                 end;
  end;
  FPageTop := FALSE;
  FCurBandHeight := 0;

  CountCommand( FALSE );

end;

procedure TNormalReport.AddFooter(Band: TWinControl; BandLayout: TBandLayout = blLeft);
var MF: TMF;
    OldW: Integer;
begin
  {$IFDEF NRLOGS}
    LogOut( 'AddFooter ' + Band.Name + '(' + IntToStr( Band.Width ) + 'x' + IntToStr( Band.Height ) + ')' );
  {$ENDIF}
  CountCommand( TRUE );

  if Band.Height > HeightAvailable then
    NewPage;
  MF := CurrentPage;
  case BandLayout of
  blLeft:  FBottom := FBottom - PaintBand( MF, Band, 0, FBottom - Band.Height );
  blRight: FBottom := FBottom - PaintBand( MF, Band, PagePixelsSize.cx - Band.Width, FBottom - Band.Height );
  blCenter: FBottom := FBottom - PaintBand( MF, Band, (PagePixelsSize.cx - Band.Width) div 2, FBottom - Band.Height );
  blExpandRight: begin
                   OldW := Band.Width;
                   try
                     Band.Width := PagePixelsSize.cx;
                     FBottom := FBottom - PaintBand( MF, Band, 0, FBottom - Band.Height );
                   finally
                     Band.Width := OldW;
                   end;
                 end;
  end;

  CountCommand( FALSE );
end;

function TNormalReport.AddPage: TMF;
var MF: TMF;
    R: TRect;
    DC0: HDC;

    procedure SimulateProgress;
    begin
      if FProgress.ProgressBar1.Position mod 2 = 0 then
      begin
        FProgress.ProgressBar1.Position := FProgress.ProgressBar1.Position + 10;
        if FProgress.ProgressBar1.Position >= 100 then
          FProgress.ProgressBar1.Position := 99;
      end
        else
      begin
        FProgress.ProgressBar1.Position := FProgress.ProgressBar1.Position - 10;
        if FProgress.ProgressBar1.Position <= 0 then
          FProgress.ProgressBar1.Position := 0;
      end;
    end;

begin
  AddingPage := TRUE;
  TRY
    if  FPages.Count > 0 then
    begin
        FY := 0;
        FX := 0;
    end;
    DC0 := GetDC( 0 );
    R := Rect( 0, 0,
         MulDiv(PagePixelsSize.cx, GetDeviceCaps(DC0, HORZSIZE)*100, GetDeviceCaps(DC0, HORZRES)),
         MulDiv(PagePixelsSize.cy, GetDeviceCaps(DC0, VERTSIZE)*100, GetDeviceCaps(DC0, VERTRES)) );
    MF := CreateEnhMetaFile( DC0, nil, @ R, '' );
    ReleaseDC( 0, DC0 );
    FPages.Add( Pointer( MF ) );
    FCurrentPageIdx := FPages.Count-1;
    FBottom := PagePixelsSize.cy;
    Result := MF;
    if not fNewPageHandling then
    if Assigned( OnNewPage ) then
    begin
      fNewPageHandling := TRUE;
      OnNewPage( Self );
      fNewPageHandling := FALSE;
    end;
    FPageTop := TRUE;
    if FProgress <> nil then
    begin
      if TwoStagePrinting then
      begin
        if FStage = 1 then
        begin
          FProgress.Label1.Caption := rsStage + ' 1 ' + rsPage + ' ' + IntToStr( FPages.Count );
          SimulateProgress;
        end
        else
        begin
          FProgress.Label1.Caption := rsStage + ' 2 ' + rsPage + ' ' + IntToStr( FPages.Count ) +
                                   rsFrom + IntToStr( FPagesAfterStage1 );
          FProgress.ProgressBar1.Position := FPages.Count * 100 div FPagesAfterStage1;
        end;
      end
        else
      begin
        FProgress.Label1.Caption := rsPage + ' ' + IntToStr( FPages.Count );
        SimulateProgress;
      end;
    end;
    Application.ProcessMessages;
    if Assigned( FProgress ) and FProgress.Cancelled then
    begin
      FCancelled := TRUE;
      raise ENormalReportException.Create( rsErr0001 );
    end;
  FINALLY
    AddingPage := FALSE;
  END;
end;

procedure TNormalReport.AddRight(Band: TWinControl; AllowWrapBand: Boolean = TRUE);
var MF: TMF;
begin
  {$IFDEF NRLOGS}
    LogOut( 'AddRight ' + Band.Name + '(' + IntToStr( Band.Width ) + 'x' + IntToStr( Band.Height ) + ')' );
  {$ENDIF}

  CountCommand( TRUE );

  if Band.Height > HeightAvailable then
    NewPage;
  MF := CurrentPage;
  if AllowWrapBand and (Band.Width > PagePixelsSize.cx - X) then
  begin
    if not fEndBandHandling then
    if Assigned( OnEndBand ) then
    begin
      fEndBandHandling := TRUE;
      OnEndBand( Self );
      fEndBandHandling := FALSE;
    end;
    FX := 0;
    FY := FY + FCurBandHeight;
    FCurBandHeight := 0;
    if  Assigned( OnNewLine ) then
        OnNewLine( Self );
  end;
  PaintBand( MF, Band, X, Y );
  FX := X + Band.Width;
  if FCurBandHeight < Band.Height then
    FCurBandHeight := Band.Height;
  FPageTop := Y = 0;

  CountCommand( FALSE );
end;

procedure TNormalReport.Cancel( const Reason: String );
begin
  FCancelled := TRUE;
  raise ENormalReportException.Create( Reason );
end;

procedure TNormalReport.Clear;
var I: Integer;
begin
  FPagePixelsSize.cx := 0;
  FPagePixelsSize.cy := 0;
  for I := FPages.Count-1 downto 0 do
    DeleteEnhMetaFile( Images[ I ] );
  FImages.Clear;
  FPages.Clear;
end;

procedure TNormalReport.CountCommand( before_command: Boolean );
begin
  if AddingPage then Exit;
  if before_command and (FStage = 1) then
    inc( FCmdCount )
  else if not before_command and (FStage = 2) and SamePageCount then
  begin
    inc( FCmdCount );
    if FCmdCount = FCmdCount1 then
    begin
      if  not DirectAcceptNewPage then
      begin
          inc( FCmdPage );
          if FCmdPage < CommandsList.Count then
          begin
            FCmdCount1 := Integer( CommandsList[ FCmdPage ] );
            FCmdCount := 0;
          end;
          AddPage;
      end;
    end;
  end;
end;

constructor TNormalReport.Create(AOwner: TComponent);
begin
  inherited;
  FPages := TList.Create;
  FImages := TList.Create;
  FCustomPaperSize.cx := 210;
  FCustomPaperSize.cy := 270;
  FExactScale := TRUE;
  FShowProgress := TRUE;
  FOrientation := poPortrait;
  FPreviewScale := psFitWidth;
  CommandsList := TList.Create;
  SamePageCount := TRUE;
end;

destructor TNormalReport.Destroy;
begin
  Clear;
  FPages.Free;
  FImages.Free;
  CommandsList.Free;
  inherited;
end;

procedure TNormalReport.DoPreview;
begin
  if FCancelled then Exit;
  if PagesCount = 0 then Exit;
  PreviewForm.ScrollBox1.DoubleBuffered := DoubleBufferedPreview;
  PreviewForm.Report := Self;
  if FProgress <> nil then
    FProgress.Hide;
  if Assigned( OnShowPreviewForm ) then
    OnShowPreviewForm( PreviewForm );
  PreviewForm.Show;
end;

procedure TNormalReport.DoPreviewModal;
begin
  if FCancelled then Exit;
  PreviewForm.ScrollBox1.DoubleBuffered := DoubleBufferedPreview;
  PreviewForm.Report := Self;
  if FProgress <> nil then
    FProgress.Hide;
  if Assigned( OnShowPreviewForm ) then
    OnShowPreviewForm( PreviewForm );
  PreviewForm.ShowModal;
  FreeAndNil( FPreviewForm );
end;

procedure TNormalReport.DoPrint;
begin
  if FPages.Count > 0 then
    PrintPages( 0, FPages.Count );
end;

procedure TNormalReport.DoPrintPreview(Proc: TProcOfObj);
begin
  FCancelled := FALSE;
  if ShowProgress then
  begin
    FProgress := TfmNormalProgress.Create( Application );
    FProgress.Show;
  end
  else
    FProgress := nil;
  TRY
    if Printing then Abort;
    Clear;
    PreparePages;
    if FCancelled or (PageCount = 0) then Exit;
    Proc;
  FINALLY
    FreeAndNil( FProgress );
  END;
end;

function TNormalReport.GetCurrentPage: TMF;
begin
  if FPages.Count = 0 then
    Result := AddPage
  else
    Result := TMF( FPages[ FCurrentPageIdx ] );
end;

function TNormalReport.GetImages(Idx: Integer): HEnhMetafile;
begin
  while FImages.Count <= Idx do
        FImages.Add( nil );
  if  FImages.Items[ Idx ] = nil then
  begin
      FImages.Items[ Idx ] := Pointer( CloseEnhMetafile( Pages[ Idx ] ) );
      FPages.Items[ Idx ] := nil;
  end;
  Result := HENHMETAFILE( FImages.Items[ Idx ] );
end;

function TNormalReport.GetOrientation: TPrinterOrientation;
begin
  if FPrinter = nil then
    Result := FOrientation
  else
  begin
    if Printer <> nil then
      TRY
          Result := Printer.Orientation
      EXCEPT
          Result := FOrientation;
      END
    else
      Result := FOrientation;
  end;
end;

function TNormalReport.GetPage: Integer;
begin
  Result := FPages.Count;
end;

function TNormalReport.GetPageCount: Integer;
begin
  Result := FPages.Count;
end;

function TNormalReport.GetPageHeight: Integer;
begin
  GetPageWidthHeight;
  Result := FPageHeight;
end;

function TNormalReport.GetPagePixelsSize: TSize;
var I: Integer;
    P: TPoint;
    DC0: HDC;
begin
  if (FPagePixelsSize.cx = 0) or (FPagePixelsSize.cy = 0) then
  begin
    case PaperSize of
    psA3: P := Point( 297, 420 );
    psA4: P := Point( 210, 297 );
    psA5: P := Point( 148, 210 );
    psA6: P := Point( 105, 148 );
    psLetter: P := Point( 216, 280 );
    else      P := Point( FCustomPaperSize.cx, FCustomPaperSize.cy );
    end;
    DC0 := GetDC( 0 );
    FPagePixelsSize.cx := Round( (P.x * {0.04} 0.039370 ) * GetDeviceCaps( DC0, LOGPIXELSX ) );
    FPagePixelsSize.cy := Round( (P.y * {0.04} 0.039370 ) * GetDeviceCaps( DC0, LOGPIXELSY ) );
    ReleaseDC( 0, DC0 );
  end;
  Result := FPagePixelsSize;
  if fOrientation = poLandscape then
  begin
    I := Result.cx;
    Result.cx := Result.cy;
    Result.cy := I;
  end;
end;

function TNormalReport.GetPages(Idx: Integer): TMF;
begin
  Result := TMF( FPages[ Idx ] );
end;

function TNormalReport.GetPagesCount: Integer;
begin
  Result := FPages.Count;
  if Result = 0 then
  begin
    if Assigned( OnPrint ) then
    begin
      FFastAndDraft := TRUE;
      TRY
      Clear;
      FStage := 1;
      FOnPrint( Self );
      if FPages.Count = 0 then Exit;
      FPagesAfterStage1 := FPages.Count;
      if TwoStagePrinting then
      begin
        Clear;
        FStage := 2;
        FOnPrint( Self );
      end;
      Result := FPages.Count;
      FINALLY
        Clear;
        FFastAndDraft := FALSE;
      END;
      Exit;
    end;
  end;
  Result := FPages.Count;
end;

function TNormalReport.GetPageWidth: Integer;
begin
  GetPageWidthHeight;
  Result := FPageWidth;
end;

procedure TNormalReport.GetPageWidthHeight;
begin
  if (FPageWidth <> 0) and (FPageHeight <> 0) then Exit;
  if Assigned( FPrinter ) and Printer.Printing then
  begin
    FPageWidth := Printer.PageWidth;
    FPageHeight := Printer.PageHeight;
  end
    else
  begin
    Printer.BeginDoc;
    TRY
      FPageWidth := Printer.PageWidth;
      FPageHeight := Printer.PageHeight;
    FINALLY
      Printer.Abort;
    END;
  end;
end;

function TNormalReport.GetPreviewForm: TfmNormalPreview;
begin
  if FPreviewForm = nil then
  begin
    FPreviewForm := TfmNormalPreview.Create( Application );
    FPreviewForm.Report := Self;
    if DocumentName <> '' then
      FPreviewForm.Caption := DocumentName + ' - ' + FPreviewForm.Caption;
    FPreviewForm.FitMode := Integer( FPreviewScale );
    if Assigned( OnCreatePreviewForm ) then
      OnCreatePreviewForm( FPreviewForm );
  end;
  Result := FPreviewForm;
end;

function TNormalReport.GetPrinter: TPrinter;
begin
  if FPrinter = nil then
  begin
    TRY
      FPrinter := Printers.Printer;
    EXCEPT
      FPrinter := nil;
    END;
    if FPrinter <> nil then
    begin
      TRY
        FPrinter.Orientation := FOrientation;
      EXCEPT
        //FPrinter := nil;
      END;
    end;
  end;
  Result := FPrinter;
end;

function TNormalReport.HeightAvailable: Integer;
begin
  Result := FBottom - FY;
end;

procedure TNormalReport.LogOut(const s: String);
var FName: String;
    F: HFile;
    L: String;
begin
  FName := 'C:\logNR' + IntToStr( FStage ) + '.txt';
  F := CreateFile( PChar( FName ), GENERIC_READ or GENERIC_WRITE,
    0, nil, OPEN_ALWAYS, FILE_ATTRIBUTE_NORMAL, 0 );
  FileSeek( F, 0, 2 );
  L := s + #13#10;
  FileWrite( F, L[1], Length( L ) );
  CloseHandle( F );
end;

procedure TNormalReport.NewLine;
begin
  {$IFDEF NRLOGS}
    LogOut( 'NewLine' );
  {$ENDIF}
  CountCommand( TRUE );

  if FCurBandHeight > HeightAvailable then
    NewPage;
  if not fEndBandHandling then
  if Assigned( OnEndBand ) then
  begin
    fEndBandHandling := TRUE;
    OnEndBand( Self );
    fEndBandHandling := FALSE;
  end;
  FX := 0;
  FY := FY + FCurBandHeight;
  FCurBandHeight := 0;
  FPageTop := FALSE;

  {if  Assigned( OnNewLine ) then
      OnNewLine( Self );}

  CountCommand( FALSE );
end;

procedure TNormalReport.NewPage;
begin
  {$IFDEF NRLOGS}
    LogOut( 'NewPage' );
  {$ENDIF}
  if (FStage = 1) or not SamePageCount or DirectAcceptNewPage then
  begin
      if FStage = 1 then
      begin
        CommandsList.Add( Pointer( FCmdCount ) );
        FCmdCount := 0;
      end;
      AddPage;
  end;
end;

function TNormalReport.PaintBand(MF: TMF; Band: TWinControl; X,
  Y: Integer): Integer;

var OldParent: TWinControl;
    WasVisible: Boolean;
begin
  Result := Band.Height;

  if FFastAndDraft then Exit;

  OldParent := Band.Parent;
  WasVisible := Band.Visible;
  try
    SetParent( Band.Handle, Application.Handle );
    Band.Visible := TRUE;
    LastX := X;
    LastY := Y;
    Band.PaintTo( MF, X, Y );
  finally
    Band.Visible := WasVisible;
    SetParent( Band.Handle, OldParent.Handle );
  end;
end;

procedure TNormalReport.PreparePages;
begin
  if FPages.Count > 0 then Exit; //  
  if Assigned( FOnPrint ) then
  begin
    Clear;
    FStage := 1;
    if  TwoStagePrinting then
        AddPage;
    CommandsList.Clear;
    FCmdCount := 0;
    Y := 0;
    TRY
    FOnPrint( Self );
    EXCEPT on E: Exception do
           begin
             if E is ENormalReportException then
             begin
               FCancelled := TRUE;
               Exit;
             end;
             raise;
           end;
    END;
    FPagesAfterStage1 := FPages.Count;
    if FPagesAfterStage1 = 0 then Exit;
    if FCancelled then Exit;
    if TwoStagePrinting then
    begin
      Clear;
      FStage := 2;
      AddPage;
      FCmdCount := 0;
      FCmdPage := 0;
      Y := 0;
      if CommandsList.Count > 0 then
      begin
        FCmdCount1 := Integer( CommandsList[ 0 ] );
        inc( FCmdPage );
      end;
      TRY
      FOnPrint( Self );
      //Assert( FPagesAfterStage1 = FPages.Count,
      //        rsErr0002 );
      EXCEPT on E: Exception do
             begin
               if E is ENormalReportException then
               begin
                 FCancelled := TRUE;
                 Exit;
               end;
               raise;
             end;
      END;
    end;
  end;
end;

procedure TNormalReport.Preview;
begin
  DoPrintPreview( DoPreview );
end;

procedure TNormalReport.PreviewModal;
begin
  DoPrintPreview( DoPreviewModal );
end;

procedure TNormalReport.Print;
begin
  DoPrintPreview( DoPrint );
end;

procedure TNormalReport.PrintPages(Start, Count: Integer);
var I, j, ToIdx: Integer;
    MF: HEnhMetafile;
    PrintingStarted: Boolean;
    RemoveProgress: Boolean;
    {$IFDEF DEBUG_PRINT}
    Bmp: TBitmap;
    {$ENDIF}
begin
  if PagesCount = 0 then Exit;
  if FCancelled then Exit;
  RemoveProgress := FALSE;
  if (FProgress = nil) and ShowProgress then
  begin
    FProgress := TfmNormalProgress.Create( Application );
    RemoveProgress := TRUE;
  end;
  if FProgress <> nil then
  begin
    FProgress.Caption := rsPrintingInProgress;
    FProgress.Label1.Caption := rsPage + ' 1' + rsFrom + IntToStr( Count );
    FProgress.ProgressBar1.Position := 0;
    FProgress.Show;
    Application.ProcessMessages;
  end;
  TRY
    if FPages.Count = 0 then
    begin
      PreparePages;
      if FCancelled then Exit;
    end;
    PrintingStarted := FALSE;
    if PrintingStarted then;
    j := 0;
    TRY
      ToIdx := Start + Count - 1;
      if ToIdx >= PagesCount then
        ToIdx := PagesCount - 1;
      for I := Start to ToIdx do
      begin
        MF := HEnhMetafile( Images[ I ] );
        if j = 0 then
        begin
          //----{$IFNDEF DEBUG_PRINT}
          Printer.Title := DocumentName;
          Printer.BeginDoc;
          PrintingStarted := TRUE;
          //----{$ENDIF}
        end;
        {$IFDEF DEBUG_PRINT}
        Bmp := TBitmap.Create;
        TRY
          Bmp.Width := PageWidth div 4;
          Bmp.Height := PageHeight div 4;
          SetStretchBltMode( Bmp.Canvas.Handle, halftone );
          PlayEnhMetaFile( Bmp.Canvas.Handle, MF,
                           Rect( 0, 0, Bmp.Width, Bmp.Height ) );
          Bmp.SaveToFile( ExtractFilePath( Application.ExeName ) + 'page' + IntToStr( I ) + '.bmp' );
        FINALLY
          Bmp.Free;
        END;
        //--------------{$ELSE}
        {$ENDIF}
        Printer.Canvas.Brush.Color := clWhite;
        Printer.Canvas.Brush.Style := bsSolid;
        PlayEnhMetaFile( Printer.Canvas.Handle, MF,
                         Rect( 0, 0, PageWidth, PageHeight ) );
        //---------------
        Inc( j );
        if FProgress <> nil then
        begin
          FProgress.Label1.Caption := rsPage + ' ' + IntToStr( J + 1 ) + rsFrom +
                                   IntToStr(Count);
          FProgress.ProgressBar1.Position := (J + 1) * 100 div Count;
          Application.ProcessMessages;
          if FProgress.Cancelled then
          begin
            //----{$IFNDEF DEBUG_PRINT}
            if j > 0 then
              Printer.Abort;
            //----{$ENDIF}
            FCancelled := TRUE;
            Exit;
          end;
        end;
        //--------{$IFNDEF DEBUG_PRINT}
        if I < ToIdx then
          Printer.NewPage;
        //--------{$ENDIF}
      end;
    FINALLY
      //--------{$IFNDEF DEBUG_PRINT}
      if PrintingStarted and not FCancelled then
        Printer.EndDoc;
      //--------{$ENDIF}
    END;
  FINALLY
    if RemoveProgress then
      FreeAndNil( FProgress );
  END;
end;

procedure TNormalReport.ResetProgress;
begin
  FreeAndNil( FProgress );
end;

function TNormalReport.ScaleX(X: Integer): Integer;
begin
  Result := Round( X * Printer.PageWidth / PagePixelsSize.cx );
end;

function TNormalReport.ScaleY(Y: Integer): Integer;
begin
  Result := Round( Y * Printer.PageHeight / PagePixelsSize.cy );
end;

procedure TNormalReport.SetCustomPaperSize(const Value: TSize);
begin
  FCustomPaperSize := Value;
  PaperSize := psCustom;
end;

procedure TNormalReport.SetDirectAcceptNewPage(const Value: Boolean);
begin
  FDirectAcceptNewPage := Value;
end;

procedure TNormalReport.SetOrientation(const Value: TPrinterOrientation);
begin
  FSetOrientation := TRUE;
  FOrientation := Value;
  if Printer <> nil then
  begin
    TRY
        Printer.Orientation := Value;
    EXCEPT

    END;
  end;
end;

procedure TNormalReport.SetPage(const Value: Integer);
var //MF: TMF;
    R: TRect;
    DC0: HDC;

    procedure SimulateProgress;
    begin
      if FProgress.ProgressBar1.Position mod 2 = 0 then
      begin
        FProgress.ProgressBar1.Position := FProgress.ProgressBar1.Position + 10;
        if FProgress.ProgressBar1.Position >= 100 then
          FProgress.ProgressBar1.Position := 99;
      end
        else
      begin
        FProgress.ProgressBar1.Position := FProgress.ProgressBar1.Position - 10;
        if FProgress.ProgressBar1.Position <= 0 then
          FProgress.ProgressBar1.Position := 0;
      end;
    end;

begin
  AddingPage := TRUE;
  TRY
    if  (Value < 1) or (Value > PagesCount) then Exit;
    FY := 0;
    FX := 0;
    DC0 := GetDC( 0 );
    R := Rect( 0, 0,
         MulDiv(PagePixelsSize.cx, GetDeviceCaps(DC0, HORZSIZE)*100, GetDeviceCaps(DC0, HORZRES)),
         MulDiv(PagePixelsSize.cy, GetDeviceCaps(DC0, VERTSIZE)*100, GetDeviceCaps(DC0, VERTRES)) );
    //MF := HDC( FPages[Value-1] );
    FCurrentPageIdx := Value-1;
    ReleaseDC( 0, DC0 );
    FBottom := PagePixelsSize.cy;
    FPageTop := TRUE;
    if FProgress <> nil then
    begin
      if TwoStagePrinting then
      begin
        if FStage = 1 then
        begin
          FProgress.Label1.Caption := rsStage + ' 1 ' + rsPage + ' ' + IntToStr( FPages.Count );
          SimulateProgress;
        end
        else
        begin
          FProgress.Label1.Caption := rsStage + ' 2 ' + rsPage + ' ' + IntToStr( FPages.Count ) +
                                   rsFrom + IntToStr( FPagesAfterStage1 );
          FProgress.ProgressBar1.Position := FPages.Count * 100 div FPagesAfterStage1;
        end;
      end
        else
      begin
        FProgress.Label1.Caption := rsPage + ' ' + IntToStr( FPages.Count );
        SimulateProgress;
      end;
    end;
    Application.ProcessMessages;
    if Assigned( FProgress ) and FProgress.Cancelled then
    begin
      FCancelled := TRUE;
      raise ENormalReportException.Create( rsErr0001 );
    end;
  FINALLY
    AddingPage := FALSE;
  END;
end;

procedure TNormalReport.SetPageTop(const Value: Boolean);
begin
  FPageTop := Value;
end;

procedure TNormalReport.SetPaperSize(const Value: TPaperSize);
begin
  if FPaperSize = Value then Exit;
  if FPrinting then Abort;
  FPaperSize := Value;
end;

procedure TNormalReport.SetPreviewForm(const Value: TfmNormalPreview);
begin
  if FPreviewForm = Value then Exit;
  if FPreviewForm <> nil then
    FPreviewForm.Free;
  FPreviewForm := Value;
end;

procedure TNormalReport.SetPreviewScale(const Value: TPreviewScale);
begin
  FPreviewScale := Value;
end;

procedure TNormalReport.SetPrinter(const Value: TPrinter);
begin
  if FPrinter = Value then Exit;
  if FPrinting then Abort;
  FPrinter := Value;
  FPageWidth := 0;
  FPageHeight := 0;
end;

procedure TNormalReport.SetSamePageCount(const Value: Boolean);
begin
  FSamePageCount := Value;
end;

procedure TNormalReport.SetShowProgress(const Value: Boolean);
begin
  FShowProgress := Value;
end;

procedure TNormalReport.SetTwoStagePrinting(const Value: Boolean);
begin
  FTwoStagePrinting := Value;
end;

procedure TNormalReport.SetX(const Value: Integer);
begin
  if FX = Value then Exit;
  FX := Value;
end;

procedure TNormalReport.SetY(const Value: Integer);
begin
  if FY = Value then Exit;
  FY := Value;
  FCurBandHeight := 0;
end;

end.
