//**  ,     VCL Delphi.
//**  Vladimir Kladov, 2000-2010
//**
//**      .
//**     /     
//**  2          
//**   GNU,  Free Software Foundation.
//**
//**       ,    
//** ,       ,   
//**         
//**   .      
//**     GNU.
//**
//**          
//**   GNU.     ,     Free
//** Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
//** MA 02111-1307, USA.
//******************************************************************************

{
  OdbcQuery.pas (c) by Vladimir Kladov, 2005
  11.03.2005, 27.12.2005
    -  2010

  :
    EODBC         -     ODBC
    TODBCDatabase -        ODBC
    TODBCQuery    -         ODBC
}

unit OdbcQuery;

interface

{$I VCL_VERSIONS.INC}

uses Windows, Classes, {$IFDEF KOL}KOL, KOLmath, err
    {$ELSE}SysUtils, Forms, Db, math{$ENDIF};

type
    EODBC = class(Exception)
    protected
        constructor CreateError(HandleType: SmallInt; fHandle: THandle; ErrCode: SmallInt; const Msg: String);
    public
        constructor CreateEnvError(fHandle: THandle; ErrCode: SmallInt; const Msg: String);
        constructor CreateConnError(fHandle: THandle; ErrCode: SmallInt; const Msg: String);
        constructor CreateStmtError(fHandle: THandle; ErrCode: SmallInt; const Msg: String);
    end;

    {$IFDEF KOL}
    PODBCDatabase = ^TODBCDatabase;
    {$ENDIF}
    TODBCDatabase = {$IFDEF KOL}{$I object_TObj.inc}
    {$ELSE}class(TObject)
    private
        fDoSetAsync: Boolean;
        {$ENDIF}
    protected
        fConnected: Boolean;
        fEnvHandle: THandle;
        fConnHandle: THandle;
        fQueryList: {$IFDEF KOL}PList{$ELSE}TList{$ENDIF};
        FAsync: Boolean;
        function GetAsync: Boolean;
        procedure SetAsync(const Value: Boolean);
        {$IFDEF KOL}
        procedure Init; virtual;
        {$ENDIF}
    public
        ConnectionString: String;
        property EnvironmentHandle: THandle read fEnvHandle;
        property Connection: THandle read fConnHandle;
        property Connected: Boolean read fConnected;
        function Connect: Boolean;
        function DSNConnect(const DSN, User, Pwd: String): Boolean;
        procedure Disconnect;
        destructor Destroy; {$IFDEF KOL} virtual{$ELSE} override{$ENDIF};
        {$IFNDEF KOL}
        constructor Create;
        {$ENDIF}
        procedure BeginTrans;
        procedure RollbackTrans;
        procedure CommitTrans;
        property Async: Boolean read GetAsync write SetAsync;
        property DoSetAsync: Boolean read fDoSetAsync write fDoSetAsync;
    end;

var DefaultODBCDataSource: {$IFDEF KOL}PODBCDatabase
    {$ELSE}TODBCDatabase
    {$ENDIF};

    {$IFDEF KOL}
function NewODBCDatabase: PODBCDatabase;
{$ENDIF}

type
    {$IFDEF KOL}
    PODBCQuery = ^TODBCQuery;
    PSQLStrList = ^TSQLStrList;
    TSQLStrList = object(TStrList)
        FOwnerODBCQuery: PODBCQuery;
    end;
    {$ELSE}
    TODBCQuery = class;

    TSQLStringList = class(TStringList)
    protected
        FOwnerODBCQuery: TODBCQuery;
        procedure Changed; override;
    end;
    {$ENDIF}

    {$IFDEF KOL}
    TFieldType = (ftUnknown, ftString, ftSmallint, ftInteger, ftWord,
        ftBoolean, ftFloat, ftCurrency, ftBCD, ftDate, ftTime, ftDateTime,
        ftBytes, ftVarBytes, ftAutoInc, ftBlob, ftMemo, ftGraphic, ftFmtMemo,
        ftParadoxOle, ftDBaseOle, ftTypedBinary, ftCursor, ftFixedChar, ftWideString,
        ftLargeint, ftADT, ftArray, ftReference, ftDataSet, ftOraBlob, ftOraClob,
        ftVariant, ftInterface, ftIDispatch, ftGuid);
    {$ENDIF}

    TFieldDescription = packed record
        DataType: SmallInt;
        ColumnSize: Integer;
        DecimalDigits: SmallInt;
        Nullable: SmallInt;
        BindPtr: Pointer;
        StrLenInd: PInteger;
    end;
    PFieldDescription = ^TFieldDescription;

    TStrList = {$IFDEF KOL}PStrListEx{$ELSE}TStringList{$ENDIF};
    TSQLStrList = {$IFDEF KOL}PSQLStrList{$ELSE}TSQLStringList{$ENDIF};
    T_ODBC_Database = {$IFDEF KOL}PODBCDatabase{$ELSE}TODBCDatabase{$ENDIF};

    TODBCQuery = {$IFDEF KOL}{$I object_TObj}
    {$ELSE}class(TComponent)
    private
        FSQLErrors: Boolean;
        procedure SetBParam(ParamNum: Integer; const Value: String);
        {$ENDIF}
    protected
        function GetColCount: Integer;
        function GetFieldNames(Idx: Integer): String;
        function GetSField(const S: String): String;
        function GetSFieldN(N: Integer): String;
        function GetIField(const S: String): Integer;
        function GetIFieldN(N: Integer): Integer;
        function GetFField(const S: String): Extended;
        function GetFFieldN(N: Integer): Extended;
        function GetDField(const S: String): TDateTime;
        function GetDFieldN(N: Integer): TDateTime;
        function GetNullField(const S: String): Boolean;
        function GetNullFieldN(N: Integer): Boolean;
        function GetFieldTypes(Idx: Integer): TFieldType;
        function GetFieldPrecision(Idx: Integer): Integer;
    protected
        ODBC: T_ODBC_Database;
        fHandle: Integer;
        fColCount: SmallInt;
        fColumns: TStrList;
        fBindBuffer: Pointer;
        fEof, fBof: Boolean;
        procedure ClearColumns;
        function ColumnByName(const S: String): Integer;
        function ColumnByNameMustBe(const S: String): Integer;
        procedure InternalOpenResultset;
        {$IFDEF KOL}
        procedure Init; virtual;
        {$ENDIF}
        function GetEof: Boolean;
    public
        SQL: TSQLStrList;
        {$IFNDEF KOL}
        constructor Create(AOwner: TComponent); override;
        {$ENDIF}
        destructor Destroy; {$IFDEF KOL} virtual{$ELSE} override{$ENDIF};
        procedure ExecSQL;
        procedure Open;
        procedure Close;
        property FieldCount: Integer read GetColCount;
        property FieldNames[Idx: Integer]: String read GetFieldNames;
        property FieldTypes[Idx: Integer]: TFieldType read GetFieldTypes;
        property FieldPrecision[Idx: Integer]: Integer read GetFieldPrecision;
        function FindFieldIndex(const S: String): Integer;
        property Eof: Boolean read GetEof;
        property Bof: Boolean read fBof;
        procedure First;
        procedure Next;
        procedure Prev;
        procedure Last;
        function More: Boolean;
        procedure Cancel;
        property SFieldN[N: Integer]: String read GetSFieldN;
        property SField[const S: String]: String read GetSField;
        property IFieldN[N: Integer]: Integer read GetIFieldN;
        property IField[const S: String]: Integer read GetIField;
        property FFieldN[N: Integer]: Extended read GetFFieldN;
        property FField[const S: String]: Extended read GetFField;
        property DFieldN[N: Integer]: TDateTime read GetDFieldN;
        property DField[const S: String]: TDateTime read GetDField;
        property NullFieldN[N: Integer]: Boolean read GetNullFieldN;
        property NullField[const S: String]: Boolean read GetNullField;
        property Connection: TODBCDatabase read ODBC write ODBC;
    public
        Parameters: TStringList;
        property BParam[ParamNum: Integer]: String write SetBParam;
        procedure Prepare;
        property SQLErrors: Boolean read FSQLErrors write FSQLErrors;
    end;

    {$IFDEF KOL}
function NewODBCQuery: PODBCQuery;
{$ENDIF}

var ODBC_raise_errors: Boolean = TRUE;
    SqlLenDataAtExec: Integer = -100;

implementation

//-------------------------   API ODBC --------------------------
const ODBC32 = 'ODBC32.dll';
    SQL_OV_ODBC3 = 3;

const
    SQL_CP_ONE_PER_DRIVER = 1;
    SQL_CP_ONE_PER_HENV = 2;
    SQL_ATTR_ODBC_VERSION = 200;
    SQL_IS_INTEGER = -6;
    SQL_ATTR_ASYNC_ENABLE = 4;
    SQL_ASYNC_ENABLE_OFF = 0;
    SQL_NTS = -3;

type SQLResult = Word;
    TSqlDateTimeStamp = packed record
        Year, month, day, hour, minute, second: Word;
        fraction: DWORD; // 1/1000 000 000 sec
    end;
    PSqlDateTimeStamp = ^TSqlDateTimeStamp;

function SqlDateTimeStampToDateTime(TS: PSqlDateTimeStamp): TDateTime;
var ST: TSystemTime;
begin
    ST.wYear := TS.Year;
    ST.wMonth := TS.month;
    ST.wDay := TS.day;
    ST.wHour := TS.hour;
    ST.wMinute := TS.minute;
    ST.wSecond := TS.second;
    ST.wMilliseconds := TS.fraction div 1000000;
    {$IFDEF KOL}
    SystemTime2DateTime(ST, Result);
    {$ELSE}
    Result := SystemTimeToDateTime(ST);
    {$ENDIF}
end;

function StrToFloat(const S: String): Extended;
begin
    if Trim(S) = '' then
        Result := 0
    else
    TRY
        Result :=
            {$IFDEF KOL}
        Str2Double(S);
        {$ELSE}
        SysUtils.StrToFloat(
            StringReplace(
            StringReplace(S, '.', DecimalSeparator, []),
            ',', DecimalSeparator, []));
        {$ENDIF}
    EXCEPT
        Result := 0;
    END;
end;

{$IFDEF KOL}

function IntToStr(i: Integer): String;
begin
    Result := Int2Str(i);
end;

function StrToInt(const s: String): Integer;
begin
    Result := Str2Int(s);
end;
{$ENDIF}

function SQLSetEnvAttr(
    EnvironmentHandle: THandle;
    Attribute: Integer;
    ValuePtr: Pointer;
    StringLength: Integer
    ): SQLResult; stdcall;
    external ODBC32 name 'SQLSetEnvAttr';

const
    SQL_HANDLE_ENV = 1;
    SQL_HANDLE_DBC = 2;
    SQL_HANDLE_STMT = 3;
    SQL_HANDLE_DESC = 4;

function SQLAllocHandle(
    HandleType: SmallInt;
    InputHandle: THandle;
    OutputHandlePtr: PHandle
    ): SQLResult; stdcall;
    external ODBC32 name 'SQLAllocHandle';

function SQLFreeHandle(
    HandleType: SmallInt;
    Handle: THandle
    ): SQLResult; stdcall;
    external ODBC32 name 'SQLFreeHandle';

function SQLDriverConnect(
    ConnectionHandle: THandle;
    WindowHandle: HWnd;
    InConnectionString: PChar;
    StringLength1: SmallInt;
    OutConnectionString: PChar;
    BufferLength: SmallInt;
    StringLength2Ptr: PSmallInt;
    DriverCompletion: Word
    ): SQLResult; stdcall;
    external ODBC32 name 'SQLDriverConnect';

function SQLConnect(
    ConnectionHandle: THandle;
    ServerName: PChar;
    NameLength1: SmallInt;
    UserName: PChar;
    NameLength2: SmallInt;
    Authentication: PChar;
    NameLength3: SmallInt
    ): SQLResult; stdcall;
    external ODBC32 name 'SQLConnect';

function SQLDisconnect(
    ConnectionHandle: THandle
    ): SQLResult; stdcall;
    external ODBC32 name 'SQLDisconnect';

const
    SQL_COMMIT = 0;
    SQL_ROLLBACK = 1;

function SQLSetConnectAttr(
    ConnectionHandle: THandle;
    Attribute: Integer;
    ValuePtr: Pointer;
    StringLength: Integer
    ): SQLResult; stdcall;
    external ODBC32 name 'SQLSetConnectAttr';

function SQLGetConnectAttr(
    ConnectionHandle: THandle;
    Attribute: Integer;
    ValuePtr: Pointer;
    BufferLength: Integer;
    StringLengthPtr: PInteger
    ): SQLResult; stdcall;
    external ODBC32 name 'SQLGetConnectAttr';

function SQLEndTran(
    HandleType: SmallInt;
    Handle: THandle;
    CompletionType: SmallInt
    ): SQLResult; stdcall;
    external ODBC32 name 'SQLEndTran';

function SQLGetDiagRec(
    HandleType: SmallInt;
    Handle: THandle;
    RecNumber: Integer;
    Sqlstate: PChar;
    NativeErrorPtr: PInteger;
    MessageText: PChar;
    BufferLength: SmallInt;
    TextLengthPtr: PSmallInt
    ): SQLResult; stdcall;
    external ODBC32 name 'SQLGetDiagRec';

function SQLExecDirect(
    StatementHandle: THandle;
    StatementText: PChar;
    TextLength: Integer
    ): SQLResult; stdcall;
    external ODBC32 name 'SQLExecDirect';

function SQLExecute(
    StatementHandle: THandle
    ): SQLResult; stdcall;
    external ODBC32 name 'SQLExecute';

function SQLCancel(
    StatementHandle: THandle
    ): SQLResult; stdcall;
    external ODBC32 name 'SQLCancel';

function SQLPrepare(
    StatementHandle: THandle;
    StatementText: PChar;
    TextLength: Integer
    ): SQLResult; stdcall;
    external ODBC32 name 'SQLPrepare';

function SQLMoreResults(
    StatementHandle: THandle
    ): SQLResult; stdcall;
    external ODBC32 name 'SQLMoreResults';

const
    SQL_ATTR_AUTOCOMMIT = 102;
    SQL_AUTOCOMMIT_OFF = 0;
    SQL_AUTOCOMMIT_ON = 1;
    SQL_CLOSE = 0;

function SQLFreeStmt(
    StatementHandle: THandle;
    Option: SmallInt
    ): SQLResult; stdcall;
    external ODBC32 name 'SQLFreeStmt';

function SQLNumResultCols(
    StatementHandle: THandle;
    ColumnCountPtr: PSmallInt
    ): SQLResult; stdcall;
    external ODBC32 name 'SQLNumResultCols';

function SQLDescribeCol(
    StatementHandle: THandle;
    ColumnNumber: SmallInt;
    ColumnName: PChar;
    BufferLength: SmallInt;
    NameLengthPtr: PSmallInt;
    DataTypePtr: PSmallInt;
    ColumnSizePtr: PDWORD;
    DecimalDigitsPtr: PSmallInt;
    NullablePtr: PSmallInt
    ): SQLResult; stdcall;
    external ODBC32 name 'SQLDescribeCol';

const
    SQL_UNKNOWN_TYPE = 0;
    SQL_CHAR = 1;
    SQL_NUMERIC = 2;
    SQL_DECIMAL = 3;
    SQL_INTEGER = 4;
    SQL_SMALLINT = 5;
    SQL_FLOAT = 6;
    SQL_REAL = 7;
    SQL_DOUBLE = 8;
    SQL_DATE = 9;
    SQL_TIME = 10;
    SQL_TYPE_TIMESTAMP = 93;
    SQL_VARCHAR = 12;
    SQL_WVARCHAR = -96;
    SQL_BINARY = -2;
    SQL_VARBINARY = -4;

function SQLBindCol(
    StatementHandle: THandle;
    ColumnNumber: Word;
    TargetType: SmallInt;
    TargetValuePtr: Pointer;
    BufferLength: Integer;
    StrLen_or_Ind: PInteger
    ): SQLResult; stdcall;
    external ODBC32 name 'SQLBindCol';

const
    SQL_ATTR_ROW_BIND_TYPE = 5;
    SQL_ATTR_ROW_ARRAY_SIZE = 27;

function SQLSetStmtAttr(
    StatementHandle: THandle;
    Attribute: Integer;
    ValuePtr: Pointer;
    StringLength: Integer
    ): SQLResult; stdcall;
    external ODBC32 name 'SQLSetStmtAttr';

const
    SQL_FETCH_NEXT = 1;
    SQL_FETCH_FIRST = 2;
    SQL_FETCH_LAST = 3;
    SQL_FETCH_PRIOR = 4;
    SQL_FETCH_ABSOLUTE = 5;
    SQL_FETCH_RELATIVE = 6;

    SQL_NULL_DATA = -1;
    SQL_NO_DATA = 100;

function SQLFetch(
    StatementHandle: THandle
    ): SQLResult; stdcall;
    external ODBC32 name 'SQLFetch';

function SQLFetchScroll(
    StatementHandle: THandle;
    FetchOrientation: SmallInt;
    FetchOffset: Integer
    ): SQLResult; stdcall;
    external ODBC32 name 'SQLFetchScroll';

function SQLGetData(
    StatementHandle: THandle;
    ColumnNumber: SmallInt;
    TargetType: SmallInt;
    TargetValuePtr: Pointer;
    BufferLength: Integer;
    StrLen_or_IndPtr: PInteger
    ): SQLResult; stdcall;
    external ODBC32 name 'SQLGetData';

function SQLGetInfo(
    ConnectionHandle: THandle;
    InfoType: SmallInt;
    InfoValuePtr: Pointer;
    BufferLength: Integer;
    StringLengthPtr: PSmallInt
    ): SQLResult; stdcall;
    external ODBC32 name 'SQLGetInfo';

function SQLBindParameter(
    StatementHandle: THandle;
    ParameterNumber: Word;
    InputOutputType: SmallInt;
    ValueType: SmallInt;
    ParameterType: SmallInt;
    ColumnSize: DWORD;
    DecimalDigits: SmallInt;
    ParameterValuePtr: Pointer;
    BufferLength: Integer;
    StrLen_or_IndPtr: PInteger
    ): SQLResult; stdcall;
    external ODBC32 name 'SQLBindParameter';

function SQLParamData(
    StatementHandle: THandle;
    ValuePtrPtr: Pointer
    ): SQLResult; stdcall;
    external ODBC32 name 'SQLParamData';

function SQLPutData(
    StatementHandle: THandle;
    DataPtr: Pointer;
    StrLen_or_Ind: Integer
    ): SQLResult; stdcall;
    external ODBC32 name 'SQLPutData';

const
    SQL_PARAM_INPUT = 1;
    SQL_NEED_DATA = 99;


    {$IFDEF KOL}

function NewODBCDatabase: PODBCDatabase;
begin
    new(Result, Create);
end;
{$ENDIF}

{ TODBCDatabase }

procedure TODBCDatabase.BeginTrans;
begin
    SQLSetConnectAttr(fConnHandle, SQL_ATTR_AUTOCOMMIT,
        Pointer(SQL_AUTOCOMMIT_OFF), SQL_IS_INTEGER);
end;

procedure TODBCDatabase.CommitTrans;
begin
    SQLEndTran(SQL_HANDLE_DBC, fConnHandle, SQL_COMMIT);
    SQLSetConnectAttr(fConnHandle, SQL_ATTR_AUTOCOMMIT,
        Pointer(SQL_AUTOCOMMIT_ON), SQL_IS_INTEGER);
end;

function TODBCDatabase.Connect: Boolean;
var OutConnString: array[0..1023] of Char;
    OutLen: SmallInt;
    R: Word;
begin
    Result := TRUE;
    if Connected then Exit;
    R := SQLAllocHandle(SQL_HANDLE_DBC, fEnvHandle, @fConnHandle);
    if R > 1 then
        raise EODBC.CreateEnvError(fEnvHandle, R, '   ');
    R := SQLDriverConnect(fConnHandle,
        {$IFDEF KOL}Applet.Handle{$ELSE}Application.Handle{$ENDIF},
        PChar(ConnectionString),
        Length(ConnectionString), @OutConnString[0], Sizeof(OutConnString),
        @OutLen, 0);
    if R <= 1 then
    begin
        fConnected := TRUE;
        Async := FAsync;
    end
    else
    begin
        TRY
            raise EODBC.CreateConnError(fConnHandle, R, '   ');
        FINALLY
            SQLFreeHandle(SQL_HANDLE_DBC, fConnHandle);
            fConnHandle := 0;
        END;
    end;
    Result := fConnected;
end;

{$IFDEF KOL}

procedure TODBCDatabase.Init;
var R: WORD;
begin
    inherited;
    fQueryList := NewList;
    R := SQLAllocHandle(SQL_HANDLE_ENV, 0, @fEnvHandle);
    if R > 1 then
        raise EODBC.CreateError(0, 0, R, '    ODBC');
    if DefaultODBCDataSource = nil then
        DefaultODBCDataSource := @Self;
    R := SQLSetEnvAttr(fEnvHandle, SQL_ATTR_ODBC_VERSION,
        Pointer(SQL_OV_ODBC3), SQL_IS_INTEGER);
    if R > 1 then
        raise EODBC.CreateEnvError(fEnvHandle, R, '    ODBC 3.0');
end;
{$ELSE}

constructor TODBCDatabase.Create;
var R: WORD;
begin
    inherited;
    fQueryList := TList.Create;
    R := SQLAllocHandle(SQL_HANDLE_ENV, 0, @fEnvHandle);
    if R > 1 then
        raise EODBC.CreateError(0, 0, R, '    ODBC');
    if DefaultODBCDataSource = nil then
        DefaultODBCDataSource := Self;
    R := SQLSetEnvAttr(fEnvHandle, SQL_ATTR_ODBC_VERSION,
        Pointer(SQL_OV_ODBC3), SQL_IS_INTEGER);
    if R > 1 then
        raise EODBC.CreateEnvError(fEnvHandle, R, '    ODBC 3.0');
end;
{$ENDIF}

destructor TODBCDatabase.Destroy;
begin
    if DefaultODBCDataSource = {$IFDEF KOL}@{$ENDIF}Self then
        DefaultODBCDataSource := nil;
    if Connected then
        Disconnect;
    SQLFreeHandle(SQL_HANDLE_ENV, fEnvHandle);
    fQueryList.Free;
    inherited;
end;

procedure TODBCDatabase.Disconnect;
var i: Integer;
    Q: {$IFDEF KOL}PODBCQuery{$ELSE}TODBCQuery{$ENDIF};
begin
    if not Connected then Exit;
    for i := fQueryList.Count - 1 downto 0 do
    begin
        Q := fQueryList.Items[i];
        Q.Close;
    end;
    SQLDisconnect(fConnHandle);
    SQLFreeHandle(SQL_HANDLE_DBC, fConnHandle);
    fConnHandle := 0;
    fConnected := FALSE;
end;

function TODBCDatabase.DSNConnect(const DSN, User, Pwd: String): Boolean;
var R: Word;
begin
    Result := TRUE;
    if Connected then Exit;
    R := SQLAllocHandle(SQL_HANDLE_DBC, fEnvHandle, @fConnHandle);
    if R > 1 then
        raise EODBC.CreateEnvError(fEnvHandle, R, '   ');
    R := SQLConnect(fConnHandle, PChar(DSN), Length(DSN),
        PChar(User), Length(User), PChar(Pwd), Length(Pwd));
    if R <= 1 then
    begin
        fConnected := TRUE;
        Async := FAsync;
    end
    else
    begin
        TRY
            raise EODBC.CreateConnError(fConnHandle, R,
                '   ');
        FINALLY
            SQLFreeHandle(SQL_HANDLE_DBC, fConnHandle);
            fConnHandle := 0;
        END;
    end;
    Result := fConnected;
end;

function TODBCDatabase.GetAsync: Boolean;
var Value: Integer;
    R: Word;
begin
    Result := FAsync;
    if not Connected then Exit;
    R := SQLGetConnectAttr(fConnHandle, SQL_ATTR_ASYNC_ENABLE, @Value,
        0, nil);
    if R > 1 then Exit;
    FAsync := Value <> 0;
end;

procedure TODBCDatabase.RollbackTrans;
begin
    SQLEndTran(SQL_HANDLE_DBC, fConnHandle, SQL_ROLLBACK);
    SQLSetConnectAttr(fConnHandle, SQL_ATTR_AUTOCOMMIT,
        Pointer(SQL_AUTOCOMMIT_ON), SQL_IS_INTEGER);
end;

procedure TODBCDatabase.SetAsync(const Value: Boolean);
var R: Word;
begin
    FAsync := Value;
    if not Connected then Exit;
    if not DoSetAsync then Exit;
    R := SQLSetConnectAttr(fConnHandle, SQL_ATTR_ASYNC_ENABLE,
        Pointer(FAsync), SQL_NTS);
    if R > 1 then
        raise EODBC.CreateError(SQL_HANDLE_DBC, fConnHandle, R,
            '     = ' +
            IntToStr(Integer(Value)));
end;

{ EODBC }

constructor EODBC.CreateConnError(fHandle: THandle; ErrCode: SmallInt;
    const Msg: String);
begin
    CreateError(SQL_HANDLE_DBC, fHandle, ErrCode, Msg);
end;

constructor EODBC.CreateEnvError(fHandle: THandle; ErrCode: SmallInt; const Msg: String);
begin
    CreateError(SQL_HANDLE_ENV, fHandle, ErrCode, Msg);
end;

constructor EODBC.CreateError(HandleType: SmallInt; fHandle: THandle;
    ErrCode: SmallInt; const Msg: String);
var ErrStringBuffer: array[0..4095] of Char;
    ErrStrLen: SmallInt;
    SQLState: array[0..5] of Char;
    NativeError: Integer;
begin
    ErrStringBuffer[0] := #0;
    ErrStrLen := 0;
    SQLGetDiagRec(HandleType, fHandle, 1, SQLState, @NativeError,
        ErrStringBuffer, Sizeof(ErrStringBuffer), @ErrStrLen);
    ErrStringBuffer[ErrStrLen] := #0;
    SQLState[5] := #0;
    inherited Create({$IFDEF KOL}e_custom, {$ENDIF}
        Msg + ',  : ' + IntToStr(ErrCode) +
        #13#10 + SQLState + ' ' + ErrStringBuffer);
end;

constructor EODBC.CreateStmtError(fHandle: THandle; ErrCode: SmallInt;
    const Msg: String);
begin
    CreateError(SQL_HANDLE_STMT, fHandle, ErrCode, Msg);
end;

{$IFDEF KOL}

function NewODBCQuery: PODBCQuery;
begin
    new(Result, Create);
end;
{$ENDIF}

{ TODBCQuery }

procedure TODBCQuery.ClearColumns;
var i: Integer;
    FD: PFieldDescription;
begin
    for i := fColumns.Count - 1 downto 0 do
    begin
        FD := Pointer(fColumns.Objects[i]);
        FreeMem(FD);
    end;
    fColumns.Clear;
    fColCount := 0;
    if fBindBuffer <> nil then
    begin
        FreeMem(fBindBuffer);
        fBindBuffer := nil;
    end;
end;

procedure TODBCQuery.Close;
begin
    fColumns.Clear;
    FreeAndNil(Parameters);
    if fHandle = 0 then Exit;
    //SQLFreeStmt( fHandle, SQL_CLOSE );
    SQLFreeHandle(SQL_HANDLE_STMT, fHandle);
    fHandle := 0;
end;

function TODBCQuery.ColumnByName(const S: String): Integer;
var i: Integer;
begin
    for i := 0 to fColumns.Count - 1 do
    begin
        if AnsiCompareText(S, fColumns{$IFDEF KOL}.Items{$ENDIF} [i]) = 0 then
        begin
            Result := i;
            Exit;
        end;
    end;
    Result := -1;
end;

function TODBCQuery.ColumnByNameMustBe(const S: String): Integer;
begin
    Result := ColumnByName(S);
    if Result < 0 then
        raise EODBC.Create({$IFDEF KOL}e_custom, {$ENDIF}
            ' ' + S + '     ');
end;

{$IFDEF KOL}

procedure TODBCQuery.Init;
begin
    inherited;
    new(SQL, Create);
    SQL.FOwnerODBCQuery := @Self;
    ODBC := DefaultODBCDataSource;
    if ODBC <> nil then
        ODBC.fQueryList.Add(@Self);
    fColumns := NewStrListEx;
end;
{$ELSE}

constructor TODBCQuery.Create;
begin
    inherited;
    SQLErrors := TRUE;
    SQL := TSQLStringList.Create;
    SQL.FOwnerODBCQuery := Self;
    ODBC := DefaultODBCDataSource;
    if ODBC <> nil then
        ODBC.fQueryList.Add(Self);
    fColumns := TStringList.Create;
end;
{$ENDIF}

function TODBCQuery.GetEof: Boolean;
begin
    Result := fEof or (fColumns = nil) or (fColumns.Count = 0);
end;

destructor TODBCQuery.Destroy;
begin
    Close;
    if ODBC <> nil then
        ODBC.fQueryList.Remove({$IFDEF KOL}@{$ENDIF}Self);
    SQL.Free;
    fColumns.Free;
    inherited;
end;

procedure TODBCQuery.ExecSQL;
var R, R2: SmallInt;
    Txt: String;
    i, j, NativeError, MsgLen: Integer;
    SqlState: array[0..6] of Char;
    Message1: array[0..512] of Char;
    found: Boolean;
begin
    ClearColumns;
    if fHandle = 0 then
    begin
        R := SQLAllocHandle(SQL_HANDLE_STMT, ODBC.fConnHandle, @fHandle);
        if R > 1 then
            raise EODBC.CreateConnError(ODBC.fConnHandle, R,
                '   :' + Copy(SQL.Text, 1, 100));
    end;
    Txt := SQL.Text;
    R := SmallInt(SQLExecDirect(fHandle, PChar(Txt), Length(Txt)));
    while (R = SQL_NEED_DATA) {and (Parameters <> nil)} do
    begin
        if Parameters = nil then
            raise Exception.Create('  ');
        R := SQLParamData(fHandle, @i);
        if R = SQL_NEED_DATA then
        begin
            found := FALSE;
            for j := 0 to Parameters.Count - 1 do
            begin
                if Parameters.Objects[j] = Pointer(i) then
                begin
                    found := TRUE;
                    R2 := SQLPutData(fHandle, PChar(Parameters[j]), Length(Parameters[j]));
                    if R2 > 1 then
                        raise EODBC.CreateStmtError(fHandle, R2,
                            '    ' + IntToStr(i));
                    break;
                end;
            end;
            if not found then
                raise Exception.Create('   ' + IntToStr(i));
        end;
    end;
    if ODBC_raise_errors and SQLErrors then
        if (R = -1) {or (R = 1)} then
        begin
            Txt := Copy(Txt, 1, 100) + #13#10;
            i := 0;
            FillChar(SqlState, Sizeof(SqlState), 0);
            while TRUE do
            begin
                inc(i);
                if 100 = SQLGetDiagRec(SQL_HANDLE_STMT, fHandle, i, SqlState, @NativeError,
                    Message1, Sizeof(Message1), @MsgLen) then break;
                Txt := Txt + #13#10 + Copy(Message1, 1, MsgLen);
            end;
            //SQL.SaveToFile( 'c:\bad.sql.txt' );
            raise EODBC.Create('   :' + Txt);
        end;
    FreeAndNil(Parameters);
end;

function TODBCQuery.FindFieldIndex(const S: String): Integer;
var i: Integer;
begin
    Result := -1;
    for i := 0 to fColumns.Count - 1 do
        if AnsiCompareText(fColumns{$IFDEF KOL}.Items{$ENDIF} [i], S) = 0 then
        begin
            Result := i;
            break;
        end;
end;

procedure TODBCQuery.First;
var R: Word;
begin
    fEof := FALSE;
    fBof := FALSE;
    R := //SQLFetchScroll( fHandle, SQL_FETCH_FIRST, 0 );
    SQLFetch(fHandle);
    if R > 1 then
    begin
        fEof := TRUE;
        fBof := TRUE;
    end;
end;

function TODBCQuery.GetColCount: Integer;
begin
    Result := fColumns.Count;
end;

procedure TODBCQuery.Cancel;
begin
    SQLCancel(fHandle);
end;

function TODBCQuery.GetDField(const S: String): TDateTime;
var i: Integer;
begin
    i := ColumnByNameMustBe(S);
    Result := 0;
    if i >= 0 then
        Result := DFieldN[i];
end;

function TODBCQuery.GetDFieldN(N: Integer): TDateTime;
var FD: PFieldDescription;
    s: String;
begin
    Result := 0;
    FD := Pointer(fColumns.Objects[N]);
    if FD.StrLenInd^ <> SQL_NULL_DATA then
        CASE FD.DataType OF
            {SQL_NUMERIC:
              Result := NumericToExtended( PSqlNumeric( FD.BindPtr )^ );}
            SQL_DECIMAL:
                begin
                    // todo
                    asm
        int 3
                    end;
                end;
            SQL_CHAR,
                SQL_VARCHAR:
                begin
                    TRY
                        Result := StrToFloat(Copy(PChar(FD.BindPtr), 1, FD.StrLenInd^));
                    EXCEPT
                        raise; //   
                    END;
                end;
            SQL_WVARCHAR:
                begin
                    TRY
                        s := PWideChar(FD.BindPtr)^;
                        Result := StrToFloat(s);
                    EXCEPT
                        raise;
                    END;
                end;
            SQL_INTEGER:
                Result := PInteger(FD.BindPtr)^;
            SQL_SMALLINT:
                Result := PSmallInt(FD.BindPtr)^;
            SQL_DOUBLE,
                SQL_FLOAT:
                Result := PDouble(FD.BindPtr)^;
            SQL_REAL:
                Result := PSingle(FD.BindPtr)^;
            {SQL_DATETIME:
              Result := PDateTime( FD.BindPtr )^;}
            SQL_TYPE_TIMESTAMP:
                Result := SqlDateTimeStampToDateTime(FD.BindPtr);
        else
            raise EODBC.Create({$IFDEF KOL}e_custom, {$ENDIF}
                '     ' + fColumns{$IFDEF KOL}.Items{$ENDIF} [N]);
        END;
end;

function TODBCQuery.GetFField(const S: String): Extended;
var i: Integer;
begin
    i := ColumnByNameMustBe(S);
    Result := 0;
    if i >= 0 then
        Result := FFieldN[i];
end;

function TODBCQuery.GetFFieldN(N: Integer): Extended;
var FD: PFieldDescription;
    s: String;
begin
    Result := 0;
    FD := Pointer(fColumns.Objects[N]);
    if FD.StrLenInd^ <> SQL_NULL_DATA then
        CASE FD.DataType OF
            {SQL_NUMERIC:
              Result := NumericToExtended( PSqlNumeric( FD.BindPtr )^ );}
            SQL_DECIMAL:
                begin
                    // todo
                    asm
        int 3
                    end;
                end;
            SQL_CHAR,
                SQL_VARCHAR:
                begin
                    TRY
                        Result := StrToFloat(Copy(PChar(FD.BindPtr), 1, FD.StrLenInd^));
                    EXCEPT
                        raise; //   
                    END;
                end;
            SQL_WVARCHAR:
                begin
                    TRY
                        s := PWideChar(FD.BindPtr)^;
                        Result := StrToFloat(s);
                    EXCEPT
                        raise;
                    END;
                end;
            SQL_INTEGER:
                Result := PInteger(FD.BindPtr)^;
            SQL_SMALLINT:
                Result := PSmallInt(FD.BindPtr)^;
            SQL_DOUBLE,
                SQL_FLOAT:
                Result := PDouble(FD.BindPtr)^;
            SQL_REAL:
                Result := PSingle(FD.BindPtr)^;
            SQL_TYPE_TIMESTAMP:
                Result := SqlDateTimeStampToDateTime(FD.BindPtr);
        else
            raise EODBC.Create({$IFDEF KOL}e_custom, {$ENDIF}
                '     ' + fColumns{$IFDEF KOL}.Items{$ENDIF} [N]);
        END;
end;

function TODBCQuery.GetFieldNames(Idx: Integer): String;
begin
    Result := fColumns{$IFDEF KOL}.Items{$ENDIF} [Idx];
end;

function TODBCQuery.GetFieldPrecision(Idx: Integer): Integer;
var FD: PFieldDescription;
begin
    FD := Pointer(fColumns.Objects[Idx]);
    Result := FD.DecimalDigits;
end;

function TODBCQuery.GetFieldTypes(Idx: Integer): TFieldType;
var FD: PFieldDescription;
begin
    {ftUnknown, ftString, ftSmallint, ftInteger, ftWord,
      ftBoolean, ftFloat, ftCurrency, ftBCD, ftDate, ftTime, ftDateTime,
      ftBytes, ftVarBytes, ftAutoInc, ftBlob, ftMemo, ftGraphic, ftFmtMemo,
      ftParadoxOle, ftDBaseOle, ftTypedBinary, ftCursor, ftFixedChar, ftWideString,
      ftLargeint, ftADT, ftArray, ftReference, ftDataSet, ftOraBlob, ftOraClob,
      ftVariant, ftInterface, ftIDispatch, ftGuid}
    FD := Pointer(fColumns.Objects[Idx]);
    CASE FD.DataType OF
        SQL_NUMERIC: Result := ftFloat;
        SQL_DECIMAL: Result := ftBCD;
        SQL_CHAR,
            SQL_VARCHAR: Result := ftString;
        SQL_WVARCHAR: Result := ftWideString;
        SQL_INTEGER: Result := ftInteger;
        SQL_SMALLINT: Result := ftSmallint;
        SQL_DOUBLE,
            SQL_FLOAT,
            SQL_REAL: Result := ftFloat;
        SQL_TYPE_TIMESTAMP: Result := ftDateTime;
        SQL_VARBINARY,
            SQL_BINARY,
            -1: Result := ftString;
    else
        raise EODBC.Create({$IFDEF KOL}e_custom, {$ENDIF}
            '     ' + fColumns{$IFDEF KOL}.Items{$ENDIF} [Idx]);
    END;
end;

function TODBCQuery.GetIField(const S: String): Integer;
var i: Integer;
begin
    i := ColumnByNameMustBe(S);
    Result := 0;
    if i >= 0 then
        Result := IFieldN[i];
end;

function TODBCQuery.GetIFieldN(N: Integer): Integer;
var FD: PFieldDescription;
    s: String;
begin
    Result := 0;
    FD := Pointer(fColumns.Objects[N]);
    if FD.StrLenInd^ <> SQL_NULL_DATA then
        CASE FD.DataType OF
            {SQL_NUMERIC:
              Result := Trunc( NumericToExtended( PSqlNumeric( FD.BindPtr )^ ) );}
            SQL_DECIMAL:
                begin
                    // todo
                    asm
        int 3
                    end;
                end;
            SQL_CHAR,
                SQL_VARCHAR:
                begin
                    TRY
                        Result := StrToInt(Copy(PChar(FD.BindPtr), 1, FD.StrLenInd^));
                    EXCEPT
                        raise; //   
                    END;
                end;
            SQL_WVARCHAR:
                begin
                    TRY
                        s := PWideChar(FD.BindPtr)^;
                        Result := StrToInt(s);
                    EXCEPT
                        raise;
                    END;
                end;
            SQL_INTEGER:
                Result := PInteger(FD.BindPtr)^;
            SQL_SMALLINT:
                Result := PSmallInt(FD.BindPtr)^;
            SQL_DOUBLE,
                SQL_FLOAT:
                Result := Trunc(PDouble(FD.BindPtr)^);
            SQL_REAL:
                Result := Trunc(PSingle(FD.BindPtr)^);
            SQL_TYPE_TIMESTAMP:
                Result := Trunc(SqlDateTimeStampToDateTime(FD.BindPtr));
        else
            raise EODBC.Create({$IFDEF KOL}e_custom, {$ENDIF}
                '     ' + fColumns{$IFDEF KOL}.Items{$ENDIF} [N]);
        END;
end;

function TODBCQuery.GetNullField(const S: String): Boolean;
var i: Integer;
begin
    i := ColumnByNameMustBe(S);
    Result := TRUE;
    if i >= 0 then
        Result := NullFieldN[i];
end;

function TODBCQuery.GetNullFieldN(N: Integer): Boolean;
var FD: PFieldDescription;
    StrLenOrInd: Integer;
begin
    FD := Pointer(fColumns.Objects[N]);
    if (FD <> nil) and (FD.StrLenInd <> nil) then
        Result := FD.StrLenInd^ = SQL_NULL_DATA
    else
    begin
        if (SQLGetData(fHandle, N, FD.DataType, nil, 0, @StrLenOrInd)
            = SQL_NO_DATA) or (StrLenOrInd = SQL_NULL_DATA) then
            Result := TRUE
        else
            Result := FALSE;
    end;
end;

function TODBCQuery.GetSField(const S: String): String;
var i: Integer;
begin
    i := ColumnByNameMustBe(S);
    Result := '';
    if i >= 0 then
        Result := SFieldN[i];
end;

function TODBCQuery.GetSFieldN(N: Integer): String;
var FD: PFieldDescription;
    R: Word;
    L: Integer;
    ShortBuf: array[0..255] of Byte;
    targetType: SmallInt;
begin
    FD := Pointer(fColumns.Objects[N]);
    if (FD.ColumnSize > 0) and (FD.StrLenInd^ = SQL_NULL_DATA) then
        Result := ''
    else
        CASE FD.DataType OF
            (*
            SQL_NUMERIC:
              begin
                E := NumericToExtended( PSqlNumeric( FD.BindPtr )^ );
                Result :=
                     FloatToStr( E );
              end;
                Numeric   SQL_DOUBLE !
            *)
            SQL_DECIMAL,
                SQL_CHAR,
                SQL_VARCHAR:
                SetString(Result, PChar(FD.BindPtr), FD.StrLenInd^);
            SQL_VARBINARY, SQL_BINARY, -1:
                if FD.ColumnSize > 0 then
                    SetString(Result, PChar(FD.BindPtr), FD.StrLenInd^)
                else
                begin
                    targetType := SQL_BINARY;
                    if SmallInt(FD.DataType) = -1 then
                        targetType := SQL_CHAR;

                    //R := SQLGetData( fHandle, N+1, SQL_CHAR, @ShortBuf[0], Sizeof(ShortBuf), FD.StrLenInd );
                    R := SQLGetData(fHandle, N + 1, targetType, @ShortBuf[0], 0, FD.StrLenInd);
                    L := FD.StrLenInd^;
                    (*R := SQLGetData( fHandle, N, SQL_BINARY, @ShortBuf[0], 256, @L );*)
                    if (R > 1) and (R <> SQL_NO_DATA) then
                    begin
                        raise EODBC.CreateStmtError(fHandle, R, '    ' +
                            fColumns{$IFDEF KOL}.Items{$ENDIF} [N]);
                    end;
                    SetLength(Result, L);
                    //Move( ShortBuf, Result[1], min( 256, L ) );
                    if L > 0 then
                    begin

                        R := SQLGetData(fHandle, N + 1, targetType, @Result[1], L, @L);
                        if (R > 1) and (R <> SQL_NO_DATA) then
                        begin
                            //SQLGetInfo( fHandle, 81{SQL_GETDATA_EXTENSIONS},@Ext,4, nil );
                            raise EODBC.CreateStmtError(fHandle, R, '    ' +
                                fColumns{$IFDEF KOL}.Items{$ENDIF} [N]);
                        end;
                    end;
                end;
            SQL_WVARCHAR:
                begin
                    Result := PWideChar(FD.BindPtr)^;
                end;
            SQL_INTEGER:
                Result := IntToStr(PInteger(FD.BindPtr)^);
            SQL_SMALLINT:
                Result := IntToStr(PSmallInt(FD.BindPtr)^);
            SQL_DOUBLE,
                SQL_FLOAT:
                Result :=
                    {$IFDEF KOL}
                Double2Str(PDouble(FD.BindPtr)^);
            {$ELSE}
                FloatToStr(PDouble(FD.BindPtr)^);
            {$ENDIF}
            SQL_REAL:
                Result :=
                    {$IFDEF KOL}
                Double2Str(PSingle(FD.BindPtr)^);
            {$ELSE}
                FloatToStr(PSingle(FD.BindPtr)^);
            {$ENDIF}
            SQL_TYPE_TIMESTAMP:
                begin
                    {$IFDEF KOL}
                    Result := DateTime2StrShort(SqlDateTimeStampToDateTime(FD.BindPtr));
                    {$ELSE}
                    Result := FormatDateTime('dd.mm.yyyy',
                        SqlDateTimeStampToDateTime(FD.BindPtr));
                    {$ENDIF}
                end;
        else Result := '';
        END;
end;

procedure TODBCQuery.InternalOpenResultset;
    function FieldSize(FD: PFieldDescription): Integer;
    begin
        CASE FD.DataType OF
            SQL_NUMERIC: Result := 8; // 19;
            SQL_DECIMAL: Result := 8;
            //SQL_CHAR,
            //SQL_VARCHAR:
            //SQL_WVARCHAR:
            SQL_INTEGER: Result := 4;
            SQL_SMALLINT: Result := 2;
            SQL_DOUBLE: Result := 8;
            SQL_FLOAT: Result := 8;
            SQL_REAL: Result := 10;
            SQL_DATE, SQL_TIME,
                SQL_TYPE_TIMESTAMP: Result := 16;
        else Result := FD.ColumnSize + 1;
        END;
        //Result := 256;
    end;
var i, L, L1, N: Integer;
    FD: PFieldDescription;
    ColName: array[0..4095] of Char;
    ColLen: SmallInt;
    s: String;
    R: Word;
    P: PByte;
    t: SmallInt;
begin
    L := 0;
    N := fColCount - 1;
    for i := 0 to fColCount - 1 do
    begin
        FD := AllocMem(Sizeof(FD^));
        SQLDescribeCol(fHandle, i + 1, ColName, Sizeof(ColName), @ColLen,
            @FD.DataType, @FD.ColumnSize, @FD.DecimalDigits, @FD.Nullable);
        L1 := FieldSize(FD);
        if (L1 < 0) or (L1 >= $40000000) then
        begin
            L1 := 0; //1024;
            if N = fColCount - 1 then
                N := i - 1;
            FD.ColumnSize := -1;
        end
        else
        begin
            {if  i > N then
                raise Exception.Create( '      ' +
                '      ' );}
        end;
        SetString(s, ColName, ColLen);
        Inc(L, L1 + Sizeof(Integer));
        fColumns.AddObject(s, {$IFDEF KOL}DWORD(FD){$ELSE}Pointer(FD){$ENDIF});
    end;
    //SQLSetStmtAttr( fHandle, SQL_ATTR_ROW_BIND_TYPE, Pointer( L ), 0 );
    fBindBuffer := AllocMem(L);
    P := fBindBuffer;
    for i := 0 to fColCount - 1 do
    begin
        FD := Pointer(fColumns.Objects[i]);
        FD.BindPtr := P;
        t := FD.DataType;
        L := FieldSize(FD);
        if L < 0 then
            L := 0; //1024;
        CASE t OF
            SQL_VARCHAR: t := SQL_CHAR;
            SQL_NUMERIC: t := SQL_DOUBLE;
            SQL_FLOAT: t := SQL_DOUBLE;
            SQL_DECIMAL: t := SQL_DOUBLE;
            SQL_VARBINARY: t := SQL_BINARY; //???
            SQL_DATE, SQL_TIME,
                SQL_TYPE_TIMESTAMP: t := SQL_TYPE_TIMESTAMP;
            SQL_INTEGER, SQL_SMALLINT, SQL_CHAR: ;
            -1: t := -1;
        else t := SQL_CHAR;
        END;
        FD.DataType := t;
        Inc(P, L);
        FD.StrLenInd := Pointer(P);
        Inc(P, Sizeof(Integer));
        if L > 0 then
        begin
            R := SQLBindCol(fHandle, i + 1, t, FD.BindPtr, L, FD.StrLenInd);
            if R > 1 then
                raise EODBC.CreateStmtError(fHandle, R, '    ' +
                    fColumns{$IFDEF KOL}.Items{$ENDIF} [i]);
        end
        else
        begin
            FD.BindPtr := nil;
        end;
    end;
    Next;
    //First;
end;

procedure TODBCQuery.Last;
var R: Word;
begin
    fEof := FALSE;
    fBof := FALSE;
    R := SQLFetchScroll(fHandle, SQL_FETCH_LAST, 0);
    if R > 1 then
    begin
        fEof := TRUE;
        fBof := TRUE;
    end;
end;

function TODBCQuery.More: Boolean;
var R: Word;
begin
    Result := FALSE;
    ClearColumns;
    if fHandle = 0 then Exit;
    REPEAT
        R := SQLMoreResults(fHandle);
        if R > 1 then
        begin
            if R = 100 then Exit;
            raise EODBC.CreateStmtError(fHandle, R, '  More');
        end;
        if R = 0 then
            SQLNumResultCols(fHandle, @fColCount);
    UNTIL fColCount > 0;
    Result := fColCount > 0;
    if Result then
        InternalOpenResultset;
end;

procedure TODBCQuery.Next;
var R: Word;
begin
    fEof := FALSE;
    R := //SQLFetchScroll( fHandle, SQL_FETCH_NEXT, 0 );
    SQLFetch(fHandle);
    if R > 1 then
        fEof := TRUE
    else
        fBof := FALSE;
end;

procedure TODBCQuery.Open;
begin
    //if Assigned( BeforeOpen ) then
    //  BeforeOpen( nil );
    ExecSQL;
    SQLNumResultCols(fHandle, @fColCount);
    if fColCount = 0 then
        More
    else
        InternalOpenResultset;
end;

procedure TODBCQuery.Prev;
var R: Word;
begin
    fBof := FALSE;
    R := SQLFetchScroll(fHandle, SQL_FETCH_PRIOR, 0);
    if R > 1 then
        fBof := TRUE
    else
        fEof := FALSE;
end;

procedure TODBCQuery.SetBParam(ParamNum: Integer; const Value: String);
var R: Word;
begin
    if Parameters = nil then
        Parameters := TStringList.Create;
    Parameters.AddObject(Value, Pointer(ParamNum));
    R := SQLBindParameter(fHandle, ParamNum, SQL_PARAM_INPUT, SQL_BINARY, //99{SQL_C_DEFAULT},
        SQL_VARBINARY,
        Length(Value),
        0 {Length(Value)},
        Pointer(ParamNum),
        0 {Length(Value)},
        @SqlLenDataAtExec);
    if R > 1 then
        raise EODBC.CreateStmtError(fHandle, SmallInt(R), '   ' +
            IntToStr(ParamNum));
end;

procedure TODBCQuery.Prepare;
var s: String;
    R: Word;
begin
    if fHandle = 0 then
    begin
        if fHandle <> 0 then
        begin
            SQLFreeHandle(SQL_HANDLE_STMT, fHandle);
            fHandle := 0;
        end;
        R := SQLAllocHandle(SQL_HANDLE_STMT, ODBC.fConnHandle, @fHandle);
        if R > 1 then
            raise EODBC.CreateConnError(ODBC.fConnHandle, R,
                '   :' + Copy(SQL.Text, 1, 100));
    end;
    s := SQL.Text;
    R := SQLPrepare(fHandle, PChar(s), SQL_NTS);
    if R > 1 then
        raise EODBC.CreateStmtError(fHandle, SmallInt(R), '  SQLPrepare');
end;

{ TSQLStringList }

{$IFDEF KOL}
{$ELSE}

procedure TSQLStringList.Changed;
begin
    inherited;
    if Assigned(FOwnerODBCQuery) then
        FOwnerODBCQuery.Close;
end;
{$ENDIF}

end.
