//------------------------------------------------------------------------------
//  
//  :
// -  TParamData -   
// -  TParams -       
//                 (   TVars -    
//                 )
//
//  .. 03.08.2004 - 
//                01.2005 -       
//------------------------------------------------------------------------------
unit ParamUnit;

interface

{$I VCL_VERSIONS.INC}

uses Windows, Classes, Dialogs, SysUtils, FormulaInterpreterUnit,
    AdoSqlUnit, ADOint,
    QSort, math, FastTable, ODBCQuery, UtilsUnit
    {$IFNDEF ODBC_DIRECT}, Adodb{$ENDIF}, Forms
    {$IFDEF _D6orHigher}, Variants{$ENDIF};

//{$DEFINE DEBUG_SUM} //   

const
    MAX_PRECISION = 1000000;
    _NULL = ''; // 'NULL' -  6.5  !
    Value_Correction = 10000000;

var PackageTableName: String = 'Package';
    Total_params_loaded: Integer;

type
    Param_Exception = class(Exception)
    end;

    TList1 = class(TList)
    protected
        procedure Grow; override;
    end;

    TParams = class;

    TParamDef = class;

    TParamsSet = class(TComponent)
    private
        function GetConstParam(Anomenklid: Integer): TParams;
        function GetConstParamsCount: Integer;
        function GetConstParamsByIdx(idx: Integer): TParams;
        procedure SetConstParamsByIdx(idx: Integer; const Value: TParams);
    protected
        function SumCached(P: TParams; Idx, fPos: Integer; var F: String;
            calc_by_batch: Boolean): Boolean;
            virtual;
        procedure DoSumCache(P: TParams; Idx, fPos: Integer; const F: String;
            Value: Double; calc_by_batch: Boolean);
            virtual;
    protected
        ParamsDefs: TList;
        ConstParams: TList;
        function Def(paramid: Integer; const param_name: String;
            nomenklid: Integer; const op: String; AddIfNotExists: Boolean = TRUE): TParamDef;
    public
        constructor Create(AOwner: TComponent); override;
        destructor Destroy; override;
        procedure Clear(ClearAll: Boolean); virtual;
        property ConstParam[Anomenklid: Integer]: TParams read GetConstParam;
        property ConstParamsByIdx[idx: Integer]: TParams read GetConstParamsByIdx write SetConstParamsByIdx;
        property ConstParamsCount: Integer read GetConstParamsCount;
    end;

    //----------------------------------------------------------------------------
    //     (,  ,  ,  )
    TParamSetData = class
        //--------------------------------------------------------------------------
        // ID   nomenklinfo
        nomenklid: Integer;
        //--------------------------------------------------------------------------
        // ID 
        packid: Integer;
        //--------------------------------------------------------------------------
        // ID   
        grpid: Integer;
        //--------------------------------------------------------------------------
        //  
        numcon: String;
        //--------------------------------------------------------------------------
        // ID 
        batchid: Integer;
        Sign_final: Boolean;
        //--------------------------------------------------------------------------
        // ID 
        opid: Integer;
        //--------------------------------------------------------------------------
        //         
        ignore_conditions: Boolean;
        //--------------------------------------------------------------------------
        //  
        code_operation: String;
        //--------------------------------------------------------------------------
        //  
        code_trans: Integer;
        //--------------------------------------------------------------------------
        //  "   " ( ,  
        //    -      
        // ,  -        
        //  ,       )
        NoCalc: Boolean;
        //--------------------------------------------------------------------------
        //  "    " -  ,   
        //        ( 
        //   . ).
        NoSave: Boolean;
        procedure Assign(From: TParamSetData); virtual;
    end;

    //----------------------------------------------------------------------------
    //   
    TParamDef = class
    protected
        loaded: Boolean;
    public
        FDepending: TList;
        constructor Create;
        destructor Destroy; override;
    public
        Param_name: String;
        paramid: Integer;
        nomenklid: Integer;
        //--------------------------------------------------------------------------
        //   ,    ,  
        //        :
        Analyse_copy_from_param: String;
        Is_Batch_Analyse: Boolean; // TRUE =  
        //--------------------------------------------------------------------------
        // . .
        Units: String;
        //--------------------------------------------------------------------------
        //  : 1 - , 2 - , 3 -  , 4 - 
        param_type: Integer;
        //--------------------------------------------------------------------------
        //   
        param_format: Double;
        //--------------------------------------------------------------------------
        //   (  )  -
        precision: Integer;
        //--------------------------------------------------------------------------
        //     2
        formula: String;
        //--------------------------------------------------------------------------
        //      (  
        //  )
        read_only: Boolean;
        //--------------------------------------------------------------------------
        //      
        //show_in_op: TStringList;
        show_in_ops: array[0..29] of Byte; // todo: 19->99
        show_in_op_count: Byte;
        //--------------------------------------------------------------------------
        //     
        param_orders: array[0..29] of Integer; // todo: 19->99
        function param_ord_count: Byte;
    public
        //--------------------------------------------------------------------------
        //       (   
        //  TParamController)
        //used_in_op: TStringList;
        used_in_ops: array[0..29] of Byte; // todo: 19->99
        used_in_op_count: Byte;
        //--------------------------------------------------------------------------
        //    
        calc_by_batch: Boolean;
        //--------------------------------------------------------------------------
        //   ,   
        recalc: Boolean;
        procedure Assign(From: TParamDef); virtual;
        function formula_empty: Boolean;
        procedure Set_show_in_op(const op: String);
        procedure Set_used_in_op(const op: String; op_param_order: Integer);
        function Is_show_in_op(const op: String): Integer;
        function Is_used_in_op(const op: String): Integer;
    end;

    //----------------------------------------------------------------------------
    //   
    TParamData = class
    private
        function Get_calc_by_batch: Boolean;
        function Get_recalc: Boolean;
        procedure Set_calc_by_batch(const Value: Boolean);
        procedure Set_recalc(const Value: Boolean);
        function get_used_in_ops(idx: Integer): Byte;
        function get_param_orders(idx: Integer): Integer;
        function get_show_in_ops(idx: Integer): Byte;
        function get_read_only: Boolean;
        procedure set_read_only(const Value: Boolean);
        function get_formula: String;
        procedure set_formula(const Value: String);
        function get_precision: Integer;
        procedure set_precision(const Value: Integer);
        function get_param_format: Double;
        procedure set_param_format(const Value: Double);
        function get_param_type: Integer;
        procedure set_param_type(const Value: Integer);
        function get_Units: String;
        procedure set_Units(const Value: String);
        function get_nomenklid: Integer;
    protected
        //--------------------------------------------------------------------------
        //        paramid,  nomenklid
        // ( nomenklid <> 0),     
        param_def: TParamDef;
        //--------------------------------------------------------------------------
        //     ( :   
        //  ,    )
        param_started_calc: Boolean;
        //--------------------------------------------------------------------------
        //
        param_error: Boolean;
    public
        //--------------------------------------------------------------------------
        //  ,   
        param_ready: Boolean;
        //FDepending: TList;
        constructor Create;
        destructor Destroy; override;
    public
        //--------------------------------------------------------------------------
        // . .
        property Units: String read get_Units write set_Units;
        //--------------------------------------------------------------------------
        //  : 1 - , 2 - , 3 -  , 4 - 
        property param_type: Integer read get_param_type write set_param_type;
        //--------------------------------------------------------------------------
        //   
        property param_format: Double read get_param_format write set_param_format;
        //--------------------------------------------------------------------------
        //   (  )  -
        property precision: Integer read get_precision write set_precision;
        //--------------------------------------------------------------------------
        //     2
        property formula: String read get_formula write set_formula;

        property nomenklid: Integer read get_nomenklid;
    public
        //--------------------------------------------------------------------------
        //  -      
        save_formula: String;
        //--------------------------------------------------------------------------
        //      (  
        //  )
        property read_only: Boolean read get_read_only write set_read_only;
        //--------------------------------------------------------------------------
        //      
        //show_in_op: TStringList;
        property show_in_ops[idx: Integer]: Byte read get_show_in_ops;
        function show_in_op_count: Byte;
        //--------------------------------------------------------------------------
        //     
        property param_orders[idx: Integer]: Integer read get_param_orders;
        function param_ord_count: Byte;
        //--------------------------------------------------------------------------
        //       (   
        //  TParamController)
        //used_in_op: TStringList;
        property used_in_ops[idx: Integer]: Byte read get_used_in_ops;
        function used_in_op_count: Byte;
        //--------------------------------------------------------------------------
        //    
        property calc_by_batch: Boolean read Get_calc_by_batch write Set_calc_by_batch;
        //--------------------------------------------------------------------------
        //   ,   
        property recalc: Boolean read Get_recalc write Set_recalc;
        procedure Assign(From: TParamData); virtual;
        function formula_empty: Boolean;
        procedure Set_show_in_op(const op: String);
        procedure Set_used_in_op(const op: String; op_param_order: Integer);
        function Is_show_in_op(const op: String): Integer;
        function Is_used_in_op(const op: String): Integer;
        function Param_name: String;
        property Def: TParamDef read param_def;
    end;

    //----------------------------------------------------------------------------
    //  .    TVars -  
    TParams = class(TVars)
    private
        FParamsController: TParamsSet;
        FParamSetData: TParamSetData;
        function GetParamData(Idx: Variant): TParamData;
        function GetCapacity: Integer;
        procedure SetCapacity(const Value: Integer);
        function GetNomenklid: Integer;
    protected
        FParamData: TList;
        procedure WhenSummed(ToWhat: TVars; const AddText: String); override;
        function SumCached(Idx, fPos: Integer; var F: String; calc_by_batch: Boolean): Boolean;
        procedure DoCacheSum(Idx, fPos: Integer; const F: String; Value: Double;
            calc_by_batch: Boolean);
        procedure BuildDependings;
        function GetNames(Idx: Integer): String; override;
        function GetValues(IdxOrName: Variant): Extended; override;
        function GetID(IdxOrName: Variant): Integer; override;
        function GetDescr(Idx: Integer): String; override;
        function HasVar(const Name: String): Integer;
    public
        property nomenklid: Integer read GetNomenklid;
        constructor Create; override;
        constructor CreateOwned(Owner: TParamsSet);
        destructor Destroy; override;
        procedure Clear; override;
        function Add(const Name: String; nomenklid: Integer; Value: Extended; ID: Integer;
            const Description: String; const op: String; const A_copy_from: String;
            Is_Batch_Analyses: Boolean)
            : Pointer; reintroduce; //override;
        function AddConst(const Name: String; nomenklid: Integer; Value: Extended; ID: Integer;
            const Description: String; const op: String): Pointer;
        function findParamByParamNameAndNomenklid(const AParam_name: String;
            Anomenklid: Integer): Integer;
        procedure Assign(From: TVars); override;
        //--------------------------------------------------------------------------
        //  
        property ParamData[Idx: Variant]: TParamData read GetParamData;
        property ParamSetData: TParamSetData read FParamSetData;
        procedure SwapItems(i1, i2: Integer); override;
        function FormattedParameter(Idx: Variant): String;
        function OK: Boolean; //         
    public
        Group: Integer; //    (   )
        GrpId: Integer; // , ID     ( GrpVal)
        function calc_by_batch(Idx: Integer): Boolean; //override;
        property Capacity: Integer read GetCapacity write SetCapacity;
        function DefsTheSame(P: TParams): Boolean;
        function ConstCount: Integer;
        procedure DebugOutput(const filename: String);
    end;

    //------------------------------------------------------------------------------
    //  .
    // :
    //   ParamsList -    TParams, 
    //              sum( )   
    //            ,     
    //            . :  
    //              ,     ,
    //               .
    //   Error - . ,    ,    
    //            .
type TOnCalcParam = procedure(const s: String) of object;
var TimePrepare, TimeCalc: DWORD;
    OnCalcParam: TOnCalcParam;
    CancelCalcParams: Boolean;
function CalcParams(ParamsList: TList; var Error: String;
    RaiseError: Boolean = FALSE; AllowSum: TAllowSum = nil;
    InitInputVals: Boolean = FALSE; const opname: String = '';
    nomenklid: Integer = 0; Ignore_packid0: Boolean = FALSE): Extended;

//------------------------------------------------------------------------------
//    
function LoadParams_by_ConList(const ListCon, Code_Operation: String;
    AllowSum: TAllowSum = nil; AllowRaise: Boolean = FALSE;
    ParamsController: Pointer = nil; AdditionalConds: String = '';
    LoadParentParams: Boolean = FALSE; only_nomenklid: Integer = -1): TList;

//------------------------------------------------------------------------------
//     +  
function LoadParams_by_ConList_TranspackId(
    const ListConditions, Code_Operation: String;
    AllowSum: TAllowSum = nil; AllowRaise: Boolean = FALSE;
    DateFrom: TDateTime = 0; DateTo: TDateTime = 0;
    ParamsController: Pointer = nil;
    LoadParentParams: Boolean = FALSE): TList;

//------------------------------------------------------------------------------
//    
function LoadParams_by_NomList(const NomList, Code_Operation: String;
    AllowSum: TAllowSum = nil; ParamsController: Pointer = nil;
    only_nomenklid: Integer = -1): TList;

//------------------------------------------------------------------------------
//    
function LoadParams_by_BatList(const BatchList, Code_Operation: String;
    AllowSum: TAllowSum = nil; ParamsController: Pointer = nil): TList;

function ParamDependOnParam(P: TParams; I, J: Integer): Boolean;

type TOnParamLoad = procedure(const S: String) of object;
var OnParamLoad: TOnParamLoad;
    DoNotCalcParamsWhileLoad: Boolean = FALSE;
var
    Operations: TStringList; //  show_in_op, used_in_op,  
    // show_in_op_flags  used_in_op_flags,   
    //   ,      

function GetLogsPath: String;

implementation

var LogsPath: String;

function GetLogsPath: String;
    procedure CheckPath(const s: String);
    var SL: TStringList;
    begin
        SL := TStringList.Create;
        TRY
            SL.Text := 'test';
            SL.SaveToFile(s + 'test_create.txt');
            if FileExists(s + 'test_create.txt') then
            begin
                LogsPath := s;
                DeleteFile(PChar(s + 'test_create.txt'));
            end;
        FINALLY
            SL.Free;
        END;
    end;
begin
    if LogsPath = '' then
        CheckPath('c:\');
    if LogsPath = '' then
        CheckPath(ExtractFilePath(Application.ExeName));
    Result := LogsPath;
end;


function Drop(const TableName: String): String;
begin
    Result := '';
    {#13'if exists(select name from sysobjects where name=''' +
           TableName + ''') begin ' +
    #13'   drop table ' + TableName + '' +
    #13'end'#13;}
end;

function EndDrop(const TableName: String): String;
begin
    Result := #13'drop table ' + TableName;
    {#13'if exists(select name from sysobjects where name=''' +
           TableName + ''') begin ' +
    #13'   drop table ' + TableName + '' +
    #13'end'#13;}
end;

function RemoveSpaces(const S: String): String;
var I: Integer;
begin
    Result := S;
    for I := Length(S) downto 1 do
        if S[I] <= ' ' then
            Delete(Result, I, 1);
end;

function IfThenElse(cond: Boolean; const Val1, Val2: String): String;
begin
    if cond then Result := Val1
    else Result := Val2;
end;

function Year(DT: TDateTime): Integer;
var Y, M, D: Word;
begin
    DecodeDate(DT, Y, M, D);
    Result := Y;
end;

function Month(DT: TDateTime): Integer;
var Y, M, D: Word;
begin
    DecodeDate(DT, Y, M, D);
    Result := M;
end;

function Day(DT: TDateTime): Integer;
var Y, M, D: Word;
begin
    DecodeDate(DT, Y, M, D);
    Result := D;
end;

function DateToSQL(D: TDateTime): String;
begin
    Result := '''' +
        Format('%.04d%.02d%.02d', [Year(D), Month(D), Day(D)]) + '''';
end;


function CompareItems(const Data: Pointer; const e1, e2: Dword): Integer;
var L: TList;
    I1, I2: Integer;
begin
    L := Data;
    I1 := Integer(L[e1]);
    I2 := Integer(L[e2]);
    Result := I1 - I2;
end;

procedure SwapItems(const Data: Pointer; const e1, e2: Dword);
var L: TList;
    Temp: Pointer;
begin
    L := Data;
    Temp := L[e1];
    L[e1] := L[e2];
    L[e2] := Temp;
end;

function Compare_grpid(const Data: Pointer; const e1, e2: Dword): Integer;
var List: TList;
    P1, P2: TParams;
begin
    List := Data;
    P1 := List[e1];
    P2 := List[e2];
    Result := P1.ParamSetData.grpid - P2.ParamSetData.grpid;
end;

function Compare_packid(const Data: Pointer; const e1, e2: Dword): Integer;
var List: TList;
    P1, P2: TParams;
begin
    List := Data;
    P1 := List[e1];
    P2 := List[e2];
    Result := P1.ParamSetData.packid - P2.ParamSetData.packid;
end;

{$IFDEF ODBC_DIRECT}
{$DEFINE ALL_IN_ONE}
{$ENDIF}

function LoadParams_by_ConList1(ListCon_has_Conditions: Boolean;
    ListConditions: String; const Code_Operation: String;
    AllowSum: TAllowSum = nil; AllowRaise: Boolean = FALSE;
    DateFrom: TDateTime = 0; DateTo: TDateTime = 0;
    ParamsController: Pointer = nil; AdditionalConds: String = '';
    LoadParentParams: Boolean = FALSE; only_nomenklid: Integer = -1): TList;
//------------------------------------------------------------------------------
//    
var{$IFDEF ALL_IN_ONE}
    Qall: TQuery;
    {$ENDIF}
    Q1, Q1A, Q2, Q3, Q3A, Q3B, Q4, Q4B, Q4C, Q5: TQuery;
    QFin, Q0: TQuery;
    ListCondShort: String;

    function q(Q_param: TQuery): TQuery;
    begin
        {$IFDEF ALL_IN_ONE}
        Result := Qall;
        {$ELSE}
        Result := Q_param;
        {$ENDIF}
    end;

var P, P1, P2: TParams;
    I, J, K: Integer;
    ListCon1: String;
    //\\ListParamId: TList;
    //\\ListParamIdAsStr: String;
    ErrorStr: String;
    Ops: TStringList;
    S: String;
    Param_found: Boolean;
    ConList: TList;
    g1, g2: Integer;

type
    TParamLoad = packed record
        // Param
        paramid: Integer;
        param_descr: String;
        param_name: String;
        nomenkild: Integer; // +++
        param_type: Integer;
        formula: String;
        param_val: Double;
        Ed_ism: String;
        Format_representation: Double;
        Precis: Integer;
        //---- Operation_params
        param_order: Integer;
        hidden: Integer;
        Operation_param_id: Integer;
        calc_by_batch: Boolean;
        recalc: Boolean;
        //----- Operation
        opid: Integer;
        Code_operation: String;
        Sign_clear: Integer;
        Analyse_copy_from: String;
        Is_Analyse_of_Batch: Integer;
    end;
    PParamLoad = ^TParamLoad;

    TNomenkLoad = packed record
        nomenklid: Integer;
        nominal, elempct, wrand, wsys, urand, usys, srand, ssys,
            percentrich, Min_percent, Max_percent,
            Accuracy_element, Accuracy_isotope: Double;
    end;
    PNomenkLoad = ^TNomenkLoad;
var
    ParamsLoaded: TList; // of TParamLoad
    PL: PParamLoad;

    procedure AddFixedParam(P: TParams; Prefix: Char; Name: String; Descr: String);
    var K, j: Integer;
        ConstParamsSet: TParams;
    begin
        if P.FParamsController.ConstParams = nil then
            P.FParamsController.ConstParams := TList.Create;
        for j := 0 to P.FParamsController.ConstParams.Count - 1 do
        begin
            ConstParamsSet := P.FParamsController.ConstParams[j];
            if ConstParamsSet.nomenklid = P.nomenklid then
            begin
                if ConstParamsSet.IndexOf(Prefix + '.' + Name) < 0 then
                begin
                    ConstParamsSet.Add(Prefix + '.' + Name, 0, q(Q5).FField[Name], 0,
                        Descr, '', '', FALSE);
                    K := ConstParamsSet.Count - 1;
                    ConstParamsSet.ParamData[K].param_type := 1;
                    ConstParamsSet.ParamData[K].read_only := TRUE;
                end;
                //break;
                Exit;
            end;
        end;
        P.AddConst(Prefix + '.' + Name, P.nomenklid, q(Q5).FField[Name], 0, Descr, '');
        {P.Add( Prefix + '.' + Name, Q5.FField[ Name ], 0, Descr, '' );
        K := P.Count-1;
        P.ParamData[ K ].param_type := 1;
        P.ParamData[ K ].read_only := TRUE;}
    end;

    procedure OptimizeListCon;
    var i, j: Integer;
        ListIn: String;
    begin
        ConList := TList.Create;
        i := 1;
        while i <= Length(ListConditions) do
        begin
            j := i + 1;
            while (j <= Length(ListConditions)) and (ListConditions[j] <> ',') do
                Inc(j);
            if j > i then
            begin
                ConList.Add(Pointer(StrToInt(Copy(ListConditions, i, j - i))));
            end;
            i := j + 1;
        end;
        SortData(ConList, ConList.Count, CompareItems, SwapItems);
        ListConditions := '';
        if (ConList.Count > 0)
            and (ConList.Count <= 100)
            and (Integer(ConList[ConList.Count - 1]) - Integer(ConList[0]) <= 1000) then
            ListCondShort := '(pk.packid >= ' + IntToStr(Integer(ConList[0])) +
                ' and pk.packid <= ' + IntToStr(Integer(ConList[ConList.Count - 1])) +
                ')'
        else
            ListCondShort := '';
        ListIn := '';
        i := 0;
        while i <= ConList.Count - 1 do
        begin
            for j := i + 1 to ConList.Count do
            begin
                if (j >= ConList.Count) or
                    (Integer(ConList[j]) <> Integer(ConList[j - 1]) + 1) then
                begin
                    if j > i + 1 then
                    begin
                        if ListConditions <> '' then
                            ListConditions := ListConditions + ' or ';
                        ListConditions := ListConditions + ' pk.packid >= ' +
                            IntToStr(Integer(ConList[i])) + ' and ' +
                            'pk.packid <= ' + IntToStr(Integer(ConList[j - 1]));
                    end
                    else
                    begin
                        for i := i to j - 1 do
                            ListIn := ListIn + ',' + IntToStr(Integer(ConList[i]));
                    end;
                    i := j;
                end;
            end;
        end;
        if ListIn <> '' then
        begin
            Delete(ListIn, 1, 1);
            ListIn := 'pk.packid in (' + ListIn + ')';
        end;
        if ListConditions <> '' then
        begin
            if ListIn <> '' then
                ListConditions := ListConditions + ' or ' + ListIn;
            ListConditions := '(' + ListConditions + ')';
        end
        else
            ListConditions := ListIn;
    end;

type
    TParamInfo = packed record
        param_name, param_descr, formula, Ed_ism, Code_operation: String;
        nomenklid, paramid, param_type, precis, calc_by_batch, recalc,
            param_order, hidden: Integer;
        format_representation, param_val: Extended;
        A_Copy_f: String; Is_A: Integer;
        //really_need: Boolean;
    end;

var NoNomenklId: Integer;
    q4_nomenklid, q4_grpid, q4_paramid, q4_packid: Integer;
    q4_param_calc: Double;
    Cache: TFastTable;
    NomenLoad: TList;
    ParamLoad: array of TParamInfo;
    Lst: TList;
    Ptr: Pointer;
    found: Boolean;
    ConstParamsSet: TParams;
    PD: TParamData;
    total_val_grps: Integer;
    I1, I2: Integer;
    opids: String;

    Hints: array of Integer;

    procedure AssignParamsToConteiner;
    var k: Integer;
    begin
        inc(total_val_grps);
        if total_val_grps mod 1000 = 0 then
            if Assigned(OnParamLoad) then
                OnParamLoad('13 -     (' +
                    IntToStr(total_val_grps) + ')');
        Param_Found := FALSE;
        J := 0;
        while J < P.Count do
        begin
            if P.ID[J] = q4_paramid then
            begin
                Param_Found := TRUE;
                P.Values[J] := q4_param_calc / Value_Correction;
                break;
            end;
            Inc(J);
        end;
        if not Param_Found then
        begin
            for k := 0 to ParamsLoaded.Count - 1 do
            begin
                PL := ParamsLoaded[k];
                if PL.paramid = q4_paramid then
                begin
                    {if (PL.param_type = 3) and (PL.formula = '') then
                    P.AddConst( PL.param_name, PL.nomenkild, q4_param_calc / Value_Correction,
                           PL.paramid, PL.param_descr, '' )
                    else}
                    P.Add(PL.param_name, PL.nomenkild, q4_param_calc / Value_Correction,
                        PL.paramid, PL.param_descr, '', PL.Analyse_copy_from,
                        PL.Is_Analyse_of_Batch = 1);
                    J := P.Count - 1;
                    Param_Found := TRUE;
                    BREAK;
                end;
            end;
        end;
        if not Param_Found then
        begin
            {asm
              int 3
            end;}
            Exit;
        end;

        k := -1;
        if Length(Hints) >= P.Count then
        begin
            k := Hints[J];
            PL := ParamsLoaded[k];
            if PL.paramid <> q4_paramid then
                k := -1;
        end;

        if k < 0 then
        begin
            k := 0;
            while k < ParamsLoaded.Count do
            begin
                PL := ParamsLoaded[k];
                if PL.paramid = q4_paramid then
                    break;
                inc(k);
            end;
        end;
        if (k >= 0) and (k < ParamsLoaded.Count) then
        begin
            if Length(Hints) < P.Count then
                SetLength(Hints, P.Count);
            Hints[J] := K;

            P.ParamData[J].param_type := 3;
            P.ParamData[J].Units := PL.Ed_ism;

            if PL.Format_representation <> 0 then
                P.ParamData[J].param_format := PL.Format_representation;


            P.ParamData[J].param_ready := PL.paramid <> 0;

            if (PL.hidden = 0) and
                (PL.Operation_param_id > 0) then
                P.ParamData[J].Set_show_in_op(PL.Code_operation);
            P.ParamData[J].Set_used_in_op(PL.Code_operation,
                PL.param_order);
            P.ParamData[J].calc_by_batch := PL.calc_by_batch;
            P.ParamData[J].recalc := PL.recalc;
        end;
    end;
var SkipMore: Boolean;
    drop_E, drop_D, drop_Pks: Boolean;
    {$IFDEF ALL_IN_ONE}
    procedure More;
    begin
        if not SkipMore then
            Qall.More;
        SkipMore := FALSE;
    end;
    {$ENDIF}

    function Fix_Condition_Sign_PresenceIsNull(const s: String): String;
    var i: Integer;
        prev: Integer;
        patt: String;
    begin
        Result := s;
        for i := Length(s) - 1 downto 1 do
            if Copy(Result, i, 2) = '  ' then
                Delete(Result, i, 1);
        prev := -1;
        patt := 'and pk.Sign_presence is null';
        for i := Length(Result) downto 1 do
        begin
            if Copy(Result, i, Length(patt)) = patt then
            begin
                if prev > 0 then
                    Result := Copy(Result, 1, prev - 1) +
                        Copy(Result, prev + Length(patt), MaxInt);
                prev := i;
            end;
        end;
    end;
begin
    SkipMore := FALSE;
    {$IFDEF DEBUG_EXCEPTION}
    Lst := TList.Create;
    TRY
        while TRUE do
        begin
            GetMem(Ptr, 10000000);
            Lst.Add(Ptr);
        end;
    FINALLY
        for I := 0 to Lst.Count - 1 do
            FreeMem(Lst[I]);
        Lst.Free;
    END;
    {$ENDIF}
    //----------------------------------------------------------------------------
    //   
    ConList := nil;
    ListCondShort := '';
    if not ListCon_has_Conditions then
        OptimizeListCon;
    ListConditions := ListConditions +
        IfThenElse((ListConditions <> '') and (AdditionalConds <> ''), ' and ', '') +
        AdditionalConds;
    if ListCondShort = '' then
        ListCondShort := ListConditions;

    //----------------------------------------------------------------------------
    //    - (   )
    //                                            ?
    Ops := TStringList.Create;
    Cache := TFastTable.Create;
    Cache.AutoAllocateColumns := TRUE;
    NomenLoad := TList.Create;
    {$IFDEF USE_NEWQUERY} // ODBC_DIRECT}
    DefaultODBCDataSource.BeginTrans;
    {$ELSE}
    DefaultConnection.BeginTrans;
    {$ENDIF}
    TRY

        S := Code_Operation;
        //for I := 1 to High( Ops ) do Ops[ I ] := '';
        //I := 1;
        while {(I <= High( Ops )) and}(S <> '') do
        begin
            J := pos(',', S);
            if J <= 0 then J := pos(';', S);
            if J <= 0 then
            begin
                //Ops[ I ] := S;
                Ops.Add(S);
                S := '';
            end
            else
            begin
                //Ops[ I ] := Trim( copy( S, 1, J - 1 ) );
                Ops.Add(copy(S, 1, J - 1));
                S := Copy(S, J + 1, MaxInt);
            end;
            //Inc( I );
        end;

        Result := TList.Create;
        {$IFDEF ALL_IN_ONE}
        Qall := TQuery.Create(nil);
        {$ENDIF}
        Q1 := TQuery.Create(nil); // Q1.CursorType := ctOpenForwardOnly; Q1.CursorLocation := clUseServer; Q1.LockType := ltReadOnly; Q1.ExecuteOptions := [ eoAsyncFetch ];
        Q1A := TQuery.Create(nil); // Q1A.CursorType := ctOpenForwardOnly; Q1A.CursorLocation := clUseServer; Q1A.LockType := ltReadOnly; Q1A.ExecuteOptions := [ eoAsyncFetch ];
        Q2 := TQuery.Create(nil); // Q2.CursorType := ctOpenForwardOnly; Q2.CursorLocation := clUseServer; Q2.LockType := ltReadOnly; Q2.ExecuteOptions := [ eoAsyncFetch ];
        Q3 := TQuery.Create(nil); // Q3.CursorType := ctOpenForwardOnly; Q3.CursorLocation := clUseServer; Q3.LockType := ltReadOnly; Q3.ExecuteOptions := [ eoAsyncFetch ];
        Q3A := TQuery.Create(nil); // Q3.CursorType := ctOpenForwardOnly; Q3.CursorLocation := clUseServer; Q3.LockType := ltReadOnly; Q3.ExecuteOptions := [ eoAsyncFetch ];
        Q3B := TQuery.Create(nil); // Q3.CursorType := ctOpenForwardOnly; Q3.CursorLocation := clUseServer; Q3.LockType := ltReadOnly; Q3.ExecuteOptions := [ eoAsyncFetch ];
        Q4 := TQuery.Create(nil); // Q4.CursorType := ctOpenForwardOnly; Q4.CursorLocation := clUseServer; Q4.LockType := ltReadOnly; Q4.ExecuteOptions := [ eoAsyncFetch ];
        Q4B := TQuery.Create(nil); // Q4B.CursorType := ctOpenForwardOnly; Q4B.CursorLocation := clUseServer; Q4B.LockType := ltReadOnly; Q4B.ExecuteOptions := [ eoAsyncFetch ];
        Q4C := TQuery.Create(nil); // Q4C.CursorType := ctOpenForwardOnly; Q4C.CursorLocation := clUseServer; Q4C.LockType := ltReadOnly; Q4C.ExecuteOptions := [ eoAsyncFetch ];
        Q5 := TQuery.Create(nil); // Q5.CursorType := ctOpenForwardOnly; Q5.CursorLocation := clUseServer; Q5.LockType := ltReadOnly; Q5.ExecuteOptions := [ eoAsyncFetch ];
        Q0 := TQuery.Create(nil);
        //\\ListParamId := TList.Create;
        drop_E := FALSE;
        drop_D := FALSE;
        drop_Pks := FALSE;
        TRY
            Q0.Text :=
                'PRINT ''--------------- GET nomenklid OF   ----------''' +
                '!select nomenklid from nomenklinfo where nomenkl = '' ''';
            Q0.Open;
            if not Q0.Eof then
                NoNomenklId := Q0.IFieldN[0]
            else
                NoNomenklId := 0;
            Q0.Close;
            opids := '';
            if Code_operation <> '' then
            begin
                Q0.SQL.Text := 'select opid from Operation where Code_operation in (' +
                    Code_operation + ')';
                Q0.Open;
                while not Q0.Eof do
                begin
                    opids := opids + ',' + Q0.SFieldN[0];
                    Q0.Next;
                end;
                Q0.Close;
                Delete(opids, 1, 1);
            end;

            if Trim(ListConditions) <> '' then
            begin
                //    #B  ,     
                Q1.SQL.Text :=
                    #13'print ''------------------------- Q1 ------------------------''' +
                    Drop('#B') +
                    #13'create table #B (batchid int)' +


                #13'insert into #B ' +
                    IfThenElse(pos('nomenklid', ListConditions) > 0,
                    #13'select ' +
                    //'distinc ' +
                    #13'pk.batchid from ' + PackageTableName + ' pk, Batch B' +
                    #13'where pk.batchid = B.batchid ' +
                    #13'group by pk.batchid'
                    ,
                    IfThenElse(ListCon_has_Conditions and
                    (ListConditions <> 'pk.packid = T.packid'),
                    #13'select ' +
                    //'distinc ' +
                    'T.batchid from #TempRslt T' +
                    ' where ' + ListConditions +
                    #13'group by T.batchid',
                    IfThenElse(ListCon_has_Conditions,
                    #13'select ' +
                    //'distinc ' +
                    'batchid from #TempRslt where batchid is not null' +
                    #13'group by batchid',
                    #13'select ' +
                    //'distinct ' +
                    'pk.batchid from ' + PackageTableName + ' pk ' +
                    IfThenElse(ListConditions <> '', ' where ' + ListConditions,
                    ' where pk.batchid is not null') +
                    #13'group by pk.batchid'))
                    ) +

                #13'select count(*) from #B';

                Q1A.SQL.Text :=
                    Fix_Condition_Sign_PresenceIsNull(
                    #13'print ''--------------------------- Q1A ---------------------''' +
                    #13 + Drop('#Pks') +
                    //     ,   #B
                    #13'create table #Pks (packid int, batchid int NULL' +
                    #13', nomenklid int NULL' +
                    #13', grpid int, code_trans int ' + _NULL +
                    #13') ' +
                    #13 +

                    #13'insert into #Pks' +
                    #13'select pk.packid, null, null, pk.grpid, pk.code_trans' +
                    #13'from ' + PackageTableName + ' pk' +
                    IfThenElse(pos('nomenklid', AdditionalConds) > 0,
                    ', Batch B', '') +
                    IfThenElse(ListCon_has_Conditions and (ListConditions <> ''), ', #TempRslt T ', '') +
                    #13'where pk.batchid is null and ' + ListConditions +
                    IfThenElse(pos('nomenklid', AdditionalConds) > 0,
                    ' and pk.batchid = B.batchid', '') +
                    IfThenElse(ListCon_has_Conditions,
                    ' and T.rslt > 0 ', '') +
                    #13 +
                    #13'insert into #Pks' +
                    #13'select pk.packid, B.batchid, B.nomenklid, pk.grpid, pk.code_trans' +
                    #13'from ' + PackageTableName + ' pk, Batch b' +
                    IfThenElse(ListCon_has_Conditions and (ListConditions <> ''), ', #TempRslt T ', '') +
                    #13'where pk.batchid = B.batchid ' +
                    IfThenElse(ListConditions <> '', 'and ' + ListConditions, '') +
                    IfThenElse(ListCon_has_Conditions,
                    ' and pk.packid = T.packid and T.rslt > 0',
                    ' and pk.newpack is null' +
                    ' and pk.sign_exist is null' +
                    ' and pk.sign_virtual is null' +
                    ' and pk.Sign_presence is null') +
                    IfThenElse(AdditionalConds <> '', ' and ' + AdditionalConds, '')
                    );

                //     ?
                if LoadParentParams then
                    Q1A.SQL.Text := Q1A.SQL.Text +
                        #13'print ''------------------------- Q1-1 ----------------------''' +
                        #13'declare @@c int' +
                        #13'select @@c = 1' +

                    #13'while @@c > 0 begin' +
                        #13'insert into #Pks' +
                        #13'select pk.Parent_pack, null, null, pk.grpid, 0' +
                        #13'from ' + PackageTableName + ' pk, #Pks' +
                        #13'where pk.packid = #Pks.packid and pk.Parent_pack is not null' +
                        #13'and not exists( select * from #Pks s where' +
                        #13's.packid = pk.Parent_pack )' +
                        #13'select @@c = @@rowcount' +
                        #13'end';

                Q1A.SQL.Text := Q1A.SQL.Text +
                    #13'print ''------------------------- Q1-2 ----------------------''' +
                    #13'update #Pks set nomenklid=n.nomenklid' +
                    #13' from #Pks p, Nomenklinfo N ' +
                    #13'where p.nomenklid is null and n.nomenkl='' ''';

                if ListCon_has_Conditions then
                    Q1A.SQL.Text := Q1A.SQL.Text +
                        #13'print ''------------------------- Q1-3 ----------------------''' +
                        #13'select pks.*, P.numcon, P.Sign_final ' +
                        #13'from #Pks pks, ' + PackageTableName + ' P' +
                        #13'where pks.packid = P.packid ' +
                        #13'group by pks.packid, pks.batchid, pks.nomenklid, pks.grpid, ' +
                        #13'      pks.code_trans, P.numcon, P.sign_final'
                else
                    Q1A.SQL.Text := Q1A.SQL.Text +
                        #13'print ''------------------------- Q1-4 ----------------------''' +
                        #13'select pks.*, P.numcon, P.Sign_final ' +
                        #13'from #Pks pks, ' + PackageTableName + ' P' +
                        #13'where pks.packid = P.packid' +
                        #13'group by pks.packid, pks.batchid, pks.nomenklid, pks.grpid, ' +
                        #13'      pks.code_trans,P.numcon, P.sign_final';

                //    Param,    
                Q2.SQL.Text :=
                    #13'print ''------------------------- Q2 ------------------------''' +
                    #13'create table #NN (nid int)' +
                    #13'insert into #NN select ' +
                    //'distinc ' +
                'pks.nomenklid from #Pks pks' +
                    #13'group by pks.nomenklid' +
                    #13 +
                    #13'create table #PP (paramid int, param_order int, op_hidden int NULL,' +
                    #13'ofr numeric(12,7) NULL,' +
                    #13'recalc int NULL, calc_by_batch int NULL, Code_Operation varchar(20),' +
                    #13'Sign_clear int NULL,' +
                    #13'nomenklid int, np_Ed_ism varchar(10) NULL, npfr numeric(12,7) NULL, ' +
                    #13'nppr int NULL)' +

                #13'insert into #PP' +
                    #13'select p.paramid,' +
                    #13'op.param_order, op.hidden, ' +
                    #13'op.Format_representation,' +
                    #13'op.recalc, op.calc_by_batch, o.Code_Operation, o.Sign_clear,' +
                    #13'np.nomenklid, np.Ed_ism as np_Ed_ism, np.Format_representation as npfr,' +
                    #13'np.Precis as nppr' +
                    #13'from Operation_Params op, Operation o, #NN,' +
                    #13'Param p RIGHT OUTER join Nomencl_param np on(' +
                    #13'     np.paramid = p.paramid )' +
                    #13'where' +
                    IfThenElse(Code_operation <> '',
                    #13'o.Code_Operation in (' + Code_Operation + ') and', '') +
                    #13'      o.opid = op.opid and op.paramid = p.paramid' +
                    #13' and  np.nomenklid = #NN.nid' +
                    #13'group by p.paramid,' +
                    #13'  op.param_order, op.hidden, op.Operation_param_id,' +
                    #13'  op.Format_representation, ' +
                    #13'  op.recalc, op.calc_by_batch, np.Format_representation,' +
                    #13'  o.Code_operation, o.Sign_clear, np.nomenklid, np.Ed_ism, np.Precis' +
                    #13'order by op.Operation_param_id' +

                #13'select p.paramid, p.param_name, p.param_descr, p.formula, ' +
                    #13'p.param_type, convert( varchar(33), p.param_val ) as param_val, ' +
                    #13'p.Ed_ism, p.Precis, ' +
                    #13'p.Format_representation as pfr, p.copy_from_param, p.Is_Analyse,' +
                    #13'pp.op_hidden as hidden, ' +
                    #13'pp.param_order, pp.op_hidden, pp.ofr,' +
                    #13'pp.recalc, pp.calc_by_batch,' +
                    #13'pp.Code_Operation, pp.Sign_clear,' +
                    #13'pp.nomenklid, pp.np_Ed_ism, pp.npfr, pp.nppr' +
                    #13'from Param p, #PP pp where p.paramid = pp.paramid' +
                    #13'drop table #PP' +

                #13 +
                    #13'drop table #NN';


                Q3.SQL.Text := //Drop( '#D' ) +
                #13'print ''------------------------- Q3 ------------------------''' +
                    #13'create table #D (paramid int,' +
                    #13' nomenklid int ' + _NULL +
                    #13', packid int, grpid int)' +
                    #13'create table #C (paramid int, nomenklid int, np_Ed_ism varchar(10) NULL)' +
                    #13 +

                #13'print ''======================= insert #D''' +
                    #13'insert into #D (paramid, packid, grpid, nomenklid) ' +
                    #13'select distinct V.paramid, pk.packid, pk.grpid, B.nomenklid ' +
                    #13'from Value V, Param p, ' +
                    #13'' + PackageTableName + ' pk left outer join Batch B on (pk.batchid = B.batchid) ' +
                    #13', Operation_params o_p' +
                    IfThenElse(ListCon_has_Conditions and (ListConditions <> ''), ', #TempRslt T ', '') +
                    #13'where ' +
                    IfThenElse(ListConditions <> '',
                    #13 + IfThenElse(ListCondShort <> ListConditions, ListCondShort, ListConditions), '') +
                    #13' and  V.paramid = p.paramid and V.grpid = pk.grpid ' +
                    IfThenElse(ListCon_has_Conditions, 'and T.rslt > 0 ', '') + //    = 1 ?
                #13' and o_p.paramid = p.paramid ' +
                    IfThenElse(opids <> '',
                    #13' and o_p.opid  in (' + opids + ')',
                    '') +
                    //#13'group by V.paramid, pk.packid, pk.grpid, B.nomenklid' +

                IfThenElse((ListConditions <> '') and (ListCondShort <> ListConditions),
                    #13'print ''===================== delete #D''' +
                    #13'delete #D from #D d, Package pk where d.packid = pk.packid and not' +
                    #13'    ' + ListConditions, ''
                    ) +

                #13 +
                    #13'print ''======================= update #D''' +
                    #13'update #D set nomenklid = N.nomenklid from #D d, nomenklinfo N' +
                    #13'where d.nomenklid is null and N.nomenkl = '' ''' +

                #13'print ''======================= insert #C''' +
                    #13'insert into #C (paramid, nomenklid, np_Ed_ism)' +
                    #13'select distinct #D.paramid, #D.nomenklid, np.Ed_ism ' +
                    #13'from nomencl_param np, #D ' +
                    #13'where np.nomenklid = #D.nomenklid and' +
                    #13'      np.paramid = #D.paramid' +
                    //#13'group by #D.paramid, #D.nomenklid, np.Ed_ism' +

                #13'create table #E (nomenklid int, paramid int)' +
                    #13'print ''======================= insert #E''' +
                    #13'insert into #E select nomenklid, paramid from #D' +
                    #13'group by nomenklid, paramid' +

                #13'create table #PPP ( paramid int, param_order int, hidden int NULL,' +
                    #13'op_Format numeric(12,7) NULL, Operation_param_id int, ' +
                    #13'calc_by_batch int NULL, recalc int NULL, opid int, ' +
                    #13'Code_operation varchar(20), Sign_clear int NULL,' +
                    #13'npfr numeric(12,7) NULL, nppr int NULL)' +

                #13'print ''======================= insert #PPP''' +
                    #13'insert into #PPP' +
                    #13'select p.paramid,' +
                    #13' o_p.param_order, o_p.hidden, ' +
                    #13' o_p.Format_representation as op_Format,' +
                    #13' o_p.Operation_param_id, o_p.calc_by_batch, o_p.recalc,' +
                    #13' o.opid, o.Code_operation, o.Sign_clear, ' +
                    #13'n_p.Format_representation as npfr, n_p.Precis as nppr' +
                    #13'from Param p, #E,' +
                    #13' Operation_Params o_p, Operation o, Nomencl_param n_p' +
                    #13'where p.paramid = #E.paramid' +
                    #13'and #E.nomenklid = n_p.nomenklid and p.paramid = n_p.paramid' +
                    #13' and o_p.paramid = p.paramid' +
                    #13' and o_p.opid = o.opid' +
                    IfThenElse(Code_operation <> '',
                    #13' and o.Code_operation in (' + Code_Operation + ')', '') +
                    #13'group by p.paramid,' +
                    #13'  o_p.param_order, o_p.hidden, ' +
                    #13'  o_p.Format_representation,' +
                    #13'o.Code_operation, o.Sign_clear, o_p.Operation_param_id, ' +
                    #13'o_p.calc_by_batch, o_p.recalc, o.opid, n_p.Format_representation,' +
                    #13'n_p.Precis';

                Q3A.SQL.Text :=
                    #13'select p.paramid, p.param_name, p.param_descr, p.formula, ' +
                    #13'p.param_type, convert( varchar(33), p.param_val ) as param_val, ' +
                    #13'p.Ed_ism, ' +
                    #13'p.Format_representation, ' +
                    #13'p.precis, ppp.* from Param p, #PPP ppp' +
                    #13'where p.paramid = ppp.paramid';

                Q3B.SQL.Text :=
                    #13'drop table #PPP'
                    ;

                //     Value
                Q4.SQL.Text :=
                    #13'print ''------------------------- Q4 ------------------------''' +
                    #13'select ' + {'DISTINCT ' +} '#D.nomenklid, #D.grpid, #D.packid' +
                #13'from #D' +
                    #13'group by #D.nomenklid, #D.grpid, #D.packid'
                    ;

                Q4B.SQL.Text :=
                    #13'print ''------------------------- Q4B -----------------------''' +
                    #13'select #D.grpid, #D.paramid,' +
                    //#13'convert( varchar(33), V.param_calc ) as param_calc' +
                #13'V.param_calc' +
                    #13'from #D, Value V' +
                    #13'where V.grpid = #D.grpid and V.paramid = #D.paramid' +
                    #13'group by #D.grpid, #D.paramid , V.param_calc' +
                    #13 //'order by #D.grpid'
                ;

                Q4C.SQL.Text :=
                    #13'print ''------------------------- Q4C -----------------------''' +
                    #13'select ''Q4C'', paramid, nomenklid, np_Ed_ism from #C';

            end
            else //-------------------------------------------------------------------
            begin //  :      
                //       ,   
                //  
                Q1.SQL.Text :=
                    #13'print ''------------------------- Q1-else -------------------''' +
                    Drop('#B') +
                    #13'create table #B (b int)' +
                    #13'select nomenklid from nomenklinfo' +
                    #13 + IfThenElse(only_nomenklid > 0,
                    #13'where  nomenklid = ' + IntToStr(only_nomenklid), '') +
                    #13;
                if Trim(Code_operation) <> '' then
                begin
                    Q2.SQL.Text :=
                        #13'print ''------------------------- Q2 ----------------------''' +
                        #13'select p.paramid, p.param_name, p.param_descr, p.formula, p.param_type,' +
                        #13'convert(varchar(33),p.param_val)as param_val, p.Ed_ism,' +
                        #13'p.Format_representation as pfr, p.copy_from_param, p.Is_Analyse,' +
                        #13'p.precis,' +
                        #13'n.nomenklid, op.param_order, ' +
                        #13'op.hidden as op_hidden, op.hidden, ' +
                        #13'op.Format_representation as ofr, ' +
                        #13'op.calc_by_batch, op.recalc, o.Code_Operation, ' +
                        #13'np.Ed_ism as np_Ed_ism, np.Format_representation as npfr, ' +
                        #13'np.Precis as nppr' +

                    #13'  from Param p, ' +
                        #13'       Operation_Params op, Operation o, ' +
                        #13'       Nomencl_param np, Nomenklinfo n' +
                        #13'where np.paramid = p.paramid ' +
                        #13' and  n.nomenklid = np.nomenklid ' +
                        IfThenElse(Code_operation <> '',
                        #13' and o.Code_Operation in (' + Code_Operation + ')',
                        '') +
                        #13' and o.opid = op.opid and op.paramid = p.paramid ' +
                        IfThenElse(only_nomenklid > 0,
                        #13' and n.nomenklid = ' + IntToStr(only_nomenklid), '');
                end
                else
                begin
                    Q2.SQL.Text :=
                        #13'print ''------------------------- Q2-else -----------------''' +
                        #13'create table #P4 (paramid int, pfr numeric(12,7), nomenklid int,' +
                        #13' np_Ed_ism varchar(10), hidden int, Code_operation varchar(20),' +
                        #13' param_order int, op_hidden int, ofr numeric(12,7),' +
                        #13' calc_by_batch int, recalc int, npfr numeric(12,7), nppr int)' +
                        #13'insert into #P4' +

                    #13'select p.paramid, ' +
                        #13'p.Format_representation, ' +
                        #13'n.nomenklid, ' +
                        #13'np.Ed_ism, null, o.Code_operation, op.param_order, op.hidden, ' +
                        #13'op.Format_representation as ofr, ' +
                        #13'op.calc_by_batch, op.recalc, np.Format_representation as npfr,' +
                        #13'np.Precis as nppr' +
                        #13'ORDER BY n.nomenklid, p.paramid' + //+++++++++++++++++
                    //  ????

                    #13'from Nomencl_param np, Nomenklinfo n, Param p' +
                        #13' LEFT OUTER JOIN Operation_params op' +
                        #13' ON( op.paramid = p.paramid ) ' +
                        #13'LEFT OUTER JOIN Operation o on(op.opid = o.opid)' +
                        #13'where np.paramid = p.paramid' +
                        #13' and  n.nomenklid = np.nomenklid' +

                    #13'group by p.paramid, p.Format_representation, n.nomenklid, ' +
                        #13'np.Ed_ism, o.Code_operation, op.param_order, op.hidden, ' +
                        #13'op.Format_representation, op.calc_by_batch, op.recalc, ' +
                        #13'np.Format_representation' +

                    #13'select p.param_name, p.param_descr, p.formula, p.param_type, ' +
                        #13'convert(varchar(33),param_val) as param_val, p.Ed_ism, p.precis,' +
                        #13'p.copy_from_param, p.Is_Analyse,' +
                        #13'pppp.* from Param p, #P4 pppp ' +
                        #13'where p.paramid = pppp.paramid' +

                    #13'drop table #P4';
                end;
            end;
            Q5.SQL.Text :=
                #13'print ''------------------------- Q5 --------------------------''' +
                #13'select n.nomenklid, n.nomenkl, ' +
                #13'n.nominal, ' +
                #13'n.elempct, ' +
                #13'n.wrand, ' +
                #13'n.wsys, ' +
                #13'n.urand, ' +
                #13'n.usys, ' +
                #13'n.ssys, ' +
                #13'n.srand, ' +
                #13'm.percentenrich, ' +
                #13'm.Min_percent, ' +
                #13'm.Max_percent, ' +
                #13'm.Accuracy_element, ' +
                #13'm.Accuracy_isotope' +
                #13'from Nomenklinfo n, Mattype m' +
                #13'where n.mattypeid = m.mattypeid ';

            {$IFDEF ALL_IN_ONE}
            Qall.SQL.Text :=
                Q1.SQL.Text + #13 +
                Q1A.SQL.Text + #13 +
                Q2.SQL.Text + #13 +
                Q3.SQL.Text + #13 +
                Q3A.SQL.Text + #13 +
                Q3B.SQL.Text + #13 +
                Q4.SQL.Text + #13 +
                Q4B.SQL.Text + #13 +
                Q4C.SQL.Text + #13 +
                Q5.SQL.Text;
            drop_D := TRUE;
            drop_E := TRUE;
            drop_Pks := TRUE;
            if DebugMode then
                Qall.SQL.SaveToFile(GetLogsPath + 'LastQuery.txt');
            {$ENDIF}

            if Trim(ListConditions) <> '' then
            begin
                if Assigned(OnParamLoad) then
                    OnParamLoad('1 -   ');
                I := 0;
                //TRY
                q(Q1).Open;
                if not q(Q1).Eof then
                begin
                    I := q(Q1).IFieldN[0];
                end;
                //EXCEPT
                //END;
                Q1.Close;
                if Assigned(OnParamLoad) then
                    OnParamLoad('1A -    (' + IntToStr(I) +
                        ' )');
                {$IFDEF ALL_IN_ONE}
                More;
                {$ELSE}
                drop_Pks := TRUE;
                Q1A.Open;
                {$ENDIF}
                if Assigned(OnParamLoad) then
                    OnParamLoad('2 -   ');

                //--------------------------------------------------------------------------
                //   ,   ,   TParams
                while not q(Q1A).Eof do
                begin
                    P := TParams.Create;
                    P.FParamsController := ParamsController;
                    Result.Add(P);
                    P.ParamSetData.packid := q(Q1A).IField['packid'];
                    P.ParamSetData.numcon := q(Q1A).SField['numcon'];
                    P.ParamSetData.batchid := q(Q1A).IField['batchid'];
                    P.ParamSetData.Sign_final := q(Q1A).IField['Sign_final'] <> 0;
                    P.ParamSetData.nomenklid := q(Q1A).IField['nomenklid'];
                    P.ParamSetData.grpid := q(Q1A).IField['grpid'];
                    P.ParamSetData.code_trans := q(Q1A).IField['code_trans'];
                    if Result.Count mod 100 = 0 then
                        if Assigned(OnParamLoad) then
                            OnParamLoad('2 -    (' + IntToStr(Result.Count) + ')');
                    q(Q1A).Next;
                end;
                Q1A.Close;

                //--------------------------------------------------------------------------
                //   
                if Assigned(OnParamLoad) then
                    OnParamLoad('3 -   ');
                {$IFDEF ALL_IN_ONE}
                More;
                {$ELSE}
                Q2.Open;
                {$ENDIF}
                if Assigned(OnParamLoad) then
                    OnParamLoad('4 -   ');
            end
            else
            begin
                if Assigned(OnParamLoad) then
                    OnParamLoad('2A -   ');
                q(Q1).Open;
                if Assigned(OnParamLoad) then
                    OnParamLoad('2B -   ');
                while not q(Q1).Eof do
                begin
                    P := TParams.Create;
                    P.FParamsController := ParamsController;
                    Result.Add(P);
                    P.ParamSetData.nomenklid := q(Q1).IField['nomenklid'];
                    //P.nomenklid := P.ParamSetData.nomenklid;
                    if Result.Count mod 100 = 0 then
                        if Assigned(OnParamLoad) then
                            OnParamLoad('2B -    (' + IntToStr(Result.Count) + ')');
                    q(Q1).Next;
                end;
                Q1.Close;
                if Assigned(OnParamLoad) then
                    OnParamLoad('3 -   ');
                {$IFDEF ALL_IN_ONE}
                More;
                {$ELSE}
                Q2.Open;
                {$ENDIF}
                if Assigned(OnParamLoad) then
                    OnParamLoad('4 -   ');
            end;

            //--------------------------------------------------------------------------
            //  :   
            Cache.Clear;
            K := 0;
            while not q(Q2).Eof do
            begin
                Cache[K, 'param_name'] := q(Q2).SField['param_name'];
                Cache[K, 'param_descr'] := q(Q2).SField['param_descr'];
                Cache[K, 'formula'] := q(Q2).SField['formula'];
                if q(Q2).NullField['np_Ed_ism'] then
                    Cache[K, 'Ed_ism'] := q(Q2).SField['Ed_ism']
                else Cache[K, 'Ed_ism'] := q(Q2).SField['np_Ed_ism'];
                Cache[K, 'Code_operation'] := q(Q2).SField['Code_operation'];
                Cache.IntVal[K, 'nomenklid'] := q(Q2).IField['nomenklid'];
                Cache.IntVal[K, 'paramid'] := q(Q2).IField['paramid'];
                Cache.IntVal[K, 'param_type'] := q(Q2).IField['param_type'];
                if not q(Q2).NullField['nppr'] then
                    Cache.IntVal[K, 'precis'] := q(Q2).IField['nppr']
                else
                    if q(Q2).NullField['precis'] then
                        Cache.IntVal[K, 'precis'] := MAX_PRECISION
                    else
                        Cache.IntVal[K, 'precis'] := q(Q2).IField['precis'];
                Cache.IntVal[K, 'calc_by_batch'] := q(Q2).IField['calc_by_batch'];
                Cache.IntVal[K, 'recalc'] := q(Q2).IField['recalc'];
                Cache.IntVal[K, 'param_order'] := q(Q2).IField['param_order'];
                Cache.IntVal[K, 'hidden'] := q(Q2).IField['hidden'];
                if not q(Q2).NullField['npfr'] then
                    Cache[K, 'format_representation'] := q(Q2).SField['npfr']
                else
                    if q(Q2).NullField['ofr'] then
                        Cache[K, 'format_representation'] := q(Q2).SField['pfr']
                    else
                        Cache[K, 'format_representation'] := q(Q2).SField['ofr'];
                Cache[K, 'param_val'] := q(Q2).SField['param_val'];
                Cache[K, 'copy_from_param'] := q(Q2).SField['copy_from_param'];
                Cache.IntVal[K, 'Is_Analyse'] := q(Q2).IField['Is_Analyse'];
                q(Q2).Next;
                inc(K);
                if K mod 100 = 0 then
                    if Assigned(OnParamLoad) then
                        OnParamLoad('4 -    (' + IntToStr(K) + ')');
            end;
            inc(Total_params_loaded, K);
            Q2.Close;
            if Assigned(OnParamLoad) then
                OnParamLoad('5 -   ');
            SetLength(ParamLoad, K + 1);
            for K := 0 to Cache.Count - 1 do
            begin
                ParamLoad[K].param_name := Cache[K, 'param_name'];
                ParamLoad[K].nomenklid := Cache.IntVal[K, 'nomenklid'];
                ParamLoad[K].param_descr := Cache[K, 'param_descr'];
                ParamLoad[K].formula := Cache[K, 'formula'];
                ParamLoad[K].Ed_ism := Cache[K, 'Ed_ism'];
                ParamLoad[K].Code_operation := Cache[K, 'Code_operation'];
                ParamLoad[K].nomenklid := Cache.IntVal[K, 'nomenklid'];
                ParamLoad[K].paramid := Cache.IntVal[K, 'paramid'];
                ParamLoad[K].param_type := Cache.IntVal[K, 'param_type'];
                ParamLoad[K].Precis := Cache.IntVal[K, 'precis'];
                ParamLoad[K].calc_by_batch := Cache.IntVal[K, 'calc_by_batch'];
                ParamLoad[K].recalc := Cache.IntVal[K, 'recalc'];
                ParamLoad[K].param_order := Cache.IntVal[K, 'param_order'];
                ParamLoad[K].hidden := Cache.IntVal[K, 'hidden'];
                ParamLoad[K].Format_representation := StrTofloat(Cache[K, 'format_representation']);
                ParamLoad[K].param_val := StrTofloat(Cache[K, 'param_val']);
                ParamLoad[K].A_Copy_f := Cache[K, 'copy_from_param'];
                ParamLoad[K].Is_A := Cache.IntVal[K, 'Is_Analyse'];
                if K mod 100 = 0 then
                    if Assigned(OnParamLoad) then
                        OnParamLoad('5 -    (' + IntToStr(K) + ')');
            end;
            if Assigned(OnParamLoad) then
                OnParamLoad('6 -    ');
            for I := 0 to Result.Count - 1 do
            begin
                P := Result[I];
                //       
                //     ,    
                //   ,     
                // . (       
                //  ,   ).
                if (P.Capacity = P.Count) and (I > 0) then
                begin
                    P1 := Result[I - 1];
                    //if P1.Capacity > P.Capacity then
                    if P1.Count > P.Capacity then
                        //P.Capacity := P1.Capacity;
                        P.Capacity := P1.Count;
                end;
                for K := 0 to Cache.Count - 1 do
                begin
                    if (P.ParamSetData.nomenklid = ParamLoad[K].nomenklid) or
                        (P.ParamSetData.nomenklid = 0) and (ParamLoad[K].nomenklid = NoNomenklId) {or
                    (Code_operation <> '')}then
                    begin
                        //P.ParamSetData.NoCalc := not ParamLoad[ K ].really_need;
                        //if (P.IndexOf( ParamLoad[ K ].param_name ) < 0) then
                        if (P.findParamByParamNameAndNomenklid(ParamLoad[K].param_name,
                            ParamLoad[K].nomenklid) < 0) then
                        begin
                            if (ParamLoad[K].param_type = 1) {or
                            (ParamLoad[ K ].param_type = 3) and (ParamLoad[ K ].formula = '')}
                            then
                                PD := P.AddConst(ParamLoad[K].param_name, ParamLoad[K].nomenklid,
                                    ParamLoad[K].param_val,
                                    ParamLoad[K].paramid, ParamLoad[K].param_descr,
                                    ParamLoad[K].Code_operation)
                            else
                                PD := P.Add(ParamLoad[K].param_name, ParamLoad[K].nomenklid,
                                    ParamLoad[K].param_val,
                                    ParamLoad[K].paramid, ParamLoad[K].param_descr,
                                    ParamLoad[K].Code_operation,
                                    ParamLoad[K].A_Copy_f, ParamLoad[K].Is_A = 1);
                            if not PD.param_def.loaded then
                            begin
                                PD.param_type := ParamLoad[K].param_type;
                                PD.formula := ParamLoad[K].formula;
                                PD.Units := ParamLoad[K].Ed_ism;
                                PD.param_format := ParamLoad[K].format_representation;
                                PD.precision := ParamLoad[K].precis;

                                PD.calc_by_batch := ParamLoad[K].calc_by_batch = 0;
                                PD.recalc := ParamLoad[K].recalc <> 0;

                                PD.param_def.loaded := TRUE;
                            end;

                            if ParamLoad[K].hidden = 0 then
                                PD.Set_show_in_op(ParamLoad[K].Code_operation);
                            PD.Set_used_in_op(ParamLoad[K].Code_operation,
                                ParamLoad[K].param_order);
                            PD.param_ready := PD.param_type = 1;
                        end
                        else //     (,   
                            // ,     ( ))
                        begin
                            J := P.findParamByParamNameAndNomenklid(
                                ParamLoad[K].param_name, ParamLoad[K].nomenklid);
                            if (ParamLoad[K].hidden = 0) then
                                P.ParamData[J].Set_show_in_op(ParamLoad[K].Code_operation);
                            P.ParamData[J].Set_used_in_op(ParamLoad[K].Code_operation,
                                ParamLoad[K].param_order);
                        end;
                    end;
                end;
                if I mod 100 = 0 then
                    if Assigned(OnParamLoad) then
                        OnParamLoad('6 -     (' + IntToStr(I) +
                            '/' + IntToStr(Result.Count) + ')');
            end;
            if Assigned(OnParamLoad) then
                OnParamLoad('7 -     ');
            for K := 0 to Cache.Count - 1 do
            begin
                ParamLoad[K].param_name := '';
                ParamLoad[K].param_descr := '';
                ParamLoad[K].formula := '';
                ParamLoad[K].Ed_ism := '';
                ParamLoad[K].Code_operation := '';
            end;
            SetLength(ParamLoad, 0);

            if Trim(ListConditions) <> '' then
            begin
                //------------------------------------------------------------------------
                //  (  hidden=1)    
                // ( Param_values),     ,   
                // hidden = (  Operation_params) ,  

                //  1:     ( )
                TRY
                    if Assigned(OnParamLoad) then
                        OnParamLoad('8 -    ');
                    {$IFDEF ALL_IN_ONE}
                    //  More         
                    {$ELSE}
                    drop_E := TRUE;
                    drop_D := TRUE;
                    Q3.ExecSQL;
                    {$ENDIF}
                    if Assigned(OnParamLoad) then
                        OnParamLoad('8A -    ');
                    Q3.Close;
                    {$IFDEF ALL_IN_ONE}
                    More;
                    {$ELSE}
                    Q3A.Open;
                    {$ENDIF}
                EXCEPT on E: Exception do
                    begin
                        ShowMessage(' ' + E.Message);
                        raise;
                    end;
                END;

                if Assigned(OnParamLoad) then
                    OnParamLoad('9 -   ');
                ParamsLoaded := TList.Create;
                TRY
                    K := -1;
                    while not q(Q3A).EOF do
                    begin
                        //if Ops.IndexOf( Q3A.SField[ 'Code_operation' ] ) >= 0 then
                        begin
                            inc(K);
                            Cache[K, 'param_descr'] := q(Q3A).SField['param_descr'];
                            Cache[K, 'param_name'] := q(Q3A).SField['param_name'];
                            Cache[K, 'formula'] := q(Q3A).SField['formula'];
                            Cache[K, 'Ed_ism'] := q(Q3A).SField['Ed_ism'];
                            Cache[K, 'Code_operation'] := q(Q3A).SField['Code_operation'];

                            PL := AllocMem(Sizeof(PL^));
                            PL.paramid := q(Q3A).IField['paramid'];
                            PL.param_descr := Cache[K, 'param_descr'];
                            PL.param_name := Cache[K, 'param_name'];
                            //PL.nomenklid := 0;
                            PL.param_type := q(Q3A).IField['param_type'];
                            PL.formula := Cache[K, 'formula'];
                            PL.param_val := q(Q3A).FField['param_val'];
                            PL.Ed_ism := Cache[K, 'Ed_ism'];
                            if not q(Q3A).NullField['npfr'] then
                                PL.Format_representation := StrTofloat(q(Q3A).SField['npfr'])
                            else if q(Q3A).FField['op_Format'] <> 0 then
                                PL.Format_representation := StrTofloat(q(Q3A).SField['op_Format'])
                            else
                                PL.Format_representation := StrTofloat(q(Q3A).SField['Format_representation']);
                            if not q(Q3A).NullField['nppr']
                                then PL.Precis := q(Q3A).IField['nppr']
                            else PL.Precis := q(Q3A).IField['precis'];
                            PL.param_order := q(Q3A).IField['param_order'];
                            PL.hidden := q(Q3A).IField['hidden'];
                            PL.Operation_param_id := q(Q3A).IField['Operation_param_id'];
                            PL.opid := q(Q3A).IField['opid'];
                            PL.Code_operation := Cache[K, 'Code_operation'];
                            PL.Sign_clear := q(Q3A).IField['Sign_clear'];
                            PL.calc_by_batch := q(Q3A).IField['calc_by_batch'] = 0;
                            PL.recalc := q(Q3A).IField['recalc'] <> 0;
                            ParamsLoaded.Add(PL);
                            if K mod 100 = 0 then
                                if Assigned(OnParamLoad) then
                                    OnParamLoad('9 -    (' + IntToStr(K) + ')');
                        end;
                        q(Q3A).Next;
                    end;
                    Q3A.Close;

                    if Result.Count > 0 then
                    begin
                        if Assigned(OnParamLoad) then
                            OnParamLoad('10 -    ');
                        {$IFDEF ALL_IN_ONE}
                        More;
                        {$ELSE}
                        Q4.Open;
                        {$ENDIF}
                        if Assigned(OnParamLoad) then
                            OnParamLoad('11 -    ');
                        K := 0;
                        SortData(Result, Result.Count, Compare_packid, SwapItems);

                        while not q(Q4).Eof do
                        begin
                            inc(K);
                            q4_packid := q(Q4).IField['packid'];
                            q4_nomenklid := q(Q4).IField['nomenklid'];

                            I1 := 0;
                            I2 := Result.Count - 1;
                            I := I1;
                            P := Result[I];
                            if P.ParamSetData.packid <> q4_packid then
                                I := I2;
                            P := Result[I];
                            while (P.ParamSetData.packid <> q4_packid) and (I2 - I1 > 1) do
                            begin
                                I := (I1 + I2) div 2;
                                P := Result[I];
                                if P.ParamSetData.packid = q4_packid then break;
                                if P.ParamSetData.packid < q4_packid then
                                    I1 := I
                                else
                                    I2 := I;
                            end;
                            P := Result[I];
                            if P.ParamSetData.packid <> q4_packid then
                                I := I1;
                            P := Result[I];
                            if P.ParamSetData.packid <> q4_packid then
                                I := I2;
                            P := Result[I];
                            if P.ParamSetData.packid <> q4_packid then
                                I := (I1 + I2) div 2;
                            P := Result[I];
                            if P.ParamSetData.packid = q4_packid then
                            begin
                                P.ParamSetData.grpid := q(Q4).IField['grpid'];
                                P.ParamSetData.nomenklid := q4_nomenklid;
                                //P.nomenklid := q4_nomenklid;
                            end
                                {else
                                asm
                                  int 3 //    !
                                end};
                            if K mod 100 = 0 then
                                if Assigned(OnParamLoad) then
                                    OnParamLoad('11 -     (' + IntToStr(K) + ')');
                            q(Q4).Next;
                        end;
                        Q4.Close;
                    end;

                    if Result.Count > 0 then
                    begin
                        if Assigned(OnParamLoad) then
                            OnParamLoad('12 -    ');
                        {$IFDEF ALL_IN_ONE}
                        More;
                        {$ELSE}
                        Q4B.Open;
                        {$ENDIF}
                        if Assigned(OnParamLoad) then
                            OnParamLoad('13 -    ');
                        total_val_grps := 0;
                        SortData(Result, Result.Count, Compare_grpid, SwapItems);
                        I := -1;
                        while not q(Q4B).Eof do
                        begin
                            if CancelCalcParams then break;
                            q4_param_calc := StrTofloat(q(Q4B).SField['param_calc']);
                            q4_paramid := q(Q4B).IField['paramid'];
                            q4_grpid := q(Q4B).IField['grpid'];
                            //---        Result
                            //   grpid (     )
                            if (I >= 0) and (I < Result.Count) then
                            begin
                                P := Result[I];
                                if P.ParamSetData.grpid <> q4_grpid then
                                begin
                                    if I < Result.Count - 1 then
                                    begin
                                        inc(I);
                                        P := Result[I];
                                        if P.ParamSetData.grpid <> q4_grpid then
                                            I := -1;
                                    end
                                    else I := -1;
                                end;
                            end
                            else I := -1;
                            if I < 0 then
                            begin
                                I1 := 0;
                                I2 := Result.Count - 1;
                                I := I1;
                                P := Result[I];
                                if P.ParamSetData.grpid <> q4_grpid then
                                    I := I2;
                                P := Result[I];
                                if P.ParamSetData.grpid <> q4_grpid then
                                    while I2 - I1 > 1 do
                                    begin
                                        if I2 - I1 >= 10 then
                                        begin
                                            P1 := Result[I1]; g1 := P1.ParamSetData.grpid;
                                            P2 := Result[I2]; g2 := P2.ParamSetData.grpid;
                                            I := I1 + (q4_grpid - g1) * (I2 - I1) div (g2 - g1);
                                            if I <= I1 then I := I1 + 1
                                            else
                                                if I >= I2 then I := I2 - 1;
                                        end
                                        else
                                            I := (I1 + I2) div 2;
                                        P := Result[I];
                                        if P.ParamSetData.grpid = q4_grpid then
                                        begin // !
                                            break;
                                        end;
                                        if P.ParamSetData.grpid < q4_grpid then
                                            I1 := I
                                        else
                                            I2 := I;
                                    end;
                            end;
                            P := Result[I];
                            if P.ParamSetData.grpid = q4_grpid then
                            begin //     1,          grpid
                                I1 := I;
                                while I >= 0 do
                                begin
                                    P := Result[I];
                                    if P.ParamSetData.grpid <> q4_grpid then break;
                                    AssignParamsToConteiner;
                                    dec(I);
                                end;
                                I := I1 + 1;
                                while I < Result.Count do
                                begin
                                    P := Result[I];
                                    if P.ParamSetData.grpid <> q4_grpid then break;
                                    AssignParamsToConteiner;
                                    inc(I);
                                end;
                                dec(I);
                            end;

                            q(Q4B).Next;
                        end;
                        Q4B.Close;
                    end;
                    SortData(Result, Result.Count, Compare_packid, SwapItems);
                    SetLength(Hints, 0);

                    if Assigned(OnParamLoad) then
                        OnParamLoad('14 -  ..   ');
                    {$IFDEF ALL_IN_ONE}
                    More;
                    {$ELSE}
                    Q4C.Open;
                    {$ENDIF}
                    if Assigned(OnParamLoad) then
                        OnParamLoad('15 -  ..   ');
                    if not q(Q4C).Eof and (q(Q4C).SFieldN[0] = 'Q4C') then
                        while not q(Q4C).Eof do
                        begin
                            q4_nomenklid := q(Q4C).IField['nomenklid'];
                            q4_paramid := q(Q4C).IField['paramid'];
                            if q(Q4C).SField['np_Ed_ism'] <> '' then
                            begin
                                for I := 0 to Result.count - 1 do
                                begin
                                    P := Result[I];
                                    if P.ParamSetData.nomenklid = q4_nomenklid then
                                    begin
                                        J := 0;
                                        while J < P.Count do
                                        begin
                                            if P.ID[J] = q4_paramid then
                                            begin
                                                P.ParamData[J].Units := q(Q4C).SField['np_Ed_ism'];
                                                break;
                                            end;
                                            Inc(J);
                                        end;
                                    end;
                                end;
                            end;
                            q(Q4C).Next;
                        end
                    else
                        SkipMore := TRUE;

                FINALLY
                    if Assigned(OnParamLoad) then
                        OnParamLoad('16 -  ');
                    //    
                    for i := 0 to ParamsLoaded.Count - 1 do
                    begin
                        PL := ParamsLoaded[i];
                        PL.param_descr := '';
                        PL.param_name := '';
                        PL.formula := '';
                        PL.Ed_ism := '';
                        PL.Code_operation := '';
                        PL.Analyse_copy_from := '';
                        FreeMem(PL);
                    end;
                    ParamsLoaded.Free;
                END;

                {$IFDEF ALL_IN_ONE}
                {$ELSE}
                Q3B.ExecSQL;
                {$ENDIF}
            end;
            //--------------------------------------------------------------------------
            //  ""    nomenklinfo  mattypeid
            //todo:    ,   
            if Assigned(OnParamLoad) then
                OnParamLoad('18 -    ');

            {$IFDEF ALL_IN_ONE}
            More;
            {$ELSE}
            Q5.Open;
            {$ENDIF}
            if Assigned(OnParamLoad) then
                OnParamLoad('18 -    ');
            while not q(Q5).Eof do
            begin
                for I := 0 to Result.Count - 1 do
                begin
                    P := Result[I];
                    if P.ParamSetData.nomenklid = q(Q5).IField['nomenklid'] then
                    begin
                        AddFixedParam(P, 'N', 'nominal', '. ');
                        AddFixedParam(P, 'N', 'elempct', '.  ');
                        AddFixedParam(P, 'N', 'wrand', '.  ');
                        AddFixedParam(P, 'N', 'wsys', '.  ');
                        AddFixedParam(P, 'N', 'urand', '.  U235 ');
                        AddFixedParam(P, 'N', 'usys', '.  U235 ');
                        AddFixedParam(P, 'N', 'srand', '.    ');
                        AddFixedParam(P, 'N', 'ssys', '.    ');
                        AddFixedParam(P, 'M', 'percentenrich', '.  ');
                        AddFixedParam(P, 'M', 'Min_percent', '.  ');
                        AddFixedParam(P, 'M', 'Max_percent', '.  ');
                        AddFixedParam(P, 'M', 'Accuracy_element', '.   ');
                        AddFixedParam(P, 'M', 'Accuracy_isotope', '.   ');
                        break;
                    end;
                end;
                q(Q5).Next;
            end;

            {$IFDEF ALL_IN_ONE}
            Qall.Close;
            {$ELSE}
            Q5.Close;
            {$ENDIF}

            if Assigned(OnParamLoad) then
                OnParamLoad('19 -   OP');
            for I := 0 to Result.Count - 1 do
            begin
                P := Result[I];
                if P.ParamSetData.code_trans <> 0 then
                begin
                    P.Add('OP', P.ParamSetData.nomenklid,
                        P.ParamSetData.code_trans, 0, ' ', '', '', FALSE);
                    K := P.Count - 1;
                    P.ParamData[K].param_type := 1;
                    P.ParamData[K].read_only := TRUE;
                end;
            end;
            //--------------------------------------------------------------------------
            //   ,  :
            if (ListConditions <> '') and not DoNotCalcParamsWhileLoad then
            begin
                if ConList <> nil then
                begin
                    for I := 0 to Result.Count - 1 do
                    begin
                        P := Result[I];
                        if ConList.IndexOf(Pointer(P.ParamSetData.packid)) < 0 then
                            P.ParamSetData.NoCalc := TRUE;
                    end;
                end;
                if Assigned(OnParamLoad) then
                    OnParamLoad('20 -   ');
                CalcParams(Result, ErrorStr, FALSE, AllowSum, FALSE, Code_Operation);
                if (ErrorStr <> '') and AllowRaise then
                begin
                    raise Exception.Create('     : ' +
                        ErrorStr);
                end;
            end;

        FINALLY
            if Assigned(OnParamLoad) then
                OnParamLoad('21 -   ');
            ConList.Free;
            {$IFDEF ALL_IN_ONE}
            Qall.Free;
            {$ENDIF}
            Q1.Free;
            Q1A.Free;
            Q2.Free;
            Q3.Free;
            Q3A.Free;
            Q3B.Free;
            Q4.Free;
            Q4B.Free;
            Q4C.Free;
            Q5.Free;
            Q0.Free;
            QFin := TQuery.Create(nil);
            QFin.SQLErrors := FALSE;
            TRY
                Query_raise_errors := FALSE;
                ODBC_raise_errors := FALSE;
                if drop_E then
                begin
                    QFin.SQL.Text := EndDrop('#E');
                    TRY
                        QFin.ExecSQL;
                    EXCEPT
                    END;
                end;
                if drop_D then
                begin
                    QFin.SQL.Text := EndDrop('#D');
                    TRY
                        QFin.ExecSQL;
                    EXCEPT
                    END;
                end;
                QFin.SQL.Text := EndDrop('#C');
                TRY
                    QFin.ExecSQL;
                EXCEPT
                END;
                QFin.Close;
                QFin.SQL.Text := EndDrop('#B');
                TRY
                    QFin.ExecSQL;
                EXCEPT
                END;
                QFin.Close;
                if drop_Pks then
                begin
                    ODBC_raise_errors := FALSE;
                    Query_raise_errors := FALSE;
                    QFin.SQL.Text := EndDrop('#Pks');
                    TRY
                        QFin.ExecSQL;
                    EXCEPT
                    END;
                end;
            FINALLY
                QFin.Free;
                ODBC_raise_errors := TRUE;
                Query_raise_errors := TRUE;
            END;
            //\\ListParamId.Free;
        END;
    FINALLY
        Ops.Free;
        Cache.Free;
        for i := 0 to NomenLoad.Count - 1 do
            FreeMem(NomenLoad[i]);
        NomenLoad.Free;
        {$IFDEF USE_NEWQUERY} // ODBC_DIRECT}
        DefaultODBCDataSource.CommitTrans;
        {$ELSE}
        DefaultConnection.CommitTrans;
        {$ENDIF}
    END;
end;

function LoadParams_by_ConList(const ListCon: String; const Code_Operation: String;
    AllowSum: TAllowSum = nil; AllowRaise: Boolean = FALSE;
    ParamsController: Pointer = nil; AdditionalConds: String = '';
    LoadParentParams: Boolean = FALSE; only_nomenklid: Integer = -1): TList;
begin
    Result := LoadParams_by_ConList1(FALSE, Listcon, Code_Operation, AllowSum,
        AllowRaise, 0, 0, ParamsController, AdditionalConds, LoadParentParams,
        only_nomenklid);
end;

function LoadParams_by_ConList_TranspackId(
    const ListConditions, Code_Operation: String;
    AllowSum: TAllowSum = nil; AllowRaise: Boolean = FALSE;
    DateFrom: TDateTime = 0; DateTo: TDateTime = 0;
    ParamsController: Pointer = nil;
    LoadParentParams: Boolean = FALSE): TList;
begin
    Result := LoadParams_by_ConList1(TRUE, ListConditions, Code_Operation, AllowSum,
        AllowRaise, DateFrom, DateTo, ParamsController, '', LoadParentParams);
end;

function LoadParams_by_NomList(const NomList: String; const Code_Operation: String;
    AllowSum: TAllowSum = nil; ParamsController: Pointer = nil;
    only_nomenklid: Integer = -1): TList;
//------------------------------------------------------------------------------
//    
begin
    if Trim(NomList) <> '' then
    begin
        Result := LoadParams_by_ConList('', Code_Operation,
            AllowSum, FALSE, ParamsController, ' B.nomenklid in (' + NomList + ')');
    end
    else //    ,   
    begin
        Result := LoadParams_by_ConList('', Code_Operation, AllowSum,
            FALSE, ParamsController, '', false, only_nomenklid);
    end;
end;

function LoadParams_by_BatList(const BatchList, Code_Operation: String;
    AllowSum: TAllowSum = nil; ParamsController: Pointer = nil): TList;
//------------------------------------------------------------------------------
//    
var Q: TQuery;
    ConList: String;
begin
    Q := TQuery.Create(nil);
    TRY
        Q.SQL.Text := 'select pk.packid from ' + PackageTableName + ' pk where pk.batchid in (' +
            BatchList + ') ' +
            #13'and pk.newpack is null' +
            #13'and pk.sign_exist is null' +
            #13'and pk.sign_virtual is null' +
            #13'and pk.Sign_presence is null'; //   ,  
        //     .   
        //          .
        Q.Open;
        ConList := '';
        while not Q.Eof do
        begin
            ConList := ConList + ',' + Q.SFieldN[0];
            Q.Next;
        end;
        Delete(ConList, 1, 1);
        Result := LoadParams_by_ConList(ConList, Code_Operation, AllowSum, FALSE,
            ParamsController);
    FINALLY
        Q.Free;
    END;
end;

function ParamDependOnParam(P: TParams; I, J: Integer): Boolean;
//------------------------------------------------------------------------------
// TRUE,     I      J
begin
    Result := Assigned(P.ParamData[I].param_def.FDepending) and
        (P.ParamData[I].param_def.FDepending.IndexOf(Pointer(J)) >= 0);
end;

function Compare_Batchid_Packid(const Data: Pointer; const e1, e2: Dword): Integer;
var PL: TList;
    P1, P2: TParams;
begin
    PL := Data;
    P1 := PL[e1];
    P2 := PL[e2];
    Result := P1.ParamSetData.batchid - P2.ParamSetData.batchid;
    if Result = 0 then
        Result := P1.ParamSetData.packid - P2.ParamSetData.packid;
end;

//------------------------------------------------------------------------------
//        

function CalcParams(ParamsList: TList; var Error: String; RaiseError: Boolean = FALSE;
    AllowSum: TAllowSum = nil; InitInputVals: Boolean = FALSE;
    const opname: String = ''; nomenklid: Integer = 0; Ignore_packid0: Boolean = FALSE): Extended;

    function LastChar(const s: String): Char;
    begin
        if s = '' then Result := #0
        else Result := s[Length(s)];
    end;

    function CountOfBatch(P: TParams; Idx: Integer): Integer;
    var i: Integer;
        P1: TParams;
    begin
        Result := 0;
        for i := Idx to ParamsList.Count - 1 do
        begin
            P1 := ParamsList[i];
            if P1.ParamSetData.batchid <> P.ParamSetData.batchid then break;
            inc(Result);
            if i > ParamsList.Count div 10 then break;
        end;
    end;

var CountFormulas: Integer;

    procedure CalcSingleFormula(P_idx: Integer; P: TParams; Idx: Integer);
    var fpos, L, J1: Integer;
        F: String;
        DName: String;
        I, J, ParamIdx, SumLevel, Level, precis: Integer;
        P1: TParams;
        will_sum: Boolean;
        X: Extended;
        PD, PD1: TParamData;
        P_Lst: TList;
    begin
        PD := P.ParamData[Idx];
        if (PD.param_type = 1) then
        begin
            PD.param_ready := TRUE;
            Exit;
        end;
        if PD.param_error then Exit;
        if PD.param_started_calc then
        begin
            //if Error = '' then
            Error := '   ' + P.Names[Idx];
            //Exit; //
            raise Exception.Create(Error);
        end;
        if DebugMode then
            LogFileOutPut(GetLogsPath + 'sum.log', 'CalcSingleFormula: ' +
                ' packid=' + IntToStr(P.ParamSetData.packid) +
                ' batchid=' + IntToStr(P.ParamSetData.batchid) +
                ' nomenklid=' + IntToStr(P.ParamSetData.nomenklid) +
                ' param_name=' + P.Names[Idx] +
                ' formula=' + PD.formula);
        PD.param_started_calc := TRUE;
        inc(CountFormulas);
        if CountFormulas mod 10 = 0 then
            if Assigned(OnCalcParam) then
                OnCalcParam('3 -   (' + IntToStr(CountFormulas) + ')');
        TRY
            //------------------------------------------------------------------------
            //     ,    
            //   .   ,    
            //   .
            F := RemoveSpaces(PD.formula);
            I := pos(';', F);
            if I > 0 then
                F := Trim(Copy(F, 1, I - 1));
            if F = '' then
            begin
                PD.param_ready := TRUE;
                //PD.param_OK := TRUE;
                Exit;
            end;
            fpos := 1;
            SumLevel := -1;
            Level := 0;
            while fpos <= Length(F) do
            begin
                if F[fpos] in [' ', '+', '-', '*', '?', ':', '/', '<', '>', '=', '(', ')', #1..' ', '0'..'9', DecimalSeparator] then
                begin
                    if F[fpos] = '(' then Inc(Level)
                    else
                        if F[fpos] = ')' then
                        begin
                            Dec(Level);
                            if Level <= SumLevel then
                                SumLevel := -1;
                        end;
                    Inc(fpos);
                end
                else
                    if (F[fpos] in ['s', 'S']) and
                        (CompareText(Copy(F, fpos, 4), 'sum(') = 0) then
                    begin
                        if not P.SumCached(Idx, fpos, F, PD.calc_by_batch) then
                        begin
                            SumLevel := Level;
                            Inc(Level);
                            Inc(fpos, 4);
                        end;
                    end
                    else
                    begin
                        L := fpos;
                        while L <= Length(F) do
                        begin
                            Inc(L);
                            if (L > Length(F)) or (F[L] in [' ', '+', '-', '*', ':', '?', '/', '<', '>', '=', '(', ')', #1..#31, ',', ';']) then
                            begin
                                DName := Copy(F, fpos, L - fpos);
                                if (DName <> '') and (AnsiCompareText(DName, 'OP') <> 0) and
                                    not (DName[1] in ['.', ',', DecimalSeparator, '0'..'9', '-', '+']) then
                                    if SumLevel >= 0 then //  ,  
                                        //    sum( ) -  
                                        //     .
                                    begin
                                        //{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{
                                        for ParamIdx := 0 to ParamsList.Count - 1 do
                                        begin
                                            P1 := ParamsList[ParamIdx];
                                            if P.ParamData[Idx].calc_by_batch and
                                                (nomenklid <> 0) and
                                                (P1.ParamSetData.nomenklid <> nomenklid) then continue;
                                            will_sum := TRUE;
                                            if Assigned(AllowSum) then
                                                AllowSum(P, P1, will_sum, P.ParamData[Idx].calc_by_batch);
                                            if will_sum and
                                                (P.ParamData[Idx].calc_by_batch and
                                                (P1.ParamSetData.batchid = P.ParamSetData.batchid)
                                                or
                                                not P.ParamData[Idx].calc_by_batch) then
                                            begin
                                                J := P1.HasVar(DName);
                                                if (J < 0) and (Error = '') and (P1.ParamSetData.batchid <> 0) then
                                                begin
                                                    //J := P1.HasVar( DName );
                                                    Error := ' ' + DName + '       ' +
                                                        P.Names[Idx] + ' ( ' + IntToStr(P.ParamSetData.nomenklid) + ')' +
                                                        ' -   ' + P1.ParamSetData.numcon +
                                                        IfThenElse(P.ParamSetData.Sign_final,
                                                        '()', '()');
                                                    if P.ParamSetData.packid <> 0 then
                                                        Error := Error + ',  ' + P.ParamSetData.numcon;
                                                    raise Exception.Create(Error);
                                                end;
                                                if (J >= 0) and not P1.ParamData[J].param_ready AND
                                                    not P1.ParamData[J].param_error then
                                                begin
                                                    CalcSingleFormula(ParamIdx, P1, J);
                                                    PD1 := P1.ParamData[J];
                                                    if not PD1.param_ready and (Error = '') then
                                                    begin
                                                        //if Error = '' then
                                                        Error := ' ' + DName + '    ' +
                                                            ' ' + P.Names[Idx] + ',      ' +
                                                            ', ..     ' + opname;
                                                        P1.ParamData[J].param_error := TRUE;
                                                        //Exit; //
                                                        raise Exception.Create(Error);
                                                    end;
                                                end;
                                            end;
                                        end
                                    end
                                        //}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}
                                    else //      , ..
                                        //    .
                                        if ((nomenklid = 0) or (nomenklid = P.ParamSetData.nomenklid)) {and
                                        not Ignore_packid0 or Ignore_packid0 and (P.ParamSetData.packid <> 0)}then
                                        begin
                                            if (DName <> '') and (DName[1] in ['0'..'9', DecimalSeparator, '-']) then
                                            begin
                                                J := 99999999; L := fpos;
                                            end
                                            else
                                                J := P.HasVar(DName);
                                            if (J < 0) and P.ParamData[Idx].calc_by_batch then
                                            begin
                                                //if Error = '' then
                                                begin
                                                    Error := ' ' + DName + '      ' +
                                                        P.Names[Idx] + ' ( ' + IntToStr(P.ParamSetData.nomenklid) + ')';
                                                    if P.ParamSetData.packid <> 0 then
                                                        Error := Error + ',  ' + P.ParamSetData.numcon;
                                                end;
                                                //Exit; //
                                                raise Exception.Create(Error);
                                            end;
                                            if (J <> 99999999) AND (J >= 0) AND
                                                not P.ParamData[J].param_ready and
                                                not P.ParamData[J].param_error then
                                                CalcSingleFormula(P_Idx, P, J);
                                        end;
                                fpos := L;
                                break;
                            end;
                        end;
                    end;
            end;
            //------------------------------------------------------------------------
            //     
            TRY
                //----------------------------------------------------------------------
                //  ,      sum( ... ),
                //    ,   :
                fpos := 1;
                while fpos < Length(F) do
                begin
                    if (F[fpos] in ['s', 'S']) and
                        (CompareText(Copy(F, fpos, 4), 'sum(') = 0) then
                    begin
                        J := fpos; SumLevel := 0;
                        while J < Length(F) do
                        begin
                            CASE F[J] OF
                                '(': Inc(SumLevel);
                                ')': begin
                                        Dec(SumLevel);
                                        if SumLevel = 0 then break;
                                    end;
                            END;
                            Inc(J);
                        end;
                        PD := P.ParamData[Idx];

                        if PD.calc_by_batch and (CountOfBatch(P, P_idx) < ParamsList.Count div 10) then
                        begin
                            P_Lst := TList.Create;
                            TRY
                                for J1 := P_Idx to ParamsList.Count - 1 do
                                begin
                                    P1 := ParamsList[J1];
                                    if P1.ParamSetData.batchid <> P.ParamSetData.batchid then
                                        break;
                                    P_Lst.Add(ParamsList[J1]);
                                end;
                                P.DoCacheSum(Idx, fpos, F,
                                    CalcExpression(Copy(F, fpos, J + 1 - fpos),
                                    P, P_Lst, AllowSum, FALSE),
                                    TRUE {PD.calc_by_batch});
                            FINALLY
                                P_Lst.Free;
                            END;
                        end
                        else
                        begin
                            P.DoCacheSum(Idx, fpos, F,
                                CalcExpression(Copy(F, fpos, J + 1 - fpos),
                                P, ParamsList, AllowSum, FALSE,
                                PD.calc_by_batch), PD.calc_by_batch);
                        end;

                        if P.SumCached(Idx, fpos, F, PD.calc_by_batch) then
                        begin
                            fpos := fpos - 1;
                            //PD.param_ready := TRUE;  //|||\\
                        end;
                    end;
                    Inc(fpos);
                end;

                //----------------------------------------------------------------------
                //    
                P.Values[Idx] := CalcExpression(F, P, ParamsList, AllowSum,
                    P.ParamData[Idx].calc_by_batch);
                precis := PD.precision;
                if precis < MAX_PRECISION then
                begin
                    X := P.Values[Idx];
                    X := Round_Fmt(Power(10, precis) * X)
                        /
                        Power(10, precis);
                    P.Values[Idx] := X;
                end;
                PD.param_ready := TRUE;
                //PD.param_OK := TRUE;
            EXCEPT on E: Exception do
                begin
                    if (opname <> '') and (PD.Is_used_in_op(opname) < 0) or
                        (nomenklid <> 0) and (P.ParamSetData.nomenklid <> nomenklid) or
                        P.ParamSetData.ignore_conditions then
                    begin
                        Error := '';
                        E.Message := '';
                        PD.param_ready := TRUE;
                        Exit; //      -   
                    end;
                    if (E is EFormula_Condition) and (P.Descriptions[Idx] <> '') then
                    begin
                        if (P.ParamSetData.nomenklid <> nomenklid) and
                            (nomenklid <> 0) then
                        begin
                            Error := '';
                            E.Message := '';
                            PD.param_ready := TRUE;
                            Exit;
                        end;
                        Error := P.Descriptions[Idx] + ': ' + E.Message +
                            ' : ' + P.ParamSetData.numcon +
                            '  .: ' + IntToStr(P.ParamSetData.nomenklid);
                        E.Message := Error;
                        PD.param_error := TRUE;
                    end
                    else
                        Error := E.Message;
                    raise;
                end;
            END;
        FINALLY
            PD.param_started_calc := FALSE;
            if PD.save_formula <> '' then
            begin
                PD.formula := TrimLeft(PD.save_formula);
                PD.save_formula := '';
            end;
            if DebugMode then
                LogFileOutPut(GetLogsPath + 'sum.log', 'CalcSingleFormula: ' +
                    ' Result=' + FloatToStr(P.Values[Idx]));
        END;
    end;

var OrderList: TList; //  :    TParams  
    // ParamsList -      
    I, J, K, N, MinDep, I1: Integer;
    Added: Boolean;
    P, P1: TParams;
    PD, PD1: TParamData;
    ParamOrderList: TList; //   - (
    //   TParams   ParamsList)
    ParamOrderListList: TList;
    FirstError: String;
    TimeStart: DWORD;
begin
    Result := 0;
    Error := '';
    OrderList := TList.Create;
    ParamOrderListList := TList.Create;
    TRY
        TimeStart := GetTickCount;
        if Assigned(OnCalcParam) then
            OnCalcParam('1 -   ');
        //--------------------------------------------------------------------------
        //      
        for I := 0 to ParamsList.Count - 1 do
        begin
            P := ParamsList[I];
            for K := 0 to P.Count - 1 do
            begin
                PD := P.ParamData[K];
                if (PD.Is_used_in_op(opname) < 0) or
                    (nomenklid <> 0) and (P.ParamSetData.nomenklid <> nomenklid) then
                begin
                    PD.param_ready := TRUE;
                end
                else
                    if (PD.param_type = 2) or
                        (PD.param_type = 3) and
                        not PD.param_ready then
                    begin
                        PD.param_ready := FALSE;
                        PD.param_error := FALSE;
                    end;
            end;
        end;
        if Assigned(OnCalcParam) then
            OnCalcParam('2 -  ');
        SortData(ParamsList, ParamsList.Count, Compare_Batchid_Packid, SwapItems);
        for I := 0 to ParamsList.Count - 1 do
        begin
            //------------------------------------------------------------------------
            //         
            //    OrderList   .
            P := ParamsList[I];
            {if (nomenklid <> 0) and (P.ParamSetData.nomenklid <> 0) and (P.ParamSetData.nomenklid <> nomenklid) then
            begin
              OrderList.Add( nil );
              continue;
            end;}
            if (I > 0) and (P.DefsTheSame(ParamsList[I - 1])) then
            begin
                OrderList.Add(OrderList[I - 1]);
            end
            else
            begin
                ParamOrderList := TList.Create;
                ParamOrderListList.Add(ParamOrderList);
                OrderList.Add(ParamOrderList);
                //if Ignore_packid0 and (P.ParamSetData.packid = 0) then continue;
                P.BuildDependings;
                //------------------------------------------------------------------------
                //       P: TParams   
                //     : Param1 < Param2,  Param2 
                //  Param1.
                for J := 0 to P.Count - 1 do
                begin
                    Added := FALSE;
                    MinDep := 0;
                    for K := 0 to ParamOrderList.Count - 1 do
                    begin
                        if ParamDependOnParam(P, Integer(ParamOrderList[K]), J) then
                        begin
                            Added := TRUE;
                            ParamOrderList.Insert(K, Pointer(J));
                            break;
                        end;
                        if ParamDependOnParam(P, J, Integer(ParamOrderList[K])) then
                            MinDep := K;
                    end;
                    if not Added then
                    begin //----------------------------------------------------------------
                        //   ,      J.
                        // :   J   
                        if MinDep >= ParamOrderList.Count then
                            ParamOrderList.Add(Pointer(J))
                        else
                            ParamOrderList.Insert(MinDep + 1, Pointer(J));
                    end;
                    K := ParamOrderList.IndexOf(Pointer(J));
                    //----------------------------------------------------------------------
                    //   ,        K+1,
                    //      .     ,
                    //          , 
                    //   .
                    PD := P.ParamData[J];
                    for N := K + 1 to ParamOrderList.Count - 1 do
                        if (PD.save_formula = '') and
                            ParamDependOnParam(P, J, Integer(ParamOrderList[N])) and
                            (P.Values[Integer(ParamOrderList[N])] = 0)
                        and ParamDependOnParam(P, Integer(ParamOrderList[N]), J)
                            then
                        begin
                            if Error = '' then
                                Error := ' ' + P.Names[J] + '  ' +
                                    P.Names[Integer(ParamOrderList[N])] + '   ' +
                                    '    .';
                            if (opname <> '') and (PD.Is_used_in_op(opname) < 0) or
                                (nomenklid <> 0) and (P.ParamSetData.nomenklid <> nomenklid) then
                            else
                            begin
                                if RaiseError then
                                    raise Exception.Create(Error)
                                else
                                    Exit;
                            end;
                        end;
                end;
            end;
        end;
        TimePrepare := TimePrepare + GetTickCount - TimeStart;
        TimeStart := GetTickCount;
        //--------------------------------------------------------------------------
        //          .
        FirstError := '';
        if Assigned(OnCalcParam) then
            OnCalcParam('3 -  ');
        TRY
            for I := 0 to ParamsList.Count - 1 do
            begin
                if CancelCalcParams then break;
                P := ParamsList[I];
                if (nomenklid <> 0) and (P.ParamSetData.nomenklid <> nomenklid) then
                    continue;
                if P.ParamSetData.NoCalc then
                begin
                    asm
            nop
                    end;
                    continue;
                end;
                //if Ignore_packid0 and (P.ParamSetData.packid = 0) then continue;
                ParamOrderList := OrderList[I];
                if ParamOrderList = nil then
                    asm
          nop
                    end;
                for J := 0 to ParamOrderList.Count - 1 do
                begin
                    K := Integer(ParamOrderList[J]);
                    PD := P.ParamData[K];
                    if (PD.param_type = 2) or
                        (PD.param_type = 3) then
                        if {PD.recalc and} not PD.param_ready and not PD.param_error then
                        TRY
                            CalcSingleFormula(I, P, K);
                            if PD.calc_by_batch and
                                (pos('sum(', RemoveSpaces(LowerCase(PD.formula))) = 1) and
                                (LastChar(RemoveSpaces(PD.formula)) = ')') then
                            begin
                                for I1 := I + 1 to ParamsList.Count - 1 do
                                begin
                                    P1 := ParamsList[I1];
                                    if (nomenklid <> 0) and (P1.ParamSetData.nomenklid <> nomenklid) then
                                        break;
                                    if P1.ParamSetData.NoCalc then
                                        continue;
                                    if P1.ParamSetData.batchid = P.ParamSetData.batchid then
                                    begin
                                        {Assert( P1.Names[ K ] = P.Names[ K ], '  !' +
                                          '<' + P1.Names[ K ] + ' ' + P.Names[ K ] + '>' );}
                                        if P1.Names[K] <> P.Names[K] then continue;
                                        PD1 := P1.ParamData[K];
                                        PD1.param_ready := TRUE;
                                        ///////////////PD1.recalc := FALSE;
                                        P1.Values[K] := P.Values[K];
                                    end else break;
                                end;
                            end
                            else
                                if not PD.calc_by_batch and
                                    (pos('sum(', RemoveSpaces(LowerCase(PD.formula))) = 1) and
                                    (LastChar(RemoveSpaces(PD.formula)) = ')') then
                                begin
                                    for I1 := I + 1 to ParamsList.Count - 1 do
                                    begin
                                        P1 := ParamsList[I1];
                                        //if (nomenklid <> 0) and (P1.ParamSetData.nomenklid <> nomenklid) then
                                        //   break;
                                        if P1.ParamSetData.NoCalc then
                                            continue;
                                        if P1.Names[K] <> P.Names[K] then break;
                                        //if P1.ParamSetData.batchid = P.ParamSetData.batchid then
                                        begin
                                            //Assert( P1.Names[ K ] = P.Names[ K ], '  !' +
                                            //  P1.Names[ K ] + ' ' + P.Names[ K ] );
                                            PD1 := P1.ParamData[K];
                                            PD1.param_ready := TRUE;
                                            ///////////////PD1.recalc := FALSE;
                                            P1.Values[K] := P.Values[K];
                                        end; // else break;
                                    end;
                                end;
                        EXCEPT on E: Exception do
                            begin
                                if (opname <> '') and (PD.Is_used_in_op(opname) < 0) or
                                    (nomenklid <> 0) and (P.ParamSetData.nomenklid <> nomenklid) or
                                    P.ParamSetData.ignore_conditions then
                                begin
                                    //      -   
                                    Error := '';
                                    E.Message := '';
                                    PD.param_ready := TRUE;
                                end
                                else
                                begin
                                    PD.param_error := TRUE;
                                    FirstError := E.Message;
                                    //Error := '';
                                end;
                            end;
                        END;
                end;
            end;
            {}
            if FirstError <> '' then
            begin
                Error := FirstError;
                raise Param_Exception.Create(FirstError);
            end;
            {}
            if Error <> '' then Exit;
        EXCEPT on E: Exception do
            begin
                E.Message := Error;
                //    , Error    
                if RaiseError then raise;
            end;
        END;
        TimeCalc := GetTickCount - TimeStart;
    FINALLY
        if Assigned(OnCalcParam) then
            if CancelCalcParams then
                OnCalcParam('5 - ')
            else
                OnCalcParam('4 - ');
        for I := 0 to ParamOrderListList.Count - 1 do
            TObject(ParamOrderListList[I]).Free;
        OrderList.Free;
        ParamOrderListList.Free;
    END;
end;

{ TParams }

function TParams.Add(const Name: String; nomenklid: Integer; Value: Extended; ID: Integer;
    const Description: String; const op: String; const A_copy_from: String;
    Is_Batch_Analyses: Boolean): Pointer;
var PData: TParamData;
    Def: TParamDef;
begin
    Result := inherited Add(Name, Value, ID, Description, op);
    if FParamsController = nil then Exit;
    Def := FParamsController.Def(ID, Name, {FParamSetData.} nomenklid, op);
    PData := TParamData.Create;
    PData.param_def := Def;
    Def.Analyse_copy_from_param := A_copy_from;
    Def.Is_Batch_Analyse := Is_Batch_Analyses;
    FParamData.Add(PData);
    Result := PData;
end;

procedure TParams.Assign(From: TVars);
var F: TParams;
    I: Integer;
    PD: TParamData;
begin
    if From <> Self then
        Clear;
    inherited;
    if From = Self then
    begin
        for I := 0 to Count - 1 do
        begin
            PD := ParamData[I];
            PD.Assign(PD);
        end;
        Exit;
    end;
    if not (From is TParams) then Exit;
    F := From as TParams;
    ParamSetData.Assign(F.ParamSetData);
    for I := 0 to F.Count - 1 do
    begin
        PD := TParamData.Create;
        FParamData.Add(PD);
        PD.Assign(F.ParamData[I]);
    end;
end;

//{$DEFINE DEP_LOG}

procedure TParams.BuildDependings;
var I, J, L: Integer;
    F, N: String;
    fpos: Integer;
    {$IFDEF DEP_LOG}
    SL: TStringList;
    {$ENDIF}
begin
    for I := 0 to Count - 1 do
        if Assigned(ParamData[I].param_def.FDepending) then
            ParamData[I].param_def.FDepending.Clear;
    {$IFDEF DEP_LOG}
    SL := TStringList.Create;
    {$ENDIF}
    for I := 0 to Count - 1 do
    begin
        if ((ParamData[I].param_type = 2) or
            (ParamData[I].param_type = 3)
            ) and {ParamData[ I ].recalc and} not ParamData[I].param_ready then
        begin
            fpos := 1;
            F := RemoveSpaces(ParamData[I].formula);
            J := pos(';', F);
            if J > 0 then
                F := Trim(Copy(F, 1, J - 1));
            while fpos <= Length(F) do
            begin
                if (F[fpos] in [' ', '0'..'9', '+', '-', '<', '>', '=', '*', '/', '<', '>', '=', '(', ')', #1..#31, DecimalSeparator]) then
                    Inc(fpos)
                else
                    if (F[fpos] in ['s', 'S']) and (CompareStr(Copy(F, fpos, 4), 'str(') = 0) then
                        Inc(fpos, 4)
                    else
                    begin
                        for L := fpos + 1 to Length(F) + 1 do
                            if (L > Length(F)) or (F[L] in [' ', '-', '+', '*', '/', '<', '>', '=', '(', ')']) then
                            begin
                                N := Copy(F, fpos, L - fpos);
                                for J := 0 to Count - 1 do
                                begin
                                    if N = Names[J] then
                                    begin
                                        if not Assigned(ParamData[I].param_def.FDepending) then
                                            ParamData[I].param_def.FDepending := TList.Create;
                                        ParamData[I].param_def.FDepending.Add(Pointer(J));
                                        {$IFDEF DEP_LOG}
                                        SL.Add(' ' + F + '   ' + N);
                                        {$ENDIF}
                                    end;
                                end;
                                fpos := L;
                                break;
                            end;
                    end;
            end;
        end;
    end;
    {$IFDEF DEP_LOG}
    SL.SaveToFile('C:\dep.log');
    SL.Free;
    {$ENDIF}
end;

function TParams.calc_by_batch(Idx: Integer): Boolean;
begin
    Result := ParamData[Idx].calc_by_batch;
end;

procedure TParams.Clear;
//------------------------------------------------------------------------------
// .   .
var I: Integer;
begin
    inherited;
    if Assigned(FParamData) then
    begin
        for I := 0 to FParamData.Count - 1 do
            TObject(FParamData[I]).Free;
        FParamData.Clear;
    end;
end;

constructor TParams.Create;
//------------------------------------------------------------------------------
// .      .
begin
    inherited;
    FParamData := TList.Create;
    FParamSetData := TParamSetData.Create;
end;

constructor TParams.CreateOwned(Owner: TParamsSet);
begin
    Create;
    FParamsController := Owner;
end;

destructor TParams.Destroy;
//------------------------------------------------------------------------------
// .  .
begin
    Clear;
    FreeAndNil(FParamData);
    FreeAndNil(FParamSetData);
    inherited;
end;

procedure TParams.DoCacheSum(Idx, fPos: Integer; const F: String; Value: Double;
    calc_by_batch: Boolean);
//------------------------------------------------------------------------------
//      
begin
    if FParamsController <> nil then
        FParamsController.DoSumCache(Self, Idx, fPos, F, Value, calc_by_batch);
end;

function TParams.FormattedParameter(Idx: Variant): String;
//------------------------------------------------------------------------------
//  ,     
//  ,    (  Param 
// Operation_params,      ).
var I: Integer;
    pf: Double;
    fmt: String;
    PD: TParamData;
begin
    if VarType(Idx) = varString then
        I := IndexOf(Idx)
    else
        I := Idx;
    PD := ParamData[I];
    if PD.param_format = 0 then
        Result := Format('%f', [Values[I]])
    else
    begin
        pf := PD.param_format;
        fmt := Format('%f', [pf]);
        fmt := StringReplace(fmt, DecimalSeparator, '.', []);
        while (Length(fmt) > 0) and (fmt[Length(fmt)] = '0') do
            System.Delete(fmt, Length(fmt), 1);
        TRY
            Result := Trim(Format('%' + fmt + 'f', [Values[I]]));
        EXCEPT
            Result := Format('%f', [Values[I]]);
        END;
    end;
end;

function TParams.GetCapacity: Integer;
begin
    Result := FParamData.Capacity;
end;

function TParams.GetParamData(Idx: Variant): TParamData;
//------------------------------------------------------------------------------
//      
var I: Integer;
begin
    if VarType(Idx) = varString then
        I := IndexOf(Idx)
    else
        I := Idx;
    if (I >= Count) then
    begin
        if Assigned(FParamsController) and
            Assigned(FParamsController.ConstParams) and
            Assigned(FParamsController.ConstParam[nomenklid]) and
            (I - Count < FParamsController.ConstParam[nomenklid].Count) then
        begin
            Result := FParamsController.ConstParam[nomenklid].ParamData[I - Count];
            Exit;
        end;
        {asm
          int 3
        end;}
    end;
    if I < Count then
        Result := FParamData[I]
    else
        Result := nil;
end;

function TParams.OK: Boolean;
//         
var I: Integer;
    PD: TParamData;
begin
    for I := 0 to Count - 1 do
    begin
        PD := ParamData[I];
        if (PD.param_type > 1) and not PD.param_ready then
        begin
            Result := FALSE;
            Exit;
        end;
    end;
    Result := TRUE;
end;

function TParams.DefsTheSame(P: TParams): Boolean;
var i: Integer;
begin
    Result := Count = P.Count;
    if not Result then Exit;
    for i := 0 to Count - 1 do
    begin
        Result := ParamData[i].param_def = P.ParamData[i].param_def;
        if not Result then Exit;
    end;
end;

procedure TParams.SetCapacity(const Value: Integer);
begin
    if Value > Capacity then
    begin
        FParamData.Capacity := Value;
        FNames.Capacity := Value;
        FDescriptions.Capacity := Value;
        FValues.Capacity := Value;
    end;
end;

function TParams.SumCached(Idx, fPos: Integer; var F: String; calc_by_batch: Boolean): Boolean;
//------------------------------------------------------------------------------
// ,       F, 
//   fPos:
begin
    Result := FALSE;
    if FParamsController = nil then Exit;
    Result := FParamsController.SumCached(Self, Idx, fPos, F, calc_by_batch);
end;

procedure TParams.SwapItems(i1, i2: Integer);
//------------------------------------------------------------------------------
//   
var TmpD: Pointer;
begin
    inherited;
    TmpD := FParamData[i1];
    FParamData[i1] := FParamData[i2];
    FParamData[i2] := TmpD;
end;

procedure TParams.WhenSummed(ToWhat: TVars; const AddText: String);
begin
    {$IFDEF DEBUG_SUM}
    if DebugMode then
        LogFileOutPut(GetLogsPath + 'sum.log', IntToStr(Integer(Self)) +
            ' packid=' + IntToStr(ParamSetData.packid) +
            ' batchid=' + IntToStr(ParamSetData.batchid) +
            ' added to ' +
            IntToStr(Integer(ToWhat)) +
            ' batchid=' + IntToStr((ToWhat as TParams).ParamSetData.batchid) +
            ' ' + AddText);
    {$ENDIF}
end;

function TParams.GetValues(IdxOrName: Variant): Extended;
//------------------------------------------------------------------------------
//   
// :   ,   
//        ,   
//    (  ?)
var I: Integer;
    ConstParamsSet: TParams;
begin
    if VarType(IdxOrName) = varString then
    begin
        I := IndexOf(IdxOrName);
        if (I < 0) and (FParamsController.ConstParams <> nil) and
            (FParamsController.ConstParam[nomenklid] <> Self) then
        begin
            //Result := 0;
            ConstParamsSet := FParamsController.ConstParam[nomenklid];
            if ConstParamsSet <> nil then
                Result := ConstParamsSet.Values[IdxOrName]
            else
                raise EFormula_NoVar.Create('   ' + IdxOrName);
            Exit;
        end
        else
            if I >= 0 then
                Result := inherited GetValues(I)
            else
                raise EFormula_NoVar.Create('   ' + IdxOrName +
                    ' ( : ' + IntToStr(nomenklid) + ')');
        Exit;
    end
    else
    begin
        I := IdxOrName;
        if I >= Count then
        begin
            if Assigned(FParamsController) and
                Assigned(FParamsController.ConstParams) then
            begin
                ConstParamsSet := FParamsController.ConstParam[nomenklid];
                if ConstParamsSet <> nil then
                begin
                    if I - Count < ConstParamsSet.Count then
                    begin
                        Result := ConstParamsSet.Values[I - Count];
                        Exit;
                    end;
                end;
            end;
            Result := 0;
            {asm
              int 3
            end;}
            Exit;
        end;
    end;
    Result := inherited GetValues(IdxOrName);
    {  I := IdxOrName;
    if (I < 0) or (I >= Count) then
      raise EFormula_NoVar.Create( ' ' + IdxOrName + '  .' );
    Result := FValues[ I ];}
end;

function TParams.HasVar(const Name: String): Integer;
var i: Integer;
    ConstParamsSet: TParams;
begin
    Result := IndexOf(Name);
    if Result >= 0 then Exit;
    if Assigned(FParamsController) and
        Assigned(FParamsController.ConstParams) then
    begin
        for i := 0 to FParamsController.ConstParams.Count - 1 do
        begin
            ConstParamsSet := FParamsController.ConstParams[i];
            if nomenklid = ConstParamsSet.nomenklid then
            begin
                Result := Count + ConstParamsSet.IndexOf(Name);
                Exit;
            end;
        end;
    end;
end;

function TParams.AddConst(const Name: String; nomenklid: Integer; Value: Extended;
    ID: Integer; const Description, op: String): Pointer;
var i: Integer;
    ConstParamsSet: TParams;
begin
    if Assigned(FParamsController) then
    begin
        if not Assigned(FParamsController.ConstParams) then
            FParamsController.ConstParams := TList.Create;
        ConstParamsSet := nil;
        for i := 0 to FParamsController.ConstParams.Count - 1 do
        begin
            ConstParamsSet := FParamsController.ConstParams[i];
            if ConstParamsSet.nomenklid = nomenklid then break;
            ConstParamsSet := nil;
        end;
        if ConstParamsSet = nil then
        begin
            ConstParamsSet := TParams.Create;
            ConstParamsSet.ParamSetData.nomenklid := nomenklid;
            ConstParamsSet.FParamsController := FParamsController;
            FParamsController.ConstParams.Add(ConstParamsSet);
        end;
        if ConstParamsSet.IndexOf(Name) >= 0 then
        begin
            Result := ConstParamsSet.ParamData[ConstParamsSet.IndexOf(Name)];
            Exit;
        end;
        Result := ConstParamsSet.Add(Name, nomenklid, Value, ID, Description,
            op, '', FALSE);
        Exit;
    end;
    Result := Add(Name, nomenklid, Value, ID, Description, op, '', FALSE);
end;

function TParams.ConstCount: Integer;
begin
    Result := 0;
    if Assigned(FParamsController) and
        Assigned(FParamsController.ConstParams) and
        Assigned(FParamsController.ConstParam[nomenklid]) then
        Result := FParamsController.ConstParam[nomenklid].Count;
end;

function TParams.GetDescr(Idx: Integer): String;
var I: Integer;
    ConstParamsSet: TParams;
begin
    I := Idx;
    if I >= Count then
    begin
        if Assigned(FParamsController) and
            Assigned(FParamsController.ConstParams) then
        begin
            ConstParamsSet := FParamsController.ConstParam[nomenklid];
            if ConstParamsSet <> nil then
            begin
                if I - Count < ConstParamsSet.Count then
                begin
                    Result := ConstParamsSet.Descriptions[I - Count];
                    Exit;
                end;
            end;
        end;
        Result := '';
        {asm
          int 3
        end;}
        Exit;
    end;
    Result := inherited GetDescr(Idx);
end;

function TParams.GetNames(Idx: Integer): String;
var I: Integer;
    ConstParamsSet: TParams;
begin
    I := Idx;
    if I >= Count then
    begin
        if Assigned(FParamsController) and
            Assigned(FParamsController.ConstParams) then
        begin
            ConstParamsSet := FParamsController.ConstParam[nomenklid];
            if ConstParamsSet <> nil then
            begin
                if I - Count < ConstParamsSet.Count then
                begin
                    Result := ConstParamsSet.Names[I - Count];
                    Exit;
                end;
            end;
        end;
        Result := '';
        {asm
          int 3
        end;}
        Exit;
    end;
    Result := inherited GetNames(Idx);
end;

function TParams.GetID(IdxOrName: Variant): Integer;
var I: Integer;
    ConstParamsSet: TParams;
begin
    Result := 0;
    if VarType(IdxOrName) = varString then
    begin
        I := IndexOf(IdxOrName);
        if (I < 0) and (FParamsController.ConstParams <> nil) then
        begin
            ConstParamsSet := FParamsController.ConstParam[nomenklid];
            if ConstParamsSet <> nil then
                Result := ConstParamsSet.GetID(IdxOrName);
        end;
    end
    else
    begin
        I := IdxOrName;
        if I < Count then
            Result := inherited GetID(I)
        else
            if I - Count < ConstCount then
                Result := FParamsController.ConstParam[nomenklid].GetID(I - Count);
    end;
end;

function TParams.GetNomenklid: Integer;
begin
    Result := ParamSetData.nomenklid;
end;

function TParams.findParamByParamNameAndNomenklid(
    const AParam_name: String; Anomenklid: Integer): Integer;
var i: Integer;
    PD: TParamData;
begin
    for i := 0 to FParamData.Count - 1 do
    begin
        PD := FParamData[i];
        if (PD.param_def.nomenklid = Anomenklid) and
            (PD.param_def.Param_name = AParam_name) then
        begin
            Result := i; Exit;
        end;
    end;
    Result := -1;
end;

procedure TParams.DebugOutput(const filename: String);
var I: Integer;
    PD: TParamData;
    SL: TStringList;
begin
    SL := TStringList.Create;
    for I := 0 to Count - 1 do
    begin
        PD := ParamData[I];
        SL.Add(PD.Param_name + '=' + FloatToStr(Values[I]));
    end;
    SL.SaveToFile(filename);
    SL.Free;
end;

{ TParamData }

procedure TParamData.Assign(From: TParamData);
begin
    if Self <> From then
        param_def := From.param_def;
end;

constructor TParamData.Create;
begin
    inherited Create;
    //calc_by_batch := TRUE;
end;

destructor TParamData.Destroy;
begin
    inherited;
end;

function TParamData.formula_empty: Boolean;
begin
    Result := param_def.formula_empty;
end;

function TParamData.Get_calc_by_batch: Boolean;
begin
    Result := param_def.calc_by_batch;
end;

function TParamData.get_formula: String;
begin
    Result := param_def.formula;
end;

function TParamData.get_nomenklid: Integer;
begin
    Result := param_def.nomenklid;
end;

function TParamData.get_param_format: Double;
begin
    Result := param_def.param_format;
end;

function TParamData.get_param_orders(idx: Integer): Integer;
begin
    Result := param_def.param_orders[idx];
end;

function TParamData.get_param_type: Integer;
begin
    Result := param_def.param_type;
end;

function TParamData.get_precision: Integer;
begin
    Result := param_def.precision;
end;

function TParamData.get_read_only: Boolean;
begin
    Result := param_def.read_only;
end;

function TParamData.Get_recalc: Boolean;
begin
    Result := param_def.recalc;
end;

function TParamData.get_show_in_ops(idx: Integer): Byte;
begin
    Result := param_def.show_in_ops[idx];
end;

function TParamData.get_Units: String;
begin
    Result := param_def.Units;
end;

function TParamData.get_used_in_ops(idx: Integer): Byte;
begin
    Result := param_def.used_in_ops[idx];
end;

function TParamData.Is_show_in_op(const op: String): Integer;
begin
    Result := param_def.Is_show_in_op(op);
end;

function TParamData.Is_used_in_op(const op: String): Integer;
begin
    Result := param_def.Is_used_in_op(op);
end;

function TParamData.Param_name: String;
begin
    Result := param_def.Param_name;
end;

function TParamData.param_ord_count: Byte;
begin
    Result := param_def.param_ord_count;
end;

procedure TParamData.Set_calc_by_batch(const Value: Boolean);
begin
    param_def.calc_by_batch := Value;
end;

procedure TParamData.set_formula(const Value: String);
begin
    param_def.formula := Value;
end;

procedure TParamData.set_param_format(const Value: Double);
begin
    param_def.param_format := Value;
end;

procedure TParamData.set_param_type(const Value: Integer);
begin
    param_def.param_type := Value;
end;

procedure TParamData.set_precision(const Value: Integer);
begin
    param_def.precision := Value;
end;

procedure TParamData.set_read_only(const Value: Boolean);
begin
    param_def.read_only := Value;
end;

procedure TParamData.Set_recalc(const Value: Boolean);
begin
    param_def.recalc := Value;
end;

procedure TParamData.Set_show_in_op(const op: String);
begin
    param_def.Set_show_in_op(op);
end;

procedure TParamData.set_Units(const Value: String);
begin
    param_def.Units := Value;
end;

procedure TParamData.Set_used_in_op(const op: String; op_param_order: Integer);
begin
    param_def.Set_used_in_op(op, op_param_order);
end;

function TParamData.show_in_op_count: Byte;
begin
    Result := param_def.show_in_op_count;
end;

function TParamData.used_in_op_count: Byte;
begin
    Result := param_def.used_in_op_count;
end;

{ TParamSetData }

procedure TParamSetData.Assign(From: TParamSetData);
begin
    nomenklid := From.nomenklid;
    packid := From.packid;
    numcon := From.numcon;
    batchid := From.batchid;
    Sign_final := From.Sign_final;
    opid := From.opid;
    code_operation := From.Code_Operation;
    code_trans := From.code_trans;
end;

{ TParamsSet }

procedure TParamsSet.Clear(ClearAll: Boolean);
var i: Integer;
    ConstParamsSet: TVars;
begin
    if ClearAll then
    begin
        for i := 0 to ParamsDefs.Count - 1 do
            TObject(ParamsDefs[i]).Free;
        ParamsDefs.Clear;
    end;
    if Assigned(ConstParams) then
    begin
        for i := 0 to ConstParams.Count - 1 do
        begin
            ConstParamsSet := ConstParams[i];
            ConstParamsSet.Free;
        end;
        ConstParams.Free;
        ConstParams := nil;
    end;
end;

constructor TParamsSet.Create(AOwner: TComponent);
begin
    inherited Create(AOwner);
    ParamsDefs := TList.Create;
end;

function TParamsSet.Def(paramid: Integer; const param_name: String;
    nomenklid: Integer; const op: String; AddIfNotExists: Boolean): TParamDef;
var i: Integer;
    Def: TParamDef;
begin
    Result := nil;
    if ParamsDefs = nil then Exit;
    for i := 0 to ParamsDefs.Count - 1 do
    begin
        Def := ParamsDefs[i];
        if (Def.paramid = paramid) and
            ((nomenklid = 0) or (nomenklid <> 0) and (Def.nomenklid = nomenklid)) and
            ((Result = nil) or
            (Def.nomenklid = nomenklid) and
            ((op = '') or (Def.Is_used_in_op(op) >= 0))
            ) then
            Result := Def;
    end;
    if (Result = nil) and AddIfNotExists then
    begin
        Def := TParamDef.Create;
        ParamsDefs.Add(Def); Result := Def;
        Def.paramid := paramid;
        Def.param_name := param_name;
        Def.nomenklid := nomenklid;
    end;
end;

destructor TParamsSet.Destroy;
var i: Integer;
begin
    for i := 0 to ParamsDefs.Count - 1 do
        TObject(ParamsDefs[i]).Free;
    ParamsDefs.Free;
    ConstParams.Free;
    inherited;
end;

procedure TParamsSet.DoSumCache(P: TParams; Idx, fPos: Integer;
    const F: String; Value: Double; calc_by_batch: Boolean);
begin
    //
end;

function TParamsSet.GetConstParam(Anomenklid: Integer): TParams;
var i: Integer;
    ConstParamsSet: TParams;
begin
    Result := nil;
    if not Assigned(ConstParams) then Exit;
    for i := 0 to ConstParams.Count - 1 do
    begin
        ConstParamsSet := ConstParams[i];
        if ConstParamsSet.nomenklid = Anomenklid then
        begin
            Result := ConstParamsSet;
            Exit;
        end;
    end;
end;

function TParamsSet.GetConstParamsByIdx(idx: Integer): TParams;
begin
    Result := nil;
    if Assigned(ConstParams) and (ConstParams.Count > idx) then
        Result := ConstParams[idx];
end;

function TParamsSet.GetConstParamsCount: Integer;
begin
    Result := 0;
    if not Assigned(ConstParams) then Exit;
    Result := ConstParams.Count;
end;

procedure TParamsSet.SetConstParamsByIdx(idx: Integer;
    const Value: TParams);
var Old, NewVal: TParams;
begin
    if not Assigned(ConstParams) then
        ConstParams := TList.Create;
    while ConstParams.Count <= idx do
        ConstParams.Add(nil);
    Old := ConstParams[idx];
    if Old <> nil then Old.Free;
    NewVal := TParams.Create;
    NewVal.Assign(Value);
    ConstParams[idx] := NewVal;
end;

function TParamsSet.SumCached(P: TParams; Idx, fPos: Integer; var F: String;
    calc_by_batch: Boolean): Boolean;
begin
    Result := FALSE;
end;

{ TList1 }

procedure TList1.Grow;
begin
    if Capacity >= 64 then
        inherited
    else
        if Capacity < 10 then
            SetCapacity(10)
        else
            SetCapacity(Capacity * 2);
end;

{ TParamDef }

procedure TParamDef.Assign(From: TParamDef);
begin
    if Self <> From then
    begin
        Units := From.Units;
        param_type := From.param_type;
        param_format := From.param_format;
        precision := From.precision;
        formula := From.formula;
        read_only := From.read_only;
        used_in_ops := From.used_in_ops;
        used_in_op_count := From.used_in_op_count;

        param_orders := From.param_orders;

        show_in_ops := From.show_in_ops;
        show_in_op_count := From.show_in_op_count;
        calc_by_batch := From.calc_by_batch;
        recalc := From.recalc;
    end;
end;

constructor TParamDef.Create;
begin
    inherited Create;
    calc_by_batch := TRUE;
end;

destructor TParamDef.Destroy;
begin
    FDepending.Free;
    inherited;
end;

function TParamDef.formula_empty: Boolean;
var I: Integer;
begin
    I := pos(';', formula);
    if I > 0 then Result := Trim(Copy(formula, 1, I - 1)) = ''
    else
        Result := Trim(formula) = '';
end;

function TParamDef.Is_show_in_op(const op: String): Integer;
var i, j: Integer;
begin
    Result := -1;
    if not Assigned(Operations) then Exit;
    i := Operations.IndexOf(op);
    if i < 0 then Exit;
    for j := 0 to show_in_op_count - 1 do
    begin
        if show_in_ops[j] = i then
        begin
            Result := j;
            break;
        end;
    end;
end;

function TParamDef.Is_used_in_op(const op: String): Integer;
var i, j: Integer;
begin
    Result := -1;
    if not Assigned(Operations) then Exit;
    i := Operations.IndexOf(op);
    if i < 0 then Exit;
    for j := 0 to used_in_op_count - 1 do
    begin
        if used_in_ops[j] = i then
        begin
            Result := j;
            break;
        end;
    end;
end;

function TParamDef.param_ord_count: Byte;
begin
    Result := used_in_op_count;
end;

procedure TParamDef.Set_show_in_op(const op: String);
var i, j: Integer;
begin
    i := Operations.IndexOf(op);
    if i < 0 then
        i := Operations.Add(op);
    for j := 0 to show_in_op_count - 1 do
        if show_in_ops[j] = i then Exit;
    if show_in_op_count >= High(show_in_ops) then
    begin
        asm
      int 3
        end;
        Exit;
    end;
    show_in_ops[show_in_op_count] := i;
    inc(show_in_op_count);
end;

procedure TParamDef.Set_used_in_op(const op: String; op_param_order: Integer);
var i, j: Integer;
begin
    i := Operations.IndexOf(op);
    if i < 0 then
        i := Operations.Add(op);
    for j := 0 to used_in_op_count - 1 do
        if used_in_ops[j] = i then Exit;
    if used_in_op_count >= High(used_in_ops) then
    begin
        asm
      int 3
        end;
        Exit;
    end;
    used_in_ops[used_in_op_count] := i;
    {if param_ord_count > High( param_orders ) then
    begin
      asm
        int 3
      end;
      Exit;
    end;}
    param_orders[{param_ord_count}used_in_op_count] := op_param_order;
    //inc( param_ord_count );
    inc(used_in_op_count);
end;

initialization
    Operations := TStringList.Create;
    Operations.Add('');
    {$IFDEF DEBUG_SUM}
    TRY
        if DebugMode then DeleteFile(GetLogsPath + 'sum.log');
    EXCEPT
    END;
    {$ENDIF}

finalization
    Operations.Free;

end.
