//------------------------------------------------------------------------------
//        
//   FastListView.
//
//  .. 03.08.2004 - 
//             04.08.2004 - .  ,
//                           .
//             10.08.2004 -      ( 5) ,
//                                
//------------------------------------------------------------------------------
unit ParamsController;

interface

{$I VCL_VERSIONS.INC}

uses
    Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
    FormulaInterpreterUnit, ParamUnit, FastListview, stdctrls, commctrl, qSort;

const Nmax = 9;
    WM_USER_REFRESH_PARAMS = WM_USER + 17;

type
    TOnOutputParam = procedure(Sender: TObject; Params: TVars; Idx: Integer)
        of object;
    TOnError = procedure(Sender: TObject; const ErrorStr: String; var RaiseError: Boolean) of object;
    TOnEdit = procedure(Sender: TObject; Params: TVars; Idx: Integer) of object;

    TParamsController = class(TParamsSet)
    private
        FLVName: array[1..Nmax] of String;
        FCode_Op: array[1..Nmax] of String;
        FColumn_Name: String;
        FColumn_Description: String;
        FColumn_Value: String;
        Fpackid: Integer;
        FOnOutputParam: TOnOutputParam;
        FOnAfterOutput: TNotifyEvent;
        FOnBeforeOutput: TNotifyEvent;
        FOnError: TOnError;
        FOnAllowSum: TAllowSum;
        FRestrictSum_Batch: Boolean;
        FClear_ListView_When_Output: Boolean;
        Fnomenklid: Integer;
        FOnEdit: TOnEdit;
        FAutoRefresh: Boolean;
        FColumn_Units: String;
        FOnBeforeCalc: TNotifyEvent;
        FActive: Boolean;
        FOnClearError: TNotifyEvent;
        FPackage_TableName: String;
        procedure SetColumn_Description(const Value: String);
        procedure SetColumn_Name(const Value: String);
        procedure SetColumn_Value(const Value: String);
        procedure SetParamsList(const Value: TList);
        procedure Setpackid(const Value: Integer);
        procedure SetOnOutputParam(const Value: TOnOutputParam);
        procedure SetOnAfterOutput(const Value: TNotifyEvent);
        procedure SetOnBeforeOutput(const Value: TNotifyEvent);
        procedure SetOnError(const Value: TOnError);
        procedure SetOnAllowSum(const Value: TAllowSum);
        procedure SetRestrictSum_Batch(const Value: Boolean);
        procedure SetClear_ListView_When_Output(const Value: Boolean);
        procedure Setnomenklid(const Value: Integer);
        procedure SetOnEdit(const Value: TOnEdit);
        procedure SetAutoRefresh(const Value: Boolean);
        procedure SetColumn_Units(const Value: String);
        function GetCodeOp(const Index: Integer): String;
        procedure SetCodeOp(const Index: Integer; const Value: String);
        function GetLVName(const Index: Integer): String;
        procedure SetLVName(const Index: Integer; const Value: String);
        procedure SetOnBeforeCalc(const Value: TNotifyEvent);
        procedure SetActive(const Value: Boolean);
        procedure SetPackage_TableName(const Value: String);
        { Private declarations }
    protected
        { Protected declarations }
        FParamsList: TList;
        procedure Free_ParamsList;
        function IsIndex(const S: String): Integer;
        procedure Loaded; override;
        procedure AllowSum_In_Batch_Only(Sender, SumParam: TVars;
            var AllowSum: Boolean; calc_by_batch: Boolean = TRUE);
        function GetLV(LVIdx: Integer): TFastListView;
        procedure SetLV(LVIdx: Integer; Value: TFastListView);
    protected
        FAttached: array[1..Nmax] of Boolean;
        OldOnKeyDown: array[1..Nmax] of TKeyEvent;
        OldOnKeyChar: array[1..Nmax] of TKeyPressEvent;
        OldOnEditing: array[1..Nmax] of TFLVEditingEvent;
        OldOnEdited: array[1..Nmax] of TFLVEditedEvent;
        OldOnMessage: array[1..NMax] of TOnMessage;
        procedure AttachEvents(LVIdx: Integer);
        procedure DetachEvents(LVIdx: Integer);
        //--------------------------------------------------------------------------
        //   OnKeyDown.   
        // list view    .
        procedure LV_OnKeyDown(Sender: TObject; var Key: Word; Shift: TShiftState);
        //--------------------------------------------------------------------------
        //   OnKeyChar.   .
        procedure LV_OnKeyChar(Sender: TObject; var Key: Char);
        //--------------------------------------------------------------------------
        //   OnEditing.  ,  
        //  .
        procedure LV_OnEditing(Sender: TObject; Item: TFastListItem;
            var AllowEdit: Boolean);
        //--------------------------------------------------------------------------
        //   OnEdited.  OnEdit,  .
        procedure LV_OnEdited(Sender: TObject; Item: TFastListItem; var S: string);
        //--------------------------------------------------------------------------
        //   OnKeyChar  Edit-.
        procedure LV_OnEditKeyChar(Sender: TObject; var Key: Char);
        //--------------------------------------------------------------------------
        //   OnKeyChar  Edit-.
        function LV_OnMessage(Sender: TObject; var Msg: TMessage): Boolean;
    protected
        OldEdWndProc: Pointer;
        DoInitInputVals: Boolean;
    protected
        FCachedFormulas: TList;
        function SumCached(P: TParams; Idx, fPos: Integer; var F: String;
            calc_by_batch: Boolean): Boolean;
            override;
        procedure DoSumCache(P: TParams; Idx, fPos: Integer; const F: String;
            Value: Double; calc_by_batch: Boolean);
            override;
    public
        { Public declarations }
        property Package_TableName: String read FPackage_TableName write SetPackage_TableName;
        property ParamsList: TList read FParamsList write SetParamsList;
        constructor Create(AOwner: TComponent); override;
        destructor Destroy; override;
        function LoadParams_by_ConList(Preload_BatchValues: Boolean;
            const ConList: String; MergeParams: Boolean = FALSE;
            AdditionalConds: String = ''): Integer;
        function LoadParams_by_ConList_TranspackId(const Conditions: String;
            DateFrom, DateTo: TDateTime; LoadParentParams: Boolean = FALSE;
            MergeParams: Boolean = FALSE): Integer;
        function LoadParams_by_NomList(const NomList: String; only_nomenklid: Integer = -1): Integer;
        function LoadParams_by_BatList(const BatList: String): Integer;
        procedure Recalculate(use_recalc: Boolean);
        //--------------------------------------------------------------------------
        //   0  ,   :
        procedure ResetDependingInputParams(LVIdx: Integer; const ParamName,
            ParamFormula: String; nomenklid, batchid, packid: Integer);
        //--------------------------------------------------------------------------
        //      listview  
        procedure OutputParams; overload;
        procedure OutputParams(LVIdx: Integer); overload;
        property packid: Integer read Fpackid write Setpackid;
        // id ,   .
        //  = 0,  0-    
        //  id  packid.  . nomenklid.
        property nomenklid: Integer read Fnomenklid write Setnomenklid;
        // id ,    
        procedure Clear(ClearAll: Boolean); override;
        //--------------------------------------------------------------------------
        //    
        function Param_By_Nomenclature(nomenklid: Integer; packid: Integer = 0): TVars;
        //--------------------------------------------------------------------------
        //     
        function Code_Operation_List: String;
        //--------------------------------------------------------------------------
        function HasContainersInNomen(Anomenklid: Integer;
            ignore_packid_0: Boolean = true): Boolean;
    published
        { Published declarations }
        property Code_Operation: String index 1 read GetCodeOp write SetCodeOp;
        property Code_Op1: String index 1 read GetCodeOp write SetCodeOp;
        property Code_Op2: String index 2 read GetCodeOp write SetCodeOp;
        property Code_Op3: String index 3 read GetCodeOp write SetCodeOp;
        property Code_Op4: String index 4 read GetCodeOp write SetCodeOp;
        property Code_Op5: String index 5 read GetCodeOp write SetCodeOp;
        property Code_Op6: String index 6 read GetCodeOp write SetCodeOp;
        property Code_Op7: String index 7 read GetCodeOp write SetCodeOp;
        property Code_Op8: String index 8 read GetCodeOp write SetCodeOp;
        property Code_Op9: String index 9 read GetCodeOp write SetCodeOp;
        property ListView: TFastListView index 1 read GetLV write SetLV;
        property ListView1: TFastListView index 1 read GetLV write SetLV;
        property ListView2: TFastListView index 2 read GetLV write SetLV;
        property ListView3: TFastListView index 3 read GetLV write SetLV;
        property ListView4: TFastListView index 4 read GetLV write SetLV;
        property ListView5: TFastListView index 5 read GetLV write SetLV;
        property ListView6: TFastListView index 6 read GetLV write SetLV;
        property ListView7: TFastListView index 7 read GetLV write SetLV;
        property ListView8: TFastListView index 8 read GetLV write SetLV;
        property ListView9: TFastListView index 9 read GetLV write SetLV;
        property LV1Name: String index 1 read GetLVName write SetLVName;
        property LV2Name: String index 2 read GetLVName write SetLVName;
        property LV3Name: String index 3 read GetLVName write SetLVName;
        property LV4Name: String index 4 read GetLVName write SetLVName;
        property LV5Name: String index 5 read GetLVName write SetLVName;
        property LV6Name: String index 6 read GetLVName write SetLVName;
        property LV7Name: String index 7 read GetLVName write SetLVName;
        property LV8Name: String index 8 read GetLVName write SetLVName;
        property LV9Name: String index 9 read GetLVName write SetLVName;

        property Column_Description: String read FColumn_Description write SetColumn_Description;
        property Column_Name: String read FColumn_Name write SetColumn_Name;
        property Column_Value: String read FColumn_Value write SetColumn_Value;
        property Column_Units: String read FColumn_Units write SetColumn_Units;
        //--------------------------------------------------------------------------
        // 
        property OnBeforeCalc: TNotifyEvent read FOnBeforeCalc write SetOnBeforeCalc;
        property OnOutputParam: TOnOutputParam read FOnOutputParam write SetOnOutputParam;
        property OnBeforeOutput: TNotifyEvent read FOnBeforeOutput write SetOnBeforeOutput;
        property OnAfterOutput: TNotifyEvent read FOnAfterOutput write SetOnAfterOutput;
        //--------------------------------------------------------------------------
        // ,       (  
        //  ).   ,    raise  ,
        //       .
        property OnError: TOnError read FOnError write SetOnError;
        property OnClearError: TNotifyEvent read FOnClearError write FOnClearError;
        property OnAllowSum: TAllowSum read FOnAllowSum write SetOnAllowSum;
        //--------------------------------------------------------------------------
        //  () :
        property OnEdit: TOnEdit read FOnEdit write SetOnEdit;
        //--------------------------------------------------------------------------
        //    OnAllowSum,   RestrictSum_Batch = TRUE 
        // ,        .
        property RestrictSum_Batch: Boolean read FRestrictSum_Batch write SetRestrictSum_Batch;
        property Clear_ListView_When_Output: Boolean read FClear_ListView_When_Output
            write SetClear_ListView_When_Output;
        property AutoRefresh: Boolean read FAutoRefresh write SetAutoRefresh;
        property Active: Boolean read FActive write SetActive;
    public
        //--------------------------------------------------------------------------
        //   ""   .
        // ,         
        //    (  ,   
        //    ,      
        // ).
        function NewSetOfPack(CountPack: Integer; nomenklid: Integer;
            const BatchName: String; batchid: Integer; InputReady: Boolean;
            FirstNum: Integer; var ConteinersToDestroy: String): Boolean;

        //--------------------------------------------------------------------------
        //       
        procedure CopyContainerAnalysesFromBatchAnalyses(only_zeroes: Boolean);

        //--------------------------------------------------------------------------
        //    , ""  NewSetPack (
        // packid = batchid = -2).
        procedure ClearDummy;
        procedure ClearData;

        //--------------------------------------------------------------------------
        //  ""  batchid 
        procedure Set_batchid(Abatchid: Integer; AllPacks: Boolean);
    public
        procedure Dump(const filename: String);

    end;

    {$R *.dcr}
procedure Register;

implementation

uses UtilsUnit;

procedure Register;
//------------------------------------------------------------------------------
//  TParamsController   
begin
    RegisterComponents('Kladov', [TParamsController]);
end;

const
    WM_USER_KEYPRESS = WM_USER + 18;

type
    TCacheFormula = class
        SumFormula: String;
        bachid: Integer;
        nomenklid: Integer;
        sumbybatch: Boolean;
        //Result: Double;
        RsltStr: String;
    end;

function IsNumber(const S: String): Boolean;
var i: Integer;
begin
    Result := FALSE;
    if S = '' then exit;
    i := 1;
    if S[i] in ['+', '-'] then inc(i);
    if i > Length(S) then Exit;
    for i := i to Length(S) do
        if not (S[i] in ['0'..'9']) then Exit;
    Result := TRUE;
end;

{ TParamsController }

procedure TParamsController.AllowSum_In_Batch_Only(Sender, SumParam: TVars;
    var AllowSum: Boolean; calc_by_batch: Boolean = TRUE);
//------------------------------------------------------------------------------
//   AllowSum  .   
//  RestrictSum_Batch =  TRUE     
// OnAllowSum.     .
begin
    if not calc_by_batch then
        AllowSum := TRUE
    else
    begin
        AllowSum := (Sender is TParams) and (SumParam is TParams) and
            (TParams(Sender).ParamSetData.batchid <> 0) and
            (TParams(Sender).ParamSetData.batchid =
            TParams(SumParam).ParamSetData.batchid);
        {$IFDEF DEBUG_SUM}
        {LogFileOutPut( 'c:\sum.log', ' ' + TParams( Sender ).ParamSetData.numcon +
          ' # ' + IntToStr( TParams( Sender ).ParamSetData.batchid ) +
          ' # ' + IntToStr( TParams( Sender ).ParamSetData.nomenklid ) +
          '  ' + TParams( SumParam ).ParamSetData.numcon +
          ' # ' + IntToStr( TParams( SumParam ).ParamSetData.batchid ) +
          ' # ' + IntToStr( TParams( SumParam ).ParamSetData.nomenklid ) +
          IfThenElse( AllowSum, ' +++  +++', '---   ---' ) );}
        {$ENDIF}
    end;
    if AllowSum and calc_by_batch then
        AllowSum := AllowSum AND
            (TParams(Sender).ParamSetData.packid <> 0) and
            (TParams(SumParam).ParamSetData.packid <> 0);
end;

procedure TParamsController.AttachEvents(LVIdx: Integer);
//------------------------------------------------------------------------------
//     ListView
//         ( design-time),  
//     (run-time).
var LV: TFastListView;
begin
    if not FActive then Exit;
    if FAttached[LVIdx] then DetachEvents(LVIdx);
    if FAttached[LVIdx] then Exit;
    if csDesigning in ComponentState then Exit;
    if csLoading in ComponentState then Exit;
    LV := GetLV(LVIdx);
    if LV = nil then Exit;
    //----------------------------------------------------------------------------
    OldOnKeyDown[LVIdx] := LV.OnKeyDown;
    LV.OnKeyDown := LV_OnKeyDown;
    OldOnKeyChar[LVIdx] := LV.OnKeyPress;
    LV.OnKeyPress := LV_OnKeyChar;
    OldOnEditing[LVIdx] := LV.OnEditing;
    LV.OnEditing := LV_OnEditing;
    OldOnEdited[LVIdx] := LV.OnEdited;
    LV.OnEdited := LV_OnEdited;
    OldOnMessage[LVIdx] := LV.OnMessage;
    LV.OnMessage := LV_OnMessage;
    //----------------------------------------------------------------------------
    LV.AutoAllocateColumnNames := TRUE; // list view    !
    FAttached[LVIdx] := TRUE;
end;

procedure TParamsController.Clear(ClearAll: Boolean);
//------------------------------------------------------------------------------
//   
var I: Integer;
    SaveConstParams: Pointer;
begin
    FCachedFormulas.Clear;
    if ClearAll then
    begin
        SaveConstParams := nil;
    end
    else
    begin
        SaveConstParams := ConstParams;
        ConstParams := nil;
    end;
    inherited;
    ConstParams := SaveConstParams;
    if Assigned(ParamsList) then
    begin
        for I := 0 to ParamsList.Count - 1 do
            TObject(ParamsList[I]).Free;
        ParamsList.Clear;
    end;
end;

procedure TParamsController.ClearData;
//     
var I: Integer;
    P: TParams;
begin
    for I := ParamsList.Count - 1 downto 0 do
    begin
        P := ParamsList[I];
        if (P.ParamSetData.packid <> 0) and (P.ParamSetData.batchid <> 0) then
        begin
            ParamsList.Delete(I);
            P.Free;
        end
        else
            if P.ParamSetData.packid = 0 then
            begin
                P.ParamSetData.ignore_conditions := FALSE;
            end;
    end;
end;

procedure TParamsController.ClearDummy;
//    , ""  NewSetPack (
// packid = batchid = -2).
var I: Integer;
    P: TParams;
begin
    for I := ParamsList.Count - 1 downto 0 do
    begin
        P := ParamsList[I];
        if (P.ParamSetData.packid < 0) and (P.ParamSetData.batchid = -2) then
        begin
            ParamsList.Delete(I);
            P.Free;
        end
        else
            if P.ParamSetData.packid = 0 then
            begin
                P.ParamSetData.ignore_conditions := FALSE;
            end;
    end;
end;

function TParamsController.Code_Operation_List: String;
var I: Integer;
begin
    Result := '';
    for I := 1 to High(FCode_Op) do
        if FCode_Op[I] <> '' then
            Result := Result + ',''' + FCode_Op[I] + '''';
    if Result = '' then Exit;
    if Result[1] = ',' then
        StrDel(Result, 1, 1);
end;

constructor TParamsController.Create(AOwner: TComponent);
//------------------------------------------------------------------------------
// . .
begin
    inherited Create(AOwner);
    FClear_ListView_When_Output := TRUE;
    Column_Description := '1';
    Column_Value := '0';
    Column_Name := '2';
    Column_Units := '';
    FActive := TRUE;
    ParamsList := TList.Create;
    FCachedFormulas := TList.Create;
    FPackage_TableName := 'Package';
end;

destructor TParamsController.Destroy;
//------------------------------------------------------------------------------
// .  .
var I: Integer;
begin
    Free_ParamsList;
    for I := 0 to FCachedFormulas.Count - 1 do
        TObject(FCachedFormulas[I]).Free;
    FCachedFormulas.Free;
    inherited;
end;

procedure TParamsController.DetachEvents(LVIdx: Integer);
//------------------------------------------------------------------------------
//     ListView.
//         ( design-time),  
//     (run-time).
var LV: TFastListView;
begin
    if csDesigning in ComponentState then Exit;
    if csLoading in ComponentState then Exit;
    LV := GetLV(LVIdx);
    if LV = nil then Exit;
    if not FAttached[LVIdx] then Exit;
    //----------------------------------------------------------------------------
    LV.OnKeyDown := OldOnKeyDown[LVIdx]; OldOnKeyDown[LVIdx] := nil;
    LV.OnKeyPress := OldOnKeyChar[LVIdx]; OldOnKeyChar[LVIdx] := nil;
    LV.OnEditing := OldOnEditing[LVIdx]; OldOnEditing[LVIdx] := nil;
    LV.OnEdited := OldOnEdited[LVIdx]; OldOnEdited[LVIdx] := nil;
    LV.OnMessage := OldOnMessage[LVIdx]; OldOnMessage[LVIdx] := nil;
    //----------------------------------------------------------------------------
    FAttached[LVIdx] := FALSE;
end;

procedure TParamsController.DoSumCache(P: TParams; Idx, fPos: Integer;
    const F: String; Value: Double; calc_by_batch: Boolean);
var CF: TCacheFormula;
    I, L: Integer;
begin
    CF := TCacheFormula.Create;
    FCachedFormulas.Add(CF);
    I := fPos;
    L := 0;
    while I < Length(F) do
    begin
        CASE F[I] OF
            '(': Inc(L);
            ')': begin
                    Dec(L);
                    if L = 0 then break;
                end;
        END;
        Inc(I);
    end;
    CF.SumFormula := Copy(F, fPos, I + 1 - fPos);
    CF.bachid := P.ParamSetData.batchid;
    CF.nomenklid := P.ParamSetData.nomenklid;
    CF.sumbybatch := calc_by_batch;
    //CF.Result := R;
    Str(Value, CF.RsltStr);
end;

procedure TParamsController.Free_ParamsList;
//------------------------------------------------------------------------------
//   ParamsList -   .
begin
    Clear(FALSE);
    FreeAndNil(FParamsList);
end;

function TParamsController.GetCodeOp(const Index: Integer): String;
begin
    Result := FCode_Op[Index];
end;

function TParamsController.GetLV(LVIdx: Integer): TFastListView;
//------------------------------------------------------------------------------
//    ,       
//  listview        
//     .
var Comp: TComponent;
    S: String;
    Masterform: TComponent;
    Dotpos: Integer;
begin
    Result := nil;
    if FLVName[LVIdx] <> '' then
    begin
        //ShowMessage( 'searching: ' + FListViewName );
        Comp := nil;
        if Owner <> nil then
        begin
            S := FLVName[LVIdx];
            Masterform := Owner;
            Dotpos := pos('.', S);
            if Dotpos > 0 then
            begin
                Masterform := Application.FindComponent(Copy(S, 1, Dotpos - 1));
                if Masterform = nil then
                begin
                    Exit;
                end;
                S := Copy(S, Dotpos + 1, MaxInt);
            end;
            Comp := Masterform.FindComponent(S);
        end;
        //ShowMessage( 'found: ' + IntToStr( Integer( Pointer( Comp ) ) ) );
        if Comp is TFastListView then
            Result := Comp as TFastListView;
    end;
end;

function TParamsController.GetLVName(const Index: Integer): String;
begin
    Result := FLVName[Index];
end;

function TParamsController.HasContainersInNomen(Anomenklid: Integer;
    ignore_packid_0: Boolean = true): Boolean;
var I: Integer;
    P: TParams;
begin
    Result := TRUE;
    for I := ParamsList.Count - 1 downto 0 do
    begin
        P := ParamsList[I];
        if ignore_packid_0 and (P.ParamSetData.packid = 0) then
            continue;
        if P.ParamSetData.nomenklid = Anomenklid then Exit;
    end;
    Result := FALSE;
end;

function TParamsController.IsIndex(const S: String): Integer;
//------------------------------------------------------------------------------
// ,   S  ,    .
//   ,  -1.
begin
    TRY
        if IsNumber(S) then
            Result := StrToInt(S)
        else Result := -1;
    EXCEPT
        Result := -1;
    END;
end;

procedure TParamsController.Loaded;
//------------------------------------------------------------------------------
//   ,      .
//       list view.
var I: Integer;
begin
    inherited;
    for I := 1 to High(FLVName) do
        AttachEvents(I);
end;

function TParamsController.LoadParams_by_BatList(const BatList: String): Integer;
//------------------------------------------------------------------------------
//      (   - Code_Operation)
var RaiseErr: Boolean;
    s_PackTabName: String;
begin
    s_PackTabName := ParamUnit.PackageTableName;
    ParamUnit.PackageTableName := Package_TableName;
    Result := 0;
    TRY
        TRY
            FCachedFormulas.Clear;
            if not Assigned(OnAllowSum) and RestrictSum_Batch then
                ParamsList := ParamUnit.LoadParams_by_BatList(BatList, Code_Operation_List,
                    AllowSum_In_Batch_Only, Self)
            else
                ParamsList := ParamUnit.LoadParams_by_BatList(BatList, Code_Operation_List,
                    OnAllowSum, Self);
            Result := ParamsList.Count;
        EXCEPT on E: Exception do
            begin
                RaiseErr := TRUE;
                if Assigned(OnError) then
                begin
                    RaiseErr := FALSE;
                    OnError(Self, E.Message, RaiseErr);
                end;
                if RaiseErr then
                    raise;
            end;
        END;
    FINALLY
        ParamUnit.PackageTableName := s_PackTabName;
    END;
end;

function TParamsController.LoadParams_by_ConList(Preload_BatchValues: Boolean; const ConList: String; MergeParams: Boolean = FALSE;
    AdditionalConds: String = ''): Integer;
//------------------------------------------------------------------------------
//      (   - Code_Operation)
var RaiseErr: Boolean;
    PL: TList;
    i: Integer;
    s_PackTabName: String;
begin
    s_PackTabName := ParamUnit.PackageTableName;
    ParamUnit.PackageTableName := Package_TableName;
    Result := 0;
    TRY
        TRY
            FCachedFormulas.Clear;
            if not Assigned(OnAllowSum) and RestrictSum_Batch then
                PL := ParamUnit.LoadParams_by_ConList(ConList, Code_Operation_List, AllowSum_In_Batch_Only,
                    FALSE, Self, AdditionalConds)
            else
                PL := ParamUnit.LoadParams_by_ConList(ConList, Code_Operation_List, OnAllowSum,
                    FALSE, Self, AdditionalConds);
            Result := PL.Count;
            if MergeParams then
            begin
                TRY
                    for i := 0 to PL.Count - 1 do
                        ParamsList.Add(PL.Items[i]);
                FINALLY
                    PL.Free;
                END;
            end
            else
                ParamsList := PL;
        EXCEPT on E: Exception do
            begin
                RaiseErr := TRUE;
                if Assigned(OnError) then
                begin
                    RaiseErr := FALSE;
                    OnError(Self, E.Message, RaiseErr)
                end;
                if RaiseErr then
                    raise;
            end;
        END;
    FINALLY
        ParamUnit.PackageTableName := s_PackTabName;
    END;
end;

function TParamsController.LoadParams_by_ConList_TranspackId(
    const Conditions: String; DateFrom, DateTo: TDateTime;
    LoadParentParams: Boolean = FALSE; MergeParams: Boolean = FALSE): Integer;
var RaiseErr: Boolean;
    List: TList;
    i: Integer;
    s_PackTabName: String;
begin
    s_PackTabName := ParamUnit.PackageTableName;
    ParamUnit.PackageTableName := Package_TableName;
    Result := 0;
    TRY
        TRY
            FCachedFormulas.Clear;
            if not Assigned(OnAllowSum) and RestrictSum_Batch then
                List := ParamUnit.LoadParams_by_ConList_TranspackId(
                    Conditions, Code_Operation_List, AllowSum_In_Batch_Only, FALSE,
                    DateFrom, DateTo, Self, LoadParentParams)
            else
                List := ParamUnit.LoadParams_by_ConList_TranspackId(
                    Conditions, Code_Operation_List, OnAllowSum, FALSE,
                    DateFrom, DateTo, Self, LoadParentParams);
            Result := List.Count;

            if MergeParams then
            begin
                for i := 0 to List.Count - 1 do
                    ParamsList.Add(List[i]);
                List.Free;
            end
            else
            begin
                ParamsList := List;
            end;
        EXCEPT on E: Exception do
            begin
                RaiseErr := TRUE;
                if Assigned(OnError) then
                begin
                    RaiseErr := FALSE;
                    OnError(Self, E.Message, RaiseErr)
                end;
                if RaiseErr then
                    raise;
            end;
        END;
    FINALLY
        ParamUnit.PackageTableName := s_PackTabName;
    END;
end;

function TParamsController.LoadParams_by_NomList(const NomList: String;
    only_nomenklid: Integer = -1): Integer;
//------------------------------------------------------------------------------
//      (   - Code_Operation)
var RaiseErr: Boolean;
    s_PackTabName: String;
begin
    s_PackTabName := ParamUnit.PackageTableName;
    ParamUnit.PackageTableName := Package_TableName;
    Result := 0;
    TRY
        TRY
            FCachedFormulas.Clear;
            if not Assigned(OnAllowSum) and RestrictSum_Batch then
                ParamsList := ParamUnit.LoadParams_by_NomList(NomList, Code_Operation_List,
                    AllowSum_In_Batch_Only, Self, only_nomenklid)
            else
                ParamsList := ParamUnit.LoadParams_by_NomList(NomList, Code_Operation_List,
                    OnAllowSum, Self, only_nomenklid);
            Result := ParamsList.Count;
        EXCEPT on E: Exception do
            begin
                RaiseErr := TRUE;
                if Assigned(OnError) then
                begin
                    RaiseErr := FALSE;
                    OnError(Self, E.Message, RaiseErr);
                end;
                if RaiseErr then
                    raise;
            end;
        END;
    FINALLY
        ParamUnit.PackageTableName := s_PackTabName;
    END;
end;

function NewEdWndProc(Wnd: HWnd; Msg: DWORD; wParam, lParam: Integer): Integer;
    stdcall;
var PC: TParamsController;
    LV: TFastListView;
    Txt: String;
begin
    PC := Pointer(GetProp(Wnd, 'ParamController'));
    LV := Pointer(GetProp(Wnd, 'FastListView'));
    if PC = nil then
        Result := DefWindowProc(Wnd, Msg, wParam, lParam)
    else
    begin
        Result := CallWindowProc(PC.OldEdWndProc, Wnd, Msg, wParam, lParam);
        CASE Msg OF
            WM_SHOWWINDOW:
                if (wParam = 0) and (LV <> nil) then
                begin
                    SetLength(Txt, GetWindowTextLength(Wnd));
                    GetWindowText(Wnd, PChar(Txt), Length(Txt) + 1);
                    PC.LV_OnEdited(LV, LV.Selected, Txt);
                    RemoveProp(Wnd, 'ParamController');
                    RemoveProp(Wnd, 'FastListView');
                end;
            WM_DESTROY:
                begin
                    if LV <> nil then
                        PC.LV_OnEdited(LV, LV.Selected, Txt);
                    RemoveProp(Wnd, 'ParamController');
                    RemoveProp(Wnd, 'FastListView');
                end;
        END;
    end;
end;

procedure TParamsController.LV_OnEdited(Sender: TObject;
    Item: TFastListItem; var S: string);
//------------------------------------------------------------------------------
// :  .    
//    ,  OnEdit.
var P: TParams;
    Idx: Integer;
    Old: Double;
    NewS: String;
    I: Integer;
    LV: TFastListView;
    LVIdx: Integer;
    RaiseErr: Boolean;
begin
    LVIdx := -1;
    if Item = nil then Exit;
    for I := 1 to High(FAttached) do
        if GetLV(I) = Sender then
        begin
            LVIdx := I;
            break;
        end;
    if LVIdx <= 0 then Exit;
    TRY
        if Assigned(OldOnEdited[LVIdx]) then
            OldOnEdited[LVIdx](Sender, Item, S);
        LV := GetLV(LVIdx);
        if LV <> nil then
        begin
            P := Pointer(StrToInt(Item['param_object']));
            if (P <> nil) and (ParamsList.IndexOf(P) >= 0) and
                (P.ParamData[StrToInt(Item['param_index'])].param_type = 3) then
            begin
                Idx := StrToInt(Item['param_index']);
                //I := Item.Index;
                //TopItem := LV.TopItem.Index;
                Old := P.Values[Idx];
                TRY
                    NewS := StringReplace(S, ',', DecimalSeparator, []);
                    NewS := StringReplace(NewS, '.', DecimalSeparator, []);
                    P.Values[Idx] := StrToFloat(NewS);
                    P.ParamData[Idx].param_ready := TRUE;
                EXCEPT
                    RaiseErr := TRUE;
                    if Assigned(OnError) then
                    begin
                        RaiseErr := FALSE;
                        OnError(Self, ' : ' + S, RaiseErr);
                    end;
                    if RaiseErr then
                        raise
                    else
                        Beep;
                    P.Values[Idx] := Old;
                END;
                ResetDependingInputParams(LVIdx, P.Names[Idx], P.ParamData[Idx].formula,
                    P.ParamSetData.nomenklid, P.ParamSetData.batchid, P.ParamSetData.packid);
                if Assigned(OnEdit) then
                    OnEdit(Self, P, Idx);
                if AutoRefresh and Clear_ListView_When_Output then
                begin
                    //Item.Caption := S;
                    PostMessage(LV.Handle, WM_USER_REFRESH_PARAMS, LVIdx, Item.Index);
                    {Recalculate;
                    OutputParams;
                    if I < LV.Count then
                    begin
                      LV.Selected := nil;
                      LV.Selected := LV.Items[ I ];
                      LV.ItemFocused := LV.Selected;
                      if TopItem < LV.Count then
                      begin
                        LV.Items[ LV.Count-1 ].MakeVisible( FALSE );
                        LV.Items[ TopItem ].MakeVisible( FALSE );
                      end
                      else
                        LV.Selected.MakeVisible( FALSE );
                    end;}
                end;
            end;
        end;
    EXCEPT
        //    -   
    END;
end;

procedure TParamsController.LV_OnEditing(Sender: TObject;
    Item: TFastListItem; var AllowEdit: Boolean);
//------------------------------------------------------------------------------
// :     .
// :     ,   
var P: TParams;
    S: String;
    LVIdx: Integer;
    I: Integer;
    Ed: HWnd;
    LV: TFastListView;
begin
    LVIdx := -1;
    for I := 1 to High(FAttached) do
        if GetLV(I) = Sender then
        begin
            LVIdx := I;
            break;
        end;
    if LVIdx <= 0 then Exit;
    LV := GetLV(LVIdx);
    if Assigned(OldOnEditing[LVIdx]) then
        OldOnEditing[LVIdx](Sender, Item, AllowEdit);
    if AllowEdit then
    TRY
        if Item['param_object'] <> '' then
        begin
            I := StrToInt(Item['param_object']);
            P := Pointer(I);
            if (P <> nil) and (ParamsList.IndexOf(P) >= 0) then
            begin
                S := Item['param_index'];
                if (P.ParamData[StrToInt(S)].param_type = 3) and
                not P.ParamData[StrToInt(S)].read_only then
                begin //   
                    AllowEdit := TRUE;
                    Exit;
                end;
            end;
            AllowEdit := FALSE;
            Ed := LV.Perform(LVM_GETEDITCONTROL, 0, 0);
            if (Ed <> 0) and (GetProp(Ed, 'ParamController') = 0) then
            begin
                SetProp(Ed, 'ParamController', Integer(Self));
                SetProp(Ed, 'FastListView', Integer(LV));
                OldEdWndProc := Pointer(GetWindowLong(Ed, GWL_WNDPROC));
                SetWindowLong(Ed, GWL_WNDPROC, Integer(@NewEdWndProc));
            end;
        end;
    EXCEPT
        //    -   
    END;
end;

procedure TParamsController.LV_OnEditKeyChar(Sender: TObject;
    var Key: Char);
begin
    //
end;

procedure TParamsController.LV_OnKeyChar(Sender: TObject; var Key: Char);
//--------------------------------------------------------------------------
//   OnKeyChar.   .
var LV: TFastListView;
    LI: TListItem;
    P: TParams;
    PD: TParamData;
    LVIdx, I: Integer;
    EW: HWnd;
    S, Snew: String;
    L, SS: Integer;
    RE: Boolean;
    Err: String;
begin
    LVIdx := -1;
    for I := 1 to High(FAttached) do
        if GetLV(I) = Sender then
        begin
            LVIdx := I;
            break;
        end;
    if LVIdx <= 0 then Exit;
    LV := GetLV(LVIdx);
    if LV = nil then Exit;
    if LV.IsEditing then
    begin
        LI := LV.ItemFocused;
        if LI = nil then
            LI := LV.Selected;
        if LI = nil then Exit;
        P := Pointer(StrToInt(LI['param_object']));
        if (P <> nil) and (ParamsList.IndexOf(P) >= 0) then
        begin
            PD := P.ParamData[StrToInt(LI['param_index'])];
            if (PD.param_type = 3) and
                not PD.read_only then
            begin //  ,    
                Err := '';
                if Assigned(OnClearError) then
                    OnClearError(self);
                TRY
                    CASE Key OF
                        #8, #13, #27: ;
                        '0'..'9', '.', ',', '-', '+':
                            begin
                                TRY
                                    EW := LV.Perform(LVM_GETEDITCONTROL, 0, 0);
                                    if EW <> 0 then //      
                                    begin
                                        //---    
                                        L := SendMessage(EW, EM_LINELENGTH, 0, 0);
                                        SetLength(S, L + 1);
                                        PWord(@S[1])^ := L + 1;
                                        SendMessage(EW, EM_GETLINE, 0, Integer(@S[1]));
                                        S[L + 1] := #0;
                                        S := PChar(S);
                                        //---   (SS)   (L)    
                                        SS := SendMessage(EW, EM_GETSEL, 0, 0);
                                        L := HiWord(SS) - LoWord(SS);
                                        SS := LoWord(SS);
                                        //---    ,  , 
                                        //   :
                                        Snew := Copy(S, 1, SS) + Key + Copy(S, SS + 1 + L, MaxInt);
                                        //---    
                                        TRY
                                            Snew := Replace(Snew, '.', DecimalSeparator);
                                            Snew := Replace(Snew, ',', DecimalSeparator);
                                            SysUtils.StrToFloat(Snew);
                                        EXCEPT
                                            if (Snew = '+') or (Snew = '-') then
                                            begin //  :     
                                                Exit;
                                            end;
                                            Err := '  ';
                                            Key := #0; Beep; Exit;
                                        END;
                                        //---     
                                        //   :
                                        SS := pos(DecimalSeparator, Snew);
                                        if SS <= 0 then Exit;
                                        L := Length(Snew) - SS;
                                        if PD.precision < L then
                                        begin
                                            Err := '   ' + IntToStr(PD.precision) +
                                                '         ' +
                                                P.Names[StrToInt(LI['param_index'])] +
                                                '  ' + IntToStr(P.ParamSetData.nomenklid);
                                            Beep; Key := #0;
                                        end;
                                    end;
                                FINALLY
                                    if Err = '' then
                                    TRY
                                        LV.ItemFocused.Caption := Snew;
                                        LV_OnEdited(Sender, LV.ItemFocused, Snew);
                                    EXCEPT
                                    END;
                                END;
                            end;
                    else begin
                            Err := ' ''' + Key + '''      ';
                            Beep; Key := #0;
                        end;
                    END;
                FINALLY
                    RE := FALSE;
                    if Key = #0 then
                        if Assigned(OnError) then
                            OnError(Self, ' : ' + Err, RE);
                END;
            end;
        end;
    end;
end;

procedure TParamsController.LV_OnKeyDown(Sender: TObject; var Key: Word;
    Shift: TShiftState);
//------------------------------------------------------------------------------
// :     list view.
    procedure SupressChar(W: HWnd);
    var M1: TMsg;
    begin
        PeekMessage(M1, W, WM_CHAR, WM_CHAR, pm_Remove);
        PeekMessage(M1, W, WM_SYSCHAR, WM_CHAR, pm_Remove);
    end;
var LV: TFastListView;
    LI: TListItem;
    I: Integer;
    E: Boolean;
    P: TParams;
    LVIdx: Integer;
    Ed: HWnd;
    Ch: Char;
begin
    LVIdx := -1;
    for I := 1 to High(FAttached) do
        if GetLV(I) = Sender then
        begin
            LVIdx := I;
            break;
        end;
    if LVIdx <= 0 then Exit;
    LV := Sender as TFastListView;
    TRY
        CASE Key OF
            VK_UP, VK_DOWN: //////////////////////////////////////////////////////////////
                if (GetKeyState(VK_SHIFT) >= 0) and (GetKeyState(VK_CONTROL) >= 0) then
                begin
                    E := LV.IsEditing;
                    if (LV.ItemFocused <> nil) then
                    begin
                        LI := LV.ItemFocused;
                        while LI <> nil do
                        begin
                            if (Key = VK_UP) and (LI.Index > 0) OR
                                (Key = VK_DOWN) and (LI.Index < LV.Items.Count - 1) then
                            begin
                                if Key = VK_UP then LI := LV.Items[LI.Index - 1]
                                else LI := LV.Items[LI.Index + 1];
                                if (LV.ColumnIndexByName('param_object') >= 0)
                                    and (LI['param_object'] <> '') then
                                begin
                                    P := Pointer(StrToInt(LI['param_object']));
                                    if (P <> nil) and (ParamsList.IndexOf(P) >= 0) then
                                    begin
                                        if (P.ParamData[StrToInt(LI['param_index'])].param_type = 3) and
                                        not P.ParamData[StrToInt(LI['param_index'])].read_only then
                                        begin //    
                                            break;
                                        end;
                                    end else LI := nil;
                                end
                                else
                                    LI := nil;
                            end
                            else
                                LI := nil;
                        end;
                        if E and (LI <> nil) then //----------------------------------------------------
                        begin //      
                            // .   
                            //      .

                            LV.Selected.EndEdit;
                            SupressChar(LV.Handle);

                            PostMessage(LV.Handle, WM_USER_KEYPRESS, Key, 0);
                            Key := 0;
                        end;
                    end;
                end;
            VK_RETURN: ///////////////////////////////////////////////////////////////////
                begin
                    if LV.IsEditing then //-------------------------------------------------
                    begin // :  
                        LI := LV.Selected;
                        LI.EndEdit;
                        SupressChar(LV.Handle);
                        Key := 0;
                    end
                    else //---------------------------------------------------
                        if not LV.ReadOnly then
                        begin //  :  ,   
                            LI := LV.ItemFocused;
                            if LI = nil then LI := LV.Selected;
                            P := Pointer(StrToInt(LI['param_object']));
                            if (P <> nil) and (ParamsList.IndexOf(P) >= 0) then
                            begin
                                if (P.ParamData[StrToInt(LI['param_index'])].param_type = 3) and
                                not P.ParamData[StrToInt(LI['param_index'])].read_only then
                                begin //    
                                    Key := 0;
                                    LI.EditCaption;
                                end;
                            end;
                        end;
                end;
            WORD('0')..WORD('9'), VK_DECIMAL, VK_SEPARATOR, VK_SUBTRACT, VK_ADD:
                begin
                    if not LV.IsEditing and not LV.ReadOnly then //-------------------------
                    begin //  :  ,  
                        LI := LV.ItemFocused;
                        if LI = nil then LI := LV.Selected;
                        P := Pointer(StrToInt(LI['param_object']));
                        if (P <> nil) and (ParamsList.IndexOf(P) >= 0) then
                        begin
                            if (P.ParamData[StrToInt(LI['param_index'])].param_type = 3) and
                            not P.ParamData[StrToInt(LI['param_index'])].read_only then
                            begin //    
                                SupressChar(LV.Handle);
                                LI.Caption := Char(Key);
                                LI.EditCaption;
                                Ed := LV.Perform(LVM_GETEDITCONTROL, 0, 0);
                                if (Ed <> 0) and (GetProp(Ed, 'ParamController') = 0) then
                                begin
                                    SetProp(Ed, 'ParamController', Integer(Self));
                                    SetProp(Ed, 'FastListView', Integer(LV));
                                    OldEdWndProc := Pointer(GetWindowLong(Ed, GWL_WNDPROC));
                                    SetWindowLong(Ed, GWL_WNDPROC, Integer(@NewEdWndProc));
                                end;
                                keybd_event(VK_RIGHT, 0, 0, 0);
                                keybd_event(VK_RIGHT, 0, 0, KEYEVENTF_KEYUP);
                                TRY
                                    Ch := Char(Key);
                                    LV_OnKeyChar(LV, Ch);
                                EXCEPT
                                END;
                                Key := 0;
                            end;
                        end;
                    end;
                end;
        END;
    EXCEPT
        //    - ,    
    END;
    if Assigned(OldOnKeyDown[LVIdx]) then
        OldOnKeyDown[LVIdx](Sender, Key, Shift);
end;

function TParamsController.LV_OnMessage(Sender: TObject;
    var Msg: TMessage): Boolean;
var LVIdx, I: Integer;
    LV: TFastListView;
    TopItem: Integer;
    P: TPoint;
    LI: TListItem;
    Key: Char;
begin
    Result := FALSE;
    CASE Msg.Msg OF
        WM_USER_REFRESH_PARAMS:
            begin
                LVIdx := Msg.wParam;
                I := Msg.LParam;
                LV := GetLV(LVIdx);
                TopItem := LV.TopItem.Index;

                Recalculate(FALSE);
                OutputParams;

                if (LV <> nil) and (I < LV.Count) then
                begin
                    LV.Selected := nil;
                    LV.Selected := LV.Items[I];
                    LV.ItemFocused := LV.Selected;
                    if TopItem < LV.Count then
                    begin
                        LV.Items[LV.Count - 1].MakeVisible(FALSE);
                        LV.Items[TopItem].MakeVisible(FALSE);
                    end
                    else
                        LV.Selected.MakeVisible(FALSE);
                end;
            end;
        WM_USER_KEYPRESS:
            begin
                keybd_event(Msg.wParam, 0, 0, 0);
                keybd_event(Msg.wParam, 0, KEYEVENTF_KEYUP, 0);
            end;
        WM_LBUTTONDOWN:
            if Msg.wParam = MK_LBUTTON then
            begin
                P.X := SmallInt(LoWord(Msg.LParam));
                P.Y := SmallInt(HiWord(Msg.LParam));
                if Sender is TFastListView then
                begin
                    //Result := TRUE;
                    LV := Sender as TFastListView;
                    LI := LV.GetItemAt(P.X, P.Y);
                    if LI <> nil then
                    begin
                        LV.Selected := LI;
                        // LV.Focused := TRUE;
                        LI.Focused := TRUE;
                        Key := #13;
                        LV_OnKeyChar(Sender, Key);
                    end;
                end;
            end;
    END;
end;

function TParamsController.NewSetOfPack(CountPack, nomenklid: Integer;
    const BatchName: String; batchid: Integer; InputReady: Boolean;
    FirstNum: Integer; var ConteinersToDestroy: String): Boolean;
//   ""   .
// ,         
//    (  ,   
//    ,      
// ).
var I, j: Integer;
    P0, P: TParams;
begin
    //Result := FALSE;
    P0 := nil;
    if nomenklid <> -1 then
        for I := 0 to ParamsList.Count - 1 do
        begin
            P0 := ParamsList[I];
            if P0.ParamSetData.nomenklid = nomenklid then
                break;
            P0 := nil;
        end;
    //if P0 = nil then Exit;  bn
    for I := 1 to CountPack do
    begin
        P := TParams.CreateOwned(Self);
        ParamsList.Add(P);
        P.Assign(P0);
        if DebugMode then
            P.DebugOutput('C:\debug_pack' + IntToStr(I));
        //--  ""  ,   -   
        if InputReady then
            for j := 0 to P.Count - 1 do
                if (P.Values[j] = 0) and (P.ParamData[j].param_type = 3) then
                    P.ParamData[j].param_ready := FALSE;
        P.ParamSetData.numcon := BatchName + '.' + IntToStr(FirstNum + I - 1);
        P.ParamSetData.packid := -1;
        P.ParamSetData.batchid := batchid;
        P.ParamSetData.Sign_final := TRUE;
    end;
    ConteinersToDestroy := '';
    if CountPack < 0 then
    begin
        for I := ParamsList.Count - 1 downto ParamsList.Count + CountPack do
        begin
            P := ParamsList[I];
            if P.ParamSetData.packid > 0 then
                ConteinersToDestroy := ConteinersToDestroy + ',' + IntToStr(P.ParamSetData.packid);
            P.Free;
            ParamsList.Delete(I);
        end;
    end;
    StrDel(ConteinersToDestroy, 1, 1);
    for I := 0 to ParamsList.Count - 1 do
    begin
        P := ParamsList[I];
        if P.ParamSetData.packid = 0 then
        begin
            P.ParamSetData.ignore_conditions := TRUE;
            P.ParamSetData.NoCalc := TRUE;
        end;
    end;
    DoInitInputVals := TRUE;
    TRY
        Recalculate(TRUE);
    FINALLY
        DoInitInputVals := FALSE;
    END;
    //Dump( 'c:\afterNewSetOfPack.txt' );
    Result := TRUE;
end;

type
    TSortParams = packed record
        L: TList;
        P: TParams;
        Op: PChar;
    end;
    PSortParams = ^TSortParams;

function CompareParamOrder(const Data: Pointer; const e1, e2: Dword): Integer;
var S: PSortParams;
    i1, i2, j1, j2: Integer;
    PD1, PD2: TParamData;
begin
    S := Data;
    i1 := Integer(S.L[e1]);
    i2 := Integer(S.L[e2]);
    j1 := S.P.ParamData[i1].Is_used_in_op(S.Op);
    j2 := S.P.ParamData[i2].Is_used_in_op(S.Op);
    PD1 := S.P.ParamData[i1];
    PD2 := S.P.ParamData[i2];
    Result := 0;
    if (j1 < PD1.param_ord_Count) and (j2 < PD2.param_ord_Count) then
        Result := Integer(PD1.param_orders[j1])
            - Integer(PD2.param_orders[j2]);
end;

procedure SwapItems(const Data: Pointer; const e1, e2: Dword);
var S: PSortParams;
    tmp: Pointer;
begin
    S := Data;
    tmp := S.L[e1];
    S.L[e1] := S.L[e2];
    S.L[e2] := tmp;
end;

procedure TParamsController.OutputParams(LVIdx: Integer);
//------------------------------------------------------------------------------
//   (  )  
//   fastlistview.
var LV: TFastListView;
    LI: TListItem;
    I, J, K: Integer;
    P: TParams;
    EmptyCOPList: Boolean;
    PD: TParamData;
    L: TList;
    SP: TSortParams;
begin
    LV := GetLV(LVIdx);
    if LV = nil then Exit;
    LV.Items.BeginUpdate;
    TRY
        if Assigned(OnBeforeOutput) then
            OnBeforeOutput(Self);
        if Clear_ListView_When_Output then
            LV.Items.Clear;
        EmptyCOPList := Code_Operation_List = '';
        for I := 0 to ParamsList.Count - 1 do
        begin
            P := ParamsList[I];
            if //(nomenklid = 0) OR

            (packid = 0) and (nomenklid <> 0) and
                (nomenklid = P.ParamSetData.nomenklid) OR

            (packid <> 0) AND (packid = P.ParamSetData.packid) AND (nomenklid = 0) OR

            (packid <> 0) AND (packid = P.ParamSetData.packid) AND (P.ParamSetData.nomenklid = nomenklid)

            {and
            (nomenklid = P.ParamSetData.nomenklid)}then
            begin
                //----------------------------------------------------------------------
                //   ,    .
                //      .  , 
                //   ,   hidden = TRUE.
                L := TList.Create;
                //Dump( 'c:\dump_params_output.txt' );
                TRY
                    //     
                    J := P.Count - 1;
                    if P.ConstCount > 0 then
                        inc(J, P.ConstCount);
                    for J := 0 to {P.Count-1} J do
                    begin
                        //if P.ID[ J ] = 0 then continue;
                        PD := P.ParamData[J];
                        if PD = nil then continue;
                        if (nomenklid <> 0) and (PD.nomenklid <> nomenklid) then
                            continue;
                        if (PD.Is_show_in_op(FCode_op[LVIdx]) >= 0) and
                            (PD.Is_used_in_op(FCode_Op[LVIdx]) >= 0)
                            or EmptyCOPList then
                            L.Add(Pointer(J));
                    end;
                    //      
                    SP.L := L;
                    SP.P := P;
                    SP.Op := PChar(FCode_Op[LVIdx]);
                    SortData(@SP, L.Count, CompareParamOrder, SwapItems);
                    for K := 0 to L.Count - 1 do
                    begin
                        J := Integer(L[K]);
                        PD := P.ParamData[J];
                        if PD = nil then continue;
                        if (PD.Is_show_in_op(FCode_op[LVIdx]) >= 0)
                            or EmptyCOPList then
                        begin
                            LI := LV.Items.Add;
                            LI['param_index'] := IntToStr(J);
                            LI['param_object'] := IntToStr(Integer(Pointer(P)));

                            if IsIndex(Column_Description) >= 0 then
                                LI[IsIndex(Column_Description)] := P.Descriptions[J]
                            else
                                if Column_Description <> '' then
                                    LI[Column_Description] := P.Descriptions[J];

                            if IsIndex(Column_Name) >= 0 then
                                LI[IsIndex(Column_Name)] := P.Names[J]
                            else
                                if Column_Name <> '' then
                                    LI[Column_Name] := P.Names[J];

                            if IsIndex(Column_Value) >= 0 then
                                LI[IsIndex(Column_Value)] := P.FormattedParameter(J)
                            else
                                if Column_Value <> '' then
                                    LI[Column_Value] := P.FormattedParameter(J);

                            if IsIndex(Column_Units) >= 0 then
                                LI[IsIndex(Column_Units)] := P.ParamData[J].Units
                            else
                                if Column_Units <> '' then
                                    LI[Column_Units] := P.ParamData[J].Units;
                            //--------------------------------------------------------------------
                            //    list view    
                            // OnOutputParam,  .
                            if Assigned(OnOutputParam) then
                                OnOutputParam(Self, P, J);
                        end;
                    end;
                FINALLY
                    L.Free;
                END;
                //----------------------------------------------------------------------
                //     
                break;
            end;
        end;
    FINALLY
        LV.Items.EndUpdate;
        if Assigned(OnAfterOutput) then
            OnAfterOutput(Self);
    END;
end;

procedure TParamsController.OutputParams;
var I: Integer;
begin
    if Owner = nil then Exit;
    if csLoading in Owner.ComponentState then Exit;
    for I := 1 to High(FLVName) do
        if GetLV(I) <> nil then
            OutputParams(I);
end;

function TParamsController.Param_By_Nomenclature(
    nomenklid: Integer; packid: Integer = 0): TVars;
//--------------------------------------------------------------------------
//    
var I: Integer;
    P: TParams;
begin
    Result := nil;
    for I := 0 to ParamsList.Count - 1 do
    begin
        P := ParamsList[I];
        if (P.ParamSetData.nomenklid = nomenklid) or (nomenklid = -1) then
        begin
            if (packid = 0) or (packid = P.ParamSetData.packid) then
            begin
                Result := P;
                break;
            end;
        end;
    end;
end;

procedure TParamsController.Recalculate(use_recalc: Boolean);
//------------------------------------------------------------------------------
//  . ,    :
//  ,      .  
//  ,  -    .
var ErrorStr: String;
    RaiseErr: Boolean;
    I, J: Integer;
    P: TParams;
    PD: TParamData;
begin
    TimePrepare := 0;
    TimeCalc := 0;
    for I := 0 to FCachedFormulas.Count - 1 do
        TObject(FCachedFormulas[I]).Free;
    FCachedFormulas.Clear;
    if csLoading in ComponentState then Exit;
    TRY
        for I := 0 to ParamsList.Count - 1 do
        begin
            P := ParamsList[I];
            for J := 0 to P.Count - 1 do
            begin
                PD := P.ParamData[J];
                if (PD.used_in_op_count > 0) and
                    ((PD.param_type = 2) or
                    (PD.param_type = 3) and
                    (use_recalc and PD.recalc) and
                    (PD.formula <> '') {and
                    (P.Values[ J ] = 0)}) then
                    PD.param_ready := FALSE
                else
                    if use_recalc then
                        PD.param_ready := TRUE;
                {if PD.recalc and (PD.param_type = 3) then
                  PD.param_ready := FALSE;}
            end;
        end;
        if Assigned(OnBeforeCalc) then
            OnBeforeCalc(Self);
        for I := 1 to High(FCode_op) do
            if (FCode_op[I] <> '') and ((GetLV(I) <> nil) and
                (GetLV(I).Owner <> nil) and (GetLV(I).Owner is TWinControl) and
                (GetLV(I).Owner as TWinControl).Visible
                or
                (GetLVName(I) = '!')) then
            begin
                if not Assigned(OnAllowSum) and RestrictSum_Batch then
                    ParamUnit.CalcParams(ParamsList, ErrorStr, TRUE, AllowSum_In_Batch_Only,
                        DoInitInputVals, FCode_op[I], nomenklid,
                        HasContainersInNomen(nomenklid))
                else
                    ParamUnit.CalcParams(ParamsList, ErrorStr, TRUE, OnAllowSum,
                        DoInitInputVals, FCode_op[I], nomenklid,
                        HasContainersInNomen(nomenklid));
            end;
        OutputParams;
    EXCEPT on E: Exception do
        begin
            RaiseErr := TRUE;
            if Assigned(OnError) then
            begin
                RaiseErr := FALSE;
                OnError(Self, E.Message, RaiseErr);
            end;
            if RaiseErr then
                raise;
        end;
    END;
    {ShowMessage( 'TimePrepare=' + IntToStr( TimePrepare ) + ', ' +
                 'TimeCalc=' + IntToStr( TimeCalc ) );}
end;

procedure TParamsController.ResetDependingInputParams(LVIdx: Integer;
    const ParamName, ParamFormula: String; nomenklid, batchid, packid: Integer);
//--------------------------------------------------------------------------
//   0  ,   :
    function FormulaReferenced(const formula, parameter: String;
        InSumOnly: Boolean = FALSE): Boolean;
    var I, J, SumLev, L: Integer;
    begin
        SumLev := 0;
        L := 0;
        I := 1;
        while I <= Length(formula) do
        begin
            CASE formula[I] OF
                '(': Inc(L);
                ')': begin
                        if SumLev = L then SumLev := 0;
                        Dec(L);
                    end;
                's', 'S':
                    begin
                        if comparetext(copy(formula, I, 4), 'sum(') = 0 then
                        begin
                            Inc(I, 3); Inc(L); SumLev := L;
                        end;
                    end;
            END;
            if not (formula[I] in [#1..' ', '+', '-', '*', '/', '<', '>', '=', '(', ')', ';',
                '0'..'9', '.', ',']) and
                (not InSumOnly or (SumLev > 0)) then
            begin
                for J := I + 1 to Length(formula) + 1 do
                begin
                    if (J > Length(formula)) or (formula[J] in [#1..' ', '+', '-', '*', '/', '<', '>', '=', '(', ')', ';',
                        '0'..'9', '.', ',']) then
                    begin
                        if ParamName = Copy(formula, I, J - I) then
                        begin
                            Result := TRUE; Exit;
                        end;
                    end;
                end;
            end;
            Inc(I);
        end;
        Result := FALSE;
    end;

    function FormulaReferencedISum(const formula, parameter: String): Boolean;
    begin
        Result := FormulaReferenced(formula, parameter, TRUE);
    end;

    function RemoveSpaces(const S: String): String;
    var I: Integer;
    begin
        Result := S;
        for I := Length(S) downto 1 do
            if S[I] in [#1..' '] then
                StrDel(Result, I, 1);
    end;

    function OwnFormula(const ParmName: String): String;
    var S, f: String;
        I, J: Integer;
        P: TParams;
    begin
        Result := '';
        S := '';
        for I := 0 to ParamsList.Count - 1 do
        begin
            P := ParamsList[I];
            J := P.IndexOf(ParmName);
            if J < 0 then continue;
            f := P.ParamData[J].formula;
            if pos(';', f) > 0 then
                f := Trim(Copy(f, 1, pos(';', f) - 1));
            if f = '' then continue;
            if (S <> '') and (S <> f) then Exit;
            S := f;
        end;
        Result := S;
    end;

var I, J, K, N: Integer;
    P: TParams;
    PD: TParamData;
    ListDepending, ListDepFormulas: TStringList;
    S, S1: String;

    ParamsList_DependList: TList;
    PL_DependFromList: TList;
begin
    ListDepending := TStringList.Create;
    ListDepFormulas := TStringList.Create;
    ParamsList_DependList := TList.Create;
    for I := 0 to ParamsList.Count - 1 do
    begin
        PL_DependFromList := TList.Create;
        ParamsList_DependList.Add(PL_DependFromList);
    end;
    TRY
        S := ParamFormula;
        I := pos(';', S);
        if I <= 0 then Exit;
        S := Copy(S, I + 1, MaxInt);
        while S <> '' do
        begin
            I := pos(',', S);
            if I <= 0 then
                I := pos(';', S);
            if I <= 0 then
            begin
                S1 := Trim(S);
                S := '';
            end
            else
            begin
                S1 := Trim(Copy(S, 1, I - 1));
                S := Copy(S, I + 1, MaxInt);
            end;
            if S1 <> '' then
            begin
                I := pos('=', S1);
                if (I <= 0) and (OwnFormula(S1) <> '') then
                begin
                    ListDepending.Add(S1);
                    ListDepFormulas.Add(OwnFormula(S1));
                end
                else
                begin
                    if I <= 0 then I := Length(S1) + 1;
                    ListDepending.Add(Trim(Copy(S1, 1, I - 1)));
                    ListDepFormulas.Add(Trim(Copy(S1, I + 1, MaxInt)));
                end;
            end;
        end;

        for I := 0 to ParamsList.Count - 1 do
        begin
            P := ParamsList[I];
            for J := 0 to P.Count - 1 do
            begin
                PD := P.ParamData[J];
                if ((PD.param_type = 3) or (PD.param_type = 2)) and
                    (ListDepending.IndexOf(P.Names[J]) >= 0) and
                    (PD.formula <> '') and
                    ((nomenklid = P.ParamSetData.nomenklid)
                    or (P.ParamSetData.nomenklid = 0)) and
                    ((batchid = P.ParamSetData.batchid)
                    or (P.ParamSetData.batchid = 0)) then
                begin
                    K := ListDepending.IndexOf(P.Names[J]);
                    if ((packid = 0) or (P.ParamSetData.packid = 0) or
                        (packid = P.ParamSetData.packid)) then
                    begin
                        if FormulaReferenced(RemoveSpaces(ListDepFormulas[K]), ParamName) then
                        begin
                            if (ListDepFormulas[K] <> '') and (PD.save_formula = '') then
                            begin
                                PD.save_formula := ' ' + PD.formula;
                                PD.formula := ListDepFormulas[K];
                            end;
                            if PD.formula <> '' then
                            begin
                                PD.param_ready := FALSE;
                                PL_DependFromList := ParamsList_DependList[I];
                                PL_DependFromList.Add(Pointer(J));
                            end;
                            //P.Values[ J ] := 0;
                        end;
                    end
                    else
                    begin
                        if FormulaReferencedISum(RemoveSpaces(ListDepFormulas[K]), ParamName) then
                        begin
                            if (ListDepFormulas[K] <> '') and (PD.save_formula = '') then
                            begin
                                PD.save_formula := ' ' + PD.formula;
                                PD.formula := ListDepFormulas[K];
                            end;
                            if PD.formula <> '' then
                            begin
                                P.ParamData[J].param_ready := FALSE;
                                PL_DependFromList := ParamsList_DependList[I];
                                PL_DependFromList.Add(Pointer(J));
                            end;
                            //P.Values[ J ] := 0;
                        end;
                    end;
                end;
            end;
        end;

        N := 1;
        while N > 0 do
        begin
            N := 0;
            for I := 0 to ParamsList.Count - 1 do
            begin
                P := ParamsList[I];
                PL_DependFromList := ParamsList_DependList[I];
                for J := 0 to P.Count - 1 do
                begin
                    for K := 0 to PL_DependFromList.Count - 1 do
                    begin
                        if ParamDependOnParam(P, J, Integer(PL_DependFromList[K])) and
                            (PL_DependFromList.IndexOf(Pointer(J)) < 0) and
                            (P.Names[J] <> ParamName) then
                        begin
                            PL_DependFromList.Add(Pointer(J));
                            Inc(N);
                            P.ParamData[J].param_ready := FALSE;
                        end;
                    end;
                end;
            end;
        end;

    FINALLY
        ListDepending.Free;
        ListDepFormulas.Free;
        for I := 0 to ParamsList_DependList.Count - 1 do
            TObject(ParamsList_DependList[I]).Free;
        ParamsList_DependList.Free;
    END;
end;

procedure TParamsController.SetActive(const Value: Boolean);
//------------------------------------------------------------------------------
//   Active
var I: Integer;
begin
    FActive := Value;
    for I := 1 to High(FLVName) do
        if Value then
            AttachEvents(I)
        else
            DetachEvents(I);
end;

procedure TParamsController.SetAutoRefresh(const Value: Boolean);
//------------------------------------------------------------------------------
//   AutoRefresh
begin
    FAutoRefresh := Value and Clear_ListView_When_Output;
end;

procedure TParamsController.SetClear_ListView_When_Output(
    const Value: Boolean);
//------------------------------------------------------------------------------
//   "   "
begin
    FClear_ListView_When_Output := Value;
    if not FClear_ListView_When_Output then
        AutoRefresh := FALSE
    else
        OutputParams;
end;

procedure TParamsController.SetCodeOp(const Index: Integer; const Value: String);
//------------------------------------------------------------------------------
//    
begin
    FCode_Op[Index] := Value;
end;

procedure TParamsController.SetColumn_Description(const Value: String);
//------------------------------------------------------------------------------
//     (  ) ,   
// Description (  1)
begin
    FColumn_Description := Value;
end;

procedure TParamsController.SetColumn_Name(const Value: String);
//------------------------------------------------------------------------------
//     (  ) ,   
// Name (  2)
begin
    FColumn_Name := Value;
end;

procedure TParamsController.SetColumn_Units(const Value: String);
//------------------------------------------------------------------------------
//    (  ),    Ed_ism
// (  , ..   ).
begin
    FColumn_Units := Value;
end;

procedure TParamsController.SetColumn_Value(const Value: String);
//------------------------------------------------------------------------------
//     (  ) ,   
// Value (  0).    
//        ,   
//  , ..     0 (, 
//     Order,  ).    
//       .
begin
    FColumn_Value := Value;
end;

procedure TParamsController.Setpackid(const Value: Integer);
//------------------------------------------------------------------------------
//  packid -  ,      
//  ParamsList.
begin
    //if Fpackid = Value then Exit;
    Fpackid := Value;
    if AutoRefresh then
        OutputParams;
end;

procedure TParamsController.SetLV(LVIdx: Integer; Value: TFastListView);
//------------------------------------------------------------------------------
//    TFastListView.     ,
//    (!!!    , ..  
// :  TFastListView     design-time  
//   ). . GetListView .
begin
    if Value = nil then
    begin
        DetachEvents(LVIdx);
        FLVName[LVIdx] := '';
    end
    else
    begin
        DetachEvents(LVIdx);
        if (Value.Owner <> nil) and (Value.Owner <> Owner) then
            FLVName[LVIdx] := Value.Owner.Name + '.' + Value.Name
        else
            FLVName[LVIdx] := Value.Name;
        AttachEvents(LVIdx);
    end;
end;

procedure TParamsController.SetLVName(const Index: Integer;
    const Value: String);
begin
    FLVName[Index] := Value;
end;

procedure TParamsController.Setnomenklid(const Value: Integer);
//------------------------------------------------------------------------------
//  nomenklid - id ,    
var I: Integer;
    P: TParams;
begin
    Fnomenklid := Value;
    for I := 0 to ParamsList.Count - 1 do
    begin
        P := ParamsList[I];
        P.Assign(P);
    end;
    if AutoRefresh then
        OutputParams;
end;

procedure TParamsController.SetOnAfterOutput(const Value: TNotifyEvent);
//------------------------------------------------------------------------------
//   OnAfterOutput
begin
    FOnAfterOutput := Value;
end;

procedure TParamsController.SetOnAllowSum(const Value: TAllowSum);
//------------------------------------------------------------------------------
//   OnAllowSum
begin
    FOnAllowSum := Value;
    //Recalculate;
end;

procedure TParamsController.SetOnBeforeCalc(const Value: TNotifyEvent);
begin
    FOnBeforeCalc := Value;
end;

procedure TParamsController.SetOnBeforeOutput(const Value: TNotifyEvent);
//------------------------------------------------------------------------------
//   OnBeforeOutput
begin
    FOnBeforeOutput := Value;
end;

procedure TParamsController.SetOnEdit(const Value: TOnEdit);
//------------------------------------------------------------------------------
//   OnEdit
begin
    FOnEdit := Value;
end;

procedure TParamsController.SetOnError(const Value: TOnError);
//------------------------------------------------------------------------------
//   OnError
begin
    FOnError := Value;
end;

procedure TParamsController.SetOnOutputParam(const Value: TOnOutputParam);
//------------------------------------------------------------------------------
//   OnOutputParam
begin
    FOnOutputParam := Value;
end;

procedure TParamsController.SetParamsList(const Value: TList);
//------------------------------------------------------------------------------
//   ParamsList (,     
//    ParamUnit).
begin
    if FParamsList = Value then Exit;
    if FParamsList <> nil then
        Free_ParamsList;
    FParamsList := Value;
    //----------------------------------------------------------------------------
    //       list view.
    OutputParams;
end;

procedure TParamsController.SetRestrictSum_Batch(const Value: Boolean);
begin
    if FRestrictSum_Batch = Value then Exit;
    FRestrictSum_Batch := Value;
    //Recalculate;
end;

procedure TParamsController.Set_batchid(Abatchid: Integer; AllPacks: Boolean);
//--------------------------------------------------------------------------
//  ""  batchid 
var I: Integer;
    P: TParams;
begin
    for I := 0 to FParamsList.Count - 1 do
    begin
        P := FParamsList[I];
        if AllPacks or (P.ParamSetData.packid < 0) then
            P.ParamSetData.batchid := Abatchid;
    end;
end;

function TParamsController.SumCached(P: TParams; Idx, fPos: Integer; var F: String;
    calc_by_batch: Boolean): Boolean;
//------------------------------------------------------------------------------
//   ,   -    
// N       N     
//  (     , ..   
//   N-1 ,   sum(...). )
var fSumFormula: String;
    fLastIdx: Integer;
    function SumFormula: String;
    var CalcLevel: Integer;
    begin
        if fSumFormula = '' then
        begin
            CalcLevel := 0;
            fLastIdx := fPos;
            while fLastIdx <= Length(F) do
            begin
                CASE F[fLastIdx] OF
                    '(': Inc(CalcLevel);
                    ')': begin
                            Dec(CalcLevel);
                            if CalcLevel = 0 then break;
                        end;
                else ;
                END;
                Inc(fLastIdx);
            end;
            fSumFormula := Copy(F, fPos, fLastIdx - fPos + 1);
        end;
        Result := fSumFormula;
    end;

var I: Integer;
    CF: TCacheFormula;
begin
    Result := FALSE;
    for I := 0 to FCachedFormulas.Count - 1 do
    begin
        CF := FCachedFormulas[I];
        if (CF.bachid = P.ParamSetData.batchid) and
            (CF.nomenklid = P.ParamSetData.nomenklid) and
            (CF.sumbybatch = calc_by_batch) and
            (CF.SumFormula = SumFormula) then
        begin
            F := Copy(F, 1, fPos - 1) + CF.RsltStr +
                Copy(F, fLastIdx + 1, MaxInt);
            Result := TRUE;
            break;
        end;
    end;
end;

procedure TParamsController.Dump(const filename: String);
var SL: TStringList;
    s: String;
    I, J, K: Integer;
    P: TParams;
    PD: TParamData;
    L: TList;
    //SP: TSortParams;
    ii: Integer;
begin
    SL := TStringList.Create;
    TRY
        for I := 0 to ParamsList.Count - 1 do
        begin
            P := ParamsList[I];
            if TRUE
                {(packid = 0) and (nomenklid <> 0) and
            (nomenklid = P.ParamSetData.nomenklid) OR

            (packid <> 0) AND (packid = P.ParamSetData.packid)}then
            begin
                //----------------------------------------------------------------------
                //   ,    .
                //      .  , 
                //   ,   hidden = TRUE.
                L := TList.Create;
                TRY
                    //     
                    for J := 0 to P.Count - 1 do
                    begin
                        //if P.ID[ J ] = 0 then continue;
                        PD := P.ParamData[J];
                        if PD = nil then continue;
                        if TRUE then
                            L.Add(Pointer(J));
                    end;
                    {//      
                    SP.L := L;
                    SP.P := P;
                    SP.Op := '';
                    SortData( @ SP, L.Count, CompareParamOrder, SwapItems );}
                    for K := 0 to L.Count - 1 do
                    begin
                        J := Integer(L[K]);
                        PD := P.ParamData[J];
                        if PD = nil then continue;
                        if TRUE then
                        begin
                            s := IntToStr(J) + ':' +
                                P.ParamSetData.numcon +
                                ' nom:' + IntToStr(P.ParamSetData.nomenklid) +
                                ' bat:' + IntToStr(P.ParamSetData.batchid);

                            s := s + ' ' + P.Names[J]
                                //+ ' ' + P.Descriptions[ J ]
                            + ' ' + P.FormattedParameter(J)
                                + ' ' + P.ParamData[J].Units;

                            s := s + ' show:';
                            for ii := 0 to PD.show_in_op_count - 1 do
                                s := s + ' ' + Operations[PD.show_in_ops[ii]];
                            s := s + ' use:';
                            for ii := 0 to PD.used_in_op_count - 1 do
                                s := s + ' ' + Operations[PD.used_in_ops[ii]];

                            s := s + ' format:' + FloatToStr(PD.param_format);
                            SL.Add(s);
                        end;
                    end;
                FINALLY
                    L.Free;
                END;
            end;
        end;
    FINALLY
        SL.SaveToFile(filename);
        SL.Free;
    END;
end;

procedure TParamsController.SetPackage_TableName(const Value: String);
begin
    if FPackage_TableName = Value then Exit;
    FPackage_TableName := Value;
end;

procedure TParamsController.CopyContainerAnalysesFromBatchAnalyses(only_zeroes: Boolean);
var i, j, k: Integer;
    P: TParams;
    D: TParamDef;
    from: String;
begin
    for i := 0 to ParamsList.Count - 1 do
    begin
        P := ParamsList[I];
        if P.ParamSetData.nomenklid <> nomenklid then
            continue;
        if DebugMode then
            P.DebugOutput('C:\parsofconteiner_debug.txt');
        for j := 0 to P.Count - 1 do
        begin
            D := P.ParamData[j].Def;
            from := D.Analyse_copy_from_param;
            if (from <> '') and
                (not only_zeroes or (P.Values[j] = 0)) then
            begin
                if DebugMode then
                    LogFileOutput('c:\test_init_analyses.txt',
                        'P.ParamData[' + IntToStr(j) + '].Def.Param_name=' +
                        P.ParamData[j].Def.Param_name + ' <= init from ' +
                        from);
                for k := 0 to P.Count - 1 do
                begin
                    if DebugMode then
                        LogFileOutput('c:\test_init_analyses.txt',
                            'P.ParamData[' + IntToStr(k) + '].Def.Param_name=' +
                            P.ParamData[k].Def.Param_name);
                    if //P.ParamData[k].Def.Is_Batch_Analyse and
                    (P.ParamData[k].Def.Param_name = from) then
                    begin
                        if DebugMode then
                            LogFileOutput('c:\test_init_analyses.txt',
                                'P.ParamData[' + IntToStr(k) + '].Def.Param_name=' +
                                P.ParamData[k].Def.Param_name + ' Value = ' +
                                FloatToStr(P.Values[k]));
                        P.Values[j] := P.Values[k];
                    end;
                end;
            end;
        end;
    end;
end;

end.
