//**  ,     VCL Delphi.
//**  Vladimir Kladov, 2000-2010
//**
//**      .
//**     /     
//**  2          
//**   GNU,  Free Software Foundation.
//**
//**       ,    
//** ,       ,   
//**         
//**   .      
//**     GNU.
//**
//**          
//**   GNU.     ,     Free
//** Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
//** MA 02111-1307, USA.
//******************************************************************************

unit ReplaceExistingKnopka;

interface

{$I VCL_VERSIONS.INC}

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  comctrls, imglist, commctrl, StdCtrls, menus, Extctrls, registry,
  math, buttons,
  {$IFDEF _D6} designintf {$IFDEF IN_PACKAGE}, DesignEditors {$ENDIF}
  {$ELSE}
     {$IFDEF _D7orHigher} designintf {$IFDEF IN_PACKAGE}, DesignEditors {$ENDIF}
     {$ELSE} dsgnintf {$ENDIF}
  {$ENDIF},
  Knopka;

resourceString
    {$IFDEF RUSSIAN_LANGUAGE}
    CompEditReplaceKnopka = '   TKnopka/TPnKnopka';
    OnClickCallVerb = ' /   OnClick ( !)';
    Err0101 = ' TKnopka   .  .';
    WhatToReplace_Caption = '     TKnopka/TPnKnopka';
    ButtonReplace_Caption = '';
    AllItems_text = '()';
    OpenProjectFolderVerb = '  ';
    {$ELSE not RUSSIAN_LANGUAGE}
    CompEditReplaceKnopka = 'Replace button(s) onto TKnopka/TPnKnopka';
    OnClickCallVerb = 'OnClick create / go to (Double click!)';
    Err0101 = 'The owner of TKnopka/TPnKnopka is not TForm. Can''t execute.';
    WhatToReplace_Caption = 'Select button(s) to replace onto TKnopka/TPnKnopka';
    ButtonReplace_Caption = 'Replace';
    AllItems_text = '(All)';
    OpenProjectFolderVerb = 'Open project folder';
    {$ENDIF not RUSSIAN_LANGUAGE}

type
  TReplaceKnopkaEditor = class( TDefaultEditor )
  private
  protected
    procedure ReplaceKnopka;
    procedure DoReplace( B: TButton; K: TKnopka ); overload;
    procedure DoReplace( B: TButton; K: TPnKnopka ); overload;
    procedure DoReplace( B: TSpeedButton; K: TKnopka ); overload;
    procedure DoReplace( B: TSpeedButton; K: TPnKnopka ); overload;
    procedure DoOpenProjectFolder;
  public
    procedure ExecuteVerb(Index: Integer); override;
    function GetVerb(Index: Integer): string; override;
    function GetVerbCount: Integer; override;
  end;

  TKnopka_Access = class( TKnopka )
  end;

  TPnKnopka_Access = class( TPnKnopka )
  end;

procedure Register;

implementation

uses
  WhatToReplace, shellapi;

procedure Register;
begin
  RegisterComponentEditor( TKnopka, TReplaceKnopkaEditor );
  RegisterComponentEditor( TPnKnopka, TReplaceKnopkaEditor );
end;

{ TReplaceKnopkaEditor }

procedure TReplaceKnopkaEditor.DoReplace(B: TButton; K: TKnopka);
var NewName, OldName: String;
    NewParent: TWinControl;
    NewBoundsRect: TRect;
    NewTabOrder: Integer;
    NewComponentIndex: Integer;
    I: SmallInt;
    {$IFDEF _D6orHigher}
    {$ELSE}
    BI: IComponent;
    {$ENDIF}
    BB: TBitBtn;
    BK: TKnopka;
    Bmp: TBitmap;
    s: String;
begin
  try

  with K do
  BEGIN

    if B = nil then Exit;
    NewName := B.Name;
    NewParent := B.Parent;
    NewBoundsRect := B.BoundsRect;
    NewTabOrder := B.TabOrder;
    NewComponentIndex := B.ComponentIndex;
    s := B.Caption;
    //if  s = '' then s := ' ';

    OldName := '';
    I := 1;
    while I > 0 do
    begin
      try
        OldName := NewName + '_Old' + IntToStr( I );
        B.Name := OldName;
        I := -1;
      except
      end;
      Inc( I );
    end;
    Name := NewName;

    if NewParent <> nil then
      Parent := NewParent
    else
    if B.ParentWindow <> 0 then
      ParentWindow := B.ParentWindow
    else
    begin
      {$IFDEF _D6orHigher}
      //ParentWindow := Designer.ContainerWindow.Handle;
      {$ELSE}
      ParentWindow := Designer.ContainerWindow.Handle;
      {$ENDIF}
      //windows.Beep( 400, 200 );
    end;

    Caption := s;
    //Color := B.Color;
    Cursor := B.Cursor;
    Font.Assign( B.Font );
    //ParentColor := B.ParentColor;
    ParentFont := B.ParentFont;
    Constraints.MaxHeight := B.Constraints.MaxHeight;
    Constraints.MaxWidth := B.Constraints.MaxWidth;
    Constraints.MinHeight := B.Constraints.MinHeight;
    Constraints.MinWidth := B.Constraints.MinWidth;
    Anchors := B.Anchors;

    DragCursor := B.DragCursor;
    DragKind := B.DragKind;
    DragMode := B.DragMode;
    Enabled := B.Enabled;
    Hint := B.Hint;
    ParentBiDiMode := B.ParentBiDiMode;
    ParentShowHint := B.ParentShowHint;
    PopupMenu := B.PopupMenu;
    ShowHint := B.showHint;
    Visible := B.Visible;
    Tag := B.Tag;
    ModalResult := B.ModalResult;

    OnClick := B.OnClick;
    OnMouseDown := B.OnMouseDown;
    OnMouseUp := B.OnMouseUp;
    OnMouseMove := B.OnMouseMove;
    OnKeyDown := B.OnKeyDown;
    OnKeyUp := B.OnKeyUp;
    OnKeyPress := B.OnKeyPress;
    OnEnter := B.OnEnter;
    OnExit := B.OnExit;
    OnContextPopup := B.OnContextPopup;
    //OnResize := B.OnResize;
    OnStartDock := B.OnStartDock;
    OnStartDrag := B.OnStartDrag;

    if  B is TBitBtn then
    begin
        BB := B as TBitBtn;
        CASE BB.Layout OF
        blGlyphLeft:   Layout := klLabelRight;
        blGlyphRight:  Layout := klLabelLeft;
        blGlyphTop:    Layout := klLabelBottom;
        blGlyphBottom: Layout := klLabelTop;
        END;
        if  (BB.Glyph <> nil) and not BB.Glyph.Empty then
        begin
            Bmp := TBitmap.Create;
            TRY
                Bmp.Assign( BB.Glyph );
                if  BB.NumGlyphs > 1 then
                    Bmp.Width := max( 1, Bmp.Width div BB.NumGlyphs );
                Picture.Assign( Bmp );
            FINALLY
                Bmp.Free;
            END;
        end;
    end;

    if  B is TKnopka then
    begin
        BK := B as TKnopka;
        DisableFormWhileClick := BK.DisableFormWhileClick;
        DisableWhileClick := BK.DisableWhileClick;
        FocusDotFrameAroundText := BK.FocusDotFrameAroundText;
        ImageIndex := BK.ImageIndex;
        ImageList := BK.ImageList;
        ImagesDisabled := BK.ImagesDisabled;
        Layout := BK.Layout;
        MarginsPic := BK.MarginsPic;
        MarginsText := BK.MarginsText;
        RepeatDelay := BK.RepeatDelay;
        RepeatFirst := BK.RepeatFirst;
        RepeatPress := BK.RepeatPress;
        TextAlignment := BK.TextAlignment;
        TextScale := BK.TextScale;
        Wrapping := BK.Wrapping;
    end;

    Left := NewBoundsRect.Left;
    Top := NewBoundsRect.Top;
    Width := NewBoundsRect.Right - NewBoundsRect.Left;
    Height := NewBoundsRect.Bottom - NewBoundsRect.Top;

    TabOrder := NewTabOrder;
    ComponentIndex := NewComponentIndex;

    if TRUE {DeleteOld} then
    begin
      try

        {$IFDEF _D6orHigher}
        B.Free;
        {$ELSE}
        BI := MakeIComponent( B );
        BI.DestroyObject;
        {$ENDIF}

      finally
      end;
      Name := OldName;
      Name := NewName;
    end;

    END;

    //windows.Beep( 600, 600 );
  except
  end;
end;

procedure TReplaceKnopkaEditor.DoReplace(B: TButton; K: TPnKnopka);
var NewName, OldName: String;
    NewParent: TWinControl;
    NewBoundsRect: TRect;
    NewTabOrder: Integer;
    NewComponentIndex: Integer;
    I: SmallInt;
    {$IFDEF _D6orHigher}
    {$ELSE}
    BI: IComponent;
    {$ENDIF}
    BB: TBitBtn;
    Bmp: TBitmap;
    BK: TKnopka;
    s: String;
begin
  //LogFileOutPut( 'C:\test_repl_knopka.txt', 'Start replaceing' );
  try

  with K do
  BEGIN

    if B = nil then Exit;
    NewName := B.Name;
    NewParent := B.Parent;
    NewBoundsRect := B.BoundsRect;
    NewTabOrder := B.TabOrder;
    NewComponentIndex := B.ComponentIndex;

    s := B.Caption;
    //if  s = '' then s := ' ';

    OldName := '';
    I := 1;
    while I > 0 do
    begin
      try
        OldName := NewName + '_Old' + IntToStr( I );
        B.Name := OldName;
        I := -1;
      except
      end;
      Inc( I );
    end;
    Name := NewName;

    if NewParent <> nil then
      Parent := NewParent
    else
    if B.ParentWindow <> 0 then
      ParentWindow := B.ParentWindow
    else
    begin
      {$IFDEF _D6orHigher}
      //ParentWindow := Designer.ContainerWindow.Handle;
      {$ELSE}
      ParentWindow := Designer.ContainerWindow.Handle;
      {$ENDIF}
      //windows.Beep( 400, 200 );
    end;

    Action := B.Action;
    Caption := s;
    //Color := B.Color;
    Cursor := B.Cursor;
    Font.Assign( B.Font );
    //ParentColor := B.ParentColor;
    ParentFont := B.ParentFont;
    Constraints.MaxHeight := B.Constraints.MaxHeight;
    Constraints.MaxWidth := B.Constraints.MaxWidth;
    Constraints.MinHeight := B.Constraints.MinHeight;
    Constraints.MinWidth := B.Constraints.MinWidth;
    Anchors := B.Anchors;

    DragCursor := B.DragCursor;
    DragKind := B.DragKind;
    DragMode := B.DragMode;
    Enabled := B.Enabled;
    Hint := B.Hint;
    ParentBiDiMode := B.ParentBiDiMode;
    ParentShowHint := B.ParentShowHint;
    PopupMenu := B.PopupMenu;
    ShowHint := B.showHint;
    Visible := B.Visible;
    Tag := B.Tag;
    ModalResult := B.ModalResult;

    OnClick := B.OnClick;
    OnMouseDown := B.OnMouseDown;
    OnMouseUp := B.OnMouseUp;
    OnMouseMove := B.OnMouseMove;
    OnKeyDown := B.OnKeyDown;
    OnKeyUp := B.OnKeyUp;
    OnKeyPress := B.OnKeyPress;
    OnEnter := B.OnEnter;
    OnExit := B.OnExit;
    OnContextPopup := B.OnContextPopup;
    //OnResize := B.OnResize;
    OnStartDock := B.OnStartDock;
    OnStartDrag := B.OnStartDrag;

    if  B is TBitBtn then
    begin
        BB := B as TBitBtn;
        CASE BB.Layout OF
        blGlyphLeft:   Layout := klLabelRight;
        blGlyphRight:  Layout := klLabelLeft;
        blGlyphTop:    Layout := klLabelBottom;
        blGlyphBottom: Layout := klLabelTop;
        END;
        //LogFileOutPut( 'C:\test_repl_knopka.txt', 'BB is TBitBtn' );
        if  (BB.Glyph <> nil) then
        begin
            //LogFileOutPut( 'C:\test_repl_knopka.txt', 'BB.Glyph <> nil' );
        end;
        if  (BB.Glyph <> nil) then
        if not BB.Glyph.Empty then
        begin
            //LogFileOutPut( 'C:\test_repl_knopka.txt', 'Create Bmp' );
            Bmp := TBitmap.Create;
            TRY
                Bmp.Assign( BB.Glyph );
                //LogFileOutPut( 'C:\test_repl_knopka.txt', 'Bmp.SaveToFile' );
                Bmp.SaveTofile( 'c:\test_bitbtn.bmp' );
                if  BB.NumGlyphs > 1 then
                begin
                    //LogFileOutPut( 'C:\test_repl_knopka.txt', 'Bmp.Width :=' );
                    Bmp.Width := max( 1, Bmp.Width div BB.NumGlyphs );
                end;
                //LogFileOutPut( 'C:\test_repl_knopka.txt', 'Picture.Graphic.Assign( Bmp );' );
                Picture.Assign( Bmp );
            FINALLY
                //LogFileOutPut( 'C:\test_repl_knopka.txt', 'finally' );
                Bmp.Free;
            END;
        end;
    end;

    if  B is TKnopka then
    begin
        BK := B as TKnopka;
        DisableFormWhileClick := BK.DisableFormWhileClick;
        DisableWhileClick := BK.DisableWhileClick;
        FocusDotFrameAroundText := BK.FocusDotFrameAroundText;
        ImageIndex := BK.ImageIndex;
        ImageList := BK.ImageList;
        ImagesDisabled := BK.ImagesDisabled;
        Layout := BK.Layout;
        MarginsPic := BK.MarginsPic;
        MarginsText := BK.MarginsText;
        RepeatDelay := BK.RepeatDelay;
        RepeatFirst := BK.RepeatFirst;
        RepeatPress := BK.RepeatPress;
        TextAlignment := BK.TextAlignment;
        TextScale := BK.TextScale;
        Wrapping := BK.Wrapping;
        StdImage := BK.StdImage;
        StdImgShowSize := BK.StdImgShowSize;
    end;

    Left := NewBoundsRect.Left;
    Top := NewBoundsRect.Top;
    Width := NewBoundsRect.Right - NewBoundsRect.Left;
    Height := NewBoundsRect.Bottom - NewBoundsRect.Top;

    TabOrder := NewTabOrder;
    ComponentIndex := NewComponentIndex;

    if TRUE {DeleteOld} then
    begin
      try

        {$IFDEF _D6orHigher}
        B.Free;
        {$ELSE}
        BI := MakeIComponent( B );
        BI.DestroyObject;
        {$ENDIF}

      finally
      end;
      Name := OldName;
      Name := NewName;
    end;

    END;

    //windows.Beep( 600, 600 );
  except
  end;
end;

procedure TReplaceKnopkaEditor.DoReplace(B: TSpeedButton; K: TKnopka);
var NewName, OldName: String;
    NewParent: TWinControl;
    NewBoundsRect: TRect;
    //NewTabOrder: Integer;
    NewComponentIndex: Integer;
    I: SmallInt;
    {$IFDEF _D6orHigher}
    {$ELSE}
    BI: IComponent;
    {$ENDIF}
    //BB: TBitBtn;
    //BK: TKnopka;
    Bmp: TBitmap;
    s: String;
begin
  try

  with K do
  BEGIN

    if B = nil then Exit;
    NewName := B.Name;
    NewParent := B.Parent;
    NewBoundsRect := B.BoundsRect;
    //NewTabOrder := B.TabOrder;
    NewComponentIndex := B.ComponentIndex;
    s := B.Caption;
    //if  s = '' then s := ' ';

    OldName := '';
    I := 1;
    while I > 0 do
    begin
      try
        OldName := NewName + '_Old' + IntToStr( I );
        B.Name := OldName;
        I := -1;
      except
      end;
      Inc( I );
    end;
    Name := NewName;

    if NewParent <> nil then
      Parent := NewParent
    else
    {if B.ParentWindow <> 0 then
      ParentWindow := B.ParentWindow
    else}
    begin
      {$IFDEF _D6orHigher}
      //ParentWindow := Designer.ContainerWindow.Handle;
      {$ELSE}
      ParentWindow := Designer.ContainerWindow.Handle;
      {$ENDIF}
      //windows.Beep( 400, 200 );
    end;

    Caption := s;
    //Color := B.Color;
    Cursor := B.Cursor;
    Font.Assign( B.Font );
    //ParentColor := B.ParentColor;
    ParentFont := B.ParentFont;
    Constraints.MaxHeight := B.Constraints.MaxHeight;
    Constraints.MaxWidth := B.Constraints.MaxWidth;
    Constraints.MinHeight := B.Constraints.MinHeight;
    Constraints.MinWidth := B.Constraints.MinWidth;
    Anchors := B.Anchors;

    //DragCursor := B.DragCursor;
    //DragKind := B.DragKind;
    //DragMode := B.DragMode;
    Enabled := B.Enabled;
    Hint := B.Hint;
    ParentBiDiMode := B.ParentBiDiMode;
    ParentShowHint := B.ParentShowHint;
    PopupMenu := B.PopupMenu;
    ShowHint := B.showHint;
    Visible := B.Visible;
    Tag := B.Tag;

    OnClick := B.OnClick;
    OnMouseDown := B.OnMouseDown;
    OnMouseUp := B.OnMouseUp;
    OnMouseMove := B.OnMouseMove;
    //OnKeyDown := B.OnKeyDown;
    //OnKeyUp := B.OnKeyUp;
    //OnKeyPress := B.OnKeyPress;
    //OnEnter := B.OnEnter;
    //OnExit := B.OnExit;
    //OnContextPopup := B.OnContextPopup;
    //OnResize := B.OnResize;
    //OnStartDock := B.OnStartDock;
    //OnStartDrag := B.OnStartDrag;
    //ModalResult := B.ModalResult;

    TabStop := FALSE;
    SpeedButton := TRUE;

    CASE B.Layout OF
    blGlyphLeft:   Layout := klLabelRight;
    blGlyphRight:  Layout := klLabelLeft;
    blGlyphTop:    Layout := klLabelBottom;
    blGlyphBottom: Layout := klLabelTop;
    END;
    if  (B.Glyph <> nil) and not B.Glyph.Empty then
    begin
        Bmp := TBitmap.Create;
        TRY
            Bmp.Assign( B.Glyph );
            if  B.NumGlyphs > 1 then
                Bmp.Width := max( 1, Bmp.Width div B.NumGlyphs );
            Picture.Assign( Bmp );
        FINALLY
            Bmp.Free;
        END;
    end;

    Left := NewBoundsRect.Left;
    Top := NewBoundsRect.Top;
    Width := NewBoundsRect.Right - NewBoundsRect.Left;
    Height := NewBoundsRect.Bottom - NewBoundsRect.Top;

    //TabOrder := NewTabOrder;
    ComponentIndex := NewComponentIndex;

    if TRUE {DeleteOld} then
    begin
      try

        {$IFDEF _D6orHigher}
        B.Free;
        {$ELSE}
        BI := MakeIComponent( B );
        BI.DestroyObject;
        {$ENDIF}

      finally
      end;
      Name := OldName;
      Name := NewName;
    end;

    END;

    //windows.Beep( 600, 600 );
  except
  end;
end;

procedure TReplaceKnopkaEditor.DoOpenProjectFolder;
var s, s1, s2, s3: String;
    FD: {$IFDEF _D6orHigher} IDesigner {$ELSE} IFormDesigner {$ENDIF};
begin
    FD := inherited GetDesigner;
    if  FD <> nil then
    begin
        FD.ModuleFileNames( s1, s2, s3 );
        s := '';
        if  s1 <> '' then
            s := s1
        else
        if  s2 <> '' then
            s := s2
        else
        if  s3 <> '' then
            s := s3;
    end;
    if  s <> '' then
    begin
        s := ExtractFilePath( s );
        ShellExecute( 0, 'open', PChar( s ), nil, nil, SW_SHOW );
    end;
end;

procedure TReplaceKnopkaEditor.DoReplace(B: TSpeedButton; K: TPnKnopka);
var NewName, OldName: String;
    NewParent: TWinControl;
    NewBoundsRect: TRect;
    //NewTabOrder: Integer;
    NewComponentIndex: Integer;
    I: SmallInt;
    {$IFDEF _D6orHigher}
    {$ELSE}
    BI: IComponent;
    {$ENDIF}
    //BB: TBitBtn;
    //BK: TKnopka;
    Bmp: TBitmap;
    s: String;
begin
  try

  with K do
  BEGIN

    if B = nil then Exit;
    NewName := B.Name;
    NewParent := B.Parent;
    NewBoundsRect := B.BoundsRect;
    //NewTabOrder := B.TabOrder;
    NewComponentIndex := B.ComponentIndex;
    s := B.Caption;
    //if  s = '' then s := ' ';

    OldName := '';
    I := 1;
    while I > 0 do
    begin
      try
        OldName := NewName + '_Old' + IntToStr( I );
        B.Name := OldName;
        I := -1;
      except
      end;
      Inc( I );
    end;
    Name := NewName;

    if NewParent <> nil then
      Parent := NewParent
    else
    {if B.ParentWindow <> 0 then
      ParentWindow := B.ParentWindow
    else}
    begin
      {$IFDEF _D6orHigher}
      //ParentWindow := Designer.ContainerWindow.Handle;
      {$ELSE}
      ParentWindow := Designer.ContainerWindow.Handle;
      {$ENDIF}
      //windows.Beep( 400, 200 );
    end;

    Caption := s;
    //Color := B.Color;
    Cursor := B.Cursor;
    Font.Assign( B.Font );
    //ParentColor := B.ParentColor;
    ParentFont := B.ParentFont;
    Constraints.MaxHeight := B.Constraints.MaxHeight;
    Constraints.MaxWidth := B.Constraints.MaxWidth;
    Constraints.MinHeight := B.Constraints.MinHeight;
    Constraints.MinWidth := B.Constraints.MinWidth;
    Anchors := B.Anchors;

    //DragCursor := B.DragCursor;
    //DragKind := B.DragKind;
    //DragMode := B.DragMode;
    Enabled := B.Enabled;
    Hint := B.Hint;
    ParentBiDiMode := B.ParentBiDiMode;
    ParentShowHint := B.ParentShowHint;
    PopupMenu := B.PopupMenu;
    ShowHint := B.showHint;
    Visible := B.Visible;
    Tag := B.Tag;

    OnClick := B.OnClick;
    OnMouseDown := B.OnMouseDown;
    OnMouseUp := B.OnMouseUp;
    OnMouseMove := B.OnMouseMove;
    //OnKeyDown := B.OnKeyDown;
    //OnKeyUp := B.OnKeyUp;
    //OnKeyPress := B.OnKeyPress;
    //OnEnter := B.OnEnter;
    //OnExit := B.OnExit;
    //OnContextPopup := B.OnContextPopup;
    //OnResize := B.OnResize;
    //OnStartDock := B.OnStartDock;
    //OnStartDrag := B.OnStartDrag;
    //ModalResult := B.ModalResult;

    TabStop := FALSE;
    SpeedButton := TRUE;

    CASE B.Layout OF
    blGlyphLeft:   Layout := klLabelRight;
    blGlyphRight:  Layout := klLabelLeft;
    blGlyphTop:    Layout := klLabelBottom;
    blGlyphBottom: Layout := klLabelTop;
    END;
    if  (B.Glyph <> nil) and not B.Glyph.Empty then
    begin
        Bmp := TBitmap.Create;
        TRY
            Bmp.Assign( B.Glyph );
            if  B.NumGlyphs > 1 then
                Bmp.Width := max( 1, Bmp.Width div B.NumGlyphs );
            Picture.Assign( Bmp );
        FINALLY
            Bmp.Free;
        END;
    end;

    Left := NewBoundsRect.Left;
    Top := NewBoundsRect.Top;
    Width := NewBoundsRect.Right - NewBoundsRect.Left;
    Height := NewBoundsRect.Bottom - NewBoundsRect.Top;

    //TabOrder := NewTabOrder;
    ComponentIndex := NewComponentIndex;

    if TRUE {DeleteOld} then
    begin
      try

        {$IFDEF _D6orHigher}
        B.Free;
        {$ELSE}
        BI := MakeIComponent( B );
        BI.DestroyObject;
        {$ENDIF}

      finally
      end;

      Name := OldName;
      Name := NewName;
    end;

    END;

    //windows.Beep( 600, 600 );
  except
  end;
end;

procedure TReplaceKnopkaEditor.ExecuteVerb(Index: Integer);
begin
  if Component = nil then Exit;
  case Index of
  0: inherited ExecuteVerb(0);
  1: ReplaceKnopka;
  2: DoOpenProjectFolder;
  end;
end;

function TReplaceKnopkaEditor.GetVerb(Index: Integer): string;
begin
  case Index of
  0:   Result := OnClickCallVerb;
  1:   Result := CompEditReplaceKnopka;
  2:   Result := OpenProjectFolderVerb;
  end;
end;

function TReplaceKnopkaEditor.GetVerbCount: Integer;
begin
  Result := 3;
end;

procedure TReplaceKnopkaEditor.ReplaceKnopka;
var WTR: TfmWhatToReplace;
    F: TForm;
    Fr: TFrame;
    C, Comp2: TComponent;
    I: Integer;
    s: String;

    procedure DoReplace1( C: TComponent; Comp: TComponent );
    begin
        if  Comp is TPnKnopka then
        begin
            if  C is TButton then
                DoReplace( C as TButton, Comp as TPnKnopka )
            else
            if  C is TSpeedButton then
                DoReplace( C as TSpeedButton, Comp as TPnKnopka );
        end
        else
        begin
            if  C is TButton then
                DoReplace( C as TButton, Comp as TKnopka )
            else
            if  C is TSpeedButton then
                DoReplace( C as TSpeedButton, Comp as TKnopka );
        end;
    end;
begin
  if Component.Owner = nil then Exit;
  if not( Component.Owner is TForm ) and not( Component.Owner is TFrame ) then
  begin
    ShowMessage( Err0101 );
    Exit;
  end;
  F := nil;
  Fr := nil;
  if Component.Owner is TForm
    then F := Component.Owner as TForm
    else Fr := Component.Owner as TFrame;
  WTR := TfmWhatToReplace.Create( Application );
  TRY
      //       TButton, TBitBtn   
      //    ( TPnKnopka,   TKnopka):
      WTR.cbNames.Items.Add( AllItems_text );
      if Assigned( F ) then
          for I := 0 to F.ComponentCount-1 do
          begin
            C := F.Components[ I ];
            if  C is TButton then
                s := ' ''' + (C as TButton).Caption + ''''
            else
            if  C is TBitBtn then
                s := ' ''' + (C as TBitBtn).Caption + ''''
            else
            if  C is TSpeedButton then
                s := ' ''' + (C as TSpeedButton).Caption + ''''
            else
                s := '';
            if (C is TButton) or (C is TBitBtn) or (C is TSpeedButton) then
            begin
              if  (C is TKnopka) and (Component is TKnopka) then continue;
              WTR.cbNames.Items.AddObject( C.Name + ': ' + C.ClassName + s, C );
            end;
          end
            else
          for I := 0 to Fr.ComponentCount-1 do
          begin
            C := Fr.Components[ I ];
            if (C is TButton) or (C is TBitBtn) then
            begin
              if  (C is TKnopka) and (Component is TKnopka) then continue;
              WTR.cbNames.Items.AddObject( C.Name + ': ' + C.ClassName + s, C );
            end;
          end;
      if WTR.cbNames.Items.Count = 2 then
      begin
        WTR.cbNames.ItemIndex := 1;
      end
        else
      begin
        WTR.cbNames.ItemIndex := 0;
      end;
      WTR.btnOK.Enabled := TRUE;
      if WTR.ShowModal = mrOK then
      begin
        if  WTR.cbNames.ItemIndex = 0 then
        begin
            for I := 1 to WTR.cbNames.Items.Count-1 do
            begin
                Comp2 := Component;
                if  I < WTR.cbNames.Items.Count-1 then
                begin
                    if  Component is TPnKnopka then
                        Comp2 := TPnKnopka.Create( Component.Owner )
                    else
                        Comp2 := TKnopka.Create( Component.Owner );
                end;
                C := TComponent( WTR.cbNames.Items.Objects[ I ] );
                DoReplace1( C, Comp2 );
            end;
        end
          else
        begin
            C := TComponent( WTR.cbNames.Items.Objects[ WTR.cbNames.ItemIndex ] );
            DoReplace1( C, Component );
        end;
      end;
  FINALLY
      WTR.Free;
  END;
end;

end.
