//**  ,     VCL Delphi.
//**  Vladimir Kladov, 2000-2010
//**
//**      .
//**     /     
//**  2          
//**   GNU,  Free Software Foundation.
//**
//**       ,    
//** ,       ,   
//**         
//**   .      
//**     GNU.
//**
//**          
//**   GNU.     ,     Free
//** Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
//** MA 02111-1307, USA.
//******************************************************************************

unit Scanner;

interface

{$I VCL_VERSIONS.INC}

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  extctrls;

type
  TPort = ( pCom1, pCom2, pCom3, pCom4, pCom5, pCom6, pCom7, pCom8 );
  TOnRead = procedure( Sender: TComponent; const S: String ) of Object;

  TScanner = class(TComponent)
  private
    FActive: Boolean;
    FOnRead: TOnRead;
    FPort: TPort;
    FAutoPort: Boolean;

    FThread: TThread;
    FHandle: THandle;
    FAsyncCall: Boolean;
    FWaitCR: Boolean;
    FSpeed: Integer;
    FClearHiBit: Boolean;

    procedure SetActive(const Value: Boolean);
    procedure SetPort(const Value: TPort);
    procedure SetAsyncCall(const Value: Boolean);
    procedure SetSpeed(const Value: Integer);
    procedure SetClearHiBit(const Value: Boolean);
    { Private declarations }
  protected
    { Protected declarations }
  public
    { Public declarations }
    ErrText: String;
    Sema: THandle;
  published
    { Published declarations }
    constructor Create( AOwner: TComponent ); override;
    destructor Destroy; override;
    property AutoPort: Boolean read FAutoPort write FAutoPort;
    property Port: TPort read FPort write SetPort;
    property Speed: Integer read FSpeed write SetSpeed;
    property Active: Boolean read FActive write SetActive;
    property OnRead: TOnRead read FOnRead write FOnRead;
    property AsyncCall: Boolean read FAsyncCall write SetAsyncCall;
    property WaitCR: Boolean read FWaitCR write FWaitCR;
    property ClearHiBit: Boolean read FClearHiBit write SetClearHiBit;
    procedure Write( const s: String );
  end;

{$R *.DCR}

procedure Register;

implementation

procedure Register;
begin
  RegisterComponents('Kladov', [TScanner]);
end;

type
  TScanThread = class( TThread )
  private
    FScanner: TScanner;
    FStopping: Boolean;
    FStopped: Boolean;
  protected
    InString: String;
    OutString: String;
    //ErrText: String;
    procedure PostChars;
  public
    property Scanner: TScanner read FScanner;
    procedure Execute; override;
    property Stopped: Boolean read FStopped;
  end;

{ TScanner }

constructor TScanner.Create(AOwner: TComponent);
begin
  inherited;
  FAutoPort := TRUE;
  FWaitCR := TRUE;
  FSpeed := 9600;
  ClearHiBit := FALSE;
  Sema := CreateSemaphore( nil, 1, 1, nil );
end;

destructor TScanner.Destroy;
begin
  Active := FALSE;
  if FHandle <> INVALID_HANDLE_VALUE then
  begin
    CloseHandle( FHandle );
    FHandle := 0;
  end;
  CloseHandle( Sema );
  inherited;
end;

procedure TScanner.SetActive(const Value: Boolean);
var N: Integer;
begin
  if FActive = Value then Exit;
  FActive := Value;
  if csDesigning in ComponentState then Exit;
  if Value then
  begin
    FThread := TScanThread.Create( TRUE );
    (FThread as TScanThread).FScanner := Self;
    FThread.Resume;
  end
    else
  begin
    (FThread as TScanThread).FStopping := TRUE;
    N := 4;
    while not (FThread as TScanThread).Stopped and (N > 0) do
    begin
      Sleep( 100 );
      Dec( N );
    end;
    if not (FThread as TScanThread).Stopped then
      TerminateThread( FThread.Handle, 0 );
    FThread.Free;
    FThread := nil;
    FActive := FALSE;
  end;
end;

procedure TScanner.SetAsyncCall(const Value: Boolean);
begin
  FAsyncCall := Value;
end;

procedure TScanner.SetClearHiBit(const Value: Boolean);
begin
  FClearHiBit := Value;
end;

procedure TScanner.SetPort(const Value: TPort);
begin
  ErrText := '';
  FPort := Value;
end;

procedure TScanner.SetSpeed(const Value: Integer);
begin
  FSpeed := Value;
end;

procedure TScanner.Write(const s: String);
begin
  WaitForSingleObject( Sema, INFINITE );
  TRY
    TScanThread( FThread ).OutString := TScanThread( FThread ).OutString + s;
  FINALLY
    ReleaseSemaphore( Sema, 1, nil );
  END;
end;

{ TScanThread }

function SysErrorMessage(ErrorCode: Integer): string;
var
  Len: Integer;
  Buffer: array[0..255] of Char;
begin
  Len := FormatMessage(FORMAT_MESSAGE_FROM_SYSTEM or
    FORMAT_MESSAGE_ARGUMENT_ARRAY, nil, ErrorCode, 0, Buffer,
    SizeOf(Buffer), nil);
  while (Len > 0) and (Buffer[Len - 1] in [#0..#32 {, '.'}]) do Dec(Len);
  SetString(Result, Buffer, Len);
end;

procedure TScanThread.Execute;
var P: TPort;
    Buf: PChar;
    Ev, L: DWORD;
    Ov: TOverlapped;
    Stat: TComStat;
    Er: DWORD;
    CC: TCommConfig;
    i: Integer;
begin
  GetMem( Buf, 65536 );
  TRY

    FillChar( Ov, Sizeof( Ov ), 0 );

    P := FScanner.FPort;
    FScanner.FHandle := INVALID_HANDLE_VALUE;

    repeat
      FScanner.FHandle := CreateFile( PChar( 'COM' + IntToStr( Ord( P ) + 1 ) ),
                       GENERIC_READ or GENERIC_WRITE, 0, nil, OPEN_EXISTING,
                       0 {FILE_FLAG_OVERLAPPED}, 0 );
      if FScanner.FHandle <> INVALID_HANDLE_VALUE then break;
      if P = High( P ) then
        P := Low( P )
      else
        Inc( P );
    until (FScanner.FHandle <> INVALID_HANDLE_VALUE) or (P=FScanner.FPort);

    if FScanner.FHandle <> INVALID_HANDLE_VALUE then
    begin
      CC.dwSize := Sizeof( CC );
      if GetCommConfig( FScanner.FHandle, CC, CC.dwSize ) then
      begin

        CC.dcb.BaudRate := FScanner.Speed; //CBR_9600;
        CC.dcb.ByteSize := 8;
        CC.dcb.Flags := CC.dcb.Flags and not 2; //  Parity
        CC.dcb.Parity := NOPARITY;
        CC.dcb.StopBits := ONESTOPBIT;
        if SetCommConfig( FScanner.FHandle, CC, Sizeof( CC ) ) then
        BEGIN

          if SetCommMask( FScanner.FHandle, EV_CTS or EV_DSR or EV_RXCHAR ) then
          begin
            Ov.hEvent := CreateEvent( nil, FALSE, FALSE, nil );
            while not FStopping do
            begin
              //Sleep( 10 );
              if WaitCommEvent( FScanner.FHandle, Ev, @ Ov ) then
              BEGIN

                  if LongBool( Ev and EV_CTS ) then
                  begin
                    ClearCommError( FScanner.FHandle, Er, @ Stat );
                    WaitForSingleObject( FScanner.Sema, INFINITE );
                    TRY
                      if OutString <> '' then
                      begin
                        if Stat.cbOutQue > 0 then
                        begin
                          TRY
                            L := Length( OutString );
                            if L > Stat.cbOutQue then L := Stat.cbOutQue;
                            move( OutString[ 1 ], Buf^, L );
                            WriteFile( FScanner.FHandle, Buf^, L, Er, @ Ov );
                            if Er <> 0 then
                            begin
                              Delete( OutString, 1, Er );
                            end;
                          EXCEPT on E: Exception do
                               begin
                                 ShowMessage( '    -  : ' +
                                   E.Message );
                               end;
                          END;
                        end;
                      end;
                    FINALLY
                      ReleaseSemaphore( FScanner.Sema, 1, nil );
                    END;
                  end
                    else
                  if LongBool( Ev and (EV_DSR or EV_RXCHAR) ) then
                  begin
                    ClearCommError( FScanner.FHandle, Er, @ Stat );
                    if Stat.cbInQue > 0 then
                    begin
                      TRY
                        ReadFile( FScanner.FHandle, Buf^, Stat.cbInQue, Er, @ Ov );
                        if Er <> 0 then
                        begin
                          Buf[ Er ] := #0;
                          if FScanner.ClearHiBit then
                          begin
                            for i := 0 to Er-1 do
                              Buf[ i ] := Char( Integer( Buf[ i ] ) and $7F );
                          end;
                          InString := InString + Buf;
                          if not FScanner.WaitCR or (pos( #13, InString ) > 0) then
                          if FScanner.FAsyncCall then
                            PostChars
                          else
                            Synchronize( PostChars );
                        end;
                      EXCEPT on E: Exception do
                             begin
                               ShowMessage( '    -  : ' +
                                 E.Message );
                             end;
                      END;
                    end;
                  end;
              END;
            end;
            CloseHandle( Ov.hEvent );
          end;

        END
          else
        begin
          FScanner.ErrText := SysErrorMessage( GetLastError );
          Windows.Beep( 100, 100 );
        end;

      end;

      CloseHandle( FScanner.FHandle );
      FScanner.FHandle := INVALID_HANDLE_VALUE;
    end
      else
    begin
      Windows.Beep( 1000, 100 );
    end;
    FStopped := TRUE;
  FINALLY
    FreeMem( Buf );
  END;
end;

procedure TScanThread.PostChars;
var I: Integer;
begin
  TRY
    while InString <> '' do
    begin
      if FScanner.WaitCR then I := pos( #13, InString )
      else I := Length( InString ) + 1;
      if I > 0 then
      begin
        if Assigned( FScanner.OnRead ) then
          FScanner.OnRead( FScanner, Copy( InString, 1, I - 1 ) );
        InString := Copy( InString, I + 1, MaxInt );
      end
        else break;
    end;
  EXCEPT on E: Exception do
         begin
           ShowMessage( '   : ' + E.Message );
         end;
  END;
end;

end.
