//**  ,     VCL Delphi.
//**  Vladimir Kladov, 2000-2010
//**
//**      .
//**     /     
//**  2          
//**   GNU,  Free Software Foundation.
//**
//**       ,    
//** ,       ,   
//**         
//**   .      
//**     GNU.
//**
//**          
//**   GNU.     ,     Free
//** Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
//** MA 02111-1307, USA.
//******************************************************************************

unit StdImgPropEditor;

interface

{$I VCL_VERSIONS.INC}

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  comctrls, imglist, commctrl, QSort, ColorListView, menus, Extctrls, registry,
  {$IFDEF _D6} designintf {$IFDEF IN_PACKAGE}, DesignEditors {$ENDIF}
  {$ELSE}
     {$IFDEF _D7orHigher} designintf {$IFDEF IN_PACKAGE}, DesignEditors {$ENDIF}
     {$ELSE} dsgnintf {$ENDIF}
  {$ENDIF},
  UniPic, Knopka, UniEdit;

type
  TStdImgEditor = class( TPropertyEditor )
  private
  protected
  public
    procedure Edit; override;
    function GetAttributes: TPropertyAttributes; override;
    function GetValue: string; override;
    procedure SetValue(const Value: string); override;
  end;

  TUniPic_Access = class( TUniPic )
  end;

procedure Register;

implementation

uses
  WhatToImport, StdImgSelectIcon;

procedure Register;
begin
  RegisterPropertyEditor( TypeInfo( TStdImage ), nil {TUniPic}, '' {'StdImage'},
                          TStdImgEditor );
end;

{ TStdImgEditor }

procedure TStdImgEditor.Edit;
var Comp: TPersistent;
    i: Integer;
begin
  for i := 0 to PropCount-1 do
  begin
      Comp := GetComponent( i );
      if  not (Comp is TUniPic)
      and not (Comp is TUniEdit)
      and not (Comp is TKnopka)
      and not (Comp is TPnKnopka)
      and not (Comp is TGrKnopka) then
      begin
          ShowMessage( 'Only TUniPic, TUniEdit, TKnopka, TPnKnopka and ' +
                       'TGrKnopka are supported!' );
          Exit;
      end;
  end;

  if  fmSelectIcon = nil then
      fmSelectIcon := TfmSelectIcon.Create( Application );
  TRY
      if  fmSelectIcon.ShowModal = mrOK then
      begin
          for i := 0 to PropCount-1 do
          begin
              Comp := GetComponent( i );
              if  Comp is TUniPic then
              begin
                  (Comp as TUniPic).StdImage := fmSelectIcon.Value;
              end
              else if  Comp is TUniEdit then
              begin
                  (Comp as TUniEdit).StdImage := fmSelectIcon.Value;
                  if  (Comp as TUniEdit).StdImgShowSize = 0 then
                      (Comp as TUniEdit).StdImgShowSize := 16;
              end
              else if  Comp is TKnopka then
              begin
                  (Comp as TKnopka).StdImage := fmSelectIcon.Value;
                  if  (Comp as TKnopka).StdImgShowSize = 0 then
                      (Comp as TKnopka).StdImgShowSize := 16;
              end
              else if  Comp is TPnKnopka then
              begin
                  (Comp as TPnKnopka).StdImage := fmSelectIcon.Value;
                  if  (Comp as TPnKnopka).StdImgShowSize = 0 then
                      (Comp as TPnKnopka).StdImgShowSize := 16;
              end
              else if  Comp is TGrKnopka then
              begin
                  (Comp as TGrKnopka).StdImage := fmSelectIcon.Value;
                  if  (Comp as TGrKnopka).StdImgShowSize = 0 then
                      (Comp as TGrKnopka).StdImgShowSize := 16;
              end;
          end;
      end;
  FINALLY
      FreeAndNil( fmSelectIcon );
  END;
end;

function TStdImgEditor.GetAttributes: TPropertyAttributes;
begin
  Result := [ paDialog ];
end;

function TStdImgEditor.GetValue: string;
var Comp: TPersistent;
    v: TStdImage;
begin
    Result := '';
    if  not AllEqual then Exit;
    Comp := GetComponent( 0 );
    if  Comp is TUniPic then
        v := (Comp as TUniPic).StdImage
    else if  Comp is TUniEdit then
        v := (Comp as TUniEdit).StdImage
    else if  Comp is TKnopka then
        v := (Comp as TKnopka).StdImage
    else if  Comp is TPnKnopka then
        v := (Comp as TPnKnopka).StdImage
    else if  Comp is TGrKnopka then
        v := (Comp as TGrKnopka).StdImage
    else Exit;
    Result := StdImageNames[ v ];
end;

procedure TStdImgEditor.SetValue(const Value: string);
var i: TStdImage;
begin
    for i := Low( i ) to High( i ) do
        if  CompareText( StdImageNames[i], Value ) = 0 then
        begin
            SetOrdValue( Integer( i ) );
            Modified;
            Exit;
        end;
end;

end.
