//**  ,     VCL Delphi.
//**  Vladimir Kladov, 2000-2010
//**
//**      .
//**     /     
//**  2          
//**   GNU,  Free Software Foundation.
//**
//**       ,    
//** ,       ,   
//**         
//**   .      
//**     GNU.
//**
//**          
//**   GNU.     ,     Free
//** Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
//** MA 02111-1307, USA.
//******************************************************************************

//------------------------------------------------------------------------------
// UtilsUnit -   
//
// :  .., 2003-2004
//
unit UtilsUnit;

{$I VCL_VERSIONS.inc}

interface

uses Windows, Messages, SysUtils, Dialogs, DB, {UniEdit, AdoSqlUnit,} Forms,
     Classes, Controls, registry;


const _IDENTITY = //'SCOPE_IDENTITY()' -  MS SQL 6.5   !
                  '@@identity';
      _NULL = //'NULL' -  6.5  !
              '';
  WM_IMITATE = WM_USER + 333;

var Upper: array[ Char ] of Char;


function GetDesktopRect : TRect;
function Replace( const S: String; CFrom, CTo: Char ): String;
function ReplaceExt( const S: String; const newExt: String ): String;
function StrToFloat( const S: String ): Extended;
function FloatToSQL( E: Extended ): String;
function StrToDate( const S: String ): TDateTime;
function StrToSQL( const S: String ): String;
function StrToSQLNullEmpty( const S: String ): String;

var date_fmt: Integer = 0;
function DateToSQL( D: TDateTime ): String;
function DateTimeToSQL( D: TDateTime ): String;
function NullIfEmpty( const S: String ): String;
function NullIfEmptyStr( const S: String ): String;
function NullIf0( const S: String ): String;
function ZeroIfEmpty( const S: String ): String;
function Year( DT: TDateTime ): Integer;
function Month( DT: TDateTime ): Integer;
function Day( DT: TDateTime ): Integer;
function Hour( DT: TDateTime ): Integer;
function Minute( DT: TDateTime ): Integer;
function Second( DT: TDateTime ): Integer;
function Years( D1, D2: TDateTime ): Integer;
function MonthStart( d: TDateTime ): TDateTime;
function ConcatIfNotEmpty( const S1, S2: String ): String;
function RemoveLast( WhatRemove: Char; const S: String ): String;
function RemoveLastDot( const S: String ): String;
function RemoveCRs( const S: String ): String;
function IfNot0( const S: String ): String;
function Triads( S: String ): String;
function RemoveSpaces( const S: String ): String;
function IfNotEmpty( const S1: String; const S2: String = '' ): String;
function IfEmpty0( const S1: String ): String;
function IfThenElse( Cond: Boolean; const S1: String; const S2: String = '' ): String; overload;
function IfThenElse( Cond: Boolean; V1: Integer; V2: Integer = 0 ): Integer; overload;
function RemoveLastComma( const S: String ): String;
function MaxVals( const Vals: array of Integer ): Integer;
function IntIn( I: Integer; const Vals: array of Integer ): Boolean;
procedure IntSwap( var i1, i2: Integer );
function IfDateToSQL( const D: String ): String;
function SqlUpdIns( Upd: Boolean; const Table: String;
         const FieldsAndValues: array of String; Distinct: Boolean = FALSE ): String;
function SqlUpd( const Table: String; const FieldsAndValues: array of String ): String;
function SqlIns( const Table: String; const FieldsAndValues: array of String ): String;
function SqlInsDistinct( const Table: String; const FieldsAndValues: array of String ): String;
function Drop( const TableName: String ): String;
function EndDrop( const TableName: String ): String;
function CmpInt( x, y: Integer ): Integer;
function CmpStr( s1, s2: String ): Integer;
function CmpStrNum( s1, s2: String ): Integer;
procedure StrDel( var s: String; from, count: Integer );
function AddValue( var Dest: Double; Src: Double; var DestEdiz: String;
          const SrcEdiz: String; var ErrString: String ): Boolean;
function OnlyLetters( const s: String ): String;
function StrIn( const s: String; const A: array of String ): Boolean;
function PathTrail( const s: String ): String;
function PathLead( const s: String ): String;
function Max( const Vals: array of Integer ): Integer;
procedure SaveForm( F: TForm );
procedure RestoreForm( F: TForm );
function LastChar( const s: String ): Char;

function min( x, y: Integer ): Integer;
function Second2Time( sec: DWORD ): String;
function Extended2Str( E: Extended ): String;
function Round_Fmt( E: Extended ): Int64;

function RussianSummaPropis( i: Int64 ): String;

function Parse( var s: String; const d: String ): String;
function ParseChars( var s: String; const d: String ): String;

function CheckFileCanRewrite( const Filename: String ): Boolean;
function IsFloatNum( const s: String ): Boolean;
function ScanNumber( const s: String ): Integer;

function Ask( const Question: String ): Boolean;
procedure Msg( const Text: String );
procedure Warning( const Text: String );
procedure ShowError( const Text: String );

function DirectoryExists( const path: String ): Boolean;
procedure LogFileOutput( const filepath, str: AnsiString );

type
  TWindowsVersion = ( wv31, wv95, wv98, wvNT, wvY2K, wvXP, wvLongHorn );
  {* Windows versions constants. }
  TWindowsVersions = Set of TWindowsVersion;
  {* Set of Windows version (e.g. to define a range of versions supported by the
     application). }
function WinVer : TWindowsVersion;

procedure Beep;


var RegistrySaveForm: String = 'Software\';

implementation

procedure InitUpper;
var c: Char;
begin
  for c := #0 to #255 do
    Upper[ c ] := ( AnsiUpperCase( c ) + ' ' )[ 1 ];
end;

function Replace( const S: String; CFrom, CTo: Char ): String;
var I: Integer;
begin
  Result := S;
  for I := 1 to Length( Result ) do
    if Result[ I ] = CFrom then Result[ I ] := CTo;
end;

function ReplaceExt( const S: String; const newExt: String ): String;
var i: Integer;
begin
    Result := S;
    for i := Length(S) downto 1 do
        if  S[i] = '.' then
        begin
            Result := Copy( Result, 1, i-1 ) + newExt;
            Exit;
        end;
end;

function StrToFloat( const S: String ): Extended;
begin
  if Trim( S ) = '' then
    Result := 0
  else
  TRY
    Result := SysUtils.StrToFloat(
      StringReplace(
      StringReplace( S, '.', DecimalSeparator, [] ),
                        ',', DecimalSeparator, [] ) );
  EXCEPT
    Result := 0;
  END;
end;

function FloatToSQL( E: Extended ): String;
begin
  Result := Extended2Str( E ); //Format( '%f', [ E ] );
  Result := Replace( Result, ',', '.' );
end;

function StrToDate( const S: String ): TDateTime;

    function ScanNumber( var i: Integer; endChar: Char  ): Integer;
    begin
        Result := 0;
        while S[i] in [ '0'..'9' ] do
        begin
            Result := Result * 10 + Ord( S[i] ) - Ord( '0' );
            inc( i );
        end;
        if  S[i] = endChar then
            inc( i );
    end;

var D, M, Y: Word;
    i: Integer;
begin
  if  S = '' then
      Result := 0
  else
  if  (Length(S) > 3) and ((S[2] = '/') or (S[3] = '/')) then
  begin
      TRY
          i := 1;
          D := ScanNumber( i, '/' );
          M := ScanNumber( i, '/' );
          Y := ScanNumber( i, ' ' );
          if  M > 12 then
          begin
              i := M;
              M := D;
              D := i;
          end;
          Result := EncodeDate( Y, M, D );
      EXCEPT
          Result := SysUtils.StrToDate( S );
      END;
  end
  else
  if (Length( S ) = 10) and (S[ 3 ] = '.') and (S[ 6 ] = '.') then
  begin
    TRY
      D := StrToInt( Copy( S, 1, 2 ) );
      M := StrToInt( Copy( S, 4, 2 ) );
      Y := StrToInt( Copy( S, 7, 4 ) );
      Result := EncodeDate( Y, M, D );
    EXCEPT
      Result := SysUtils.StrToDate( S );
    END;
  end
  else
    TRY
        Result := SysUtils.StrToDate( S );
    EXCEPT
        Result := 0;
    END;
end;

function StrToSQL( const S: String ): String;
var I: Integer;
begin
  Result := '''' + S + '''';
  for I := Length( Result )-1 downto 2 do
  begin
    if Result[ I ] = '''' then
      Result := Copy( Result, 1, I ) + Copy( Result, I, MaxInt );
  end;
end;

function StrToSQLNullEmpty( const S: String ): String;
begin
  if S = '' then
    Result := 'null'
  else
    Result := StrToSQL( S );
end;

function NullIfEmpty( const S: String ): String;
begin
  if Trim( S ) = '' then Result := 'null'
                    else Result := S;
end;

function NullIfEmptyStr( const S: String ): String;
begin
  if Trim( S ) = '''''' then Result := 'null'
                        else Result := S;
end;

function NullIf0( const S: String ): String;
var D: Double;
begin
  TRY
    D := StrToFloat( S );
  EXCEPT
    D := 0;
  END;
  if D = 0 then Result := 'null'
           else Result := Replace( S, ',', '.' );
end;

function ZeroIfEmpty( const S: String ): String;
begin
    Result := S;
    if  Result = '' then
        Result := '0';
end;

function DateToSQL( D: TDateTime ): String;
begin
  CASE date_fmt OF
  0:   Result := '''' + //FormatDateTime( 'yyyymmdd', D ) + '''';
           Format( '%.04d%.02d%.02d', [ Year( D ), Month( D ), Day( D ) ] ) + '''';
  1:   Result := '''' + //FormatDateTime( 'yyyymmdd', D ) + '''';
           Format( '%.02d/%.02d/%.04d', [ Month( D ), Day( D ), Year( D ) ] ) + '''';
  2:   Result := '#' + //FormatDateTime( 'yyyymmdd', D ) + '''';
           Format( '%.04d.%.02d.%.02d', [ Year( D ), Month( D ), Day( D ) ] ) + '#';
  else Result := '#' + //FormatDateTime( 'yyyymmdd', D ) + '''';
           Format( '%.02d-%.02d-%.04d', [ Month( D ), Day( D ), Year( D ) ] ) + '#';
  END;
end;

function DateTimeToSQL( D: TDateTime ): String;
begin
  CASE date_fmt OF
  0:   Result := '''' + //FormatDateTime( 'yyyymmdd', D ) + '''';
           Format( '%.04d%.02d%.02d ', [ Year( D ), Month( D ), Day( D ) ] ) +
           ' ' + Format( '%.02d:%.02d:%.02d', [ Hour( D ), Minute( D ), Second( D ) ] ) +
           '''';
  1:   Result := '''' + //FormatDateTime( 'yyyymmdd', D ) + '''';
           Format( '%.02d/%.02d/%.04d ', [ Month( D ), Day( D ), Year( D ) ] ) +
           ' ' + Format( '%.02d:%.02d:%.02d', [ Hour( D ), Minute( D ), Second( D ) ] ) +
           '''';
  2:   Result := '#' + //FormatDateTime( 'yyyymmdd', D ) + '''';
           Format( '%.04d.%.02d.%.02d ', [ Year( D ), Month( D ), Day( D ) ] ) +
           ' ' + Format( '%.02d:%.02d:%.02d', [ Hour( D ), Minute( D ), Second( D ) ] ) +
           '#';
  else Result := '#' + //FormatDateTime( 'yyyymmdd', D ) + '''';
           Format( '%.02d-%.02d-%.04d ', [ Month( D ), Day( D ), Year( D ) ] ) +
           ' ' + Format( '%.02d:%.02d:%.02d', [ Hour( D ), Minute( D ), Second( D ) ] ) +
           '#';
  END;
end;


function Year( DT: TDateTime ): Integer;
var Y, M, D: Word;
begin
  DecodeDate( DT, Y, M, D );
  Result := Y;
end;

function Month( DT: TDateTime ): Integer;
var Y, M, D: Word;
begin
  DecodeDate( DT, Y, M, D );
  Result := M;
end;

function Day( DT: TDateTime ): Integer;
var Y, M, D: Word;
begin
  DecodeDate( DT, Y, M, D );
  Result := D;
end;

function Hour( DT: TDateTime ): Integer;
var H,M,S, Ms: Word;
begin
  DecodeTime( DT, H, M, S, Ms );
  Result := H;
end;

function Minute( DT: TDateTime ): Integer;
var H,M,S, Ms: Word;
begin
  DecodeTime( DT, H, M, S, Ms );
  Result := M;
end;

function Second( DT: TDateTime ): Integer;
var H,M,S, Ms: Word;
begin
  DecodeTime( DT, H, M, S, Ms );
  Result := S;
end;

function Years( D1, D2: TDateTime ): Integer;
var Yb, Mb, Db, Ye, Me, De: Word;
    Yn, Mn, Dn, Yk, Mk, Dk, Y, M, D: Integer;
    D0: TDateTime;
begin
  if D1 > D2 then
  begin
    D0 := D1;
    D1 := D2;
    D2 := D0;
  end;
  DecodeDate( D1, Yb, Mb, Db );
  Yn := Yb; Mn := Mb; Dn := Db;
  if Yn <= 1900 then
  begin
    Result := 0;
    Exit;
  end;
  DecodeDate( D2, Ye, Me, De );
  Yk := Ye; Mk := Me; Dk := De;
  D := Dk - Dn;
  if D < 0 then
  begin
    D := D + 30;
    Mk := Mk - 1;
  end;
  M := Mk - Mn;
  if M < 0 then
  begin
    M := M + 12;
    Yk := Yk - 1;
  end;
  Y := Yk - Yn;
  D := D + 1;
  if D >= 30 then
  begin
    //D := 0;
    M := M + 1;
  end;
  if M = 12 then
  begin
    //M := 0;
    Y := Y + 1;
  end;
  Result := Y;
end;

function MonthStart( d: TDateTime ): TDateTime;
var dd, mm, yy: Word;
begin
  DecodeDate( d, yy, mm, dd );
  Result := EncodeDate( yy, mm, 1 );
end;

function ConcatIfNotEmpty( const S1, S2: String ): String;
begin
  if Trim( S1 ) <> '' then Result := S1 + S2
  else Result := '';
end;

function RemoveLast( WhatRemove: Char; const S: String ): String;
begin
  Result := S;
  if S = '' then Exit;
  if S[ Length( S ) ] = WhatRemove then
    Delete( Result, Length( S ), 1 );
end;

function RemoveLastDot( const S: String ): String;
begin
  Result := RemoveLast( '.', S );
end;

function RemoveCRs( const S: String ): String;
var I: Integer;
begin
  Result := S;
  for I := Length( S ) downto 1 do
    if S[ I ] < ' ' then Delete( Result, I, 1 );
end;

function IfNot0( const S: String ): String;
begin
  if S = '0' then Result := ''
             else Result := S;
end;

function Triads( S: String ): String;
var F: String;
    I: Integer;
    SLast: String;
begin
  F := RemoveSpaces( S );
  I := pos( '.', F );
  if I <= 0 then
    I := pos( ',', F );

  SLast := '';
  if I > 0 then
  begin
    SLast := Copy( F, I, MaxInt );
    F := Copy( F, 1, I - 1 );
  end;

  Result := '';
  while F <> '' do
  begin
    Result := Copy( F, Length( F ) - 2, 3 ) + ' ' + Result;
    F := Copy( F, 1, Length( F ) - 3 );
  end;
  Result := Trim( Result ) + SLast;
end;

function RemoveSpaces( const S: String ): String;
var I: Integer;
begin
  Result := S;
  for I := Length( S ) downto 1 do
    if S[ I ] <= ' ' then
      Delete( Result, I , 1 );
end;

function IfNotEmpty( const S1: String; const S2: String = '' ): String;
begin
  if Trim( S1 ) <> '' then
  begin
    Result := S2;
    if Result = '' then Result := S1;
  end
  else Result := '';
end;

function IfEmpty0( const S1: String ): String;
begin
  if Trim( S1 ) = '' then Result := '0'
  else Result := S1;
end;

function IfThenElse( Cond: Boolean; const S1: String; const S2: String = '' ): String; overload;
begin
  if Cond then Result := S1 else Result := S2;
end;

function IfThenElse( Cond: Boolean; V1: Integer; V2: Integer = 0 ): Integer; overload;
begin
  if Cond then Result := V1 else Result := V2;
end;

function RemoveLastComma( const S: String ): String;
begin
  Result := Trim( S );
  if Result = '' then Exit;
  if Result[ Length( Result ) ] = ',' then Delete( Result, Length( Result ), 1 );
end;

function MaxVals( const Vals: array of Integer ): Integer;
var I: Integer;
begin
  Result := Vals[ 0 ];
  for I := 1 to High( Vals ) do
    if Result < Vals[ I ] then Result := Vals[ I ];
end;

function IntIn( I: Integer; const Vals: array of Integer ): Boolean;
var J: Integer;
begin
  Result := FALSE;
  for J := 0 to High( Vals ) do
    if I = Vals[ J ] then
    begin
      Result := TRUE;
      Exit;
    end;
end;

procedure IntSwap( var i1, i2: Integer );
var tm: Integer;
begin
    tm := i1;
    i1 := i2;
    i2 := tm;
end;

  function IfDateToSQL( const D: String ): String;
  begin
    Result := D;
    if Trim( D ) <> '' then
      Result := DateToSQL( StrToDate( D ) );
  end;

function SqlUpd( const Table: String; const FieldsAndValues: array of String ): String;
begin
  Result := SqlUpdIns( TRUE, Table, FieldsAndValues );
end;

function SqlIns( const Table: String; const FieldsAndValues: array of String ): String;
begin
  Result := SqlUpdIns( FALSE, Table, FieldsAndValues );
end;

function SqlInsDistinct( const Table: String; const FieldsAndValues: array of String ): String;
begin
  Result := SqlUpdIns( FALSE, Table, FieldsAndValues, TRUE );
end;

function SqlUpdIns( Upd: Boolean; const Table: String;
         const FieldsAndValues: array of String; Distinct: Boolean=FALSE ): String;
var I: Integer;
begin
  if Upd then
  begin
    Result := #13'update ' + Table + ' set ';
    for I := 0 to (High( FieldsAndValues )+1) div 2 - 1 do
    begin
      if  FieldsAndValues[ I * 2 ] = '' then
          continue;
      if I > 0 then
        Result := Result + ', ';
      Result := Result + FieldsAndValues[ I * 2 ] + ' = ' +
                         FieldsAndValues[ I * 2 + 1 ];
    end;
  end
    else
  begin
    Result := #13'insert into ' + Table + ' (';
    for I := 0 to (High( FieldsAndValues )+1) div 2 - 1 do
    begin
      if  FieldsAndValues[ I * 2 ] = '' then
          continue;
      if I > 0 then
        Result := Result + ', ';
      Result := Result + FieldsAndValues[ I * 2 ];
    end;
    Result := Result + ')'#13' select ';
    if Distinct then
      Result := Result + 'distinct ';
    for I := 0 to (High( FieldsAndValues )+1) div 2 - 1 do
    begin
      if  FieldsAndValues[ I * 2 ] = '' then
          continue;
      if I > 0 then
        Result := Result + ', ';
      Result := Result + FieldsAndValues[ I * 2 + 1 ];
    end;
    Result := Result + ' ';
  end;
end;

function Drop( const TableName: String ): String;
begin
  Result := ''
  {#13'if exists(select name from sysobjects where name=''' +
         TableName + ''') begin ' +
  #13'   drop table ' + TableName + '' +
  #13'end'#13;}
end;

function EndDrop( const TableName: String ): String;
begin
  Result := #13'drop table ' + TableName + #13;
  {#13'if exists(select name from sysobjects where name=''' +
         TableName + ''') begin ' +
  #13'   drop table ' + TableName + '' +
  #13'end'#13;}
end;

function CmpInt( x, y: Integer ): Integer;
begin
  if x < y then Result := -1 else if x > y then Result := 1 else Result := 0;
end;

function CmpStr( s1, s2: String ): Integer;
begin
  Result := AnsiCompareStr( s1, s2 );
end;

function CmpStrNum( s1, s2: String ): Integer;
var i1, i2, n1, n2: Integer;
begin
  Result := 0;
  i1 := 1;
  i2 := 1;
  while (i1 <= Length( s1 )) and (i2 <= Length( s2 )) do
  begin
    if (s1[ i1 ] in [ '0'..'9' ]) and (s2[ i2 ] in [ '0'..'9' ]) then
    begin
      n1 := 0;
      while (i1 <= Length( s1 )) and (s1[ i1 ] in [ '0'..'9' ]) do
      begin
        n1 := n1 * 10 + Ord( s1[ i1 ] ) - Ord( '0' );
        inc( i1 );
      end;
      n2 := 0;
      while (i2 <= Length( s2 )) and (s2[ i2 ] in [ '0'..'9' ]) do
      begin
        n2 := n2 * 10 + Ord( s2[ i2 ] ) - Ord( '0' );
        inc( i2 );
      end;
      Result := n1 - n2;
    end
      else
    begin
      Result := Ord( Upper[ s1[ i1 ] ] ) - Ord( Upper[ s2[ i2 ] ] );
      if Result = 0 then
        Result := Ord( s1[ i1 ] ) - Ord( s2[ i2 ] );
      inc( i1 );
      inc( i2 );
    end;
    if Result <> 0 then Exit;
  end;
  if (i1 > Length( s1 )) and (i2 <= Length( s2 )) then Result := 1
  else
  if (i1 <= Length( s1 )) and (i2 > Length( s2 )) then Result := -1;
end;

procedure StrDel( var s: String; from, count: Integer );
begin
    Delete( s, from, count );
end;

function AddValue( var Dest: Double; Src: Double; var DestEdiz: String;
          const SrcEdiz: String; var ErrString: String ): Boolean;
begin
  Result := TRUE;
  if Src = 0 then
  begin
    if Dest = 0 then
      DestEdiz := SrcEdiz;
    Exit;
  end;
  if Dest = 0 then
  begin
    Dest := Src;
    DestEdiz := SrcEdiz;
    Exit;
  end;
  if DestEdiz = SrcEdiz then
  begin
    Dest := Dest + Src;
    Exit;
  end;
  if StrIn( OnlyLetters( DestEdiz ), [ '', 'kg' ] ) and
     StrIn( OnlyLetters( SrcEdiz ), [ '', 'g' ] ) then
  begin
    Dest := Dest + Src / 1000;
    Exit;
  end
    else
  if StrIn( OnlyLetters( DestEdiz ), [ '' ] ) and
     (SrcEdiz = '%') then
  begin
    Dest := Dest + Src / 100;
    Exit;
  end;
  if StrIn( OnlyLetters( DestEdiz ), [ '', 'g' ] ) and
     StrIn( OnlyLetters( SrcEdiz ), [ '', 'kg' ] ) then
  begin
    DestEdiz := SrcEdiz;
    Dest := Dest / 1000 + Src;
    Exit;
  end
    else
  if (DestEdiz = '%') and StrIn( OnlyLetters( SrcEdiz ), [ '' ] ) then
  begin
    DestEdiz := SrcEdiz;
    Dest := Dest / 100 + Src;
    Exit;
  end;
  Result := FALSE;
  ErrString := '   ' + DestEdiz + '  ' + SrcEdiz;
end;

function OnlyLetters( const s: String ): String;
var i: Integer;
begin
  Result := AnsiLowerCase( s );
  for i := Length( Result ) downto 1 do
    if not( Result[ i ] in [ ''..'', '', 'a'..'z' ] ) then
      Delete( Result, i, 1 );
end;

function StrIn( const s: String; const A: array of String ): Boolean;
var i: Integer;
begin
  Result := TRUE;
  for i := 0 to High( A ) do
    if AnsiSameStr( s, A[ i ] ) then Exit;
  Result := FALSE;
end;

function PathTrail( const s: String ): String;
var i: Integer;
begin
  for i := Length( s ) downto 1 do
    if s[ i ] = '/' then
    begin
      Result := Trim( Copy( s, i+1, MaxInt ) );
      Exit;
    end;
  Result := s;
end;

function PathLead( const s: String ): String;
var i: Integer;
begin
  for i := Length( s ) downto 1 do
    if s[ i ] = '/' then
    begin
      Result := Trim( Copy( s, 1, i-1 ) );
      Exit;
    end;
  Result := s;
end;

function Max( const Vals: array of Integer ): Integer;
var i: Integer;
begin
  Result := Vals[ 0 ];
  for i := 1 to High( Vals ) do
    if Vals[ i ] > Result then Result := Vals[ i ];
end;

function GetDesktopRect : TRect;
var W1, W2 : HWnd;
begin
  Result := Rect( 0, 0, GetSystemMetrics( SM_CXSCREEN ), GetSystemMetrics( SM_CYSCREEN ) );
  W2 := findwindow(nil,'Program Manager');
  W1 := findwindowex(W2,0,'SHELLDLL_DefView',nil);
  if W1 = 0 then Exit;
  GetWindowRect( W1, Result );
end;

procedure SaveForm( F: TForm );
var R: TRegistry;
begin
  if (F = nil) or (F.Name = '') then Exit;
  R := TRegistry.Create;
  TRY
    R.OpenKey( RegistrySaveForm, TRUE );
    R.WriteInteger( F.Name + '.Left', F.Left );
    R.WriteInteger( F.Name + '.Top', F.Top );
    R.WriteInteger( F.Name + '.Width', F.Width );
    R.WriteInteger( F.Name + '.Height', F.Height );
    R.WriteInteger( F.Name + '.WindowState', Integer( F.WindowState ) )
  FINALLY
    R.Free;
  END;
end;

procedure RestoreForm( F: TForm );
var DR: TRect;
    R: TRegistry;
    FR: TRect;
    WS: TWindowState;
begin
  DR := GetDesktopRect;
  R := TRegistry.Create;
  TRY
    R.OpenKeyReadOnly( RegistrySaveForm );
    if R.ValueExists( F.Name + '.Left' ) then
    begin
      FR.Left := R.ReadInteger( F.Name + '.Left' );
      FR.Top := R.ReadInteger( F.Name + '.Top' );
      if F.BorderStyle = bsSizeable then
      begin
        FR.Right := FR.Left + R.ReadInteger( F.Name + '.Width' );
        FR.Bottom := FR.Top + R.ReadInteger( F.Name + '.Height' );
      end
        else
      begin
        FR.Right := FR.Left + F.Width;
        FR.Bottom := FR.Top + F.Height;
      end;
      WS := TWindowState( R.ReadInteger( F.Name + '.WindowState' ) );
      if WS = wsMaximized then
      begin
        //F.BoundsRect := DR;
        F.WindowState := WS;
      end
      else
      begin
        if FR.Bottom > DR.Bottom then
          OffsetRect( FR, 0, DR.Bottom - FR.Bottom );
        if FR.Right > DR.Right then
          OffsetRect( FR, DR.Right - FR.Right, 0 );
        if FR.Top < DR.Top then
          OffsetRect( FR, 0, DR.Top - FR.Top );
        if FR.Left < DR.Left then
          OffsetRect( FR, DR.Left - FR.Left, 0 );
        if FR.Right - FR.Left > DR.Right - DR.Left then
          FR.Right := FR.Left + DR.Right - DR.Left;
        if FR.Bottom - FR.Top > DR.Bottom - DR.Top then
          FR.Bottom := FR.Top + DR.Bottom - DR.Top;
        F.BoundsRect := FR;
      end;
    end;
  FINALLY
    R.Free;
  END;
end;

function min( x, y: Integer ): Integer;
asm
  cmp eax, edx
  jle @@exit
  mov eax, edx
@@exit:
end;

function Second2Time( sec: DWORD ): String;
begin
  if sec <= 59 then
    Result := IntToStr( sec ) + ' '
  else
    Result := IntToStr( sec div 60 ) + '  ' + IntToStr( sec mod 60 ) + ' ';
end;

function LastChar( const s: String ): Char;
begin
  if s = '' then Result := #0
  else Result := s[ Length( s ) ];
end;

function Extended2Str( E: Extended ): String;
    function UnpackFromBuf( const Buf: array of Byte; N: Integer ): String;
    var I, J, K, L: Integer;
    begin
      SetLength( Result, 16 );
      J := 1;
      for I := 7 downto 0 do
      begin
        K := Buf[ I ] shr 4;
        Result[ J ] := Char( Ord('0') + K );
        Inc( J );
        K := Buf[ I ] and $F;
        Result[ J ] := Char( Ord('0') + K );
        Inc( J );
      end;

      Assert( Result[ 1 ] = '0', 'error!' );
      Delete( Result, 1, 1 );

      if N <= 0 then
      begin
        while N < 0 do
        begin
          Result := '0' + Result;
          Inc( N );
        end;
        Result := '0.' + Result;
      end
        else
      if N < Length( Result ) then
      begin
        Result := Copy( Result, 1, N ) + '.' + Copy( Result, N + 1, MaxInt );
      end
        else
      begin
        while N > Length( Result ) do
        begin
          Result := Result + '0';
        end;
        Exit;
      end;

      L := Length( Result );
      while L > 1 do
      begin
        if not (Result[ L ] in ['0','.']) then break;
        Dec( L );
        if Result[ L + 1 ] = '.' then break;
      end;
      if L < Length( Result ) then Delete( Result, L + 1, MaxInt );

    end;

var
  S: Boolean;
var F: Extended;
    N: Integer;
    Buf1: array[ 0..9 ] of Byte;
    I10: Integer;
begin
  Result := '0';
  if E = 0 then Exit;
  S := E < 0;
  if S then E := -E;

  N := 15;
  F := 5E12;
  I10 := 10;
  while E < F do
  begin
    Dec( N );
    E := E * I10;
  end;
  if N = 15 then
  while E >= 1E13 do
  begin
    Inc( N );
    E := E / I10;
  end;

  while TRUE do
  begin
    asm
      FLD    [E]
      FBSTP  [Buf1]
    end;
    if Buf1[ 7 ] <> 0 then break;
    E := E * I10;
    Dec( N );
  end;

  Result := UnpackFromBuf( Buf1, N );

  if S then Result := '-' + Result;
end;

function Round_Fmt( E: Extended ): Int64;
var s: String;
    i: Integer;
    Negate: Boolean;
begin
    s := Format( '%.0f', [ E ] );
    //Result := StrToInt( s );
    Negate := FALSE;
    i := 1;
    if (s[1] = '-') then
    begin
      i := 2;
      Negate := TRUE;
    end;
    Result := 0;
    for i := i to Length(s) do
    begin
        Result := Result * 10 + Ord( s[i] ) - Ord( '0' );
    end;
    if Negate then
       Result := -Result;
end;

function RussianSummaPropis( i: Int64 ): String;
//  
const Dig: array[ 0..1, 1..9 ] of String = ( ( '', '', '', '',
  '', '', '', '', '' ),
  ( '', '', '', '', '', '', '', '', '' ) );

  function Trojka( i: Integer; nul: Boolean; const end1, end234, end567890: String;
           rod: Integer ): String;
  const Hun: array[ 1..9 ] of String = ( '', '', '', '',
    '', '', '', '', '' );
  const Duz: array[ 1..9 ] of String = ( '', '', '',
    '', '', '', '', '', '' );
  const Duz1: array[ 1..9 ] of String = ( '', '', '',
    '', '', '', '', '',
    '' );
  begin
    Result := '';
    if i < 0 then Exit;
    if (i = 0) then
    if i = 0 then
    begin
      if nul then
        Result := '';
      Exit;
    end;
    if i >= 100 then
    begin
      Result := Result + ' ' + Hun[ i div 100 ];
      i := i mod 100;
    end;
    if i >= 10 then
    begin
      if (i div 10 = 1) and (i <> 10) then
      begin
        Result := Result + ' ' + Duz1[ i - 10 ];
        i := 0;
      end
      else
      begin
        Result := Result + ' ' + Duz[ i div 10 ];
        i := i mod 10;
      end;
    end;
    if i > 0 then
      Result := Result + ' ' + Dig[ rod, i ];
    CASE i OF
    1: Result := Result + ' ' + end1;
    2,3,4: Result := Result + ' ' + end234;
    else   Result := Result + ' ' + end567890;
    END;
  end;
var nul: Boolean;
    j: Integer;
begin
  Result := '';
  nul := TRUE;
  if i >= 1000000000000 then
  begin
    Result := Result + Trojka( i div 1000000000000, FALSE, '', '',
      '', 0 );
    i := i mod 1000000000;
    nul := FALSE;
  end;
  if i >= 1000000000 then
  begin
    Result := Result + Trojka( i div 1000000000, FALSE, '', '',
      '', 0 );
    i := i mod 1000000000;
    nul := FALSE;
  end;
  if i >= 1000000 then
  begin
    Result := Result + ' ' + Trojka( i div 1000000, FALSE, '', '',
      '', 0 );
    i := i mod 1000000;
    nul := FALSE;
  end;
  if i >= 1000 then
  begin
    Result := Result + ' ' + Trojka( i div 1000, FALSE, '', '',
      '', 1 );
    i := i mod 1000;
    nul := FALSE;
  end;
  Result := Result + ' ' + Trojka( i, nul, '', '', '', 0 );
  for j := Length( Result ) - 1 downto 1 do
  begin
    if (Result[ j ] = ' ') and (Result[ j+1 ] = ' ') then
      Delete( Result, j, 1 );
  end;
  Result := Trim( Result );
end;

function Parse( var s: String; const d: String ): String;
var i: Integer;
begin
  i := pos( d, s );
  if i <= 0 then
  begin
    Result := s;
    s := '';
    Exit;
  end;
  Result := Copy( s, 1, i-1 );
  s := Copy( s, i + Length( d ), MaxInt );
end;

function ParseChars( var s: String; const d: String ): String;
var i, j, k: Integer;
begin
  i := -1;
  for j := 1 to Length( d ) do
  begin
    k := pos( d[ j ], s );
    if (i <= 0) or (k <= i) and (k > 0) then i := k;
  end;
  if i <= 0 then
  begin
    Result := s;
    s := '';
    Exit;
  end;
  Result := Copy( s, 1, i-1 );
  s := Copy( s, i + 1, MaxInt );
end;

function CheckFileCanRewrite( const Filename: String ): Boolean;
var F: THandle;
begin
  Result := TRUE;
  TRY
    if not FileExists( Filename ) then Exit;
    Result := FALSE;
    F := CreateFile( PChar( Filename ), GENERIC_READ or GENERIC_WRITE,
      FILE_SHARE_READ, nil, OPEN_EXISTING, FILE_ATTRIBUTE_ARCHIVE, 0 );
    if F = INVALID_HANDLE_VALUE then Exit;
    CloseHandle( F );
    Result := TRUE;
  FINALLY
    if not Result then
      ShowError( ' ' + ExtractFilename( Filename ) + '  , ' +
        '     ,    ' +
        ' .' );
  END;
end;

function IsFloatNum( const s: String ): Boolean;
var i: Integer;
begin
  Result := FALSE;
  if s = '' then Exit;
  for i := 1 to Length( s ) do
    if not(s[ i ] in [ '0'..'9','-',',','.' ]) then Exit;
  TRY
    StrToFloat( s );
    Result := TRUE;
  EXCEPT
  END;
end;

function ScanNumber( const s: String ): Integer;
var i: Integer;
begin
  Result := 0;
  for i := 1 to Length( s ) do
    if s[ i ] in ['0'..'9'] then
      Result := Result * 10 + Ord( s[ i ] ) - Ord( '0' )
    else
      break;
end;

var SaveWinVer: Byte = $FF;

function WinVer : TWindowsVersion;
{* Returns Windows version. }
var OVI: TOsVersionInfo;
begin
  if SaveWinVer <> $FF then Result := TWindowsVersion( SaveWinVer )
  else
  begin
    OVI.dwOSVersionInfoSize := Sizeof( OVI );
    GetVersionEx( OVI );
    with OVI do
    if dwPlatformId = VER_PLATFORM_WIN32_NT then
    begin
      Result := wvNT;
      if dwMajorVersion >= 6 then
        Result := wvLongHorn
      else begin
             if dwMajorVersion >= 5 then
               if dwMinorVersion >=1 then
                 Result := wvXP
               else
                 Result := wvY2K;
           end;
    end
       else
    if dwPlatformId = VER_PLATFORM_WIN32_WINDOWS then
    begin
      Result := wv95;
      if (dwMajorVersion > 4) or (dwMajorVersion = 4)
      and (dwMinorVersion >= 10)  then
        Result := wv98;
    end
       else
       Result := wv31; // Windows 3.1 (WIN32s)
    SaveWinVer := Ord( Result );
  end;
end;

procedure Beep;
begin
  MessageBeep( DWORD( -1 ) );
end;

function Ask( const Question: String ): Boolean;
begin
  Result := MessageBox( Application.Handle, PChar( Question ),
    PChar( Application.Title ), MB_YESNO or MB_DEFBUTTON2 or MB_ICONQUESTION ) = ID_YES;
end;

procedure Msg( const Text: String );
begin
  MessageBox( Application.Handle, PChar( Text ),
    PChar( Application.Title ), MB_OK or MB_SETFOREGROUND or MB_ICONINFORMATION );
end;

procedure Warning( const Text: String );
begin
  MessageBox( Application.Handle, PChar( Text ),
    PChar( Application.Title ), MB_OK or MB_SETFOREGROUND or MB_ICONWARNING );
end;

procedure ShowError( const Text: String );
begin
  MessageBox( Application.Handle, PChar( Text ),
    PChar( Application.Title ), MB_OK or MB_SETFOREGROUND or MB_ICONERROR );
end;

function DirectoryExists( const path: String ): Boolean;
var
  Handle: THandle;
  FindData: TWin32FindData;
begin
  Result := FALSE;
  Handle := FindFirstFile(PChar(path), FindData);
  if Handle <> INVALID_HANDLE_VALUE then
  begin
    Windows.FindClose(Handle);
    if (FindData.dwFileAttributes and FILE_ATTRIBUTE_DIRECTORY) <> 0 then
        Result := TRUE;
  end;
end;

procedure
  LogFileOutput( const filepath, str: AnsiString );
var F: THandle;
    Tmp: AnsiString;
begin
  F := CreateFile( PChar(filepath), GENERIC_WRITE, 0, nil, OPEN_ALWAYS,
                   FILE_ATTRIBUTE_NORMAL, 0 );
  if F = INVALID_HANDLE_VALUE then Exit;
  FileSeek( F, 0, 2 );
  Tmp := str + {$IFDEF LIN} #10 {$ELSE} #13#10 {$ENDIF};
  FileWrite( F, PAnsiChar( Tmp )^, Length( Tmp ) );
  FileClose( F );
end;



initialization
  InitUpper;

end.



