//**  ,     VCL Delphi.
//**  Vladimir Kladov, 2000-2010
//**
//**      .
//**     /     
//**  2          
//**   GNU,  Free Software Foundation.
//**
//**       ,    
//** ,       ,   
//**         
//**   .      
//**     GNU.
//**
//**          
//**   GNU.     ,     Free
//** Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
//** MA 02111-1307, USA.
//******************************************************************************

unit fmuFastLVColsEdit;

interface

{$I VCL_VERSIONS.INC}

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms,
  Dialogs, ExtCtrls, ComCtrls, StdCtrls, FLVCompEditor,
  {$IFDEF _D6orHigher} Variants, DesignIntf {$ELSE} dsgnintf {$ENDIF};

const
  WM_DOSELECTIONS = WM_USER + 100;

type
  {$IFDEF _D6orHigher}
  IFormDesigner = IDesigner;
  {$ENDIF}

  TfmFLVColsEdit = class(TForm)
    Panel1: TPanel;
    btnAdd: TButton;
    btnDel: TButton;
    btnUp: TButton;
    btnDown: TButton;
    chkOnTop: TCheckBox;
    Panel2: TPanel;
    LVCols: TListView;
    procedure LVColsResize(Sender: TObject);
    procedure chkOnTopClick(Sender: TObject);
    procedure btnAddClick(Sender: TObject);
    procedure FormShow(Sender: TObject);
    procedure LVColsSelectItem(Sender: TObject; Item: TListItem;
      Selected: Boolean);
    procedure FormClose(Sender: TObject; var Action: TCloseAction);
    procedure LVColsEdited(Sender: TObject; Item: TListItem;
      var S: String);
    procedure btnDelClick(Sender: TObject);
    procedure FormKeyDown(Sender: TObject; var Key: Word;
      Shift: TShiftState);
    procedure btnUpClick(Sender: TObject);
    procedure btnDownClick(Sender: TObject);
  protected
    procedure WMSelections( var M: TMessage ); message WM_DOSELECTIONS;
  private
    { Private declarations }
    procedure DoSelections;
  public
    { Public declarations }
    DesignColumns: TListColumns;
    DesignObject: TObject;
    procedure ColumnChanged( ColumnObject: TFLVColumn );
    procedure DesignObject_Changed;
  end;

var
  fmFLVColsEdit: TfmFLVColsEdit;

function QueryFormDesigner( D: IDesigner; var FD: IFormDesigner ): Boolean;
procedure RemoveSelection( FD: IFormDesigner );

implementation

{$R *.dfm}

function QueryFormDesigner( D: IDesigner; var FD: IFormDesigner ): Boolean;
begin
  {$IFDEF _D4orHigher}
    Result := D.QueryInterface( IFormDesigner, FD ) = 0;
  {$ELSE}
    Result := False;
    if D is TFormDesigner then
    begin
      FD := D as TFormDesigner;
      Result := True;
    end;
  {$ENDIF}
end;

procedure RemoveSelection( FD: IFormDesigner );
{$IFDEF _D2orD3}
var L: TComponentList;
{$ENDIF}
begin
  try
    {$IFDEF _D5orHigher}
    FD.NoSelection;
    {$ELSE}
      {$IFDEF _D2orD3}
        L := TComponentList.Create;
        FD.SetSelections( L );
        L.Free;
      {$ELSE _D4}
        FD.SetSelections( nil );
      {$ENDIF}
    {$ENDIF}
  except
    ShowMessage( 'EXCEPTION - Could not remove current selection' );
  end;
end;

{ fmFLVColsEdit }

procedure TfmFLVColsEdit.LVColsResize(Sender: TObject);
begin
  LVCols.Columns[ 0 ].Width := LVCols.ClientWidth;
end;

procedure TfmFLVColsEdit.chkOnTopClick(Sender: TObject);
begin
  if chkOnTop.Checked then
  begin
    FormStyle := fsStayOnTop;
    BringToFront;
  end
  else
    FormStyle := fsNormal;
end;

procedure TfmFLVColsEdit.btnAddClick(Sender: TObject);
var Item: TListItem;
    ColObj: TFLVColumn;
    Col: TListColumn;
begin
  Item := LVCols.Selected;
  if Item <> nil then
  begin
    Item := LVCols.Items.Insert(Item.Index + 1);
    Col := DesignColumns.Insert( Item.Index ) as TListColumn;
  end
  else
  begin
    Item := LVCols.Items.Add;
    Col := DesignColumns.Add;
  end;
  FLVColumn_InsertFieldName( DesignObject, Item.Index, '' );
  ColObj := TFLVColumn.Create(Self);
  ColObj.Caption := Col.Caption;
  ColObj.ImageIndex := -1;
  ColObj.Visible := TRUE;
  ColObj.Order := -1;
  ColObj.FDesigner := Self;
  Item.Data := ColObj;
  LVCols.Selected := Item;
  Item.EditCaption;
end;

procedure TfmFLVColsEdit.FormShow(Sender: TObject);
var I: Integer;
    LI: TListItem;
    ColObj: TFLVColumn;
begin
  LVCols.Items.Clear;
  for I := 0 to DesignColumns.Count-1 do
  begin
    LI := LVCols.Items.Add;
    LI.Caption := DesignColumns[ I ].Caption;
    ColObj := TFLVColumn.Create( Self );
    //ColObj.Name := TComponent( DesignObject ).Name + '.Columns[' + IntToStr( I ) + ']';
    ColObj.Caption    := DesignColumns[ I ].Caption;
    ColObj.Alignment  := DesignColumns[ I ].Alignment;
    ColObj.Autosize   := DesignColumns[ I ].AutoSize;
    ColObj.ImageIndex := DesignColumns[ I ].ImageIndex;
    ColObj.MaxWidth   := DesignColumns[ I ].MaxWidth;
    ColObj.MinWidth   := DesignColumns[ I ].MinWidth;
    ColObj.Width      := DesignColumns[ I ].Width;
    ColObj.SortAttrs  := GetFLVColumn_SortAttrs( DesignObject, I );
    ColObj.Filtered   := GetFLVColumn_Filtered( DesignObject, I );
    ColObj.Visible    := GetFLVColumn_Visible( DesignObject, I );
    ColObj.Order      := GetFLVColumn_Order( DesignObject, I );
    ColObj.FieldName  := GetFLVColumn_FieldName( DesignObject, I );

    ColObj.CanEdit    := GetFLVColumn_CanEdit( DesignObject, I );
    ColObj.EditKind   := GetFLVColumn_EditKind( DesignObject, I );
    ColObj.ComboItems := GetFLVColumn_ComboItems( DesignObject, I );

    ColObj.FDesigner  := Self;
    LI.Data := ColObj;
  end;
end;

procedure TfmFLVColsEdit.ColumnChanged(ColumnObject: TFLVColumn);
var I: Integer;
    LI: TListItem;
begin
  for I := 0 to LVCols.Items.Count-1 do
  begin
    LI := LVCols.Items[ I ];
    if LI.Data = ColumnObject then
    begin
      LI.Caption := ColumnObject.Caption;
      DesignColumns[ I ].Caption := ColumnObject.Caption + ' ';
      DesignColumns[ I ].Alignment := ColumnObject.Alignment;
      DesignColumns[ I ].AutoSize := ColumnObject.Autosize;
      DesignColumns[ I ].ImageIndex := ColumnObject.ImageIndex;
      DesignColumns[ I ].MaxWidth := ColumnObject.MaxWidth;
      DesignColumns[ I ].MinWidth := ColumnObject.MinWidth;
      DesignColumns[ I ].Tag := ColumnObject.Tag;
      DesignColumns[ I ].Width := ColumnObject.Width;

      SetFLVColumn_SortAttrs( DesignObject, I, ColumnObject.SortAttrs );
      SetFLVColumn_Filtered( DesignObject, I, ColumnObject.Filtered );
      SetFLVColumn_Visible( DesignObject, I, ColumnObject.Visible );
      SetFLVColumn_Order( DesignObject, I, ColumnObject.Order );
      SetFLVColumn_FieldName( DesignObject, I, ColumnObject.FieldName );

      SetFLVColumn_CanEdit( DesignObject, I, ColumnObject.CanEdit );
      SetFLVColumn_EditKind( DesignObject, I, ColumnObject.EditKind );
      SetFLVColumn_ComboItems( DesignObject, I, ColumnObject.ComboItems );

      DesignObject_Changed;
      DesignColumns[ I ].Caption := ColumnObject.Caption;
      break;
    end;
  end;
end;

procedure TfmFLVColsEdit.LVColsSelectItem(Sender: TObject; Item: TListItem;
  Selected: Boolean);
var M: TMsg;
begin
  //DoSelections;
  if not PeekMessage( M, Handle, WM_DOSELECTIONS, WM_DOSELECTIONS, PM_NOREMOVE ) then
    PostMessage( Handle, WM_DOSELECTIONS, 0, 0 );
end;

procedure TfmFLVColsEdit.FormClose(Sender: TObject;
  var Action: TCloseAction);
begin
  LVCols.Selected := nil;
  LVCols.Items.Clear;
  LVColsSelectItem(Self, nil, FALSE);
end;

procedure TfmFLVColsEdit.LVColsEdited(Sender: TObject; Item: TListItem;
  var S: String);
var ColObj: TFLVColumn;
begin
  ColObj := Item.Data;
  ColObj.Caption := S;
  LVColsSelectItem(Self, nil, FALSE);
end;

procedure TfmFLVColsEdit.DoSelections;
var F: TForm;
    Fr: TFrame;
    D: IDesigner;
    FD: IFormDesigner;
    I, J: Integer;
    LI: TListItem;
    Sels: IDesignerSelections;
    ColObj: TFLVColumn;
begin
  if DesignObject <> nil then
  begin
    Fr := nil;
    F := nil;
    if (DesignObject as TControl).Owner is TForm then
      F := (DesignObject as TControl).Owner as TForm
    else
      Fr := (DesignObject as TControl).Owner as TFrame;
    if F <> nil then
    begin
        //Rpt( 'Form found: ' + F.Name );

  {$IFDEF _D6orHigher}                                  //
        F.Designer.QueryInterface(IFormDesigner,D);     //
  {$ELSE}                                               //
        D := F.Designer;
  {$ENDIF}                                              //
        if D <> nil then
        begin
          //Rpt( 'IDesigner interface returned' );
          if QueryFormDesigner( D, FD ) then
          begin
            //Rpt( 'IFormDesigner interface quered' );
            try
              if LVCols.Items.Count = 0 then
                RemoveSelection( FD )
              else
              begin
                Sels := CreateSelectionList;
                TRY
                FD.GetSelections( Sels );
                EXCEPT
                ShowMessage( 'Could not get selection!' );
                END;
                if Assigned( Sels) and (Sels.Count > 0) then
                begin
                  for I := 0 to Sels.Count-1 do
                  begin
                    for J := 0 to LVCols.Items.Count-1 do
                    begin
                      LI := LVCols.Items[ J ];
                      if Pointer( Sels.Items[ I ] ) = LI.Data then Continue
                      else
                      begin
                        Sels := nil;
                        RemoveSelection( FD );
                        break;
                      end;
                    end;
                    if Sels = nil then break;
                  end;
                end;
                Sels := nil;
              end;
              J := 0;
              if LVCols.SelCount = 1 then
              for I := 0 to LVCols.Items.Count-1 do
              begin
                LI := LVCols.Items[ I ];
                if LI.Selected then
                begin
                  FD.SelectComponent( LI.Data );
                  Inc( J );
                end;
              end
                else
              if LVCols.SelCount > 1 then
              begin
                Sels := CreateSelectionList;
                TRY
                  for I := 0 to LVCols.Items.Count-1 do
                  begin
                    LI := LVCols.Items[ I ];
                    if LI.Selected then
                    begin
                      TRY
                      {$IFDEF _D6orHigher}
                      ColObj := LI.Data;
                      Sels.Add( ColObj );
                      {$ELSE}
                      ColObj := LI.Data;
                      Sels.Add( MakeIPersistent( ColObj ) );
                      {$ENDIF}
                      EXCEPT
                      ShowMessage( 'Can not add item to SelectionList, I=' + IntToStr( I ) );
                      raise;
                      END;
                    end;
                  end;
                  FD.SetSelections( Sels );
                  J := Sels.Count;
                FINALLY
                  Sels := nil;
                END;
              end;
              if J = 0 then
                FD.SelectComponent( DesignObject as TPersistent );
            except
              ShowMessage( 'EXCEPTION *** Could not clear selection!' )
            end;
          end;

        end;
    end
      else
    if Assigned( Fr.Owner ) and (Fr.Owner is TForm) then
    begin
        //Rpt( 'Form found: ' + F.Name );

  {$IFDEF _D6orHigher}                                  //
        (Fr.Owner as TForm).Designer.QueryInterface(IFormDesigner,D);     //
  {$ELSE}                                               //
        D := (Fr.Owner as TForm).Designer;
  {$ENDIF}                                              //
        if D <> nil then
        begin
          //Rpt( 'IDesigner interface returned' );
          if QueryFormDesigner( D, FD ) then
          begin
            //Rpt( 'IFormDesigner interface quered' );
            try
              if LVCols.Items.Count = 0 then
                RemoveSelection( FD )
              else
              begin
                Sels := CreateSelectionList;
                TRY
                FD.GetSelections( Sels );
                EXCEPT
                ShowMessage( 'Could not get selection!' );
                END;
                if Assigned( Sels) and (Sels.Count > 0) then
                begin
                  for I := 0 to Sels.Count-1 do
                  begin
                    for J := 0 to LVCols.Items.Count-1 do
                    begin
                      LI := LVCols.Items[ J ];
                      if Pointer( Sels.Items[ I ] ) = LI.Data then Continue
                      else
                      begin
                        Sels := nil;
                        RemoveSelection( FD );
                        break;
                      end;
                    end;
                    if Sels = nil then break;
                  end;
                end;
                Sels := nil;
              end;
              J := 0;
              if LVCols.SelCount = 1 then
              for I := 0 to LVCols.Items.Count-1 do
              begin
                LI := LVCols.Items[ I ];
                if LI.Selected then
                begin
                  FD.SelectComponent( LI.Data );
                  Inc( J );
                end;
              end
                else
              if LVCols.SelCount > 1 then
              begin
                Sels := CreateSelectionList;
                TRY
                  for I := 0 to LVCols.Items.Count-1 do
                  begin
                    LI := LVCols.Items[ I ];
                    if LI.Selected then
                    begin
                      TRY
                      {$IFDEF _D6orHigher}
                      ColObj := LI.Data;
                      Sels.Add( ColObj );
                      {$ELSE}
                      ColObj := LI.Data;
                      Sels.Add( MakeIPersistent( ColObj ) );
                      {$ENDIF}
                      EXCEPT
                      ShowMessage( 'Can not add item to SelectionList, I=' + IntToStr( I ) );
                      raise;
                      END;
                    end;
                  end;
                  FD.SetSelections( Sels );
                  J := Sels.Count;
                FINALLY
                  Sels := nil;
                END;
              end;
              if J = 0 then
                FD.SelectComponent( DesignObject as TPersistent );
            except
              ShowMessage( 'EXCEPTION *** Could not clear selection!' )
            end;
          end;

        end;
    end;
  end;
end;

procedure TfmFLVColsEdit.WMSelections(var M: TMessage);
begin
  DoSelections;
  btnDel.Enabled := LVCols.SelCount > 0;
  btnUp.Enabled := (LVCols.SelCount = 1) and (LVCols.Selected.Index > 0);
  btnDown.Enabled := (LVCols.SelCount = 1) and (LVCols.Selected.Index < LVCols.Items.Count-1);
end;

procedure TfmFLVColsEdit.btnDelClick(Sender: TObject);
var I: Integer;
    LI: TListItem;
    LI1: TListItem;
    ColObj: TFLVColumn;
begin
  for I := LVCols.Items.Count-1 downto 0 do
  begin
    LI := LVCols.Items[ I ];
    if not LI.Selected then Continue;

    LVCols.Selected := nil;
    DoSelections;
    if I < LVCols.Items.Count-1 then
      LI1 := LVCols.Items[ I + 1 ]
    else if I > 0 then
      LI1 := LVCols.Items[ I - 1 ]
    else
      LI1 := nil;
    if LI1 <> nil then
      LVCols.Selected := LI1;

    ColObj := LI.Data;
    ColObj.Free;
    LI.Data := nil;
    LI.Free;
    DesignColumns.Delete( I );

    FLVColumn_DeleteFieldName( DesignObject, I );
    break;

  end;
end;

procedure TfmFLVColsEdit.FormKeyDown(Sender: TObject; var Key: Word;
  Shift: TShiftState);
begin
  if (Key = VK_DELETE) and (btnDel.Enabled) then
  begin
    btnDel.Click;
    Key := 0;
  end;
end;

procedure TfmFLVColsEdit.btnUpClick(Sender: TObject);
var LI, LI1: TListItem;
    LC, LC1: TListColumn;
    ColObj: TFLVColumn;
begin
  LVColsSelectItem( Self, nil, FALSE );
  LI := LVCols.Selected;
  if (LI = nil) or (LI.Index <= 0) then Exit;
  FLVColumn_MoveFieldName( DesignObject, LI.Index, LI.Index - 1 );
  ColObj := LI.Data;
  LC := DesignColumns.Items[ LI.Index ];
  LC1 := DesignColumns.Insert( LI.Index - 1 ) as TListColumn;
  LC1.Assign( LC );
  LC.Free;
  LI1 := LVCols.Items.Insert( LI.Index - 1);
  LI1.Caption := LI.Caption;
  LI1.Data := ColObj;
  LI.Data := nil;
  LI.Free;
  LI1.Selected := TRUE;
  DesignObject_Changed;
end;

procedure TfmFLVColsEdit.btnDownClick(Sender: TObject);
var LI, LI1: TListItem;
    LC, LC1: TListColumn;
    ColObj: TFLVColumn;
begin
  LVColsSelectItem( Self, nil, FALSE );
  LI := LVCols.Selected;
  if (LI = nil) or (LI.Index >= LVCols.Items.Count-1) then Exit;
  FLVColumn_MoveFieldName( DesignObject, LI.Index, LI.Index + 1 );
  ColObj := LI.Data;
  LC := DesignColumns.Items[ LI.Index ];
  LC1 := DesignColumns.Insert( LI.Index + 2 ) as TListColumn;
  LC1.Assign( LC );
  LC.Free;
  LI1 := LVCols.Items.Insert( LI.Index + 2);
  LI1.Caption := LI.Caption;
  LI1.Data := ColObj;
  LI.Data := nil;
  LI.Free;
  LI1.Selected := TRUE;
  DesignObject_Changed;
end;

procedure TfmFLVColsEdit.DesignObject_Changed;
begin
  if ((DesignObject as TComponent).Owner <> nil) and
     ((DesignObject as TComponent).Owner is TForm) then
  begin
    ((DesignObject as TComponent).Owner as TForm).Designer.Modified;
  end
    else
  if ((DesignObject as TComponent).Owner <> nil) and
     ((DesignObject as TComponent).Owner is TFrame) then
  begin
    (((DesignObject as TComponent).Owner as TFrame).Owner as TForm).Designer.Modified;
  end;
end;

end.
