//**  ,     VCL Delphi.
//**  Vladimir Kladov, 2000-2010
//**
//**      .
//**     /     
//**  2          
//**   GNU,  Free Software Foundation.
//**
//**       ,    
//** ,       ,   
//**         
//**   .      
//**     GNU.
//**
//**          
//**   GNU.     ,     Free
//** Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
//** MA 02111-1307, USA.
//******************************************************************************

unit fmuFastLVFilter;

interface

{$I VCL_VERSIONS.INC}

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  ComCtrls, ColorListView, FastListview, ExtCtrls, Menus, CLiPBRD;

type
  TfmFilterList = class(TForm)
    Panel1: TPanel;
    Shape1: TShape;
    PopupMenu1: TPopupMenu;
    pmiSelectAll: TMenuItem;
    pmiInvertSel: TMenuItem;
    N1: TMenuItem;
    pmiSave: TMenuItem;
    SaveDialog1: TSaveDialog;
    pmiCopy: TMenuItem;
    procedure FormCreate(Sender: TObject);
    procedure FormShow(Sender: TObject);
    procedure FormResize(Sender: TObject);
    procedure pmiSelectAllClick(Sender: TObject);
    procedure pmiInvertSelClick(Sender: TObject);
    procedure PopupMenu1Popup(Sender: TObject);
    procedure lvValuesCompare(Sender: TObject; Item1,
      Item2: TFastListItem; var Compare: Integer);
    procedure FormKeyDown(Sender: TObject; var Key: Word;
      Shift: TShiftState);
    procedure pmiSaveClick(Sender: TObject);
    procedure pmiCopyClick(Sender: TObject);
    procedure FormCloseQuery(Sender: TObject; var CanClose: Boolean);
    procedure FastListview1Resize(Sender: TObject);
    procedure FormClose(Sender: TObject; var Action: TCloseAction);
  private
    FPreventClose: Integer;
    WantClose: Boolean;
    { Private declarations }
  protected
    procedure CutWindow;
  public
    lvValues: TFastListview;
    FilteredLV: TFastListview;
    Filteredcolumn: Integer;
    { Public declarations }
    function CaptionHeigth: Integer;
    procedure PreventClose(const Value: Boolean );
    function CanClose: Boolean;
  end;

var
  fmFilterList: TfmFilterList;

implementation

uses
  FLVResources;

type
  TFakeFastLV = class( TFastListview )
  end;

{$R *.DFM}

{ TfmFilterList }

function TfmFilterList.CaptionHeigth: Integer;
var P: TPoint;
begin
  P := ClientToScreen( Point( 0, 0 ) );
  Result := P.y - Top;
end;

procedure TfmFilterList.CutWindow;
var Rgn: HRgn;
begin
  Rgn := CreateRectRgn( 0, CaptionHeigth, Screen.Width, Screen.Height );
  SetWindowRgn( Handle, Rgn, TRUE );
end;

procedure TfmFilterList.FormCreate(Sender: TObject);
begin
  lvValues := TFastListView.Create( Self );
  lvValues.ViewStyle := vsReport;
  //DisableAutoRange;
  lvValues.Parent := Self;
  //EnableAutoRange;
  lvValues.Align := alClient;
  lvValues.ShowColumnHeaders := FALSE;
  lvValues.ReadOnly := TRUE;
  lvValues.RowSelect := TRUE;
  lvValues.MultiSelect := TRUE;
  lvValues.GridLines := TRUE;
  lvValues.Columns.Add;
  lvValues.Columns.Add;
  lvValues.AutoExpandColumn := 0;
  lvValues.PopupMenu := PopupMenu1;
  lvValues.OnCompare := lvValuesCompare;
  lvValues.OnResize := FastListview1Resize;

  //  - :
  pmiSelectAll.Caption := FiltPopupMenu1;
  pmiInvertSel.Caption := FiltPopupMenu2;
  pmiSave.Caption := FiltPopupMenu3;
  pmiCopy.Caption := FiltPopupMenu4;
  SaveDialog1.Title := FiltSaveDlgTitle;
  SaveDialog1.Filter := FiltSaveDlgFilter;
end;

procedure TfmFilterList.FormShow(Sender: TObject);
begin
  CutWindow;
end;

procedure TfmFilterList.FormResize(Sender: TObject);
begin
  CutWindow;
end;

procedure TfmFilterList.pmiSelectAllClick(Sender: TObject);
begin
  lvValues.SelectAll;
end;

procedure TfmFilterList.pmiInvertSelClick(Sender: TObject);
begin
  lvValues.InvertSelection;
end;

procedure TfmFilterList.PopupMenu1Popup(Sender: TObject);
begin
  pmiSelectAll.Enabled := lvValues.SelCount < lvValues.Items.Count;
  pmiInvertSel.Enabled := lvValues.Items.Count > 0;
end;

procedure TfmFilterList.lvValuesCompare(Sender: TObject; Item1,
  Item2: TFastListItem; var Compare: Integer);
begin
  if Item1.FontColor <> Item2.FontColor then
  begin
    if Item1.FontColor = clNone then
      Compare := -1
    else
      Compare := 1
  end
    else
    Compare := 0;
    {else
    Compare := CompareText( Item1.Caption, Item2.Caption );}
end;

procedure TfmFilterList.FormKeyDown(Sender: TObject; var Key: Word;
  Shift: TShiftState);
var I, FromCol, ToCol, Step: Integer;
begin
  case Key of
  VK_LEFT:
    begin
      FromCol := Filteredcolumn - 1;
      if FromCol < 0 then
        FromCol := FilteredLV.Columns.Count-1;
      ToCol := Filteredcolumn + 1;
      if Tocol >= FilteredLV.Columns.Count then
        ToCol := 0;
      Step := -1;
    end;
  VK_RIGHT:
    begin
      FromCol := Filteredcolumn + 1;
      if FromCol >= FilteredLV.Columns.Count then
        FromCol := 0;
      ToCol := Filteredcolumn - 1;
      if Tocol < 0 then
        ToCol := FilteredLV.Columns.Count-1;
      Step := 1;
    end;
  VK_ESCAPE:
    begin
      Key := 0;
      close;
      Exit;
    end;
  else
    Exit;
  end;
  I := FromCol;
  while TRUE do
  begin
    if Step > 0 then
    begin
      if (I < FromCol) and (I > ToCol) then Exit;
      if I >= FilteredLV.Columns.Count then I := 0;
    end;
    if Step < 0 then
    begin
      if (I > FromCol) and (I < ToCol) then Exit;
      if (I < 0) then I := FilteredLV.Columns.Count - 1;
    end;
    if FilteredLV.ColumnVisible[ I ] then
    if FilteredLV.ColumnFiltered[ I ] then
    begin
      PostMessage( FilteredLV.Handle, WM_USER_CALLFIRSTFILTER, I, 0 );
      Exit;
    end;
    if I = ToCol then Exit;
    Inc( I, Step );
  end;
end;

procedure TfmFilterList.pmiSaveClick(Sender: TObject);
var SL: TStringList;
    SD: TSaveDialog;
begin
  SD := TSaveDialog.Create( nil );
  SL := TStringList.Create;
  TRY
    SL.Text := lvValues.ContentAsText;
    SD.DefaultExt := SaveDialog1.DefaultExt;
    SD.Filter := SaveDialog1.Filter;
    SD.Options := SaveDialog1.Options;
    SD.Title := SaveDialog1.Title;
    SD.InitialDir := SaveDialog1.InitialDir;
    if SD.Execute then
      SL.SaveToFile( SD.FileName );
  FINALLY
    SD.Free;
    SL.Free;
  END;
end;

procedure TfmFilterList.pmiCopyClick(Sender: TObject);
var CB: TClipboard;
begin
  CB := TClipboard.Create;
  TRY
    CB.Open;
    CB.SetTextBuf( PChar( lvValues.ContentAsText ) );
    CB.Close;
  FINALLY
    CB.Free;
  END;
end;

procedure TfmFilterList.PreventClose(const Value: Boolean);
begin
  If Value then
    Inc( FPreventClose )
  else
  begin
    Dec( FPreventClose );
    if FPreventClose <= 0 then
    if WantClose then
      Close;
  end;
end;

procedure TfmFilterList.FormCloseQuery(Sender: TObject;
  var CanClose: Boolean);
begin
  if not CanClose then
  begin
    WantClose := TRUE;
    CanClose := FALSE;
    Hide;
  end;
end;

procedure TfmFilterList.FastListview1Resize(Sender: TObject);
begin
  if lvValues = nil then Exit;
  if WantClose then Exit;
  if not Visible then Exit;
  if lvValues.Columns = nil then Exit;
  if lvValues.Columns.Count < 2 then Exit;
  lvValues.Columns[ 0 ].Width := lvValues.ClientWidth - lvValues.Columns[ 1 ].Width;
end;

function TfmFilterList.CanClose: Boolean;
begin
  Result := FPreventClose <= 0;
end;

procedure TfmFilterList.FormClose(Sender: TObject;
  var Action: TCloseAction);
begin
   //FFilterListForm := nil;
   TFakeFastLV( FilteredLV ).NilFFilterListForm;
end;

end.
