//------------------------------------------------------------------------------
//  
//  :
// -  TParamData -   
// -  TParams -       
//                 (   TVars -    
//                 )
//
//  .. 03.08.2004 - 
//             10.08.2004 -  used_in_op -    TParamController
//------------------------------------------------------------------------------
unit ParamUnit;

interface

uses Windows, Classes, SysUtils, FormulaInterpreterUnit, AdoSqlUnit, ADOint,
     QSort, math;

//{$DEFINE DEBUG_SUM} //   

const
  MAX_PRECISION = 1000000;

type
  TParams = class;

  TParamsSet = class( TComponent )
  protected
    function SumCached( P: TParams; Idx, fPos: Integer; var F: String ): Boolean;
      virtual;
    procedure DoSumCache( P: TParams; Idx, fPos: Integer; const F: String; Value: Double );
      virtual;
  end;

  //----------------------------------------------------------------------------
  //     (,  ,  ,  )
  TParamSetData = class
    //--------------------------------------------------------------------------
    // ID   nomenklinfo
    nomenklid: Integer;
    //--------------------------------------------------------------------------
    // ID 
    con_id: Integer;
    //--------------------------------------------------------------------------
    //  
    numcon: String;
    //--------------------------------------------------------------------------
    // ID 
    batchid: Integer;
    //--------------------------------------------------------------------------
    // ID 
    opid: Integer;
    //--------------------------------------------------------------------------
    //         
    ignore_conditions: Boolean;
    //--------------------------------------------------------------------------
    //  
    code_operation: String;
    procedure Assign( From: TParamSetData ); virtual;
  end;

  //----------------------------------------------------------------------------
  //   
  TParamData = class
  protected
    //--------------------------------------------------------------------------
    //     ( :   
    //  ,    )
    param_started_calc: Boolean;
    //--------------------------------------------------------------------------
    //
    param_error: Boolean;
  public
    //--------------------------------------------------------------------------
    //  ,   
    param_ready: Boolean;
    FDepending: TList;
    constructor Create;
    destructor Destroy; override;
  public
    //--------------------------------------------------------------------------
    // . .
    Units: String;
    //--------------------------------------------------------------------------
    //  : 1 - , 2 - , 3 -  , 4 - 
    param_type: Integer;
    //--------------------------------------------------------------------------
    //   
    param_format: Double;
    //--------------------------------------------------------------------------
    //   (  )  -
    precision: Integer;
    //--------------------------------------------------------------------------
    //     2
    formula: String;
    //--------------------------------------------------------------------------
    //  -      
    save_formula: String;
    //--------------------------------------------------------------------------
    //      (  
    //  )
    read_only: Boolean;
    //--------------------------------------------------------------------------
    //      
    show_in_op: TStringList;
    //--------------------------------------------------------------------------
    //      5  (   
    //  TParamController)
    used_in_op: TStringList;
    //--------------------------------------------------------------------------
    //     
    param_orders: TList;
    procedure Assign( From: TParamData ); virtual;
    function formula_empty: Boolean;
  end;

  //----------------------------------------------------------------------------
  //  .    TVars -  
  TParams = class( TVars )
  private
    FParamsController: TParamsSet;
    FParamSetData: TParamSetData;
    function GetParamData(Idx: Variant): TParamData;
  protected
    FParamData: TList;
    procedure WhenSummed( ToWhat: TVars; const AddText: String ); override;
    function SumCached( Idx, fPos: Integer; var F: String ): Boolean;
    procedure DoCacheSum( Idx, fPos: Integer; const F: String; Value: Double );
    procedure BuildDependings;
  public
    constructor Create; override;
    constructor CreateOwned( Owner: TParamsSet );
    destructor Destroy; override;
    procedure Clear; override;
    procedure Add( const Name: String; Value: Extended; ID: Integer = 0;
              const Description: String = '' ); override;
    procedure Assign( From: TVars ); override;
    //--------------------------------------------------------------------------
    //  
    property ParamData[ Idx: Variant ]: TParamData read GetParamData;
    property ParamSetData: TParamSetData read FParamSetData;
    procedure SwapItems( i1, i2: Integer ); override;
    function FormattedParameter( Idx: Variant ): String;
    function OK: Boolean; //         
  end;

//------------------------------------------------------------------------------
//  .
// :
//   ParamsList -    TParams, 
//              sum( )   
//            ,     
//            . :  
//              ,     ,
//               .
//   Error - . ,    ,    
//            .
var TimePrepare, TimeCalc: DWORD;
function CalcParams( ParamsList: TList; var Error: String;
         RaiseError: Boolean = FALSE; AllowSum: TAllowSum = nil;
         InitInputVals: Boolean = FALSE; const opname: String = '';
         nomenklid: Integer = 0; Ignore_con_id0: Boolean = FALSE ): Extended;

//------------------------------------------------------------------------------
//    
function LoadParams_by_ConList( const ListCon, Code_Operation: String;
         AllowSum: TAllowSum = nil ): TList;

//------------------------------------------------------------------------------
//    
function LoadParams_by_NomList( const NomList, Code_Operation: String;
         AllowSum: TAllowSum = nil ): TList;

//------------------------------------------------------------------------------
//    
function LoadParams_by_BatList( const BatchList, Code_Operation: String;
         AllowSum: TAllowSum = nil ): TList;

function ParamDependOnParam( P: TParams; I, J: Integer ): Boolean;

implementation

function RemoveSpaces( const S: String ): String;
var I: Integer;
begin
  Result := S;
  for I := Length( S ) downto 1 do
    if S[ I ] <= ' ' then
      Delete( Result, I , 1 );
end;

{function CompareOrders(const Data: Pointer; const e1,e2 : Dword) : Integer;
//------------------------------------------------------------------------------
//     
var P: TParams;
    i1, i2: Integer;
begin
  P := Data;
  i1 := e1;
  i2 := e2;
  if P.ParamData[ i1 ].param_order < P.ParamData[ i2 ].param_order then
    Result := -1
  else
  if P.ParamData[ i1 ].param_order > P.ParamData[ i2 ].param_order then
    Result := 1
  else
    Result := 0;
end;

procedure SwapItems(const Data : Pointer; const e1,e2 : Dword);
//------------------------------------------------------------------------------
//   
var P: TParams;
begin
  P := Data;
  P.SwapItems( e1, e2 );
end;}

function LoadParams_by_ConList( const ListCon: String; const Code_Operation: String;
         AllowSum: TAllowSum = nil ): TList;
//------------------------------------------------------------------------------
//    
var Q1, Q2, Q3, Q5: TQuery;
    P: TParams;
    I, J: Integer;
    ListCon1: String;
    ListParamId: TList;
    ListParamIdAsStr: String;
    ErrorStr: String;
    Ops: array[ 1..5 ] of String;
    S: String;
    Param_found: Boolean;

  procedure AddFixedParam( P: TParams; Prefix: Char; Name: String; Descr: String );
  var K: Integer;
  begin
    P.Add( Prefix + '.' + Name, Q5.FField[ Name ], 0, Descr );
    K := P.Count-1;
    P.ParamData[ K ].param_type := 1;
    P.ParamData[ K ].read_only := TRUE;
  end;

var NoNomenklId: Integer;
begin
  //----------------------------------------------------------------------------
  //    - (   )
  S := Code_Operation;
  for I := 1 to High( Ops ) do Ops[ I ] := '';
  I := 1;
  while (I <= High( Ops )) and (S <> '') do
  begin
    J := pos( ',', S );
    if J <= 0 then J := pos( ';', S );
    if J <= 0 then
    begin
      Ops[ I ] := S;
      S := '';
    end
      else
    begin
      Ops[ I ] := Trim( copy( S, 1, J - 1 ) );
      S := Copy( S, J+1, MaxInt );
    end;
    Inc( I );
  end;

  Result := TList.Create;
  Q1 := TQuery.Create( nil );
  Q2 := TQuery.Create( nil );
  Q3 := TQuery.Create( nil );
  //Q4 := TQuery.Create( nil );
  Q5 := TQuery.Create( nil );
  ListParamId := TList.Create;
  TRY
    Q1.SQL.Text := 'select nomenklid from nomenklinfo where nomenkl = '' ''';
    Q1.Open;
    if not Q1.Eof then
      NoNomenklId := Q1.IFieldN[ 0 ]
    else
      NoNomenklId := 0;
    Q1.Close;
    if Trim( ListCon ) <> '' then
    begin
      //    #B  ,     
      Q1.SQL.Text := 'create table #B (batchid int)' +
        #13'insert into #B ' +
        #13'select distinct pk.batchid from Packaging pk where con_id in (' + ListCon + ')' +
        #13 +
      //     ,   #B
        #13'create table #Pks (con_id int, batchid int null, nomenklid int null) ' +
        #13 +


        #13'insert into #Pks' +
        #13'select pk.con_id, null, null' +
        #13'from Packaging pk' +
        #13'where pk.batchid is null and pk.con_id in' +
        #13'(' + ListCon + ')' +

        #13'insert into #Pks' +
        #13'select pk.con_id, b.batchid, b.nomenklid' +
        #13'from Packaging pk, #B, Batch b' +
        #13'where pk.batchid = #B.batchid' +
        #13'and  #B.batchid = b.batchid' +

        #13'declare @@c int' +
        #13'select @@c = 1' +

        #13'while @@c > 0 begin' +
        #13'insert into #Pks' +
        #13'select pk.Parent_pack, null, null' +
        #13'from Packaging pk, #Pks' +
        #13'where pk.con_id = #Pks.con_id and pk.Parent_pack is not null' +
        #13'and not exists( select * from #Pks s where' +
        #13's.con_id = pk.Parent_pack )' +
        #13'select @@c = @@rowcount' +
        #13'end' +

        #13 +
        #13'select * from #Pks' +
        #13 +
        #13'drop table #Pks';
      //    Param,    
      Q2.SQL.Text :=
        #13'select distinct p.*, p.Format_representation as pfr, ' +
        #13'op.*, op.hidden as op_hidden, op.Format_representation as ofr, o.Code_Operation, ' +
        #13'np.nomenklid, np.Ed_ism as np_Ed_ism ' +
        #13'from Param p, Operation_Params op, Operation o, ' +
        #13'Nomencl_param np, Packaging pk, Batch b, #B' +
        #13'where o.Code_Operation in (' + Code_Operation + ')' +
        #13' and  o.opid = op.opid and op.param_id = p.param_id ' +
        #13' and  np.param_id = p.param_id ' +
        #13' and  (pk.batchid is not NULL and #B.batchid = b.batchid and ' +
        #13'       b.nomenklid = np.nomenklid OR ' +
        #13'      pk.batchid is NULL) ' +
        #13' and p.param_type <> 3' +
        #13'order by op.Operation_param_id';
      //    Param_values_pack
      Q3.SQL.Text :=
        #13'select distinct p.*, np.nomenklid, np.Ed_ism as np_Ed_ism, ' +
        #13'       pv.*, o_p.*, ' +
        #13'(select Code_Operation from Operation o where o.opid = o_p.opid) ' +
        #13'as Code_operation, ' +
        #13'       pk.con_id as pk_con_id ' +
        #13'from ' +
        #13'Nomencl_param np, Packaging pk, Batch b, #B, Param_values_pack pv, ' +
        #13'Param p FULL OUTER JOIN Operation_params o_p on(' +
        #13'o_p.param_id = p.param_id and o_p.opid=(' +
        #13'select opid from Operation where Code_Operation in (' + Code_Operation +  ')))' +
        #13'where np.param_id = p.param_id ' +
        #13' and  (pk.batchid is not NULL and #B.batchid = b.batchid and ' +
        #13'       b.nomenklid = np.nomenklid OR ' +
        #13'      pk.batchid is NULL) ' +
        #13' and  pv.param_id = p.param_id ' +
        #13' and  pv.transpack_id = pk.transpack_id ' +
        #13' and  p.param_type = 3';
      {//    Param_values_batch
      Q4.SQL.Text :=
        #13'select distinct p.*, np.nomenklid, np.Ed_ism as np_Ed_ism, ' +
        #13'       pb.*, o_p.*, ' +
        #13'(select Code_Operation from Operation o where o.opid = o_p.opid), ' +
        #13'       pk.con_id as pk_con_id ' +
        #13'from ' +
        #13'Nomencl_param np, Packaging pk, Batch b, #B, Param_values_batch pb, ' +
        #13'Param p FULL OUTER JOIN Operation_params o_p on(' +
        #13'o_p.param_id = p.param_id and o_p.opid=(' +
        #13'select opid from Operation where Code_Operation in (' + Code_Operation +  ')))' +
        #13'where np.param_id = p.param_id ' +
        #13' and  pk.batchid = b.batchid ' +
        #13' and  b.nomenklid = np.nomenklid ' +
        #13' and  b.batchid = #B.batchid ' +
        #13' and  pb.param_id = p.param_id ' +
        #13' and  pb.batchid = b.batchid ' +
        #13' and  p.param_type = 3';}
      Q1.Open;

      //--------------------------------------------------------------------------
      //   ,   ,   TParams
      while not Q1.Eof do
      begin
        P := TParams.Create;
        Result.Add( P );
        P.ParamSetData.con_id := Q1.IField[ 'con_id' ];
        P.ParamSetData.batchid := Q1.IField[ 'batchid' ];
        P.ParamSetData.nomenklid := Q1.IField[ 'nomenklid' ];
        Q1.Next;
      end;

      //--------------------------------------------------------------------------
      //   
      Q2.Open;
    end
      else //-------------------------------------------------------------------
    begin  //  :      
           //       ,   
           //  
      Q1.SQL.Text := 'create table #B (b int)' +
        #13'select nomenklid from nomenklinfo'#13;
      if Trim( Code_operation ) <> '' then
      begin
        Q2.SQL.Text :=
          #13'select p.*, p.Format_representation as pfr, ' +
          #13'n.nomenklid, op.*, op.hidden as op_hidden, op.Format_representation as ofr, o.Code_Operation, ' +
          #13'np.Ed_ism as np_Ed_ism ' +
          #13'from Param p, Operation_Params op, Operation o, ' +
          #13'Nomencl_param np, Nomenklinfo n' +
          #13'where np.param_id = p.param_id ' +
          #13' and  n.nomenklid = np.nomenklid ' +
          #13' and o.Code_Operation in (' + Code_Operation + ')' +
          #13' and o.opid = op.opid and op.param_id = p.param_id ';
      end
        else
      begin
        Q2.SQL.Text :=
          #13'select distinct p.*, p.Format_representation as pfr, ' +
          #13'n.nomenklid, np.Ed_ism as np_Ed_ism ' +
          #13', null as ofr, null as hidden, null as Code_operation' +
            #13', op.param_order, ' +
            //#13'op.*, ' +
            #13'op.hidden as op_hidden, op.Format_representation as ofr, ' +
            #13'(select o.Code_Operation from Operation o ' +
            #13'where o.opid = op.opid) ' +
          #13'from Nomencl_param np, Nomenklinfo n, Param p ' +
            #13' LEFT OUTER JOIN Operation_params op' +
            #13' ON( op.param_id = p.param_id )' +
          #13'where np.param_id = p.param_id ' +
          #13' and  n.nomenklid = np.nomenklid ';
      end;
      Q1.Open;
      while not Q1.Eof do
      begin
        P := TParams.Create;
        Result.Add( P );
        P.ParamSetData.nomenklid := Q1.IField[ 'nomenklid' ];
        Q1.Next;
      end;
      Q1.Close;
      Q2.Open;
    end;
    //--------------------------------------------------------------------------
    //  :   
    while not Q2.Eof do
    begin
      for I := 0 to Result.Count-1 do
      begin
        P := Result[ I ];
        if (P.ParamSetData.nomenklid = Q2.IField[ 'nomenklid' ]) or
           (P.ParamSetData.nomenklid = 0) and (Q2.IField[ 'nomenklid' ] = NoNomenklId) {or
           (Code_operation <> '')} then
        begin
          if (P.IndexOf( Q2.SField[ 'param_name' ] ) < 0) then
          begin
            P.Add( Q2.SField[ 'param_name' ], Q2.FField[ 'param_val' ],
                   Q2.IField[ 'param_id' ], Q2.SField[ 'param_descr' ] );
            J := P.Count-1;
            P.ParamData[ J ].param_type  := Q2.IField[ 'param_type' ];
            P.ParamData[ J ].formula     := Q2.SField[ 'formula' ];
            if Q2.NullField[ 'np_Ed_ism' ] then
              P.ParamData[ J ].Units := Q2.SField[ 'Ed_ism' ]
            else
              P.ParamData[ J ].Units := Q2.SField[ 'np_Ed_ism' ];

            if Q2.NullField[ 'ofr' ] then
              P.ParamData[ J ].param_format := Q2.FField[ 'pfr' ]
            else
              P.ParamData[ J ].param_format := Q2.FField[ 'ofr' ];

            if Q2.NullField[ 'precis' ] then
              P.ParamData[ J ].precision := MAX_PRECISION
            else
              P.ParamData[ J ].precision := Q2.IField[ 'precis' ];

            if (Q2.IField[ 'hidden' ] = 0) then
              P.ParamData[ J ].show_in_op.Add( Q2.SField[ 'Code_operation' ] );
            P.ParamData[ J ].used_in_op.Add( Q2.SField[ 'Code_operation' ] );

            if Trim( ListCon ) <> '' then
            begin
              P.ParamData[ J ].param_orders.Add( Pointer( Q2.IField[ 'param_order' ] ) );
            end
              else
            begin
              P.ParamData[ J ].param_orders.Add( Pointer( Q2.IField[ 'param_order' ] ) );
            end;
            P.ParamData[ J ].param_ready := P.ParamData[ J ].param_type = 1;
          end
            else //     (,   
                 // ,     ( ))
          begin
            J := P.IndexOf( Q2.SField[ 'param_name' ] );
            if (Q2.IField[ 'hidden' ] = 0) then
              P.ParamData[ J ].show_in_op.Add( Q2.SField[ 'Code_operation' ] );
            P.ParamData[ J ].used_in_op.Add( Q2.SField[ 'Code_operation' ] );
            P.ParamData[ J ].param_orders.Add( Pointer( Q2.IField[ 'param_order' ] ) );
          end;
        end;
      end;
      Q2.Next;
    end;
    Q2.Close;

    if Trim( ListCon ) <> '' then
    begin
      //------------------------------------------------------------------------
      //  (  hidden)    
      // ( Param_values),     ,   
      // hidden   Operation_params ,  
      Q3.Open;
      while not Q3.Eof do
      begin
        for I := 0 to Result.count-1 do
        begin
          P := Result[ I ];
          if (P.ParamSetData.nomenklid = Q3.IField[ 'nomenklid' ]) and
             (P.ParamSetData.con_id = Q3.IField[ 'pk_con_id' ]) then
          begin
            P.Add( Q3.SField[ 'param_name' ], Q3.FField[ 'param_calc' ],
                   Q3.IField[ 'param_id' ], Q3.SField[ 'param_descr' ] );
            J := P.Count-1;
            P.ParamData[ J ].param_type := 3;
            if Q3.NullField[ 'np_Ed_ism' ] then
              P.ParamData[ J ].Units := Q3.SField[ 'Ed_ism' ]
            else
              P.ParamData[ J ].Units := Q3.SField[ 'np_Ed_ism' ];
            P.ParamData[ J ].param_ready := Q3.IField[ 'param_id' ] <> 0;
            //P.ParamData[ J ].param_OK := P.ParamData[ J ].param_ready;

            if (Q3.IField[ 'hidden' ] = 0) and
                         (Q3.IField[ 'Operation_param_id' ] > 0) then
              P.ParamData[ J ].show_in_op.Add( Q3.SField[ 'Code_operation' ] );
            P.ParamData[ J ].used_in_op.Add( Q3.SField[ 'Code_operation' ] );
            P.ParamData[ J ].param_orders.Add( Pointer( Q3.IField[ 'param_order' ] ) );
          end;
        end;
        Q3.Next;
      end;
      Q3.Close;

      //--------------------------------------------------------------------------
      //     ,   
      ListCon1 := '';
      for I := 0 to Result.Count-1 do
      begin
        P := Result[ I ];
        ListCon1 := ListCon1 + ',' + IntToStr( P.ParamSetData.con_id );
        for J := 0 to P.Count-1 do
        begin
          if P.ParamData[ J ].param_type = 3 then
          begin
            P.ParamData[ J ].param_ready := FALSE;
            if ListParamId.IndexOf( Pointer( P.ID[ J ] ) ) < 0 then
              ListParamId.Add( Pointer( P.ID[ J ] ) );
          end;
        end;
      end;
      Delete( ListCon1, 1, 1 );

      ListParamIdAsStr := '';
      for I := 0 to ListParamId.Count-1 do
        ListParamIdAsStr := ListParamIdAsStr + ',' + IntToStr( Integer( ListParamId[ I ] ) );
      Delete( ListParamIdAsStr, 1, 1 );

      if (ListParamIdAsStr <> '') and (ListCon1 <> '') then
      begin
        Q3.SQL.Text := 'select pv.*, pk.con_id, ' + //'np.nomenklid, np.Ed_ism as np_Ed_ism, ' +

          #13'       pv.*, o_p.*, ' +
          #13'(select Code_Operation from Operation o where o.opid = o_p.opid) ' +
          #13'as Code_operation, ' +
          #13'       pk.con_id as pk_con_id, ' +
          #13' p.param_descr, p.param_name, p.precis, p.Format_representation, p.Ed_ism' +

          #13'from Param_values_pack pv, Packaging pk, ' + //'Nomencl_param np, ' +
          #13'Param p FULL OUTER JOIN Operation_params o_p on(' +
          #13'o_p.param_id = p.param_id and o_p.opid=(' +
          #13'select opid from Operation where Code_Operation in (' +
          Code_Operation +  ')))' +
          #13'where ' + //'np.param_id = p.param_id and  ' +
          #13' p.param_type = 3' +
          #13' and  pv.transpack_id = pk.transpack_id ' +
          #13' and  pv.param_id in (' + ListParamIdAsStr + ')' +
          #13' and  (pk.con_id in (' + ListCon1 + ') or ' +
          #13'       pk.con_id in (' + ListCon + '))' +
          #13' and p.param_id = pv.param_id';
        Q3.Open;
        while not Q3.Eof do
        begin
          for I := 0 to Result.Count-1 do
          begin
            P := Result[ I ];
            if P.ParamSetData.con_id = Q3.IField[ 'con_id' ] then
            begin
              Param_found := FALSE;
              for J := 0 to P.Count-1 do
              begin
                if P.ID[ J ] = Q3.IField[ 'param_id' ] then
                begin
                  P.ParamData[ J ].param_ready := TRUE;
                  //P.ParamData[ J ].param_OK := TRUE;
                  P.Values[ J ] := Q3.FField[ 'param_calc' ];
                  Param_found := TRUE;
                  break;
                end;
              end;
              if not Param_found then
              begin
                J := P.Count;
                P.Add( Q3.SField[ 'param_name' ], Q3.FField[ 'param_calc' ],
                  Q3.IField[ 'param_id' ], Q3.SField[ 'param_descr' ] );
                P.ParamData[ J ].param_format := Q3.FField[ 'Format_representation' ];
                P.ParamData[ J ].precision := Q3.IField[ 'Precis' ];
                P.ParamData[ J ].param_type := 3;

                P.ParamData[ J ].Units := Q3.SField[ 'Ed_ism' ];
                P.ParamData[ J ].param_ready := Q3.IField[ 'param_id' ] <> 0;
                //P.ParamData[ J ].param_OK := P.ParamData[ J ].param_ready;

                if (Q3.IField[ 'hidden' ] = 0) and
                             (Q3.IField[ 'Operation_param_id' ] > 0) then
                  P.ParamData[ J ].show_in_op.Add( Q3.SField[ 'Code_operation' ] );
                P.ParamData[ J ].used_in_op.Add( Q3.SField[ 'Code_operation' ] );
                P.ParamData[ J ].param_orders.Add( Pointer( Q3.IField[ 'param_order' ] ) );
              end;
            end;
          end;
          Q3.Next;
        end;
      end;

    end;
    //--------------------------------------------------------------------------
    //  ""    nomenklinfo  mattypeid
    Q5.SQL.Text :=
      #13'select n.*, m.* from Nomenklinfo n, Mattype m' +
      #13'where n.mattypeid = m.mattypeid ' +
      #13 +
      #13'drop table #B';

    Q5.Open;
    while not Q5.Eof do
    begin
      for I := 0 to Result.Count-1 do
      begin
        P := Result[ I ];
        if P.ParamSetData.nomenklid = Q5.IField[ 'nomenklid' ] then
        begin
          AddFixedParam( P, 'N', 'nominal', '. ' );
          AddFixedParam( P, 'N', 'elempct', '.  ' );
          AddFixedParam( P, 'N', 'wrand', '.  ' );
          AddFixedParam( P, 'N', 'wsys' , '.  ' );
          AddFixedParam( P, 'N', 'urand', '.  U235 ' );
          AddFixedParam( P, 'N', 'usys' , '.  U235 ' );
          AddFixedParam( P, 'N', 'srand', '.    ' );
          AddFixedParam( P, 'N', 'ssys' , '.    ' );
          AddFixedParam( P, 'M', 'percentenrich', '.  ' );
          AddFixedParam( P, 'M', 'Accuracy_element', '.   ' );
          AddFixedParam( P, 'M', 'Accuracy_isotope', '.   ' );
        end;
      end;
      Q5.Next;
    end;
    Q5.Close;
    //--------------------------------------------------------------------------
    //   ,  :
    if ListCon <> '' then
    begin
      CalcParams( Result, ErrorStr, FALSE, AllowSum );
      if ErrorStr <> '' then
        raise Exception.Create( '     : ' +
              ErrorStr );
    end;

    {//--------------------------------------------------------------------------
    //  ,    
    for I := 0 to Result.Count-1 do
    begin
      P := Result[ I ];
      SortData( P, P.Count, CompareOrders, SwapItems );
    end;}

  FINALLY
    Q1.Free;
    Q2.Free;
    Q3.Free;
    //Q4.Free;
    Q5.Free;
    ListParamId.Free;
  END;
end;

function LoadParams_by_NomList( const NomList: String; const Code_Operation: String;
         AllowSum: TAllowSum = nil ): TList;
//------------------------------------------------------------------------------
//    
var Q: TQuery;
    ConList: String;
begin
  if Trim( NomList ) <> '' then
  begin
    Q := TQuery.Create( nil );
    TRY
      Q.SQL.Text := 'select pk.con_id from Packaging pk, Batch b ' +
        #13'where pk.batchid = b.batchid and b.nomenklid in (' + NomList + ')';
      Q.Open;
      ConList := '';
      while not Q.Eof do
      begin
        ConList := ConList + ',' + Q.SFieldN[ 0 ];
        Q.Next;
      end;
      Delete( ConList, 1, 1 );
      Result := LoadParams_by_ConList( ConList, Code_Operation, AllowSum );
    FINALLY
      Q.Free;
    END;
  end
    else //    ,   
  begin
    Result := LoadParams_by_ConList( '', Code_Operation, AllowSum );
  end;
end;

function LoadParams_by_BatList( const BatchList, Code_Operation: String;
         AllowSum: TAllowSum = nil ): TList;
//------------------------------------------------------------------------------
//    
var Q: TQuery;
    ConList: String;
begin
  Q := TQuery.Create( nil );
  TRY
    Q.SQL.Text := 'select pk.con_id from Packaging pk where pk.batchid in (' +
      BatchList + ')';
    Q.Open;
    ConList := '';
    while not Q.Eof do
    begin
      ConList := ConList + ',' + Q.SFieldN[ 0 ];
      Q.Next;
    end;
    Delete( ConList, 1, 1 );
    Result := LoadParams_by_ConList( ConList, Code_Operation, AllowSum );
  FINALLY
    Q.Free;
  END;
end;

function ParamDependOnParam( P: TParams; I, J: Integer ): Boolean;
//------------------------------------------------------------------------------
// TRUE,     I      J
begin
  Result := P.ParamData[ I ].FDepending.IndexOf( Pointer( J ) ) >= 0;
end;

//------------------------------------------------------------------------------
//        
function CalcParams( ParamsList: TList; var Error: String; RaiseError: Boolean = FALSE;
         AllowSum: TAllowSum = nil; InitInputVals: Boolean = FALSE;
         const opname: String = ''; nomenklid: Integer = 0; Ignore_con_id0: Boolean = FALSE ): Extended;

  procedure CalcSingleFormula( P: TParams; Idx: Integer );
  var fpos, L: Integer;
      F: String;
      DName: String;
      I, J, ParamIdx, SumLevel, Level, precis: Integer;
      P1: TParams;
      will_sum: Boolean;
      X: Extended;
      PD, PD1: TParamData;
  begin
    PD := P.ParamData[ Idx ];
    if (PD.param_type = 1) then Exit;
    if PD.param_error then Exit;
    if PD.param_started_calc then
    begin
      Error := '   ' + P.Names[ Idx ];
      raise Exception.Create( Error );
    end;
    PD.param_started_calc := TRUE;
    TRY
      //------------------------------------------------------------------------
      //     ,    
      //   .   ,    
      //   .
      F := RemoveSpaces( PD.formula );
      I := pos( ';', F );
      if I > 0 then
        F := Trim( Copy( F, 1, I-1 ) );
      if F = '' then
      begin
        PD.param_ready := TRUE;
        //PD.param_OK := TRUE;
        Exit;
      end;
      fpos := 1;
      SumLevel := -1;
      Level := 0;
      while fpos <= Length( F ) do
      begin
        if F[ fpos ] in [ ' ','+','-','*','/','<','>','=','(',')',#1..' ', '0'..'9', DecimalSeparator ] then
        begin
          if F[ fpos ] = '(' then Inc( Level )
          else
          if F[ fpos ] = ')' then
          begin
            Dec( Level );
            if Level <= SumLevel then
              SumLevel := -1;
          end;
          Inc( fpos );
        end
        else
        if (F[ fpos ] in [ 's', 'S' ]) and
           (CompareText( Copy( F, fpos, 4 ), 'sum(' ) = 0) then
        begin
          if not P.SumCached( Idx, fpos, F ) then
          begin
            SumLevel := Level;
            Inc( Level );
            Inc( fpos, 4 );
          end;
        end
        else
        begin
          L := fpos;
          while L <= Length( F ) do
          begin
            Inc( L );
            if (L > Length( F )) or (F[ L ] in [ ' ','+','-','*','/','<','>','=','(',')',#1..#31,',',';' ]) then
            begin
              DName := Copy( F, fpos, L - fpos );
              if (DName <> '') and
                 not(DName[ 1 ] in [ '.', ',', DecimalSeparator, '0'..'9', '-', '+' ]) then
                if SumLevel >= 0 then //  ,  
                                 //    sum( ) -  
                                 //     .
                  for ParamIdx := 0 to ParamsList.Count-1 do
                  begin
                    P1 := ParamsList[ ParamIdx ];
                    J := P1.IndexOf( DName );
                    if J < 0 then
                    begin
                      will_sum := TRUE;
                      if Assigned( AllowSum ) then
                        AllowSum( P, P1, will_sum );
                      if will_sum then
                      begin
                        Error := ' ' + DName + '       ' +
                              P.Names[ Idx ] + ' ( ' + IntToStr( nomenklid ) + ')';
                        if P.ParamSetData.con_id <> 0 then
                          Error := Error + ',  ' + P.ParamSetData.numcon;
                        raise Exception.Create( Error );
                      end
                        else J := -1;
                    end;
                    if (J >= 0) and not P1.ParamData[ J ].param_ready AND
                       not P1.ParamData[ J ].param_error then
                    begin
                      CalcSingleFormula( P1, J );
                      PD1 := P1.ParamData[ J ];
                      if not PD1.param_ready then
                      begin
                        Error := ' ' + DName + '    ' +
                          ' ' + P.Names[ Idx ] + ',      ' +
                          ', ..     ' + opname;
                        P1.ParamData[ J ].param_error := TRUE;
                        raise Exception.Create( Error );
                      end;
                    end;
                  end
                else //      , ..
                     //    .
                  if ( (nomenklid = 0) or (nomenklid = P.ParamSetData.nomenklid) ) and
                     not Ignore_con_id0 or Ignore_con_id0 and (P.ParamSetData.con_id <> 0) then
                  begin
                    if (DName <> '') and (DName[ 1 ] in [ '0'..'9', DecimalSeparator, '-' ]) then
                    begin
                      J := 999; L := fpos;
                    end
                    else
                      J := P.IndexOf( DName );
                    if J < 0 then
                    begin
                      Error := ' ' + DName + '      ' +
                            P.Names[ Idx ] + ' ( ' + IntToStr( nomenklid ) + ')';
                      if P.ParamSetData.con_id <> 0 then
                        Error := Error + ',  ' + P.ParamSetData.numcon;
                      raise Exception.Create( '' );
                    end;
                    if not P.ParamData[ J ].param_ready and
                       not P.ParamData[ J ].param_error then
                      CalcSingleFormula( P, J );
                  end;
              fpos := L;
              break;
            end;
          end;
        end;
      end;
      //------------------------------------------------------------------------
      //     
      TRY
        //----------------------------------------------------------------------
        //  ,      sum( ... ),
        //    ,   :
        fpos := 1;
        while fpos < Length( F ) do
        begin
          if (F[ fpos ] in [ 's', 'S' ]) and
             (CompareText( Copy( F, fpos, 4 ), 'sum(' ) = 0) then
          begin
            J := fpos; SumLevel := 0;
            while J < Length( F ) do
            begin
              CASE F[ J ] OF
              '(': Inc( SumLevel );
              ')': begin
                     Dec( SumLevel );
                     if SumLevel = 0 then break;
                   end;
              END;
              Inc( J );
            end;
            P.DoCacheSum( Idx, fpos, F,
              CalcExpression( Copy( F, fpos, J + 1 - fpos ),
                              P, ParamsList, AllowSum ) );
            if P.SumCached( Idx, fpos, F ) then
              fpos := fpos - 1;
          end;
          Inc( fpos );
        end;

        //----------------------------------------------------------------------
        //    
        P.Values[ Idx ] := CalcExpression( F, P, ParamsList, AllowSum );
        precis := PD.precision;
        if precis < MAX_PRECISION then
        begin
          X := P.Values[ Idx ];
          X := Round( Power( 10, precis ) * X )
                             /
                             Power( 10, precis );
          P.Values[ Idx ] := X;
        end;
        PD.param_ready := TRUE;
        //PD.param_OK := TRUE;
      EXCEPT on E: Exception do
             begin
               if (opname <> '') and (PD.used_in_op.IndexOf( opname ) < 0) or
                  (nomenklid <> 0) and (P.ParamSetData.nomenklid <> nomenklid) or
                  P.ParamSetData.ignore_conditions then
               begin
                 Error := '';
                 E.Message := '';
                 PD.param_ready := TRUE;
                 Exit; //      -   
               end;
               if (E is EFormula_Condition) and (P.Descriptions[ Idx ] <> '') then
               begin
                 if (P.ParamSetData.nomenklid <> nomenklid) and
                    (nomenklid <> 0) then
                 begin
                   Error := '';
                   E.Message := '';
                   PD.param_ready := TRUE;
                   Exit;
                 end;
                 Error := P.Descriptions[ Idx ];
                 E.Message := Error;
                 PD.param_error := TRUE;
               end
               else
                 Error := E.Message;
               raise;
             end;
      END;
    FINALLY
      PD.param_started_calc := FALSE;
      if PD.save_formula <> '' then
      begin
        PD.formula := TrimLeft( PD.save_formula );
        PD.save_formula := '';
      end;
    END;
  end;

var OrderList: TList; //  :    TParams  
                      // ParamsList -      
    I, J, K, N, MinDep: Integer;
    Added: Boolean;
    P: TParams;
    PD: TParamData;
    ParamOrderList: TList; //   - (
                           //   TParams   ParamsList)

    FirstError: String;
    TimeStart: DWORD;
begin
  {$IFDEF DEBUG_SUM}
  DeleteFile( 'C:\sum.log' );
  {$ENDIF}
  Result := 0;
  Error := '';
  OrderList := TList.Create;
  TRY
    TimeStart := GetTickCount;
    //--------------------------------------------------------------------------
    //      
    for I := 0 to ParamsList.Count-1 do
    begin
      P := ParamsList[ I ];
      for K := 0 to P.Count-1 do
      begin
        PD := P.ParamData[ K ];
        if (PD.param_type = 2) or
           (PD.param_type = 3) and (InitInputVals or not PD.param_ready) and
           (P.Values[ K ] = 0) and
           (not PD.formula_empty) and not PD.read_only then
        begin
          //P.ParamData[ K ].param_ready := FALSE;
          //P.ParamData[ K ].param_OK := FALSE;
          P.ParamData[ K ].param_error := FALSE;
        end;
      end;
    end;
    for I := 0 to ParamsList.Count-1 do
    begin
      //------------------------------------------------------------------------
      //         
      //    OrderList   .
      ParamOrderList := TList.Create;
      OrderList.Add( ParamOrderList );
      P := ParamsList[ I ];
      if (nomenklid <> 0) and (P.ParamSetData.nomenklid <> nomenklid) then
        continue;
      if Ignore_con_id0 and (P.ParamSetData.con_id = 0) then continue;
      P.BuildDependings;
      //------------------------------------------------------------------------
      //       P: TParams   
      //     : Param1 < Param2,  Param2 
      //  Param1.
      for J := 0 to P.Count-1 do
      begin
        Added := FALSE;
        MinDep := 0;
        for K := 0 to ParamOrderList.Count-1 do
        begin
          if ParamDependOnParam( P, Integer( ParamOrderList[ K ] ), J ) then
          begin
            Added := TRUE;
            ParamOrderList.Insert( K, Pointer( J ) );
            break;
          end;
          if ParamDependOnParam( P, J, Integer( ParamOrderList[ K ] ) ) then
            MinDep := K;
        end;
        if not Added then
        begin //----------------------------------------------------------------
              //   ,      J.
              // :   J   
          if MinDep >= ParamOrderList.Count then
            ParamOrderList.Add( Pointer( J ) )
          else
            ParamOrderList.Insert( MinDep+1, Pointer( J ) );
        end;
        K := ParamOrderList.IndexOf( Pointer( J ) );
        //----------------------------------------------------------------------
        //   ,        K+1,
        //      .     ,
        //          , 
        //   .
        PD := P.ParamData[ J ];
        for N := K+1 to ParamOrderList.Count-1 do
          if (PD.save_formula = '') and
             ParamDependOnParam( P, J, Integer( ParamOrderList[ N ] ) ) and
             (P.Values[ Integer( ParamOrderList[ N ] ) ] = 0) then
          begin
            Error := ' ' + P.Names[ J ] + '  ' +
              P.Names[ Integer( ParamOrderList[ N ] ) ] + '   ' +
              '    .';
            if (opname <> '') and (PD.used_in_op.IndexOf( opname ) < 0) or
               (nomenklid <> 0) and (P.ParamSetData.nomenklid <> nomenklid) then
            else
              raise Exception.Create( Error );
          end;
      end;
    end;
    TimePrepare := TimePrepare + GetTickCount - TimeStart;
    TimeStart := GetTickCount;
    //--------------------------------------------------------------------------
    //          .
    FirstError := '';
    TRY
      for I := 0 to ParamsList.Count-1 do
      begin
        P := ParamsList[ I ];
        if (nomenklid <> 0) and (P.ParamSetData.nomenklid <> nomenklid) then
          continue;
        if Ignore_con_id0 and (P.ParamSetData.con_id = 0) then continue;
        ParamOrderList := OrderList[ I ];
        for J := 0 to ParamOrderList.Count-1 do
        begin
          K := Integer( ParamOrderList[ J ] );
          PD := P.ParamData[ K ];
          if (PD.param_type = 2) or
             (PD.param_type = 3) then
            if not PD.param_ready and not PD.param_error then
            TRY
              CalcSingleFormula( P, K );
            EXCEPT on E: Exception do
                   begin
                     if (opname <> '') and (PD.used_in_op.IndexOf( opname ) < 0) or
                        (nomenklid <> 0) and (P.ParamSetData.nomenklid <> nomenklid) or
                        P.ParamSetData.ignore_conditions then
                     begin
                       //      -   
                       Error := '';
                       E.Message := '';
                       PD.param_ready := TRUE;
                     end
                     else
                     begin
                       PD.param_error := TRUE;
                       FirstError := Error;
                       Error := '';
                     end;
                   end;
            END;
        end;
      end;
      if FirstError <> '' then
      begin
        Error := FirstError;
        raise Exception.Create( FirstError );
      end;
    EXCEPT on E: Exception do
           begin
             E.Message := Error;
             //    , Error    
             if RaiseError then raise;
           end;
    END;
    TimeCalc := GetTickCount - TimeStart;
  FINALLY
    for I := 0 to OrderList.Count-1 do
      TObject( OrderList[ I ] ).Free;
    OrderList.Free;
  END;
end;

{ TParams }

procedure TParams.Add(const Name: String; Value: Extended; ID: Integer;
  const Description: String);
begin
  inherited;
  FParamData.Add( TParamData.Create );
end;

procedure TParams.Assign( From: TVars );
var F: TParams;
    I: Integer;
    PD: TParamData;
begin
  if From <> Self then
    Clear;         
  inherited;
  if From = Self then
  begin
    for I := 0 to Count-1 do
    begin
      PD := ParamData[ I ];
      PD.Assign( PD );
    end;
    Exit;
  end;
  if not( From is TParams ) then Exit;
  F := From as TParams;
  ParamSetData.Assign( F.ParamSetData );
  for I := 0 to F.Count-1 do
  begin
    PD := TParamData.Create;
    FParamData.Add( PD );
    PD.Assign( F.ParamData[ I ] );
  end;
end;

procedure TParams.BuildDependings;
var I, J, L: Integer;
    F, N: String;
    fpos: Integer;
begin
  for I := 0 to Count-1 do
    ParamData[ I ].FDepending.Clear;
  for I := 0 to Count-1 do
  begin
    if (ParamData[ I ].param_type = 2) or
       (ParamData[ I ].param_type = 3) then
    begin
      fpos := 1;
      F := RemoveSpaces( ParamData[ I ].formula );
      J := pos( ';', F );
      if J > 0 then
        F := Trim( Copy( F, 1, J-1 ) );
      while fpos <= Length( F ) do
      begin
        if (F[ fpos ] in [ ' ','0'..'9','+','-','<','>','=','*','/','<','>','=','(',')', #1..#31, DecimalSeparator ]) then
          Inc( fpos )
        else
        if (F[ fpos ] in [ 's', 'S' ]) and (CompareStr( Copy( F, fpos, 4 ), 'str(') = 0) then
          Inc( fpos, 4 )
        else
        begin
          for L := fpos+1 to Length( F )+1 do
            if (L > Length( F )) or (F[ L ] in [ ' ', '-', '+', '*', '/', '<','>','=', '(', ')' ]) then
            begin
              N := Copy( F, fpos, L - fpos );
              for J := 0 to Count-1 do
              begin
                if N = Names[ J ] then
                  ParamData[ I ].FDepending.Add( Pointer( J ) );
              end;
              fpos := L;
              break;
            end;
        end;
      end;
    end;
  end;
end;

procedure TParams.Clear;
//------------------------------------------------------------------------------
// .   .
var I: Integer;
begin
  inherited;
  if Assigned( FParamData ) then
  begin
    for I := 0 to FParamData.Count-1 do
      TObject( FParamData[ I ] ).Free;
    FParamData.Clear;
  end;
end;

constructor TParams.Create;
//------------------------------------------------------------------------------
// .      .
begin
  inherited;
  FParamData := TList.Create;
  FParamSetData := TParamSetData.Create;
end;

constructor TParams.CreateOwned(Owner: TParamsSet);
begin
  Create;
  FParamsController := Owner;
end;

destructor TParams.Destroy;
//------------------------------------------------------------------------------
// .  .
begin
  Clear;
  FreeAndNil( FParamData );
  FreeAndNil( FParamSetData );
  inherited;
end;

procedure TParams.DoCacheSum(Idx, fPos: Integer; const F: String; Value: Double);
//------------------------------------------------------------------------------
//      
begin
  if FParamsController <> nil then
    FParamsController.DoSumCache( Self, Idx, fPos, F, Value );
end;

function TParams.FormattedParameter(Idx: Variant): String;
//------------------------------------------------------------------------------
//  ,     
//  ,    (  Param 
// Operation_params,      ).
var I: Integer;
    pf: Double;
    fmt: String;
begin
  if VarType( Idx ) = varString then
    I := IndexOf( Idx )
  else
    I := Idx;
  if ParamData[ I ].param_format = 0 then
    Result := Format( '%f', [ Values[ I ] ] )
  else
  begin
    pf := ParamData[ I ].param_format;
    fmt := Format( '%f', [ pf ] );
    fmt := StringReplace( fmt, DecimalSeparator, '.', [ ] );
    while (Length( fmt ) > 0) and (fmt[ Length( fmt ) ] = '0') do
      System.Delete( fmt, Length( fmt ), 1 );
    TRY
      Result := Trim( Format( '%' + fmt + 'f', [ Values[ I ] ] ) );
    EXCEPT
      Result := Format( '%f', [ Values[ I ] ] );
    END;
  end;
end;

function TParams.GetParamData(Idx: Variant): TParamData;
//------------------------------------------------------------------------------
//      
var I: Integer;
begin
  if VarType( Idx ) = varString then
    I := IndexOf( Idx )
  else
    I := Idx;
  Result := FParamData[ I ];
end;

function TParams.OK: Boolean;
//         
var I: Integer;
    PD: TParamData;
begin
  for I := 0 to Count-1 do
  begin
    PD := ParamData[ I ];
    if (PD.param_type > 1) and not PD.param_ready then
    begin
      Result := FALSE;
      Exit;
    end;
  end;
  Result := TRUE;
end;

function TParams.SumCached(Idx, fPos: Integer; var F: String): Boolean;
//------------------------------------------------------------------------------
// ,       F, 
//   fPos:
begin
  Result := FALSE;
  if FParamsController = nil then Exit;
  Result := FParamsController.SumCached( Self, Idx, fPos, F );
end;

procedure TParams.SwapItems(i1, i2: Integer);
//------------------------------------------------------------------------------
//   
var TmpD: Pointer;
begin
  inherited;
  TmpD := FParamData[ i1 ];
  FParamData[ i1 ] := FParamData[ i2 ];
  FParamData[ i2 ] := TmpD;
end;

{$IFDEF DEBUG_SUM}
procedure LogFileOutPut( const Filename, Text: String );
var F: TFileStream;
    Flg: DWORD;
begin
  Flg := fmOpenReadWrite;
  if not FileExists( Filename ) then
    Flg := Flg or fmCreate;
  F := TFileStream.Create( Filename, Flg );
  F.Position := F.Size;
  F.Write( (Text + #10#13)[ 1 ], Length( Text ) + 2 );
  F.Free;
end;
{$ENDIF}

procedure TParams.WhenSummed(ToWhat: TVars; const AddText: String);
begin
  {$IFDEF DEBUG_SUM}
  LogFileOutPut( 'c:\sum.log', IntToStr( Integer( Self ) ) +
                 ' batchid=' + IntToStr( ParamSetData.batchid ) +
                 ' added to ' +
                 IntToStr( Integer( ToWhat ) ) +
                 ' batchid=' + IntToStr( (ToWhat as TParams).ParamSetData.batchid ) +
                 ' ' + AddText );
  {$ENDIF}
end;

{ TParamData }

procedure TParamData.Assign( From: TParamData);
var I: Integer;
begin
  if Self <> From then
  begin
    Units       := From.Units;
    param_type  := From.param_type;
    param_format:= From.param_format;
    precision   := From.precision;
    formula     := From.formula;
    read_only   := From.read_only;
    used_in_op.Assign( From.used_in_op );
    //param_orders.Assign( From.param_order );
    param_orders.Clear;
    for I := 0 to From.param_orders.Count-1 do
      param_orders.Add( From.param_orders[ I ] );
    show_in_op.Assign( From.show_in_op );
  end;
end;

constructor TParamData.Create;
begin
  inherited Create;
  used_in_op := TStringList.Create;
  param_orders := TList.Create;
  show_in_op := TStringList.Create;
  FDepending := TList.Create;
end;

destructor TParamData.Destroy;
begin
  used_in_op.Free;
  param_orders.Free;
  show_in_op.Free;
  FDepending.Free;
  inherited;
end;

function TParamData.formula_empty: Boolean;
var I: Integer;
begin
  I := pos( ';', formula );
  if I > 0 then Result := Trim( Copy( formula, 1, I -1 ) ) = ''
  else
    Result := Trim( formula ) = '';
end;

{ TParamSetData }

procedure TParamSetData.Assign(From: TParamSetData);
begin
  nomenklid := From.nomenklid;
  con_id    := From.con_id;
  numcon    := From.numcon;
  batchid   := From.batchid;
  opid      := From.opid;
  code_operation := From.Code_Operation;
end;

{ TParamsSet }

procedure TParamsSet.DoSumCache(P: TParams; Idx, fPos: Integer;
  const F: String; Value: Double);
begin
  //
end;

function TParamsSet.SumCached(P: TParams; Idx, fPos: Integer; var F: String): Boolean;
begin
  Result := FALSE;
end;

end.
