//**  ,     VCL Delphi.
//**  Vladimir Kladov, 2000-2010
//**
//**      .
//**     /     
//**  2          
//**   GNU,  Free Software Foundation.
//**
//**       ,    
//** ,       ,   
//**         
//**   .      
//**     GNU.
//**
//**          
//**   GNU.     ,     Free
//** Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
//** MA 02111-1307, USA.
//******************************************************************************

{************************************************}
{                                                }
{     FastListView              }
{   " /             }
{   -"                              }
{   Copyright (c) 2000-2003 by Vladimir Kladov   }
{                                                }
{           19.02.2010         }
{                                                }
{************************************************}
unit FastListview;
{*     TListView,   
    :
   |<br> -      ,   
           ,     ()
             ;
   |<br> -    ,      
         ;
   |<br> -    (    ,
           );
   |<br> -        (
           ,    ,   
                );
   |<br> -         
           (      
            );
   |<br> -        
          /  (    
         OnGetColors);
   |<br> -       ,   
         (   ,     fieldName 
         );
   |<br> -       ( 
           ) ,      
         , .  AutoAllocateColumnNames.
   |<br> -        
         ( TQuery, TADOQuery)        
         FastListView ( Assign).       
                    
           Variant,      
         TFastListItem.Values[ ] (      
              );
   |<br> -          
           Excel (   Microsoft Office, \
           Office 97, Office 2000, Office XP);
   |<br> -        (SelectAll -
          , InvertSelection -  ; NextSelected -
                ;
         DeleteSelection -    );
   |<br> -    TListView,     ""
         list view, ..       ,
              .
   |<br> -  TFastListView     TListView
           ,      TListView   
            ,      
         "" listview    .   
                TListView 
          TFastListView   :    
         TFastListView,  Import,     
           ,   TFastListView    
           TListView.      
              (  
                  
         ).
}
{/ This component is a descendant from TListView, but it has a lot of additional
   capabilities:
   |<br> - Sort by columns by multiple criteria, including sorting by several
         columns at a time and sorting only certain items range;
   |<br> - Filter by certain columns including filtering by several columns at
         a time;
   |<br> - Hidden columns (programmatically or by request from the user, using
         context menu);
   |<br> - Change column order transparently for a code: columns are preserving
         its index known at design time, son there is no necessity to take its
         change in a code);
   |<br> - Multi-level header for a table with fixed columns order;
   |<br> - Programming change of colors and fonts for certain items (or/and
         cells, using an event OnGetColors);
   |<br> - Access to cells of items by column field names or column titles
         (the second case, if special field names are not assigned to columns);
   |<br> - Storing additional fields in invisible columns including allocating
         those automatically (see AutoAllocateColumnNames property).
   |<br> - It is possible to fill items immediately from items of TFastListView
         instance or from a record of TQuery / TADOQuery, using Assign method.
         When loading data from query, native values are stored together with
         strings and those are available later via property
         TFastListItem.Values[ ];
   |<br> - A method exists to store all visible content to Excel work book
         (all versions of MS Office are supported, including Office 97,
         Office 2000, Office XP);
   |<br> - There are methods to perform some operations fast: SelectAll,
         InvertSelection, NextSelected, DeleteSelection etc.;
   |<br> - The component is based on TListView, but it is implemented already
         as "virtual" list view, i.e. it allows to work fast with huge data
         arrays containing millions of items.
   |<br> - It is implemented as much compatible with TListView as possible,
         so TListView object can be replaced by TFastListView instance without
         necessity to change (almost) any code. There is a special design-time
         operation to perform such replacement: item "Import" in context menu.
         Though some events are not moved automatically (to move it, create
         such event handler for the TFastListView object and move code to its
         body manually).
}

interface

{$I VCL_VERSIONS.INC}
{$IFNDEF DEBUG_NOOPTIMIZE}
{$O+}
{$ENDIF}

uses
    Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
    comctrls, imglist, commctrl, QSort, ColorListView, menus, Extctrls, registry,
    ComObj, db, dbtables, ADOdb, AdoSqlUnit, FLVProgress, math,
    {$IFDEF _D6orHigher}designintf, Variants{$ELSE}dsgnintf{$ENDIF};

const
    WM_USER_KILLFILTERLISTFORM = CM_BASE + 99;
    WM_USER_FILTERLISTDBLCLK = CM_BASE + 98;
    WM_USER_CALLFIRSTFILTER = CM_BASE + 97;
    WM_USER_SETFOCUS = CM_BASE + 96;
    WM_FREE_NIL_EDITCELLCTL = WM_USER + 2002;

    SaveToExcelViaClipboard: Boolean = TRUE;
var DisableFrameAroundFocusedListView: Boolean = TRUE;

type
    FLV_Exception = class(Exception)
    end;
    TEditKind = (ekEdit, ekEditNum, ekEditFloat, ekEditCurrency, ekEditDate,
        ekEditPeriod, ekCombo, ekCheck);

    PVariant = ^Variant;

    TSortByIcon = (sbiNone, sbiClickOnIcon, sbiAlways);

    TOldListItem = comctrls.TListItem;
    TOldListView = comctrls.TListView;

    TFastListview = class;

    TFastListItems = class;

    TFastListItem = class;
    TListItem = TFastListItem;

    TOnGetColors = procedure(Sender: TObject; var ForeColor, BackColor: TColor;
        var FontStyle: TFontStyles; Item: TFastListItem;
        SubItemIdx: integer) of object;

    TLVDeletedEvent = procedure(Sender: TObject; Item: TFastListItem) of object;
    TLVOnCustomDrawSubItem = procedure(Sender: TCustomListView;
        Item: TListItem; SubItem: Integer; State: TCustomDrawState;
        var DefaultDraw: Boolean) of object;
    TLVAdvancedCustomDrawsubItem = procedure(
        Sender: TCustomListView; Item: TListItem; SubItem: Integer;
        State: TCustomDrawState; Stage: TCustomDrawStage;
        var DefaultDraw: Boolean) of object;

    TFastListSubItems = class(TStringList)
    private
        FOwnerItem: TFastListItem;
        function GetByIndexOrName(IndexOrName: Variant): string;
        procedure PutByIndexOrName(IndexOrName: Variant; const Value: string);
    protected
        procedure Put(Index: Integer; const S: string); override;
        procedure UpdateOwner;
    public
        constructor Create(AOwner: TFastListItem);
        procedure Insert(Index: Integer; const S: string); override;
        function Add(const S: string): Integer; override;
        procedure Clear; override;
        procedure Delete(Index: Integer); override;
        property Strings[IndexOrName: Variant]: string read GetByIndexOrName
        write PutByIndexOrName; default;
    end;

    TFilterStrings = class(TStringList)
        destructor Destroy; override;
    end;

    TFastListItem = class
        {*         . 
          , - (), , .     -
              TListItem,    .
        }
        {/ An object to store data of each item in fast list view. It stores Caption,
           sub-items, icon(s), state. And a pointer to actual TListItem if it is
           already created. }
    private
        FIndex: Integer;
        FOwnerFastListview: TFastListView;
        FOwnerItemsHolder: TFastListItems;
        FCaption: String;
        FListItem: TOldListItem;
        FSubItems: TFastListSubItems;
        FImageIndex: Integer;
        FOverlayIndex: Integer;
        FStateIndex: Integer;
        FData: Pointer;
        FSelected: Boolean;
        FIndent: Integer;
        fBackColor: TColor;
        fBackColorSelected: TColor;
        fVisible: Boolean;
        FFontColor: TColor;
        FFontStyle: TFontStyles;
        FCustomFontStyle: Boolean;
        FFontColorSelected: TColor;
        FEditingColIdx: Integer;
        procedure SetCaption(const Value: String);
        function GetCaption: String;
        procedure SetSubitemsList(const Value: TStrings);
        function GetIndexInListView: Integer;
        procedure Setselected(const Value: Boolean);
        function GetImageIndex: Integer;
        procedure SetImageIndex(const Value: Integer);
        function Get_Focused: Boolean;
        procedure Set_Focused(const Value: Boolean);
        function GetCut: Boolean;
        procedure SetCut(const Value: Boolean);
        function GetDropTarget: Boolean;
        procedure SetDropTarget(const Value: Boolean);
        function GetSubitemsList: TStrings;
        function GetOverlayIndex: Integer;
        procedure SetOverlayIndex(const Value: Integer);
        function GetStateIndex: Integer;
        procedure SetStateIndex(const Value: Integer);
        function GetIndent: Integer;
        procedure SetIndent(const Value: Integer);
        procedure SetBackColor(const Value: TColor);
        procedure SetBackColorSelected(const Value: TColor);
        procedure SetVisible(const Value: Boolean);
        procedure SetFontColor(const Value: TColor);
        procedure SetFontColorSelected(const Value: TColor);
        procedure SetFontStyle(const Value: TFontStyles);
        procedure SetCustomFontStyle(const Value: Boolean);
        function GetFontStyle: TFontStyles;
        function GetIsChecked: Boolean;
        procedure SetIsChecked(const Value: Boolean);
        function GetSubItems(IdxOrName: Variant): String;
        procedure SetSubItems(IdxOrName: Variant; const Value: String);
        function GetValues(IdxOrName: Variant): Variant;
        procedure SetValues(IdxOrName: Variant; const Value: Variant);
        function GetData: Pointer;
        function GetSelected: Boolean;
    protected
        FValues: TList;
        procedure DoEndEditCell(Sender: TObject);
    public
        constructor Create(AOwner: TFastListItems);
        destructor Destroy; override;
        property Caption: String read GetCaption write SetCaption;
        property SubItemsList: TStrings read GetSubitemsList write SetSubitemsList;
        property SubItems: TStrings read GetSubitemsList write SetSubitemsList;

        property Index: Integer read GetIndexInListView;
        property Selected: Boolean read GetSelected write Setselected;
        property ImageIndex: Integer read GetImageIndex write SetImageIndex;
        property OverlayIndex: Integer read GetOverlayIndex write SetOverlayIndex;
        property StateIndex: Integer read GetStateIndex write SetStateIndex;
        property Data: Pointer read GetData write FData;
        property Focused: Boolean read Get_Focused write Set_Focused;
        procedure MakeVisible(PartialOK: Boolean);
        function MirrorItem: TOldListItem;
        property Cut: Boolean read GetCut write SetCut;
        property DropTarget: Boolean read GetDropTarget write SetDropTarget;
        function GetRect(SubItemIndex: Integer = -1): TRect;
        { SubItemIndex = -1 -  Item, 0 - Caption, 1 - SubItems[ 0 ]... }
        property Indent: Integer read GetIndent write SetIndent;
        property BackColor: TColor read fBackColor write SetBackColor;
        property BackColorSelected: TColor read fBackColorSelected write SetBackColorSelected;
        property FontColor: TColor read FFontColor write SetFontColor;
        property FontColorSelected: TColor read FFontColorSelected write SetFontColorSelected;
        property Visible: Boolean read fVisible write SetVisible;
        function ListView: TFastListView;
        function EditCaption: Boolean;
        function EditCell(Idx: Integer): Boolean;
        {* Starts editing cell defined. OnBeginEditCell not called, editing is started
           always even if a column can not be edited. Idx is -1 - based column index,
           or 0-based SubItem index. }
        procedure EndEditCell(Cancel: Boolean = FALSE);
        {* Stops editing cell, currently being editing (if any). OnEndEdit is called
           if assigned, for owner fast list view control. }
        property FontStyle: TFontStyles read GetFontStyle write SetFontStyle;
        property CustomFontStyle: Boolean read FCustomFontStyle write SetCustomFontStyle;
        property Checked: Boolean read GetIsChecked write SetIsChecked;

        property SItems[IdxOrName: Variant]: String read GetSubItems write SetSubItems;
        DEFAULT;
        {*    ()  fastlistview   ,
            ,   
           (-1 - Caption, 0..n - SubItems).   
            DEFAULT,      
                 SItems,  
                 . }
        {/ Access of subitems either by index (-1 means Caption, 0..n
           mean subitems) or by column's field names (if not assigned for a
           column, by column title). }
        property Values[IdxOrName: Variant]: Variant read GetValues write SetValues;
        {*    ()  fastlistview (  
                SItems[])   
             .     
                TQuery  TADOQuery,
           .  Assign . }
        {/ Access to values associated with fast list view cells, also
           either by index of by filed name (or column title). Those
           values are given automatically when assigning an item from
           TQuery or TADOQuery current record (see Assign methods). }
        procedure Assign(FromItem: TFastListItem; const AlterAssignments: String = '');
            overload;
        {*      FastListView  .
            AlterAssignments  ,  
                 
           ,  
           __1=__1;...
            ,     ,  
           .    list view   AutoAllocateColumns,
                 .
        }
        {/ Assigning an item from one FastListView item to another
           (instances of TFastListView could be different). AlterAssignments
           parameter allows to define which columns of the source should
           be assigned to which target columns, in a form
           Output_column_name1=Input_column_name;...
           |<br> Fields, which are not listed are assigned as is. If a
           property AutoAllocateColumnNames is FALSE for target
           TFastListView, and desired column is not found, an exception
           is generated. }
        procedure Assign(FromQuery: TQuery; const AlterAssignments: String = '');
            overload;
        {*       
            TQuery.  ,      
           AlterAssignments,     ,   
             ( ,    ,
             list view  AutoAllocateColumns=FALSE). 
            AlterAssignments  ,   Assign   -
            list view - . }
        {/ Assignment of the entire item from given TQuery current record.
           For those fields which names are not present in AlterAssignments,
           same named columns are used (or having same named field name
           assigned, if any). It is possible that an exception will be
           generated if such column is not exist and the property
           AutoAllocateColumnNames is FALSE for target TFastListView
           object. }
        procedure Assign(FromADOQuery: TADOQuery; const AlterAssignments: String = '');
            overload;
        {*  Assign   TQuery,   TADOQuery. }
        {/ Like Assign for TQuery but for TADOQuery. }

        procedure EndEdit;
        {*   ,   
           () . }
    end;

    TFastListItems = class
        {* }
        {/ }
    private
        FOwner: TFastListview;
        F_Items: TList;
        FDestroying: Boolean;
        FCount: Integer;
        FTmpItem1, FTmpItem2: TFastListItem;
        FSaveSelList: TList;
        function Get_Items(Idx: Integer): TFastListItem;
        function GetItemsCount: Integer;
        procedure SetItemsCount(const Value: Integer);
    protected
        AccessingOnData: Boolean;
        Clearing: Boolean;
    public
        constructor Create;
        destructor Destroy; override;
        property Items[Idx: Integer]: TFastListItem read Get_Items; default;
        procedure Clear;
        function Add: TFastListItem;
        function Insert(Index: Integer): TFastListItem;
        procedure BeginUpdate;
        procedure EndUpdate;
        property Count: Integer read GetItemsCount write SetItemsCount;
        procedure Delete(Index: Integer);
    end;

    TFastLVCompareEvent = procedure(Sender: TObject; Item1, Item2: TFastListItem;
        var Compare: Integer) of object;
    {*    ,   . }
    {/ An event of comparing two items, used while sorting. }

    TLVSelectItemEvent = procedure(Sender: TObject; Item: TFastListItem;
        Selected: Boolean) of object;
    {*  /  . }
    {/ An event to set or reset selection. }

    TFastLVDrawItemEvent = procedure(Sender: TCustomListView; Item: TFastListItem;
        Rect: TRect; State: TOwnerDrawState) of object;
    {*   . }
    {/ An event of drawing item. }

    TRangeSelect = procedure(Sender: TObject; IdxFrom, IdxTo: Integer;
        Selected: Boolean) of object;
    {*   . }
    {/ An event of range selection. }

    TSelectItemEvent = procedure(Sender: TObject; Item: TListItem;
        Selected: Boolean) of object;
    {*    . }
    {/ An event of item selection. }

    TFLVEditingEvent = procedure(Sender: TObject; Item: TFastListItem;
        var AllowEdit: Boolean) of object;
    {*     . }
    {/ An event of start editing item Caption. }

    TFLVEditedEvent = procedure(Sender: TObject; Item: TFastListItem;
        var S: string) of object;
    {*     . }
    {/ An event of end editing item Caption. }

    TOnMessage = function(Sender: TObject; var Msg: TMessage): Boolean of object;
    {*     . }
    {/ An event of any message handling. }

    TOnSelecting = function(Sender: TObject; Item: TFastListItem): Boolean of object;
    {*  ,      (
        FALSE   ). }
    {/ An event sect item, allows to prevent selection of certain items if
       FALSE is returned. }

    TFastLVOwnerDataEvent = procedure(Sender: TObject; Item: TFastListItem) of object;
    {*        TListView,  
       . }
    {/ An event of getting data in request of TListView, used internally. }

    TNotifyAccept = procedure(Sender: TObject; var Cancel: Boolean) of object;
    {*          (
            ). }
    {/ An event type to notify about start of sorting of filtering (allows to
       do something in such case ). }

    {$IFDEF ENGLISH}
    TColumnSortAttr = (
        saImageIndex, // Sort first by image index
        saStateImage, // Then by static image index
        saCheckedFirst, // First are items having checked=TRUE
        saImgDescending, // Sort by image index in image index descending order
        saLenghtFirst, // First are items having shorter text in the column
        saNumber, // When possible, sequences of digits are sorted as integer numbers
        saCase, // Case sensitive (aBC > ABC)
        saDate, // Dates in format DD.MM.YYYY or DD.MM.YY are compared as dates
        saMonthYear, // Dates in format MM.YYYY or MM.YY
        saMonthName, // Month names in Russian and in English ('January' < 'Febraury')
        saMonthShortName, // The same but short month names: Jan, Feb, Mar, Apr, Mat, ...
        saDayWeekName, // Day week names: (Wednesday < Saturday)
        saDayWeekShortName, // Short wek day names: Mn, Tu, Wd ... and Mon, Teu, Wed, ...
        saIgnoreSpaces, // Ignore internal and external spaces.
        saDescending, // Descending sorting
        saEmptyFirst, // Empty strings are first (if there are no saEmptyFirst,
        // such strings are at the end).
        saNotRusLatEq); // Do not account Russian letters similar to so-shaped
    // English one, the same (default those are).
{/ Sort column attributes. Can be set at design time or programmatically
for certain columns using TColumnSortAttrs.
|<pre>
saImageIndex,       // Sort first by image index
saStateImage,       // Then by static image index
saCheckedFirst,     // First are items having checked=TRUE
saImgDescending,    // Sort by image index in image index descending order
saLenghtFirst,      // First are items having shorter text in the column
saNumber,           // When possible, sequences of digits are sorted as integer numbers
saCase,             // Case sensitive (aBC > ABC)
saDate,             // Dates in format DD.MM.YYYY or DD.MM.YY are compared as dates
saMonthYear,        // Dates in format MM.YYYY or MM.YY
saMonthName,        // Month names in Russian and in English ('January' < 'Febraury')
saMonthShortName,   // The same but short month names: Jan, Feb, Mar, Apr, Mat, ...
saDayWeekName,      // Day week names: (Wednesday < Saturday)
saDayWeekShortName, // Short wek day names: Mn, Tu, Wd ... and Mon, Teu, Wed, ...
saIgnoreSpaces,     // Ignore internal and external spaces.
saDescending,       // Descending sorting
saEmptyFirst,       // Empty strings are first (if there are no saEmptyFirst,
    // such strings are at the end).
saNotRusLatEq );    // Do not account Russian letters similar to so-shaped
    // English one, the same (default those are).
</pre>
}
    {$ELSE RUSSIAN}
    TColumnSortAttr = (
        saImageIndex, //     
        saStateImage, //     
        saCheckedFirst, //   ,    checked=TRUE
        saImgDescending, //    -   ,
        saLenghtFirst, //    ,      
        saNumber, //        ('C9' < 'C10')
        saCase, //   (aBC > ABC)
        saDate, //     DD.MM.YYYY  DD.MM.YY   ,
        //          
        saMonthYear, //     MM.YYYY  MM.YY
        saMonthName, //    -  - ('' < '')
        saMonthShortName, //  ,    : , , , , , ...
        saDayWeekName, //    : ( < )
        saDayWeekShortName, //     : , , ...  , , , ...
        saIgnoreSpaces, //       
        saDescending, //   
        saEmptyFirst, //          
        //    ( saEmptyFirst  , 
        //     ).
        saNotRusLatEq); //      ,   , 
    {*   .      
       .
    |<pre>
      saImageIndex,       //     
      saStateImage,       //     
      saCheckedFirst,     //   ,    checked=TRUE
      saImgDescending,    //    -   ,
      saLenghtFirst,      //    ,      
      saNumber,           //        ('C9' < 'C10')
      saCase,             //   (aBC > ABC)
      saDate,             //     DD.MM.YYYY  DD.MM.YY   ,
                          //          
      saMonthYear,        //     MM.YYYY  MM.YY
      saMonthName,        //    -  - ('' < '')
      saMonthShortName,   //  ,    : , , , , , ...
      saDayWeekName,      //    : ( < )
      saDayWeekShortName, //     : , , ...  , , , ...
      saIgnoreSpaces,     //       
      saDescending,       //   
      saEmptyFirst,       //          
                          //    ( saEmptyFirst  , 
                          //     ).
      saNotRusLatEq );    //      ,   , 
    |</pre>
       }
    {$ENDIF}

    TColumnSortAttrs = Set of TColumnSortAttr;
    {*   .   ,  
        .  }
    {/ A set of sort column attributes. Used in aproprty of a column controlling
       sorting of a certain column. }

    PColumnSortAttrs = ^TColumnSortAttrs;

    TColAdditionalAttrs = packed record
        {*   .    
           ,       . }
        {/ Additional column attributes. This structure is used internally to store
           additional column attributes in a form resource. }
        SortAttrs: TColumnSortAttrs;
        Filtered: Boolean;
        VisualOrder: Integer;
        Visible: Boolean;
    end;

    TOnSwapItems = procedure(Sender: TFastListview; e1, e2: Integer)
        of object;

    TOnBeginEditCell = procedure(Sender: TFastListView; Item: TFastListItem;
        SubItem: Integer; var CanBeginEdit: Boolean; var EditKind: TEditKind)
        of object;

    TOnEndEditCell = procedure(Sender: TFastListView; Item: TFastListItem;
        SubItem: Integer; UniEdit: TControl; var S: String) of object;

    TOnCreateEditCell = procedure(Sender: TFastListView; Item: TFastListItem;
        SubItem: Integer; UniEdit: TControl) of object;

    TOnLineToExcel = function(Sender: TFastListView; Item: TFastListItem;
        WB, WS: Variant; xLeft, xRight: Integer; Progress: TFLV_fmProgress = nil): String of object;

    TOnDataCell = procedure(Sender: TFastListview; ItemIdx, ColIdx: Integer;
        var Txt: String) of object;
    TOnDataCellData = procedure(Sender: TFastListview; ItemIdx: Integer;
        var Data: Pointer) of object;
    TOnDataCellImage = procedure(Sender: TFastListview; ItemIdx: Integer;
        var Image_Index: Integer) of object;
    TOnDataCellNamed = procedure(Sender: TFastListview; ItemIdx: Integer;
        const ColName: String; var Txt: String) of Object;
    TOnEndOfSaveToExcel = procedure(Sender: TFastListview; WS: Variant)
        of object;

    TFastListview = class(TColorListView)
        {*!  TFastListView    
           (   )    . }
        {/! A component TFastListView is designed to work fast (mainly to view)
            with huge tables of data. }
    private
        F_ItemsList: TFastListItems;
        F_OnCompare: TFastLVCompareEvent;
        F_OnColClick: TLVColumnClickEvent;
        FSortByIcons: TSortByIcon;
        FSortByNumber: Boolean;
        FSortByCase: Boolean;
        fOnSelectRange: TRangeSelect;
        FSortFrom: Integer;
        FSortTo: Integer;
        fOnDraw: TFastLVDrawItemEvent;
        fOnSelectItem: TSelectItemEvent;
        FPreviouslySelectedIdx: Integer;
        F_OwnerData: Boolean;
        FSortByDate: Boolean;
        FOnDeletion: TLVDeletedEvent;
        FOnCustomDrawSubItem: TLVOnCustomDrawSubItem;
        FOnAdvancedCustomDrawSubItem: TLVAdvancedCustomDrawsubItem;
        fOnGetColors: TOnGetColors;
        fOndividerDblClk: TNotifyHeaderDividerDblClk;
        FOnHdrPaint: TNotifyHeaderPaint;
        fOnKillFocus: TNotifyEvent;
        FHiddenItems: TList;
        fOnFilterChange: TNotifyEvent;
        fOnMessage: TOnMessage;
        fOnSelecting: TOnSelecting;
        FLVOwnerDataEvent: TFastLVOwnerDataEvent;
        fOnBeforeSort: TNotifyAccept;
        fOnAfterSort: TNotifyEvent;
        fOnBeforeFilter: TNotifyAccept;
        FAllowColumnDrag: Boolean;
        FRegistrySaveRestore: Boolean;
        fOnEdited: TFLVEditedEvent;
        fOnEditing: TFLVEditingEvent;
        FAllowColumnHide: Boolean;
        FAutoAllocateColumnNames: Boolean;
        FUniSourcesCount: Integer;
        FOnEndEditCell: TOnEndEditCell;
        FOnBeginEditCell: TOnBeginEditCell;
        FOnCreateEditCell: TOnCreateEditCell;
        FOnLineToExcel: TOnLineToExcel;
        FVirtualTotally: Boolean;
        FOnDataCell: TOnDataCell;
        FOnSwapItems: TOnSwapItems;
        FOnDataCellData: TOnDataCellData;
        FOnDataCellImage: TOnDataCellImage;
        FOnDataCellNamed: TOnDataCellNamed;
        FAutoSelectOnEnter: Boolean;
        FGridLines1: Boolean;
        FNoMultiColumnSort: Boolean;
        FOnEndOfSaveToExcel: TOnEndOfSaveToExcel;
        F_MultiSelect: Boolean;
        function GetInheritedItemsCount: Integer;
        function GetOldItems(Idx: Integer): TOldListItem;
        procedure SetOldItems(Idx: Integer; const Value: TOldListItem);
        procedure SetSortByIcons(const Value: TSortByIcon);
        procedure SetSortByNumber(const Value: Boolean);
        procedure SetSortByCase(const Value: Boolean);
        function GetSelected: TListItem;
        procedure SetSelected(const Value: TListItem);
        procedure SetSortFrom(const Value: Integer);
        procedure SetSortTo(const Value: Integer);
        procedure Set_OnDraw(const Value: TFastLVDrawItemEvent);
        procedure DoDrawItem(Sender: TCustomListView; Item: TOldListItem;
            ARect: TRect; State: TOwnerDrawState);
        function GetFocusedItem: TListItem;
        procedure SetFocusedItem(const Value: TListItem);
        procedure DoOnSelectItem(Sender: TObject; Item: TOldListItem; ASelected: Boolean);
        procedure SetSortByDate(const Value: Boolean);
        procedure SetOnCustomDrawSubItem(const Value: TLVOnCustomDrawSubItem);
        procedure DoCustomDrawSubitem(Sender: TCustomListView;
            Item: TOldListItem; SubItem: Integer; State: TCustomDrawState;
            var DefaultDraw: Boolean);
        procedure SetOnAdvancedCustomDrawSubItem(
            const Value: TLVAdvancedCustomDrawsubItem);
        procedure DoAdvancedDrawSubItem(Sender: TCustomListView; Item: TOldListItem; SubItem: Integer;
            State: TCustomDrawState; Stage: TCustomDrawStage;
            var DefaultDraw: Boolean);
        function GetHiddenItems(Idx: Integer): TFastListItem;
        function GetHiddenItemsCount: Integer;
        function GetItemData(Idx: Integer): Pointer;
        procedure setItemData(Idx: Integer; const Value: Pointer);
        procedure SetColumnVisible(colIdx: Integer; const Value: Boolean);
        procedure SetAllowColumnDrag(const Value: Boolean);
        function GetIsColumnFiltered(colIdx: Integer): Boolean;
        function GetVisibleColumns(Idx: Integer): TListColumn;
        function GetVisibleColumnsCount: Integer;
        procedure SetMultiHeader(const Value: String);
        function GetColumnSortAttrs(ColIdx: Integer): TColumnSortAttrs;
        procedure SetIsColumnFiltered(colIdx: Integer; const Value: Boolean);
        procedure SetAllowColumnHide(const Value: Boolean);
        function GetColFNames(Idx: Integer): String;
        procedure SetColFNames(Idx: Integer; const Value: String);
        procedure SetAutoAllocateColumnNames(const Value: Boolean);
        function GetUniSources(Idx: Integer): TComponent;
        procedure SetUniSourcesCount(const Value: Integer);
        function GetUniSourcesCount: Integer;
        function GetColComboItems(Idx: Integer): TStrings;
        function GetColKinds(Idx: Integer): TEditKind;
        function GetColumnCanEdit(Idx: Integer): Boolean;
        procedure SetColKinds(Idx: Integer; const Value: TEditKind);
        procedure SetColumnCanEdit(Idx: Integer; const Value: Boolean);
        procedure SetColumnComboItems(Idx: Integer; const Value: TStrings);
        function GetFilterValues: TFilterStrings;
        procedure SetFilterValues(const Value: TFilterStrings);
        procedure SetVirtualTotally(const Value: Boolean);
        procedure SetOnDataCell(const Value: TOnDataCell);
        procedure SetOnSwapItems(const Value: TOnSwapItems);
        procedure SetOnDataCellData(const Value: TOnDataCellData);
        procedure SetOnDataCellImage(const Value: TOnDataCellImage);
        procedure SetOnDataCellNamed(const Value: TOnDataCellNamed);
        procedure SetHdrHeight(const Value: Integer);
        function GetHdrHeight: Integer;
        function GetMultiHeader: String;
        procedure SetAutoSelectOnEnter(const Value: Boolean);
        procedure SetGridLines1(const Value: Boolean);
        procedure SetMultiColumnSort(const Value: Boolean);
        procedure SetOnEndOfSaveToExcel(const Value: TOnEndOfSaveToExcel);
        function Get_ReadOnly: Boolean;
        procedure Set_ReadOnly(const Value: Boolean);
        procedure Set_MultiSelect(const Value: Boolean);
    private
        FFakeHdrFont: TFont;
        procedure Do_OnEditing(Sender: TObject; Item: TOldListItem; var AllowEdit: Boolean);
        procedure Do_OnEdited(Sender: TObject; Item: TOldListItem; var S: string);
        function RecalcHeaderHeight: Integer;
    protected
        function GetSubItemRect(ItemIdx, ColIdx: Integer): TRect; override;
    protected
        FColumnclicked: Integer;
        FIconclicked: Boolean;
        FColumnsToSort: TList;
        FColumnsToFilter: TStringList;
        FSaveFocusedIdx: Integer;
        FTriangleBmp, FTrianglePressedBmp, FTriangleFilteredBmp: TBitmap;
        FFilterButtonClicked: Integer;
        FColumnSortAttrs: TList;
        FColPopup: TPopupMenu;
        miSortOneColumn, miAddcolumn, miRemoveColumn, miIncreaseOrder, miDecreaseOrder,
            miAscending, miDescending: TMenuItem;
        miFilterMenu, miFilter, miResetFilter, miResetAllFilters: TMenuItem;
        miHideMenu, miHideshowSep1, miHideShowSep2: TMenuItem;
        miHideColumn, miShowAllColumns: TMenuItem;
        miRestoreOrder, miSizeProportional: TMenuItem;
        FHiddenColumns: TStringList;
        FOriginalHitTestHeader: Boolean;
        FDisabledItemChanging: Boolean;
        FFocusWasSet: Boolean;
        FParentFormName: String;
        FEndDragTime: DWORD;
        FDisableAutoExpand: Boolean;
        FOrderByFollowsColumns: Boolean;
        IsClosing: Boolean;
        FKillingListForm: Boolean;
        procedure miSortOneColumnClick(Sender: TObject);
        procedure miAddColumnClick(Sender: TObject);
        procedure miRemoveColumnClick(Sender: TObject);
        procedure miIncreaseOrderClick(Sender: TObject);
        procedure miDecreaseOrderClick(Sender: TObject);
        procedure miAscendingClick(Sender: TObject);
        procedure miDescendingClick(Sender: TObject);
        procedure miFilterClick(Sender: TObject);
        procedure miResetFilterClick(Sender: TObject);
        procedure miResetAllFiltersClick(Sender: TObject);
        procedure miHideColumnClick(Sender: TObject);
        procedure miShowAllColumnsClick(Sender: TObject);
        procedure miShow1Columnclick(Sender: TObject);
        procedure miRestoreOrderClick(Sender: TObject);
        procedure miSizeProportionalClick(Sender: TObject);
        procedure ColPopupOnPopup(Sender: TObject);
        procedure ClearInheritedItems;
        procedure Get_Data(Sender: TObject; Item: TOldListItem);
        procedure SetCount(C: Integer);
        function Providecount(C: Integer): Boolean;
        property InheritedItems[Idx: Integer]: TOldListItem read GetOldItems write SetOldItems;
        procedure CompareItemsFast(Sender: TObject; Item1,
            Item2: TFastListItem; var Compare: Integer);
        procedure ColumnClickFast(Sender: TObject; Column: TListColumn);
        procedure RangeSelected(var Msg: TWMNotify); message CN_NOTIFY;
        procedure SaveSelection;
        procedure RestoreSelection;
        function AlphaSort1(allcolumns: Boolean): Boolean;
        procedure doHdrDivider(Sender: TObject; columnIdx: Integer);
        procedure doHdrPaint(Sender: TObject; Canvas: TCanvas; R: TRect; ColIdx: Integer);
        function HdrMarginLeft(ColIdx: Integer): Integer; override;
        function HdrMarginRight(ColIdx: Integer): Integer; override;
        procedure HdrHitTest(var HTI: THDHitTestInfo); override;
        function ProcessWMSetCursor(Wnd: HWnd; lParam: Integer): Boolean; override;
        procedure LoadTriangleBmp;
        function FilterButtonClick(IgnoreCursorPos: Boolean; ItemIdx: Integer = -1): Boolean; override;
        function OnFilterButton: Boolean; override;
        procedure DoFilterButtonClick(ColIdx: Integer; const Pt1, Pt2: TPoint);
        procedure KillFilterListform(Sender: TObject);
        procedure WMKillFilterListForm(var M: TMessage); message WM_USER_KILLFILTERLISTFORM;
        procedure FilterListKeyPress(Sender: TObject; var Key: Char);
        procedure FilterListDblClk(Sender: TObject);
        procedure WMFilterListDblClk(var M: TMessage); message WM_USER_FILTERLISTDBLCLK;
        procedure ApplyFilters;
        procedure ApplyAllFilters(CanExtend: Boolean);
    protected
        FHdrTrkIdx: Integer;
        procedure WndProc(var M: TMessage); override;
        function FilterListSelecting(Sender: TObject; Item: TFastListItem): Boolean;
        function ItemIndent(Item: TOldListItem): Integer; override;
        function HdrRightClick(ColIdx: Integer; Down: Boolean): Boolean; override;
        procedure CreateWnd; override;
        procedure DoAutoExpand(ExcludeColumn: Integer); override;
        procedure CustDrawItem(
            Sender: TCustomListView;
            Item: TOldListItem;
            State: TCustomDrawState;
            var DefaultDraw: Boolean); override;
        procedure CustDraw(Rect: TRect;
            Item: TListItem;
            Capt: string;
            SubItemIdx: Integer;
            Stat: TCustomDrawState;
            Alig: TAlignment);
        procedure InvalidateSelected;
        function EnabledAutoexpand: Boolean; override;
        procedure Loaded; override;
        function ParentForm: TForm;
        function GetColumnVisible(colIdx: Integer): Boolean; override;
        procedure PaintHdr(W: Hwnd; DC: HDC); override;
        procedure WMHScroll(var M: TMessage); message WM_HSCROLL;
        procedure WMVScroll(var M: TMessage); message WM_VSCROLL;
        procedure CMEnter(var Message: TCMEnter); message CM_ENTER;
        procedure CMLeave(var Message: TCMExit); message CM_EXIT;
    protected
        FColumnsEditor_DesignTime: TObject;
        FColumnToLoadStoreAdditional: Integer;
        procedure DefineProperties(Filer: TFiler); override;
        procedure LoadColumnAdditional(Stream: TStream);
        procedure StoreColumnAdditional(Stream: TStream);
        procedure NilFFilterListForm;
    public
        constructor Create(AOwner: TComponent); override;
        destructor Destroy; override;

        procedure RestoreSettingsFromRegistry;
        {*    .     
               (    
           ).      
            RegistryKeyPath (      ,
             initialization - ,  , 
                -   ).
           |<br> .    RegistrySaveRestore. }
        {/ Restores settings from the Registry. This method is called automatically
           when an obkect is created (but it can certainly be called programmatically
           if necessary). To work correctly it is necessary to assign a value to
           global variable RegistryKeyPath (once only, in initialization of
           the main unit).
           |<br> See also property RegistrySaveRestore. }

        procedure SaveSettingsInRegistry;
        {*    ,     
             RegistryKeyPath, . .   
               TFastListView,     
             (     ).
            , ,  .
           |<br> .    RegistrySaveRestore. }
        {/ Saves settings in the registry, the global variable RegistryKeyPath is
           used as a base path. This method is called automatically when the  object
           is destroyed, but it can also be called programmatically when necessary.
           Orders, visibility and width of columns are stored.
           |<br> See also method RestoreSettingsFromRegistry. }

        procedure Clear; {$IFDEF _D6orHigher} override; {$ENDIF}
        {*  .   ,  Items.Clear (
            , .. , ). }
        {/ Clear content. The same as Items.Clear (it is recommended still
           use the second, original variant). }

        function GetItemAt(X, Y: Integer): TFastListItem;
        {*     . X, Y    
              TFastListView, (0,0)  
                fast list view. }
        {/ Returns an item at the given coordinates (client originated). }

    protected
        procedure QuickSort(L, R: Integer);
        procedure QuickSort2(L, R: Integer);
    public
        function AlphaSort: Boolean;
        {*         
            ( ,     , 
                 ).  }
        {/ Sort items in correspondance with currently assigned criteria and
           sorting event handlers assigned (for items and subitems not having special
           sorting attributes assigned default ones are used). }

        function Sort: Boolean;
        {*    ,  . }

        function CompareItems(Item1, Item2: TFastListItem): Integer;
        {*    fast list view   , 
             . }

        property ColumnsToSort: TList read FColumnsToSort;
        {*    .     $80000000,
                  .
                 
            CompareItems,      
             . }

        procedure SortColumns(const ColumnsArray: array of Integer;
            OrderByFollowsColumns: Boolean = TRUE);
        {*         .
              ,    
           ,         
            ( ,     
              ). }
        {/ This method allows to sort columns taking into attention only (mostly)
           given columns. additional parameter allows to define if to take into
           account columns not listed in ColumnsArray (if not, an original item
           indwex is used to determine which items to place first). }

        procedure SetSortColumnAttributes(const ColumnIndeces: array of Integer;
            Attrs: TColumnSortAttrs);
        {*     .  
              (0   Caption, 1 - SubItems[0],
           ...)    ,    
              .     
              ,     .
                    ,
                  , 
               .
           |<br> .    ColumnSortAttrs. }
        {/ Programmatical setting columns sort attributes. In ColumnIndeces array,
           0 means Caption, 1 - subitem[ 0 ], etc. }

        property ColumnSortAttrs[ColIdx: Integer]: TColumnSortAttrs read GetColumnSortAttrs;
        {*    .     
           SetSortColumnAttributes. }
        {/ Sorting attributes for a given column. }

        property SortFrom: Integer read FSortFrom write SetSortFrom;
        {*       , 
           .   0,     . .
             SortTo . }
        {/ This propery allows to set up the first item in range to be sorted.
           If a range is defined with SortFrom and SortTo, only such items are
           sorted between each others. }

        property SortTo: Integer read FSortTo write SetSortTo;
        {*       , 
           .     SortFrom   
               ,    . }
        {/ This property allows to set up upper index of a range of items to be
           sorted.
           If a range is defined with SortFrom and SortTo, only such items are
           sorted between each others. }

        procedure FilterColumns(ColumnsList: array of Integer);
        {*      ,  
               .   
                    
                  . 
               ,  
            .
           |<br>
                   
           , ..       
           ,  . .    ColumnFiltered
           . }
        {/ This method allows to define a set of columns available for filtering
           from user context menu. At design-time this attribute can be set separately
           for each column. If columns are created dynamically, use this method or
           property ColumnFiltered. }

        property ColumnFiltered[colIdx: Integer]: Boolean read GetIsColumnFiltered
        write SetIsColumnFiltered;
        {*      . .   
           FilterColumns . }
        {/ If a column is filtered. See also method FilterColumns. }

        function IndexOfColumn(const ColName: String): Integer;

        property FilterValues: TFilterStrings read GetFilterValues write SetFilterValues;
        {* }

        procedure FilterItems;
        {*          
                  
                  (
             ). }
        {/ This method can be called to filter fast list view in correspondance
           with current user preferences and allows to refresh object state after
           making some changes in its cells (programmatically or by user). }

        procedure ClearFilters;
        {*  ()    . 
               . }
        {/ Reset all assigned filters. List of columns to filter is not changed. }

        property ItemFocused: TListItem read GetFocusedItem write SetFocusedItem;
        {* /     .  
                 . }
        {/ Currently focused item. When an item is set into focus, it becomes
           visible. }

        property ItemData[Idx: Integer]: Pointer read GetItemData write setItemData;
        {*     Data  .  , 
            Item.Data (..      ). }
        {/ Fast acces to item's data. Works faster than Item.Data. }

        function FindData(StartIndex: Integer; Value: Pointer;
            Inclusive, Wrap: Boolean): TListItem;
        {*      Item.Data. }
        {/ Fast search of an item by Item.Data. }

        function GetNextItem(StartItem: TListItem;
            Direction: TSearchDirection; States: TItemStates): TListItem;
        {*   (  )    
           TItemStates.        
            NextSelected,   .  }
        {/ Search next available (if any) item in a given direction with ceratin
           attributes TItemStates. To search next selected item (forward) a method
           NextSelected can be used. }

        function NextSelected(LI: TFastListItem): TFastListItem;
        {*      . :
           ! LI := LV.Selected;
           ! while LI <> nil do
           ! begin
           !   //    () LI
           !   LI := LV.NextSelected( LI );
           ! end;
        }
        {/ Allows to list fast selected items. E.g.:
           ! LI := LV.Selected;
           ! while LI <> nil do
           ! begin
           !   // working with an item LI here ...
           !   LI := LV.NextSelected( LI );
           ! end;
        }

        property Selected: TListItem read GetSelected write SetSelected;
        {* /  .   MultiSelect=TRUE,
                Selected    ,
                .     
               nil  Selected.
           |<br>    Selected  ItemFocused,    
            ,    ,    -.
               - ,
                (  ). }
        {/ Selected item. If MultiSelect=TRUE, assigning another item as selected
           does not cancel current selection but just selects ones more item. It
           is possible to reset selection assigning nil to this property.
           |<br> Note: do not confuce with ItemFocused which could be any item not
           only selected, but always single. }

        procedure SelectAll; {$IFDEF _D6orHigher} override; {$ENDIF}
        {*    ( )   
            fast list view (  , ,  
               ). }
        {/ Selects all the items in the fast list view, fast. Items, which are
           filtered out, are not selected. }

        procedure InvertSelection;
        {*     .  ( 
             )        ). }
        {/ Inverts current selection, fast. Items which are filtered out, are not
           selected anyway. }

        procedure DeleteSelection;
        {*       (). }
        {/ Delets currently selected items, fast. }

    published
        property OnSelecting: TOnSelecting read fOnSelecting write fOnSelecting;
        {* ,        
             (  ).  
             FALSE,    . }
        {/ An event which is called when an item is to be selected or deselected.
           To prevent selection of the item, an event handler can return FALSE. }

    public
        property HiddenItems[Idx: Integer]: TFastListItem read GetHiddenItems;
        {*   ( : ) .   
               Items,   
                -   
                  
               .   ,  
              -        
              , ,     
              ).
           |<br> .   HiddenItemsCount, DeleteHiddenItems. }
        {/ A list of hidden (filtered out) items. Such items are not available for a
           code by the usual way, via property Items. }

        property HiddenItemsCount: Integer read GetHiddenItemsCount;
        {*   . . HiddenItems. }
        {/ Number of hidden items. }

        procedure DeleteHiddenItems;
        {*    . }
        {/ Deletes hidden items. }

        function HasEmptyCaptions: Boolean;
        {*    TRUE,       
           Caption=''. }
        {/ TRUE if at least a single item exists with a caption equal to '' (empty). }

        property ColumnVisible[colIdx: Integer]: Boolean read GetColumnVisible write SetColumnVisible;
        {*    . }
        {/ Visibility of a column. }

        property VisibleColumnsCount: Integer read GetVisibleColumnsCount;
        {*   . }
        {/ Visible columns count.}

        property VisibleColumns[Idx: Integer]: TListColumn read GetVisibleColumns;
        {*   .      
           . }
        {/ An array of visible columns. Can be used to handle visible columns only. }

    public
        Excel_Progress: TFLV_fmProgress;
        procedure SaveToExcel(const Filename: String; const ImageNames: String = '';
            const StateImageNames: String = ''; FromIdx: Integer = 0; ToIdx: Integer = MaxInt;
            SelectedOnly: Boolean = FALSE);
        {*   FastListview    " Excel". 
           ImageNames  StateImageNames    , 
               (  ).
              ,    .
                    .
        }
        {/ Stores visible TFastListview content to the Excel worksheet file.
           ImageNames and StateImageNames are strings containing lists of
           icon descriptions (separated with ',') stored in first column(s)
           of the Excel book. }
        procedure SaveToWorksheet(Excel, WB, WS: Variant; TopLine: Integer = 1;
            const ImageNames: String = '';
            const StateImageNames: String = ''; FromIdx: Integer = 0; ToIdx: Integer = MaxInt;
            SelectedOnly: Boolean = FALSE; Progress: TFLV_fmProgress = nil);
        procedure SaveToWorksheetFast(Excel, WB, WS: Variant; TopLine: Integer = 1;
            const ImageNames: String = '';
            const StateImageNames: String = ''; FromIdx: Integer = 0; ToIdx: Integer = MaxInt;
            SelectedOnly: Boolean = FALSE; Progress: TFLV_fmProgress = nil);
        {*        Excel. }
    published
        property OnLineToExcel: TOnLineToExcel read FOnLineToExcel
            write FOnLineToExcel;
        property OnEndOfSaveToExcel: TOnEndOfSaveToExcel read FOnEndOfSaveToExcel write SetOnEndOfSaveToExcel;
    public
        function ContentAsText(const ColumnSeparator: String = ''#9;
            const LineSeparator: String = ''#13#10): String;
        {*   FastListView     ( 
           , ,    ,     . }
        {/ Retrieving TFastListview visible content in a form of text string
           (later it can be saved to atext file, for instance). }

    protected
        FUniSources: TList;
        property UniSourcesCount: Integer read GetUniSourcesCount write SetUniSourcesCount;
        property UniSources[Idx: Integer]: TComponent read GetUniSources;
    public
        procedure AddUniSource(UniSource: TComponent);
        procedure RemoveUniSource(UniSource: TComponent);
    published

        property BorderStyle;
        property Columns;
        property ColumnClick;
        property ReadOnly: Boolean read Get_ReadOnly write Set_ReadOnly;
        property HideSelection;
        property IconOptions;
        property AllocBy;
        property HoverTime;
        property LargeImages;
        property MultiSelect: Boolean read F_MultiSelect write Set_MultiSelect;
        property OwnerData: Boolean read F_OwnerData write F_OwnerData;
        property OwnerDraw;
        property OnAdvancedCustomDraw;
        property OnAdvancedCustomDrawItem;
        property OnAdvancedCustomDrawSubItem: TLVAdvancedCustomDrawsubItem read FOnAdvancedCustomDrawSubItem write SetOnAdvancedCustomDrawSubItem;
        property OnChange;
        property OnChanging;
        property OnClick;
        property OnColumnClick: TLVColumnClickEvent read F_OnColClick write F_OnColClick;
        property OnColumnDragged;
        property OnColumnRightClick;
        property OnCompare: TFastLVCompareEvent read F_OnCompare write F_OnCompare;
        property OnContextPopup;
        property OnCustomDraw;
        property OnCustomDrawItem;
        property OnCustomDrawSubItem: TLVOnCustomDrawSubItem read FOnCustomDrawSubItem write SetOnCustomDrawSubItem;
        property OnData: TFastLVOwnerDataEvent read FLVOwnerDataEvent write FLVOwnerDataEvent;
        property OnDataFind;
        property OnDataStateChange;
        property OnDblClick;
        property OnDeletion: TLVDeletedEvent read FOnDeletion write FOnDeletion;
        property OnDragDrop;
        property OnDragOver;
        property OnDrawItem: TFastLVDrawItemEvent read fOnDraw write Set_OnDraw;
        property OnEdited: TFLVEditedEvent read fOnEdited write fOnEdited;
        property OnEditing: TFLVEditingEvent read fOnEditing write fOnEditing;
        property OnEndDock;
        property OnEndDrag;
        property OnEnter;
        property OnExit;
        property OnInfoTip;
        property OnInsert;
        property OnGetImageIndex;
        property OnGetSubItemImage;
        property OnKeyDown;
        property OnKeyPress;
        property OnKeyUp;
        property OnMouseDown;
        property OnMouseMove;
        property OnMouseUp;
        property OnResize;
        property OnSelectItem: TSelectItemEvent read fOnSelectItem write fOnSelectItem;
        property OnSelectRange: TRangeSelect read fOnSelectRange write fOnSelectRange;
        property OnStartDock;
        property OnStartDrag;
        property ShowColumnHeaders;
        property ShowWorkAreas;
        property SmallImages;
        property SortType;
        property StateImages;
        property ViewStyle;
        property Align;
        property Anchors;
        property BiDiMode;
        property BorderWidth;
        property Checkboxes;
        property Color;
        property Constraints;
        property DragCursor;
        property DragKind;
        property DragMode;
        property Enabled;
        property FlatScrollBars;
        property Font;
        property FullDrag;
        property GridLines: Boolean read FGridLines1 write SetGridLines1
            default TRUE;
        property HotTrack;
        property HotTrackStyles;
        property ParentBiDiMode;
        property ParentColor;
        property ParentFont;
        property ParentShowHint;
        property PopupMenu;
        property RowSelect;
        property ShowHint;
        property TabOrder;
        property TabStop;
        property Visible;

        property Items: TFastListItems read F_ItemsList;
        property Count: Integer read GetInheritedItemsCount;
        property SortByIcons: TSortByIcon read FSortByIcons write SetSortByIcons;
        property SortByNumber: Boolean read FSortByNumber write SetSortByNumber;
        property SortByCase: Boolean read FSortByCase write SetSortByCase;
        property SortByDate: Boolean read FSortByDate write SetSortByDate;

        procedure GetColumnLeftRight(NCol: Integer; var L, R: Integer);
        property OnGetColors: TOnGetColors read fOnGetColors write fOnGetColors;

        property OnHdrDividerDblClk: TNotifyHeaderDividerDblClk read fOndividerDblClk write fOndividerDblClk;
        property OnHdrPaint: TNotifyHeaderPaint read FOnHdrPaint write FOnHdrPaint;
        property MultiHeader: String read GetMultiHeader write SetMultiHeader;

        property OnKillFocus: TNotifyEvent read fOnKillFocus write fOnKillFocus;
        property OnMessage: TOnMessage read fOnMessage write fOnMessage;
        {*       , 
             TFastListView,     
           . }
        {/ This event allows to filter out all the messages retrieved by the
           TFastListView control, and to modify its handling. }

        property OnBeforeSort: TNotifyAccept read fOnBeforeSort write fOnBeforeSort;
        {*        
            . }
        {/ This event allows to do something at the beginning of strings sorting. }

        property OnAfterSort: TNotifyEvent read fOnAfterSort write fOnAfterSort;
        {*        
            . }
        {/ This event allows to do something at the end of strings sorting. }

        property OnBeforeFilter: TNotifyAccept read fOnBeforeFilter write fOnBeforeFilter;
        {*         
           ( ). .   OnFilterChange . }
        {/ This event allows to do something when filter is to be changed. See also
           OnFilterChange. }

        property OnFilterChange: TNotifyEvent read fOnFilterChange write fOnFilterChange;
        {*         
           ( ). .   OnBeforeFilter . }
        {/ This event allows to do something when filter is changed (after change
           it). See also OnBeforeFilterChange event. }

        property AllowColumnDrag: Boolean read FAllowColumnDrag write SetAllowColumnDrag;
        {*        
                (   FALSE).
                ,   
           .         
           (MultiHeader). }
        {/ This event allows to disable columns order change by the user (by
           dragging its title by mouse). Such disabling can be useful in sume
           cases. Also it is always on (AllowColumnDrag=FALSE), when MultiHeader=TRUE. }

        property AllowColumnHide: Boolean read FAllowColumnHide write SetAllowColumnHide;
        {*        
              .}
        {/ This property allows to disable hiding separate columns in the TFastListView
           object by user (from the context menu). }

        property RegistrySaveRestore: Boolean read FRegistrySaveRestore write FRegistrySaveRestore;
        {*       
             (, , ) / , 
            ,    RegistryKeyPath. }
        {/ This property allows to disable saving and restoring column attributes
           (order, visibility, width) to/from the registry, independently either
           there is RegistryKeyPath variable is assigned. }
        property RegistrySaveResore: Boolean read FRegistrySaveRestore write FRegistrySaveRestore;

    protected
        FColumnNames: TStringList;
        FColumnCanEdit: TList;
        FColumnEditKind: TList;
        FColumnComboItems: TList; // of TStrings;
        FEditCellCtl: TControl; // TUniEdit-    
        FEditingItem: TFastListItem;
        FEditing: Boolean;
        procedure LoadColumnFieldNames(Stream: TStream);
        procedure StoreColumnFieldNames(Stream: TStream);
        procedure LoadColumnCanEdit(Stream: TStream);
        procedure StoreColumnCanEdit(Stream: TStream);
        procedure LoadColumnEditKind(Stream: TStream);
        procedure StoreColumnEditKind(Stream: TStream);
        procedure LoadColumnComboItems(Stream: TStream);
        procedure StoreColumnComboItems(Stream: TStream);
        procedure WMMouseDown(var M: TMessage); message WM_LBUTTONDOWN;
        procedure EditCellKeyPress(Sender: TObject; var Key: Char);
        procedure WMKeyDown(var M: TMessage); message WM_KEYDOWN;
        procedure WMScroll(var M: TMessage); message WM_SYSCOMMAND;
    public

        function ColumnIndexByName(const S: String): Integer;
        {*  0..Columns.Count   ,  .
                    
            (  ,       ),
           . TFastListItem.SItems. .   ColumnFieldNames.  }
        {/ Returns 0..Columns.Count by field name correspondent to a column (or
           by column title, if a field name is not assigned to the column). }

        property ColumnFieldNames[Idx: Integer]: String
        read GetColFNames write SetColFNames;
        {*  ,  . .   ColumnIndexByName . }
        {/ Fileld names correspondent to columns. See also ColumnIndexByName above.}

        property ColumnCanEdit[Idx: Integer]: Boolean read GetColumnCanEdit
        write SetColumnCanEdit;
        {* TRUE,     (  FALSE).  ,
                  , 
                -   OnBeginEditCell. }

        function CanEditByCells: Boolean;
        {*  TRUE,        . }

        function EditingCell: Boolean;
        {*  TRUE,       . }

        procedure EndEditCell;
        {*       . }

        property ColumnEditKind[Idx: Integer]: TEditKind read GetColKinds
        write SetColKinds;
        {*  ,    ,  .
                OnBeginEditCell -   . }

        property ColumnComboItems[Idx: Integer]: TStrings read GetColComboItems
        write SetColumnComboItems;
        {*  ,    , 
             .      
               OnCreateEditCell,    
            . }

    published
        property OnBeginEditCell: TOnBeginEditCell read FOnBeginEditCell write FOnBeginEditCell;
        property OnEndEditCell: TOnEndEditCell read FOnEndEditCell write FOnEndEditCell;
        property OnCreateEditCell: TOnCreateEditCell read FOnCreateEditCell write FOnCreateEditCell;

    published
        property AutoAllocateColumnNames: Boolean read FAutoAllocateColumnNames write SetAutoAllocateColumnNames;
        {*     TRUE (  FALSE)   
               ,   (,  
           ,   TQuery  TADOQuery    ),   
             fast list view  .     
             ,    TRUE. }
        {/ Assigning TRUE to this property allows to not care about adding all the
           (invisible) columns what are necessary (e.g. to store data, obtained from
           TQuery or TADOQuery current record via Assign method). New columns are
           added when necessary if this property is set to TRUE. }
        property VirtualTotally: Boolean read FVirtualTotally write SetVirtualTotally;
        property OnDataCell: TOnDataCell read FOnDataCell write SetOnDataCell;
        property OnDataCellData: TOnDataCellData read FOnDataCellData write SetOnDataCellData;
        property OnDataCellImage: TOnDataCellImage read FOnDataCellImage write SetOnDataCellImage;
        property OnDataCellNamed: TOnDataCellNamed read FOnDataCellNamed write SetOnDataCellNamed;
        property OnSwapItems: TOnSwapItems read FOnSwapItems write SetOnSwapItems;
    protected
        FSortLineIdexes: array of Integer;
        FSelectedItemIdx: Integer;
        procedure Do_SwapItems(n, m: Integer);
    public
        property ColumnClicked: Integer read FColumnclicked;
        property EditCellControl: TControl read FEditCellCtl;
    protected
        //procedure WMNotify01( var M: TWMNotify ); message WM_NOTIFY;
    published
        property HdrHeight: Integer read GetHdrHeight write SetHdrHeight;
        property AutoSelectOnEnter: Boolean read FAutoSelectOnEnter write SetAutoSelectOnEnter
            default TRUE;
        property NoMultiColumnSort: Boolean read FNoMultiColumnSort write SetMultiColumnSort;
    protected
        procedure Wm_freenileditcellctl(var M: TMsg); message WM_FREE_NIL_EDITCELLCTL;
    end;

    TListView = TFastListView;

procedure TextToClipboard(const s: String);

var
    RegistryKeyPath: String;
    {*  ,      
         fast list view  . ,
       'Software\MyCompany\MyCoolApp\Settings\'.   
            ,    
          .  ,  
        , /    
        .    list view  
         .  }
    {/ A global variable, which defines a base path to store columns attributes
       (of TFastListView items) in the registry. E.g.,
       'Software\MyCompany\MyCoolApp\Settings\'. It is necessary to assign a value
       to this variable only once, in the initialization section of the main unit.
       If this global variable is not initialized, storing and restoring of
       column attributes is not used. }

    {$R *.DCR}

function CompareStringsFast(ByNumber, ByCase, ByDate: Boolean; const S1, S2: String): Integer;

implementation

uses
    fmuFastLVFilter, utRus2Eng, FLVResources, UniEdit, FastTable, clipbrd,
    utPhotoExc, ShellAPI, ActiveX;

var
    FFilterListForm: TfmFilterList;
    Upper: array[Char] of Char;

procedure InitUpper;
var c: Char;
begin
    for c := Low(c) to High(c) do
        Upper[c] := (AnsiUpperCase(c) + ' ')[1];
end;

function IndexOfAbs(List: TList; Value: Integer): Integer;
begin
    Result := List.IndexOf(Pointer(Value));
    if Result < 0 then
        Result := List.IndexOf(Pointer(DWORD(Value) or $80000000));
end;

function CompareFun(const Sender_: Pointer; const e1, e2: DWORD): Integer;
var Item1, Item2: TFastListItem;
    Sender: TFastListview;
begin
    Result := 0;
    if e1 = e2 then Exit;
    Sender := Sender_;
    Item1 := Sender.Items[Integer(e1 + DWORD(Sender.FSortFrom))];
    Item2 := Sender.Items[Integer(e2 + DWORD(Sender.FSortFrom))];
    if Assigned(Sender.OnCompare) then
        Sender.OnCompare(Sender, Item1, Item2, Result);
    if Result = 0 then
        Sender.CompareItemsFast(Sender, Item1, Item2, Result);
    if Result = 0 then
        Result := e1 - e2;
end;

procedure SwapFun(const Sender_: Pointer; const e1, e2: DWORD);
var Sender: TFastListview;
    Tmp: Integer;
begin
    Sender := Sender_;

    Tmp := Sender.FSortLineIdexes[e1 + DWORD(Sender.FSortFrom)];

    Sender.FSortLineIdexes[e1 + DWORD(Sender.FSortFrom)] :=
        Sender.FSortLineIdexes[e2 + DWORD(Sender.FSortFrom)];

    Sender.FSortLineIdexes[e2 + DWORD(Sender.FSortFrom)] := Tmp;

    Sender.Do_SwapItems(e1 + DWORD(Sender.FSortFrom),
        e2 + DWORD(Sender.FSortFrom));
end;

type
    TSortedData = packed record
        ListSorted: TList;
        FLVSorted: TFastListview;
        ColIdx: Integer;
    end;
    PSortedData = ^TSortedData;

function CompareSorted(const Data: Pointer; const e1, e2: Dword): Integer;
var D: PSortedData;
    S1, S2: String;
begin
    D := Data;
    CASE D.ColIdx OF
        0: begin
                S1 := D.FLVSorted.Items[Integer(D.ListSorted.Items[e1])].Caption;
                S2 := D.FLVSorted.Items[Integer(D.ListSorted.Items[e2])].Caption;
            end;
    else
        begin
            TRY
                S1 := D.FLVSorted.Items[Integer(D.ListSorted.Items[e1])].SubItemsList[D.ColIdx - 1];
            EXCEPT
                S1 := '';
            END;
            TRY
                S2 := D.FLVSorted.Items[Integer(D.ListSorted.Items[e2])].SubItemsList[D.ColIdx - 1];
            EXCEPT
                S2 := '';
            END;
        end;
    END;
    Result := AnsiCompareText(S1, S2);
end;

function CompareHiddenSorted(const Data: Pointer; const e1, e2: Dword): Integer;
var D: PSortedData;
    S1, S2: String;
begin
    D := Data;
    CASE D.ColIdx OF
        0: begin
                S1 := D.FLVSorted.HiddenItems[Integer(D.ListSorted.Items[e1])].Caption;
                S2 := D.FLVSorted.HiddenItems[Integer(D.ListSorted.Items[e2])].Caption;
            end;
    else
        begin
            S1 := D.FLVSorted.HiddenItems[Integer(D.ListSorted.Items[e1])].SubItemsList[D.ColIdx - 1];
            S2 := D.FLVSorted.HiddenItems[Integer(D.ListSorted.Items[e2])].SubItemsList[D.ColIdx - 1];
        end;
    END;
    Result := AnsiCompareText(S1, S2);
end;

procedure SwapSorted(const Data: Pointer; const e1, e2: Dword);
var D: PSortedData;
    Ptr: Pointer;
begin
    D := Data;
    Ptr := D.ListSorted.Items[e1];
    D.ListSorted.Items[e1] := D.ListSorted.Items[e2];
    D.ListSorted.Items[e2] := Ptr;
end;

{procedure LogCompare( const S1, S2: String; compare: Integer );
var F: TFileStream;
    s: String;
    Mode: Integer;
begin
  s := 'C:\Compare_log.txt';
  if not FileExists( s ) then Mode := fmCreate
  else Mode := fmOpenReadWrite;
  F := TFileStream.Create( s, Mode );
  TRY
    F.Position := F.Size;
    if compare < 0 then s := S1 + ' < ' + S2
    else
    if compare > 0 then s := S1 + ' > ' + S2
    else
    s := S1 + ' = ' + S2;
    s := s + #13#10;
    F.Write( s[1], Length( s ) );
  FINALLY
    F.Free;
  END;
end;}

procedure TFastListView.QuickSort(L, R: Integer);
var
    I, J, P, cmp: Integer;
    T: Pointer;
    Tmp: Integer;
begin
    repeat
        I := L;
        J := R;
        P := (L + R) shr 1;
        repeat
            OnCompare(Self, Items[I], Items[P], cmp);
            while (cmp < 0) and (I < P) do begin
                Inc(I);
                OnCompare(Self, Items[I], Items[P], cmp);
            end;
            OnCompare(Self, Items[J], Items[P], cmp);
            while (cmp > 0) and (J > P) do begin
                Dec(J);
                OnCompare(Self, Items[J], Items[P], cmp);
            end;
            if I <= J then
            begin
                //ExchangeItems(I, J);
                if VirtualTotally then
                begin
                    //if Assigned( OnSwapItems ) then
                    OnSwapItems(Self, I, J);
                end
                else
                begin
                    T := F_ItemsList.F_Items[I];
                    F_ItemsList.F_Items[I] := F_ItemsList.F_Items[J];
                    F_ItemsList.F_Items[J] := T;
                end;
                begin
                    Tmp := FSortLineIdexes[I];
                    FSortLineIdexes[I] := FSortLineIdexes[J];
                    FSortLineIdexes[J] := Tmp;
                end;
                if P = I then
                    P := J
                else if P = J then
                    P := I;
                Inc(I);
                Dec(J);
            end;
        until I > J;
        if L < J then QuickSort(L, J);
        L := I;
    until I >= R;
end;

{ TFastListview }

function TFastListview.AlphaSort: Boolean;
begin
    Result := AlphaSort1(TRUE);
end;

function TFastListview.AlphaSort1(allcolumns: Boolean): Boolean;
var N: Integer;
    Cancelsort: Boolean;
begin
    Result := FALSE;
    if SortType = stNone then Exit;
    if Assigned(OnBeforeSort) then
    begin
        Cancelsort := FALSE;
        //SendMessage( Getwindow( Handle, GW_CHILD ), WM_LBUTTONUP, 0, 0 );
        OnBeforeSort(Self, Cancelsort);
        if Cancelsort then Exit;
    end;
    if allcolumns then
        FColumnsToSort.Clear;
    if (FSortTo >= 0) then
        N := FSortTo - FSortFrom + 1
    else
        N := Items.Count;
    if N > Items.Count then
        N := Items.Count;
    if N > 1 then
    begin
        SaveSelection;
        if Assigned(OnCompare) then
            QuickSort(0, F_ItemsList.Count - 1)
        else
            SortData(Self, N, CompareFun, SwapFun)
                //QuickSort2( 0, F_ItemsList.Count-1 )
            ;
        RestoreSelection;
    end;
    Result := TRUE;
    if Assigned(OnAfterSort) then
        OnAfterSort(Self);
    if FUpdateCount = 0 then
    begin
        InvalidateRect(Handle, nil, TRUE);
        InvalidateHeader;
    end;
end;

procedure TFastListview.Clear;
begin
    //{$IFDEF _D6orHigher}
    //inherited;
    //{$ELSE}
    Items.BeginUpdate;
    try
        Items.Clear;
    finally
        Items.EndUpdate;
    end;
    //{$ENDIF}
end;
(*var I: Integer;
begin
  F_ItemsList.Clear;
  for I := FHiddenItems.Count-1 downto 0 do
    TObject( FHiddenItems[ I ] ).Free;
  FHiddenItems.Clear;
  {$IFDEF _D6orHigher}
  inherited;
  {$ENDIF}
end;*)

procedure TFastListview.ClearInheritedItems;
begin
    inherited Items.Clear;
end;

procedure TFastListview.ColumnClickFast(Sender: TObject;
    Column: TListColumn);
var P, PScreen: TPoint;
    I, W, N: Integer;
    HdrWnd: HWnd;
    ClassNameBuf: array[0..49] of Char;
    HdHti: THDHitTestInfo;
    //R: TRect;
    iCC: TLVColumnClickEvent;
begin
    iCC := OnColumnClick;
    if Assigned(iCC) then
    begin
        iCC(Self, Column);
        Exit;
    end;
    if SortType = stNone then Exit;
    try

        if (GetKeyState(VK_MENU) < 0) and
            (GetKeyState(VK_CONTROL) >= 0) and
            (GetKeyState(VK_SHIFT) >= 0) then
        begin
            I := IndexOfAbs(FColumnsToSort, Column.Index);
            if I >= 0 then
                FColumnsToSort[I] := Pointer($80000000 xor DWORD(FColumnsToSort[I]))
            else
                if FColumnsToSort.Count = 0 then
                    FColumnsToSort.Add(Pointer($80000000 or DWORD(Column.Index)))
                else
                    FColumnsToSort.Insert(1, Pointer($80000000 or DWORD(Column.Index)));
        end
        else
            if (GetKeyState(VK_CONTROL) < 0)
                and (not NoMultiColumnSort) then
            begin
                N := Column.Index;
                if GetKeyState(VK_MENU) < 0 then
                    N := $80000000 or DWORD(N);
                I := IndexOfAbs(FColumnsToSort, Column.Index);
                if I < 0 then
                    FColumnsToSort.Add(Pointer(N))
                else
                begin
                    FColumnsToSort.Delete(I);
                    if FColumnsToSort.Count = 0 then Exit;
                end;
            end else
                if (GetKeyState(VK_SHIFT) < 0)
                    and not NoMultiColumnSort then
                begin
                    if FColumnsToSort.Count = 0 then
                    begin
                        if FColumnclicked < 0 then
                            FColumnsToSort.Add(nil)
                        else FColumnsToSort.Add(Pointer(FColumnclicked));
                    end else
                    begin
                        if (FColumnsToSort.Count = 1) and
                            ((Integer(FColumnsToSort[0]) and $FFFF) = Column.Index) then
                        begin
                            if GetKeyState(VK_MENU) < 0 then
                                //FColumnsToSort[ 0 ] := Pointer( $80000000 xor DWORD( FColumnsToSort[ 0 ] ) )
                            else Exit; //    
                        end;
                    end;

                    I := IndexOfAbs(FColumnsToSort, Column.Index);
                    if (I = 1) or (I = 0) and (FColumnsToSort.Count = 1) or
                        (GetKeyState(VK_MENU) < 0) then
                        FColumnsToSort[I] := Pointer($80000000 xor DWORD(FColumnsToSort[I]))
                    else if I >= 0 then
                    begin
                        if FColumnsToSort.Count > 0 then
                            FColumnsToSort.Move(I, 1);
                        if GetKeyState(VK_MENU) < 0 then
                            FColumnsToSort[1] := Pointer($80000000 xor DWORD(FColumnsToSort[1]));
                    end else
                    begin
                        FColumnsToSort.Insert(1, Pointer(Column.Index));
                        if GetKeyState(VK_MENU) < 0 then
                            FColumnsToSort[1] := Pointer($80000000 xor DWORD(FColumnsToSort[1]));
                    end;
                    FColumnclicked := Column.Index;
                    FIconclicked := FALSE;
                end else
                begin
                    if (FColumnsToSort.Count = 1) and (IndexOfAbs(FColumnsToSort, Column.Index) = 0) then
                        FColumnsToSort[0] := Pointer($80000000 xor DWORD(FColumnsToSort[0]))
                    else
                    begin
                        FColumnsToSort.Clear;
                        FColumnsToSort.Add(Pointer(Column.Index));
                    end;
                end;

        //if Column.ImageIndex >= 0 then
        begin
            GetCursorPos(PScreen);
            P := PScreen;
            windows.ScreenToClient(Handle, P);
            HdrWnd := childWindowFromPoint(Handle, P);
            if (HdrWnd <> 0) and (HdrWnd <> Handle) then
            begin

                GetClassName(HdrWnd, ClassNameBuf, 50);
                if ClassNameBuf = 'SysHeader32' then
                begin
                    P := PScreen;
                    windows.ScreenToClient(HdrWnd, P);
                    HdHti.Point := P;
                    SendMessage(HdrWnd, HDM_HITTEST, 0, Integer(@HdHti));
                    I := HdHti.Item;
                    if I = Column.Index then
                    begin
                        if Column.ImageIndex >= 0 then
                        begin
                            W := 26;
                            if SmallImages <> nil then
                                W := SmallImages.Width + 10;
                            Dec(PScreen.x, W);
                            P := ScreenToClient(PScreen);
                            if P.x < 0 then
                                FIconclicked := TRUE
                            else
                            begin
                                P := PScreen;
                                windows.ScreenToClient(HdrWnd, P);
                                HdHti.Point := P;
                                SendMessage(HdrWnd, HDM_HITTEST, 0, Integer(@HdHti));
                                I := HdHti.Item;
                                if I <> Column.Index then
                                    FIconclicked := TRUE;
                            end;
                        end;
                    end;
                end;

            end;
        end;
        FColumnclicked := Column.Index;

        if not Assigned(F_OnColClick) then
            if SortType <> stNone then
            begin
                SortType := stBoth;
                FOrderByFollowsColumns := TRUE;
                AlphaSort1(FColumnsToSort.Count = 0);
            end;

        InvalidateHeader;

    except on E: Exception do
        begin
            ShowMessage('ERROR: ' + E.Message);
        end;
    end;

end;

function IsDigit(C: Char): Boolean;
asm
  SUB AL, '0'
  CMP AL, 9
  SETBE AL
end;

function IsDate(const S: String; I: Integer): Boolean;
var Y, J: Integer;
begin
    Result := FALSE;
    if Length(S) < I + 8 then Exit;
    {if IsDigit( S[ I ] ) and IsDigit( S[ I + 1 ] ) and (S[ I + 2 ]='.') and
              IsDigit( S[ I + 3 ] ) and IsDigit( S[ I + 4 ] ) and (S[ I + 5 ]='.') and
              IsDigit( S[ I + 6 ] ) and IsDigit( S[ I + 7 ] )}
    if (S[I] in ['0'..'9']) and (S[I + 1] in ['0'..'9']) and (S[I + 2] = '.') and
        (S[I + 3] in ['0'..'9']) and (S[I + 4] in ['0'..'9']) and (S[I + 5] = '.') and
        (S[I + 6] in ['0'..'9']) and (S[I + 7] in ['0'..'9'])
        then begin
        Result := TRUE;
        if Length(S) = I + 7 then Exit;
        if not IsDigit(S[I + 8]) then Exit;
        Result := FALSE;
        if Length(S) < I + 9 then Exit;
        //if IsDigit( S[ I + 8 ] ) and IsDigit( S[ I + 9 ] ) then
        if (S[I + 8] in ['0'..'9']) and (S[I + 9] in ['0'..'9']) then
        begin
            Y := 0;
            for J := I + 6 to I + 9 do
                Y := Y * 10 + Ord(S[J]) - Ord('0');
            //if StrToInt( Copy( S, I + 6, 4 ) ) <= 1900 then
            if Y <= 1900 then
                Result := FALSE
            else
                Result := TRUE;
        end;
    end;
end;

function IsMonthYear(const S: String; I: Integer): Boolean;
begin
    Result := FALSE;
    if Length(S) < I + 5 then Exit;
    if IsDigit(S[I]) and IsDigit(S[I + 1]) and (S[I + 2] = '.') and
        IsDigit(S[I + 3]) and IsDigit(S[I + 4]) then
    begin
        Result := TRUE;
        if Length(S) = I + 5 then Exit;
        if not IsDigit(S[I + 5]) then Exit;
        Result := FALSE;
        if Length(S) < I + 6 then Exit;
        if IsDigit(S[I + 6]) then
        begin
            TRY
                if StrToInt(Copy(S, I + 3, 4)) <= 1900 then Exit;
                Result := TRUE;
            EXCEPT
                Result := FALSE;
            END;
        end;
    end;
end;

function Dig(C: Char): Integer;
asm
  SUB AL, '0'
end;

function GetDate(const S: String; var I: Integer): TDateTime;
var D, M, Y: Integer;
begin
    D := Dig(S[I]) * 10 + Dig(S[I + 1]);
    M := Dig(S[I + 3]) * 10 + Dig(S[I + 4]);
    Y := Dig(S[I + 6]) * 10 + Dig(S[I + 7]);
    Inc(I, 8);
    if (Length(S) > I) and IsDigit(S[I]) then
    begin
        Y := Y * 100 + Dig(S[I]) * 10 + Dig(S[I + 1]);
        Inc(I, 2);
    end
    else
    begin
        if Y < 50 then
            Y := Y + 2000
        else
            Y := Y + 1900;
    end;
    Result := EncodeDate(Y, M, D);
end;

function GetMonthYear(const S: String; var I: Integer): TDateTime;
var M, Y: Integer;
begin
    M := Dig(S[I]) * 10 + Dig(S[I + 1]);
    Y := Dig(S[I + 3]) * 10 + Dig(S[I + 4]);
    Inc(I, 5);
    if (Length(S) > I) and IsDigit(S[I]) then
    begin
        Y := Y * 100 + Dig(S[I]) * 10 + Dig(S[I + 1]);
        Inc(I, 2);
    end
    else
    begin
        if Y < 50 then
            Y := Y + 2000
        else
            Y := Y + 1900;
    end;
    Result := EncodeDate(Y, M, 1);
end;

function GetNumber(const S: String; var I: Integer): Integer;
begin
    Result := 0;
    while (I <= Length(S)) and IsDigit(S[I]) do
    begin
        Result := Result * 10 + Dig(S[I]);
        Inc(I);
    end;
end;

function HasDigits(const S: String): Boolean;
var I: Integer;
begin
    Result := TRUE;
    for I := 1 to Length(S) do
        if S[I] in ['0'..'9'] then
            Exit;
    Result := FALSE;
end;

function StrIsStartingFrom(const S: String; var I: Integer; const P: String): Boolean;
var J, L: Integer;
begin
    Result := FALSE;
    L := Length(P);
    if Length(S) < I + L - 1 then Exit;
    for J := I + L - 1 downto I do
    begin
        if S[J] <> P[L] then Exit;
        Dec(L);
    end;
    Inc(I, Length(P) - 1);
    Result := TRUE;
end;

function IsMonthName(const S: String; var I, MN: Integer): Boolean;
const RusMonths: array[1..12] of String = ('', '', '', '',
        '', '', '', '', '', '', '', '');
    EngMonths: array[1..12] of String = ('JANUARY', 'FEBRUARY', 'MARCH', 'APRIL',
        'MAY', 'JUNE', 'JULAY', 'AUGUST', 'SEPTEMBER', 'OCTOBER', 'NOVEMBER', 'DECEMBER');
var K: Integer;
begin
    for K := 1 to 12 do
    begin
        if StrIsStartingFrom(S, I, RusMonths[K]) or
            StrIsStartingFrom(S, I, EngMonths[K]) then
        begin
            MN := K;
            Result := TRUE;
            Exit;
        end;
    end;
    Result := FALSE;
end;

function IsMonthShortName(const S: String; var I, MN: Integer): Boolean;
const RusMonths: array[1..12] of String = ('', '', '', '',
        '', '', '', '', '', '', '', '');
    EngMonths: array[1..12] of String = ('JAN', 'FEB', 'MAR', 'APR',
        'MAY', 'JUN', 'JUL', 'AUG', 'SEP', 'OCT', 'NOV', 'DEC');
var K: Integer;
begin
    for K := 1 to 12 do
    begin
        if StrIsStartingFrom(S, I, RusMonths[K]) or
            StrIsStartingFrom(S, I, EngMonths[K]) then
        begin
            MN := K;
            Result := TRUE;
            Exit;
        end;
    end;
    Result := FALSE;
end;

function IsDayWeekName(const S: String; var I, WD: Integer): Boolean;
const RusDays: array[1..7] of String = ('', '', '',
        '', '', '', '');
    EngDays: array[1..7] of String = ('SUNDAY', 'MONDAY', 'TEUSDAY',
        'WEDNESDAY', 'THIRSDAY', 'FRIDAY', 'SATURDAY');
var K: Integer;
begin
    for K := 1 to 7 do
    begin
        if StrIsStartingFrom(S, I, RusDays[K]) or
            StrIsStartingFrom(S, I, EngDays[K]) then
        begin
            WD := K;
            Result := TRUE;
            Exit;
        end;
    end;
    Result := FALSE;
end;

function IsDayWeekShortName(const S: String; var I, WD: Integer): Boolean;
const RusDays: array[1..7] of String = ('', '', '', '', '', '', '');
    EngDays: array[1..7] of String = ('SU', 'MO', 'TE', 'WE', 'TH', 'FR', 'SA');
var K: Integer;
begin
    for K := 1 to 7 do
    begin
        if StrIsStartingFrom(S, I, RusDays[K]) or
            StrIsStartingFrom(S, I, EngDays[K]) then
        begin
            WD := K;
            Result := TRUE;
            Exit;
        end;
    end;
    Result := FALSE;
end;

function CompareNumbered(const s1, s2: String; ByCase: Boolean): Integer;
var i1, i2: Integer;
    n1, n2: Integer;
begin
    i1 := 1;
    i2 := 1;
    while (i1 <= Length(s1)) and (i2 <= Length(s2)) do
    begin
        if (s1[i1] in ['0'..'9']) and (s2[i2] in ['0'..'9']) then
        begin
            n1 := 0;
            while s1[i1] in ['0'..'9'] do
            begin
                n1 := n1 * 10 + Ord(s1[i1]) - Ord('0');
                inc(i1);
            end;
            n2 := 0;
            while s2[i2] in ['0'..'9'] do
            begin
                n2 := n2 * 10 + Ord(s2[i2]) - Ord('0');
                inc(i2)
            end;
            Result := n1 - n2;
            if Result <> 0 then Exit;
        end
        else
        begin
            if ByCase then
                Result := Ord(s1[i1]) - Ord(s2[i2])
            else
                Result := Ord(Upper[s1[i1]]) - Ord(Upper[s2[i2]]);
            if Result <> 0 then Exit;
            inc(i1);
            inc(i2);
        end;
    end;
    if (i1 <= Length(s1)) then Result := 1
    else
        if (i2 <= Length(s2)) then Result := -1
        else
            Result := 0;
end;

function min(i, j: Integer): Integer;
asm
  cmp eax, edx
  jle @@exit
  xchg eax, edx
@@exit:
end;

function CompareFastStr(const s1, s2: String): Integer;
var i: Integer;
    p1, p2: PChar;
begin
    p1 := PChar(s1);
    p2 := PChar(s2);
    if p1 = p2 then
    begin
        Result := 0;
        Exit;
    end;
    for i := 1 to min(Length(s1), Length(s2)) do
    begin
        Result := Ord(s1[i]) - Ord(s2[i]);
        if Result <> 0 then Exit;
    end;
    Result := Length(s1) - Length(s2);
end;

function CompareFastStrUpper(const s1, s2: String): Integer;
var i: Integer;
    p1, p2: PChar;
begin
    p1 := PChar(s1);
    p2 := PChar(s2);
    if p1 = p2 then
    begin
        Result := 0;
        Exit;
    end;
    for i := 1 to min(Length(s1), Length(s2)) do
    begin
        Result := Ord(Upper[s1[i]]) - Ord(Upper[s2[i]]);
        if Result <> 0 then Exit;
    end;
    Result := Length(s1) - Length(s2);
end;

function CompareStringsFast(ByNumber, ByCase, ByDate: Boolean; const S1, S2: String): Integer;
var I1, I2: Integer;
    D1, D2: TDateTime;
    //p1, p2: PChar;
begin
    {p1 := PChar( s1 );
    p2 := PChar( s2 );
    if p1 = p2 then
    begin
      Result := 0;
      Exit;
    end;}
    I1 := 1;
    I2 := 1;
    Result := 0;
    if ByDate and IsDate(S1, I1) and IsDate(S2, I2) then
    begin
        D1 := GetDate(S1, I1);
        D2 := GetDate(S2, I2);
        if D1 < D2 then
        begin
            Result := -1; Exit;
        end
        else
            if D1 > D2 then
            begin
                Result := 1; Exit;
            end;
    end
    else
        if ByNumber and HasDigits(S1) and HasDigits(S2) then
        begin
            {if Length( S1 ) < Length( S2 ) then
              Result := -1
            else if Length( S1 ) > Length( S2 ) then
              Result := 1
            else}
            begin
                Result := //AnsiCompareText( S1, S2 );
                CompareNumbered(S1, S2, ByCase);
            end;
        end;
    if Result <> 0 then Exit;
    if not ByCase then
    begin
        {SS1 := S1;
        SS2 := S2;
        if SS1 <> '' then CharUpper( @ SS1[ 1 ] );
        if SS2 <> '' then CharUpper( @ SS2[ 1 ] );}
        Result := //AnsiCompareText( S1, S2 );
        CompareFastStrUpper(S1, S2)
    end
    else
        Result := //AnsiCompareStr( S1, S2 );
        CompareFastStr(S1, S2);
end;

function CompareIntegers(I1, I2: Integer): Integer;
asm
  CMP EAX, EDX
  SETNZ AL
  MOVZX EAX, AL
  JGE   @@exit
  NEG   EAX
@@exit:
end;

function CompareFloats(D1, D2: Double): Integer;
begin
    Result := 0;
    if D1 < D2 then Result := -1 else
        if D1 > D2 then Result := 1;
end;

function CompareStringsDetailed(const S1, S2, U1, U2: String; Attr: TColumnSortAttrs): Integer;
var I, J: Integer;
    D1, D2: TDateTime;
    N1, N2: Integer;
begin
    Result := 0;
    I := 1;
    J := 1;
    PrepareTableRussianToEnglish;
    while (I <= Length(S1)) and (J <= Length(S2)) do
    begin
        if saIgnoreSpaces in Attr then
        begin
            while (I <= Length(S1)) and (S1[I] <= ' ') do Inc(I);
            while (J <= Length(S2)) and (S2[J] <= ' ') do Inc(J);
            if (I > Length(S1)) then break;
            if (J > Length(S2)) then break;
        end;
        if IsDigit(S1[I]) and IsDigit(S2[J]) then
        begin
            if (saDate in Attr) and IsDate(S1, I) and IsDate(S2, J) then
            begin
                D1 := GetDate(S1, I);
                D2 := GetDate(S2, J);
                Result := CompareFloats(D1, D2);
                if Result <> 0 then Exit;
            end;
            if (saMonthYear in Attr) then
                if IsMonthYear(S1, I) and IsMonthYear(S2, J) then
                begin
                    D1 := GetMonthYear(S1, I);
                    D2 := GetMonthYear(S2, J);
                    Result := CompareFloats(D1, D2);
                    if Result <> 0 then Exit;
                end;
            if saNumber in Attr then
            begin
                N1 := GetNumber(S1, I);
                N2 := GetNumber(S2, J);
                Result := CompareIntegers(N1, N2);
                if Result <> 0 then Exit;
            end;
        end;
        if (saMonthName in Attr) and IsMonthName(U1, I, N1) and IsMonthName(U2, J, N2) then
        begin
            Result := CompareIntegers(N1, N2);
            if Result <> 0 then Exit;
        end;
        if (saMonthShortName in Attr) and IsMonthShortName(U1, I, N1) and IsMonthShortName(U2, J, N2) then
        begin
            Result := CompareIntegers(N1, N2);
            if Result <> 0 then Exit;
        end;
        if (saDayWeekName in Attr) and IsDayWeekName(U1, I, N1) and IsDayWeekName(U2, J, N2) then
        begin
            Result := CompareIntegers(N1, N2);
            if Result <> 0 then Exit;
        end;
        if (saDayWeekShortName in Attr) and IsDayWeekShortName(U1, I, N1) and IsDayWeekShortName(U2, J, N2) then
        begin
            Result := CompareIntegers(N1, N2);
            if Result <> 0 then Exit;
        end;
        if (I > Length(S1)) then break;
        if (J > Length(S2)) then break;
        if saNotRusLatEq in Attr then
        begin
            if S1[I] < S2[J] then Result := -1 else
                if S1[I] > S2[J] then Result := 1;
        end
        else
        begin
            if Eng2RusTable[S1[I]] < Eng2RusTable[S2[J]] then Result := -1 else
                if Eng2RusTable[S1[I]] > Eng2RusTable[S2[J]] then Result := 1;
        end;
        if Result <> 0 then Exit;
        Inc(I);
        Inc(J);
    end;
    if I <= Length(S1) then Result := 1 else
        if J <= Length(S2) then Result := -1
        else
        begin
            if S1 < S2 then Result := -1 else
                if S1 > S2 then Result := 1 else
                    Result := 0;
        end;
end;

procedure TFastListview.CompareItemsFast(Sender: TObject; Item1,
    Item2: TFastListItem; var Compare: Integer);
var FirstPass: Boolean;

    procedure DoCompareSubitems(ColIdx: Integer);
    var S1, S2, U1, U2: String;
        I1, I2: Integer;
        Attr: TColumnSortAttrs;
        Dw: DWORD;
    begin
        if FirstPass and (SortByIcons = sbiClickOnIcon) and FIconclicked or
            (SortByIcons = sbiAlways) then
        begin
            if Item1.ImageIndex <> Item2.ImageIndex then
            begin
                if Item1.ImageIndex < Item2.ImageIndex then
                    if Item1.ImageIndex >= 0 then
                        Compare := -1
                    else
                        Compare := 1
                else
                    if Item2.ImageIndex >= 0 then
                        Compare := 1
                    else
                        Compare := -1;
            end;
        end;
        if Compare <> 0 then
            Exit;
        if ColIdx = 0 then
        begin
            S1 := Item1.Caption;
            S2 := Item2.Caption;
        end
        else
        begin
            if VirtualTotally then
            begin
                S1 := '';
                S2 := '';
                if Assigned(OnDataCell) then
                begin
                    OnDataCell(Self, Item1.Index, ColIdx, S1);
                    OnDataCell(Self, Item2.Index, ColIdx, S2);
                end;
            end
            else
            begin
                if Item1.SubItemsList.Count > ColIdx - 1 then
                    S1 := Item1.SubItemsList[ColIdx - 1]
                else
                    S1 := '';
                if Item2.SubItemsList.Count > ColIdx - 1 then
                    S2 := Item2.SubItemsList[ColIdx - 1]
                else
                    S2 := '';
            end;
        end;
        TRY

            if (ColIdx < FColumnSortAttrs.Count) and (FColumnSortAttrs[ColIdx] <> nil) then
            begin
                DW := DWORD(FColumnSortAttrs[ColIdx]);
                Attr := PColumnSortAttrs(@DW)^;
                if saImageIndex in Attr then
                begin
                    if ColIdx = 0 then
                    begin
                        I1 := Item1.ImageIndex;
                        I2 := Item2.ImageIndex;
                        Compare := CompareIntegers(I1, I2);
                    end;
                    {else
                  begin
                    I1 := -1; if Item1.SubItems.Count > ColIdx - 1 then I1 := Item1.SubItems[ ColIdx - 1 ].ImageIndex;
                    I2 := -1; if Item2.SubItems.Count > ColIdx - 1 then I2 := Item2.SubItems[ ColIdx - 1 ].ImageIndex;
                  end;}
                end;
                if (Compare = 0) and (saStateImage in Attr) then
                begin
                    if ColIdx = 0 then
                    begin
                        I1 := Item1.StateIndex;
                        I2 := Item2.StateIndex;
                        Compare := CompareIntegers(I1, I2);
                    end;
                end;
                {if (Compare = 0) and (saCheckedFirst in Attr) then
                begin
                  if ColIdx = 0 then
                    Compare := CompareIntegers( Integer( not Item1.Checked ), Integer( not Item2.Checked ) );
                end;}
                if (Compare <> 0) and (saImgDescending in Attr) then
                    Compare := -Compare;
                if (Compare = 0) and (saLenghtFirst in Attr) then
                    Compare := CompareIntegers(Length(S1), Length(S2));
                if (Compare = 0) and (saEmptyFirst in Attr) then
                    if ((S1 <> '') or (S2 <> '')) and ((S1 = '') or (S2 = '')) then
                        Compare := CompareIntegers(Length(S2), Length(S1));
                if Compare = 0 then
                begin
                    S1 := TrimRight(S1);
                    S2 := TrimRight(S2);
                    if (S1 <> S2) then
                    begin
                        U1 := ''; U2 := '';
                        if (saCase in Attr) or (saDayWeekName in Attr) or (saDayWeekShortName in Attr) or
                            (saMonthName in Attr) or (saMonthShortName in Attr) then
                        begin
                            U1 := AnsiUpperCase(S1);
                            U2 := AnsiUpperCase(S2);
                            if saCase in Attr then
                            begin
                                S1 := U1;
                                S2 := U2;
                            end;
                        end;
                        Compare := CompareStringsDetailed(S1, S2, U1, U2, Attr);
                        if (Compare <> 0) and (saDescending in Attr) then
                            Compare := -Compare;
                    end;
                end;
            end
            else
                if (Compare = 0) {and (S1 <> S2)} then
                    Compare := CompareStringsFast(FSortByNumber, FSortByCase, FSortByDate,
                        S1, S2);
        FINALLY
            //LogCompare( S1, S2, Compare );
        END;
    end;

    procedure CompareSubitems(ColIdx: Integer; InvertOrder: Boolean);
    begin
        DoCompareSubitems(ColIdx);
        if InvertOrder then
            Compare := -Compare;
    end;

var I, StartColumn: Integer;
begin

    FirstPass := TRUE;
    Compare := 0;

    if FColumnsToSort.Count > 0 then
    begin
        for I := 0 to FColumnsToSort.Count - 1 do
        begin
            CompareSubitems($FFFF and (Integer(FColumnsToSort[I])),
                Integer(FColumnsToSort[I]) < 0);
            FirstPass := FALSE;
            if Compare <> 0 then Exit;
        end;
        StartColumn := $FFFF and Integer(FColumnsToSort[FColumnsToSort.Count - 1]);
        if not FOrderByFollowsColumns then
        begin
            //Compare := Item1.Index - Item2.Index;
            Exit;
        end;
    end
    else
    begin
        StartColumn := FColumnclicked;
    end;


    if StartColumn < 0 then
        StartColumn := 0;
    I := StartColumn;
    if I < 0 then
        I := 0;
    repeat
        CompareSubitems(I, FALSE);
        if Compare <> 0 then
            Exit;

        Inc(I);
        if (I > Item1.SubItemsList.Count) or (I > Item2.SubItemsList.Count) then
        begin
            if Item1.SubItemsList.Count <> Item2.SubItemsList.Count then
            begin
                if I > Item1.SubItemsList.Count then
                    Compare := 1
                else
                    Compare := -1;
                if Compare <> 0 then Exit;
            end;
            if I = 1 then Exit;
            I := 0;
        end;
        if Compare <> 0 then
            Exit;
        FirstPass := FALSE;
    until I = StartColumn;
end;

constructor TFastListview.Create(AOwner: TComponent);
    function AddMenuItem(AParentMenu: TMenuItem; const ACaption: String; Event: TNotifyEvent): TMenuItem;
    var MI: TMenuItem;
    begin
        MI := TMenuItem.Create(Self);
        MI.Caption := ACaption;
        MI.OnClick := Event;
        if AParentMenu = nil then
            FColPopup.Items.Add(MI)
        else
            AParentMenu.Add(MI);
        Result := MI;
    end;
var MI: TMenuItem;
begin
    {$IFDEF LOG_FLV}
    LogFileOutput('c:\log_flv.txt', 'FastListView_Create');
    {$ENDIF}
    inherited;
    FHdrTrkIdx := -1;
    ViewStyle := vsReport;
    FGridLines1 := TRUE;
    FAutoSelectOnEnter := TRUE;
    AllowColumnDrag := TRUE;
    AllowColumnHide := TRUE;
    FRegistrySaveRestore := TRUE;
    FColumnsToSort := TList.Create;
    FHiddenItems := TList.Create;
    FHiddenColumns := TStringList.Create;
    FFilterButtonClicked := -1;
    FColumnsToFilter := TStringList.Create;
    FColumnSortAttrs := TList.Create;
    FSortTo := -1;
    F_ItemsList := TFastListItems.Create;
    F_ItemsList.FOwner := Self;
    FSortByNumber := TRUE;
    FSortByDate := TRUE;
    inherited OnSelectItem := DoOnSelectItem;
    inherited OnHdrDividerDblClk := doHdrDivider;
    inherited OnHdrPaint := doHdrPaint;
    FColPopup := TPopupMenu.Create(Self);
    FColPopup.OnPopup := ColPopupOnPopup;
    {$IFDEF LOG_FLV}
    LogFileOutput('c:\log_flv.txt', 'FastListView_Create 1');
    {$ENDIF}
    {$IFDEF CHK_COLUMN_VISIBLE}
    LogFileOutput('c:\column_vis.txt', 'creating FastListView csDesigning:' +
        IntToStr(Integer(csDesigning in ComponentState)));
    {$ENDIF}
    if not (csDesigning in ComponentState) then
    begin
        inherited OnColumnClick := ColumnClickFast;
        inherited OwnerData := TRUE;
        inherited OnData := Get_Data;
        inherited OnEditing := Do_OnEditing;
        inherited OnEdited := Do_OnEdited;
        MI := AddMenuItem(nil, HeadPopupSort, nil);
        miSortOneColumn :=
            AddMenuItem(MI, HeadPopupOnlyByThisCol, miSortOneColumnClick);
        miAddColumn :=
            AddMenuItem(MI, HeadPopupAddColToSort, miAddColumnClick);
        miRemoveColumn :=
            AddMenuItem(MI, HeadPopupDelColFromSort, miRemoveColumnClick);
        AddMenuItem(MI, '-', nil);
        miIncreaseOrder :=
            AddMenuItem(MI, HeadPopupIncColOrder, miIncreaseOrderClick);
        miDecreaseOrder :=
            AddMenuItem(MI, HeadPopupDecColOrder, miDecreaseOrderClick);
        AddMenuItem(MI, '-', nil);
        miAscending :=
            AddMenuItem(MI, HeadPopupSortAscending, miAscendingClick);
        miDescending :=
            AddMenuItem(MI, HeadPopupSortDescending, miDescendingClick);
        MI := AddMenuItem(nil, HeadPopupFilter, nil);
        miFilterMenu := MI;
        miFilter :=
            AddMenuItem(MI, HeadPopupFilter_Filter, miFilterClick);
        miResetFilter :=
            AddMenuItem(MI, HeadPopupFilter_Reset, miResetFilterClick);
        miResetAllFilters :=
            AddMenuItem(MI, HeadPopupFilter_ResetAllCols, miResetAllFiltersClick);
        MI := AddMenuItem(nil, HeadPopupColumns, nil);
        {$IFDEF CHK_COLUMN_VISIBLE}
        LogFileOutput('c:\column_vis.txt', 'Creating miHideMenu');
        {$ENDIF}
        miHideMenu := MI;
        miHideColumn :=
            AddMenuItem(MI, HeadPopupHide, miHideColumnClick);
        miHideColumn.Tag := -1;
        miHideShowSep1 :=
            AddMenuItem(MI, '-', nil);
        miHideShowSep1.Tag := -1;
        miHideShowSep2 :=
            AddMenuItem(MI, '-', nil);
        miHideShowSep2.Tag := -1;
        miShowAllColumns :=
            AddMenuItem(MI, HeadPopupShowAllCols, miShowAllColumnsClick);
        miShowAllColumns.Tag := -1;
        AddMenuItem(MI, '-', nil).Tag := -1;
        miRestoreOrder :=
            AddMenuItem(MI, HeadPopupRestoreColsOrder, miRestoreOrderClick);
        miRestoreOrder.Tag := -1;
        miSizeProportional :=
            AddMenuItem(MI, HeadPopupProportional, miSizeProportionalClick);
        miSizeProportional.Tag := -1;
    end;
    FUniSources := TList.Create;
    {$IFDEF LOG_FLV}
    LogFileOutput('c:\log_flv.txt', 'FastListView_Create exit');
    {$ENDIF}
end;

procedure TFastListview.DeleteSelection;
var I: Integer;
    L: TList;
    LI: TFastListItem;
    Foc: Integer;
begin
    L := TList.Create;
    Items.BeginUpdate;
    try
        LI := Selected;
        if LI <> nil then
        begin
            Foc := LI.Index;
            while LI <> nil do
            begin
                L.Add(LI);
                LI := //GetNextItem( LI, sdAll, [ isSelected ] );
                NextSelected(LI);
            end;
            Selected := nil;
            SetCount(Count - L.Count);
            for I := L.Count - 1 downto 0 do
            begin
                LI := L[I];
                LI.Free;
            end;
            if Foc >= Count then
                Foc := Count - 1;
            if Foc >= 0 then
                ItemFocused := Items[Foc];
        end;
    finally
        L.Free;
        Items.EndUpdate;
    end;
end;

destructor TFastListview.Destroy;
var I: Integer;
    LI: TFastListItem;
begin
    FColumnsEditor_DesignTime.Free;
    SaveSettingsInRegistry;
    for I := FHiddenItems.Count - 1 downto 0 do
    begin
        LI := FHiddenItems[I];
        LI.Free;
    end;
    FHiddenItems.Free;
    F_ItemsList.Free;
    for I := 0 to FColumnsToFilter.Count - 1 do
        TStringList(FColumnsToFilter.Objects[I]).Free;
    FColumnsToFilter.Free;
    FTriangleBmp.Free;
    FTrianglePressedBmp.Free;
    FTriangleFilteredBmp.Free;
    FColumnSortAttrs.Free;
    FColumnsToSort.Free;
    FHiddenColumns.Free;
    FFakeHdrFont.Free;
    FColumnNames.Free;
    for I := 0 to FUniSources.Count - 1 do
        (UniSources[I] as TUniSource).FastListViewExists := FALSE;
    FUniSources.Free;
    if Assigned(FColumnComboItems) then
    begin
        for I := 0 to FColumnComboItems.Count - 1 do
            TObject(FColumnComboItems[I]).Free;
        FColumnComboItems.Free;
    end;
    FColumnCanEdit.Free;
    FColumnEditKind.Free;
    inherited;
end;

{procedure TFastListview.DoDataFind(Sender: TObject; Find: TItemFind;
  const FindString: string; const FindPosition: TPoint; FindData: Pointer;
  StartIndex: Integer; Direction: TSearchDirection; Wrap: Boolean;
  var Index: Integer);
begin
  if Assigned( F_DataFind ) then
    F_DataFind( Self, Find, FindString, FindPosition, FindData, StartIndex,
                Direction, Wrap, Index )
  else
  begin

  end;
end;}

procedure TFastListview.DoAdvancedDrawSubItem(Sender: TCustomListView;
    Item: TOldListItem; SubItem: Integer; State: TCustomDrawState;
    Stage: TCustomDrawStage; var DefaultDraw: Boolean);
begin
    {if Assigned( OnAdvancedCustomDrawSubItem ) then
     OnAdvancedCustomDrawSubItem( self, Items[ Item.Index ], SubItem, State,
     Stage, DefaultDraw );}
end;

procedure TFastListview.DoCustomDrawSubitem(Sender: TCustomListView;
    Item: TOldListItem; SubItem: Integer; State: TCustomDrawState;
    var DefaultDraw: Boolean);
begin
    {if assigned( FOnCustomDrawSubItem ) then
      FOnCustomDrawSubItem( Self, Items[ Item.Index ], SubItem, State, DefaultDraw );}
end;

procedure TFastListview.DoDrawItem(Sender: TCustomListView;
    Item: TOldListItem; ARect: TRect; State: TOwnerDrawState);
var FItem: TFastListItem;
begin
    Assert(Item <> nil, Err0001 +
        Sender.Name + ': TFastListView');
    FItem := F_ItemsList[Item.Index];
    if Assigned(fOnDraw) then
        fOnDraw(Sender, FItem, ARect, State);
end;

procedure TFastListview.DoFilterButtonClick(ColIdx: Integer; const Pt1, Pt2: TPoint);
var W: Integer;
    I, J: Integer;
    vLI: TFastListItem;
    DW: DWORD;
    IH, H: Integer;
    D: TSortedData;
    S, S1: String;
begin
    if Items.Count + HiddenItemsCount = 0 then Exit;
    if Assigned(FFilterListForm) then
    begin
        if FFilterButtonClicked = ColIdx then
        begin
            if FFilterListForm.lvValues.SelCount > 0 then
            begin
                FFilterListForm.PreventClose(TRUE);
                TRY
                    ApplyFilters;
                FINALLY
                    FFilterListForm.PreventClose(FALSE);
                END;
                FreeAndNil(FFilterListForm);
                InvalidateHeader;
                Exit;
            end;
        end;
    end
    else
    begin
        FFilterListForm := TfmFilterList.Create(Self);
        FFilterListForm.lvValues.OnKillFocus := KillFilterListform;
        FFilterListForm.lvValues.OnKeyPress := FilterListKeyPress;
        FFilterListForm.lvValues.OnDblClick := FilterListDblClk;
        FFilterListForm.lvValues.Onselecting := FilterListSelecting;
    end;
    FFilterListForm.PreventClose(TRUE);
    TRY

        FFilterListForm.FilteredLV := Self;
        FFilterListForm.Filteredcolumn := ColIdx;
        FFilterButtonClicked := ColIdx;
        InvalidateHeader;
        W := Pt2.x - Pt1.x;
        if W < FFilterListForm.Width then
            W := FFilterListForm.Width;
        FFilterListForm.Width := W;
        if Pt1.x + FFilterListForm.Width < Screen.Width then
            FFilterListForm.Left := Pt1.x
        else
            FFilterListForm.Left := Pt2.x - FFilterListForm.Width;
        if Pt2.y + FFilterListForm.Height - FFilterListForm.CaptionHeigth < Screen.Height then
            FFilterListForm.Top := Pt2.y - FFilterListForm.CaptionHeigth
        else
            FFilterListForm.Top := Pt1.y - FFilterListForm.Height; // - FFilterListForm.CaptionHeigth;
        FFilterListForm.lvValues.Items.BeginUpdate;
        TRY
            FFilterListForm.lvValues.Items.Clear;

            D.ListSorted := TList.Create;
            D.FLVSorted := Self;
            D.ColIdx := ColIdx;

            TRY
                for I := 0 to Items.Count - 1 do
                    D.ListSorted.Add(Pointer(I));
                SortData(@D, D.ListSorted.Count, CompareSorted, SwapSorted);
                J := -1;
                S := '';
                vLI := nil;
                for I := 0 to D.ListSorted.Count - 1 do
                begin
                    if ColIdx = 0 then
                        S1 := Items[Integer(D.ListSorted[I])].Caption
                    else
                        if ColIdx - 1 < Items[Integer(D.ListSorted[I])].SubItemsList.Count then
                            S1 := Items[Integer(D.ListSorted[I])].SubItemsList[ColIdx - 1]
                        else
                            S1 := '';
                    if (I = 0) or (S1 = S) then
                    begin
                        Inc(J); S := S1;
                        if I = 0 then
                        begin
                            J := 1;
                            vLI := FFilterListForm.lvValues.Items.Add;
                            vLI.SubItemsList.Add('');
                            vLI.Caption := S1;
                        end;
                    end
                    else
                    begin
                        if vLI <> nil then
                        begin
                            vLI.SubItemsList[0] := IntToStr(J);
                            vLI.Data := Pointer(J);
                        end;
                        vLI := FFilterListForm.lvValues.Items.Add;
                        vLI.SubItemsList.Add('');
                        vLI.Caption := S1;
                        J := 1;
                        S := S1;
                    end;
                end;
                if vLI <> nil then
                begin
                    vLI.SubItemsList[0] := IntToStr(J);
                    vLI.Data := Pointer(J);
                end;

                D.ListSorted.Clear;
                for I := 0 to FHiddenItems.Count - 1 do
                    D.ListSorted.Add(Pointer(I));
                SortData(@D, D.ListSorted.Count, CompareHiddenSorted, SwapSorted);
                J := -1;
                S := '';
                vLI := nil;
                for I := 0 to D.ListSorted.Count - 1 do
                begin
                    if ColIdx = 0 then
                        S1 := HiddenItems[Integer(D.ListSorted[I])].Caption
                    else
                        S1 := HiddenItems[Integer(D.ListSorted[I])].SubItemsList[ColIdx - 1];
                    if (I = 0) or (S1 = S) then
                    begin
                        Inc(J); S := S1;
                        if I = 0 then
                        begin
                            J := 1;
                            vLI := FFilterListForm.lvValues.Items.Add;
                            vLI.FontColor := clBtnShadow;
                            vLI.FontColorSelected := clBtnFace;
                            vLI.SubItemsList.Add('');
                            vLI.Caption := S1;
                        end;
                    end
                    else
                    begin
                        if vLI <> nil then
                        begin
                            vLI.SubItemsList[0] := IntToStr(J);
                            vLI.Data := Pointer(J);
                        end;
                        vLI := FFilterListForm.lvValues.Items.Add;
                        vLI.FontColor := clBtnShadow;
                        vLI.FontColorSelected := clBtnFace;
                        vLI.SubItemsList.Add('');
                        vLI.Caption := S1;
                        J := 1;
                        S := S1;
                    end;
                end;
                if vLI <> nil then
                begin
                    vLI.SubItemsList[0] := IntToStr(J);
                    vLI.Data := Pointer(J);
                end;

            FINALLY
                D.ListSorted.Free;
            END;
            //FFilterListForm.lvValues.SortType := stBoth;

            if FColumnSortAttrs <> nil then
                if FColumnSortAttrs.Count > ColIdx then
                    if FColumnSortAttrs.Items[ColIdx] <> nil then
                    begin
                        DW := DWORD(FColumnSortAttrs[ColIdx]);
                        FFilterListForm.lvValues.SetSortColumnAttributes([0],
                            PColumnSortAttrs(@DW)^);
                    end;

            FFilterListForm.lvValues.AlphaSort;
            FFilterListForm.lvValues.SortType := stNone;
            IH := 0;
            if FFilterListForm.lvValues.Parent <> nil then
                IH := FFilterListForm.lvValues.Canvas.TextHeight('Ap_^\');
            if IH <= 0 then IH := 20;
            if (FFilterListForm.lvValues.Items.Count > 0) and
                (FFilterListForm.lvValues.Items.Count <= 10) then
            begin
                FFilterListForm.Height :=
                    (FFilterListForm.lvValues.Items.Count + 2) * (IH + 1)
                    + FFilterListForm.Height
                    - FFilterListForm.lvValues.Height + 5
            end
            else
            begin
                H := Height div 2;
                if H < 100 then H := 100;
                if H > (FFilterListForm.lvValues.Items.Count + 2) * (IH + 1)
                    + FFilterListForm.Height
                    - FFilterListForm.lvValues.Height + 5 then
                    H := (FFilterListForm.lvValues.Items.Count + 2) * (IH + 1)
                        + FFilterListForm.Height
                        - FFilterListForm.lvValues.Height + 5;
                FFilterListForm.Height := H;
            end;
            if Pt2.y + FFilterListForm.Height - FFilterListForm.CaptionHeigth < Screen.Height then
                FFilterListForm.Top := Pt2.y - FFilterListForm.CaptionHeigth
            else
                FFilterListForm.Top := Pt1.y - FFilterListForm.Height; // - FFilterListForm.CaptionHeigth;
            vLI := FFilterListForm.lvValues.Items.Insert(0);
            vLI.Caption := FiltAll;
            vLI.SubItemsList.Add(IntToStr(Items.Count + FHiddenItems.Count));
        FINALLY
            FFilterListForm.lvValues.Items.EndUpdate;
        END;

    FINALLY
        FFilterListForm.PreventClose(FALSE);
    END;
    FFilterListForm.Show;
end;

procedure TFastListview.doHdrDivider(Sender: TObject; columnIdx: Integer);
var I, W, X: Integer;
    LI: TFastListItem;
    S: String;
    bmp: TBitmap;
begin
    if Assigned(OnHdrDividerDblClk) then
        OnHdrDividerDblClk(Sender, columnIdx)
    else
    begin
        Bmp := TBitmap.Create;
        TRY
            Bmp.Width := 1;
            Bmp.Height := 1;
            Bmp.Canvas.Font.Assign(Canvas.Font);
            if BoldIfFocused then
                Bmp.Canvas.Font.Style := Bmp.Canvas.font.Style + [fsBold];

            W := 0;
            for I := TopItem.Index to TopItem.Index + VisibleRowCount - 1 do
            begin
                if I >= Items.Count then break;
                LI := Items[I];
                if columnIdx = 0 then
                begin
                    X := 6;
                    S := LI.Caption;
                    if SmallImages <> nil then
                        X := X + 4 + SmallImages.Width;
                    if StateImages <> nil then
                        X := X + 4 + StateImages.Width;
                end
                else
                begin
                    S := '';
                    X := 6;
                    if LI.SubItemsList.Count > columnIdx - 1 then
                    begin
                        S := LI.SubItemsList[columnIdx - 1];
                    end;
                end;
                X := X + Bmp.Canvas.TextWidth(S);
                if X > W then
                    W := X;
            end;

            if W = 0 then
                W := 4;
            Columns[columnIdx].Width := W;

        FINALLY
            Bmp.Free;
        END;

    end;
end;

{$R FLV_TRIANGLE.RES}

procedure TFastListview.doHdrPaint(Sender: TObject; Canvas: TCanvas;
    R: TRect; ColIdx: Integer);
var J, L, H: Integer;
    OldBrushColor: TColor;
    Bmp: TBitmap;
    SL: TStringList;
    S: String;
    TotalRect, R1: TRect;

    procedure DrawMark;
    begin
        if Integer(FColumnsToSort[J]) >= 0 then
        begin
            Canvas.Polygon([Point(R.Left + 3, R.Bottom - 5), Point(R.Left + 4 + L div 2, R.Bottom - 3),
                Point(R.Left + 5 + L, R.Bottom - 5), Point(R.Left + 6 + L, R.Bottom - 6),
                    Point(R.Left + 6 + L, R.Top + 3), Point(R.Left + 5 + L, R.Top + 2),
                    Point(R.Left + 3, R.Top + 2), Point(R.Left + 2, R.Top + 3),
                    Point(R.Left + 2, R.Bottom - 6)]);
        end
        else
        begin
            Canvas.Polygon([Point(R.Left + 2, R.Top + 5), Point(R.Left + 3, R.Top + 4),
                Point(R.Left + 4 + L div 2, R.Top + 2),
                    Point(R.Left + 5 + L, R.Top + 4), Point(R.Left + 6 + L, R.Top + 5),
                    Point(R.Left + 6 + L, R.Bottom - 4), Point(R.Left + 5 + L, R.Bottom - 3),
                    Point(R.Left + 3, R.Bottom - 3), Point(R.Left + 2, R.Bottom - 4)]);
        end;
    end;

    function IAbs(I: Integer): Integer;
    begin
        Result := I;
        if I < 0 then Result := -I;
    end;

var OldFontHeight: Integer;
    I: Integer;
    pts_triangle: array[0..2] of TPoint;
    x, y: Integer;
    R00: TRect;
begin
    R00 := R;
    if (MultiHeader <> '') and not (csDesigning in ComponentState) then
        RecalcHeaderHeight;
    TotalRect := R;
    I := Canvas.TextHeight('Ap_^');
    if I < 18 then I := 18;
    R.Top := R.Bottom - I;
    R1 := R;
    J := FColumnsToFilter.IndexOf(IntToStr(ColIdx));
    if J >= 0 then
    begin
        LoadTriangleBmp;
        L := R.Right - FTriangleBmp.Width - 6;
        if L >= R.Left + FTriangleBmp.Width then
        begin
            R.Left := L;
            Bmp := FTriangleBmp;
            if FFilterButtonClicked = ColIdx then
                Bmp := FTrianglePressedBmp
            else
            begin
                SL := TStringList(FColumnsToFilter.Objects[J]);
                if (SL <> nil) and (SL.Count > 0) then
                    Bmp := FTriangleFilteredBmp;
            end;
            Bmp.Transparent := TRUE;
            if Canvas.Handle = ForceCanvas then
            begin
                Canvas.Brush.Color := clBlack;
                Canvas.Brush.Style := bsSolid;
                pts_triangle[0] := Point(L + 1, R.Top + (R.Bottom - R.Top - FTriangleBmp.Height) div 2);
                pts_triangle[1] := Point(L + 9, R.Top + (R.Bottom - R.Top - FTriangleBmp.Height) div 2);
                pts_triangle[2] := Point(L + 5, R.Top + (R.Bottom - R.Top - FTriangleBmp.Height) div 2 + 8);
                Canvas.Polygon(pts_triangle);
                if Bmp = FTriangleFilteredBmp then
                begin
                    Canvas.Brush.Color := clLime;
                    x := (pts_triangle[0].x + pts_triangle[1].x) div 2 + 1;
                    y := (pts_triangle[0].y + pts_triangle[2].y) div 2;
                    Canvas.Ellipse(x - 3, y - 3, x + 2, y + 2);
                end;
            end
            else
            begin
                Canvas.Brush.Style := bsClear;
                Canvas.Draw(L + 1,
                    R.Top + (R.Bottom - R.Top - FTriangleBmp.Height) div 2,
                    Bmp); //uuu
            end;
        end;
    end;
    J := IndexOfAbs(FColumnsToSort, ColIdx);
    if (J > 0) and not NoMultiColumnSort
        or (J = 0) then
    begin
        R := R1;
        if (Columns[ColIdx].ImageIndex >= 0) and Assigned(SmallImages) then
            R.Left := R.Left + 4 + SmallImages.Width;
        S := IntToStr(J + 1);
        OldFontHeight := Canvas.Font.Height;
        Canvas.font.Height := IAbs(Canvas.Font.Height) - 2;
        L := Canvas.TextWidth(S) or 1;
        OldBrushColor := Canvas.Brush.Color;
        Canvas.Brush.Color := clWindow;
        DrawMark;
        {Canvas.Pen.Style := psClear;
        DrawMark;}
        Canvas.Pen.Style := psSolid;
        Canvas.Brush.Style := bsClear;
        H := Canvas.TextHeight(S);
        Canvas.TextRect(Rect(R.Left + 4, R.Top + 3, R.Left + 4 + L + 3, R.Bottom - 3),
            R.Left + 4, R.Top + (R.Bottom - R.Top - H) div 2, S);
        Canvas.Brush.Style := bsSolid;
        Canvas.Brush.Color := OldBrushColor;
        Canvas.Font.Height := OldFontHeight;
    end;
    if not DisableFrameAroundFocusedListView then
        if Focused then
        begin
            Canvas.Brush.Color := clBlue;
            inc(R00.Right);
            Canvas.FrameRect(R00);
        end;
    if Assigned(OnHdrPaint) then
        OnHdrPaint(Sender, Canvas, R, ColIdx);
end;

procedure TFastListview.DoOnSelectItem(Sender: TObject;
    Item: TOldListItem; ASelected: Boolean);
var FastItem: TFastListItem;
begin
    if Item <> nil then
        FastItem := Items[Item.Index]
    else
        FastItem := Selected;
    if FastItem <> nil then
    begin
        if ASelected then
            if Assigned(OnSelecting) then
                if not Onselecting(Self, FastItem) then
                begin
                    if Item = nil then
                        Item := FastItem.MirrorItem;
                    Item.Selected := FALSE;
                    FastItem.FSelected := FALSE;
                    Exit;
                end;
        FastItem.FSelected := ASelected;
    end;
    if Assigned(OnSelectItem) then
    begin
        OnSelectItem(Self, FastItem, ASelected);
    end;
end;

function TFastListview.FilterButtonClick(IgnoreCursorPos: Boolean; ItemIdx: Integer = -1): Boolean;
var I: Integer;
    HdrWnd: HWnd;
    ClassNameBuf: array[0..49] of Char;
    HdHti: THDHitTestInfo;
    P, PScreen: TPoint;
    R: TRect;
    Cancelled: Boolean;
begin
    Result := FALSE;
    if Assigned(fOnBeforeFilter) then
    begin
        Cancelled := FALSE;
        SendMessage(GetHeaderCtlWindow, WM_LBUTTONUP, 0, 0);
        OnBeforeFilter(Self, Cancelled);
        if Cancelled then Exit;
    end;

    GetCursorPos(PScreen);
    P := PScreen;
    windows.ScreenToClient(Handle, P);
    if IgnoreCursorPos then
        HdrWnd := GetHeaderCtlWindow
    else
        HdrWnd := childWindowFromPoint(Handle, P);

    if (HdrWnd <> 0) and (HdrWnd <> Handle) then
    begin

        GetClassName(HdrWnd, ClassNameBuf, 50);
        if ClassNameBuf = 'SysHeader32' then
        begin
            P := PScreen;
            windows.ScreenToClient(HdrWnd, P);
            HdHti.Point := P;
            SendMessage(HdrWnd, HDM_HITTEST, 0, Integer(@HdHti));
            if not IgnoreCursorPos then
                if (HdHti.Flags = HHT_ONDIVIDER) or (HdHti.Flags = HHT_ONDIVOPEN) then
                    Exit;
            if ItemIdx >= 0 then
                I := ItemIdx
            else
                I := HdHti.Item;

            if FColumnsToFilter.IndexOf(IntToStr(I)) >= 0 then
            begin
                LoadTriangleBmp;
                SendMessage(HdrWnd, HDM_GETITEMRECT, I, Integer(@R));
                if IgnoreCursorPos or
                    (PtInRect(R, P) and (P.x > R.Right - 6 - FTriangleBmp.Width) or
                    Assigned(FFilterListForm) and FFilterListForm.Visible
                    ) then
                begin
                    Result := TRUE;
                    windows.ClientToScreen(HdrWnd, R.TopLeft);
                    windows.ClientToScreen(HdrWnd, R.BottomRight);
                    DoFilterButtonClick(I, R.TopLeft, R.BottomRight);
                    Exit;
                end;
            end;

        end;

    end;
end;

procedure TFastListview.FilterColumns(ColumnsList: array of Integer);
var I: Integer;
    SL: TStringList;
    DoReApply: Boolean;
begin
    if High(ColumnsList) < 0 then
    begin
        DoReApply := FALSE;
        for I := 0 to FColumnsToFilter.Count - 1 do
        begin
            SL := Pointer(FColumnsToFilter.Objects[I]);
            if SL <> nil then
            begin
                SL.Free;
                DoReApply := TRUE;
            end;
        end;
        FColumnsToFilter.Clear;
        if DoReApply then
            ApplyAllFilters(TRUE);
    end
    else
        for I := 0 to High(ColumnsList) do
        begin
            if FColumnsToFilter.IndexOf(IntToStr(ColumnsList[I])) < 0 then
                FColumnsToFilter.Add(IntToStr(ColumnsList[I]));
        end;
    InvalidateHeader;
end;

function TFastListview.FindData(StartIndex: Integer; Value: Pointer;
    Inclusive, Wrap: Boolean): TListItem;
var LI: TOldListItem;
begin
    LI := inherited FindData(StartIndex, Value, Inclusive, Wrap);
    if LI <> nil then
        Result := Items[LI.Index]
    else
        Result := nil;
end;

procedure TFastListview.GetColumnLeftRight(NCol: Integer; var L,
    R: Integer);
var HdrWnd: HWnd;
    Buf: array[0..49] of Char;
    Bounds: TRect;
begin
    HdrWnd := GetHeaderCtlWindow;
    if HdrWnd = 0 then Exit;
    GetClassName(HdrWnd, Buf, 50);
    if Buf <> 'SysHeader32' then Exit;
    if SendMessage(HdrWnd, HDM_GETITEMRECT, Ncol, Integer(@Bounds)) = 0 then Exit;
    Windows.ClientToScreen(HdrWnd, Bounds.TopLeft);
    Windows.ClientToScreen(HdrWnd, Bounds.BottomRight);
    Windows.ScreenToClient(Handle, Bounds.TopLeft);
    Windows.ScreenToClient(Handle, Bounds.BottomRight);
    L := Bounds.Left;
    R := Bounds.Right;
    if NCol = 0 then
    begin
        if R > ClientWidth then
            R := ClientWidth;
    end;
end;

function TFastListview.GetFocusedItem: TListItem;
var LI: TOldListItem;
begin
    Result := nil;
    LI := inherited ItemFocused;
    if LI = nil then Exit;
    Result := F_ItemsList[LI.Index];
end;

function TFastListview.GetInheritedItemsCount: Integer;
begin
    Result := inherited Items.Count;
end;

function TFastListview.GetItemAt(X, Y: Integer): TFastListItem;
var LI: TOldListItem;
begin
    LI := inherited GetItemAt(X, Y);
    Result := nil;
    if LI <> nil then
        Result := Items[LI.Index];
end;

function TFastListview.GetNextItem(StartItem: TListItem;
    Direction: TSearchDirection; States: TItemStates): TListItem;
var LI: TOldListItem;
begin
    LI := StartItem.MirrorItem;
    LI := inherited GetNextItem(LI, Direction, States);
    if (LI <> nil) and (LI.Index <> StartItem.Index) then
        Result := Items[LI.Index]
    else
        Result := nil;
end;

function TFastListview.GetOldItems(Idx: Integer): TOldListItem;
begin
    Result := inherited Items[Idx];
end;

function TFastListview.GetSelected: TListItem;
var LI: TOldListItem;
begin
    Result := nil;
    if not Visible and not MultiSelect then
    begin
        if FSelectedItemIdx < 0 then
            Result := nil
        else
            Result := Items[FSelectedItemIdx];
        Exit;
    end;
    LI := inherited Selected;
    if LI = nil then Exit;
    Result := F_ItemsList[LI.Index];
end;

procedure TFastListview.Get_Data(Sender: TObject; Item: TOldListItem);
var I, j: Integer;
    FastItem: TFastListItem;
    Txt: String;
    ImgIdx: Integer;
    Data: Pointer;
begin
    Assert(Item <> nil, Err0002 +
        (Sender as TComponent).Name + ': TFastListView');
    I := Item.Index;
    if VirtualTotally and Assigned(OnDataCell) then
    begin
        Txt := '';
        ImgIdx := 0;
        Data := nil;
        OnDataCell(Self, I, 0, Txt);
        ImgIdx := 0;
        if Assigned(OnDataCellImage) then
            OnDataCellImage(Self, I, ImgIdx);
        Data := nil;
        if Assigned(OnDataCellData) then
            OnDataCellData(Self, I, Data);
        Item.Caption := Txt;
        Item.ImageIndex := ImgIdx;
        Item.Data := Data;
        Item.SubItems.Clear;
        for j := 1 to Columns.Count - 1 do
        begin
            OnDataCell(Self, I, j, Txt);
            Item.SubItems.Add(Txt);
        end;
        Exit;
    end;

    if (I < 0) or (I >= F_ItemsList.Count) then
        FastItem := nil
    else
        FastItem := F_ItemsList[I];
    if FastItem <> nil then
    begin
        if Assigned(FLVOwnerDataEvent) then
            FLVOwnerDataEvent(Self, FastItem);
        FastItem.FListItem := Item;
        Item.Caption := FastItem.FCaption;
        Item.SubItems.Clear;
        Item.Data := FastItem.Data;
        for I := 0 to FastItem.SubItemsList.Count - 1 do
        begin
            {J :=}Item.SubItems.Add(FastItem.SubItemsList[I]);
            //Item.SubItems.Objects[ J ] := FastItem.SubItems.Objects[ I ];
        end;
        Item.ImageIndex := FastItem.ImageIndex;
        Item.StateIndex := FastItem.StateIndex;
        Item.OverlayIndex := FastItem.OverlayIndex;
        FastItem.FListItem := nil;
    end
    else
    begin
        Item.Caption := '';
        Item.SubItems.Clear;
        Item.Data := nil;
        Item.ImageIndex := -1;
    end;
end;

function TFastListview.HdrMarginLeft(ColIdx: Integer): Integer;
var J: Integer;
    S: String;
begin
    Result := 0;
    if ColIdx < 0 then Exit;
    Result := inherited HdrMarginLeft(ColIdx);
    J := IndexOfAbs(FColumnsToSort, ColIdx);
    if J >= 0 then
    begin
        S := IntToStr(J + 1);
        Result := Result + Canvas.TextWidth(S) + 6;
    end;
end;

function TFastListview.HdrMarginRight(ColIdx: Integer): Integer;
begin
    Result := 0;
    if ColIdx < 0 then Exit;
    Result := inherited HdrMarginRight(ColIdx);
    if FColumnsToFilter.IndexOf(IntToStr(ColIdx)) >= 0 then
    begin
        LoadTriangleBmp;
        if FTriangleBmp.Width < Columns[ColIdx].Width + 4 then
            Inc(Result, FTriangleBmp.Width + 4);
    end;
end;

procedure TFastListview.InvertSelection;
var I, J: Integer;
    A: array of Integer;
    LI: TListItem;
begin
    Items.BeginUpdate;
    TRY
        if (SelCount < Items.Count div 3) or
            (SelCount > Items.Count * 2 div 3) then
        begin
            if SelCount < Items.Count div 3 then
            begin
                //  :  ,      
                SetLength(A, SelCount);
                I := 0;
                LI := Selected;
                while LI <> nil do
                begin
                    A[I] := LI.Index;
                    Inc(I);
                    LI := NextSelected(LI);
                end;
                Assert(I = High(A) + 1, Err0003);
                SelectAll;
                for I := 0 to High(A) do
                    Items[A[I]].Selected := FALSE;
            end
            else
            begin
                //  :    ,    ,
                //     
                SetLength(A, Items.Count - SelCount);
                I := 0;
                for J := 0 to Items.Count - 1 do
                    if not Items[J].FSelected then
                    begin
                        A[I] := J;
                        Inc(I);
                    end;
                Assert(I = High(A) + 1, Err0003);
                //Selected := nil;
                ListView_SetItemState(Handle, -1, 0, LVIS_SELECTED);
                for I := 0 to High(A) do
                    Items[A[I]].Selected := TRUE;
            end;
        end
        else
            for I := 0 to Items.Count - 1 do
                if Items[I].Selected then
                    ListView_SetItemState(Handle, i, 0, LVIS_SELECTED)
                else
                    ListView_SetItemState(Handle, i, LVIS_SELECTED, LVIS_SELECTED);
    FINALLY
        Items.EndUpdate;
    END;
end;

procedure TFastListview.KillFilterListform(Sender: TObject);
begin
    if FFilterButtonClicked >= 0 then
        SendMessage(Handle, WM_USER_KILLFILTERLISTFORM, FFilterButtonClicked, 0);
end;

procedure TFastListview.LoadTriangleBmp;
begin
    if not Assigned(FTriangleBmp) then
    begin
        FTriangleBmp := TBitmap.Create;
        FTriangleBmp.LoadFromResourceName(hInstance, 'FAST_LV_TRIANGLE');
        FTriangleBmp.Transparent := TRUE;
        FTrianglePressedBmp := TBitmap.Create;
        FTrianglePressedBmp.LoadFromResourceName(hInstance, 'FAST_LV_TRIANGLE_PRESSED');
        FTrianglePressedBmp.Transparent := TRUE;
        FTriangleFilteredBmp := TBitmap.Create;
        FTriangleFilteredBmp.LoadFromResourceName(hInstance, 'FAST_LV_TRIANGLE_FILTERED');
        FTriangleFilteredBmp.Transparent := TRUE;
    end;
end;

function TFastListview.NextSelected(LI: TFastListItem): TFastListItem;
//var I: Integer;
begin
    Result := GetNextItem(LI, sdBelow, [isSelected]);
    {for I := LI.Index + 1 to Items.Count-1 do
    begin
      LI := Items.Items[ I ];
      if LI.FSelected then
      begin
        Result := LI;
        Exit;
      end;
    end;
    Result := nil;}
end;

function TFastListview.Providecount(C: Integer): Boolean;
begin
    Result := FALSE;
    if inherited Items.Count <> C then
    begin
        inherited Items.Count := C;
        Result := TRUE;
    end;
end;

procedure TFastListview.RangeSelected(var Msg: TWMNotify);

    procedure DoDoDo;
    var Selecting, Deselecting: Boolean;
        I: Integer;
        OD: PNMLVODStateChange;
        LV: PNMListView;
        Item: TFastListItem;
    begin
        if Application.Terminated then Exit;
        if csDestroying in ComponentState then Exit;
        if Msg.NMHdr.code = LVN_ODSTATECHANGED then
        begin
            OD := Pointer(Msg.NMHdr);
            Selecting := LongBool(OD.uNewState and LVIS_SELECTED) and
                not LongBool(OD.uOldState and LVIS_SELECTED);
            Deselecting := not LongBool(OD.uNewState and LVIS_SELECTED) and
                LongBool(OD.uOldState and LVIS_SELECTED);
            if Selecting or Deselecting then
            begin
                for I := OD.iFrom to OD.iTo do
                begin
                    Item := Items.Items[I];
                    if Selecting and Assigned(OnSelecting) then
                        if not OnSelecting(Self, Item) then
                        begin
                            Item.MirrorItem.Selected := FALSE;
                            continue;
                        end;
                    Item.FSelected := Selecting;
                end;
                if Assigned(OnSelectItem) then
                begin
                    OnSelectItem(Self, Items[OD.iFrom], Selecting);
                end;
                if Assigned(OnSelectRange) then
                    OnSelectRange(Self, OD.iFrom, OD.iTo, Selecting);
            end;
        end
        else
            if Msg.NMHdr.code = LVN_ITEMCHANGED then
            begin
                LV := Pointer(Msg.NMHdr);
                //if LV.iItem >= 0 then
                begin
                    Selecting := LongBool(LV.uNewState and LVIS_SELECTED) and
                        not LongBool(LV.uOldState and LVIS_SELECTED);
                    Deselecting := not LongBool(LV.uNewState and LVIS_SELECTED) and
                        LongBool(LV.uOldState and LVIS_SELECTED);
                    if Selecting or Deselecting then
                    begin
                        if Deselecting and (LV.iItem < 0) then
                        begin
                            if not VirtualTotally then
                                for I := 0 to Count - 1 do
                                begin
                                    Item := F_ItemsList.F_Items[I];
                                    Item.FSelected := FALSE;
                                end;
                        end;
                        if Assigned(OnSelectItem) then
                        begin
                            Item := nil;
                            if LV.iItem >= 0 then
                            begin
                                Item := Items[LV.iItem];
                                if Item = nil then
                                    if Deselecting then
                                        if FPreviouslySelectedIdx > 0 then
                                            Item := Items[FPreviouslySelectedIdx - 1];
                            end;
                            OnSelectItem(Self, Item, Selecting);
                            if lv.iItem >= 0 then
                                if Selecting then
                                    FPreviouslySelectedIdx := Item.Index + 1;
                        end;
                        if not VirtualTotally then
                            Exit;
                    end;
                end;
            end;
        inherited;
    end;
begin
    DoDoDo;
    {
    TRY
    EXCEPT
        //DoDoDo;
    END;
    }
end;

procedure TFastListview.RestoreSelection;
type
    TOldLVSelectItemEvent = procedure(Sender: TObject; Item: TOldListItem;
        Selected: Boolean) of object;
var I: Integer;
    FLI: TFastListItem;
    FSaveOnSel: TLVSelectItemEvent;
    Buf: PChar;
    NewPositions: array of Integer;
begin
    FSaveOnSel := OnSelectItem;
    Items.BeginUpdate;
    Buf := AllocMem(F_ItemsList.Count);
    SetLength(NewPositions, F_ItemsList.Count);
    TRY
        OnSelectItem := nil;
        for I := 0 to F_ItemsList.Count - 1 do
            NewPositions[FSortLineIdexes[I]] := I;
        for I := 0 to F_ItemsList.FSaveSelList.Count - 1 do
            Buf[NewPositions[Integer(F_ItemsList.FSaveSelList[I])]] := #1;
        if SelCount < F_ItemsList.Count div 2 + 2 then
        begin
            Selected := nil;
            for I := 0 to F_ItemsList.Count - 1 do
            begin
                if Buf[I] <> #0 then
                    Items[I].Selected := TRUE;
            end;
        end
        else
        begin
            SelectAll;
            for I := 0 to F_ItemsList.Count - 1 do
            begin
                if Buf[I] = #0 then
                    Items[I].Selected := FALSE;
            end;
        end;
        FLI := Selected;
        if FSaveFocusedIdx >= 0 then
            //ItemFocused := Items[ FSaveFocusedIdx ]
            ItemFocused := Items[NewPositions[FSaveFocusedIdx]]
        else
            if FLI <> nil then
                FLI.Focused := TRUE;
    FINALLY
        FreeMem(Buf);
        OnSelectItem := FSaveOnSel;
        Items.EndUpdate;
        SetLength(FSortLineIdexes, 0);
        FreeAndNil(F_ItemsList.FSaveSelList);
        SetLength(NewPositions, 0);
    END;
end;

procedure TFastListview.SaveSelection;
var I: Integer;
    FLI: TFastListItem;
begin
    FreeAndNil(F_ItemsList.FSaveSelList);
    F_ItemsList.FSaveSelList := TList.Create;
    for I := 0 to F_ItemsList.Count - 1 do
    begin
        FLI := F_ItemsList[I];
        if FLI.Selected then
            F_ItemsList.FSaveSelList.Add(Pointer(I));
        //FLI.FSaveSel := FLI.FSelected;
    end;
    FSaveFocusedIdx := -1;
    if ItemFocused <> nil then
        FSaveFocusedIdx := ItemFocused.Index;
    SetLength(FSortLineIdexes, F_ItemsList.Count);
    for I := 0 to F_ItemsList.Count - 1 do
        FSortLineIdexes[I] := I;
end;

procedure TFastListview.SelectAll;
var lvItem: TLVItem;
    I: Integer;
begin
    for I := 0 to Items.Count - 1 do
        Items[I].FSelected := TRUE;
    FillChar(lvItem, sizeof(lvItem), 0);
    lvItem.Mask := LVIF_STATE or LVIF_DI_SETITEM;
    lvItem.iItem := -1;
    lvItem.state := LVIS_SELECTED;
    lvItem.stateMask := LVIS_SELECTED;
    Inherited Perform(LVM_SETITEMSTATE, -1, Integer(@lvItem));
end;

procedure TFastListview.SetCount(C: Integer);
begin
    if FUpdateCount = 0 then
        if Providecount(C) then
            Invalidate;
end;

procedure TFastListview.SetFocusedItem(const Value: TListItem);
begin
    if Providecount(Items.Count) then
        Invalidate;
    if Value = nil then
        inherited ItemFocused := nil
    else
    begin
        Assert((Value.Index >= 0) and (Value.Index < Items.Count),
            Err0004 + ' (' +
            IntToStr(Value.Index) + ' ): ' + Value.Caption);
        inherited ItemFocused := inherited Items[Value.Index];
        Value.MakeVisible(FALSE);
    end;
end;

procedure TFastListview.SetOldItems(Idx: Integer;
    const Value: TOldListItem);
begin
    inherited Items[Idx] := Value;
end;

procedure TFastListview.WMKillFilterListForm(var M: TMessage);
begin
    if not FKillingListForm then
    begin
        if FFilterButtonClicked = M.wParam then
        begin
            if FFilterListForm <> nil then
            begin

                {if not FFilterListForm.PreventClose then
                begin
                  PostMessage( Handle, WM_USER_KILLFILTERLISTFORM, M.wParam, M.lParam );
                  Exit;
                end;}

                FKillingListForm := TRUE;
                TRY

                    FFilterButtonClicked := -1;
                    InvalidateHeader;
                    //FreeAndNil( FFilterListForm );
                    FFilterListForm.Close;
                    //SendMessage( Handle, WM_USER_SETFOCUS, 0, 0 );

                FINALLY
                    FKillingListForm := FALSE;
                END;

            end;
        end;
    end;
end;

procedure TFastListview.FilterListKeyPress(Sender: TObject; var Key: Char);
begin
    if Key = #13 then
    begin
        if FFilterListForm.lvValues.SelCount > 0 then
            ApplyFilters;
        FFilterButtonClicked := -1;
        InvalidateHeader;
        FreeAndNil(FFilterListForm);
        Key := #0;
        PostMessage(Handle, WM_USER_SETFOCUS, 0, 0);
    end;
end;

procedure TFastListview.ApplyFilters;
var K: Integer;
    SL: TStringList;
    LI: TListItem;
    CanExtend: Boolean;
begin
    if not Assigned(FFilterListForm) then Exit;
    K := FColumnsToFilter.IndexOf(IntToStr(FFilterButtonClicked));
    CanExtend := FALSE;
    if FColumnsToFilter.Objects[K] = nil then
        FColumnsToFilter.Objects[K] := TStringList.Create;
    SL := Pointer(FColumnsToFilter.Objects[K]);
    if SL.Count > 0 then
        CanExtend := TRUE;
    SL.Clear;
    if (FFilterListForm.lvValues.SelCount > 1) or
        (FFilterListForm.lvValues.Selected.Index > 0) then
    begin
        LI := FFilterListForm.lvValues.Selected;
        while LI <> nil do
        begin
            SL.Add(LI.Caption);
            LI := FFilterListForm.lvValues.NextSelected(LI);
        end;
    end
    else
        CanExtend := TRUE;
    ApplyAllFilters(CanExtend);
end;

procedure TFastListview.FilterListDblClk(Sender: TObject);
begin
    PostMessage(Handle, WM_USER_FILTERLISTDBLCLK, 0, 0);
end;

function TFastListview.OnFilterButton: Boolean;
var I: Integer;
    HdrWnd: HWnd;
    ClassNameBuf: array[0..49] of Char;
    HdHti: THDHitTestInfo;
    P, PScreen: TPoint;
    R: TRect;
begin
    Result := FALSE;

    GetCursorPos(PScreen);
    P := PScreen;
    windows.ScreenToClient(Handle, P);
    HdrWnd := childWindowFromPoint(Handle, P);

    if (HdrWnd <> 0) and (HdrWnd <> Handle) then
    begin

        GetClassName(HdrWnd, ClassNameBuf, 50);
        if ClassNameBuf = 'SysHeader32' then
        begin
            P := PScreen;
            windows.ScreenToClient(HdrWnd, P);
            HdHti.Point := P;
            SendMessage(HdrWnd, HDM_HITTEST, 0, Integer(@HdHti));
            if (HdHti.Flags = HHT_ONDIVIDER) or (HdHti.Flags = HHT_ONDIVOPEN) then
                Exit;
            I := HdHti.Item;

            if FColumnsToFilter.IndexOf(IntToStr(I)) >= 0 then
            begin
                LoadTriangleBmp;
                SendMessage(HdrWnd, HDM_GETITEMRECT, I, Integer(@R));
                if PtInRect(R, P) and
                    ((P.x > R.Right - 6 - FTriangleBmp.Width) or
                    Assigned(FFilterListForm) and FFilterListForm.Visible
                    ) then
                begin
                    Result := TRUE;
                end;
            end;

        end;

    end;
end;

procedure TFastListview.ApplyAllFilters(CanExtend: Boolean);
var I, J, K, L: Integer;
    LI: TListItem;
    SL: TStringList;
    Present, PresentAllFilters: Boolean;
begin
    Items.BeginUpdate;
    TRY

        for I := Items.Count - 1 downto 0 do
        begin
            LI := Items[I];
            for J := 0 to FColumnsToFilter.Count - 1 do
            begin
                K := StrToInt(FColumnsToFilter[J]);
                SL := Pointer(FColumnsToFilter.Objects[J]);
                if (SL <> nil) and (SL.Count > 0) then
                begin
                    Present := FALSE;
                    for L := 0 to SL.Count - 1 do
                    begin
                        if K = 0 then
                            Present := SL[L] = LI.Caption
                        else
                            if LI.SubItemsList.Count > K - 1 then
                                Present := SL[L] = LI.SubItemsList[K - 1]
                            else
                                Present := FALSE;
                        if Present then
                            break;
                    end;
                    if not Present then
                    begin
                        LI.Visible := FALSE;
                        break;
                    end;
                end;
            end;
        end;
        if CanExtend then
            for I := HiddenItemsCount - 1 downto 0 do
            begin
                LI := HiddenItems[I];
                PresentAllFilters := TRUE;
                for J := 0 to FColumnsToFilter.Count - 1 do
                begin
                    K := StrToInt(FColumnsToFilter[J]);
                    SL := Pointer(FColumnsToFilter.Objects[J]);
                    if (SL <> nil) and (SL.Count > 0) then
                    begin
                        Present := FALSE;
                        for L := 0 to SL.Count - 1 do
                        begin
                            if K = 0 then
                                Present := SL[L] = LI.Caption
                            else
                                if LI.SubItemsList.Count > K - 1 then
                                    Present := SL[L] = LI.SubItemsList[K - 1]
                                else
                                    Present := FALSE;
                            if Present then
                                break;
                        end;
                        if not Present then
                        begin
                            PresentAllFilters := FALSE;
                            break;
                        end;
                    end;
                end;
                if PresentAllFilters then
                    LI.Visible := TRUE;
            end;

        if SortType <> stNone then
            AlphaSort;
        if Assigned(OnFilterChange) then
            OnFilterChange(Self);

    FINALLY
        Items.EndUpdate;
    END;
end;

function TFastListview.GetHiddenItems(Idx: Integer): TFastListItem;
begin
    Result := FHiddenItems[Idx];
end;

function TFastListview.GetHiddenItemsCount: Integer;
begin
    Result := FHiddenItems.Count;
end;

procedure TFastListview.WMFilterListDblClk(var M: TMessage);
begin
    if not Assigned(FFilterListForm) then Exit;
    if FFilterListForm.lvValues.SelCount > 0 then
        ApplyFilters;
    FFilterButtonClicked := -1;
    InvalidateHeader;
    FreeAndNil(FFilterListForm);
    PostMessage(Handle, WM_USER_SETFOCUS, 0, 0);
end;

procedure TFastListview.HdrHitTest(var HTI: THDHitTestInfo);
var R: TRect;
    HdrWnd: HWnd;
    ClassNameBuf: array[0..49] of Char;
begin
    if FOriginalHitTestHeader then
        Exit;
    if FColumnsToFilter.IndexOf(IntToStr(HTI.Item)) >= 0 then
    begin
        HdrWnd := childWindowFromPoint(Handle, HTI.Point);
        if (HdrWnd <> 0) and (HdrWnd <> Handle) then
        begin

            GetClassName(HdrWnd, ClassNameBuf, 50);
            if ClassNameBuf = 'SysHeader32' then
            begin
                SendMessage(HdrWnd, HDM_GETITEMRECT, HTI.Item, Integer(@R));
                if HTI.Point.x < R.Right - 2 then
                begin
                    HTI.Flags := HTI.Flags and not (HHT_ONDIVIDER or HHT_ONDIVOPEN) or
                        HHT_ONHEADER;
                end;
            end;

        end;
    end;
    if LongBool(HTI.Flags and HHT_ONDIVOPEN) then
    begin
        if FHiddenColumns.IndexOf(IntToStr(HTI.Item)) >= 0 then
        begin
            HTI.Flags := HTI.Flags or HHT_ONDIVIDER;
        end;
    end;
end;

function TFastListview.ProcessWMSetCursor(Wnd: HWnd;
    lParam: Integer): Boolean;
var P, PScreen: TPoint;
    HdrWnd: Hwnd;
    ClassNameBuf: array[0..49] of Char;
    HTI: THDHitTestInfo;
    R: TRect;
begin
    Result := FALSE;
    //FColumnResizeDisabled := FALSE;
    if (HiWord(lParam) = WM_MOUSEMOVE) or
        (HiWord(lParam) = WM_LBUTTONDOWN) or
        (HiWord(lParam) = WM_LBUTTONUP) then
    begin

        GetCursorPos(PScreen);
        P := PScreen;
        windows.ScreenToClient(Handle, P);
        HdrWnd := childWindowFromPoint(Handle, P);
        if (HdrWnd <> 0) and (HdrWnd <> Handle) then
        begin

            GetClassName(HdrWnd, ClassNameBuf, 50);
            if ClassNameBuf = 'SysHeader32' then
            begin
                P := PScreen;
                windows.ScreenToClient(HdrWnd, P);
                HTI.Point := P;
                SendMessage(HdrWnd, HDM_HITTEST, 0, Integer(@HTI));
                if LongBool(HTI.Flags and (HHT_ONHEADER or HHT_ONDIVIDER or HHT_ONDIVOPEN)) then
                begin
                    if FColumnsToFilter.IndexOf(IntToStr(HTI.Item)) >= 0 then
                    begin
                        SendMessage(HdrWnd, HDM_GETITEMRECT, HTI.Item, Integer(@R));
                        if P.x < R.Right - 2 then
                        begin
                            windows.SetCursor(Screen.Cursors[crDefault]);
                            Result := TRUE;
                        end
                        else
                            if LongBool(HTI.Flags and HHT_ONDIVIDER and HHT_ONDIVOPEN) then
                            begin
                                HTI.Flags := HTI.Flags and not HHT_ONDIVOPEN;
                                //windows.SetCursor( Screen.Cursors[ crHSplit ] );
                                //FColumnResizeDisabled := TRUE;
                                Result := TRUE;
                            end;
                    end;
                    if FHiddenColumns.IndexOf(IntToStr(HTI.Item)) >= 0 then
                    begin
                        HTI.Flags := HTI.Flags and not HHT_ONDIVOPEN or HHT_ONDIVIDER;
                        Result := TRUE;
                    end;
                end;
            end;

        end;
    end;
end;

procedure TFastListview.FilterItems;
begin
    ApplyAllFilters(TRUE);
end;

procedure TFastListview.WndProc(var M: TMessage);
    function MouseInHeader: Boolean;
    var P: TPoint;
        HdrWnd: HWnd;
        R: TRect;
    begin
        Result := FALSE;
        GetCursorPos(P);
        HdrWnd := GetHeaderCtlWindow;
        if HdrWnd = 0 then Exit;
        GetWindowRect(HdrWnd, R);
        Result := PtInRect(R, P);
    end;
type
    TNMHeader = packed record
        hdr: TNMHdr;
        iItem: Integer;
        iButton: Integer;
        pitem: PHDItem;
    end;
    PNMHeader = ^TNMHeader;
var Header: PNMHeader;
    DoInherited: Boolean;
    P: TPoint;
    R: TRect;
    HdrWnd: HWnd;
    I: Integer;
begin
    {$IFDEF LOG_FLV}
    LogFileOutput('c:\log_flv.txt', 'FLV_WndProc ' + Name + ' msg:' +
        IntToStr(M.Msg));
    {$ENDIF}
    if csDesigning in ComponentState then
    begin
        inherited;
        Exit;
    end;
    if Assigned(OnMessage) then
        if OnMessage(Self, M) then Exit;
    case M.Msg of
        WM_KILLFOCUS:
            begin
                InvalidateSelected;
                if FFocusWasSet then
                    if Visible then
                        if Assigned(OnKillfocus) then
                            OnKillfocus(Self);
            end;
        WM_SETFOCUS:
            begin
                FFocusWasSet := TRUE;
                InvalidateSelected;
            end;
        WM_USER_SETFOCUS:
            begin
                if Visible and not IsClosing then
                    SetFocus;
            end;
        WM_SYSKEYDOWN:
            if (M.wParam = Integer('F')) and (GetKeyState(VK_MENU) < 0) then
            begin
                // ALT-F :      
                PostMessage(Handle, WM_USER_CALLFIRSTFILTER, 0, 0);
            end
            else
                if (M.wParam = VK_APPS) and (GetKeyState(VK_MENU) < 0) then
                begin
                    HdrWnd := GetHeaderCtlWindow;
                    if HdrWnd <> 0 then
                    begin
                        GetWindowRect(HdrWnd, R);
                        FColPopup.Popup(R.Left + (R.Right - R.Left) div 2,
                            R.Top + (R.Bottom - R.Top) div 2);
                    end;
                end;
        WM_USER_CALLFIRSTFILTER:
            begin
                for I := M.wParam to Columns.Count - 1 do
                begin
                    if ColumnVisible[I] and (FColumnsToFilter.IndexOf(IntToStr(I)) >= 0) then
                    begin
                        HdrWnd := GetHeaderCtlWindow;
                        GetWindowRect(HdrWnd, R);
                        P := R.TopLeft;
                        //Windows.ScreenToClient( Handle, P );
                        SendMessage(HdrWnd, HDM_GETITEMRECT, I, Integer(@R));
                        OffsetRect(R, P.x, P.y);
                        DoFilterButtonClick(I, R.TopLeft, R.BottomRight);
                        break;
                    end;
                end;
            end;
        WM_NOTIFY:
            begin
                Header := Pointer(M.lParam);
                case Header.hdr.code of
                    HDN_BEGINTRACK, HDN_BEGINTRACKW:
                        begin
                            DoInherited := TRUE;
                            while FHiddenColumns.IndexOf(IntToStr(Header.iItem)) >= 0 do
                            begin
                                FHdrTrkIdx := Header.iItem;
                                Dec(Header.iItem);
                                DoInherited := FALSE;
                                if Header.iItem < 0 then
                                begin
                                    M.Result := 1;
                                    Exit;
                                end;
                            end;
                            if not DoInherited then Exit;
                        end;
                    HDN_ENDTRACK, HDN_ENDTRACKW:
                        FHdrTrkIdx := -1;
                    HDN_ENDDRAG:
                        FEndDragTime := GetTickCount;
                    HDN_ITEMCHANGING, HDN_ITEMCHANGINGW, HDN_TRACK, HDN_TRACKW:
                        if not fDisabledItemChanging then
                        begin
                            DoInherited := TRUE;
                            while FHiddenColumns.IndexOf(IntToStr(Header.iItem)) >= 0 do
                            begin
                                DoInherited := FALSE;
                                Columns[Header.iItem].Width := 0;
                                Dec(Header.iItem);
                                if Header.iItem < 0 then
                                begin
                                    M.Result := 1;
                                    DoInherited := TRUE;
                                    break;
                                end;
                            end;
                            if not DoInherited then
                                if (Header.iItem >= 0) and (Header.iItem = FHdrTrkIdx) then
                                begin
                                    HdrWnd := GetHeaderCtlWindow;
                                    SendMessage(HdrWnd, HDM_GETITEMRECT, Header.iItem, Integer(@R));
                                    GetCursorPos(P);
                                    Windows.ScreenToClient(HdrWnd, P);
                                    FDisabledItemChanging := TRUE;
                                    if not LBtnPressed then //+++++++++++++++++++++++++++++++++++++++
                                    TRY
                                        Windows.GetClientRect(HdrWnd, R);
                                        if PtInRect(R, P) then
                                            if GetTickCount - FEndDragTime > 350 then
                                                Columns[Header.iItem].Width := P.x - R.Left;
                                    FINALLY
                                        FDisabledItemChanging := FALSE;
                                    END;
                                end;
                            if not DoInherited then
                            begin
                                M.Result := 1;
                                Exit;
                            end;
                        end;
                end;
            end;
        WM_PAINT: //, WM_NCPAINT:
            if FUpdateCount > 0 then
            begin
                M.Result := 0;
                ValidateRect(Handle, nil);
                Exit;
            end;
        WM_ERASEBKGND:
            if FUpdateCount > 0 then
            begin
                M.Result := 1;
                Exit;
            end;
        WM_MOUSEFIRST..WM_MOUSELAST, WM_SETCURSOR, WM_NCHITTEST:
            if (csDesigning in ComponentState) and MouseInHeader then
            begin
                //M.Result := DefWindowProc( Handle, M.Msg, M.WParam, M.LParam );
                M.Result := CallWindowProc(DefWndProc, Handle, M.Msg, M.WParam, M.LParam);
                {P.x := SmallInt( LoWord( M.LParam ) );
                P.y := SmallInt( HiWord( M.LParam ) );
                DefWindowProc( GetHeaderCtlWindow, M.Msg, M.WParam,
                               (P.x and $FFFF) or (P.y shl 16) );}
                //Dispatch( M );
                Exit;
            end;
    end;
    inherited;
    {$IFDEF LOG_FLV}
    LogFileOutput('c:\log_flv.txt', 'FLV_WndProc ' + Name + ' - exit');
    {$ENDIF}
end;

function TFastListview.FilterListSelecting(Sender: TObject;
    Item: TFastListItem): Boolean;
begin
    Result := Item.FontColor = clNone;
end;

procedure TFastListview.ClearFilters;
var SL: TStringList;
    I: Integer;
begin
    for I := 0 to FColumnsToFilter.Count - 1 do
    begin
        SL := TStringList(FColumnsToFilter.Objects[I]);
        if SL <> nil then
            SL.Clear;
    end;
    FilterItems;
end;

procedure TFastListview.DeleteHiddenItems;
var I: Integer;
begin
    for I := FHiddenItems.Count - 1 downto 0 do
        TFastListItem(FHiddenItems[I]).Free;
    FHiddenItems.Clear;
end;

function TFastListview.ItemIndent(Item: TOldListItem): Integer;
var FastItem: TFastListItem;
begin
    FastItem := Items[Item.Index];
    if FastItem <> nil then
        Result := FastItem.Indent
    else
        Result := Item.Indent;
end;

procedure TFastListview.SetSortColumnAttributes(
    const ColumnIndeces: array of Integer; Attrs: TColumnSortAttrs);
var I, Col: Integer;
    Dw: DWORD;
begin
    for I := 0 to High(ColumnIndeces) do
    begin
        Col := ColumnIndeces[I];
        Assert((Col >= 0) and (Col < Columns.Count), Err0005 + ' ( ' +
            IntToStr(Col) + ' )');
        while Col >= FColumnSortAttrs.Count do
            FColumnSortAttrs.Add(nil);
        Dw := PDWORD(@Attrs)^;
        FColumnSortAttrs.Items[Col] := Pointer(Dw);
    end;
end;

function TFastListview.GetItemData(Idx: Integer): Pointer;
var Item: TFastListItem;
begin
    Item := F_ItemsList.Items[Idx];
    Result := Item.FData;
end;

procedure TFastListview.setItemData(Idx: Integer; const Value: Pointer);
var Item: TFastListItem;
begin
    Item := F_ItemsList.Items[Idx];
    Item.FData := Value;
end;

function TFastListview.HasEmptyCaptions: Boolean;
var I: Integer;
begin
    for I := Items.Count - 1 downto 0 do
    begin
        if Items[I].FCaption = '' then
        begin
            Result := TRUE;
            Exit;
        end;
    end;
    Result := FALSE;
end;

procedure TFastListview.miAddColumnClick(Sender: TObject);
var I: Integer;
begin
    if NoMultiColumnSort then
    begin
        {$IFDEF RUSSIAN_LANGUAGE}
        ShowMessage('      .');
        {$ELSE}
        ShowMessage('Multiple sorting is not available for this view');
        {$ENDIF}
        Exit;
    end;
    I := IndexOfAbs(FColumnsToSort, FColumnclicked);
    if I < 0 then
        FColumnsToSort.Add(Pointer(FColumnclicked))
    else
        if FColumnsToSort.Count > 0 then
            FColumnsToSort.Move(I, 1);

    InvalidateHeader;
    if SortType = stNone then
        SortType := stBoth
    else
        AlphaSort1(FALSE);
end;

procedure TFastListview.miAscendingClick(Sender: TObject);
var I: Integer;
begin
    I := IndexOfAbs(FColumnsToSort, FColumnclicked);
    if I < 0 then
        if FColumnsToSort.Count < 2 then
            I := FColumnsToSort.Add(Pointer(FColumnclicked))
        else
        begin
            I := 1;
            FColumnsToSort.Insert(1, Pointer(FColumnclicked));
        end;
    FColumnsToSort[I] := Pointer(DWORD(FColumnsToSort[I]) and $FFFF);

    InvalidateHeader;
    if SortType = stNone then
        SortType := stBoth
    else
        AlphaSort1(FALSE);
end;

procedure TFastListview.miDescendingClick(Sender: TObject);
var I: Integer;
begin
    I := IndexOfAbs(FColumnsToSort, FColumnclicked);
    if I < 0 then
        if FColumnsToSort.Count < 2 then
            I := FColumnsToSort.Add(Pointer(FColumnclicked))
        else
        begin
            I := 1;
            FColumnsToSort.Insert(1, Pointer(FColumnclicked));
        end;
    FColumnsToSort[I] := Pointer(DWORD(FColumnsToSort[I]) or $80000000);

    InvalidateHeader;
    if SortType = stNone then
        SortType := stBoth
    else
        AlphaSort1(FALSE);
end;

procedure TFastListview.miIncreaseOrderClick(Sender: TObject);
var I: Integer;
begin
    I := IndexOfAbs(FColumnsToSort, FColumnclicked);
    if I < 0 then
        FColumnsToSort.Add(Pointer(FColumnclicked))
    else
        if FColumnsToSort.Count > I + 1 then
            FColumnsToSort.Move(I, I + 1);

    InvalidateHeader;
    if SortType = stNone then
        SortType := stBoth
    else
        AlphaSort1(FALSE);
end;

procedure TFastListview.miDecreaseOrderClick(Sender: TObject);
var I: Integer;
begin
    I := IndexOfAbs(FColumnsToSort, FColumnclicked);
    if I < 0 then
        FColumnsToSort.Add(Pointer(FColumnclicked))
    else
        if (FColumnsToSort.Count > 1) and (I > 0) then
            FColumnsToSort.Move(I, I - 1);

    InvalidateHeader;
    if SortType = stNone then
        SortType := stBoth
    else
        AlphaSort1(FALSE);
end;

procedure TFastListview.miRemoveColumnClick(Sender: TObject);
var I: Integer;
begin
    I := IndexOfAbs(FColumnsToSort, FColumnclicked);
    if I >= 0 then
        FColumnsToSort.Delete(I);

    InvalidateHeader;
    if FColumnsToSort.Count = 0 then Exit;
    if SortType = stNone then
        SortType := stBoth
    else
        AlphaSort1(FALSE);
end;

procedure TFastListview.miFilterClick(Sender: TObject);
var I: Integer;
begin
    I := FColumnsToFilter.IndexOf(IntToStr(FColumnclicked));
    if I < 0 then Exit;
    FilterButtonClick(TRUE, FColumnclicked);
    InvalidateHeader;
end;

procedure TFastListview.miResetFilterClick(Sender: TObject);
var I: Integer;
    SL: TStringList;
begin
    I := FColumnsToFilter.IndexOf(IntToStr(FColumnclicked));
    if I < 0 then Exit;
    SL := Pointer(FColumnsToFilter.Objects[I]);
    FColumnsToFilter.Objects[I] := nil;
    SL.Free;
    InvalidateHeader;
    ApplyAllFilters(TRUE);
end;

procedure TFastListview.miResetAllFiltersClick(Sender: TObject);
var I: Integer;
    SL: TStringList;
begin
    for I := 0 to FColumnsToFilter.Count - 1 do
    begin
        SL := Pointer(FColumnsToFilter.Objects[I]);
        FColumnsToFilter.Objects[I] := nil;
        SL.Free;
    end;
    InvalidateHeader;
    ApplyAllFilters(TRUE);
end;

procedure TFastListview.miSortOneColumnClick(Sender: TObject);
begin
    FColumnsToSort.Clear;
    FColumnsToSort.Add(Pointer(FColumnclicked));

    InvalidateHeader;
    {if SortType = stNone then
      SortType := stBoth
    else}
    AlphaSort1(FALSE);
end;

procedure TFastListview.miHideColumnClick(Sender: TObject);
begin
    ColumnVisible[FColumnclicked] := FALSE;
end;

procedure TFastListview.miShowAllColumnsClick(Sender: TObject);
var I: Integer;
begin
    for I := Columns.Count - 1 downto 0 do
        ColumnVisible[I] := TRUE;
end;

function TFastListview.HdrRightClick(ColIdx: Integer;
    Down: Boolean): Boolean;
var P: TPoint;
begin
    Result := TRUE;
    GetCursorPos(P);

    if ShowColumnHeaders and (GetKeyState(VK_CONTROL) < 0) and
        (GetKeyState(VK_RBUTTON) < 0) then
        ColumnVisible[ColIdx] := False
    else
        if GetKeyState(VK_RBUTTON) < 0 then
        begin
            FColumnclicked := ColIdx;
            FColPopup.Popup(P.x, P.y);
        end;
end;

procedure TFastListview.ColPopupOnPopup(Sender: TObject);
var I, colIdx: Integer;
begin
    miFilterMenu.Visible := FColumnsToFilter.Count > 0;
    I := FColumnsToFilter.IndexOf(IntToStr(FColumnclicked));
    miFilter.Enabled := I >= 0;
    miResetFilter.Enabled := (I >= 0) and (TStringList(FColumnsToFilter.Objects[I]) <> nil) and
        (TStringList(FColumnsToFilter.Objects[I]).Count > 0);
    miResetAllFilters.Enabled := FALSE;
    for I := 0 to FColumnsToFilter.Count - 1 do
    begin
        if TStringList(FColumnsToFilter.Objects[I]) <> nil then
            if TStringList(FColumnsToFilter.Objects[I]).Count > 0 then
            begin
                miResetAllFilters.Enabled := TRUE;
                break;
            end;
    end;

    I := IndexOfAbs(FColumnsToSort, FColumnclicked);
    miAscending.Checked := (I >= 0) and not LongBool(DWORD(FColumnsToSort[I]) and $80000000);
    miDescending.Checked := (I >= 0) and LongBool(DWORD(FColumnsToSort[I]) and $80000000);

    if SortType = stNone then
    begin
        miSortOneColumn.Enabled := FALSE;
        miAddcolumn.Enabled := FALSE;
        miRemoveColumn.Enabled := FALSE;
        miAscending.Enabled := FALSE;
        miDescending.Enabled := FALSE;
    end
    else
    begin
        miSortOneColumn.Enabled := TRUE;
        miRemoveColumn.Enabled := I >= 0;
        miAddColumn.Enabled := I < 0;
        miAscending.Enabled := TRUE;
        miDescending.Enabled := TRUE;
    end;

    miIncreaseOrder.Enabled := (I >= 0) and (I < FColumnsToSort.Count - 1);
    miDecreaseOrder.Enabled := I > 0;

    miHideColumn.Visible := AllowColumnHide;
    miShowAllColumns.Visible := AllowColumnHide;

    miHideColumn.Enabled := FHiddenColumns.Count < Columns.Count - 1;
    for I := miHideMenu.Count - 2 downto 1 do
    begin
        colIdx := miHideMenu.Items[I].Tag;
        if colIdx >= 0 then
            if FHiddenColumns.IndexOf(IntToStr(colIdx)) < 0 then
                miHideMenu.Items[I].Free;
    end;
    miHideShowSep1.Visible := miHideMenu.Count > 4;
    miHideShowSep2.Visible := miHideMenu.Count > 4;

    //miHideColumn.Enabled := TRUE;
    miShowAllColumns.Enabled := FHiddenColumns.Count > 0;

    miRestoreOrder.Enabled := FALSE;
    for I := 0 to Columns.Count - 1 do
        if ColumnOrder[I] <> I then
        begin
            miRestoreOrder.Enabled := TRUE;
            break;
        end;
end;

procedure TFastListview.CreateWnd;
begin
    {$IFDEF LOG_FLV}
    LogFileOutput('c:\log_flv.txt', 'FLV_CreateWnd ' + Name);
    {$ENDIF}
    inherited;
    if not (csDesigning in ComponentState) then
        SetWindowLong(Handle, GWL_STYLE, GetWindowLong(Handle, GWL_STYLE) or HDS_FULLDRAG);
    {$IFDEF LOG_FLV}
    LogFileOutput('c:\log_flv.txt', 'FLV_CreateWnd exit');
    {$ENDIF}
end;

procedure TFastListview.miShow1Columnclick(Sender: TObject);
var MI: TMenuItem;
    ColIdx: Integer;
    I: Integer;
begin
    MI := Sender as TMenuItem;
    ColIdx := MI.Tag;
    I := FHiddenColumns.IndexOf(IntToStr(ColIdx));
    if I < 0 then Exit;
    FDisabledItemChanging := TRUE;
    TRY
        Columns[ColIdx].Width := Integer(FHiddenColumns.Objects[I]);
    FINALLY
        FDisabledItemChanging := FALSE;
    END;
    FHiddenColumns.Delete(I);
end;

function TFastListview.GetColumnVisible(colIdx: Integer): Boolean;
begin
    Result := TRUE;
    if FHiddenColumns <> nil then
        Result := FHiddenColumns.IndexOf(IntToStr(colIdx)) < 0;
end;

procedure TFastListview.SetColumnVisible(colIdx: Integer;
    const Value: Boolean);
var I: Integer;
    MI: TMenuItem;
    S: String;
begin
    if ColumnVisible[colIdx] = Value then Exit;
    FDisabledItemChanging := TRUE;
    TRY
        if Value then
        begin
            {$IFDEF CHK_COLUMN_VISIBLE}
            LogFileOutput('c:\column_vis.txt', 'Setting ColumnVisible[' +
                IntToStr(colIdx) + '] = TRUE');
            {$ENDIF}
            I := FHiddenColumns.IndexOf(IntToStr(colIdx));
            if I >= 0 then
            begin
                Columns[colIdx].Width := Integer(FHiddenColumns.Objects[I]);
                FHiddenColumns.Delete(I);
            end;
            for I := miHideMenu.Count - 1 downto 1 do
            begin
                MI := miHideMenu.Items[I];
                if MI.Tag >= 0 then
                    if FHiddenColumns.IndexOf(IntToStr(MI.Tag)) < 0 then
                        MI.Free;
            end;
            {$IFDEF CHK_COLUMN_VISIBLE}
            LogFileOutput('c:\column_vis.txt', 'Setting ColumnVisible[' +
                IntToStr(colIdx) + '] = TRUE -- finished OK');
            {$ENDIF}
        end
        else
            if AllowColumnHide then
            begin
                {$IFDEF CHK_COLUMN_VISIBLE}
                LogFileOutput('c:\column_vis.txt', 'Setting ColumnVisible[' +
                    IntToStr(colIdx) + '] = FALSE');
                {$ENDIF}
                if FHiddenColumns.Count >= Columns.Count - 1 then Exit;
                I := FHiddenColumns.Add(IntToStr(colIdx));
                FHiddenColumns.Objects[I] := Pointer(Columns[colIdx].Width);
                Columns[colIdx].Width := 0;
                {$IFDEF CHK_COLUMN_VISIBLE}
                LogFileOutput('c:\column_vis.txt', 'Setting ColumnVisible[' +
                    IntToStr(colIdx) + '] = FALSE -- getting Caption');
                {$ENDIF}
                S := Columns[ColIdx].Caption;
                if (MultiHeader <> '') and (pos(MultiHeader, S) > 0) then
                    S := Copy(S, 1, pos(MultiHeader, S) - 1);
                {$IFDEF CHK_COLUMN_VISIBLE}
                LogFileOutput('c:\column_vis.txt', 'Setting ColumnVisible[' +
                    IntToStr(colIdx) + '] = FALSE -- work with menu');
                {$ENDIF}
                MI := TMenuItem.Create(Self);
                MI.Caption := '[ ' + Trim(S) + ' ]';
                MI.Tag := ColIdx;
                MI.OnClick := miShow1ColumnClick;
                {$IFDEF CHK_COLUMN_VISIBLE}
                LogFileOutput('c:\column_vis.txt', 'Setting ColumnVisible[' +
                    IntToStr(colIdx) + '] = FALSE -- insert menu item, miHideMenu = ' +
                    IntToStr(Integer(miHideMenu)));
                {$ENDIF}
                if Assigned(miHideMenu) then
                begin
                    miHideMenu.Insert(miHideMenu.Count - 4, MI);
                    {$IFDEF CHK_COLUMN_VISIBLE}
                    LogFileOutput('c:\column_vis.txt', 'Setting ColumnVisible[' +
                        IntToStr(colIdx) + '] = FALSE -- post message WM_USER+66');
                    {$ENDIF}
                    PostMessage(Handle, WM_USER + 66, -1, 0);
                end;
                {$IFDEF CHK_COLUMN_VISIBLE}
                LogFileOutput('c:\column_vis.txt', 'Setting ColumnVisible[' +
                    IntToStr(colIdx) + '] = FALSE -- finished OK');
                {$ENDIF}
            end;
    FINALLY
        FDisabledItemChanging := FALSE;
    END;
end;

procedure TFastListview.CustDrawItem(Sender: TCustomListView;
    Item: TOldListItem; State: TCustomDrawState; var DefaultDraw: Boolean);
var HdrWnd: HWnd;
    R, HR: TRect;
    I, HdrOffsetX: Integer;
    FastItem: TFastListItem;
    S: String;
    FrColor: TColor;
    maxX, minX: Integer;

    procedure DoGrids(var R: TRect; LeftOffset: Boolean);
    begin
        if not GridLines then Exit;
        if LeftOffset then
        begin
            //inc( R.Left );
        end;
        Canvas.Pen.Style := psSolid;
        Canvas.Pen.Color := clBtnFace;
        Canvas.Pen.Width := 0;
        if LeftOffset then
        begin
            Canvas.MoveTo(R.Left, R.Top);
            Canvas.LineTo(R.Left, R.Bottom - 1);
        end
        else
        begin
            Canvas.MoveTo(R.Left, R.Bottom - 1);
        end;
        Canvas.LineTo(R.Right, R.Bottom - 1);

        Canvas.MoveTo(R.Right + 1, R.Bottom - 1);
        Canvas.LineTo(R.Right + 1, R.Top);

        if R.Left < minX then
            minX := R.Left;
        if R.Right > maxX then
            maxX := R.Right;
        //dec( R.Right );
        //dec( R.Bottom );
    end;
var SaveFont: TFont;
    SaveBrush: TBrush;
begin
    DefaultDraw := FALSE;
    FastItem := Items[Item.Index];
    R := Item.DisplayRect(drBounds);
    Canvas.Font.Assign(Font);
    Canvas.Brush.Color := PswColor;
    Canvas.Brush.Style := bsSolid;
    TRY
        if (FastItem <> nil) and (FastItem.BackColor <> clNone) then
            Canvas.Brush.Color := FastItem.BackColor;
    EXCEPT
        Exit;
        asm
      nop
        end;
    END;
    if (FastItem <> nil) and (FastItem.FontColor <> clNone) then
        Canvas.Font.Color := FastItem.FontColor;
    if (FastItem <> nil) and FastItem.FCustomFontStyle then
        Canvas.Font.Style := FastItem.FontStyle;
    if Item.Selected then
    begin
        Canvas.Font.Color := SelColorText;
        Canvas.Brush.Color := SelColor;
        if (FastItem <> nil) and (FastItem.BackColorSelected <> clNone) then
            Canvas.Brush.Color := FastItem.BackColorSelected;
        if BoldIfFocused and (FastItem <> nil) and FastItem.Focused then
            Canvas.Font.Style := Canvas.Font.Style + [fsBold];
        if not Focused then
        begin
            Canvas.Font.Color := SelColorTextNoFocus;
            Canvas.Brush.Color := SelColorNoFocus;
        end;
        if (FastItem <> nil) and (FastItem.FontColorSelected <> clNone) then
            Canvas.Font.Color := FastItem.FontColorSelected;
    end;
    HdrWnd := GetHeaderCtlWindow;
    GetWindowRect(HdrWnd, HR);
    HR.TopLeft := ScreenToClient(HR.TopLeft);
    HdrOffsetX := HR.Left;
    if ViewStyle = vsReport then
    begin
        SendMessage(HdrWnd, HDM_GETITEMRECT, 0, Integer(@HR));
        R.Left := HR.Left + HdrOffsetX;
        R.Right := HR.Right + HdrOffsetX;
    end;
    if (FastItem <> nil) and (FastItem.Indent > 0) then
    begin
        if (SmallImages = nil) and (StateImages = nil) then I := 20
        else
        begin
            I := 4;
            if SmallImages <> nil then
                I := I + SmallImages.Width;
            if StateImages <> nil then
                I := I + StateImages.Width;
        end;
        HR := R;
        R.Left := R.Left + I * FastItem.Indent;
        HR.Right := R.Left;
        Canvas.FillRect(HR);
    end;
    if StateImages <> nil then
    begin
        HR := R;
        R.Left := R.Left + StateImages.Width + 2;
        HR.Right := R.Left;
        Canvas.FillRect(HR);
        if Item.StateIndex >= 0 then
        begin
            I := (HR.Bottom - HR.Top - StateImages.Height) div 2;
            if Canvas.Handle = ForceCanvas then
            begin
                SaveFont := TFont.Create;
                TRY
                    SaveFont.Assign(Canvas.Font);
                    Canvas.Font.Color := clRed;
                    Canvas.Font.Name := 'Arial Narrow';
                    Canvas.Font.Size := 9;
                    Canvas.Refresh;
                    Canvas.TextOut(HR.Left, HR.Top + I - 1, '<' + IntToStr(Item.StateIndex) + '>')
                FINALLY
                    Canvas.Font.Assign(SaveFont);
                    SaveFont.Free;
                    Canvas.Refresh;
                END;
            end
            else
                StateImages.Draw(Canvas, HR.Left + 1, HR.Top + I, Item.StateIndex);
        end;
    end;
    if SmallImages <> nil then
    begin
        HR := R;
        R.Left := R.Left + SmallImages.Width + 2;
        HR.Right := R.Left;
        Canvas.FillRect(HR);
        if Item.ImageIndex >= 0 then
        begin
            I := (HR.Bottom - HR.Top - SmallImages.Height) div 2;
            if Canvas.Handle = ForceCanvas then
            begin
                SaveFont := TFont.Create;
                SaveBrush := TBrush.Create;
                TRY
                    SaveFont.Assign(Canvas.Font);
                    Canvas.Font.Color := clRed;
                    if Item.ImageIndex = 0 then
                        Canvas.Font.Color := clBlue;
                    Canvas.Font.Name := 'Arial Narrow';
                    Canvas.Font.Size := 9;
                    Canvas.Brush.Style := bsSolid;
                    Canvas.Brush.Color := clYellow;
                    Canvas.Refresh;
                    Canvas.TextOut(HR.Left, HR.Top + I - 1, '<' + IntToStr(Item.ImageIndex) + '>')
                FINALLY
                    Canvas.Font.Assign(SaveFont);
                    Canvas.Brush.Assign(SaveBrush);
                    SaveFont.Free;
                    SaveBrush.Free;
                    Canvas.Refresh;
                END;
            end
            else
                SmallImages.Draw(Canvas, HR.Left + 1, HR.Top + I, Item.ImageIndex); //uuu
        end;
    end;
    if (FastItem <> nil) then
        Custdraw(R, FastItem, FastItem.Caption, -1, State, ItmAlign);
    maxX := 0;
    minX := MaxInt;
    DoGrids(R, FALSE {TRUE});
    if ViewStyle = vsReport then
    begin
        for I := 0 to Columns.Count - 2 do
        begin
            S := '';
            if VirtualTotally then
            begin
                if Assigned(OnDataCell) then
                    OnDataCell(Self, FastItem.Index, I + 1, S);
            end
            else
            begin
                if (FastItem <> nil) and (FastItem.SubItemsList.Count > I) then
                    S := FastItem.SubItemsList[I];
            end;
            SendMessage(HdrWnd, HDM_GETITEMRECT, I + 1, Integer(@HR));
            R.Left := HR.Left + HdrOffsetX;
            R.Right := HR.Right + HdrOffsetX;
            CustDraw(R, FastItem, S, I, State, Columns[I + 1].Alignment);
            DoGrids(R, TRUE);
        end;
    end;
    if maxX > 0 then
    begin
        Canvas.Pen.Style := psSolid;
        Canvas.Pen.Color := clBtnFace;
        Canvas.Pen.Width := 0;
        Canvas.MoveTo(maxX + 1, R.Top);
        Canvas.LineTo(maxX + 1, R.Bottom - 1);
    end;
    if (minX > 0) and (minX < MaxInt) then
    begin
        Canvas.Pen.Style := psSolid;
        Canvas.Pen.Color := clBtnFace;
        Canvas.Pen.Width := 0;
        {Canvas.MoveTo( 0, R.Top );
        Canvas.LineTo( minX, R.Top );}
        Canvas.MoveTo(0, R.Bottom - 1);
        Canvas.LineTo(minX, R.Bottom - 1);
    end;
    if (FastItem <> nil) and FastItem.Focused then
    begin
        R := Item.DisplayRect(drSelectBounds);
        if GridLines then
            Dec(R.Bottom);
        FrColor := $FFFFFF xor ColorToRGB(Canvas.Brush.Color);
        Frame3D(Canvas, R, FrColor, FrColor, 1);
    end;
end;

procedure TFastListview.CustDraw(Rect: TRect; Item: TFastListItem;
    Capt: string; SubItemIdx: Integer; Stat: TCustomDrawState;
    Alig: TAlignment);
var f: integer;
    R: TRect;
    T: string;
    DfltDraw: Boolean;
    OldBkColor, OldFwColor, BkColor, FwColor: TColor;
    OldFontStyle, FontStyle: TFontStyles;
    RestoreColors: Boolean;
begin
    RestoreColors := FALSE;
    OldBkColor := Canvas.Brush.Color;
    OldFwColor := Canvas.Font.Color;
    if Assigned(OnGetColors) then
    begin
        BkColor := OldBkColor;
        FwColor := OldFwColor;
        FontStyle := Canvas.Font.Style;
        OldFontStyle := FontStyle;
        RestoreColors := TRUE;
        OnGetColors(Self, FwColor, BkColor, FontStyle, Item, SubItemIdx);
        if BkColor = clNone then
            Canvas.Brush.Style := bsClear
        else
            Canvas.Brush.Color := BkColor;
        Canvas.Font.Color := FwColor;
        Canvas.Font.Style := FontStyle;
        Canvas.Refresh;
    end;
    TRY

        if Assigned(FOnCustomDrawSubItem) then
        begin
            DfltDraw := FALSE;
            FOnCustomDrawSubItem(Self, Item, SubItemIdx, Stat, DfltDraw);
            if not DfltDraw then Exit;
        end;
        f := DT_NOPREFIX;
        R := Rect;
        if Alig = taLeftJustify then begin
            f := f or DT_LEFT;
        end else
            if Alig = taRightJustify then begin
                f := f or DT_RIGHT;
                R.Right := R.Right - 5;
            end else
                if Alig = taCenter then begin
                    f := f or DT_CENTER;
                end;

        Canvas.FillRect(Rect);

        R.Left := R.Left + 4;
        T := Capt;
        if T <> '' then begin
            if T[Length(T)] = #0 then begin
                T := Copy(T, 1, Length(T) - 1);
            end;
        end;
        DrawText(Canvas.Handle, PChar(T), Length(T), R,
            f or DT_END_ELLIPSIS or DT_SINGLELINE);

    FINALLY
        if RestoreColors then
        begin
            Canvas.Font.Color := OldFwColor;
            Canvas.Brush.Style := bsSolid;
            Canvas.Brush.Color := OldBkColor;
            Canvas.Font.Style := OldFontStyle;
        end;
    END;
end;

procedure TFastListview.InvalidateSelected;
var I: Integer;
    LI: TFastListItem;
    R: TRect;
begin
    if TopItem = nil then Exit;
    I := TopItem.Index;
    for I := I to I + VisibleRowCount do
    begin
        if I >= Items.Count then break;
        LI := Items[I];
        if LI.FSelected then
        begin
            R := LI.MirrorItem.DisplayRect(drBounds);
            Windows.InvalidateRect(Handle, @R, TRUE);
        end;
    end;
end;

procedure TFastListview.DoAutoExpand(ExcludeColumn: Integer);
begin
    if AutoExpandColumn < 0 then Exit;
    if FHiddenColumns.IndexOf(IntToStr(AutoExpandColumn)) >= 0 then Exit;
    if FDisableAutoExpand then Exit;
    inherited;
end;

procedure TFastListview.SetAllowColumnDrag(const Value: Boolean);
begin
    FAllowColumnDrag := Value;
    if Value then
    begin
        DragMode := dmAutomatic;
        FullDrag := TRUE;
        //DragKind := dkDrag;
    end
    else
    begin
        DragMode := dmManual;
        FullDrag := FALSE;
    end;
end;

procedure TFastListview.miRestoreOrderClick(Sender: TObject);
var I: Integer;
begin
    for I := 0 to Columns.Count - 1 do
        ColumnOrder[I] := I;
end;

procedure TFastListview.miSizeProportionalClick(Sender: TObject);
var I: Integer;
    WasAutoExpand: Integer;
    WasFocused: Boolean;
begin
    WasFocused := GetFocus = Handle;
    Visible := FALSE;
    WasAutoExpand := AutoExpandColumn;
    TRY
        AutoExpandColumn := -1;
        for I := 0 to Columns.Count - 1 do
        begin
            if ColumnVisible[I] then
                Columns[I].Width := ClientWidth div (Columns.Count - FHiddenColumns.Count);
        end;
    FINALLY
        AutoExpandColumn := WasAutoExpand;
        Visible := TRUE;
        if WasFocused and not Focused then
            PostMessage(Handle, WM_USER_SETFOCUS, 0, 0);
    END;
end;

function TFastListview.EnabledAutoexpand: Boolean;
begin
    Result := not Assigned(FFilterListForm);
end;

function TFastListview.GetIsColumnFiltered(colIdx: Integer): Boolean;
begin
    Result := FColumnsToFilter.IndexOf(IntToStr(colIdx)) >= 0;
end;

procedure TFastListview.Loaded;
begin
    TRY
        inherited;
        if not (csDesigning in ComponentState) then
        begin
            if FRegistrySaveRestore and (RegistryKeyPath <> '') and (Name <> '') and
                (ParentForm <> nil) and (Parentform.Name <> '') then
                RestoreSettingsFromRegistry;
        end;
        if (MultiHeader <> '') and not (csDesigning in ComponentState) then
        begin
            RecalcHeaderHeight;
        end;
        if F_MultiSelect and (csDesigning in ComponentState) then
            inherited MultiSelect := F_MultiSelect;
    EXCEPT
        //
    END;
end;

function TFastListview.ParentForm: TForm;
var Ctl: TWinControl;
begin
    Ctl := Parent;
    while (Ctl <> nil) and not (Ctl is TForm) do
        Ctl := Ctl.Parent;
    Result := Ctl as TForm;
end;

procedure TFastListview.RestoreSettingsFromRegistry;
var R: TRegistry;
    DesktopSettings: String;
    S: String;
    I, J: Integer;
    SL: TStringList;
begin
    if Name = '' then Exit;
    FParentformName := '';
    if ParentForm = nil then Exit;
    FParentformName := ParentForm.Name;
    if FParentFormName = '' then Exit;
    R := TRegistry.Create;
    SL := TStringList.Create;
    FDisableAutoExpand := TRUE;
    TRY
        DesktopSettings := IntToStr(Screen.Width) + 'x' + IntToStr(Screen.Height);
        R.OpenKeyReadOnly(RegistryKeyPath + DesktopSettings);
        if R.ValueExists(FParentformName + '.' + Name) then
        begin
            S := R.ReadString(FParentformName + '.' + Name);
            for I := 1 to Length(S) do
                if S[I] = ';' then
                    S[I] := #13;
            SL.Text := S;
            for I := SL.Count - 1 downto 0 do
            begin
                if I >= Columns.Count then continue;
                S := SL[I] + '                                        ';
                TRY
                    J := StrToInt(Trim(Copy(S, 2, 5)));
                    if J > 0 then Columns[I].Width := J;
                    ColumnOrder[I] := StrToInt(Trim(Copy(S, 7, 3)));
                EXCEPT
                END;
            end;
            for I := SL.Count - 1 downto 0 do
            begin
                if I >= Columns.Count then continue;
                S := SL[I] + '                                        ';
                if S[1] = 'N' then
                    ColumnVisible[I] := FALSE;
            end;
        end;
    FINALLY
        R.Free;
        SL.Free;
        FDisableAutoExpand := FALSE;
    END;
end;

procedure TFastListview.SaveSettingsInRegistry;
var R: TRegistry;
    DesktopSettings: String;
    S: String;
    I, J: Integer;
    SL: TStringList;
begin
    if not FRegistrySaveRestore then Exit;
    if FParentFormName = '' then Exit;
    R := TRegistry.Create;
    SL := TStringList.Create;
    TRY
        DesktopSettings := IntToStr(Screen.Width) + 'x' + IntToStr(Screen.Height);
        R.Access := KEY_WRITE;
        R.OpenKey(RegistryKeyPath + DesktopSettings, TRUE);
        S := '';
        for I := 0 to Columns.Count - 1 do
        begin
            if ColumnVisible[I] then
                S := S + 'V'
            else
                S := S + 'N';
            if ColumnVisible[I] then
                S := S + Format('%.05d', [ColumnWidth[I]])
            else
            begin
                J := FHiddenColumns.IndexOf(IntToStr(I));
                if J >= 0 then
                    S := S + Format('%.05d', [Integer(FHiddenColumns.Objects[J])])
            end;
            S := S + Format('%.03d', [ColumnOrder[I]]) + ';';
        end;
        R.WriteString(FParentformName + '.' + Name, S);
    FINALLY
        R.Free;
        SL.Free;
    END;
end;

function TFastListview.GetVisibleColumns(Idx: Integer): TListColumn;
var I, J: Integer;
    CL: TList;
    C, CJ: TListColumn;
begin
    CL := TList.Create;
    TRY
        for I := 0 to Columns.Count - 1 do
        begin
            if ColumnVisible[I] then
            begin
                C := Columns[I];
                C.Width := ColumnWidth[I];
                for J := CL.Count - 1 downto 0 do
                begin
                    CJ := CL[J];
                    if ColumnOrder[I] > ColumnOrder[CJ.Index] then
                    begin
                        CL.Insert(J + 1, C);
                        C := nil;
                        break;
                    end;
                end;
                if C <> nil then
                    CL.Add(C);
            end;
        end;
        Result := CL[Idx];
    FINALLY
        CL.Free;
    END;
end;

function TFastListview.GetVisibleColumnsCount: Integer;
var I: Integer;
begin
    Result := 0;
    for I := 0 to Columns.Count - 1 do
        if ColumnVisible[I] then Inc(Result);
end;

procedure TFastListview.Do_OnEdited(Sender: TObject; Item: TOldListItem;
    var S: string);
var FastItem: TFastListItem;
begin
    TRY
        if Assigned(fOnEdited) then
        begin
            FastItem := Selected;
            if FastItem = nil then
            begin
                if Item <> nil then
                    FastItem := Items[Item.Index]
                else
                    FastItem := ItemFocused;
            end;
            if FastItem = nil then
                Exit
            else
                fOnEdited(Sender, FastItem, S);
        end;
        FastItem := Selected;
        if FastItem <> nil then
            FastItem.Caption := S;
    FINALLY
        FEditing := FALSE;
    END;
end;

procedure TFastListview.Do_OnEditing(Sender: TObject; Item: TOldListItem;
    var AllowEdit: Boolean);
var FastItem: TFastListItem;
begin
    AllowEdit := not ReadOnly;
    if Assigned(fOnEditing) then
    begin
        FastItem := Selected;
        if FastItem = nil then
        begin
            if Item <> nil then
                FastItem := Items[Item.Index]
            else
                FastItem := ItemFocused;
        end;
        if FastItem = nil then
            AllowEdit := FALSE
        else
            fOnEditing(Sender, FastItem, AllowEdit);
    end;
    if AllowEdit then
        FEditing := TRUE;
end;

procedure TFastListview.SetMultiHeader(const Value: String);
begin
    FMultiHeader := Value;
    if Value <> '' then
        AllowColumnDrag := FALSE;
    if not (csDesigning in ComponentState) then
        RecalcHeaderHeight;
    Invalidate;
    InvalidateHeader;
end;

//{$DEFINE HEADER_WAR_IMAGELIST}

function TFastListview.RecalcHeaderHeight: Integer;
var I, J, Levels: Integer;
    S: String;
    HdrWnd: HWnd;
    R: TRect;
    {$IFDEF HEADER_WAR_IMAGELIST}
    //HDItem: THDItem;
    {$ENDIF}
begin
    Result := 0;
    if csLoading in ComponentState then Exit;
    Levels := 1;
    if MultiHeader <> '' then
        for I := 0 to Columns.Count - 1 do
        begin
            S := Columns[I].Caption;
            J := 1;
            while pos(MultiHeader, S) > 0 do
            begin
                S := Copy(S, pos(MultiHeader, S) + Length(MultiHeader), MaxInt);
                if S = '' then break;
                Inc(J);
            end;
            if J > Levels then
                Levels := J;
        end;
    Result := Levels;

    if not (csDesigning in ComponentState) then
    begin
        HandleNeeded;
        HdrWnd := GetHeaderCtlWindow;
        if HdrWnd = 0 then Exit;
        GetWindowRect(HdrWnd, R);
        I := Canvas.TextHeight('Ap_^');
        if I < 18 then I := 18;
        I := Levels * I;
        //if R.Bottom - R.Top = I then Exit;
        FHdrHeight := max(FHdrHeight, I);
        if FFakeHdrFont = nil then
        begin
            FFakeHdrFont := TFont.Create;
            FFakeHdrFont.Name := 'Arial';
        end;
        R.Bottom := R.Top + I;
        if FFakeHdrFont.Height <> -I - 2 * (Levels - 1) then
        begin
            FFakeHdrFont.Height := -I - 2 * (Levels - 1);
            SendMessage(HdrWnd, WM_SETFONT, 0, 1);
            if Levels = 1 then Exit;
            SendMessage(HdrWnd, WM_SETFONT, FFakeHdrFont.Handle, 1);
            {SetWindowPos( HdrWnd, 0, R.Left, R.Top, R.Right - R.Left, R.Bottom - R.Top,
                          SWP_NOMOVE or SWP_NOZORDER or SWP_NOACTIVATE );
            GetWindowRect( HdrWnd, R );}
            {$IFDEF HEADER_WAR_IMAGELIST}
            if HdrImgList = 0 then
                HdrImgList := ImageList_Create(1, I, ILC_COLOR, 1, 1);
            Header_SetImageList(HdrWnd, HdrImgList);

            {HDItem.Mask := HDI_IMAGE or HDI_FORMAT;
            HDItem.iImage := 0;
            HDItem.fmt := HDF_IMAGE;
            SendMessage( HdrWnd, HDM_SETITEM, 1, Integer( @HDItem ) );}
            {$ENDIF HEADER_WAR_IMAGELIST}
        end;
        {SetWindowPos( HdrWnd, 0, 0, 0, R.Right - R.Left, R.Bottom - R.Top,
                      SWP_NOZORDER or SWP_NOMOVE or SWP_SHOWWINDOW or SWP_NOCOPYBITS
                      or SWP_NOACTIVATE );}
    end;
end;

procedure TFastListview.PaintHdr(W: Hwnd; DC: HDC);
var R, R1, R2, R3, R4: TRect;
    Cnv: TCanvas;
    I, J, J1, K, H, ToCol: Integer;
    MP: TPoint;
    Buf: array[0..511] of Char;
    Item: THDItem;
    Levels,
        L: Integer;
    S, S1, S2: String;
    ResetLevel: Boolean;
    do_draw: Boolean;
begin
    if  not ShowColumnHeaders then Exit;
    {$IFDEF LOG_FLV}
    LogFileOutput('c:\log_flv.txt', 'FLV_PaintHdr ' + Name);
    {$ENDIF}
    if ForceCanvas <> 0 then
        DC := ForceCanvas;
    do_draw := TRUE;
    if ForceCanvas <> 0 then
    begin
        I := SendMessage(W, HDM_GETITEMCOUNT, 0, 0);
        for I := 0 to I - 1 do
        begin
            Item.Mask := HDI_TEXT;
            Item.cchTextMax := Sizeof(Buf);
            Item.pszText := @Buf[0];
            SendMessage(W, HDM_GETITEM, I, Integer(@Item));
            SendMessage(W, HDM_GETITEMRECT, I, Integer(@R1));
            if R.Left < R1.Right then
                R.Left := R1.Right;

            R2 := R1;
            DrawFrameControl(ForceCanvas, R2, DFC_BUTTON, DFCS_BUTTONPUSH);
        end;
    end;
    if MultiHeader = '' then
    begin
        inherited;
        //Exit;
        do_draw := FALSE;
    end;
    Levels := RecalcHeaderHeight;
    if Levels = 0 then Levels := 1;
    windows.GetClientRect(W, R);
    GetCursorPos(MP);
    windows.ScreenToClient(W, MP);
    Cnv := TCanvas.Create;
    try
        Cnv.Handle := DC;
        Cnv.Font.Assign(Font);
        if HdrColor <> clNone then
            Cnv.Brush.Color := HdrColor
        else
            Cnv.Brush.Color := clBtnFace;
        I := SendMessage(W, HDM_GETITEMCOUNT, 0, 0);
        L := 1;
        H := Canvas.TextHeight('Ap_^');
        if H < 18 then H := 18; //   
        if H < R.Bottom div Levels then
            H := R.Bottom div Levels;
        for I := 0 to I - 1 do
        begin
            S := Columns[I].Caption;
            J := 1;
            while pos(MultiHeader, S) > 0 do
            begin
                S := Copy(S, pos(MultiHeader, S) + Length(MultiHeader), MaxInt);
                if S = '' then break;
                Inc(J);
            end;
            if J > L then L := J; // L =   
            Item.Mask := HDI_TEXT;
            Item.cchTextMax := Sizeof(Buf);
            Item.pszText := @Buf[0];
            SendMessage(W, HDM_GETITEM, I, Integer(@Item));
            SendMessage(W, HDM_GETITEMRECT, I, Integer(@R1));
            if R.Left < R1.Right then
                R.Left := R1.Right;

            R2 := R1;
            {if L < Levels then
              R2.Top := R2.Bottom - H * L;}

            //   


            R3 := R2;
            S2 := Columns[I].Caption;
            //if L = Levels then
            if L > 1 then
                R3.Top := R2.Bottom - H;
            if do_draw then
                Cnv.FillRect(R3); //<<<<<<<<<<<<
            if pos(MultiHeader, S2) > 0 then
            begin
                S2 := Copy(S2, 1, pos(MultiHeader, S2) - 1);
            end;
            if do_draw then
                BtnDraw(Cnv, R3, not FTrackingHeader and not FTrackingHdr and
                    FLBtnPressed and
                    PtInRect(R1, MP), S2,
                    Columns[I].ImageIndex,
                    SmallImages, Self, I);
            if HdrColor <> clNone then
                Cnv.Brush.Color := HdrColor
            else
                Cnv.Brush.Color := clBtnFace;
            if do_draw then
                doHdrPaint(Self, Cnv, R3, I);

            //    ,     

            ResetLevel := FALSE;
            S := Columns[I].Caption;
            for J := 2 to L do
            begin
                K := pos(MultiHeader, S);
                if K <= 0 then break; //        -
                S := Copy(S, K + Length(MultiHeader), MaxInt);
                if S = '' then
                begin
                    ResetLevel := TRUE;
                    break; //         ,
                    //      -
                end;
                K := pos(MultiHeader, S);
                S1 := S;
                if K > 0 then S1 := Copy(S, 1, K - 1); // S1 = ,      
                ToCol := I;
                for K := I + 1 to Columns.Count - 1 do
                begin
                    S2 := Columns[K].Caption;
                    J1 := 1;
                    while pos(MultiHeader, S2) > 0 do
                    begin
                        S2 := Copy(S2, pos(MultiHeader, S2) + Length(MultiHeader), MaxInt);
                        if S2 = '' then
                        begin //       -
                            //L := J1;
                            //ResetLevel := TRUE;
                            ToCol := K;
                            J1 := J;
                            break;
                        end;
                        Inc(J1);
                    end;
                    //if ResetLevel then Exit;
                    if J1 >= J then break; //  ,     J   
                    ToCol := K;
                end;
                // ,  S1      I    ToCol
                R3.Bottom := R3.Top;
                R3.Top := R3.Top - H;
                if J = L then
                    R3.Top := R2.Top;
                SendMessage(W, HDM_GETITEMRECT, ToCol, Integer(@R4));
                R3.Right := R4.Right;
                if do_draw then
                begin
                    Cnv.FillRect(R3); //<<<<<<<<<<<<
                    BtnDraw(Cnv, R3, FALSE, S1, -1, nil, Self, -1);
                end;
            end;
            if ResetLevel then L := 1;
        end;
        if HdrColor <> clNone then
            Cnv.Brush.Color := HdrColor
        else
            Cnv.Brush.Color := clBtnFace;
        Cnv.FillRect(R);
    finally
        Cnv.Free;
    end;
    {$IFDEF LOG_FLV}
    LogFileOutput('c:\log_flv.txt', 'FLV_PaintHdr exit');
    {$ENDIF}
end;

procedure TFastListview.DefineProperties(Filer: TFiler);
    function AtLeastOneComboItemsNotEmpty: Boolean;
    var I: Integer;
        SL: TStrings;
    begin
        Result := TRUE;
        for I := 0 to FColumnComboItems.Count - 1 do
        begin
            SL := FColumnComboItems[I];
            if SL.Count > 0 then Exit;
        end;
        Result := FALSE;
    end;
begin
    inherited;
    Filer.DefineBinaryProperty('AdditionalColumnProp',
        LoadColumnAdditional, StoreColumnAdditional,
        (Columns <> nil) and (Columns.Count > 0));
    Filer.DefineBinaryProperty('ColumnFieldNames',
        LoadColumnFieldNames, StoreColumnFieldNames,
        Assigned(FColumnNames) and (FColumnNames.Count > 0));
    Filer.DefineBinaryProperty('ColumnCanEdit',
        LoadColumnCanEdit, StoreColumnCanEdit,
        Assigned(FColumnCanEdit) and (FColumnCanEdit.Count > 0));
    Filer.DefineBinaryProperty('ColumnEditKind',
        LoadColumnEditKind, StoreColumnEditKind,
        Assigned(FColumnEditKind) and (FColumnEditKind.Count > 0));
    Filer.DefineBinaryProperty('ColumnComboItems',
        LoadColumnComboItems, StoreColumnComboItems,
        Assigned(FColumnComboItems) and (FColumnComboItems.Count > 0) and
        AtLeastOneComboItemsNotEmpty);
end;

procedure TFastListview.LoadColumnAdditional(Stream: TStream);
var AA: TColAdditionalAttrs;
    I: Integer;
    ColOrders: TList;
begin
    ColOrders := TList.Create;
    TRY

        for I := 0 to Columns.Count - 1 do
        begin
            Stream.Read(AA.SortAttrs, Sizeof(AA.SortAttrs));
            Stream.Read(AA.Filtered, Sizeof(AA.Filtered));
            Stream.Read(AA.VisualOrder, Sizeof(AA.VisualOrder));
            Stream.Read(AA.Visible, Sizeof(AA.Visible));
            TRY
                SetSortColumnAttributes([I], AA.SortAttrs);
            EXCEPT
                ShowMessage('Can not set SortColumnAttributes, I=' + IntToStr(I));
            END;
            TRY
                if AA.Filtered then
                    ColumnFiltered[I] := TRUE;
            EXCEPT
                ShowMessage('Can not set Filtered, I=' + IntToStr(I));
            END;
            TRY
                if AA.VisualOrder >= 0 then
                begin
                    //ShowMessage( 'Load ColumnOrder[ ' + IntToStr( I ) + ' ]=' +IntToStr( AA.VisualOrder ) );
                    //ColumnOrder[ I ] := AA.VisualOrder;
                    while ColOrders.Count < I + 1 do
                        ColOrders.Add(Pointer(-1));
                    ColOrders[I] := Pointer(AA.VisualOrder);
                end;
            EXCEPT
                ShowMessage('Can not set VisualOrder, I=' + IntToStr(I));
            END;
            TRY
                if not AA.Visible then
                    ColumnVisible[I] := FALSE;
            EXCEPT
                //ShowMessage('Can not set ColumnVisible, I=' + IntToStr(I));
            END;
        end;
        for I := ColOrders.Count - 1 downto 0 do
        begin
            if ColOrders[I] <> Pointer(-1) then
                ColumnOrder[I] := Integer(ColOrders[I]);
        end;

    FINALLY
        ColOrders.Free;
    END;
end;

procedure TFastListview.StoreColumnAdditional(Stream: TStream);
var I, J: Integer;
    SA: TColumnSortAttrs;
    B: Boolean;
begin
    for I := 0 to Columns.Count - 1 do
    begin
        TRY
            J := 0;
            if (FColumnSortAttrs <> nil) and (FColumnSortAttrs.Count > I) then
                J := Integer(FColumnSortAttrs[I]);
            SA := PColumnSortAttrs(@J)^;
            Stream.Write(SA, Sizeof(TColumnSortAttrs));
        EXCEPT
            ShowMessage('error writing ColumnSortAttrs');
        END;
        TRY
            B := (FColumnsToFilter <> nil) and
                (FColumnsToFilter.IndexOf(IntToStr(I)) >= 0);
            Stream.Write(B, Sizeof(B));
        EXCEPT
            ShowMessage('error writing Filtered');
        END;
        TRY
            J := -1;
            if ColumnOrder[I] <> -1 then
            begin
                J := ColumnOrder[I];
                //ShowMessage( 'Store ColumnOrder[ ' + IntToStr( I ) + ' ] = ' + IntToStr( J ) );
            end;
            Stream.Write(J, Sizeof(J));
        EXCEPT
            ShowMessage('error writing ColumnOrder');
        END;
        TRY
            B := ColumnVisible[I];
            Stream.Write(B, Sizeof(Boolean));
        EXCEPT
            ShowMessage('error writing ColumnVisible')
        END;
    end;
end;

function TFastListview.GetColumnSortAttrs(ColIdx: Integer): TColumnSortAttrs;
var I: Integer;
begin
    Result := [];
    if (FColumnSortAttrs <> nil) and (FColumnSortAttrs.Count > ColIdx) then
    begin
        I := Integer(FColumnSortAttrs[ColIdx]);
        Result := PColumnSortAttrs(@I)^;
    end;
end;

procedure TFastListview.SetIsColumnFiltered(colIdx: Integer;
    const Value: Boolean);
var I: Integer;
begin
    if FColumnsToFilter = nil then
    begin
        ShowMessage('FColumnsToFilter created');
        FColumnsToFilter := TStringList.Create;
    end;
    I := FColumnsToFilter.IndexOf(IntToStr(colIdx));
    if Value then
    begin
        if I < 0 then
            FColumnsToFilter.Add(IntToStr(ColIdx));
    end
    else
    begin
        if I >= 0 then
            FColumnsToFilter.Delete(I);
    end;
end;

procedure TFastListview.SetAllowColumnHide(const Value: Boolean);
var I: Integer;
begin
    FAllowColumnHide := Value;
    if not Value then
    begin
        for I := 0 to Columns.Count - 1 do
            ColumnVisible[I] := TRUE;
    end;
end;

type
    TSortColsData = record
        ColsList: TList;
        ListView: TFastListView;
    end;
    PSortColsData = ^TSortColsData;

function CompareColOrders(const Data: Pointer; const e1, e2: Dword): Integer;
var SCD: PSortColsData;
    Col1, Col2: TListColumn;
begin
    SCD := Data;
    Col1 := SCD.ColsList[e1];
    Col2 := SCD.ColsList[e2];
    Result := SCD.ListView.ColumnOrder[Col1.Index] - SCD.ListView.ColumnOrder[Col2.Index];
end;

procedure SwapItems(const Data: Pointer; const e1, e2: Dword);
var SCD: PSortColsData;
    Tmp: Pointer;
begin
    SCD := Data;
    Tmp := SCD.ColsList[e1];
    SCD.ColsList[e1] := SCD.ColsList[e2];
    SCD.ColsList[e2] := Tmp;
end;

procedure TFastListView.SaveToWorksheet(Excel, WB, WS: Variant; TopLine: Integer = 1;
    const ImageNames: String = '';
    const StateImageNames: String = ''; FromIdx: Integer = 0; ToIdx: Integer = MaxInt;
    SelectedOnly: Boolean = FALSE; Progress: TFLV_fmProgress = nil);
var i, i0, j, K, N, L: Integer;
    LI: TListItem;
    S: String;

    procedure DoRange(y1, x1, y2, x2: Integer; const Str: String = '';
        Align: Integer = -1; BorderBold: Boolean = TRUE);
    var R: Variant;
    begin
        y1 := y1 + TopLine - 1;
        y2 := y2 + TopLine - 1;
        if Str <> '' then WS.Cells[y1, x1] := Str;
        R := WS.Range[WS.Cells[y1, x1], WS.Cells[y2, x2]];
        TRY
            R.MergeCells := TRUE;
            if BorderBold then
            begin
                R.Borders.LineStyle := 1;
                R.Borders.Weight := 3;

                R.Borders[1].Color := clWhite;
                R.Borders[3].Color := clWhite;
            end;
            if (Align >= 0) and ShowColumnHeaders then
                R.HorizontalAlignment := Align;
            //if y2 > y1 then
            R.WrapText := TRUE;
            R.Interior.Color := clSilver;
        FINALLY
            R := Unassigned;
        END;
    end;

    function AlignOfCol(Idx: Integer): Integer;
    begin
        CASE Columns[Idx].Alignment OF
            taLeftJustify: Result := 1;
            taCenter: Result := 3;
            taRightJustify: Result := 4;
        else Result := 0;
        END;
    end;

    procedure FillNames(SL: TStringList; S: String);
    var I: Integer;
    begin
        while S <> '' do
        begin
            I := pos(',', S);
            if I > 0 then
            begin
                SL.Add(Copy(S, 1, I - 1));
                S := Copy(S, I + 1, MaxInt);
            end
            else
            begin
                SL.Add(S);
                S := '';
            end;
        end;
    end;

var HLevels, X: Integer;
    SHeader: String;
    Line: Variant;
    HasIcons, HasStatic: Boolean;
    Img_Names, Stt_Names: TStringList;
    SCD: TSortColsData;

    function Cols(I: Integer): TListColumn;
    begin
        Result := SCD.ColsList[I];
    end;

    function ColsCount: Integer;
    begin
        Result := SCD.ColsList.Count;
    end;

    function ItemCols(LI: TListItem; I: Integer): String;
    begin
        I := Cols(I).Index;
        if I = 0 then
            Result := LI.Caption
        else
        begin
            if Assigned(OnDataCell) then
                OnDataCell(Self, LI.Index, I, Result)
            else
                if LI.SubItemsList.Count > I - 1 then
                    Result := LI.SubItemsList[I - 1]
                else
                    Result := '';
        end;
    end;

var S1: String;
    R: Variant;
begin
    if FromIdx < 0 then FromIdx := 0;
    if (FromIdx >= Items.Count) and not (Items.Count = 0) then Exit;
    if ToIdx > Items.Count - 1 then ToIdx := Items.Count - 1;
    if (ToIdx < FromIdx) and not ((ToIdx = -1) and (Items.Count = 0)) then Exit;

    Img_Names := TStringList.Create;
    Stt_Names := TStringList.Create;
    FillNames(Img_Names, ImageNames);
    FillNames(Stt_Names, StateImageNames);
    SCD.ColsList := TList.Create;
    SCD.ListView := Self;
    TRY
        for I := 0 to VisibleColumnsCount - 1 do
        begin
            j := SCD.ColsList.Add(VisibleColumns[I]);
            TListColumn(SCD.ColsList[j]).Width :=
                VisibleColumns[I].Width;
        end;
        SortData(@SCD, SCD.ColsList.Count, CompareColOrders, SwapItems);

        if (Progress <> nil) then
        begin
            Progress.Show;
        end;
        Application.ProcessMessages;

        HLevels := 1;
        if MultiHeader <> '' then
        begin
            //     
            if ShowColumnHeaders then
                for i := 0 to Columns.Count - 1 do
                begin
                    S := Columns[i].Caption;
                    j := 1;
                    while S <> '' do
                    begin
                        if pos(MultiHeader, S) > 0 then
                            S := Copy(S, pos(MultiHeader, S) + Length(MultiHeader), MaxInt)
                        else
                            break;
                        Inc(j);
                    end;
                    if j > HLevels then HLevels := j;
                end;
        end;
        // ,   
        HasIcons := FALSE;
        if Img_Names.Count > 0 then
        begin
            if SmallImages <> nil then
                for I := FromIdx to ToIdx do
                    if Items[I].ImageIndex > -1 then
                    begin
                        HasIcons := TRUE;
                        break;
                    end;
        end;
        HasStatic := FALSE;
        if Stt_Names.Count > 0 then
        begin
            if SmallImages <> nil then
                for I := FromIdx to ToIdx do
                    if Items[I].StateIndex > -1 then
                    begin
                        HasStatic := TRUE;
                        break;
                    end;
        end;
        X := 0;
        if HasIcons then Inc(X);
        if HasStatic then Inc(X);

        //   
        if ShowColumnHeaders then
            for i := 0 to ColsCount - 1 do
            begin
                S := Cols(i).Caption;
                if Cols(i).Width <> 50 then
                begin
                    R := WS.Range[WS.Cells[1, i + 1], WS.Cells[1, i + 1]]; ;
                    R.Width := Cols(i).Width;
                end;
                for j := 1 to HLevels do
                begin
                    if S = '' then break;
                    if pos(MultiHeader, S) > 0 then
                    begin
                        SHeader := Copy(S, 1, pos(MultiHeader, S) - 1);
                        S := Copy(S, pos(MultiHeader, S) + Length(MultiHeader), MaxInt);
                    end
                    else
                    begin
                        SHeader := S;
                        S := '';
                    end;
                    if j = 1 then
                    begin
                        {if S = '' then
                        DoRange( 1, X + i + 1, HLevels - j + 1, X + i + 1, SHeader, 3 )
                        else}
                        DoRange(HLevels - j + 1, X + i + 1, HLevels - j + 1, X + i + 1, SHeader, 3);
                    end
                    else
                    begin
                        //   ,      
                        //  j:
                        N := i + 1;
                        for k := i + 1 to ColsCount do
                        begin
                            if k = ColsCount then
                            begin
                                N := k;
                                break;
                            end;
                            S1 := Cols(k).Caption;
                            //  j :
                            for L := 1 to j - 1 do
                                if pos(MultiHeader, S1) > 0 then
                                begin
                                    S1 := Copy(S1, pos(MultiHeader, S1) + Length(MultiHeader), MaxInt);
                                    if S1 = '' then
                                    begin
                                        N := k + 1; //       
                                        break;
                                    end;
                                end
                                else
                                begin
                                    S1 := '';
                                    break;
                                end;
                            if S1 <> '' then
                            begin
                                N := k; //   k      j
                                break;
                            end;
                        end;
                        {if S = '' then
                        DoRange( 1, X + N + 1, HLevels - j + 1, X + i + 1, SHeader, 3 )
                        else}
                        DoRange(HLevels - j + 1, X + i + 1, HLevels - j + 1, X + N, SHeader, 3);
                    end;
                end;
            end
        else
        begin
            HLevels := 0;
        end;

        Application.ProcessMessages;
        if Progress.Cancelled then Exit;

        if FromIdx <= ToIdx then
        begin
            //  - 
            if HLevels > 0 then
            begin
                Excel.ActiveWindow.SplitRow := HLevels;
                Excel.ActiveWindow.FreezePanes := TRUE;
            end;

            //     
            TRY
                if ShowColumnHeaders then
                    WS.Range[WS.Cells[HLevels + TopLine, X + 1],
                        WS.Cells[HLevels + ToIdx - FromIdx + TopLine,
                        X + ColsCount]].NumberFormat := '@';
            EXCEPT
                on E: EOleError do
                begin
                    ShowMessage(E.Message);
                    raise;
                end;
            END;
            if ShowColumnHeaders then
                for j := 0 to ColsCount - 1 do
                    if Cols(j).Alignment <> taLeftJustify then
                        WS.Range[WS.Cells[HLevels + TopLine, X + j + 1],
                            WS.Cells[HLevels + ToIdx - FromIdx + TopLine,
                            X + ColsCount]].HorizontalAlignment := AlignOfCol(j);
            i := -1;
            for i0 := FromIdx to ToIdx do
            begin
                if Progress <> nil then
                begin
                    Progress.ProgressBar1.Position :=
                        Progress.ProgressBar1.Position + 1;
                    Application.ProcessMessages;
                    if Progress.Cancelled then Exit;
                end;

                LI := Items[i0];
                if SelectedOnly and not LI.Selected then
                begin
                    //LI := Items[ i0 ];
                    continue;
                end;
                Inc(i);

                Line := VarArrayCreate([0, X + VisibleColumnsCount],
                    varVariant);

                if HasIcons then
                begin
                    Line[0] := '';
                    if LI.ImageIndex >= 0 then
                        if Img_Names.Count > LI.ImageIndex then
                            Line[0] := Img_Names[LI.ImageIndex];
                end;

                if HasStatic then
                begin
                    Line[X - 1] := '';
                    if LI.StateIndex >= 0 then
                        if Stt_Names.Count > LI.ImageIndex then
                            Line[X - 1] := Stt_Names[LI.StateIndex];
                end;

                for j := 0 to ColsCount - 1 do
                    if ItemCols(LI, j) <> '' then
                        Line[X + j] := ItemCols(LI, j);

                WS.Range[WS.Cells[i + HLevels + TopLine, 1],
                    WS.Cells[i + HLevels + TopLine,
                    X + ColsCount]] := Line;

                if LI.BackColor <> clNone then
                    WS.Range[WS.Cells[i + HLevels + TopLine, X + 1],
                        WS.Cells[i + HLevels + TopLine, X + ColsCount]]
                        .Interior.Color := ColorToRGB(LI.BackColor);

                if LI.FontColor <> clNone then
                    WS.Range[WS.Cells[i + HLevels + TopLine, X + 1],
                        WS.Cells[i + HLevels + TopLine, X + ColsCount]]
                        .Font.Color := ColorToRGB(LI.FontColor);

                if fsBold in LI.FontStyle then
                    WS.Range[WS.Cells[i + HLevels + TopLine, X + 1],
                        WS.Cells[i + HLevels + TopLine, X + ColsCount]]
                        .Font.Bold := TRUE;

                if fsItalic in LI.FontStyle then
                    WS.Range[WS.Cells[i + HLevels + TopLine, X + 1],
                        WS.Cells[i + HLevels + TopLine, X + ColsCount]]
                        .Font.Italic := TRUE;

                if fsUnderline in LI.FontStyle then
                    WS.Range[WS.Cells[i + HLevels + TopLine, X + 1],
                        WS.Cells[i + HLevels + TopLine, X + ColsCount]]
                        .Font.Underline := TRUE;

                if fsStrikeout in LI.FontStyle then
                    WS.Range[WS.Cells[i + HLevels + TopLine, X + 1],
                        WS.Cells[i + HLevels + TopLine, X + ColsCount]]
                        .Font.Striketrough := TRUE;

                if Assigned(OnLineToExcel) then
                    OnLineToExcel(Self, LI, WB, WS, X + 1, X + ColsCount);

            end;
        end;

    FINALLY
        Img_Names.Free;
        Stt_Names.Free;
        SCD.ColsList.Free;
    END;
end;

procedure TFastListview.SaveToExcel(const Filename: String; const ImageNames: String = '';
    const StateImageNames: String = ''; FromIdx: Integer = 0; ToIdx: Integer = MaxInt;
    SelectedOnly: Boolean = FALSE);
var Excel, WB, WS: Variant;
    {i, i0, j, K, N, L: Integer;
    LI: TListItem;
    S: String;}
    Progress: TFLV_fmProgress;

    {var HLevels, X: Integer;
        SHeader: String;
        Line: Variant;
        HasIcons, HasStatic: Boolean;
        Img_Names, Stt_Names: TStringList;
        SCD: TSortColsData;
    var S1: String;}
begin
    if FromIdx < 0 then FromIdx := 0;
    if FromIdx >= Items.Count then Exit;
    if ToIdx > Items.Count - 1 then ToIdx := Items.Count - 1;
    if ToIdx < FromIdx then Exit;

    Progress := TFLV_fmProgress.Create(Application);
    Excel_Progress := Progress;
    TRY
        Progress.ProgressBar1.Max := Items.Count + 1;
        Progress.Show;
        Application.ProcessMessages;

        CoInitialize(nil);

        Excel := CreateOleObject('Excel.Application');
        WB := Excel.Workbooks.Add;
        TRY
            WS := WB.ActiveSheet;
            if SaveToExcelViaClipboard then
                SaveToWorksheetFast(Excel, WB, WS, 1, ImageNames, StateImageNames, FromIdx,
                    ToIdx, SelectedOnly, Progress)
            else
                SaveToWorksheet(Excel, WB, WS, 1, ImageNames, StateImageNames, FromIdx,
                    ToIdx, SelectedOnly, Progress);
            if Assigned(OnEndOfSaveToExcel) then
                OnEndOfSaveToExcel(Self, WS);
            WB.SaveAs(FileName);
            //WB.Activate(  );

        FINALLY
            //ShellExecute( 0, 'open', PChar( FileName ), nil, nil, SW_SHOW );
            WB.Close;
            WS := Unassigned;
            Excel := Unassigned;
            CoUninitialize;
        END;

    FINALLY
        Excel_Progress := nil;
        Progress.Free;
    END;
end;

function TFastListview.ContentAsText(const ColumnSeparator,
    LineSeparator: String): String;
var I, J: Integer;
    LI: TListItem;
    S: String;
begin
    Result := '';
    for I := 0 to Items.Count - 1 do
    begin
        LI := Items[I];
        S := LI.Caption;
        for J := 0 to LI.SubItemsList.Count - 1 do
            S := S + ColumnSeparator + LI.SubItemsList[J];
        Result := Result + S + LineSeparator;
    end;
end;

procedure TFastListview.NilFFilterListForm;
begin
    FFilterListForm := nil;
end;

function TFastListview.ColumnIndexByName(const S: String): Integer;
var I: Integer;
begin
    Result := -1;
    if not Assigned(FColumnNames) then Exit;
    I := FColumnNames.Count - 1;
    if Columns.Count - 1 > I then
        I := Columns.Count - 1;
    for I := 0 to I do
        if (I < FColumnNames.Count) and (AnsiCompareText(FColumnNames[I], S) = 0) or
            ((I >= FColumnNames.Count) or (FColumnNames[I] = '')) and
            (I < Columns.Count) and (AnsiCompareText(Columns[I].Caption, S) = 0) then
        begin
            Result := I; Exit;
        end;
end;

function TFastListview.GetColFNames(Idx: Integer): String;
begin
    Result := '';
    if not Assigned(FColumnNames) then Exit;
    if FColumnNames.Count <= Idx then Exit;
    Result := FColumnNames[Idx];
end;

procedure TFastListview.SetColFNames(Idx: Integer; const Value: String);
begin
    if not Assigned(FColumnNames) then
        FColumnNames := TStringList.Create;
    TRY
        while FColumnNames.Count <= Idx do
            FColumnNames.Add('');
    EXCEPT on E: Exception do ShowMessage('exception: ' + E.Message);
    END;
    FColumnNames[Idx] := Value;
end;

procedure TFastListview.LoadColumnFieldNames(Stream: TStream);
var I, N: Integer;
    S: String;
begin
    TRY
        if not Assigned(FColumnNames) then
            FColumnNames := TStringList.Create;
        FColumnNames.Clear;
        Stream.Read(N, Sizeof(N));
        for N := 0 to N - 1 do
        begin
            Stream.Read(I, Sizeof(I));
            SetLength(S, I);
            if I > 0 then
                Stream.Read(S[1], I);
            FColumnNames.Add(S);
        end;
    FINALLY

    END;
end;

procedure TFastListview.StoreColumnFieldNames(Stream: TStream);
var I, K, N: Integer;
    S: String;
begin
    TRY
        N := FColumnNames.Count;
        Stream.Write(N, Sizeof(N));
        for K := 0 to N - 1 do
        begin
            I := Length(FColumnNames[K]);
            Stream.Write(I, Sizeof(I));
            if I > 0 then
            begin
                S := FColumnNames[K];
                Stream.Write(S[1], I);
            end;
        end;
    EXCEPT on E: Exception do
        begin
            ShowMessage('Store ColumnFieldNames: (FColumnNames.Count=' +
                IntToStr(FColumnNames.Count) + ',S=' + S + ') ' + E.Message);
        end;
    END;
end;

procedure TFastListview.SetAutoAllocateColumnNames(const Value: Boolean);
begin
    FAutoAllocateColumnNames := Value;
end;

function TFastListview.CompareItems(Item1, Item2: TFastListItem): Integer;
begin
    CompareItemsFast(Self, Item1, Item2, Result);
end;

procedure TFastListview.AddUniSource(UniSource: TComponent);
begin
    FUniSources.Add(UniSource);
end;

function TFastListview.GetUniSources(Idx: Integer): TComponent;
begin
    Result := FUniSources[Idx];
end;

procedure TFastListview.RemoveUniSource(UniSource: TComponent);
begin
    FUniSources.Remove(UniSource);
end;

procedure TFastListview.SetUniSourcesCount(const Value: Integer);
begin
    FUniSourcesCount := Value;
end;

function TFastListview.GetUniSourcesCount: Integer;
begin
    Result := FUniSources.Count;
end;

function TFastListview.GetColComboItems(Idx: Integer): TStrings;
begin
    if not Assigned(FColumnComboItems) then
        FColumnComboItems := TList.Create;
    while FColumnComboItems.Count - 1 < Idx do
        FColumnComboItems.Add(TStringList.Create);
    Result := FColumnComboItems[Idx];
end;

function TFastListview.GetColKinds(Idx: Integer): TEditKind;
begin
    Result := ekEdit;
    if not Assigned(FColumnEditKind) then Exit;
    if FColumnEditKind.Count - 1 < Idx then Exit;
    Result := TEditKind(FColumnEditKind[Idx]);
end;

function TFastListview.GetColumnCanEdit(Idx: Integer): Boolean;
begin
    Result := FALSE;
    if not Assigned(FColumnCanEdit) then Exit;
    Result := FColumnCanEdit.IndexOf(Pointer(Idx)) >= 0;
end;

procedure TFastListview.SetColKinds(Idx: Integer; const Value: TEditKind);
begin
    if not Assigned(FColumnEditKind) then
        FColumnEditKind := TList.Create;
    while FColumnEditKind.Count - 1 < Idx do
        FColumnEditKind.Add(Pointer(ekEdit));
    FColumnEditKind[Idx] := Pointer(Value);
end;

procedure TFastListview.SetColumnCanEdit(Idx: Integer;
    const Value: Boolean);
begin
    if not Assigned(FColumnCanEdit) then
    begin
        if not Value then Exit;
        FColumnCanEdit := TList.Create;
    end;
    if Value then
    begin
        if FColumnCanEdit.IndexOf(Pointer(Idx)) < 0 then
            FColumnCanEdit.Add(Pointer(Idx));
    end
    else
    begin
        FColumnCanEdit.Remove(Pointer(Idx));
    end;
end;

procedure TFastListview.LoadColumnCanEdit(Stream: TStream);
var I, N: Integer;
begin
    TRY
        if not Assigned(FColumnCanEdit) then
            FColumnCanEdit := TList.Create;
        Stream.Read(N, Sizeof(N));
        for N := 0 to N - 1 do
        begin
            Stream.Read(I, Sizeof(I));
            FColumnCanEdit.Add(Pointer(I));
        end;
    FINALLY

    END;
end;

procedure TFastListview.LoadColumnComboItems(Stream: TStream);
var I, N: Integer;
    S: String;
    SL: TStringList;
begin
    TRY
        if not Assigned(FColumnComboItems) then
            FColumnComboItems := TList.Create;
        Stream.Read(N, Sizeof(N));
        for N := 0 to N - 1 do
        begin
            Stream.Read(I, Sizeof(I));
            SetLength(S, I);
            if I > 0 then
                Stream.Read(S[1], I);
            SL := TStringList.Create;
            SL.Text := S;
            FColumnComboItems.Add(SL);
        end;
    FINALLY

    END;
end;

procedure TFastListview.LoadColumnEditKind(Stream: TStream);
var I, N: Integer;
begin
    TRY
        if not Assigned(FColumnEditKind) then
            FColumnEditKind := TList.Create;
        FColumnEditKind.Clear;
        Stream.Read(N, Sizeof(N));
        for N := 0 to N - 1 do
        begin
            Stream.Read(I, Sizeof(I));
            FColumnEditKind.Add(Pointer(I));
        end;
    FINALLY

    END;
end;

procedure TFastListview.StoreColumnCanEdit(Stream: TStream);
var I, K, N: Integer;
begin
    TRY
        N := FColumnCanEdit.Count;
        Stream.Write(N, Sizeof(N));
        for K := 0 to N - 1 do
        begin
            I := Integer(FColumnCanEdit[K]);
            Stream.Write(I, Sizeof(I));
        end;
    EXCEPT on E: Exception do
        begin
            ShowMessage('Store ColumnCanEdit: (FColumnCanEdit.Count=' +
                IntToStr(FColumnCanEdit.Count) + ') ' + E.Message);
        end;
    END;
end;

procedure TFastListview.StoreColumnComboItems(Stream: TStream);
var I, K, N: Integer;
    SL: TStringList;
    S: String;
begin
    S := '';
    TRY
        N := FColumnComboItems.Count;
        Stream.Write(N, Sizeof(N));
        for K := 0 to N - 1 do
        begin
            SL := FColumnComboItems[K];
            S := SL.Text;
            I := Length(S);
            Stream.Write(I, Sizeof(I));
            Stream.Write(S[1], I);
        end;
    EXCEPT on E: Exception do
        begin
            ShowMessage('Store ColumnComboItems: (FColumnComboItems.Count=' +
                IntToStr(FColumnComboItems.Count) + ', S=' + S + ') ' + E.Message);
        end;
    END;
end;

procedure TFastListview.StoreColumnEditKind(Stream: TStream);
var I, K, N: Integer;
begin
    TRY
        N := FColumnEditKind.Count;
        Stream.Write(N, Sizeof(N));
        for K := 0 to N - 1 do
        begin
            I := Integer(FColumnEditKind[K]);
            Stream.Write(I, Sizeof(I));
        end;
    EXCEPT on E: Exception do
        begin
            ShowMessage('Store ColumnEditKind: (FColumnEditKind.Count=' +
                IntToStr(FColumnEditKind.Count) + ') ' + E.Message);
        end;
    END;
end;

procedure TFastListview.SetColumnComboItems(Idx: Integer;
    const Value: TStrings);
begin
    if not Assigned(FColumnComboItems) then
        FColumnComboItems := TList.Create;
    while FColumnComboItems.Count - 1 < Idx do
        FColumnComboItems.Add(TStringList.Create);
    if Value <> nil then
        TStrings(FColumnComboItems[Idx]).Assign(Value)
    else
        TStrings(FColumnComboItems[Idx]).Clear;
end;

function TFastListview.CanEditByCells: Boolean;
var I: Integer;
begin
    Result := TRUE;
    for I := 0 to Columns.Count - 1 do
        if ColumnCanEdit[I] then Exit;
    Result := FALSE;
end;

procedure TFastListview.WMMouseDown(var M: TMessage);
var P: TPoint;
    Item: TListItem;
    I: Integer;
    R: TRect;
    Kind: TEditKind;
    Can: Boolean;
begin
    GetCursorPos(P);
    inherited;
    P := ScreenToClient(P);
    Item := GetItemAt(P.X, P.Y);
    if Item = nil then Exit;
    for I := 0 to Columns.Count - 1 do
    begin
        R := Item.GetRect(I);
        if PtInRect(R, P) then
        begin //  ,   :
            if not ColumnCanEdit[I] then Exit;
            Can := TRUE;
            Kind := ColumnEditKind[I];
            if Assigned(OnBeginEditCell) then
                OnBeginEditCell(Self, Item, I - 1, Can, Kind);
            if not Can then Exit;
            Item.EditCell(I);
            Exit;
        end;
    end;
end;

function TFastListview.EditingCell: Boolean;
begin
    Result := Assigned(FEditCellCtl) and (FEditCellCtl.Visible) or FEditing;
end;

procedure TFastListview.EditCellKeyPress(Sender: TObject; var Key: Char);
begin
    CASE Key OF
        #13: FEditingItem.EndEditCell;
        #27: FEditingItem.EndEditCell(TRUE);
        #9: //ShowMessage( 'TAB TAB TAB TAB TAB TAB TAB TAB !' );
    END;
end;

procedure TFastListview.EndEditCell;
begin
    if EditingCell then
        FEditingItem.EndEditCell;
end;

procedure TFastListview.WMKeyDown(var M: TMessage);
var I: Integer;
begin
    inherited;
    if (M.wParam = VK_RETURN) and not EditingCell and CanEditByCells and
        (Selected <> nil) then
    begin
        if SelectedColumn < 0 then
        begin
            for I := 0 to Columns.Count - 1 do
                if ColumnCanEdit[I] then
                begin
                    SelectedColumn := I;
                    break;
                end;
        end;
        Selected.EditCell(SelectedColumn);
    end
    else
        if not EditingCell and CanEditByCells and (Selected <> nil) and
            (M.wParam = VK_LEFT) then
        begin
            if SelectedColumn < 0 then SelectedColumn := Columns.Count;
            for I := SelectedColumn - 1 downto 0 do
            begin
                if ColumnCanEdit[I] then
                begin
                    SelectedColumn := I;
                    Exit;
                end;
            end;
            if Selected.Index > 0 then
            begin
                ItemFocused := Items[Selected.Index - 1];
                Selected := ItemFocused;
                SelectedColumn := Columns.Count;
                for I := SelectedColumn - 1 downto 0 do
                begin
                    if ColumnCanEdit[I] then
                    begin
                        SelectedColumn := I;
                        Exit;
                    end;
                end;
            end;
        end
        else
            if not EditingCell and CanEditByCells and (Selected <> nil) and
                (M.wParam = VK_LEFT) then
            begin
                if SelectedColumn < 0 then SelectedColumn := -1;
                for I := SelectedColumn + 1 to Columns.Count - 1 do
                begin
                    if ColumnCanEdit[I] then
                    begin
                        SelectedColumn := I;
                        Exit;
                    end;
                end;
                if Selected.Index < Items.Count - 1 then
                begin
                    ItemFocused := Items[Selected.Index + 1];
                    Selected := ItemFocused;
                    SelectedColumn := -1;
                    for I := SelectedColumn - 1 downto 0 do
                    begin
                        if ColumnCanEdit[I] then
                        begin
                            SelectedColumn := I;
                            Exit;
                        end;
                    end;
                end;
            end;
end;

function TFastListview.GetSubItemRect(ItemIdx, ColIdx: Integer): TRect;
begin
    Result := Items[ItemIdx].GetRect(ColIdx);
end;

procedure TFastListview.WMScroll(var M: TMessage);
begin
    if (M.wParam = SC_HSCROLL) or (M.wParam = SC_VSCROLL) then
        if EditingCell then EndEditCell;
    inherited;
end;

function TFastListview.Sort: Boolean;
begin
    Result := AlphaSort1(FALSE);
end;

function TFastListview.GetFilterValues: TFilterStrings;
var i: Integer;
    SL1, SL2: TStringList;
begin
    if FColumnsToFilter = nil then Result := nil
    else
    begin
        Result := TFilterStrings.Create;
        Result.Assign(FColumnsToFilter);
        for i := 0 to Result.Count - 1 do
        begin
            SL1 := Pointer(FColumnsToFilter.Objects[i]);
            SL2 := TStringList.Create;
            if SL1 <> nil then
                SL2.Assign(SL1);
            Result.Objects[i] := SL2;
        end;
    end;
end;

procedure TFastListview.SetFilterValues(const Value: TFilterStrings);
var i: Integer;
    SL1, SL2: TStringList;
begin
    if Value = nil then
    begin
        if FColumnsToFilter <> nil then
        begin
            for i := FColumnsToFilter.Count - 1 downto 0 do
                TObject(FColumnsToFilter.Objects[i]).Free;
            FColumnsToFilter.Clear;
        end;
    end
    else
    begin
        if FColumnsToFilter = nil then
            FColumnsToFilter := TStringList.Create;
        FColumnsToFilter.Assign(Value);
        for i := 0 to FColumnsToFilter.Count - 1 do
        begin
            SL1 := Pointer(Value.Objects[i]);
            SL2 := TStringList.Create;
            SL2.Assign(SL1);
            FColumnsToFilter.Objects[i] := SL2;
        end;
    end;
    FilterItems;
end;

function TFastListview.IndexOfColumn(const ColName: String): Integer;
var i: Integer;
begin
    for i := 0 to Columns.Count - 1 do
        if (AnsiCompareText(ColName, columns[i].Caption) = 0) or
            (AnsiCompareText(ColName, ColumnFieldNames[i]) = 0) then
        begin
            Result := i; Exit;
        end;
    Result := -1;
end;

procedure TFastListview.SetVirtualTotally(const Value: Boolean);
begin
    if FVirtualTotally = Value then Exit;
    FVirtualTotally := Value;
end;

procedure TFastListview.SetOnDataCell(const Value: TOnDataCell);
begin
    FOnDataCell := Value;
end;

procedure TFastListview.Do_SwapItems(n, m: Integer);
var T: Pointer;
begin
    if VirtualTotally then
    begin
        if Assigned(OnSwapItems) then
            OnSwapItems(Self, n, m);
    end
    else
    begin
        T := F_ItemsList.F_Items[n];
        F_ItemsList.F_Items[n] := F_ItemsList.F_Items[m];
        F_ItemsList.F_Items[m] := T;
    end;
end;

procedure TFastListview.SetOnSwapItems(const Value: TOnSwapItems);
begin
    FOnSwapItems := Value;
end;

procedure TFastListview.SetOnDataCellData(const Value: TOnDataCellData);
begin
    FOnDataCellData := Value;
end;

procedure TFastListview.SetOnDataCellImage(const Value: TOnDataCellImage);
begin
    FOnDataCellImage := Value;
end;

procedure TFastListview.SetOnDataCellNamed(const Value: TOnDataCellNamed);
begin
    FOnDataCellNamed := Value;
end;

procedure TFastListview.QuickSort2(L, R: Integer);
var
    I, J, P, cmp: Integer;
    T: Pointer;
    Tmp: Integer;
begin
    repeat
        I := L;
        J := R;
        P := (L + R) shr 1;
        repeat
            cmp := CompareFun(Self, I, P);
            while (cmp < 0) and (I < P) do begin
                Inc(I);
                cmp := CompareFun(Self, I, P);
            end;
            cmp := CompareFun(Self, J, P);
            while (cmp > 0) and (J > P) do begin
                Dec(J);
                cmp := CompareFun(Self, J, P);
            end;
            if I <= J then
            begin
                if VirtualTotally then
                begin
                    OnSwapItems(Self, I, J);
                end
                else
                begin
                    T := F_ItemsList.F_Items[I];
                    F_ItemsList.F_Items[I] := F_ItemsList.F_Items[J];
                    F_ItemsList.F_Items[J] := T;
                end;
                begin
                    Tmp := FSortLineIdexes[I];
                    FSortLineIdexes[I] := FSortLineIdexes[J];
                    FSortLineIdexes[J] := Tmp;
                end;
                if P = I then
                    P := J
                else if P = J then
                    P := I;
                Inc(I);
                Dec(J);
            end;
        until I > J;
        if L < J then QuickSort2(L, J);
        L := I;
    until I >= R;
end;

procedure TextToClipboard(const s: String);
var Clip: TClipboard;
begin
    Clip := TClipboard.Create;
    Clip.Open;
    Clip.Clear;
    Clip.SetTextBuf(PChar(s));
    Clip.Close;
    Clip.Free;
end;

procedure TFastListview.SaveToWorksheetFast(Excel, WB, WS: Variant;
    TopLine: Integer; const ImageNames, StateImageNames: String; FromIdx,
    ToIdx: Integer; SelectedOnly: Boolean; Progress: TFLV_fmProgress);
var i, i0, j, K, N, L: Integer;
    LI: TListItem;
    S: String;

    procedure DoRange(y1, x1, y2, x2: Integer; const Str: String = '';
        Align: Integer = -1; BorderBold: Boolean = TRUE);
    var R: Variant;
    begin
        y1 := y1 + TopLine - 1;
        y2 := y2 + TopLine - 1;
        if Str <> '' then WS.Cells[y1, x1] := Str;
        R := WS.Range[WS.Cells[y1, x1], WS.Cells[y2, x2]];
        TRY
            R.MergeCells := TRUE;
            if BorderBold then
            begin
                R.Borders.LineStyle := 1;
                R.Borders.Weight := 3;

                R.Borders[1].Color := clWhite;
                R.Borders[3].Color := clWhite;
            end;
            if (Align >= 0) and ShowColumnHeaders then
                R.HorizontalAlignment := Align;
            //if y2 > y1 then
            R.WrapText := TRUE;
            R.Interior.Color := clSilver;
        FINALLY
            R := Unassigned;
        END;
    end;

    function AlignOfCol(Idx: Integer): Integer;
    begin
        CASE Columns[Idx].Alignment OF
            taLeftJustify: Result := 1;
            taCenter: Result := 3;
            taRightJustify: Result := 4;
        else Result := 0;
        END;
    end;

    procedure FillNames(SL: TStringList; S: String);
    var I: Integer;
    begin
        while S <> '' do
        begin
            I := pos(',', S);
            if I > 0 then
            begin
                SL.Add(Copy(S, 1, I - 1));
                S := Copy(S, I + 1, MaxInt);
            end
            else
            begin
                SL.Add(S);
                S := '';
            end;
        end;
    end;

var HLevels, X: Integer;
    SHeader: String;
    //Line: Variant;
    HasIcons, HasStatic: Boolean;
    Img_Names, Stt_Names: TStringList;
    SCD: TSortColsData;

    function Cols(I: Integer): TListColumn;
    begin
        Result := SCD.ColsList[I];
    end;

    function ColsCount: Integer;
    begin
        Result := SCD.ColsList.Count;
    end;

    function ItemCols(LI: TListItem; I: Integer): String;
    begin
        I := Cols(I).Index;
        if I = 0 then
            Result := LI.Caption
        else
        begin
            if Assigned(OnDataCell) then
                OnDataCell(Self, LI.Index, I, Result)
            else
                if LI.SubItemsList.Count > I - 1 then
                    Result := LI.SubItemsList[I - 1]
                else
                    Result := '';
        end;
    end;

var S1: String;
    SLines: TStringList;
    FTable: TFastTable;
    MaxHeaderColumn: Integer;
    C1, C2, R: Variant;
begin
    if FromIdx < 0 then FromIdx := 0;
    if (FromIdx >= Items.Count) and not (Items.Count = 0) then Exit;
    if ToIdx > Items.Count - 1 then ToIdx := Items.Count - 1;
    if (ToIdx < FromIdx) and not ((ToIdx = -1) and (Items.Count = 0)) then Exit;

    Img_Names := TStringList.Create;
    Stt_Names := TStringList.Create;
    FillNames(Img_Names, ImageNames);
    FillNames(Stt_Names, StateImageNames);
    SCD.ColsList := TList.Create;
    SCD.ListView := Self;
    SLines := TStringList.Create;
    FTable := TFastTable.Create;
    FTable.AutoAllocateColumns := TRUE;
    TRY
        for I := 0 to VisibleColumnsCount - 1 do
            SCD.ColsList.Add(VisibleColumns[I]);
        SortData(@SCD, SCD.ColsList.Count, CompareColOrders, SwapItems);

        if (Progress <> nil) then
        begin
            Progress.Show;
        end;
        Application.ProcessMessages;

        HLevels := 1;
        if MultiHeader <> '' then
        begin
            //     
            if ShowColumnHeaders then
                for i := 0 to Columns.Count - 1 do
                begin
                    S := Columns[i].Caption;
                    j := 1;
                    while S <> '' do
                    begin
                        if pos(MultiHeader, S) > 0 then
                            S := Copy(S, pos(MultiHeader, S) + Length(MultiHeader), MaxInt)
                        else
                            break;
                        Inc(j);
                    end;
                    if j > HLevels then HLevels := j;
                end;
        end;
        // ,   
        HasIcons := FALSE;
        if Img_Names.Count > 0 then
        begin
            if SmallImages <> nil then
                for I := FromIdx to ToIdx do
                    if Items[I].ImageIndex > -1 then
                    begin
                        HasIcons := TRUE;
                        break;
                    end;
        end;
        HasStatic := FALSE;
        if Stt_Names.Count > 0 then
        begin
            if SmallImages <> nil then
                for I := FromIdx to ToIdx do
                    if Items[I].StateIndex > -1 then
                    begin
                        HasStatic := TRUE;
                        break;
                    end;
        end;
        X := 0;
        if HasIcons then Inc(X);
        if HasStatic then Inc(X);

        //   
        MaxHeaderColumn := 0;
        if ShowColumnHeaders then
            for i := 0 to ColsCount - 1 do
            begin
                S := Cols(i).Caption;
                for j := 1 to HLevels do
                begin
                    if S = '' then break;
                    if pos(MultiHeader, S) > 0 then
                    begin
                        SHeader := Copy(S, 1, pos(MultiHeader, S) - 1);
                        S := Copy(S, pos(MultiHeader, S) + Length(MultiHeader), MaxInt);
                    end
                    else
                    begin
                        SHeader := S;
                        S := '';
                    end;
                    if j = 1 then
                    begin
                        {if S = '' then
                        DoRange( 1, X + i + 1, HLevels - j + 1, X + i + 1, SHeader, 3 )
                        else}
                        DoRange(HLevels - j + 1, X + i + 1, HLevels - j + 1, X + i + 1, SHeader, 3);
                        FTable[HLevels - j, IntToStr(X + i + 1)] := SHeader;
                        if X + i + 1 > MaxHeaderColumn then
                            MaxHeaderColumn := X + i + 1;
                    end
                    else
                    begin
                        //   ,      
                        //  j:
                        N := i + 1;
                        for k := i + 1 to ColsCount do
                        begin
                            if k = ColsCount then
                            begin
                                N := k;
                                break;
                            end;
                            S1 := Cols(k).Caption;
                            //  j :
                            for L := 1 to j - 1 do
                                if pos(MultiHeader, S1) > 0 then
                                begin
                                    S1 := Copy(S1, pos(MultiHeader, S1) + Length(MultiHeader), MaxInt);
                                    if S1 = '' then
                                    begin
                                        N := k + 1; //       
                                        break;
                                    end;
                                end
                                else
                                begin
                                    S1 := '';
                                    break;
                                end;
                            if S1 <> '' then
                            begin
                                N := k; //   k      j
                                break;
                            end;
                        end;
                        {if S = '' then
                        DoRange( 1, X + N + 1, HLevels - j + 1, X + i + 1, SHeader, 3 )
                        else}
                        DoRange(HLevels - j + 1, X + i + 1, HLevels - j + 1, X + N, SHeader, 3);
                        FTable[HLevels - j, IntToStr(X + i + 1)] := SHeader;
                        if X + i + 1 > MaxHeaderColumn then
                            MaxHeaderColumn := X + i + 1;
                    end;
                end;
            end
        else
        begin
            HLevels := 0;
        end;

        ////////////////////////////////////////////////////////////////////////////
        //FTable -> SLines
        for i := 0 to FTable.Count - 1 do
        begin
            S := '';
            for j := 1 to MaxHeaderColumn do
            begin
                TRY
                    SHeader := FTable[i, IntToStr(j)];
                EXCEPT
                    SHeader := '';
                END;
                S := S + SHeader + #9;
            end;
            SLines.Add(S);
        end;

        Application.ProcessMessages;
        if Progress.Cancelled then Exit;

        if FromIdx <= ToIdx then
        begin
            //  - 
            if HLevels > 0 then
            begin
                Excel.ActiveWindow.SplitRow := HLevels;
                Excel.ActiveWindow.FreezePanes := TRUE;
            end;

            //     
            TRY
                if ShowColumnHeaders then
                    WS.Range[WS.Cells[HLevels + TopLine, X + 1],
                        WS.Cells[HLevels + ToIdx - FromIdx + TopLine,
                        X + ColsCount]].NumberFormat := '@';
            EXCEPT
                on E: EOleError do
                begin
                    ShowMessage(E.Message);
                    raise;
                end;
            END;
            if ShowColumnHeaders then
                for j := 0 to ColsCount - 1 do
                    if Cols(j).Alignment <> taLeftJustify then
                        WS.Range[WS.Cells[HLevels + TopLine, X + j + 1],
                            WS.Cells[HLevels + ToIdx - FromIdx + TopLine,
                            X + ColsCount]].HorizontalAlignment := AlignOfCol(j);
            i := -1;
            for i0 := FromIdx to ToIdx do
            begin
                if Progress <> nil then
                begin
                    Progress.ProgressBar1.Position :=
                        Progress.ProgressBar1.Position + 1;
                    Application.ProcessMessages;
                    if Progress.Cancelled then Exit;
                end;

                LI := Items[i0];
                if SelectedOnly and not LI.Selected then
                begin
                    //LI := Items[ i0 ];
                    continue;
                end;
                Inc(i);

                S := '';
                //Line := VarArrayCreate( [ 0, X + VisibleColumnsCount ],
                //     varVariant );

                if HasIcons then
                begin
                    //Line[ 0 ] := '';
                    S1 := '';
                    if LI.ImageIndex >= 0 then
                        if Img_Names.Count > LI.ImageIndex then
                            //Line[ 0 ] := Img_Names[ LI.ImageIndex ];
                            S1 := Img_Names[LI.ImageIndex];
                    S := S + S1 + #9;
                end;

                if HasStatic then
                begin
                    //Line[ X-1 ] := '';
                    S1 := '';
                    if LI.StateIndex >= 0 then
                        if Stt_Names.Count > LI.ImageIndex then
                            //Line[ X-1 ] := Stt_Names[ LI.StateIndex ];
                            S1 := Stt_Names[LI.StateIndex];
                    S := S + S1 + #9;
                end;

                for j := 0 to ColsCount - 1 do
                begin
                    if ItemCols(LI, j) <> '' then
                        //Line[ X + j ] := ItemCols( LI, j );
                        S := S + ItemCols(LI, j);
                    S := S + #9;
                end;

                {WS.Range[ WS.Cells[ i + HLevels + TopLine, 1 ],
                          WS.Cells[ i + HLevels + TopLine,
                          X + ColsCount ] ] := Line;}
                if LI.BackColor <> clNone then
                    WS.Range[WS.Cells[i + HLevels + TopLine, X + 1],
                        WS.Cells[i + HLevels + TopLine, X + ColsCount]]
                        .Interior.Color := ColorToRGB(LI.BackColor);

                if LI.FontColor <> clNone then
                    WS.Range[WS.Cells[i + HLevels + TopLine, X + 1],
                        WS.Cells[i + HLevels + TopLine, X + ColsCount]]
                        .Font.Color := ColorToRGB(LI.FontColor);

                if fsBold in LI.FontStyle then
                    WS.Range[WS.Cells[i + HLevels + TopLine, X + 1],
                        WS.Cells[i + HLevels + TopLine, X + ColsCount]]
                        .Font.Bold := TRUE;

                if fsItalic in LI.FontStyle then
                    WS.Range[WS.Cells[i + HLevels + TopLine, X + 1],
                        WS.Cells[i + HLevels + TopLine, X + ColsCount]]
                        .Font.Italic := TRUE;

                if fsUnderline in LI.FontStyle then
                    WS.Range[WS.Cells[i + HLevels + TopLine, X + 1],
                        WS.Cells[i + HLevels + TopLine, X + ColsCount]]
                        .Font.Underline := TRUE;

                if fsStrikeout in LI.FontStyle then
                    WS.Range[WS.Cells[i + HLevels + TopLine, X + 1],
                        WS.Cells[i + HLevels + TopLine, X + ColsCount]]
                        .Font.Striketrough := TRUE;

                S1 := S;
                if Assigned(OnLineToExcel) then
                    S1 := OnLineToExcel(Self, LI, WB, WS, X + 1, X + ColsCount, Progress);

                if S1 <> '' then S := S1;
                SLines.Add(S);

            end;
        end;
        if Trim(SLines.Text) <> '' then
        begin
            TextToClipboard(SLines.Text);
            C1 := WS.Cells[1, 1];
            C2 := WS.Cells[SLines.Count, ColsCount];

            R := WS.Range[C1, C2];
            OleVariant(R).PasteSpecial();
        end;


    FINALLY
        Img_Names.Free;
        Stt_Names.Free;
        SCD.ColsList.Free;
        SLines.Free;
        FTable.Free;
    END;
end;

procedure TFastListview.WMHScroll(var M: TMessage);
begin
    inherited;
    if LOWORD(M.wParam) = SB_ENDSCROLL then
        Invalidate;
end;

procedure TFastListview.SetHdrHeight(const Value: Integer);
begin
    FHdrHeight := Value;
end;

function TFastListview.GetHdrHeight: Integer;
begin
    Result := FHdrHeight;
end;

function TFastListview.GetMultiHeader: String;
begin
    Result := FMultiHeader;
end;

procedure TFastListview.WMVScroll(var M: TMessage);
begin
    inherited;
    if LOWORD(M.wParam) in [SB_ENDSCROLL, SB_PAGEDOWN, SB_PAGEUP] then
        Invalidate;
end;

procedure TFastListview.CMEnter(var Message: TCMEnter);
begin
    inherited;
    if AutoSelectOnEnter then
        if (Selected = nil) then
            if (Items.Count > 0) then
                Selected := Items[0];
    InvalidateHeader;
end;

procedure TFastListview.SetAutoSelectOnEnter(const Value: Boolean);
begin
    FAutoSelectOnEnter := Value;
end;

procedure TFastListview.SetGridLines1(const Value: Boolean);
begin
    inherited GridLines := FALSE;
    if FGridLines1 = Value then Exit;
    FGridLines1 := Value;
    Invalidate;
end;

procedure TFastListview.CMLeave(var Message: TCMExit);
begin
    inherited;
    InvalidateHeader;
end;

procedure TFastListview.SetMultiColumnSort(const Value: Boolean);
begin
    FNoMultiColumnSort := Value;
end;

procedure TFastListview.Wm_freenileditcellctl(var M: TMsg);
begin
    FreeAndNil(TControl(Pointer(M.lParam)));
end;

procedure TFastListview.SetOnEndOfSaveToExcel(
    const Value: TOnEndOfSaveToExcel);
begin
    FOnEndOfSaveToExcel := Value;
end;

function TFastListview.Get_ReadOnly: Boolean;
begin
    TRY
        Result := inherited ReadOnly;
    EXCEPT
        Result := FALSE;
        LogFileOutput('c:\exceptFLV.txt', '    ReadOnly   ' +
            Name);
    END;
end;

procedure TFastListview.Set_ReadOnly(const Value: Boolean);
const FT: array[Boolean] of String = ('FALSE', 'TRUE');
begin
    TRY
        inherited ReadOnly := Value;
    EXCEPT on E: Exception do
        begin
            MessageBox(0, PChar('    ReadOnly   ' +
                FT[Value] + '  ' + Name + ': TFastListview, ' + E.Message),
                nil, MB_OK);
        end;
    END;
end;

procedure TFastListview.Set_MultiSelect(const Value: Boolean);
begin
    if F_MultiSelect = Value then Exit;
    F_MultiSelect := Value;
    if not (csDesigning in ComponentState) then
    begin
        inherited MultiSelect := F_MultiSelect;
    end;
end;

{ TFastListItems }

function TFastListItems.Add: TFastListItem;
begin
    Result := TFastListItem.Create(Self);
    F_Items.Add(Result);
    FOwner.SetCount(Count);
end;

procedure TFastListItems.BeginUpdate;
begin
    Inc(FOwner.FUpdateCount);
    inherited;
end;

procedure TFastListItems.Clear;
var I: Integer;
    Item: TFastListItem;
    WasItems: TList;
    WasSortType: TSortType;
begin
    if Clearing then Exit;
    BeginUpdate;
    WasItems := TList.Create;
    WasSortType := FOwner.SortType;
    Clearing := TRUE;
    try
        FOwner.DeleteHiddenItems;
        FOwner.SortType := stNone;
        FOwner.ClearFilters;
        if FOwner.VirtualTotally then
        begin
            F_Items.Clear;
            FOwner.SetCount(0);
        end
        else
        begin
            for I := 0 to Count - 1 do
                WasItems.Add(F_Items[I]);
            F_Items.Clear;
            FOwner.SetCount(0);
            for I := WasItems.Count - 1 downto 0 do
            begin
                Item := WasItems[I];
                Item.Free;
            end;
        end;
    finally
        FOwner.SortType := WasSortType;
        EndUpdate;
        WasItems.Free;
        Clearing := FALSE;
    end;
end;

constructor TFastListItems.Create;
begin
    inherited;
    F_Items := TList.Create;
end;

procedure TFastListItems.Delete(Index: Integer);
var FastItem: TFastListItem;
begin
    FastItem := F_Items[Index];
    FastItem.Free;
end;

destructor TFastListItems.Destroy;
var I: Integer;
begin
    FDestroying := TRUE;
    FTmpItem1.Free;
    FTmpItem2.Free;
    for I := F_Items.Count - 1 downto 0 do
        TObject(F_Items[I]).Free;
    F_Items.Free;
    inherited;
end;

procedure TFastListItems.EndUpdate;
begin
    Dec(FOwner.FUpdateCount);
    if FOwner.FUpdateCount <= 0 then
    begin
        FOwner.SetCount(Count);
        InvalidateRect(FOwner.Handle, nil, TRUE);
        FOwner.InvalidateHeader;
        FOwner.FUpdateCount := 0;
    end;
    inherited;
end;

function TFastListItems.GetItemsCount: Integer;
begin
    if FOwner.VirtualTotally then
        Result := FCount
    else
        Result := F_Items.Count;
end;

function TFastListItems.Get_Items(Idx: Integer): TFastListItem;
var T: Pointer;
begin
    Result := nil;
    if (Idx >= 0) and (Idx < Count) then
    begin
        if FOwner.VirtualTotally then
        begin
            T := FTmpItem1;
            FTmpItem1 := FTmpItem2;
            FTmpItem2 := T;
            if FTmpItem1 = nil then
                FTmpItem1 := TFastListItem.Create(Self);
            FTmpItem1.FIndex := Idx;
            Result := FTmpItem1;
        end
        else
        begin
            Result := F_Items[Idx];
        end;
    end;
end;

function TFastListItems.Insert(Index: Integer): TFastListItem;
begin
    Result := TFastListItem.Create(Self);
    if Index >= Count then
        F_Items.Add(Result)
    else
        F_Items.Insert(Index, Result);
    FOwner.SetCount(Count);
end;

procedure TFastListItems.SetItemsCount(const Value: Integer);
begin
    if not FOwner.VirtualTotally then
    begin
        asm
      int 3
        end;
        raise Exception.Create('      ' +
            '   FastListview (VirtualTotally = TRUE)');
    end;
    TColorListView(FOwner).Items.Count := Value;
    FCount := Value;
end;

{ TFastListItem }

procedure TFastListItem.Assign(FromItem: TFastListItem;
    const AlterAssignments: String);
var I, J, K: Integer;
    Alters: TStringList;
    S, A: String;
begin
    Alters := TStringList.Create;
    TRY

        S := Trim(AlterAssignments);
        while S <> '' do
        begin
            I := pos(';', S);
            if I > 0 then
            begin
                Alters.Add(Trim(Copy(S, 1, I - 1)));
                S := Trim(Copy(S, I + 1, MaxInt));
            end
            else
            begin
                Alters.Add(S);
                S := '';
            end;
        end;

        for I := -1 to FromItem.SubItems.Count - 1 do
        begin
            S := FromItem.FOwnerFastListview.ColumnFieldNames[I + 1];
            if (S = '') and (FromItem.FOwnerFastListview.Columns.Count > I + 1) then
                S := FromItem.FOwnerFastListview.Columns[I + 1].Caption;
            if S = '' then continue;
            for J := 0 to Alters.Count - 1 do
            begin
                A := Alters[J];
                K := pos('=', A);
                if K > 0 then
                begin
                    if AnsiCompareText(Copy(A, K + 1, MaxInt), S) = 0 then
                    begin
                        S := Copy(A, 1, K - 1);
                        break;
                    end;
                end;
            end;
            Values[S] := FromItem.Values[I];
            Sitems[S] := FromItem[I];
        end;

    FINALLY
        Alters.Free;
    END;
end;

procedure TFastListItem.Assign(FromQuery: TQuery;
    const AlterAssignments: String = '');
var I, J, K: Integer;
    Alters: TStringList;
    S, A: String;
begin
    Alters := TStringList.Create;
    TRY

        S := Trim(AlterAssignments);
        while S <> '' do
        begin
            I := pos(';', S);
            if I > 0 then
            begin
                Alters.Add(Trim(Copy(S, 1, I - 1)));
                S := Trim(Copy(S, I + 1, MaxInt));
            end
            else
            begin
                Alters.Add(S);
                S := '';
            end;
        end;

        {if FromQuery.DataSource = nil then
          FieldDefs := FromQuery.FieldDefs
        else
          FieldDefs := FromQuery.DataSource.DataSet.FieldDefs;}
        for I := 0 to FromQuery.FieldCount - 1 do
        begin
            S := FromQuery.FieldNames[I];
            for J := 0 to Alters.Count - 1 do
            begin
                A := Alters[J];
                K := pos('=', A);
                if K > 0 then
                begin
                    if AnsiCompareText(Copy(A, K + 1, MaxInt), S) = 0 then
                    begin
                        S := Copy(A, 1, K - 1);
                        break;
                    end;
                end;
            end;
            {if FromQuery.NullFieldN[ I ] then
            begin
              Values[ S ] := Unassigned;
              Sitems[ S ] := '';
            end
              else}
            begin
                A := FromQuery.SFieldN[I];
                if A = '' then
                begin
                    Sitems[S] := '';
                    if FromQuery.NullFieldN[I] then
                        Values[S] := Unassigned
                    else
                        Values[S] := '';
                end
                else
                begin
                    Values[S] := A;
                    CASE FromQuery.FieldTypes[I] OF
                        ftCurrency: SItems[S] := Format('%.2f', [FromQuery.FFieldN[I]]);
                        ftDate: SItems[S] := FormatDateTime('dd.mm.yyyyy',
                                FromQuery.DFieldN[I]);
                        ftTime: SItems[S] := FormatDateTime('hh:nn:ss',
                                FromQuery.DFieldN[I]);
                        ftString: Sitems[S] := A;
                    else Sitems[S] := A;
                    END;
                end;
            end;
        end;

    FINALLY
        Alters.Free;
    END;
end;

procedure TFastListItem.Assign(FromADOQuery: TADOQuery;
    const AlterAssignments: String);
var I, J, K: Integer;
    Alters: TStringList;
    S, A: String;
begin
    Alters := TStringList.Create;
    TRY

        S := Trim(AlterAssignments);
        while S <> '' do
        begin
            I := pos(';', S);
            if I > 0 then
            begin
                Alters.Add(Trim(Copy(S, 1, I - 1)));
                S := Trim(Copy(S, I + 1, MaxInt));
            end
            else
            begin
                Alters.Add(S);
                S := '';
            end;
        end;

        for I := 0 to FromADOQuery.FieldCount - 1 do
        begin
            S := FromADOQuery.FieldDefs[I].Name;
            for J := 0 to Alters.Count - 1 do
            begin
                A := Alters[J];
                K := pos('=', A);
                if K > 0 then
                begin
                    if AnsiCompareText(Copy(A, K + 1, MaxInt), S) = 0 then
                    begin
                        S := Copy(A, 1, K - 1);
                        break;
                    end;
                end;
            end;
            if FromADOQuery.Fields[I].IsNull then
            begin
                Values[S] := Unassigned;
                Sitems[S] := '';
            end
            else
            begin
                Values[S] := FromADOQuery.Fields[I].AsString;
                CASE FromADOQuery.FieldDefs[I].DataType OF
                    ftCurrency: SItems[S] := Format('%.2f', [FromADOQuery.Fields[I].AsFloat]);
                    ftDate,
                        ftDateTime: SItems[S] := FormatDateTime('dd.mm.yyyyy',
                            FromADOQuery.Fields[I].AsDateTime);
                    ftTime: SItems[S] := FormatDateTime('hh:nn:ss',
                            FromADOQuery.Fields[I].AsDateTime);
                else Sitems[S] := FromADOQuery.Fields[I].AsString;
                END;
            end;
        end;

    FINALLY
        Alters.Free;
    END;
end;

constructor TFastListItem.Create(AOwner: TFastListItems);
begin
    inherited Create; //( nil );
    FOwnerItemsHolder := AOwner;
    FOwnerFastListview := AOwner.FOwner;
    FSubItems := TFastListSubItems.Create(Self);
    fBackColor := clNone;
    fBackColorSelected := clNone;
    fVisible := TRUE;
    FFontColor := clNone;
    FFontColorSelected := clNone;
end;

destructor TFastListItem.Destroy;
var I: Integer;
    pValue: PVariant;
begin
    if fVisible then
    begin
        if Assigned(FOwnerItemsHolder.FOwner.OnDeletion) then
            FOwnerItemsHolder.FOwner.OnDeletion(FOwnerItemsHolder.FOwner, Self);
        if not FOwnerFastListview.VirtualTotally then
            if not FOwnerItemsHolder.FDestroying and not FOwnerItemsHolder.Clearing then
                Selected := FALSE;
        FSubItems.Free;
        if (Index >= 0) and (Index < FOwnerItemsHolder.F_Items.Count) then
            FOwnerItemsHolder.F_Items.Delete(Index);
        if not FOwnerItemsHolder.FDestroying then
            if FOwnerItemsHolder.FOwner.FUpdateCount <= 0 then
                FOwnerItemsHolder.FOwner.SetCount(FOwnerItemsHolder.Count);
    end
    else
    begin
        if Assigned(FOwnerFastListview) then
            FOwnerFastListview.FHiddenItems.Remove(Self);
    end;
    if Assigned(FValues) then
    begin
        for I := 0 to FValues.Count - 1 do
        begin
            pValue := FValues[I];
            if pValue <> nil then
            begin
                pValue^ := Unassigned;
                FreeMem(pValue);
            end;
        end;
        FValues.Free;
    end;
    inherited;
end;

procedure TFastListItem.DoEndEditCell(Sender: TObject);
begin
    EndEditCell;
end;

function TFastListItem.EditCaption: Boolean;
begin
    if FListItem <> nil then
        Result := FListItem.EditCaption
    else
    begin
        Result :=
            FOwnerFastListview.Perform(LVM_EDITLABEL, Index, 0) <> 0;
    end;
end;

function TFastListItem.EditCell(Idx: Integer): Boolean;
var R: TRect;
    L: Integer;
begin
    if not Assigned(FOwnerFastListview.FEditCellCtl) then
    begin
        FOwnerFastListview.FEditCellCtl := TUniEdit.Create(FOwnerFastListview);
        FOwnerFastListview.FEditCellCtl.Visible := FALSE;
        FOwnerFastListview.FEditCellCtl.Parent := FOwnerFastListview;
        (FOwnerFastListview.FEditCellCtl as TUniEdit).BorderWidth := 0;
        (FOwnerFastListview.FEditCellCtl as TUniEdit).TitleWidth := 0;
    end;
    FEditingColIdx := Idx;
    MakeVisible(FALSE);
    while TRUE do
    begin
        R := GetRect(Idx);
        if R.Left > FOwnerFastListview.ClientWidth then
        begin
            L := R.Left;
            ScrollWindow(FOwnerFastListview.Handle, -10, 0, nil, nil);
            R := GetRect(Idx);
            if R.Left = L then break;
        end
        else
            if R.Right < 0 then
            begin
                L := R.Left;
                ScrollWindow(FOwnerFastListview.Handle, 10, 0, nil, nil);
                R := GetRect(Idx);
                if R.Left = L then break;
            end
            else
                break;
    end;
    FOwnerFastListview.FEditCellCtl.SetBounds(R.Left, R.Top, R.Right - R.Left, R.Bottom - R.Top);
    (FOwnerFastListview.FEditCellCtl as TUniEdit).Kind :=
        FOwnerFastListview.ColumnEditKind[Idx];
    (FOwnerFastListview.FEditCellCtl as TUniEdit).Text := SItems[Idx - 1];
    if ((FOwnerFastListview.FEditCellCtl as TUniEdit).Kind = ekCombo) then
        (FOwnerFastListview.FEditCellCtl as TUniEdit).Items.Assign(
            FOwnerFastListview.ColumnComboItems[Idx]);
    if Assigned(FOwnerFastListview.OnCreateEditCell) then
        FOwnerFastListview.OnCreateEditCell(FOwnerFastListview, Self, Idx,
            FOwnerFastListview.FEditCellCtl);
    FOwnerFastListview.FEditingItem := Self;
    FOwnerFastListview.FEditCellCtl.Visible := TRUE;
    Windows.SetFocus((FOwnerFastListview.FEditCellCtl as TUniEdit).Control.Handle);
    (FOwnerFastListview.FEditCellCtl as TUniEdit).OnLeave := DoEndEditCell;
    (FOwnerFastListview.FEditCellCtl as TUniEdit).OnKeyPress := FOwnerFastListview.EditCellKeyPress;
    Result := TRUE;
end;

procedure TFastListItem.EndEdit;
var EWnd: HWnd;
    Txt: String;
    L: Integer;
begin
    if not FOwnerFastListview.IsEditing then Exit;
    EWnd := FOwnerFastListview.Perform(LVM_GETEDITCONTROL, 0, 0);
    if EWnd = 0 then Exit;
    L := GetWindowTextLength(EWnd);
    SetLength(Txt, L);
    GetWindowText(EWnd, @Txt[1], L + 1);
    MirrorItem.CancelEdit;
    Caption := Txt;
    FOwnerFastListview.Do_OnEdited(FOwnerFastListview, MirrorItem, Txt);
    Caption := Txt;
end;

procedure TFastListItem.EndEditCell(Cancel: Boolean = FALSE);
var S: String;
begin
    TRY
        if not FOwnerFastListview.EditingCell then Exit;
        if not Assigned(FOwnerFastListview.FEditCellCtl) then Exit;
        if not Cancel then
        begin
            S := (FOwnerFastListview.FEditCellCtl as TUniEdit).Text;
            if Assigned(FOwnerFastListview.OnEndEditCell) then
                FOwnerFastListview.OnEndEditCell(FOwnerFastListview,
                    Self, FEditingColIdx, FOwnerFastListview.FEditCellCtl, S);
            try
                SItems[FEditingColIdx - 1] := S;
            except
                //
            end;
        end;
        FOwnerFastListview.FEditCellCtl.Visible := FALSE;
        FOwnerFastListview.FEditCellCtl.Parent := nil;
        PostMessage(FOwnerFastListview.Handle, WM_FREE_NIL_EDITCELLCTL, 0,
            Integer(Pointer(FOwnerFastListview.FEditCellCtl)));
        FOwnerFastListview.FEditCellCtl := nil;
        //FreeAndNil( FOwnerFastListview.FEditCellCtl );
        Windows.SetFocus(FOwnerFastListview.Handle);
    EXCEPT

    END;
end;

function TFastListItem.GetCaption: String;
begin
    if FOwnerFastListview.VirtualTotally then
    begin
        if Assigned(FOwnerFastListview.OnDataCell) then
            FOwnerFastListview.OnDataCell(FOwnerFastListview, Index, 0, Result);
    end
    else
        Result := FCaption;
end;

function TFastListItem.GetCut: Boolean;
var Mask: Integer;
begin
    Mask := LVIS_CUT;
    Result := ListView_GetItemState(FOwnerItemsHolder.FOwner.Handle, Self.Index, Mask) and Mask <> 0;
end;

function TFastListItem.GetData: Pointer;
begin
    if FOwnerFastListview.VirtualTotally then
    begin
        if Assigned(FOwnerFastListview.OnDataCellData) then
            FOwnerFastListview.OnDataCellData(FOwnerFastListview, Index, FData);
    end;
    Result := FData;
end;

function TFastListItem.GetDropTarget: Boolean;
var Mask: Integer;
begin
    Mask := LVIS_DROPHILITED;
    Result := ListView_GetItemState(FOwnerItemsHolder.FOwner.Handle, Self.Index, Mask) and Mask <> 0;
end;

function TFastListItem.GetFontStyle: TFontStyles;
begin
    if FCustomFontStyle then
        Result := FFontStyle
    else
        Result := FOwnerItemsHolder.FOwner.Font.Style;
end;

function TFastListItem.GetImageIndex: Integer;
begin
    if FOwnerFastListview.VirtualTotally then
    begin
        if Assigned(FOwnerFastListview.OnDataCellImage) then
            FOwnerFastListview.OnDataCellImage(FOwnerFastListview, Index, FImageIndex);
    end;
    Result := FImageIndex;
end;

function TFastListItem.GetIndent: Integer;
begin
    Result := FIndent;
end;

function TFastListItem.GetIndexInListView: Integer;
begin
    if FOwnerFastListview.VirtualTotally then
        Result := FIndex
    else
        if Assigned(FListItem) then
            Result := FListItem.Index
        else
            Result := FOwnerItemsHolder.F_Items.IndexOf(Self);
end;

function TFastListItem.GetIsChecked: Boolean;
var LV: TOldListItem;
begin
    Result := FALSE;
    LV := MirrorItem;
    if LV <> nil then
        Result := LV.Checked;
end;

function TFastListItem.GetOverlayIndex: Integer;
begin
    Result := FOverlayIndex;
end;

function TFastListItem.GetRect(SubItemIndex: Integer = -1): TRect;
var Hdr: HWnd;
    R, WR: TRect;
begin
    Result.Left := 0;
    FOwnerItemsHolder.FOwner.Perform(LVM_GETITEMRECT, Index, Integer(@Result));
    if SubItemIndex = -1 then Exit;
    if FOwnerItemsHolder.FOwner.ViewStyle = vsReport then
    begin
        Hdr := FOwnerFastListview.GetHeaderCtlWindow;
        if Hdr = 0 then Exit;
        SendMessage(Hdr, HDM_GETITEMRECT, SubItemIndex, Integer(@R));
        WR.TopLeft := Point(0, 0);
        ClientToScreen(Hdr, WR.TopLeft);
        WR.BottomRight := Point(0, 0);
        ClientToScreen(FOwnerItemsHolder.FOwner.Handle, WR.BottomRight);
        OffsetRect(WR, -WR.Right, -WR.Bottom); // WR.TopLeft =   Hdr . FastLV
        OffsetRect(R, WR.Left, WR.Top); // R =   SubItems[ SubItemIndex ] (Left, Right) . FastLV
        Result.Left := R.Left;
        Result.Right := R.Right;
    end
    else
    begin
        if SubItemIndex > 0 then Exit;
        if FListItem <> nil then
            Result := FListItem.DisplayRect(drLabel)
        else
            //Result := SendMessage( )
    end;
end;

function TFastListItem.GetSelected: Boolean;
begin
    //if FOwnerFastListview.VirtualTotally then
    begin
        Result := FOwnerFastListview.Perform(LVM_GETITEMSTATE, Index, LVIS_SELECTED)
            <> 0;
    end
    {else
    Result := FSelected};
end;

function TFastListItem.GetStateIndex: Integer;
begin
    Result := 0;
    if Self <> nil then
        Result := FStateIndex;
end;

function TFastListItem.GetSubItems(IdxOrName: Variant): String;
var I: Integer;
begin
    if VarType(IdxOrName) = varString then
    begin
        if FOwnerFastListview.VirtualTotally then
        begin
            if Assigned(FOwnerFastListview.OnDataCellNamed) then
            begin
                FOwnerFastListview.OnDataCellNamed(FOwnerFastListview, Index, IdxOrName,
                    Result);
                Exit;
            end;
        end;
        I := FOwnerFastListview.ColumnIndexByName(IdxOrName);
        if I < 0 then
            raise Exception.Create('Column name (' + IdxOrName + ') not found ' +
                'while obtaining a value, component: ' + FOwnerFastListview.Name +
                ', item# ' + IntToStr(Index));
    end
    else
        I := IdxOrName + 1;
    if FOwnerFastListview.VirtualTotally then
    begin
        if Assigned(FOwnerFastListview.OnDataCell) then
        begin
            FOwnerFastListview.OnDataCell(FOwnerFastListview, Index, I, Result);
            Exit;
        end;
    end;
    if I > 0 then
    begin
        if SubItemsList.Count > I - 1 then
            Result := SubItemsList[I - 1]
        else
            Result := '';
    end
    else
        if I = 0 then
            Result := Caption
        else
            Result := '';
end;

function TFastListItem.GetSubitemsList: TStrings;
begin
    if (FOwnerItemsHolder <> nil) and
        not FOwnerItemsHolder.AccessingOnData then
    begin
        if assigned(FOwnerFastListview.OnData) then
        begin
            FOwnerItemsHolder.AccessingOnData := TRUE;
            TRY
                FOwnerFastListview.OnData(FOwnerFastListview, Self);
            FINALLY
                FOwnerItemsHolder.AccessingOnData := FALSE;
            END;
        end;
    end;
    Result := FSubItems;
end;

function TFastListItem.GetValues(IdxOrName: Variant): Variant;
var I: Integer;
    pValue: PVariant;
begin
    if VarType(IdxOrName) = varString then
        I := FOwnerFastListview.ColumnIndexByName(IdxOrName)
    else
        I := IdxOrName + 1;
    if (I >= 0) and Assigned(FValues) and (I < FValues.Count) then
    begin
        pValue := FValues[I];
        if (pValue <> nil) and {(pValue^ <> Unassigned)} not VarIsEmpty(pValue^) then
        begin
            Result := pValue^;
            Exit;
        end;
    end;
    Result := SItems[IdxOrName];
end;

function TFastListItem.Get_Focused: Boolean;
var LV: TOldListItem;
begin
    Result := FALSE;
    LV := MirrorItem;
    if LV <> nil then
        Result := LV.Focused;
end;

function TFastListItem.ListView: TFastListView;
begin
    Result := FOwnerFastListview;
end;

procedure TFastListItem.MakeVisible(PartialOK: Boolean);
var LV: TOldListItem;
begin
    LV := MirrorItem;
    if LV <> nil then
        LV.MakeVisible(PartialOK);
end;

function TFastListItem.MirrorItem: TOldListItem;
begin
    Result := FOwnerItemsHolder.FOwner.InheritedItems[Index];
end;

procedure TFastListItem.SetBackColor(const Value: TColor);
begin
    fBackColor := Value;
    FOwnerItemsHolder.FOwner.Invalidate;
end;

procedure TFastListItem.SetBackColorSelected(const Value: TColor);
begin
    fBackColorSelected := Value;
    FOwnerItemsHolder.FOwner.Invalidate;
end;

procedure TFastListItem.SetCaption(const Value: String);
begin
    if FListItem <> nil then
        FListItem.Caption := Value;
    FCaption := Value;
    FOwnerItemsHolder.FOwner.Invalidate;
end;

procedure TFastListItem.SetCustomFontStyle(const Value: Boolean);
begin
    if FCustomFontStyle = Value then Exit;
    FCustomFontStyle := Value;
    if FOwnerItemsHolder.FOwner.FUpdateCount = 0 then
        FOwnerItemsHolder.FOwner.Invalidate;
end;

procedure TFastListItem.SetCut(const Value: Boolean);
var
    Mask: Integer;
    Data: Integer;
begin
    if FOwnerItemsHolder.FOwner.Providecount(FOwnerItemsHolder.Count) then
        FOwnerItemsHolder.FOwner.Invalidate;
    Mask := LVIS_CUT;
    if Value then Data := Mask
    else Data := 0;
    ListView_SetItemState(FOwnerItemsHolder.FOwner.Handle, Self.Index, Data, Mask);
end;

procedure TFastListItem.SetDropTarget(const Value: Boolean);
var
    Mask: Integer;
    Data: Integer;
begin
    if FOwnerItemsHolder.FOwner.Providecount(FOwnerItemsHolder.Count) then
        FOwnerItemsHolder.FOwner.Invalidate;
    Mask := LVIS_DROPHILITED;
    if Value then Data := Mask
    else Data := 0;
    ListView_SetItemState(FOwnerItemsHolder.FOwner.Handle, Self.Index, Data, Mask);
end;

procedure TFastListItem.SetFontColor(const Value: TColor);
begin
    FFontColor := Value;
    FOwnerItemsHolder.FOwner.Invalidate;
end;

procedure TFastListItem.SetFontColorSelected(const Value: TColor);
begin
    FFontColorSelected := Value;
    FOwnerItemsHolder.FOwner.Invalidate;
end;

procedure TFastListItem.SetFontStyle(const Value: TFontStyles);
begin
    FCustomFontStyle := TRUE;
    FFontStyle := Value;
    if FOwnerItemsHolder.FOwner.FUpdateCount = 0 then
        FOwnerItemsHolder.FOwner.Invalidate;
end;

procedure TFastListItem.SetImageIndex(const Value: Integer);
begin
    FImageIndex := Value;
    if FOwnerItemsHolder.FOwner.FUpdateCount = 0 then
        FOwnerItemsHolder.FOwner.Invalidate;
end;

procedure TFastListItem.SetIndent(const Value: Integer);
var MI: TOldListItem;
begin
    if FIndent = Value then Exit;
    FIndent := Value;
    MI := MirrorItem;
    if MI <> nil then
        MI.Indent := Value;
end;

procedure TFastListItem.SetIsChecked(const Value: Boolean);
var LV: TOldListItem;
begin
    LV := MirrorItem;
    if LV <> nil then
        LV.Checked := Value;
end;

procedure TFastListItem.SetOverlayIndex(const Value: Integer);
begin
    FOverlayIndex := Value;
    if FOwnerItemsHolder.FOwner.FUpdateCount = 0 then
        FOwnerItemsHolder.FOwner.Invalidate;
end;

procedure TFastListItem.Setselected(const Value: Boolean);
var
    Mask: Integer;
    Data: Integer;
begin
    if FOwnerItemsHolder.FOwner.Providecount(FOwnerItemsHolder.Count) then
        FOwnerItemsHolder.FOwner.Invalidate;
    Mask := LVIS_SELECTED;
    if Value then Data := Mask
    else Data := 0;
    ListView_SetItemState(FOwnerItemsHolder.FOwner.Handle, Self.Index, Data, Mask);
    FSelected := Value;
end;

procedure TFastListItem.SetStateIndex(const Value: Integer);
begin
    FStateIndex := Value;
    if FOwnerItemsHolder.FOwner.FUpdateCount = 0 then
        FOwnerItemsHolder.FOwner.Invalidate;
end;

procedure TFastListItem.SetSubItems(IdxOrName: Variant;
    const Value: String);
var I: Integer;
begin
    if FOwnerFastListview.VirtualTotally then
    begin
        asm
      int 3
        end;
        raise Exception.Create('    TFastListView  ' +
            ' VirtualTotally');
    end;
    if VarType(IdxOrName) = varString then
    begin
        I := FOwnerFastListview.ColumnIndexByName(IdxOrName);
        if I < 0 then
            if FOwnerFastListview.AutoAllocateColumnNames then
            begin
                if FOwnerFastListview.FColumnNames = nil then
                    FOwnerFastListview.FColumnNames := TStringList.Create;
                while FOwnerFastListview.FColumnNames.Count < FOwnerFastListview.Columns.Count do
                    FOwnerFastListview.FColumnNames.Add('');
                FOwnerFastListview.FColumnNames.Add(IdxOrName);
                I := FOwnerFastListview.FColumnNames.Count - 1;
            end
            else
                raise Exception.Create('Column name (' + IdxOrName + ') not found ' +
                    'while assigning a value, component: ' + FOwnerFastListview.Name +
                    ', item# ' + IntToStr(Index));
    end
    else
        I := IdxOrName + 1;
    if I > 0 then
        SubItemsList[I - 1] := Value
    else
        if I = 0 then
            Caption := Value
        else
            raise Exception.Create('Column index (' + IntToStr(I) +
                ') incorrect while assigning a value, component: ' + FOwnerFastListview.Name +
                ', item #' + IntToStr(Index));
end;

procedure TFastListItem.SetSubitemsList(const Value: TStrings);
begin
    FSubItems.Assign(Value);
end;

procedure TFastListItem.SetValues(IdxOrName: Variant;
    const Value: Variant);
var I: Integer;
    pValue: PVariant;
    S: String;
begin
    if FOwnerFastListview.VirtualTotally then
    begin
        asm
      int 3
        end;
        raise Exception.Create('    TFastListView  ' +
            ' VirtualTotally');
    end;
    if VarType(IdxOrName) = varString then
        I := FOwnerFastListview.ColumnIndexByName(IdxOrName)
    else
        I := IdxOrName + 1;
    if I >= 0 then
    begin
        if not Assigned(FValues) then
            FValues := TList.Create;
        while I >= FValues.Count do
        begin
            GetMem(pValue, sizeof(Variant));
            FillChar(pValue^, sizeof(Variant), 0);
            FValues.Add(pValue);
        end;
        pValue := FValues[I];
        pValue^ := Value;
        // ,   
        CASE VarType(pValue^) OF
            varInteger, varSmallInt, varByte: S := IntToStr(Value);
            varSingle, varDouble, varCurrency: S := FloatToStr(Value);
            varString, varOleStr: S := Value;
            varDate: S := FormatDateTime('dd.mm.yyyy', Value);
        else S := '';
        END;
        SItems[IdxOrName] := S;
    end;
end;

procedure TFastListItem.SetVisible(const Value: Boolean);
begin
    if fVisible = Value then Exit;
    fVisible := Value;
    if not Value then
    begin
        FSelected := FALSE;
        FOwnerItemsHolder.F_Items.Remove(Self);
        FOwnerFastListview.FHiddenItems.Add(Self);
    end
    else
    begin
        FOwnerItemsHolder.F_Items.Add(Self);
        FOwnerFastListview.FHiddenItems.Remove(Self);
    end;
    FOwnerFastListview.Providecount(FOwnerFastListview.Items.Count);
end;

procedure TFastListItem.Set_Focused(const Value: Boolean);
var LV: TOldListItem;
begin
    LV := MirrorItem;
    if LV <> nil then
        LV.Focused := Value;
end;

{ TFastListSubItems }

function TFastListSubItems.Add(const S: string): Integer;
begin
    Result := inherited Add(S);
    UpdateOwner;
end;

procedure TFastListSubItems.Clear;
begin
    inherited;
    UpdateOwner;
end;

constructor TFastListSubItems.Create(AOwner: TFastListItem);
begin
    inherited Create;
    FOwnerItem := AOwner;
end;

procedure TFastListSubItems.Delete(Index: Integer);
begin
    inherited;
    UpdateOwner;
end;

function TFastListSubItems.GetByIndexOrName(IndexOrName: Variant): string;
begin
    Result := FOwnerItem.SItems[IndexOrName];
end;

procedure TFastListSubItems.Insert(Index: Integer; const S: string);
begin
    inherited;
    UpdateOwner;
end;

procedure TFastListSubItems.Put(Index: Integer; const S: string);
begin
    while Index >= Count do
        Add('');
    if Index = -1 then
        FOwnerItem.Caption := S
    else
        inherited;
    UpdateOwner;
end;

procedure TFastListSubItems.PutByIndexOrName(IndexOrName: Variant;
    const Value: string);
begin
    FOwnerItem.SItems[IndexOrName] := Value;
end;

procedure TFastListSubItems.UpdateOwner;
var I: Integer;
begin
    if (FOwnerItem <> nil) and
        (FOwnerItem.FListItem <> nil) and
        (FOwnerItem.FListItem.SubItems <> nil) then
    begin
        for I := 0 to Count - 1 do
        begin
            if FOwnerItem.FListItem.SubItems.Count <= I then
            TRY
                FOwnerItem.FListItem.SubItems.Add(Strings[I])
            EXCEPT on E: Exception do
                begin
                    raise FLV_Exception('     ''' +
                        Strings[I] + '''   ' + IntToStr(I) + ': ' +
                        E.Message);
                end;
            END
            else
            TRY
                FOwnerItem.FListItem.SubItems[I] := Strings[I];
            EXCEPT on E: Exception do
                begin
                    raise FLV_Exception('     ''' +
                        Strings[I] + '''   ' + IntToStr(I) + ': ' +
                        E.Message);
                end;
            END;
        end;
    end;
    if (FOwnerItem <> nil) and (FOwnerItem.FOwnerItemsHolder <> nil) and
        (FOwnerItem.FOwnerItemsHolder.FOwner <> nil) and
        (FOwnerItem.FOwnerItemsHolder.FOwner.FUpdateCount = 0) then
        FOwnerItem.FOwnerItemsHolder.FOwner.Invalidate;
end;

procedure TFastListview.SetOnAdvancedCustomDrawSubItem(
    const Value: TLVAdvancedCustomDrawsubItem);
begin
    FOnAdvancedCustomDrawSubItem := Value;
    inherited OnAdvancedCustomDrawSubItem := DoAdvancedDrawSubItem;
end;

procedure TFastListview.SetOnCustomDrawSubItem(
    const Value: TLVOnCustomDrawSubItem);
begin
    FOnCustomDrawSubItem := Value;
    inherited OnCustomDrawSubItem := DocustomDrawSubItem;
end;

procedure TFastListview.SetSelected(const Value: TListItem);
var I: Integer;
begin
    if Value <> nil then
        FSelectedItemIdx := Value.Index
    else
        FSelectedItemIdx := -1;
    if Providecount(Items.Count) then
        Invalidate;
    if Value = nil then
    begin
        inherited Selected := nil;
        for I := 0 to Count - 1 do
            Items[I].FSelected := FALSE;
    end
    else
    begin
        Assert((Value.Index >= 0) and (Value.Index < Items.Count),
            Err0006 + ' ( ' +
            IntToStr(Value.Index) + ' ): ' + Value.Caption);
        if (Value <> nil) and (MultiSelect) then
        begin
            for I := 0 to Count - 1 do
                Items[I].FSelected := FALSE;
        end;
        inherited Selected := inherited Items[Value.Index];
        if Selected <> nil then
            Selected.FSelected := TRUE;
    end;
end;

procedure TFastListview.SetSortByCase(const Value: Boolean);
begin
    FSortByCase := Value;
end;

procedure TFastListview.SetSortByDate(const Value: Boolean);
begin
    FSortByDate := Value;
end;

procedure TFastListview.SetSortByIcons(const Value: TSortByIcon);
begin
    FSortByIcons := Value;
    Invalidate;
end;

procedure TFastListview.SetSortByNumber(const Value: Boolean);
begin
    FSortByNumber := Value;
end;

procedure TFastListview.SetSortFrom(const Value: Integer);
begin
    FSortFrom := Value;
end;

procedure TFastListview.SetSortTo(const Value: Integer);
begin
    FSortTo := Value;
    if FSortTo < 0 then
        FSortFrom := 0;
end;

procedure TFastListview.Set_OnDraw(const Value: TFastLVDrawItemEvent);
begin
    fOnDraw := Value;
    inherited OnDrawItem := DoDrawItem;
end;

procedure TFastListview.SortColumns(const ColumnsArray: array of Integer;
    OrderByFollowsColumns: Boolean = TRUE);
var I: Integer;
begin
    FColumnsToSort.Clear;
    for I := 0 to High(ColumnsArray) do
        FColumnsToSort.Add(Pointer(ColumnsArray[I]));
    FOrderByFollowsColumns := OrderByFollowsColumns;
    AlphaSort1(FALSE);
end;

{ TFilterStrings }

destructor TFilterStrings.Destroy;
var i: Integer;
begin
    for i := 0 to Count - 1 do
        TObject(Objects[i]).Free;
    inherited;
end;

initialization
    InitUpper;

end.
